__all__ = ['Blob', 'AbstractBlobDM']


class Blob (object) :

    __slots__ = ('bdm', 'oid', 'handler')

    def __init__ (self, *a, **kw) :
        raise TypeError('Use your blob data manager for creating new instances')

    def __repr__ (self) :
        return 'Blob(oid=%s)' % self.oid

    def open (self, mode='r') :
        if self.handler is not None :
            raise RuntimeError(
                '%r is already opened (%d)' % (self, self.handler)
            )
        self.handler = self.bdm.open(self, mode)

    def close (self) :
        if self.handler is None :
            raise RuntimeError('%r is not opened' % self)
        self.bdm.close(self.handler)
        self.handler = None

    def read (self, size=None) :
        if self.handler is None :
            raise RuntimeError('%r is not opened' % self)
        return self.bdm.read(self.handler, size)

    def write (self, data) :
        if self.handler is None :
            raise RuntimeError('%r is not opened' % self)
        return self.bdm.write(self.handler, data)

    def seek (self, off, whence=0) :
        if self.handler is None :
            raise RuntimeError('%r is not opened' % self)
        return self.bdm.seek(self.handler, off, whence)

    def tell (self) :
        if self.handler is None :
            raise RuntimeError('%r is not opened' % self)
        return self.bdm.tell(self.handler)

    def size (self) :
        if self.handler is None :
            raise RuntimeError('%r is not opened' % self)
        return self.bdm.size(self.handler)

    def __iter__(self):
        block = 1024 * 128
        while True:
            d = self.read(block)
            if not d:
                break
            yield d


class AbstractBlobDM :

    def new (self) :
        oid = self.newOID()
        b = object.__new__(Blob)
        b.oid = oid
        b.bdm = self
        b.handler = None
        return b

    def _preload (self, oid) :
        b = object.__new__(Blob)
        b.oid = oid
        b.bdm = self
        b.handler = None
        return b

    def __getitem__ (self, oid) :
        oid = self.find(oid)
        return self._preload(oid)

    def __delitem__ (self, oid) :
        if isinstance(oid, Blob) :
            oid = oid.oid
        return self.delete(oid)

    def newOID (self) :
        raise NotImplementedError

    def find (self, oid) :
        raise NotImplementedError

    def delete (self, oid) :
        raise NotImplementedError

    def open (self, b, mode='r') :
        raise NotImplementedError

    def close (self, h) :
        raise NotImplementedError

    def read (self, h, size=None) :
        raise NotImplementedError

    def write (self, h, data) :
        raise NotImplementedError

    def seek (self, h, off, whence=0) :
        raise NotImplementedError

    def tell (self, h) :
        raise NotImplementedError

    def size (self, h) :
        pos = self.tell(h)
        self.seek(0, 2)
        size = self.tell(h)
        self.seek(pos, 0)
        return size
