# http://pypi.python.org/pypi/bda.intellidatetime/
# http://pypi.python.org/pypi/magicdate/
# http://www.w3.org/TR/2005/NOTE-timezone-20051013/
# http://projects.amor.org/misc/browser/recur.py?format=txt
# http://code-bear.com/code/parsedatetime/
# http://cheeseshop.python.org/pypi/when/
# Actions and Events in Interval Temporal Logic (1994)
# http://citeseer.ist.psu.edu/allen94actions.html

__all__ = [
    'DatetimeNoTimeZone', 'Datetime', 'Date', 'Month', 'Year', 'TimeService',
    'DAY', 'TZLOCAL', 'TZUTC',
]


from datetime import date, datetime, timedelta, time, MINYEAR, MAXYEAR
from dateutil import parser, tz


YEAR_DAYS = (
    (0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365),
    (0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335, 366),
)

MONTH_DAYS = (
    (0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31),
    (0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31),
)

TZLOCAL = tz.tzlocal()
TZUTC   = tz.tzutc()


def leap_year(year):
    return year % 4 == 0 and (year % 100 != 0 or year % 400 == 0)









class BaseDatetime(datetime):

    @classmethod
    def parse(klass, val):
        if isinstance(val, datetime):
            args = (
                val.year, val.month, val.day,
                val.hour, val.minute, val.second,
                val.microsecond, val.tzinfo
            )
        elif isinstance(val, date):
            args = val.year, val.month, val.day
        elif isinstance(val, basestring):
            val = parser.parse(val, dayfirst=val.count('.')>1)
            args = (
                val.year, val.month, val.day,
                val.hour, val.minute, val.second,
                val.microsecond, val.tzinfo
            )
        else:
            raise ValueError("invalid argument %r" % d)
        return klass(*args)


class DatetimeNoTimeZone(BaseDatetime):

    def __new__(klass, *args, **kwds):
        self = BaseDatetime.__new__(klass, *args, **kwds)
        if self.tzinfo is not None:
            self = klass.parse(self.replace(tzinfo=None))
        return self

DatetimeNoTimeZone.min = DatetimeNoTimeZone.parse(
                            datetime.min.replace(tzinfo=None))
DatetimeNoTimeZone.max = DatetimeNoTimeZone.parse(
                            datetime.max.replace(tzinfo=None))

class Datetime(BaseDatetime):

    def __new__(klass, *args, **kwds):
        self = BaseDatetime.__new__(klass, *args, **kwds)
        if self.tzinfo is None:
            self = klass.parse(self.replace(tzinfo=TZLOCAL))
        return self

Datetime.min = Datetime.parse(datetime.min.replace(tzinfo=TZUTC))
Datetime.max = Datetime.parse(datetime.max.replace(tzinfo=TZUTC))













class Date(date):

    @classmethod
    def parse(klass, val):
        if isinstance(val, (date, datetime)):
            year = val.year
            month = val.month
            day = val.day
        elif isinstance(val, basestring):
            val = parser.parse(str(val), dayfirst=val.count('.')>1)
            year = val.year
            month = val.month
            day = val.day
        else:
            raise ValueError("invalid argument %r" % d)
        return klass(year, month, day)

    # XXX +/-

    @property
    def first_in_month(self):
        return self.__class__(self.year, self.month, 1)

    @property
    def last_in_month(self):
        year = self.year
        month = self.month
        return self.__class__(year, month, MONTH_DAYS[leap_year(year)][month])

    @classmethod
    def today(klass):
        return klass.parse(date.today())


Date.min = Date.parse(date.min)
Date.max = Date.parse(date.max)
DAY = Date.resolution









class Month(object):

    def __init__(self, year, month=None):
        if month is None and isinstance(year, (date, datetime)):
            self.month = year.month
            self.year = year.year
        else:
            self.year = int(year)
            self.month = int(month)
            if self.year < MINYEAR or self.year > MAXYEAR:
                raise ValueError('year must be in range', MINYEAR, MAXYEAR)
            if self.month < 1 or self.month > 12:
                raise ValueError('month must be in range 1-12', self.month)

    @property
    def first(self):
        return Date(self.year, self.month, 1)

    @property
    def last(self):
        year = self.year
        month = self.month
        return Date(year, month, MONTH_DAYS[leap_year(year)][month])

    def __len__(self):
        return MONTH_DAYS[leap_year(self.year)][self.month]

    def __iter__(self):
        y = self.year
        m = self.month
        for d in xrange(1, len(self)+1):
            yield Date(y, m, d)

    def __getitem__(self, day):
        if isinstance(day, int):
            if day < 1:
                raise KeyError(day)
            try:
                return Date(self.year, self.month, day)
            except:
                raise KeyError(day)
        else:
            if day.start is None:
                day = slice(1, day.stop, day.step)
            range = xrange(*day.indices(len(self)+1))
        return [Date(self.year, self.month, d) for d in range]

    for meth, op in [
        ('__eq__', '=='),
        ('__ne__', '!='),
        ('__lt__', '<'),
        ('__le__', '<='),
        ('__ge__', '>='),
        ('__gt__', '>'),
    ]:
        exec (
            'def %s(self, other):\n'
            '    return (self.year, self.month) %s (other.year, other.month)\n'
        ) % (meth, op)

    del meth, op

    def __add__(self, other):
        if not isinstance(other, int):
            raise TypeError('operation not supported for', self, other)
        if other < 0:
            return self - (-other)
        elif other == 0:
            return self
        else:
            year = self.year
            month = self.month
            year += other // 12
            month += other % 12
            if month > 12:
                year += 1
                month = month % 12
            return Month(year, month)

    def __sub__(self, other):
        if not isinstance(other, int):
            raise TypeError('operation not supported for', self, other)
        if other < 0:
            return self + (-other)
        elif other == 0:
            return self
        else:
            year = self.year
            month = self.month
            year -= other // 12
            month -= other % 12
            if month <= 0:
                year -= 1
                month = month + 12
            return Month(year, month)

    def __repr__(self):
        return 'Month(%d, %d)' % (self.year, self.month)

    @classmethod
    def current(klass):
        today = date.today()
        return klass(today.year, today.month)










class Year(int):

    def __new__(klass, year):
        if isinstance(year, date):
            year = year.year
        else:
            year = int(year)
            if year < MINYEAR or year > MAXYEAR:
                raise ValueError('year must be in range', MINYEAR, MAXYEAR)
        return int.__new__(klass, year)

    is_leap = property(leap_year)

    def __len__(self):
        return leap_year(self) and 366 or 365

    def __getitem__(self, month):
        if isinstance(month, int):
            if month < 0:
                month = 12 + month
            return Month(self.year, month)
        else:
            range = xrange(*month.indices(len(self)))
            return [Month(self.year, m) for m in range]

    for meth, op in [
        ('__eq__', '=='),
        ('__ne__', '!='),
        ('__lt__', '<'),
        ('__le__', '<='),
        ('__ge__', '>='),
        ('__gt__', '>'),
    ]:
        exec (
            'def %s(self, other):\n'
            '    return self.year %s other.year\n'
        ) % (meth, op)

    del meth, op

    def itermonths(self):
        for month in xrange(1, 13):
            yield Month(self, month)

    def iterweeks(self, firstday):
        raise NotImplementedError

    def iterdays(self):
        raise NotImplementedError










class TimeService:

    def now(self):
        return Datetime.now()

    def today(self):
        return Date.today()

    date      = staticmethod(Date)
    datetime  = staticmethod(Datetime)
    time      = staticmethod(time)
    timedelta = staticmethod(timedelta)
