#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "acl.h"

#define DELIM " \t\n\r"



typedef struct pnAclToken {
    char        *string;
    int         role;
    uchar       is_granted;
    pnAclUser   u;
    struct pnAclToken   *next;
} pnAclToken;



void pnAclPrintToken (pnAclToken *token)
{
    printf("token: >>%s<<\n", token->string);
    printf("\trole: %d\n", token->role);
    printf("\tis_granted: %d\n", token->is_granted);
    printf("\tis_policy: %d\n", token->u.is_policy);
    printf("\tis_group: %d\n", token->u.is_group);
    printf("\tuser: %d\n", token->u.user);
}



int pnAclParseToken (pnAclToken *token)
{
    char ibuf[64], *d, *s = token->string;
    int  j;

    if(strlen(s) < 4) {
        return -1;
    }

    /* Grant/Revoke */
    if(s[0] == '+') {
        token->is_granted = PACL_ALLOW;
    } else if (s[0] == '-') {
        token->is_granted = PACL_DENY;
    } else {
        return -2;
    }
            
    /* Role */
    d = index(s, ':');
    if(d == NULL) return -3;

    strncpy(ibuf, s+1, d-s-1);
    ibuf[d-s-1] = '\0';
    for(j=strlen(ibuf)-1; j >= 0; j--) {
        if(!isdigit(ibuf[j])) return -4;
    }
    token->role = strtol(ibuf, NULL, 10);

    /* User/Group */
    d++;
    if(strlen(d) < 1) return -5;
    if(strcmp(d, "*") == 0) {
        token->u.is_policy = 1;
    } else {
        if(d[0] == '@') {
            token->u.is_group = 1;
            d++;
        }
        for(j=strlen(d)-1; j >= 0; j--) {
            if(!isdigit(d[j])) return -6;
        }
        token->u.user = strtol(d, NULL, 10);
    }
    return 1;
}



/* Integer comparison function for qsort */
int cmpInt (const void *_l, const void *_r)
{
    int l = * ((int *) _l);
    int r = * ((int *) _r);

    if(l < r) return -1;
    if(l == r) return 0;
    return 1;
}



/* pnAclUser comparison function for qsort */
int cmpAclUser (const void *_l, const void *_r)
{
    pnAclUser *left = (pnAclUser *) _l;
    pnAclUser *right = (pnAclUser *) _r;

    if(left->is_policy > right->is_policy) return 1;
    if(left->is_policy < right->is_policy) return -1;

    if(left->is_group > right->is_group) return 1;
    if(left->is_group < right->is_group) return -1;

    if(left->user > right->user) return 1;
    if(left->user < right->user) return -1;

    return 0;
}


/* pnAclItem comparison function for qsort */
int cmpAclItem (const void *_l, const void *_r)
{
    pnAclItem *left = (pnAclItem *) _l;
    pnAclItem *right = (pnAclItem *) _r;
    int r;

    r = cmpAclUser(&(left->u), &(right->u));
    if(r != 0) return r;

    if(left->is_granted > right->is_granted) return 1;
    if(left->is_granted < right->is_granted) return -1;

    return 0;
}

/* pnAclToken comparison function for qsort */
int cmpAclToken (const void *_l, const void *_r)
{
    pnAclToken *left = (pnAclToken *) _l;
    pnAclToken *right = (pnAclToken *) _r;
    int r;

    if(left->role > right->role) return 1;
    if(left->role < right->role) return -1;

    r = cmpAclUser(&(left->u), &(right->u));
    if(r != 0) return r;

    if(left->is_granted > right->is_granted) return 1;
    if(left->is_granted < right->is_granted) return -1;

    return 0;
}




int pnAclCheck (pnAclCtx *ctx, pnAcl *acl)
{
    pnAclRole       *role;
    pnAclItem       *item, *last;
    int             ri, i;

    for(ri=0; ri < acl->role_num; ri++) {

        role = &(acl->roles[ri]);
        if(role->item_num < 2) continue;

        last = (pnAclItem*)((void *)acl + role->first_item);
        for(i=1, item = last+1; i<role->item_num; i++, item++, last++) {
            if( cmpAclUser( &(last->u), &(item->u) ) == 0) return 0;
        }
    }
    return 1;
}




pnAcl* pnAclFromString (pnAclCtx *ctx, const char *s)
{
    pnAclToken   *tokens=NULL, *tail=NULL, *token=NULL;
    int         role_num, token_num, token_span, space_span, r, i, j, last;
    int         *roles, asize, off, tc;
    char        *p, *scp;
    pnAcl        *acl=NULL;
    pnAclItem    *ai=NULL;


    i = strlen(s);
    p = scp = ctx->malloc(i+1);
    memcpy(scp, s, i+1);

    /* Ignore leading whitespace */
    space_span = strspn(p, DELIM);
    p += space_span;

    for(token_num=0; ; ++token_num) {

        /* Find token boundary */
        token_span = strcspn(p, DELIM);
        if(token_span == 0)
            break;
        
        /* Allocate pnAclToken and push them on top of linked list */
        token = ctx->malloc(sizeof(pnAclToken));
        memset(token, 0, sizeof(pnAclToken));
        token->string = p;
        token->next = NULL;
        if(!tokens) {
            tokens = tail = token;
        } else {
            tail->next = token;
            tail = token;
        }

        /* Ignore whitespaces after token */
        p += token_span;
        space_span = strspn(p, DELIM);
        p[0] = '\0';
        p += space_span;

        /* Parse token */
        r = pnAclParseToken(token);
        if(r != 1)
            goto dealloc_tokens;
    }
    
    /* Empty ACL? */
    if(token_num == 0) {
        acl = ctx->malloc(sizeof(pnAcl));
        memset(acl, 0, sizeof(pnAcl));
        acl->_size = sizeof(pnAcl);
        goto dealloc_scp;
    }

    /* Compute number of unique roles (role_num) */
    roles = ctx->malloc(sizeof(int) * token_num);
    for(i=0, token=tokens; i<token_num; i++) {
        roles[i] = token->role;
        token = token->next;
    }
    qsort(roles, token_num, sizeof(int), cmpInt);
    for(i=1, role_num=1, last=roles[0]; i<token_num; i++) {
        r = roles[i];
        if(r != last) {
            roles[role_num] = r;
            role_num++;
            last = r;
        }
    }

    /* Create pnAcl structure */
    off = sizeof(pnAcl) + role_num * sizeof(pnAclRole);
    asize = off + (token_num * sizeof(pnAclItem));
    acl = ctx->malloc(asize);
    memset(acl, 0, asize);
    acl->_size = asize;
    acl->role_num = role_num;

    for(i=0, tc=0; i<role_num; j++, i++) {
        acl->roles[i].role = r = roles[i];
        acl->roles[i].first_item = off + (tc*sizeof(pnAclItem));
        token = tokens;
        do {
            tail = token->next;
            if(token->role == r) {
                acl->roles[i].item_num++;
                ai = (pnAclItem*) ((void *) acl + off + (tc*sizeof(pnAclItem)));
                ai->is_granted = token->is_granted;
                memcpy(&(ai->u), &(token->u), sizeof(pnAclUser));
                tc++;
            }
            if(!tail) break;
            token = tail;
        } while (1);
    }

    /* Sort AclItems for every role */
    for(i=0; i<role_num; i++) {
        ai = (pnAclItem*)((void *)acl + acl->roles[i].first_item);
        qsort(ai, acl->roles[i].item_num, sizeof(pnAclItem), cmpAclItem);
    }
    if(pnAclCheck(ctx, acl)==1) {
        goto ok;
    } else {
        ctx->free(acl);
        acl = NULL;
    }
ok:
    ctx->free(roles);

dealloc_tokens :
    token = tokens;
    while(token) {
        tail = token->next;
        ctx->free(token);
        if(!tail) break;
        token = tail;
    }

dealloc_scp :
    ctx->free(scp);

    return acl;
}



int intStrlen (int n)
{
    int      l=0;

    if(n<0) {
        l++;
        n = - n;
    }
    if(n<10) return l+1;
    if(n<100) return l+2;
    if(n<1000) return l+3;
    if(n<10000) return l+4;
    if(n<100000) return l+5;
    if(n<1000000) return l+6;
    if(n<10000000) return l+7;
    if(n<100000000) return l+8;
    if(n<1000000000) return l+9;
    /* XXX this is incomplete */
    return -1;
}


int pnAclStrlen (pnAclCtx *ctx, pnAcl *acl)
{
    int         l=0, r, role_len, i, item_num=0;
    pnAclRole   *role;
    pnAclItem   *item;

    if(acl->role_num < 1) return 0;

    for(r=0; r<acl->role_num; r++) {
        role = &(acl->roles[r]);
        role_len = intStrlen(role->role);
        for(i=0; i<role->item_num; i++) {
            item = pnAclGetItem(acl, role, i);
            l += role_len + 2;
            if(item->u.is_policy) {
                l++;
            } else {
                if(item->u.is_group) l++;
                l += intStrlen(item->u.user);
            }
            item_num++;
        }
    }
    l += item_num -1;
    return l;
}




char * pnAclDumps (pnAclCtx *ctx, pnAcl *acl)
{
    char        *out;
    int         ri, ii, oi, ol;
    pnAclItem   *item;

    if(acl->role_num < 1) {
        out = ctx->malloc(1);
        out[0] = '\0';
        return out;
    }

    ol = pnAclStrlen(ctx, acl)+1;
    if(! (out = ctx->malloc(ol)))
        return NULL;

    memset(out, 0, ol);

    for(oi=0, ri=0; ri<acl->role_num; ri++) {

        item = pnAclGetItem(acl, &(acl->roles[ri]), 0);
        for(ii=0; ii<acl->roles[ri].item_num; ii++, item++) {
            if(oi > 0) {
                strcpy(&(out[oi]), " ");
                oi++;
            }
            if(item->u.is_policy) {
                oi += sprintf(&(out[oi]), "%c%d:*",
                        item->is_granted ? '+' : '-', acl->roles[ri].role
                );
            } else {
                oi += sprintf(&(out[oi]), "%c%d:%s%d",
                        item->is_granted ? '+' : '-',
                        acl->roles[ri].role,
                        item->u.is_group ? "@" : "",
                        item->u.user
                );
            }
        }
    }
    return out;
}





pnAcl * pnAclJoin (pnAclCtx *ctx, pnAcl *base, pnAcl *join)
{
    pnAcl       *res;
    int         br, jr, bi, ji, r, i;
    pnAclRole   *brole, *jrole, *role;
    pnAclItem   *bitem, *jitem, *item;

    res           = ctx->malloc(base->_size + join->_size - sizeof(pnAcl));
    res->_size    = base->_size + join->_size - sizeof(pnAcl);
    res->role_num = 0;

    role = &(res->roles[0]);
    item = (pnAclItem*) (
        (void *)res + 
        sizeof(pnAcl) +
        ((base->role_num+join->role_num) * sizeof(pnAclRole))
    );

    for(br=jr=r=0; br<base->role_num && jr<join->role_num; ) {

        brole = &(base->roles[br]);
        jrole = &(join->roles[jr]);

        if(brole->role < jrole->role)  {
            memcpy(role, brole, sizeof(pnAclRole));
            role->first_item = ((void *)item) - ((void *)res);
            for(i=0; i<brole->item_num; i++, item++) {
                memcpy(item, pnAclGetItem(base, brole, i), sizeof(pnAclItem));
            }
            res->role_num++; role++; br++;

        } else if(brole->role > jrole->role)  {
            memcpy(role, jrole, sizeof(pnAclRole));
            role->first_item = ((void *)item) - ((void *)res);
            for(i=0; i<jrole->item_num; i++, item++) {
                jitem = pnAclGetItem(join, jrole, i);
                memcpy(item, jitem, sizeof(pnAclItem));
            }
            res->role_num++; role++; jr++;

        } else {
            memcpy(role, brole, sizeof(pnAclRole));
            role->first_item = ((void *)item) - ((void *)res);
            role->item_num = 0;
            for(bi=ji=0; bi<brole->item_num && ji<jrole->item_num; ) {
                bitem = pnAclGetItem(base, brole, bi);
                jitem = pnAclGetItem(join, jrole, ji);
                if(cmpAclUser(&(bitem->u), &(jitem->u)) < 0) {
                    memcpy(item, bitem, sizeof(pnAclItem));
                    item++; bi++;
                } else if(cmpAclUser(&(bitem->u), &(jitem->u)) > 0) {
                    memcpy(item, jitem, sizeof(pnAclItem));
                    item++; ji++;
                } else {
                    memcpy(item, jitem, sizeof(pnAclItem));
                    item++; bi++; ji++;
                }
                role->item_num++;

            }
            for( ; bi<brole->item_num; bi++, item++) {
                memcpy(item, pnAclGetItem(base, brole, bi), sizeof(pnAclItem));
                role->item_num++;
            }
            for( ; ji<jrole->item_num; ji++, item++) {
                memcpy(item, pnAclGetItem(join, jrole, ji), sizeof(pnAclItem));
                role->item_num++;
            }
            res->role_num++; role++; br++; jr++;
        }
    }

    for( ;br<base->role_num; br++, role++, res->role_num++) {
        brole = &(base->roles[br]);
        memcpy(role, brole, sizeof(pnAclRole));
        role->first_item = ((void *)item) - ((void *)res);
        for(i=0; i<brole->item_num; i++, item++) {
            memcpy(item, pnAclGetItem(base, brole, i), sizeof(pnAclItem));
        }
    }
    for( ;jr<join->role_num; jr++, role++, res->role_num++) {
        jrole = &(join->roles[jr]);
        memcpy(role, jrole, sizeof(pnAclRole));
        role->first_item = ((void *)item) - ((void *)res);
        for(i=0; i<jrole->item_num; i++, item++) {
            memcpy(item, pnAclGetItem(join, jrole, i), sizeof(pnAclItem));
        }
    }

    return res;
}






int pnAclAllows(pnAclCtx *ctx,pnAcl *acl,int role,int user,int gnum,int *groups)
{
    pnAclItem   *item;
    pnAclRole   *_role;
    int         r, i, g;
    
    for(r=0; r < acl->role_num ; r++) {

        _role = &(acl->roles[r]);

        if(_role->role > role) 
            return PACL_DEFAULT_POLICY;

        if(_role->role == role) {

            for(i=0; i < _role->item_num; i++) {

                item = pnAclGetItem(acl, _role, i);

                if(item->u.is_policy) {
                    return item->is_granted;
                }

                if(item->u.is_group) {
                    for(g=0; g<gnum; g++) {
                        if(item->u.user == groups[g]) {
                            return item->is_granted;
                        }
                    }
                }

                if(item->u.user == user) {
                    return item->is_granted;
                }
            }
            return PACL_DEFAULT_POLICY;
        }
    }
    return PACL_DEFAULT_POLICY;
}


int pnAclEq(pnAclCtx *ctx, pnAcl *l, pnAcl *r)
{
    pnAclRole   *rl, *rr;
    int         j,i;

    if(l->role_num != r->role_num) return 0;
    for(j=0; j < r->role_num ; j++) {
        rl = &(l->roles[j]);
        rr = &(r->roles[j]);
        if (rl->role != rr->role) return 0;
        if (rl->item_num != rr->item_num) return 0;
        for(i=0; i < rl->item_num; i++) {
            if(cmpAclItem(pnAclGetItem(l, rl, i), pnAclGetItem(r, rr, i)) != 0)
                return 0;
        }
    }
    return 1;
}
