#include "Python.h"
#include "structmember.h"
#include "acl.h"


#define PyPNACL_Check(op) PyObject_TypeCheck(op, &pnacltype)



pnAclCtx  ctx = {malloc, realloc, free, printf};



typedef struct {
    PyObject_HEAD
    pnAcl   *pnacl;
} pnaclobject;

static PyTypeObject pnacltype;



static PyObject *
pnacl_new (PyTypeObject *type, PyObject *args, PyObject *kw)
{
    char        *buf, *err;
    pnAcl       *_pnacl;
    pnaclobject *pnacl;
    int         l;

    if(! PyArg_ParseTuple(args, "s", &buf)) {
        return NULL;
    }

    if( ! (_pnacl = pnAclFromString(&ctx, buf)) ) {
        l = strlen(buf) + 40;
        err = PyMem_Malloc(l);
        PyOS_snprintf(err, l, "Syntax error in pnacl argument '%s'", buf);
        PyErr_SetString(PyExc_ValueError, err);
        PyMem_Free(err);
        return NULL;
    }

    if(! (pnacl = PyObject_NEW(pnaclobject, &pnacltype)) ) {
        return NULL;
    }

    pnacl->pnacl = _pnacl;
    return (PyObject *) pnacl;
}

static void
pnacl_dealloc (pnaclobject *self)
{
    if(self->pnacl) free(self->pnacl);
}


    
static PyObject *
pnacl_join (pnaclobject *self, PyObject *args)
{
    pnaclobject *arg, *res;
    pnAcl       *_res;

    if(! PyArg_ParseTuple(args, "O!", &pnacltype, &arg)) {
        return NULL;
    }
    if( ! (_res = pnAclJoin(&ctx, self->pnacl, arg->pnacl)) ) {
        PyErr_SetString(
                PyExc_ValueError,
                "Internal error in row_acl module - pnAclJoin has returned NULL"
        );
        return NULL;
    }

    if(! (res = PyObject_NEW(pnaclobject, &pnacltype)) ) {
        return NULL;
    }
    res->pnacl = _res;
    return (PyObject *) res;
}


static PyObject *
pnacl_allows (pnaclobject *self, PyObject *args)
{
    int         role, user, *groups, gnum, allows, i;
    PyObject    *pygroups, *pygroup;

    if(! PyArg_ParseTuple(args, "iiO", &role, &user, &pygroups)) {
        return NULL;
    }
    if(!PySequence_Check(pygroups)) {
        PyErr_SetString(
            PyExc_ValueError, "Third argument to allows must be sequence"
        );
        return NULL;
    }
    gnum = PySequence_Length(pygroups);
    groups = malloc(gnum * sizeof(int));
    for(i=0; i<gnum; i++) {
        pygroup = PySequence_GetItem(pygroups, i);

        if(!PyInt_Check(pygroup)) {
            free(groups);
            PyErr_SetString(
                PyExc_ValueError, "Group sequence contains nonintegers"
            );
            Py_DECREF(pygroup);
            return NULL;
        }
        groups[i] = PyInt_AsLong(pygroup);
        Py_DECREF(pygroup);

    }

    allows = pnAclAllows(&ctx, self->pnacl, role, user, gnum, groups);
    free(groups);
    if(allows) {
        Py_INCREF(Py_True);
        return Py_True;
    } else {
        Py_INCREF(Py_False);
        return Py_False;
    }
}


static PyMethodDef pnacl_methods[] = {
    {"join",    (PyCFunction) pnacl_join, METH_VARARGS,
                    PyDoc_STR(".join(a) - join argument pnacl")},
    {"allows",  (PyCFunction) pnacl_allows, METH_VARARGS,
                    PyDoc_STR(".allows(r,u,g) - check permission")},
    {NULL,	NULL},
};


static PyObject *
pnacl_size_get(pnaclobject *self)
{
    return PyInt_FromLong(self->pnacl->_size);
}

static PyGetSetDef pnacl_getsets[] = {
    {"_size", (getter) pnacl_size_get, NULL,
              PyDoc_STR("Size of pnAcl structure (for debugging)")},
    {0}
};



static PyObject *
pnacl_repr(pnaclobject *self)
{
    char        *dump;
    PyObject    *ret;
    dump = pnAclDumps(&ctx, self->pnacl);
    ret = PyString_FromFormat("row_acl.NumAcl('%s')", dump);
    free(dump);
    return ret;
}


static PyObject *
pnacl_str(pnaclobject *self)
{
    char        *dump;
    PyObject    *ret;
    dump = pnAclDumps(&ctx, self->pnacl);
    ret = PyString_FromString(dump);
    free(dump);
    return ret;
}

static PyObject *
pnacl_richcompare(pnaclobject *self, PyObject *other, int op)
{
    pnaclobject *right;
    int cmp;
    PyObject *result;

    if(!PyPNACL_Check(other)) {
        PyErr_SetString(
                PyExc_ValueError,
                "Only row_acl.nacl instances can be compared with row_acl.nacl."
        );
        return NULL;
    }
    right = (pnaclobject *) other;
    cmp = pnAclEq(&ctx, self->pnacl, right->pnacl);
    if(op == Py_EQ) {
        result = cmp ? Py_True : Py_False;
    } else if(op == Py_NE) {
        result = cmp ? Py_False : Py_True ;
    } else {
        PyErr_SetString(
                PyExc_NotImplementedError, "Comparison not implemented"
        );
        return NULL;
    }
    Py_INCREF(result);
    return result;
}



static PyTypeObject pnacltype = {
	PyObject_HEAD_INIT(&PyType_Type)
	0,
	"adt4.row_acl.NumAcl",
	sizeof(pnaclobject),
	0,
	(destructor) pnacl_dealloc,		/* tp_dealloc */
	0,					/* tp_print */
	0,					/* tp_getattr */
	0,					/* tp_setattr */
	0,					/* tp_compare */
	(reprfunc) pnacl_repr,			/* tp_repr */
	0,					/* tp_as_number */
	0,					/* tp_as_sequence */
	0,					/* tp_as_mapping */
	0,					/* tp_hash */
	0,					/* tp_call */
	(reprfunc) pnacl_str,			/* tp_str */
	0,					/* tp_getattro */
	0,					/* tp_setattro */
	0,					/* tp_as_buffer */
	Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE, /* tp_flags */
	0,					/* tp_doc */
	0,					/* tp_traverse */
	0,					/* tp_clear */
	(richcmpfunc)pnacl_richcompare,	        /* tp_richcompare */
	0,					/* tp_weaklistoffset */
	0,					/* tp_iter */
	0,					/* tp_iternext */
	pnacl_methods,			        /* tp_methods */
	0,					/* tp_members */
	pnacl_getsets,			        /* tp_getset */
	0,                                   	/* tp_base */
	0,					/* tp_dict */
	0,					/* tp_descr_get */
	0,					/* tp_descr_set */
	0,					/* tp_dictoffset */
        0,		                        /* tp_init */
	0,					/* tp_alloc */
	pnacl_new,			        /* tp_new */
};


PyMODINIT_FUNC
initrow_acl(void)
{
    PyObject *m;

    m = Py_InitModule("adt4.row_acl", NULL);
    if (m == NULL)
        return;

    if (PyType_Ready(&pnacltype) < 0)
        return;

    Py_INCREF(&pnacltype);
    if (PyModule_AddObject(m, "NumAcl",
                           (PyObject *) &pnacltype) < 0)
        return;
}
