from unittest import *
from adt4.domain_name import DomainName as DN


class TestDomainName (TestCase) :

    def test01_Parsing (self) :
        # Invalid characters
        self.assertRaises(ValueError, DN, '#')
        self.assertRaises(ValueError, DN, 'aaa.-xxx.cz.')
        self.assertRaises(ValueError, DN, 'aaa._xxx.cz.')

        # Empty domain
        self.assertRaises(ValueError, DN, '')

        # Nonexisting top level domain is OK
        self.assertRaises(ValueError, DN, 'neex.')
        self.assertRaises(ValueError, DN, 'b.neex.')

        # Normalized is in lower case
        self.assertEqual(DN('CZ').normalized, 'cz.')
        self.assertEqual(DN('DAT.CZ').normalized, 'dat.cz.')
        self.assertEqual(DN('WWW.DAT.CZ').normalized, 'www.dat.cz.')

        # Domain root
        self.assertEqual(DN('.').normalized, '.')

        # Top level domain
        self.assertEqual(DN('cz.').normalized, 'cz.')
        self.assertEqual(DN('CZ.').normalized, 'cz.')

        # Dot at end is allowed
        self.assertEqual(DN('dat.cz.').normalized, 'dat.cz.')

        # DomainName as argument is simply returned
        cz = DN('cz')
        assert DN(cz) is cz

        # Arpa domains
        self.assertEqual(DN('127.in-addr.arpa').normalized, '127.in-addr.arpa.')

        # Wildcards
        self.assertRaises(ValueError, DN, '*')
        self.assertRaises(ValueError, DN, '*.')
        self.assertRaises(ValueError, DN, 'www.*.cz')
        self.assertRaises(ValueError, DN, 'www.*.cz')
        self.assertEqual(DN('*.cz').normalized, '*.cz.')
        self.assertEqual(DN('*.cz.').normalized, '*.cz.')
        self.assertEqual(DN('*.dat.cz').normalized, '*.dat.cz.')
        self.assertEqual(DN('*.dat.cz.').normalized, '*.dat.cz.')

    def test02_ReadOnlyness (self) :
        self.assertRaises(AttributeError, setattr, DN('cz'), 'parts', 'x')
        self.assertRaises(AttributeError, delattr, DN('cz'), 'parts')
        self.assertRaises(AttributeError, setattr, DN('cz'), 'normalized', 'x')
        self.assertRaises(AttributeError, delattr, DN('cz'), 'normalized')
        self.assertRaises(AttributeError, setattr, DN('cz'), 'level', 'x')
        self.assertRaises(AttributeError, delattr, DN('cz'), 'level')

    def test03_Level (self) :
        self.assertEqual(DN('.').level, 0)
        self.assertEqual(DN('cz').level, 1)
        self.assertEqual(DN('cz.').level, 1)
        self.assertEqual(DN('dat.cz').level, 2)
        self.assertEqual(DN('dat.cz.').level, 2)
        self.assertEqual(DN('www.dat.cz').level, 3)
        self.assertEqual(DN('www.dat.cz.').level, 3)

    def test04_IsHashable (self) :
        self.assertEqual(hash(DN('.')), hash('.'))
        self.assertEqual(hash(DN('cz')), hash('cz'))
        self.assertEqual(hash(DN('cz.')), hash('cz.'))
        self.assertEqual(hash(DN('CZ')), hash('CZ'))
        self.assertEqual(hash(DN('dat.cz')), hash('dat.cz'))
        self.assertEqual(hash(DN('DAT.CZ')), hash('DAT.CZ'))

    def test05_Parts (self) :
        self.assertEqual(DN('.').parts, ())
        self.assertEqual(DN('cz').parts, ('cz',))
        self.assertEqual(DN('CZ').parts, ('cz',))
        self.assertEqual(DN('dat.cz').parts, ('cz','dat'))
        self.assertEqual(DN('DAT.CZ').parts, ('cz','dat'))
        self.assertEqual(DN('www.dat.cz').parts, ('cz','dat','www'))
        self.assertEqual(DN('WWW.DAT.CZ').parts, ('cz','dat','www'))

    def test10_Comparison (self) :
        # Root is smallest
        self.failUnless(DN('.') == DN('.'))
        self.failUnless(DN('.') < DN('cz'))
        self.failUnless(DN('.') <= DN('cz'))
        self.failUnless(DN('.') < DN('dat.cz'))
        self.failUnless(DN('.') <= DN('dat.cz'))
        self.failUnless(DN('cz') > DN('.'))
        self.failUnless(DN('cz') >= DN('.'))
        self.failUnless(DN('dat.cz') > DN('.'))
        self.failUnless(DN('dat.cz') >= DN('.'))

        # Top level is smaller
        self.failUnless(DN('cz') == DN('cz'))
        self.failUnless(DN('cz') < DN('dat.cz'))
        self.failUnless(DN('cz') <= DN('dat.cz'))
        self.failUnless(DN('dat.cz') > DN('cz'))
        self.failUnless(DN('dat.cz') >= DN('cz'))

        # Lexicographic and reversed parts
        self.failUnless(DN('a.cz') > DN('z.ar'))
        self.failUnless(DN('a.cz') >= DN('z.ar'))
        self.failUnless(DN('z.ar') < DN('a.cz'))
        self.failUnless(DN('z.ar') <= DN('a.cz'))
        self.failUnless(DN('z.ar') != DN('a.cz'))



def suite() :
    return TestSuite((
        makeSuite(TestDomainName),
    ))


if __name__ == '__main__':
    main(defaultTest='suite')
