from unittest import *
from adt4.row_acl import NumAcl


# Groups

class GROUP : gid = None
class ADMIN (GROUP) : gid = 100
class HOSTMASTER (GROUP) : gid = 200
class WEBMASTER (GROUP) : gid = 300

# Users

class USER :
    uid     = None
    member  = None

class JAN (USER) :
    uid     = 1
    member  = (HOSTMASTER,)

class PETR (USER) :
    uid     = 2
    member  = (WEBMASTER,)

class PAUL (USER) :
    uid     = 3
    member  = (HOSTMASTER, WEBMASTER)

class ROBERT (USER) :
    uid     = 4
    member  = (ADMIN,)




# ACL case

class RowAclCase :

    def __init__ (self, acl, **kw) :
        l = globals()
        self._orig = acl
        self.acl = NumAcl(acl)
        self.res = []
        a = self.res.append
        for u,r in kw.iteritems() :
            a((l[u], r))





class TestAclNormalization (TestCase) :

    cases = [
        ("", ""),
        ("+50:100 +50:@100 +50:*", "+50:100 +50:@100 +50:*"),
        ("+50:* +50:@100 +50:100", "+50:100 +50:@100 +50:*"),
        ("+50:* +50:@100 +50:100 +40:* +40:@100 +40:100 ",
         "+40:100 +40:@100 +40:* +50:100 +50:@100 +50:*"),

        # Syntacticaly incorrect
        ("+50:*+50:*", None),
        ("+:*", None),
        ("+*", None),
        ("+:", None),
        ("+", None),
        (":", None),
        ("50:100", None),

        # Semanticaly incorrect
        ("+50:* +50:*", None),
        ("+50:* -50:*", None),
        ("-50:* +50:*", None),
        ("+50:100 +50:100", None),
        ("+50:100 -50:100", None),
        ("+50:@100 +50:@100", None),
        ("+50:@100 -50:@100", None),
    ]

    def test (self) :
        for p,n in self.cases :
            if n is None :
                self.assertRaises(ValueError, NumAcl, p)
            else :
                self.assertEqual(str(NumAcl(p)), n)






class TestDefaultScenarios (TestCase) :

    scenarios = [

        # Everybody allowed
        RowAclCase("+50:*",
                    JAN = {50:1},
                    PETR = {50:1},
                    PAUL = {50:1},
                    ROBERT = {50:1},
        ),

        # Everybody denied
        RowAclCase("-50:*",
                    JAN = {50:0},
                    PETR = {50:0},
                    PAUL = {50:0},
                    ROBERT = {50:0},
        ),

        # Everybody denied
        RowAclCase("",
                    JAN = {50:0},
                    PETR = {50:0},
                    PAUL = {50:0},
                    ROBERT = {50:0},
        ),

        # Default permission for two roles
        RowAclCase("-50:* +60:*",
                    JAN = {50:0, 60:1},
                    PETR = {50:0, 60:1},
                    PAUL = {50:0, 60:1},
                    ROBERT = {50:0, 60:1},
        ),

        # Nonexisting role
        RowAclCase("-40:* +60:*",
                    JAN = {50:0, 60:1},
                    PETR = {50:0, 60:1},
                    PAUL = {50:0, 60:1},
                    ROBERT = {50:0, 60:1},
        ),

    ]

    def test (self) :
        for scenario in self.scenarios :
            for user, resdict in scenario.res :
                for role, result in resdict.iteritems() :
                    groups = [g.gid for g in user.member]
                    r = scenario.acl.allows(role, user.uid, groups)
                    self.failIf(
                        bool(r) != bool(result),
                        'Scenario failed: row_acl("%s").allows(%d, %r, %s)' %(
                            scenario._orig, role, user.uid,
                            ", ".join(map(repr, groups))
                        )
                    )



class TestUserScenarios (TestDefaultScenarios) :

    scenarios = [

        # Jan allowed
        RowAclCase("+50:1 -50:2",
                    JAN = {50:1},
                    PETR = {50:0},
                    PAUL = {50:0},
                    ROBERT = {50:0},
        ),

        # Petr denied
        RowAclCase("+50:1 -50:2 +50:*",
                    JAN = {50:1},
                    PETR = {50:0},
                    PAUL = {50:1},
                    ROBERT = {50:1},
        ),
    ]


class TestGroupScenarios (TestDefaultScenarios) :

    scenarios = [

        # Admin group allowed, rest denied
        RowAclCase("+50:@100 -50:*",
                    JAN = {50:0},
                    PETR = {50:0},
                    PAUL = {50:0},
                    ROBERT = {50:1},
        ),

        # Admin or hostmaster group allowed
        RowAclCase("+50:@100 +50:@200",
                    JAN = {50:1},
                    PETR = {50:0},
                    PAUL = {50:1},
                    ROBERT = {50:1},
        ),
    ]


class TestCombinedScenarios (TestDefaultScenarios) :

    scenarios = [

        # Hostmaster group allowed but Jan denied
        RowAclCase("+50:@200 -50:1",
                    JAN = {50:0},
                    PETR = {50:0},
                    PAUL = {50:1},
                    ROBERT = {50:0},
        ),

        # Hostmaster and webmaster group allowed but Paul denied
        RowAclCase("+50:@200 +50:@300 -50:3",
                    JAN = {50:1},
                    PETR = {50:1},
                    PAUL = {50:0},
                    ROBERT = {50:0},
        ),

        # Hostmaster and webmaster group allowed but Paul denied
        RowAclCase("+50:@200 +50:@300 -50:3 +50:*",
                    JAN = {50:1},
                    PETR = {50:1},
                    PAUL = {50:0},
                    ROBERT = {50:1},
        ),

    ]






class TestAclJoin (TestCase) :

    cases = [
        ( "", "", ""),

        ( "", "-50:*", "-50:*"),
        ( "", "-50:100", "-50:100"),
        ( "", "-50:@100", "-50:@100"),

        ( "-50:*", "", "-50:*"),
        ( "-50:100", "", "-50:100"),
        ( "-50:@100", "", "-50:@100"),

        ( "-50:*", "+50:*", "+50:*"),
        ( "-50:100", "+50:100", "+50:100"),
        ( "-50:@100", "+50:@100", "+50:@100"),

        ( "+50:*", "-50:*", "-50:*"),
        ( "+50:100", "-50:100", "-50:100"),
        ( "+50:@100", "-50:@100", "-50:@100"),

        ( "", "+50:10 -50:*", "+50:10 -50:*"),
        ( "+50:10 -50:*", "", "+50:10 -50:*"),

        ( "+50:*",
          "+50:10 -50:*",
          "+50:10 -50:*"),

        ( "+50:100",
          "+50:200",
          "+50:100 +50:200"),

        ( "+50:100 -50:*",
          "-50:200 +50:*",
          "+50:100 -50:200 +50:*"),

        ( "+50:100 -50:*",
          "-50:100 +50:*",
          "-50:100 +50:*"),

        ( "+50:100 +50:@200 -50:*",
          "-50:100 +50:@300",
          "-50:100 +50:@200 +50:@300 -50:*"),

        ( "+50:@200 -50:*",
          "+50:@300",
          "+50:@200 +50:@300 -50:*"),

        ( "+50:@200",
          "+50:@300 -50:*",
          "+50:@200 +50:@300 -50:*"),

        ( "+50:@200 -50:*",
          "+50:@300 -50:*",
          "+50:@200 +50:@300 -50:*"),

        ( "+50:@100 +60:@100",
          "+50:@300 +60:*",
          "+50:@100 +50:@300 +60:@100 +60:*"),

        ( "+50:@100 +60:100",
          "+50:@300 +60:*",
          "+50:@100 +50:@300 +60:100 +60:*"),

    ]

    def test (self) :
        for p, n, r in self.cases :
            ap = NumAcl(p)
            an = NumAcl(n)
            ar = ap.join(an)
            self.assertEqual(str(ar), r)




def suite():
    return TestSuite((
        makeSuite(TestAclNormalization),
        makeSuite(TestDefaultScenarios),
        makeSuite(TestUserScenarios),
        makeSuite(TestGroupScenarios),
        makeSuite(TestCombinedScenarios),
        makeSuite(TestAclJoin),
    ))


if __name__ == '__main__':
    main( defaultTest='suite' )
