# -*- coding: utf-8 -*-

import re

from lib4.imports import importObject
from lib4.trees import AbstractTreeNode
from lib4.text import unicode2ascii, to_unicode

from wm4.api import *
from wm4.schema.api import Entity, Feature, Attribute, Reference, Feature
from wm4.schema.api import Default, Required, ReadOnly, PrimaryKey, Workspace
from wm4.schema.api import ComputeDefault, SequenceName
from wm4.schema.containers import PersistentList

from comp4.dbgettext.api import Message

from config import MAX_CHSET, DEFAULT_POSITION
from config import manager, changeset_manager, RootProperty
from entities import RepositoryChangeset, RepositoryNode, RepositoryElement


__all__ = [
    'ChildProxy', 'ChildrenProxy', 'Metadata',
    'NodeVersion', 'LastNode', 'PublicNode',
    'Element', 'Document', 'PublicElement',
    'MenuItem', 'Menu',
]

LETTERS_AND_NUMBERS = re.compile('[^a-zA-Z0-9]+')

##################### Element proxies #######################

class ChildProxy(object):
    
    def __init__(self, kind, name, from_obj=lambda o:o):
        self.kind = kind
        self.name = name
        self.from_obj = from_obj

    def __get__(self, obj, typ=None):
        if obj is None:
            return self
        for child in self.from_obj(obj).children:
            if child.kind == self.kind and child.name.startswith(self.name):
                return child
        return None

    def __set__(self, obj, val):
        assert val.kind == self.kind and val.name.startswith(self.name)
        obj = self.from_obj(obj)
        for idx, child in enumerate(obj.children):
            if child.kind == self.kind and child.name.startswith(self.name):
                obj.children[idx] = val
                return
        obj.children.append(val)

    def __delete__(self, obj):
        obj = self.from_obj(obj)
        for idx, child in enumerate(obj.children):
            if child.kind == self.kind and child.name.startswith(self.name):
                del obj.children[idx]

class ChildrenProxyView(object):
    
    def __init__(self, children, kind, name):
        self.children = children
        self.kind = kind
        self.name = name
        self.filtered = [
            child for child in children
                   if child.kind == self.kind
                  and child.name.startswith(self.name)
        ]

    def __getitem__(self, idx):
        return list.__getitem__(self.filtered, idx)

    def __len__(self):
        return len(self.filtered)

    def append(self, val):
        self.filtered.append(val)
        self.update_children()

    def extend(self, val):
        self.filtered.extend(val)
        self.update_children()

    def __setitem__(self, idx, value):
        list.__setitem__(self.filtered, idx, value)
        self.update_children()

    def __delitem__(self, idx):
        list.__delitem__(self.filtered, idx)
        self.update_children()

    def update_children(self):
        children = [
            child for child in self.children
                   if child.kind != self.kind
                   or not child.name.startswith(self.name)
        ]
        for e in self.filtered:
            assert e.kind == self.kind and e.name.startswith(self.name)
            children.append(e)
        self.children[:] = children

    def __repr__(self):
        return repr(self.filtered)

class ChildrenProxy(object):
    
    def __init__(self, kind, name, from_obj=lambda o:o):
        self.kind = kind
        self.name = name
        self.from_obj = from_obj

    def __get__(self, obj, typ=None):
        if obj is None:
            return self
        else:
            proxy = ChildrenProxyView(
                    self.from_obj(obj).children, self.kind, self.name)
            # caching?
            return proxy

    def __set__(self, obj, val):
        proxy = ChildrenProxyView(
                self.from_obj(obj).children, self.kind, self.name)
        proxy[:] = val




class Metadata:

    protocols.advise(instancesProvide=[web.IMetadata])

    title = None
    author = None
    keywords = None
    description = None

    def __init__(self, **kw):
        self.__dict__.update(kw)



##################### RepositoryNode facade #######################


class NodeVersion(Entity, AbstractTreeNode):
    partition      = Attribute(str, Required)
    inode          = Attribute(int, Required, Default(-1))
    changeset      = Attribute(int, Required, Default(MAX_CHSET))
    changeset_from = Attribute(int)
    changeset_till = Attribute(int)
    public_version = Attribute(int)
    parent_node    = Attribute(int)
    klass          = Attribute(str, Required, Default('cms'))

    binding.metadata(PrimaryKey('inode', 'changeset'))

    def inodepath(self):
        if RootProperty(self.partition, 'seo_url')(default=False) == True:
            return '%s-%s' % (self.inode, LETTERS_AND_NUMBERS.sub('-',
                    unicode2ascii(self.title).lower()).strip('-'))
        else:
            return str(self.inode)
    inodepath = binding.Make(inodepath)

    # utils

    partMgr        = binding.Make(lambda s: manager()[s.partition])
    isDeleted      = binding.Make(lambda s: s.changeset_till < MAX_CHSET)

    # schema

    def loadDeactivatedFeature(self, feature):
        manager().preloadNode(self)

    def onChange(self, added, removed):
        raise TypeError('Immutable instance', self)

    # tree

    parent         = binding.Make(lambda s: manager().lastParentFor(s))
    allParents     = property(lambda s: s.iterParentsRev())

    def lastChildren(self):
        return manager().lastChildrenFor(self)
    lastChildren = binding.Make(lastChildren)

    def validChildren(self):
        dpg = self.dir_positions.get
        vc = [c for c in self.lastChildren if not c.isDeleted]
        vc.sort(key=lambda c: (dpg(c.inode, DEFAULT_POSITION), c._oid))
        return vc
    validChildren = binding.Make(validChildren)

    def publicChildren(self):
        dpg = self.dir_positions.get
        #dpg = {}.get
        pc = list(manager().publicChildrenFor(self))
        pc.sort(key=lambda c: (dpg(c.inode, DEFAULT_POSITION), c._oid))
        return pc
    publicChildren = binding.Make(publicChildren)

    deletedChildren = binding.Make(
                    lambda s: [c for c in s.lastChildren if c.isDeleted])

    children = binding.Make(lambda s: s.validChildren)

    _dir_positions = ChildProxy('navig','dir.positions')
    
    def dir_positions(self):
        dp = self.__dict__.get('_preloaded_dir_positions', NOT_FOUND)
        if dp is NOT_FOUND:
            dp = self._dir_positions
        parsed = {}
        if dp:
            for v in dp.split(','):
                v = v.strip()
                if not v:
                    continue
                node, idx = v.split('=')
                parsed[int(node)] = int(idx)
        return parsed
    dir_positions = binding.Make(dir_positions)

    def publicVersions(self):
        from entities import CmsPublicVersion
        return tuple(self._workspace('''[
            p for p in CmsPublicVersion
               if p.inode == self.inode
              and query.OrderBy(query.Desc(p.pstart))
        ]'''))
    publicVersions = binding.Make(publicVersions)
    
    changesetObj = binding.Make(
                lambda s: s._workspace.get(RepositoryChangeset, s.changeset))
        
    modTime = binding.Make(
                lambda s: s.changesetObj.creation_time)

    def createdBy(self):
        from comp4.users.entities import UserContext
        return UserContext().userByLogin(self.changesetObj.created_by)
    createdBy = binding.Make(createdBy)

    _dir_positions = binding.Make(
        lambda s: s.findElementValue('navig','dir.positions',''))

    def propertyFactories(self):
        factories = []
        for pf in RootProperty(
            self.partition, 'property_factories.%s' % self.toplevel
        )():
            factories.append(importObject(pf))
        return factories
    propertyFactories = binding.Make(propertyFactories)

    def elements(self):
        return manager().getDocument(Document, self)
    elements = binding.Make(elements)

    def findElement(self, kind, name, default=NOT_GIVEN):
        for elem in self.elements.iterDepthFirst(include_self=False):
            if elem.kind == kind and elem.name == name:
                return elem
        if default is NOT_GIVEN:
            raise KeyError('element not found', kind, name)
        return default
        
    def findElements(self, kind, name):
        for elem in self.elements.iterDepthFirst(include_self=False):
            if elem.kind == kind and elem.name == name:
                yield elem
    
    def findElementValue(self, kind, name, default=NOT_GIVEN):
        for elem in self.elements.iterDepthFirst(include_self=False):
            if elem.kind == kind and elem.name == name:
                return elem.value
        if default is NOT_GIVEN:
            raise KeyError('element not found', kind, name)
        return default
        
    def findElementValues(self, kind, name):
        for elem in self.elements.iterDepthFirst(include_self=False):
            if elem.kind == kind and elem.name == name:
                yield elem.value
        
    def findElementByID(self, eid, default=NOT_GIVEN):
        eid = int(eid)
        for e in self.elements.iterBreadthFirst(include_self=False):
            if e.eid == eid:
                return e
        if default is NOT_GIVEN:
            raise ValueError('element not found', eid)
        return default

    document = ChildProxy('cms', 'toplevel.', lambda s: s.elements)
    toplevel = binding.Make(lambda s: s.document and s.document.name or None)
    style    = binding.Make(lambda s: s.document and s.document.value or None)

    title_e = ChildProxy('cms', 'head.title', lambda s: s.document)
    def title(self, dict, attr):
        t = dict.get('_preloaded_title', NOT_FOUND)
        if t is NOT_FOUND:
            if self.title_e:
                return self.title_e.value
        else:
            return to_unicode(t[0])
    title   = binding.Make(title)
    def _title(self, dict, attr):
        t = dict.get('_preloaded_title', NOT_FOUND)
        if t is NOT_FOUND:
            if self.title_e:
                return self.title_e._value
        else:
            ut = to_unicode(t[0])
            return Message(
                ut, 'comp4.cms.pagelets.Title:value/%s' % t[1],
                self._workspace
            )
    _title  = binding.Make(_title)

    abs_e     = ChildProxy('cms', 'pagelet.abstract', lambda s: s.document)
    abstract  = binding.Make(lambda s: s.abs_e and s.abs_e.value or None)
    _abstract = binding.Make(lambda s: s.abs_e and s.abs_e._value or None)

    def metadata(self):
        meta = {}
        for e in self.elements.iterBreadthFirst():
            if e.kind == 'meta':
                if e.value:
                    meta[str(e.name)] = e.value
        return Metadata(**meta)
    metadata = binding.Make(metadata)

    def __repr__(self):
        return 'NodeVersion(%d,%d)' % self._oid


@Workspace.get.when(
    'self in Workspace and klass in type and issubclass(klass, NodeVersion)')
def get_node(self, klass, key, default=NOT_GIVEN):
    inode, changeset = key
    return manager().getNode(
        NodeVersion, inode, changeset=changeset, default=default
    )



class PublicNode(NodeVersion):

    binding.metadata(PrimaryKey('inode'))

    parent   = binding.Make(lambda s: manager().publicParentFor(s))
    children = binding.Make(lambda s: s.publicChildren)

    def __repr__(self):
        return 'PublicNode(%s)' % str(self._oid)

@Workspace.get.when(
    'self in Workspace and klass in type and issubclass(klass, PublicNode)')
def get_node(self, klass, key, default=NOT_GIVEN):
    return manager().getNode(klass, key, default=default)




class LastNode(NodeVersion):

    binding.metadata(PrimaryKey('inode'))

    def new_inode(obj, attr, values):
        ws = settings.workspace()
        seq = SequenceName.of(RepositoryNode, 'inode')
        return ~ws('query.NextVal(seq)')
    inode = Attribute(int, Required, ComputeDefault(config.Value(new_inode)))

    def onChange(self, added, removed):
        assert self.changeset == MAX_CHSET
        ws = self._workspace
        if ws is None:
            return
        if self.__dict__.get('__lock__') is not None:
            return
        lock = self.__dict__['__lock__'] = manager().lockNode(self._oid)
        ws.on_change(self)

    def set_parent(self, attr, val):
        if val is None:
            self.__dict__['parent_node'] = None
        else:
            old = self.__dict__.get('parent_node')
            if old != val.inode:
                self.__dict__['parent_node'] = val.inode
        return val

    parent = binding.Make(
                lambda s: manager().lastParentFor(s), onSet = set_parent)

    children = binding.Make(lambda s: s.validChildren)

    def __repr__(self):
        return 'LastNode(%s)' % str(self._oid)


@Workspace.get.when(
    'self in Workspace and klass in type and issubclass(klass, LastNode)')
def get_node(self, klass, key, default=NOT_GIVEN):
    return manager().getNode(klass, key, default=default)

@Workspace.do_add.when('self in Workspace and obj in LastNode')
def do_add_node(self, obj):
    return manager().addLastNode(obj)

@Workspace.do_update.when('self in Workspace and obj in LastNode')
def do_update_node(self, obj):
    return manager().updateLastNode(obj)

@Workspace.do_remove.when('self in Workspace and obj in LastNode')
def do_remove_node(self, obj):
    return manager().removeLastNode(obj)









##################### RepositoryElement facade #######################


        

class ChildList(PersistentList):

    element = None

    def _p_register(self, _):
        if self.element is not None:
            self.element.onChange({}, {})

    def _p_wrap(self, v):
        if not isinstance(v, Element):
            raise AssertionError('Not element', v)
        if v.parent is None:
            v.parent = self.element
        elif v.parent is not self.element:
            v.parent.children.remove(v)
            v.parent = self.element
        return v

class Children(Feature):
    cardinality = 'list'
    refPrimitives = ()

    def get(self, obj):
        attr = self.attrName
        val = obj.__dict__.get(attr)
        if val is None:
            obj.loadDeactivatedFeature(self)
            try :
                val = obj.__dict__[attr]
            except KeyError:
                raise AttributeError(
                    'Lazy load failed', obj.__class__, attr,
                    obj.__dict__.get('_oid'), obj.__dict__.get('_workspace'),
                )
        return val

    def default(self, obj, values={}):
        return ChildList()

    def set(self, obj, val):
        attr = self.attrName
        old = obj.__dict__.get(attr)
        if isinstance(val, ChildList):
            if val.element is None:
                val.element = obj
            elif val.element is obj:
                pass
            else:
                raise AssertionError('invalid ChildList.element')
        else:
            val = ChildList(val)
            val.element = obj
        obj.onChange({}, {})
        obj.__dict__[attr] = val

    def delete(self, obj):
        self.set(obj, [])

    def afterAdd(self, obj):
        obj.root.save()

    def afterUpdate(self, obj):
        obj.root.save()
    
    def afterRemove(self, obj):
        obj.root.remove()


class BaseElement(Entity, AbstractTreeNode):
    children = Children()
    root     = property(lambda s: s.getRoot())
    inode    = property(lambda s: s.getRoot().inode)
    node     = property(lambda s: s.getRoot().node)
    partMgr  = property(lambda s: manager().partitionForInode(s.inode))

    def makeElement(self, re):
        factory = RootProperty(
            self.partMgr.partition,
            'element_factory.%s.%s' % (re.kind, re.name)
        )()
        el = Entity.__new__(factory)
        el.__dict__['_oid'] = re.eid
        el.__dict__['eid'] = re.eid
        el.__dict__['kind'] = re.kind
        el.__dict__['name'] = re.name
        el.__dict__['value'] = re.value
        el.__dict__['parent'] = self
        return el
        
    def makeElements(self, re_list):
        children = ChildList()
        children.element = self
        children._p_preload([self.makeElement(re) for re in re_list])
        return children


class Element(BaseElement):
    def new_eid(obj, attr, values):
        ws = settings.workspace()
        seq = SequenceName.of(RepositoryElement, 'eid')
        return ~ws('query.NextVal(seq)')
    eid    = Attribute(int, Required, ComputeDefault(config.Value(new_eid)))
    parent = Reference(BaseElement)
    kind   = Attribute(unicode, Required, Default('cms'))
    name   = Attribute(unicode, Required)
    value  = Attribute(unicode)

    binding.metadata(PrimaryKey('eid'))

    def position(self):
        if self.parent is not None:
            return self.parent.children.index(self)
        else:
            return -1
    position = property(position)

    def parent_id(self):
        if self.parent is not None:
            return self.parent.eid
        else:
            return None
    parent_id = property(parent_id)

    def onChange(self, added, removed):
        if self.parent is not None:
            self.parent.onChange(added, removed)

    def loadDeactivatedFeature(self, feature):
        if not isinstance(feature, Children):
            raise AssertionError(self, feature)
        elements = self.root.branches.get(self._oid,())
        self.__dict__.setdefault('children', self.makeElements(elements))

    def findElement(self, kind, name, default=NOT_GIVEN):
        for elem in self.iterDepthFirst(include_self=False):
            if elem.kind == kind and elem.name == name:
                return elem
        if default is NOT_GIVEN:
            raise KeyError('element not found', kind, name)
        return default

    def findChild(self, kind, name, default=NOT_GIVEN):
        for elem in self.children:
            if elem.kind == kind and elem.name == name:
                return elem
        if default is NOT_GIVEN:
            raise KeyError('element not found', kind, name)
        return default
        
    def __repr__(self):
        dct = self.__dict__.get
        return '%s:%s(%s)' % (dct('kind','_'), dct('name','_'), self._oid)


class Document(BaseElement):
    facade    = Attribute(object)
    node      = Reference(NodeVersion)
    branches  = Attribute(object, ReadOnly)
    elements  = Attribute(object, ReadOnly)

    binding.metadata(PrimaryKey('facade', 'node'))

    inode     = binding.Make(lambda s: s.node._oid[1])
    changeset = binding.Make(lambda s: s.node._oid[2])
    parent    = None
    root      = property(lambda s: s)
    kind      = None
    eid       = None
    value     = None

    def onChange(self, added, removed):
        assert self.changeset == MAX_CHSET, \
            'cannot edit changeset %s of document %s' % (
                self.changeset, self._oid)
        ws = self._workspace
        if ws is None:
            return
        if self.__dict__.get('__lock__') is not None:
            return
        lock = self.__dict__['__lock__'] = self.node.onChange({}, {})

    def loadDeactivatedFeature(self, feature):
        if isinstance(feature, Children):
            self.__dict__.setdefault(
                'children', self.makeElements(self.branches.get(None, ())))
        else:
            manager().preloadDocument(self)

    def __repr__(self):
        return 'document%s' % repr(self._oid)


class PublicElement(Entity):
    pass


@Workspace.get.when(
    'self in Workspace and klass in type and issubclass(klass,PublicElement)')
def get_public_element(self, klass, key, default=NOT_GIVEN):
    return manager().getPublicElement(klass, key)



class MenuItem(AbstractTreeNode):

    __metaclass__ = binding.Activator



class Menu(AbstractTreeNode):
    
    def __init__(self, root_inode, depth=2, pagelets=['head.title']):
        self.root_inode = root_inode
        self.depth = depth


