DROP TABLE cms_image CASCADE;
DROP TABLE cms_file CASCADE;
DROP TABLE cms_published CASCADE;
DROP TABLE cms_publishing_request CASCADE;
DROP TABLE cms_metadata CASCADE;
DROP TABLE cms_object CASCADE;
DROP TABLE cms_node CASCADE;

CREATE TABLE cms_node (
    inode           SERIAL NOT NULL,
    parent          INTEGER REFERENCES cms_node(inode)
                            ON UPDATE CASCADE ON DELETE CASCADE,

    path            VARCHAR(128) NOT NULL,
    abs_path        LTREE NOT NULL,

    acl             ROWNACL NOT NULL DEFAULT ''::ROWNACL,
    abs_acl         ROWNACL NOT NULL,

    sorting         INTEGER NOT NULL DEFAULT 100,

    PRIMARY KEY (inode),
    UNIQUE (parent, path),
    CHECK (path ~* '^[_a-zA-Z0-9]{1,128}$'),
    CHECK (inode >= 0)
);
CREATE INDEX cms_node_abs_path ON cms_node USING gist (abs_path);

CREATE OR REPLACE FUNCTION cms_node_set_abs() RETURNS TRIGGER AS
'
    rec = TD["new"]
    
    if not rec["path"] : plpy.error("path is NULL")
    if rec["acl"] is None : plpy.error("acl is NULL")
    if rec["inode"] == rec["parent"] : plpy.error("parent equals to inode")
        
    chg = None
    if TD["event"] == "INSERT" :
        chg = 1
    elif TD["event"] == "UPDATE" :
        old = TD["old"]
        if old["path"] != rec["path"] :
            chg = 1
        if old["parent"] != rec["parent"] :
            chg = 1
        if old["acl"] != rec["acl"] :
            chg = 1
    if not chg :
        return "OK"

    if rec["parent"] :
        q = """SELECT abs_path, join(abs_acl,''%s'') AS abs_acl
                 FROM cms_node
                WHERE inode=%d""" % (rec["acl"], rec["parent"])
        pres = plpy.execute(q, 1)
        if len(pres) != 1 :
            plpy.error("parent with inode %s not found" % rec["parent"])
        parent = pres[0]
        abs_path = "%s.%s" % (parent["abs_path"], rec["path"])
        abs_acl = parent["abs_acl"]
        if TD["event"]=="UPDATE" :
            if parent["abs_path"].startswith(old["abs_path"]) :
                plpy.error("cannot move node inside its subtree")
    else :
        parent = None
        abs_path = rec["path"]
        abs_acl = rec["acl"]

    rec["abs_path"] = abs_path
    rec["abs_acl"] = abs_acl

    return "MODIFY"
' LANGUAGE plpythonu;
    
CREATE TRIGGER "10_cms_node_set_abs"
    BEFORE INSERT OR UPDATE ON cms_node FOR EACH ROW
    EXECUTE PROCEDURE cms_node_set_abs();

CREATE OR REPLACE FUNCTION cms_node_update_subtree() RETURNS TRIGGER AS
'
    old, new = TD["old"], TD["new"]
    ret = "OK"

    if old["path"] != new["path"] or old["parent"] != new["parent"] :
        l = old["abs_path"].count(".") + 1
        q = ("UPDATE cms_node "
             "SET abs_path = ''%s''::ltree || subpath(abs_path,%d) "
             "WHERE abs_path <@ ''%s''" ) % (
                new["abs_path"], l, old["abs_path"]
            )
        plpy.execute(q)
        ret = "MODIFY"

    if old["acl"] != new["acl"] or old["parent"] != new["parent"] :
        q = """SELECT MAX(NLEVEL(abs_path)) AS lev_max
                 FROM cms_node
                WHERE abs_path <@ ''%s''""" % new["abs_path"]
        lev_max = plpy.execute(q)[0]["lev_max"]
        lev = new["abs_path"].count(".") + 2
        q = """
          UPDATE cms_node
             SET abs_acl = v_abs_acl
            FROM (
                SELECT cp.inode, cc.inode, join(cp.abs_acl, cc.acl)
                  FROM cms_node cc
                  LEFT JOIN cms_node cp ON (cc.parent = cp.inode)
            ) AS cms_acl_view (v_parent,v_inode, v_abs_acl)
           WHERE inode = v_inode AND abs_path <@ ''%s''
             AND nlevel(abs_path) = %%d;""" % new["abs_path"]
        for l in xrange(lev, lev_max + 1) :
            res = plpy.execute(q % l)
        ret = "MODIFY"

    return ret 
' LANGUAGE plpythonu;

CREATE TRIGGER "10_cms_node_update_subtree"
    AFTER UPDATE ON cms_node FOR EACH ROW
    EXECUTE PROCEDURE cms_node_update_subtree();

















CREATE TABLE cms_object (
    node            INTEGER NOT NULL REFERENCES cms_node(inode)
                    ON UPDATE CASCADE ON DELETE CASCADE,
    version         TIMESTAMP(0) WITH TIME ZONE NOT NULL DEFAULT NOW(),

    title           VARCHAR NOT NULL,
    abstract        VARCHAR,

    toplevel        VARCHAR NOT NULL,
    -- content         BYTEA NOT NULL,
    rest            VARCHAR NOT NULL,

    owner           INTEGER NOT NULL REFERENCES "user"(uid)
                    ON UPDATE CASCADE ON DELETE RESTRICT,
    PRIMARY KEY (node, version)
);

CREATE OR REPLACE FUNCTION cms_object_history_update_before()
RETURNS TRIGGER LANGUAGE plpythonu AS '
    oldRec, newRec = TD["old"], TD["new"]
    # some columns are immutable
    for f in ("node", "version"):
        if oldRec[f] != newRec[f]:
            plpy.error("column %r is immutable" % f)
    # only last version can be updated
    last = plpy.execute(
        "SELECT MAX(version) FROM cms_object WHERE node=''%s''" % oldRec["node"]
    )
    if len(last) != 1:
        plpy.error("internal error: cannot find last record (%d)" % len(last))
    last = last[0]
    if last["max"] != newRec["version"]:
        plpy.error("only last version can be updated")
    now = plpy.execute("SELECT NOW() AS NOW")[0]["now"]
    newRec["version"] = now
    return "MODIFY"
';

CREATE TRIGGER "10_cms_object_history" 
    BEFORE UPDATE ON cms_object FOR EACH ROW
    EXECUTE PROCEDURE cms_object_history_update_before();

CREATE OR REPLACE FUNCTION cms_object_history_update_after()
RETURNS TRIGGER LANGUAGE plpythonu AS '
    import pgspi
    oldRec = TD["old"]
    planId = "insert_plan_for_%s" % TD["relid"]
    if SD.has_key(planId) :
        plan = SD[planId]
    else :
        SD[planId] = plan = pgspi.InsertPlan(TD["relid"])
    plan.execute(oldRec)
';

CREATE TRIGGER "20_cms_object_history" 
    AFTER UPDATE ON cms_object FOR EACH ROW
    EXECUTE PROCEDURE cms_object_history_update_after();

CREATE VIEW cms_object_last AS
    SELECT DISTINCT ON (node) * FROM cms_object ORDER BY node, version DESC;
















CREATE TABLE cms_metadata (
    propid          SERIAL NOT NULL,
    node            INTEGER NOT NULL REFERENCES cms_node (inode)
                    ON UPDATE CASCADE ON DELETE CASCADE,
    published       VALIDITY_DATE NOT NULL
                    DEFAULT VALIDITY_DATE_IN('from today till infinity'),

    property        VARCHAR NOT NULL,
    value           VARCHAR,

    mandatory       BOOLEAN NOT NULL DEFAULT 'f',

    owner           INTEGER NOT NULL REFERENCES "user"(uid)
                    ON UPDATE CASCADE ON DELETE RESTRICT,
    created         TIMESTAMP(0) WITH TIME ZONE NOT NULL DEFAULT NOW(),

    PRIMARY KEY (propid)
);

















CREATE TABLE cms_publishing_request (
    reqid           SERIAL NOT NULL,

    node            INTEGER NOT NULL REFERENCES cms_node(inode)
                    ON UPDATE CASCADE ON DELETE CASCADE,
    version         TIMESTAMP(0) WITH TIME ZONE NOT NULL,

    owner           INTEGER NOT NULL REFERENCES "user"(uid)
                    ON UPDATE CASCADE ON DELETE RESTRICT,
    created         TIMESTAMP(0) WITH TIME ZONE NOT NULL DEFAULT NOW(),

    publish         VALIDITY_DATE NOT NULL
                    DEFAULT VALIDITY_DATE_IN('from today till infinity'),

    result          VARCHAR,
    processed       TIMESTAMP(0) WITH TIME ZONE,
    processedBy     INTEGER REFERENCES "user"(uid)
                    ON UPDATE CASCADE ON DELETE RESTRICT,

    PRIMARY KEY(reqid)
);


CREATE TABLE cms_published (
    node            INTEGER NOT NULL REFERENCES cms_node(inode)
                    ON UPDATE CASCADE ON DELETE CASCADE,
    published       VALIDITY_DATE NOT NULL,
    version         TIMESTAMP(0) WITH TIME ZONE NOT NULL,
    owner           INTEGER NOT NULL REFERENCES "user"(uid)
                    ON UPDATE CASCADE ON DELETE RESTRICT,

    PRIMARY KEY (node, published)
);
    


    




-- Obrazky a soubory pro ruzne ucely
CREATE TABLE cms_file(
    "pkey"        SERIAL NOT NULL,
    "node"        INTEGER NOT NULL REFERENCES cms_node(inode)
                  ON UPDATE CASCADE ON DELETE CASCADE,
    "description" VARCHAR NOT NULL,
    "lang"        VARCHAR NOT NULL DEFAULT 'en',
    "name"        VARCHAR NOT NULL,
    "blob"        OID NOT NULL,
    "size"        INTEGER NOT NULL,
    "mime"        VARCHAR NOT NULL,
    "encoding"    VARCHAR,
    "sorting"     INTEGER NOT NULL DEFAULT 100,

    PRIMARY KEY("pkey")
);
CREATE TRIGGER "cms_file_blob"
  BEFORE UPDATE OR DELETE ON cms_file FOR EACH ROW
  EXECUTE PROCEDURE blob_trigger('blob');


CREATE TABLE cms_image(
    "pkey"        SERIAL NOT NULL,
    "node"        INTEGER NOT NULL REFERENCES cms_node(inode)
                  ON UPDATE CASCADE ON DELETE CASCADE,
    "description" VARCHAR NOT NULL,
    "lang"        VARCHAR NOT NULL DEFAULT 'en',
    "name"        VARCHAR NOT NULL,
    "blob"        OID NOT NULL,
    "size"        INTEGER NOT NULL,
    "width"       INTEGER NOT NULL,
    "height"      INTEGER NOT NULL,
    "mime"        VARCHAR NOT NULL,
    "encoding"    VARCHAR,

    PRIMARY KEY("pkey")
);
CREATE TRIGGER "cms_image_blob"
  BEFORE UPDATE OR DELETE ON cms_image FOR EACH ROW
  EXECUTE PROCEDURE blob_trigger('blob');







