# -*- coding: ISO-8859-2 -*-

from unittest import *
from wm4.api import *
from datetime import datetime
from lib4.timeutils import *

from wm4.schema.tests.utils import MockWorkspaceCase
from comp4.cms.api import *


class MyTimeService(TimeService):
    
    def __init__(self):
        self.stamp = datetime(2006,1,1,10,15,0, tzinfo=TZUTC)

    def now(self):
        s = self.stamp
        self.stamp = s + DAY
        return s



class RepositoryCase(MockWorkspaceCase):
    
    def setUp(self):
        MockWorkspaceCase.setUp(self)
        settings.time_service(MyTimeService)
        self.mgr = manager()





class TestNodeFacade(RepositoryCase):

    def test10_Get(self):
        self.expects(transaction='begin')
        self.expects(sql=(
            "SELECT partition, inode, changeset_from, changeset_till, "
            "klass, parent_node FROM repository_node "
            "WHERE (inode = 13) "
            "AND (changeset_from <= 42) AND (42 <= changeset_till)"),
            res=[['P1', 13, 8, MAX_CHSET, 'test3', None]],
        )
        self.expects(transaction='commit')

        storage.begin()
        node = self.mgr['P1'].get(13, 42)
        self.assertEqual(node.__class__, Node)
        self.assertEqual(node._oid, (13,42))
        self.assertEqual(node._workspace, self.ws)
        self.assertEqual(node.current._oid, (13, 8))
        storage.commit()
        
    def test20_Add(self):
        self.expects(transaction='begin')
        self.expects(
            sql="SELECT NEXTVAL('repository_node_inode_seq')",
            res=[[13]],
        )
        self.expects(
            sql="SELECT NEXTVAL('repository_changeset_pkey_seq')",
            res=[[6]],
        )
        self.expects(sql=(
          "INSERT INTO repository_changeset (pkey, created_by, creation_time) "
          "VALUES (6, NULL, '2006-01-01 10:15:00+00:00')"
        ))
        self.expects(sql=(
          "INSERT INTO repository_node ("
          "partition, inode, changeset_from, changeset_till, klass, parent_node"
          ") VALUES ("
          "'P1', 13, 6, 2147483647, 'test1', NULL"
          ")"
        ))
        self.expects(sql=(
          "SELECT partition, inode, changeset_from, changeset_till, "
          "klass, parent_node FROM repository_node "
          "WHERE (inode = 13) AND (changeset_from <= 6) "
          "AND (6 <= changeset_till)"),
          res=[['P1', 13, 8, MAX_CHSET, 'test3', None]]
        )
        self.expects(transaction='commit')

        storage.begin()
        node = self.ws.add(Node(partition='P1',klass='test1'))
        self.failUnless(isinstance(node, Node))
        self.assertEqual(node.inode, 13)
        oid = (13, MAX_CHSET)
        self.assertEqual(node._oid, oid)
        self.assertEqual(node._workspace, self.ws)
        self.assertEqual(node.klass, 'test1')
        self.failUnless(oid in self.ws.contexts[Node].cache)
        storage.commit()
        self.failUnless(oid in self.ws.contexts[Node].cache)

    def test30_Preload(self):
        self.expects(transaction='begin')
        self.expects(sql=(
            "SELECT partition, inode, changeset_from, changeset_till, "
                   "klass, parent_node "
             "FROM repository_node WHERE inode = 13 ORDER BY changeset_from"),
            res=[
                ['P1', 13, 3, 6,         'test1', None],
                ['P1', 13, 7, 7,         'test2', None],
                ['P1', 13, 8, MAX_CHSET, 'test3', None],
            ],
        )
        self.expects(transaction='commit')

        oid = (13, MAX_CHSET)
        node = object.__new__(Node)
        node.__dict__['_oid'] = oid

        storage.begin()
        n = self.ws.preload(node)
        self.failUnless(n is node)
        self.assertEqual(node.partition, 'P1')
        self.assertEqual(node.inode, 13)
        self.assertEqual(node._oid, oid)
        self.assertEqual(node._workspace, self.ws)
        self.assertEqual(node.klass, 'test3')

        self.assertEqual(len(node.versions), 3)

        self.failUnless(isinstance(node.last, RepositoryNode))
        self.assertEqual(node.last.changeset_from, 8)
        self.assertEqual(node.last.changeset_till, MAX_CHSET)

        self.failUnless(
            node.current is node.last, '%s x %s' % (node.current, node.last))

        self.failUnless(oid in self.ws.contexts[Node].cache)
        self.failUnless((13,3) in self.ws.contexts[RepositoryNode].cache)
        self.failUnless((13,7) in self.ws.contexts[RepositoryNode].cache)
        self.failUnless((13,8) in self.ws.contexts[RepositoryNode].cache)

        storage.commit()

        self.failUnless(oid in self.ws.contexts[Node].cache)
        self.failIf((13,3) in self.ws.contexts[RepositoryNode].cache)
        self.failIf((13,7) in self.ws.contexts[RepositoryNode].cache)
        self.failIf((13,8) in self.ws.contexts[RepositoryNode].cache)

    def test31_PreloadDeleted(self):
        self.expects(transaction='begin')
        self.expects(sql=(
            "SELECT partition, inode, changeset_from, changeset_till, klass, "
            "parent_node FROM repository_node WHERE inode = 13 "
            "ORDER BY changeset_from"),
            res=[
                ['P1', 13, 3, 6, 'test1', None],
                ['P1', 13, 7, 7, 'test2', None],
                ['P1', 13, 8, 9, 'test3', None],
            ],
        )
        self.expects(transaction='commit')

        oid = (13, 9)
        node = object.__new__(Node)
        node.__dict__['_oid'] = oid

        storage.begin()
        n = self.ws.preload(node)
        self.failUnless(n is node)
        self.assertEqual(node.changeset, 9)
        self.assertEqual(node.inode, 13)
        self.assertEqual(node._oid, oid)
        self.assertEqual(node._workspace, self.ws)
        self.assertEqual(node.klass, 'test3')
        self.failUnless(oid in self.ws.contexts[Node].cache)

        self.failUnless(isinstance(node.last, RepositoryNode))
        self.assertEqual(node.last.changeset_from, 8)
        self.assertEqual(node.last.changeset_till, 9)

        self.assertEqual(node.isDeleted, True)

        self.assertEqual(len(node.versions), 3)

        storage.commit()


    def test50_Update(self):
        self.expects(transaction='begin')
        self.expects(sql=(
          "SELECT partition, inode, changeset_from, changeset_till, "
          "klass, parent_node "
          "FROM repository_node WHERE inode = 13 ORDER BY changeset_from"),
          res=[
                ['P1', 13, 3, 6,         'test1', None],
                ['P1', 13, 7, MAX_CHSET, 'test2', None],
          ],
        )
        self.expects(
            sql="SELECT NEXTVAL('repository_changeset_pkey_seq')",
            res=[[8]],
        )
        self.expects(sql=(
          "INSERT INTO repository_changeset (pkey, created_by, creation_time) "
          "VALUES (8, NULL, '2006-01-01 10:15:00+00:00')"
        ))
        self.expects(
          sql=(
            "SELECT partition, inode, changeset_from, changeset_till, "
            "klass, parent_node "
            "FROM repository_node WHERE (changeset_from = 7) AND (inode = 13) "
            "FOR UPDATE"),
          res=[['P1', 13, 7, MAX_CHSET, 'test2', None]],
        )
        self.expects(sql=(
            "UPDATE repository_node SET changeset_till=7 "
            "WHERE (changeset_from = 7) AND (inode = 13)"),
        )
        self.expects(sql=(
          "INSERT INTO repository_node "
          "(partition, inode, changeset_from, changeset_till, klass, "
          "parent_node) VALUES "
          "('P1', 13, 8, 2147483647, 'update1', NULL)"
        ))
        self.expects(sql=(
          "SELECT partition, inode, changeset_from, changeset_till, "
          "klass, parent_node FROM repository_node "
          "WHERE (inode = 13) AND (changeset_from <= 8) "
          "AND (8 <= changeset_till)"),
          res=[['P1', 13, 8, MAX_CHSET, 'test3', None]]
        )
        self.expects(sql=(
          "SELECT partition, inode, changeset_from, changeset_till, "
          "klass, parent_node FROM repository_node "
          "WHERE (inode = 13) AND (changeset_from <= 8) "
          "AND (8 <= changeset_till)"),
          res=[['P1', 13, 8, MAX_CHSET, 'test3', None]]
        )
        self.expects(transaction='commit')

        oid = (13, MAX_CHSET)
        node = object.__new__(Node)
        node.__dict__['_oid'] = oid

        storage.begin()
        self.ws.preload(node)
        node.klass = 'update'
        node.klass = 'update1'
        self.assertEqual(node._workspace, self.ws)
        storage.commit()

    def XXXtest51_UpdateNewlyAdded(self):
        self.expects(transaction='begin')
        self.expects(
            sql="SELECT NEXTVAL('repository_node_inode_seq')",
            res=[[13]],
        )
        self.expects(
            sql="SELECT NEXTVAL('repository_changeset_pkey_seq')",
            res=[[6]],
        )
        self.expects(sql=(
          "INSERT INTO repository_changeset (pkey, created_by, creation_time) "
          "VALUES (6, NULL, '2006-01-01 10:15:00+00:00')"
        ))
        self.expects(sql=(
          "INSERT INTO repository_node (partition, inode, changeset_from, "
          "changeset_till, klass, parent_node) VALUES ('P1', 13, 6, "
          "2147483647, 'test1', NULL)"
        ))
        self.expects(sql=(
          "SELECT partition, inode, changeset_from, changeset_till, klass, "
          "parent_node FROM repository_node WHERE inode = 13 "
          "ORDER BY changeset_from"),
          res=[['P1', 13, 6, MAX_CHSET, 'test1', None]],
        )
        self.expects(
          sql=(
            "SELECT partition, inode, changeset_from, changeset_till, klass, "
            "parent_node FROM repository_node WHERE (changeset_from = 6) "
            "AND (inode = 13) FOR UPDATE"),
          res=[['P1', 13, 6, MAX_CHSET, 'test1', None]],
        )
        self.expects(
          sql=(
            "SELECT partition, inode, changeset_from, changeset_till, "
            "klass, parent_node FROM repository_node WHERE (inode = 13) "
            "AND (changeset_from <= 6) AND (6 <= changeset_till)"),
          res=[['P1', 13, 6, MAX_CHSET, 'test1', None]],
        )
        # XXX proc dvakrat?
        self.expects(
          sql=(
            "SELECT partition, inode, changeset_from, changeset_till, "
            "klass, parent_node FROM repository_node WHERE (inode = 13) "
            "AND (changeset_from <= 6) AND (6 <= changeset_till)"),
          res=[['P1', 13, 6, MAX_CHSET, 'test1', None]],
        )
        self.expects(sql=(
            "UPDATE repository_node SET klass='update1' "
            "WHERE (changeset_from = 6) AND (inode = 13)"),
        )
        self.expects(transaction='commit')
        storage.begin()
        node = self.ws.add(Node(partition='P1', klass='test1'))

        node.klass = 'update1'

        storage.commit()

    def test60_Remove(self):
        self.expects(transaction='begin')
        self.expects(sql=(
          "SELECT partition, inode, changeset_from, changeset_till, klass, "
          "parent_node FROM repository_node WHERE inode = 13 "
          "ORDER BY changeset_from"),
          res=[
                ['P1', 13, 3, 6,         'test1', None],
                ['P1', 13, 7, MAX_CHSET, 'test2', None],
          ],
        )
        self.expects(sql=(
            "SELECT inode, eid, changeset_from, changeset_till, kind, name, "
            "value, parent_id, position FROM repository_element "
            "WHERE (inode = 13) AND (changeset_from <= 2147483647) "
            "AND (2147483647 <= changeset_till) ORDER BY position"),
            res=[]
        )
        self.expects(sql=(
            "SELECT partition, inode, changeset_from, changeset_till, "
            "klass, parent_node FROM repository_node "
            "WHERE (partition = 'P1') AND (parent_node = 13) "
            "AND (changeset_from <= 2147483647) AND "
            "(2147483647 <= changeset_till)"),
            res=[]
        )
        self.expects(
            sql="SELECT NEXTVAL('repository_changeset_pkey_seq')",
            res=[[9]],
        )
        self.expects(sql=(
          "INSERT INTO repository_changeset (pkey, created_by, creation_time) "
          "VALUES (9, NULL, '2006-01-01 10:15:00+00:00')"
        ))
        self.expects(
          sql=(
            "SELECT partition, inode, changeset_from, changeset_till, klass, "
            "parent_node FROM repository_node WHERE (changeset_from = 7) "
            "AND (inode = 13) FOR UPDATE"),
          res=[['P1', 13, 7, MAX_CHSET, 'test2', None]],
        )
        self.expects(sql=(
            "UPDATE repository_node SET changeset_till=8 "
            "WHERE (changeset_from = 7) AND (inode = 13)"),
        )
        self.expects(sql=(
          "SELECT partition, inode, changeset_from, changeset_till, "
          "klass, parent_node FROM repository_node "
          "WHERE (inode = 13) AND (changeset_from <= 9) "
          "AND (9 <= changeset_till)"),
          res=[['P1', 13, 8, MAX_CHSET, 'test3', None]]
        )
        self.expects(transaction='commit')

        oid = (13, MAX_CHSET)
        node = object.__new__(Node)
        node.__dict__['_oid'] = oid

        storage.begin()
        self.ws.preload(node)
        self.ws.remove(node)
        self.assertEqual(node._workspace, None)
        storage.commit()

    def test61_RemoveNewlyAdded(self):
        self.expects(transaction='begin')
        self.expects(
            sql="SELECT NEXTVAL('repository_node_inode_seq')",
            res=[[13]],
        )
        self.expects(
            sql="SELECT NEXTVAL('repository_changeset_pkey_seq')",
            res=[[6]],
        )
        self.expects(sql=(
          "INSERT INTO repository_changeset (pkey, created_by, creation_time) "
          "VALUES (6, NULL, '2006-01-01 10:15:00+00:00')"
        ))
        self.expects(sql=(
          "INSERT INTO repository_node (partition, inode, changeset_from, "
          "changeset_till, klass, parent_node) VALUES "
          "('P1', 13, 6, 2147483647, 'test1', NULL)"
        ))
        self.expects(sql=(
            "SELECT inode, eid, changeset_from, changeset_till, kind, name, "
            "value, parent_id, position FROM repository_element "
            "WHERE (inode = 13) AND (changeset_from <= 2147483647) "
            "AND (2147483647 <= changeset_till) ORDER BY position"),
            res=[]
        )
        self.expects(sql=(
            "SELECT partition, inode, changeset_from, changeset_till, "
            "klass, parent_node FROM repository_node "
            "WHERE (partition = 'P1') AND (parent_node = 13) "
            "AND (changeset_from <= 2147483647) AND "
            "(2147483647 <= changeset_till)"),
            res=[]
        )
        self.expects(sql=(
          "SELECT partition, inode, changeset_from, changeset_till, klass, "
          "parent_node FROM repository_node WHERE inode = 13 "
          "ORDER BY changeset_from"),
          res=[['P1', 13, 6, MAX_CHSET, 'test1', None]],
        )
        self.expects(sql=(
            "DELETE FROM repository_node "
            "WHERE (changeset_from = 6) AND (inode = 13)"),
        )
        self.expects(sql=(
          "SELECT partition, inode, changeset_from, changeset_till, "
          "klass, parent_node FROM repository_node "
          "WHERE (inode = 13) AND (changeset_from <= 6) "
          "AND (6 <= changeset_till)"),
          res=[['P1', 13, 8, MAX_CHSET, 'test3', None]]
        )
        self.expects(transaction='commit')

        storage.begin()
        node = self.ws.add(Node(partition='P1', klass='test1'))
        self.ws.remove(node)
        self.assertEqual(node._workspace, None)
        storage.commit()














class TestNodeTree(RepositoryCase):

    def test10_NodeParent(self):
        self.expects(transaction='begin')
        self.expects(sql=(
            "SELECT partition, inode, changeset_from, changeset_till, "
            "klass, parent_node FROM repository_node "
            "WHERE (inode = 13) "
            "AND (changeset_from <= 42) AND (42 <= changeset_till)"),
            res=[['P1', 13, 8, MAX_CHSET, 'test', 7]],
        )
        self.expects(sql=(
            "SELECT partition, inode, changeset_from, changeset_till, "
            "klass, parent_node FROM repository_node "
            "WHERE (partition = 'P1') AND (inode = 7) "
            "AND (changeset_from <= 42) AND (42 <= changeset_till)"),
            res=[['P1', 7, 2, MAX_CHSET, 'test', 2]],
            )
        self.expects(sql=(
            "SELECT partition, inode, changeset_from, changeset_till, "
            "klass, parent_node FROM repository_node "
            "WHERE (partition = 'P1') AND (inode = 2) "
            "AND (changeset_from <= 42) AND (42 <= changeset_till)"),
            res=[['P1', 2, 1, MAX_CHSET, 'test', None]],
            )
        self.expects(transaction='commit')

        storage.begin()
        node = self.mgr['P1'].get(13, 42)
        parent1 = node.getParent()
        self.assertEqual(parent1.__class__, Node)
        self.assertEqual(parent1._oid, (7,42))
        self.assertEqual(parent1.current._oid, (7, 2))
        parent2 = parent1.getParent()
        self.assertEqual(parent2.__class__, Node)
        self.assertEqual(parent2._oid, (2,42))
        self.assertEqual(parent2.current._oid, (2, 1))
        self.assertEqual(parent2.getParent(), None)
        self.assertEqual(
            [n._oid for n in node.iterParents()],
            [(13,42), (7,42), (2,42)],
        )
        self.assertEqual([n.inode for n in node.iterParentsRev()], [2, 7, 13])
        storage.commit()

    def test20_NodeChildren(self):
        self.expects(transaction='begin')
        self.expects(sql=(
            "SELECT partition, inode, changeset_from, changeset_till, "
            "klass, parent_node FROM repository_node "
            "WHERE (inode = 13) "
            "AND (changeset_from <= 42) AND (42 <= changeset_till)"),
            res=[['P1', 13, 8, MAX_CHSET, 'test', 7]],
        )
        self.expects(sql=(
            "SELECT partition, inode, changeset_from, changeset_till, "
            "klass, parent_node FROM repository_node "
            "WHERE (partition = 'P1') AND (parent_node = 13) "
            "AND (changeset_from <= 42) AND (42 <= changeset_till)"),
            res=[
                ['P1', 70, 20, MAX_CHSET, 'test', 13],
                ['P1', 71, 21, MAX_CHSET, 'test', 13],
            ],
        )
        self.expects(transaction='commit')

        storage.begin()
        node = self.mgr['P1'].get(13, 42)
        children = node.getChildren()
        self.assertEqual(len(children), 2)
        ch0, ch1 = children
        self.assertEqual(ch0.__class__, Node)
        self.assertEqual(ch0._oid, (70,42))
        self.assertEqual(ch0.current._oid, (70, 20))
        self.assertEqual(ch0.getParent(), node)
        self.assertEqual(ch1.__class__, Node)
        self.assertEqual(ch1._oid, (71,42))
        self.assertEqual(ch1.current._oid, (71, 21))
        self.assertEqual(ch1.getParent(), node)
        storage.commit()

    def test30_IterDepthFirst(self):
        self.expects(transaction='begin')
        self.expects(sql=(
            "SELECT partition, inode, changeset_from, changeset_till, "
            "klass, parent_node FROM repository_node "
            "WHERE (partition = 'P1') AND (parent_node ISNULL) "
            "AND (changeset_from <= 42) AND (42 <= changeset_till)"),
            res=[['P1', 1, 100, MAX_CHSET, 'test', None]],
        )
        self.expects(sql=(
            "SELECT partition, inode, changeset_from, changeset_till, "
            "klass, parent_node FROM repository_node "
            "WHERE (partition = 'P1') AND (parent_node = 1) "
            "AND (changeset_from <= 42) AND (42 <= changeset_till)"),
            res=[
                ['P1', 11, 100, MAX_CHSET, 'test', 1],
                ['P1', 12, 100, MAX_CHSET, 'test', 1],
            ],
        )
        self.expects(sql=(
            "SELECT partition, inode, changeset_from, changeset_till, "
            "klass, parent_node FROM repository_node "
            "WHERE (partition = 'P1') AND (parent_node = 11) "
            "AND (changeset_from <= 42) AND (42 <= changeset_till)"),
            res=[],
        )
        self.expects(sql=(
            "SELECT partition, inode, changeset_from, changeset_till, "
            "klass, parent_node FROM repository_node "
            "WHERE (partition = 'P1') AND (parent_node = 12) "
            "AND (changeset_from <= 42) AND (42 <= changeset_till)"),
            res=[
                ['P1', 121, 100, MAX_CHSET, 'test', 12],
            ],
        )
        self.expects(sql=(
            "SELECT partition, inode, changeset_from, changeset_till, "
            "klass, parent_node FROM repository_node "
            "WHERE (partition = 'P1') AND (parent_node = 121) "
            "AND (changeset_from <= 42) AND (42 <= changeset_till)"),
            res=[],
        )
        self.expects(transaction='commit')

        storage.begin()
        root = self.mgr['P1'].getRoot(changeset=42)
        children = list(root.iterDepthFirst())
        self.assertEqual(
            [c.inode for c in children],
            [1, 11, 12, 121],
        )
        for c in children:
            self.assertEqual(c.__class__, Node)
            if c.inode > 1:
                self.assertEqual(c.getParent().inode, (c.inode / 10) or None)
        storage.commit()
        













class TestNodeElements(RepositoryCase):
    
    def test10_NodeElementsTree(self):
        self.expects(transaction='begin')
        self.expects(sql=(
            "SELECT partition, inode, changeset_from, changeset_till, "
                   "klass, parent_node "
             "FROM repository_node WHERE inode = 13 ORDER BY changeset_from"),
            res=[['P1', 13, 8, MAX_CHSET, 'test3', None]],
        )
        self.expects(sql=(
            "SELECT inode, eid, changeset_from, changeset_till, kind, "
            "name, value, parent_id, position FROM repository_element "
            "WHERE (inode = 13) AND (changeset_from <= 9) "
            "AND (9 <= changeset_till) ORDER BY position"),
            res=[
                [13, 100, 8, 10, 'xml', 'a',   'A',   None, 0],
                [13, 101, 8, 10, 'xml', 'aa',  'AA',  100,  0],
                [13, 102, 8, 10, 'xml', 'aaa', 'AAA', 101,  0],
                [13, 105, 8, 10, 'xml', 'aca', 'ACA', 104,  0],
                [13, 103, 8, 10, 'xml', 'ab',  'AB',  100,  1],
                [13, 106, 8, 10, 'xml', 'b',   'B',   None, 1],
                [13, 104, 8, 10, 'xml', 'ac',  'AC',  100,  2],
            ],
        )
        self.expects(transaction='commit')

        oid = (13, 9)
        node = object.__new__(Node)
        node.__dict__['_oid'] = oid
        storage.begin()
        self.ws.preload(node)
        
        elems = node.elements
        self.assertEqual(
            [(n._oid,n.treeLevel,n.name) for n in elems.iterDepthFirst()],
            [(None, 0, None),
                ((100, 9), 1, u'a'),
                    ((101, 9), 2, u'aa'),
                        ((102, 9), 3, u'aaa'),
                    ((103, 9), 2, u'ab'),
                    ((104, 9), 2, u'ac'),
                        ((105, 9), 3, u'aca'),
                ((106, 9), 1, u'b')
            ]
        )
        self.failUnless((103,8) in self.ws.contexts[RepositoryElement].cache)
        self.failUnless((103,9) in self.ws.contexts[Element].cache)

        self.assertEqual(elems.children[0].children[0].name, 'aa')
        self.assertEqual(elems.children[0].children[0].value, 'AA')
        self.assertEqual(elems.children[1].name, 'b')
        self.assertEqual(elems.children[1].value, 'B')

        storage.commit()
        self.failIf((103,8) in self.ws.contexts[RepositoryElement].cache)
        self.failUnless((103,9) in self.ws.contexts[Element].cache)

    def test20_ElementAdd(self):
        pass

    def test30_ElementUpdate(self):
        self.expects(transaction='begin')
        self.expects(sql=(
            "SELECT partition, inode, changeset_from, changeset_till, "
                   "klass, parent_node "
             "FROM repository_node WHERE inode = 13 ORDER BY changeset_from"),
            res=[['P1', 13, 8, MAX_CHSET, 'test3', None]],
        )
        self.expects(sql=(
            "SELECT inode, eid, changeset_from, changeset_till, kind, "
            "name, value, parent_id, position FROM repository_element "
            "WHERE (inode = 13) AND (changeset_from <= 9) "
            "AND (9 <= changeset_till) ORDER BY position"),
            res=[
                [13, 100, 8, 10, 'xml', 'a',   'A',   None, 0],
                [13, 101, 8, MAX_CHSET, 'xml', 'aa',  'AA',  100,  0],
                [13, 102, 8, 10, 'xml', 'aaa', 'AAA', 101,  0],
            ],
        )
        self.expects(
            sql="SELECT NEXTVAL('repository_changeset_pkey_seq')", res=[[12]],
        )
        self.expects(sql=(
          "INSERT INTO repository_changeset (pkey, created_by, creation_time) "
          "VALUES (12, NULL, '2006-01-01 10:15:00+00:00')"
        ))
        self.expects(sql=(
            "SELECT inode, eid, changeset_from, changeset_till, kind, name, "
            "value, parent_id, position FROM repository_element WHERE "
            "(changeset_from = 8) AND (eid = 101) FOR UPDATE"),
            res = [[13, 101, 8, MAX_CHSET, 'xml', 'aa',  'AA',  100,  0]],
        )
        self.expects(sql=(
            "UPDATE repository_element SET changeset_till=11 "
            "WHERE (changeset_from = 8) AND (eid = 101)"
        ))
        self.expects(sql=(
            "INSERT INTO repository_element (inode, eid, changeset_from, "
            "changeset_till, kind, name, value, parent_id, position) "
            "VALUES (13, 101, 12, 2147483647, 'xml', 'aa', 'BB', 100, 0)"
        ))
        self.expects(sql=(
          "SELECT partition, inode, changeset_from, changeset_till, "
          "klass, parent_node FROM repository_node "
          "WHERE (inode = 13) AND (changeset_from <= 12) "
          "AND (12 <= changeset_till)"),
          res=[['P1', 13, 8, MAX_CHSET, 'test3', None]]
        )
        # XXX proc dvakrat?
        self.expects(sql=(
          "SELECT partition, inode, changeset_from, changeset_till, "
          "klass, parent_node FROM repository_node "
          "WHERE (inode = 13) AND (changeset_from <= 12) "
          "AND (12 <= changeset_till)"),
          res=[['P1', 13, 8, MAX_CHSET, 'test3', None]]
        )
        self.expects(transaction='commit')

        oid = (13, 9)
        node = object.__new__(Node)
        node.__dict__['_oid'] = oid
        storage.begin()
        self.ws.preload(node)
        elems = node.elements
        elems.children[0].children[0].value = 'BB'
        storage.commit()

    def test40_ElementDelete(self):
        self.expects(transaction='begin')
        self.expects(sql=(
            "SELECT partition, inode, changeset_from, changeset_till, "
                   "klass, parent_node "
             "FROM repository_node WHERE inode = 13 ORDER BY changeset_from"),
            res=[['P1', 13, 8, MAX_CHSET, 'test3', None]],
        )
        self.expects(sql=(
            "SELECT inode, eid, changeset_from, changeset_till, kind, "
            "name, value, parent_id, position FROM repository_element "
            "WHERE (inode = 13) AND (changeset_from <= 2147483647) "
            "AND (2147483647 <= changeset_till) ORDER BY position"),
            res=[
                [13, 100, 8, MAX_CHSET, 'xml', 'a',   'A',   None, 0],
                [13, 101, 8, MAX_CHSET, 'xml', 'aa',  'AA',  100,  0],
                [13, 102, 8, MAX_CHSET, 'xml', 'aaa', 'AAA', 101,  1],
            ],
        )
        self.expects(
            sql="SELECT NEXTVAL('repository_changeset_pkey_seq')", res=[[12]],
        )
        self.expects(sql=(
          "INSERT INTO repository_changeset (pkey, created_by, creation_time) "
          "VALUES (12, NULL, '2006-01-01 10:15:00+00:00')"
        ))
        self.expects(sql=(
            "SELECT inode, eid, changeset_from, changeset_till, kind, name, "
            "value, parent_id, position FROM repository_element WHERE "
            "(changeset_from = 8) AND (eid = 102) FOR UPDATE"),
            res = [[13, 102, 8, MAX_CHSET, 'xml', 'aaa',  'AAA',  101,  1]],
        )
        self.expects(sql=(
            "UPDATE repository_element SET changeset_till=11 "
            "WHERE (changeset_from = 8) AND (eid = 102)"
        ))
        self.expects(sql=(
            "SELECT inode, eid, changeset_from, changeset_till, kind, name, "
            "value, parent_id, position FROM repository_element WHERE "
            "(changeset_from = 8) AND (eid = 101) FOR UPDATE"),
            res = [[13, 101, 8, MAX_CHSET, 'xml', 'aa',  'AA',  100,  0]],
        )
        self.expects(sql=(
            "UPDATE repository_element SET changeset_till=11 "
            "WHERE (changeset_from = 8) AND (eid = 101)"
        ))
        self.expects(sql=(
          "SELECT partition, inode, changeset_from, changeset_till, "
          "klass, parent_node FROM repository_node "
          "WHERE (inode = 13) AND (changeset_from <= 12) "
          "AND (12 <= changeset_till)"),
          res=[['P1', 13, 8, MAX_CHSET, 'test3', None]]
        )
        self.expects(transaction='commit')

        oid = (13, MAX_CHSET)
        node = object.__new__(Node)
        node.__dict__['_oid'] = oid
        storage.begin()
        self.ws.preload(node)
        elems = node.elements
        self.ws.remove(elems.children[0].children[0])
        storage.commit()





def suite():
    return TestSuite((
        makeSuite(TestNodeFacade),
        makeSuite(TestNodeTree),
        makeSuite(TestNodeElements),
    ))


if __name__ == '__main__':
    main(defaultTest='suite')
