# -*- coding: utf-8 -*-

from datetime import date, datetime, timedelta
from dateutil.relativedelta import relativedelta
from lib4.timeutils import TZLOCAL
import calendar

from wm4.api import *
from wm4.schema.api import *

from adt4.serial import Serial
from adt4.enums import Enum, enum

__all__ = ['Event', 'EventCalendar', 
           'EventTime', 'EventTimeException', 'EventTimeReminder', 
           'EventResource']

_ = i18n.MessageIDFactory('comp4.event')

class Privacy(Enum):
    public = enum(_(u'Public'))
    private = enum(_(u'Private'))

class ReminderMethod(Enum):
    email = enum(_(u'E-mail'))
    sms = enum(_(u'SMS'))

class ReminderStatus(Enum):
    sent = enum(_(u'Sent'))
    pending = enum(_(u'Pending'))

class UserRole(Enum):
    owner =enum(_(u'Owner'))
    guest =enum(_(u'Guest'))

class Recurrence(Enum):
    none = enum(_('Does not repeat'))
    daily = enum(_('Daily'))
    weekly = enum(_('Weekly'))
    monthly_day = enum(_('Monthly by day of the month'))
    monthly_weekday = enum(_('Monthly by day of the week'))
    yearly = enum(_('Yearly'))

ordinals = [_(u'first'), _(u'second'), _(u'third'), _(u'fourth'), _(u'fifth')]

def nthday(nth, wday, month, year):
    dt = date(year, month, 1)
    lst = []
    while dt.weekday() != wday:
        dt += timedelta(days=1)
    while dt.month == month:
        lst.append(dt)
        dt += timedelta(days=7)
    try:
        return lst[nth-1]
    except IndexError:
        return lst[-1]


class EventCalendar(Entity):
    pkey = Attribute(Serial)

    name = Attribute(unicode, Required)
    description = Attribute(unicode)

    privacy = Attribute(Privacy, Required, Default(Privacy.public))
    owner = Reference('comp4.users.entities.User')

    binding.metadata(PrimaryKey('pkey'))





class Event(Entity):
    pkey = Attribute(Serial)
    parent = Reference(THIS_CLASS)
    calendar = Reference('comp4.event.entities.EventCalendar')

    name = Attribute(unicode, Required)
    description = Attribute(unicode)
    place = Attribute(unicode)
    privacy = Attribute(Privacy, Required, Default(Privacy.public))

    times = ReferenceEnd('comp4.event.entities.EventTime', 'event',
                ReferenceEndOrder('begin', 'end'))

    binding.metadata(PrimaryKey('pkey'))

@schema.Workspace.do_remove.before('obj in Event')
def remove_times(ws, obj):
    ws("[query.Delete(et) for et in 'event/EventTime' if et.event==obj]")



class EventTime(Entity):
    pkey = Attribute(Serial)
    event = Reference('comp4.event.entities.Event', Required)

    recurrence = Attribute(Recurrence, Required, Default(Recurrence.none))
    interval = Attribute(int, Required, Default(1))
    repeat_until = Attribute(datetime)

    begin = Attribute(datetime, Required)
    begin_date = binding.Make(lambda s: s.begin.date())
    begin_time = binding.Make(lambda s: s.begin.timetz())
    end = Attribute(datetime)
    end_date = binding.Make(lambda s: s.end.date())
    end_time = binding.Make(lambda s: s.end.timetz())

    description = Attribute(unicode)

    binding.metadata(PrimaryKey('pkey'))


    def recurrences(self, start_time=None, limit=None):
        '''generate begins or all recurrences of event starting at start_time
        '''
        if self.recurrence == Recurrence.none:
            return
        if start_time is None:
            start_time = datetime.now(TZLOCAL)

        def findNextRecurrence(start_time):
            '''returns next recurrence after start_time'''
            start_time = max(start_time, self.begin)
            if self.repeat_until and start_time > self.repeat_until:
                return None
            next_date = start_time.date()
            if start_time.time() > self.begin.time():
                next_date += timedelta(days=1)
            if self.recurrence in (Recurrence.daily,
                                   Recurrence.weekly):
                if self.recurrence == Recurrence.daily:
                    interval = self.interval
                else:
                    interval = self.interval * 7
                    while next_date.weekday() != self.begin_date.weekday():
                        next_date += timedelta(days=1)
                delta = (next_date - self.begin_date).days % interval
                if delta:
                    next_date += timedelta(days=interval-delta)
            elif self.recurrence == Recurrence.monthly_day: 
                if next_date.day > self.begin.day:
                    next_date += relativedelta(months=1)
                delta = ((next_date.year*12 + next_date.month) - \
                         (self.begin.year*12 + \
                            self.begin.month)) % self.interval
                if delta:
                    next_date += relativedelta(months=self.interval - delta)
                next_date = next_date.replace(
                    day=min(self.begin.day, 
                            calendar.monthrange(next_date.year, 
                                                next_date.month)[1]))
            elif self.recurrence == Recurrence.monthly_weekday:
                nth = (self.begin.day-1)/7 + 1
                if next_date > nthday(nth, self.begin.weekday(), 
                        next_date.month, next_date.year):
                    next_date += relativedelta(months=1)
                delta = ((next_date.year*12 + next_date.month) - \
                         (self.begin.year*12 + \
                            self.begin.month)) % self.interval
                if delta:
                    next_date += relativedelta(months=self.interval - delta)
                next_date = nthday(nth, self.begin.weekday(), 
                                    next_date.month, next_date.year)
            elif self.recurrence == Recurrence.yearly:
                years = next_date.year-self.begin.year
                if next_date.month == self.begin.month:
                    if next_date.day > self.begin.day:
                        years += 1
                elif next_date.month > self.begin.month:
                    years += 1
                delta = years % self.interval
                if delta:
                    years += self.interval - delta
                next_date = self.begin + relativedelta(years=years)

            next_time = datetime.combine(next_date, self.begin_time)
            if self.repeat_until and next_time > self.repeat_until:
                return None
            else:
                return next_time
        def nextRecurrence(start_time):
            '''returns next recurrence after start_time 
               (start_time is one of the recurrences)'''
            if self.recurrence == Recurrence.daily:
                next_time = start_time + timedelta(days=self.interval)
            elif self.recurrence == Recurrence.weekly:
                next_time = start_time + timedelta(days=7*self.interval)
            else:
                next_time = findNextRecurrence(
                        start_time + timedelta(seconds=1))

            if next_time and self.repeat_until and \
                            next_time > self.repeat_until:
                return None
            return next_time

        time = findNextRecurrence(start_time)
        i = 0
        while time:
            yield time
            i += 1
            if limit and i>=limit:
                break
            time = nextRecurrence(time)

    
    def showTime(self):
        req = web.request()
        def fmtTime(t, separ='', withdate=True):
            if withdate:
                s = '%s.%s.%s ' % (t.day, t.month, t.year)
            else:
                s = ''
            if t.hour<>0 or t.minute<>0:
                s +=  t.strftime('%H:%M')
            return s and separ + s or ''
        def wr(s):
            return web.render(req, s)

        if self.recurrence == Recurrence.none:
            s = fmtTime(self.begin)
            if self.begin != self.end:
                s += fmtTime(self.end, ' - ', 
                             self.begin.date() != self.end.date())
        else:
            s = wr(_(u'Repeats')) + ' '
            if self.recurrence == Recurrence.daily:
                if self.interval == 1:
                    s += wr(_(u'daily'))
                else:
                    s += wr(_(u'every %s days')) % self.interval
            elif self.recurrence == Recurrence.weekly:
                if self.interval == 1:
                    s += wr(_(u'weekly'))
                else:
                    s += wr(_(u'every %s weeks')) % self.interval
                s += ' ' + wr(_(u'on')) + ' ' + wr(_(self.begin.strftime('%A')))
            elif self.recurrence in (Recurrence.monthly_day, 
                                     Recurrence.monthly_weekday):
                if self.interval == 1:
                    s += wr(_(u'monthly'))
                else:
                    s += wr(_(u'every %s months')) % self.interval
                if self.recurrence == Recurrence.monthly_day:
                    s += ' ' + wr(_(u'on day')) + ' ' + str(self.begin.day)
                else:
                    s += ' ' + wr(_(u'on')) + ' ' 
                    s += wr(ordinals[(self.begin.day-1)/7]) 
                    s += ' ' + wr(_(self.begin.strftime('%A')))
            elif self.recurrence == Recurrence.yearly:
                if self.interval == 1:
                    s += wr(_(u'annually'))
                else:
                    s += wr(_(u'every %s years')) % self.interval
                s += ' ' + wr(_(u'on')) + ' ' + self.begin.strftime('%d.%m.')
            s += fmtTime(self.begin, wr(' ' + _(u'from')) + ' ')
            if self.repeat_until:
                s += fmtTime(self.repeat_until, 
                             wr(' ' + _(u'until')) + ' ')
        return s
    showTime = binding.Make(showTime)



    @schema.setattrs.after('"begin" in added')
    def set_end(self, added, removed):
        if 'end' in added and not added['end']:
            self.end = added['begin']

class EventTimeCache(Entity):
    pkey = Attribute(Serial)
    event_time = Reference('comp4.event.entities.EventTime', Required)

    begin = Attribute(datetime, Required)
    end = Attribute(datetime)

    binding.metadata(PrimaryKey('pkey'))

class EventTimeException(Entity):
    pkey = Attribute(Serial)
    event_time = Reference('comp4.event.entities.EventTime', Required)

    time = Attribute(datetime, Required)
    description = Attribute(unicode)

    binding.metadata(PrimaryKey('pkey'))

class EventTimeReminder(Entity):
    pkey = Attribute(Serial)
    event_time = Reference('comp4.event.entities.EventTime', Required)

    method = Attribute(ReminderMethod, Required)
    address = Attribute(unicode, Required)
    time = Attribute(datetime, Required)
    status = Attribute(ReminderStatus, Required)

    text = Attribute(unicode)

    binding.metadata(PrimaryKey('pkey'))




class EventResource(Entity):
    pkey = Attribute(Serial)
    event = Reference('comp4.event.entities.Event', Required)
    klass = Attribute(str)
    ident = Attribute(str)

    role = Attribute(UserRole)
    description = Attribute(unicode)

    binding.metadata(PrimaryKey('pkey'))







class EventComponent(schema.SchemaComponent):

    def eventsNonPeriodic(self, **kw):
        ws = settings.workspace()
        time_from = kw.get('time_from', datetime.now())
        q = ws.parse('''[
            (event, event_time)
                for event in 'event/Event'
                for event_time in 'event/EventTime'
                 if query.Join(event==event_time.event)
                and event_time.recurrence==Recurrence.none
                and event_time.end>=time_from
            ]''')

        conds = []

        time_to = kw.get('time_to', None)
        if isinstance(time_to, date):
            time_to += timedelta(days=1)
        elif isinstance(time_to, datetime):
            pass
        else:
            time_to = None
        if time_to:
            conds.append(q.parseSubexpr('event_time.begin<time_to'))

        if conds:
            q = q.withCondition(ws.connection, query.AND(*conds))

        if 'limit' in kw:
            q.limit = query.Const(kw['limit'])

        return ws(q) 
        
    
    def eventsPeriodic(self, time_from, time_to):
        ws = settings.workspace()
        q = ws.parse('''[
            (event, event_time)
                for event in 'event/Event'
                for event_time in 'event/EventTime'
                 if query.Join(event==event_time.event)
                and event_time.recurrence!=Recurrence.none
                and event_time.begin<time_to
                and (event_time.repeat_until is None 
                    or event_time.repeat_until>=time_from)
            ]''')
        return




