# -*- coding: iso-8859-2 -*-

from datetime import datetime, date

from wm4.api import *
from wm4.schema.api import *
from wm4.utils.files import IFileInfo, IImageInfo

from adt4.blob import Blob
from adt4.serial import Serial

import comp4.dbgettext.api as dbgettext

from lib4.units import B2KMG
from lib4.images import getFileDesc, getMIME, getStorageSize, loadImage

__all__ = ['FileBase', 'File', 'ImageBase', 'Image', 'Photo', 'PhotoRes']


class FileBase(Entity):
    
    protocols.advise(instancesProvide=[IFileInfo])

    blob     = Attribute(Blob, Required)
    size     = Attribute(int, Required)
    content_type = Attribute(str, Required, Default('application/octet-stream'))
    filename = Attribute(unicode)
    modified = Attribute(datetime, Required,
                    ComputeDefault(config.Value(lambda o,a,v:datetime.now()))
    )

    def ext(self):
        from lib4.content_types import guessExtension
        return guessExtension(self.content_type)
    ext = binding.Make(ext)

    sizeInfo = binding.Make(lambda s: B2KMG(s.size))

    downloadInfo = binding.Make(
        lambda s: u'formt: %s, velikost: %s' % (s.ext, s.sizeInof))
    
    def etag(self):
        return '"%x-%x-%s"' % (
            self.blob.oid, self.size, self.modified.isoformat()
        )
    etag = binding.Make(etag)

    undir = binding.Make(lambda s: (s.filename.split('/')[-1]).split('\\')[-1])

@Workspace.do_remove.before('obj in FileBase')
def do_remove_blobs(ws, obj):
    bdm  = settings.blob_manager()
    try:
        oid = bdm.find(obj.blob.oid)
    except KeyError:
        return
    bdm.delete(obj.blob.oid)


class File(FileBase):
    pkey = Attribute(Serial)

    binding.metadata(PrimaryKey('pkey'))

class ImageBase(FileBase):

    protocols.advise(instancesProvide=[IImageInfo])

    width  = Attribute(int, Required)
    height = Attribute(int, Required)

class Image(ImageBase):
    pkey = Attribute(Serial)

    binding.metadata(PrimaryKey('pkey'))




class PhotoRes(ImageBase):
    photo = Reference('comp4.files.entities.Photo', Required)
    resolution = Attribute(str, Required)

    binding.metadata(PrimaryKey('photo', 'resolution'))

class Photo(Entity):
    '''photography with multiple resolutions (original, preview, thumbnail...)
    '''
    pkey = Attribute(Serial)

    name = Attribute(unicode, Required)
    _name = dbgettext.Translations('name')
    author = Attribute(unicode)
    created = Attribute(date)
    quality = Attribute(int, Required, Default(5))
    description = Attribute(unicode)
    _description = dbgettext.Translations('description', useDefault=False)

    kind = Attribute(str, Required)
    convertors = binding.Make(
            lambda s: config.Property('photo.resolutions.%s' % s.kind)())

    resolutions = ReferenceEnd('comp4.files.entities.PhotoRes', 'photo')
    resdict = binding.Make(
            lambda s: dict((r.resolution, r) for r in s.resolutions))

    orig = None

    undir = binding.Make(lambda s: (s.name.split('/')[-1]).split('\\')[-1])

    binding.metadata(schema.PrimaryKey('pkey'))

def makeResolutions(ws, obj):
    img = loadImage(obj.orig.path)
    for res, conv in obj.convertors:
        i = conv(img)
        fp = getFileDesc(i)
        fp.seek(0,0)
        ws.add(PhotoRes(
            photo = obj,
            resolution = res,
            content_type = getMIME(i),
            size = getStorageSize(i, fp),
            width = i.size[0],
            height = i.size[1],
            blob = fp,
        ))

@Workspace.do_add.before('obj in Photo')
def do_add_foto_res(ws, obj):
    makeResolutions(ws, obj)


@Workspace.do_update.before('obj in Photo')
def do_update_foto_res(ws, obj):
    if obj.orig is not None:
        for res in obj.resolutions:
            ws.remove(res)
        makeResolutions(ws, obj)

@Workspace.do_remove.before('obj in Photo')
def do_remove_foto_res(ws, obj):
    for res in obj.resolutions:
        ws.remove(res)

