BEGIN;

-- Zakladni udaje o jednom emailu
CREATE TABLE mailing_subscription(
    subid       SERIAL NOT NULL,
    email       VARCHAR NOT NULL,
    name        VARCHAR NOT NULL,
    greeting    VARCHAR,
    rtime       TIMESTAMP(0) WITH TIME ZONE NOT NULL,

    PRIMARY KEY(subid)
);

-- Definice volitelnych atributu. Pokud je multival, muze byt vic
-- hodnot v tabulce mailing_value. Pokud je typ select, je vyber
-- hodnot ulozen ve sloupci options.
CREATE TABLE mailing_attr (
    attr        NAME NOT NULL,
    prompt      VARCHAR NOT NULL,
    public      BOOLEAN NOT NULL DEFAULT FALSE,
    required    BOOLEAN NOT NULL DEFAULT FALSE,
    defval      VARCHAR,
    type        VARCHAR NOT NULL DEFAULT 'varchar',
    multival    BOOLEAN NOT NULL DEFAULT FALSE,
    options     BOOLEAN NOT NULL DEFAULT FALSE,
    position    INTEGER NOT NULL DEFAULT 10,

    CHECK(type IN ('varchar', 'integer', 'date')),
    CHECK(LENGTH(attr) > 1),
    PRIMARY KEY(attr)
);


-- Hodnoty volitelnych atributu, triggery nad mailing_attr zajisti
-- pridavani a mazani sloupcu z teto tabulky.
CREATE TABLE mailing_value(
    subid       INTEGER NOT NULL REFERENCES mailing_subscription
                ON UPDATE CASCADE ON DELETE CASCADE,

    PRIMARY KEY(subid)
);

CREATE OR REPLACE FUNCTION mailing_attr_create() RETURNS TRIGGER 
LANGUAGE plpgsql AS $$
BEGIN
    EXECUTE 'ALTER TABLE mailing_value ADD COLUMN '
            || quote_ident(NEW.attr) || ' ' || NEW.type;
    IF (NEW.options) THEN
        EXECUTE 'CREATE TABLE '
            || quote_ident('mailing_value_' || NEW.attr)
            || '(option VARCHAR, PRIMARY KEY(option))';
        EXECUTE 'ALTER TABLE mailing_value ADD CONSTRAINT mailing_value_'
            || quote_ident(NEW.attr) || '_fkey '
            || 'FOREIGN KEY (' 
            || quote_ident(NEW.attr)
            || ') REFERENCES ' || quote_ident('mailing_value_' || NEW.attr);
    END IF;
    RETURN NEW;
END$$;
CREATE TRIGGER mailing_attr_create BEFORE INSERT ON mailing_attr
FOR EACH ROW EXECUTE PROCEDURE mailing_attr_create();


CREATE OR REPLACE FUNCTION mailing_attr_drop() RETURNS TRIGGER 
LANGUAGE plpgsql AS $$
BEGIN
    EXECUTE 'ALTER TABLE mailing_value DROP COLUMN ' || quote_ident(OLD.attr);
    RETURN OLD;
END$$;
CREATE TRIGGER mailing_attr_drop BEFORE DELETE ON mailing_attr
FOR EACH ROW EXECUTE PROCEDURE mailing_attr_drop();


CREATE OR REPLACE FUNCTION mailing_attr_update() RETURNS TRIGGER 
LANGUAGE plpgsql AS $$
BEGIN
    IF (OLD.type <> NEW.type) THEN
        RAISE EXCEPTION 'cannot change attribute type';
    END IF;
    IF (OLD.multival <> NEW.multival) THEN
        RAISE EXCEPTION 'cannot change attribute multiplicity';
    END IF;
    IF (OLD.attr <> NEW.attr) THEN
        EXECUTE 'ALTER TABLE mailing_value RENAME COLUMN '
            || quote_ident(OLD.attr) || ' TO ' || quote_ident(NEW.attr);
    END IF;
    RETURN NEW;
END$$;
CREATE TRIGGER mailing_attr_update BEFORE UPDATE ON mailing_attr
FOR EACH ROW EXECUTE PROCEDURE mailing_attr_update();


-- Definice skupiny. Ve sloupci rule je podminka pro vyber clenu
-- na zaklade hodnot atributu.
CREATE TABLE mailing_group(
    gid         VARCHAR NOT NULL,
    description VARCHAR,
    active      BOOLEAN NOT NULL DEFAULT TRUE,
    rule        VARCHAR,

    PRIMARY KEY(gid)
);

-- Clenove, kteri byli do urcite skupiny pridani rucne bez ohledu
-- na automaticky vyber
CREATE TABLE mailing_group_add(
    gid         VARCHAR NOT NULL REFERENCES mailing_group
                ON UPDATE CASCADE ON DELETE CASCADE,
    subid       INTEGER NOT NULL REFERENCES mailing_subscription
                ON UPDATE CASCADE ON DELETE CASCADE,

    PRIMARY KEY(gid, subid)
);

-- Clenove, kteri byli z urcite skupiny rucne vyrazeni bez ohledu
-- na automaticky vyber
CREATE TABLE mailing_group_remove(
    gid         VARCHAR NOT NULL REFERENCES mailing_group
                ON UPDATE CASCADE ON DELETE CASCADE,
    subid       INTEGER NOT NULL REFERENCES mailing_subscription
                ON UPDATE CASCADE ON DELETE CASCADE,

    PRIMARY KEY(gid, subid)
);


-- Mail, tj. jedna obesilka na jednu skupinu. Mail muze byt v HTML nebo
-- plain textu nebo muze mit oboji a pak je Content-Type mailu
-- multipart/alternative, tj. plain text by mel zhruba obsahovat to, 
-- co obsahuje HTML.
CREATE TABLE mailing_mail (
    mailid          SERIAL NOT NULL,
    tracking        VARCHAR NOT NULL, -- sledovaci kod
    fromname        VARCHAR,
    fromemail       VARCHAR NOT NULL,
    frombounce      VARCHAR,
    subject         VARCHAR NOT NULL,
    htmlbody        VARCHAR,
    plainbody       VARCHAR,
    gid             VARCHAR NOT NULL REFERENCES mailing_group
                    ON UPDATE CASCADE ON DELETE CASCADE,
    status          VARCHAR NOT NULL DEFAULT 'new',
    count           INTEGER NOT NULL DEFAULT 0,
    sent            INTEGER NOT NULL DEFAULT 0,
    started         TIMESTAMP(0) WITH TIME ZONE,
    finished        TIMESTAMP(0) WITH TIME ZONE,

    CHECK(status IN ('new','processing', 'finished', 'canceled')),
    CHECK(htmlbody NOTNULL OR plainbody NOTNULL),
    PRIMARY KEY  (mailid)
);

-- Prilohy k mailu
CREATE TABLE mailing_mail_attachment (
    mailid      INTEGER NOT NULL REFERENCES mailing_mail
                ON UPDATE CASCADE ON DELETE CASCADE,
    position    INTEGER NOT NULL,
    inline      BOOLEAN NOT NULL DEFAULT FALSE,
    contentid   VARCHAR, -- kvuli HTML mailum
    blob        VARCHAR,

    PRIMARY KEY  (mailid, position)
);

-- Emaily, ktere byly zarazeny do konkretniho mailu rucne bez ohledu
-- na automaticky vyber
CREATE TABLE mailing_mail_add(
    mailid      INTEGER NOT NULL REFERENCES mailing_mail
                ON UPDATE CASCADE ON DELETE CASCADE,
    subid       INTEGER NOT NULL REFERENCES mailing_subscription
                ON UPDATE CASCADE ON DELETE CASCADE,

    PRIMARY KEY(mailid, subid)
);

-- Emaily, ktere byly vyrazeny z konkretniho mailu rucne bez ohledu
-- na automaticky vyber
CREATE TABLE mailing_mail_remove(
    mailid      INTEGER NOT NULL REFERENCES mailing_mail
                ON UPDATE CASCADE ON DELETE CASCADE,
    subid       INTEGER NOT NULL REFERENCES mailing_subscription
                ON UPDATE CASCADE ON DELETE CASCADE,

    PRIMARY KEY(mailid, subid)
);

-- Sablony
CREATE TABLE mailing_template (
    name            VARCHAR NOT NULL,
    fromname        VARCHAR,
    fromemail       VARCHAR NOT NULL,
    frombounce      VARCHAR,
    subject         VARCHAR NOT NULL,
    htmlbody        VARCHAR,
    plainbody       VARCHAR,

    PRIMARY KEY(name)
);

END;
