# -*- coding: utf-8 -*-

from wm4.api import *
from wm4.schema.api import *
from adt4.serial import Serial
from datetime import datetime
from lib4.imports import lazyModule, importObject
from lib4.validators import EMAIL_REGEX
from wm4.config.utils import getStreamFactory

from comp4.simpleparams.api import SingleParametersFeature

__all__ = [
    'RegisteredUser', 'RegisteredUserPreferences',
    'RegisteredEmailChange', 'RegisteredSendmail', 'Registration',
]


_ = i18n.MessageIDFactory('comp4.registration')

CheckEmail = Syntax('^%s$' % EMAIL_REGEX())

class RegisteredUser(Entity):
    protocols.advise(instancesProvide=[web.IUser])

    uid         = Attribute(Serial)
    login       = Attribute(str, Required)
    password    = Attribute(str)
    email       = Attribute(str, Required, ReadOnly, CheckEmail)
    name        = Attribute(unicode, Required)
    company     = Attribute(unicode)
    phone       = Attribute(unicode)
    mobile      = Attribute(unicode)
    skype       = Attribute(unicode)
    web         = Attribute(unicode)

    home        = Attribute(str)

    created     = Attribute(datetime, Required, defaultDatetime)
    key         = Attribute(str, Required)
    verified    = Attribute(datetime)
    active      = Attribute(bool, Required, Default(True))
    approved    = Attribute(datetime)
    
    groups = () # For IUser compatibility

    _preferences = ReferenceEnd(
        'comp4.registration.entities.RegisteredUserPreferences', 'user')

    def pref_dict(self):
        pd = {}
        for p in self._preferences:
            pd.setdefault(p.pref, []).append(p)
        return pd
    pref_dict = binding.Make(pref_dict)
    
    params = SingleParametersFeature(
                    'comp4.registration.params',
                    'comp4.registration.entities.RegisteredUserParam',
                    'user')

    binding.metadata(
        PrimaryKey('uid'),
        Unique('login'), Unique('email'), Unique('key'), Unique('home')
    )



@schema.Workspace.do_add.before('obj in RegisteredUser')
def check_unique_features(ws, obj):
    ee = schema.EntityError()
    klass = obj.__class__
    for ufeat in klass.unique_features:
        cond = dict((f, getattr(obj,f)) for f in ufeat)
        if None in cond.values():
            continue
        if -ws.find(klass, **cond) is not None:
            for f in ufeat:
                ee.add(schema.FeatureError(
                    f, _(u'Registrace s těmito údaji již existuje.')
                ))
    if ee:
        raise ee

class RegisteredUserPreferences(Entity):
    user = Reference('comp4.registration.entities.RegisteredUser')
    pkey = Attribute(Serial)
    pref = Attribute(str, Required)
    value  = Attribute(unicode)

    binding.metadata(PrimaryKey('pkey'))


class RegisteredUserParam(Entity):
    pkey = Attribute(Serial)
    param = Attribute(str, Required)
    value  = Attribute(unicode)
    user = Reference('comp4.registration.entities.RegisteredUser')

    binding.metadata(PrimaryKey('pkey'))


@schema.Workspace.do_remove.before('obj in RegisteredUser')
def do_remove_user_params(ws, obj):
    ws('[query.Delete(p) for p in RegisteredUserParam if p.user==obj]')
    ws('[query.Delete(p) for p in RegisteredUserPreferences if p.user==obj]')






class RegisteredEmailChange(Entity):
    user     = Reference('comp4.registration.entities.RegisteredUser')
    email    = Attribute(str, Required)

    created  = Attribute(datetime, Required, defaultDatetime)
    key      = Attribute(str, Required)
    verified = Attribute(datetime)

    binding.metadata(PrimaryKey('user'), Unique('key'))
    

class RegisteredSendmail(Entity):
    sent       = Attribute(datetime, Required, defaultDatetime)
    sender     = Attribute(unicode, Required)
    recipients = Attribute(unicode, Required)
    subject    = Attribute(unicode, Required)
    message    = Attribute(unicode, Required)

    binding.metadata(PrimaryKey('sent'))



class Registration(schema.SchemaComponent):

    def userByLogin(self, login):
        try:
            return ~settings.workspace()(
                "[u for u in RegisteredUser if u.login == login]"
            )
        except exceptions.TooFewResults:
            return None

    def userByEmail(self, email):
        try:
            return ~settings.workspace()(
                "[u for u in RegisteredUser if u.email == email]"
            )
        except exceptions.TooFewResults:
            return None

    def userByRegKey(self, key):
        try:
            return ~settings.workspace()(
                "[u for u in RegisteredUser if u.key == key]"
            )
        except exceptions.TooFewResults:
            return None

    def userByHome(self, home):
        try:
            return ~settings.workspace()(
                "[u for u in RegisteredUser if u.home == home]"
            )
        except exceptions.TooFewResults:
            return None

   
    key_charset = ('abcdefghijklmnopqrstuvwxyz'
                   'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
                   '0123456789')

    random = lazyModule('random')

    key_length = 32

    def genKey(self, entity, attr):
        ch = self.random.choice
        kc = self.key_charset
        ws = settings.workspace()
        while True:
            key = ''.join(ch(kc) for i in xrange(self.key_length))
            res = ws(
                "[e.%s for e in entity if e.%s == key]" % (attr,attr))
            if len(res) == 0:
                break
        return key

    def cleanUnverified(self):
        raise NotImplementedError
        
    def sendVerification(self, regUser, server, url):
        assert regUser.key
        import comp4.sendmail.api as sm
        sm.sendmail(
            'registration_verification',
             regUser = regUser,
             recipients = (regUser.name, regUser.email),
             server = server,
             url = url % dict(key=regUser.key)
        )

    def sendLostPassword(self, regUser, server):
        import comp4.sendmail.api as sm
        sm.sendmail(
            'registration_lost_password', 
             regUser = regUser,
             recipients = (regUser.name, regUser.email),
             server = server,
        )

    def sendChangeEmail(self, regUser, server, newEmail):
        import comp4.sendmail.api as sm
        sm.sendmail(
            'registration_change_email',
             regUser = regUser,
             recipients = (regUser.name, regUser.email),
             server = server,
             newEmail = newEmail,
        )

    sendmail_name = 'registration_spam'
    def mailThemAll(self, subject, msg_text, sender=None, channels=None,
            attachment=None):
        import comp4.sendmail.api as sm
        ws = settings.workspace()
        recipients = []
        for usr, pre in self.workspace('''[
            (usr,pre)
                for usr in RegisteredUser
                for pre in RegisteredUserPreferences
                 if usr.active == True
                and query.LeftJoin(usr == pre.user and pre.pref == 'newsletter.channels')
                and usr.approved is not None
        ]'''):
            if channels:
                if pre:
                    for ch in pre.value.split(','):
                        if ch in channels:
                            recipients.append((usr.name, usr.email))
            else:
                recipients.append((usr.name, usr.email))
        text = sm.Plain(data=msg_text)
        if attachment:
            fn = attachment.get('filename', 'attachment').encode('utf-8')
            soubor = sm.Attachment(
                content_type = attachment.get('content_type',
                                            'application/octet-stream'),
                content_disposition = 'attachment; filename="%s"' % fn,
                data=attachment['blob'].read(),
                )
            message = sm.Multipart(text, soubor, subtype='mixed')
        else:
            message = text
        opts = dict(
            recipients = recipients,
            subject = subject,
            message = message,
            one_mail_per_recipient = True,
        )
        if sender is not None:
            opts['sender'] = sender
        sm.sendmail(self.sendmail_name, **opts)
        ws.new(RegisteredSendmail,
            recipients = ','.join([e for n,e in recipients]),
            sender = opts['sender'],
            subject = subject,
            message = msg_text,
        )
        return recipients

    def myAdminPlugins(self):
        from lib4.odict import OrderedDict
        plugins = []
        pluginDefs = config.pluginsFor('comp4.registration.myadmin')
        for name in config.Property('comp4.registration.myadmin_plugins')():
            pdef = pluginDefs.get(name, {}).copy()
            factory = importObject(
                pdef.pop('factory','comp4.registration.plugins.MyPlugin'))
            plugins.append((name, (factory, pdef)))
        return OrderedDict(plugins)
    myAdminPlugins = binding.Make(myAdminPlugins)

    def myHomePlugins(self):
        from lib4.odict import OrderedDict
        plugins = []
        pluginDefs = config.pluginsFor('comp4.registration.myhome')
        for name in config.Property('comp4.registration.myhome_plugins')():
            pdef = pluginDefs.get(name, {}).copy()
            factory = importObject(
                pdef.pop('factory','comp4.registration.plugins.MyPlugin'))
            plugins.append((name, (factory, pdef)))
        return OrderedDict(plugins)
    myHomePlugins = binding.Make(myHomePlugins)
