'''
Sendmail
--------

Support for sending email from your application.

    myapp.ini:
        
        # defaults
        [Plugin comp4.sendmail:DEFAULT]
        port = 25
        mailserver = 'localhost'
        one_mail_per_recipient = True
        
        [Plugin comp4.sendmail:shop_invoice_info]
        sender = 'invoices@myshop.cz'
        subject = '[MyShop] Invoice info'
        message = sendmail.Plain(file='myapp/email/invoice_info.tt')
    
        [Plugin comp4.sendmail:shop_admin]
        sender = (u'Jirina Bohdalova', 'invoices@myshop.cz')
        recipients = ['foo@example.com', (u'Bar', 'bar@example.com')]
        message = sendmail.Plain(data='Test from %(senders)s')

    code:
        
        from comp4.sendmail.api import sendmail

        # 'recipients' is required keyword
        sendmail(
            'shop_invoice_info',
            recipients = 'cust@example.com',
        )

        # any keyword can be specified explicitely
        sendmail(
            'shop_invoice_info',
            recipients = 'cust@example.com',
            subject   = '[MyShop] Customized subject',
        )
        
        # many recipients
        sendmail(
            'shop_invoice_info',
            recipients = ['c1@example.com', 'c2@example.com'],
        )

        # data for template
        sendmail(
            'shop_invoice_info',
            recipients = 'cust@example.com',
            invoice   = invoice,
        )

        # attachments
        sendmail(
            'shop_invoice_info',
            recipients   = 'cust@example.com',
            message = sendmail.Mixed(
                sendmail.Plain('/usr/share/myapp/invoice.tt'),
                sendmail.Attachment('/tmp/invoice.pdf'),
                sendmail.Attachment('/usr/share/myapp/signature.jpg'),
            ),
        )
    
'''

__all__ = [
    'substitute',
    'Message', 'Text', 'Plain', 'Html', 'SimpleTemplate', 'Tempita',
    'Multipart', 'Attachment', 
    'SendmailComp', 'sendmail_comp', 'message', 'sendmail',
]

from wm4.api import *
from lib4.imports import lazyModule
from wm4.config.utils import getStreamFactory

import sys
import email

if sys.version_info >= (2,5):
    eh   = lazyModule('email.header')
    emnp = lazyModule('email.mime.nonmultipart')
    emp  = lazyModule('email.mime.multipart')
    emsg = lazyModule('email.message')
    eenc = lazyModule('email.encoders')
    smtp = lazyModule('smtplib')
else:
    eh   = lazyModule('email.Header')
    emnp = lazyModule('email.MIMENonMultipart')
    emp  = lazyModule('email.MIMEMultipart')
    emsg = lazyModule('email.Message')
    eenc = lazyModule('email.Encoders')
    smtp = lazyModule('smtplib')


subst_regex = '%([a-zA-Z][._a-zA-Z0-9]*?)%'
subst_regexc = None


def substitute(txt, params):
    global subst_regexc
    if subst_regexc is None:
        import re
        subst_regexc = re.compile(subst_regex, re.M)
    def s(m):
        k = m.group(1)
        v = params.get(k)
        if v is None: v = ''
        return v
    return subst_regexc.sub(s, txt)


class Message(object):

    file = None
    data = None
    content_type = None
    
    substitute = staticmethod(substitute)

    def __init__(self, *args, **params):
        self.args = args
        self.__dict__.update(params)

    def __call__(self, params={}):
        msg = self.makeMessage(params)
        self.setHeaders(msg, params)
        return msg

    def getData(self, params):
        if self.data is None:
            stream = getStreamFactory(self.file)
            return stream.open('rb').read()
        else:
            return self.data

    def makeMessage(self, params):
        raise NotImplementedError
    
    def setHeaders(self, msg, params):
        self.setSender(msg, params)
        self.setRecipients(msg, params)
        self.setSubject(msg, params)

    def setSender(self, msg, params):
        sender = params.get('sender')
        if not sender:
            return
        elif isinstance(sender, tuple):
            name, email = sender
            name = eh.make_header([(unicode(name).encode('utf-8'), 'utf-8')])
            msg['From'] = '"%s" <%s>' % (name, email)
        else:
            msg['From'] = str(sender)

    def setRecipients(self, msg, params):
        recipients = params.get('recipients')
        if not recipients:
            return
        if not isinstance(recipients, list):
            recipients = [recipients]
        rcpts = []
        for recipient in recipients:
            if isinstance(recipient, tuple):
                name, email = recipient
                name = eh.make_header([(unicode(name).encode('utf-8'),'utf-8')])
                rcpts.append('"%s" <%s>' % (name, email))
            else:
                rcpts.append(str(recipient))
        msg['To'] = ', '.join(rcpts)

    def setSubject(self, msg, params):
        subj = params.get('subject')
        if subj:
            s = self.substitute(unicode(subj), params)
            msg['Subject'] = eh.make_header([(s, 'utf-8')])


class Text(Message):

    charset = 'utf-8'
    subtype = None
    source_encoding = 'utf-8'

    def getData(self, params):
        data = super(Text, self).getData(params)
        if not isinstance(data, unicode):
            data = unicode(data, self.source_encoding)
        return data
        
    def makeMessage(self, params):
        data = self.getData(params)
        data = data.encode(self.charset)
        msg = emnp.MIMENonMultipart('text', self.subtype, charset=self.charset)
        msg['Content-Transfer-Encoding'] = '8bit'
        msg.set_charset(self.charset)
        msg.set_payload(data, self.charset)
        return msg


class Plain(Text):
    subtype = 'plain'

class Html(Text):
    subtype = 'html'


class Template(Plain):

    def getData(self, params):
        data = super(Template, self).getData(params)
        return self.substitute(data, params)

class SimpleTemplate(Template):
    pass

class Tempita(Template):
    
    def substitute(self, data, params):
        import tempita
        tmpl = tempita.Template(data)
        return tmpl.substitute(**params)



class Multipart(Message):
    subtype = 'mixed'

    def makeMessage(self, params):
        msg = emp.MIMEMultipart(
            self.subtype, _subparts=[a(params) for a in self.args]
        )
        return msg


class Attachment(Message):
    content_type = 'application/octet-stream'
    content_disposition = None
    encoding = 'base64' # quoted-printable

    def makeMessage(self, params):
        msg = emsg.Message()
        # XXX use magic for content_type
        msg['Content-Type'] = self.content_type
        if self.content_disposition:
            msg['Content-Disposition'] = self.content_disposition
        msg['MIME-Version'] = '1.0'
        data = self.getData(params)
        msg.set_payload(data)
        if self.encoding == 'base64':
            eenc.encode_base64(msg)
        elif self.encoding in ('quopri', 'quoted-printable'):
            eenc.encode_quopri(msg)
        elif self.encoding == '8bit':
            pass
        else:
            raise ValueError('Unknown Content-Transfer-Encoding', self.encoding)
        return msg
















class SendmailComp(schema.SchemaComponent):

    logger = binding.Make(lambda: logs.getLogger('comp4.sendmail'))

    plugins = binding.Make(lambda: config.pluginsFor('comp4.sendmail'))

    def defaults(self):
        d = self.plugins.get('DEFAULT', {}).copy()
        d.setdefault('port', 25)
        d.setdefault('mailserver', 'localhost')
        d.setdefault('charset', 'utf-8')
        return d
    defaults = binding.Make(defaults)

    def _message(self, params):
        msg = params['message'](params)
        return msg
        
    def params(self, etemplate, params={}):
        _params = self.defaults.copy()
        _params.update(self.plugins.get(etemplate, {}))
        _params.update(params)
        return _params
    
    def message(self, etemplate, **params):
        params = self.params(etemplate, params)
        return self._message(params)
        
    def __call__(self, etemplate, **params):

        params = self.params(etemplate, params)

        sender = params['sender']
        if isinstance(sender, tuple):
            _, sender = sender
        sender = str(sender)

        recipients = params['recipients']
        if not isinstance(recipients, list):
            recipients = [recipients]

        con = smtp.SMTP(params['mailserver'], params['port'])

        if params.get('one_mail_per_recipient', True):

            for recipient in recipients:
                p = params.copy()
                p['recipients'] = [recipient]
                msg = self._message(p)
                if isinstance(recipient, tuple):
                    _, recipient = recipient
                con.sendmail(sender, [recipient], str(msg))

        else:

            rcpts = []
            for recipient in recipients:
                if isinstance(recipient, tuple):
                    _, recipient = recipient
                rcpts.append(recipient)
            msg = self._message(params)
            con.sendmail(sender, rcpts, str(msg))
                



sendmail_comp = config.Resource('comp4.sendmail', factory=SendmailComp)


def message(email, **params):
    return sendmail_comp().message(email, **params)
    

def sendmail(email, **params):
    return sendmail_comp()(email, **params)
