# -*- coding: utf-8 -*-

import sys
from wm4.testing import *
from wm4.api import *
from comp4.sendmail.api import *
from os.path import dirname, join
from lib4.extremes import Any
if sys.version_info >= (2, 5):
    from email.header import decode_header
else:
    from email.Header import decode_header

THIS_DIR = dirname(__file__)


INI = '''
# -*- coding: utf-8 -*-

[Import on Demand]
sm = 'comp4.sendmail.api'

[Plugin comp4.sendmail:DEFAULT]
port = 2525

[Plugin comp4.sendmail:test1]
sender = (u'Foo Bár','foo@example.com')
recipients = ['bar@example.com', (u'Baz Báz','baz@example.com')]
subject = u'Český subjekt'
message = sm.Plain(data=u'Zdravíčko')
'''


class TestSendmailConfig(ConfiguredTestCase):

    def test01_Defaults(self):
        sm = sendmail_comp()
        self.assertEqual(
            sm.params('test1'),
            dict(mailserver='localhost', port=25, charset='utf-8'),
        )
        
    @with_ini(INI)
    def test02_TestDefaults(self):
        sm = sendmail_comp()
        self.assertEqual(
            sm.params('test1'),
            dict(   
                mailserver='localhost', port=2525, charset='utf-8',
                sender=(u'Foo Bár', 'foo@example.com'),
                recipients=['bar@example.com', (u'Baz Báz','baz@example.com')],
                subject=u'Český subjekt', message = Any,
            ),
        )


class TestMessageComposition(ConfiguredTestCase):
    
    def test20_Attachment(self):
        EQ = self.assertEqual
        img = join(THIS_DIR, 'img.gif')
        m = Attachment(file=img, content_type='image/gif')()
        EQ(m.get_content_type(), 'image/gif')
        EQ(m.is_multipart(), False)
        EQ(m.get_payload(decode=True), open(img,'rb').read())
        
    def test21_AttachmentPkgURL(self):
        EQ = self.assertEqual
        img = join(THIS_DIR, 'img.gif')
        m = Attachment(
            file='pkgfile:comp4.sendmail.tests/img.gif',
            content_type='image/gif'
        )()
        EQ(m.get_content_type(), 'image/gif')
        EQ(m.is_multipart(), False)
        EQ(m.get_payload(decode=True), open(img,'rb').read())
        
    def test30_SimpleTemplate(self):
        EQ = self.assertEqual
        m = SimpleTemplate(data = 'aaa%b%c\n%d%eee%%%f%\n')
        m = m(dict(b='B', d='D', f='Foo'))
        EQ(m.is_multipart(), False)
        EQ(m.get_content_type(), 'text/plain')
        EQ(m.get_payload(), 'aaaBc\nDeee%%Foo\n')

    def test31_SimpleTemplateWithCustomSubstitution(self):
        EQ = self.assertEqual
        m = SimpleTemplate(
                data = 'aaa%b%c\n%d%eee%%%f%\n',
                substitute=lambda t,p: substitute(t,p).upper())()
        EQ(m.get_payload(), 'AAAC\nEEE%%\n')

    def test40_Multipart(self):
        EQ = self.assertEqual
        m = Multipart(
                Multipart(
                    Plain(data=u'Foo'),
                    Html(data=u'<p>Foo</p>'),
                    subtype='alternative'
                ),
                Plain(data=u'Baz'),
        )()
        EQ(m.is_multipart(), True)
        EQ(m.get_content_type(), 'multipart/mixed')

        m0 = m.get_payload(0)
        EQ(m0.is_multipart(), True)
        EQ(m0.get_content_type(), 'multipart/alternative')

        m00 = m0.get_payload(0)
        EQ(m00.is_multipart(), False)
        EQ(m00.get_content_type(), 'text/plain')
        EQ(m00.get_payload(), 'Foo')

        m01 = m0.get_payload(1)
        EQ(m01.is_multipart(), False)
        EQ(m01.get_content_type(), 'text/html')
        EQ(m01.get_payload(), '<p>Foo</p>')

        m1 = m.get_payload(1)
        EQ(m1.is_multipart(), False)
        EQ(m1.get_content_type(), 'text/plain')

    def test50_Temptia(self):
        EQ = self.assertEqual
        m = Tempita(data = u'čeština {{a}}')
        m = m(dict(a=42, subject='Sub {{a}}'))
        EQ(m.is_multipart(), False)
        EQ(m.get_content_type(), 'text/plain')
        EQ(m.get_payload(), u'čeština 42'.encode('utf-8'))
        EQ(m['Subject'], 'Sub 42')
        
        
        



class TestSendmail(ConfiguredTestCase):

    def setUp(self):
        super(TestSendmail, self).setUp()
        from wm4.testing.smtp import MockSmtpServer
        self.smtpd = MockSmtpServer(port=2525).spawn()
        
    def tearDown(self):
        super(TestSendmail, self).tearDown()
        self.smtpd.exit()
    
    @with_ini(INI)
    def test10_OneMailPerRecipient(self):
        EQ = self.assertEqual
        sendmail('test1', one_mail_per_recipient=True)
        q = self.smtpd.messages()
        EQ(len(q), 2)
        m = q[0]
        EQ(m.mailfrom, 'foo@example.com')
        EQ(m.rcpttos, ['bar@example.com'])
        EQ(m['MIME-Version'], '1.0')
        EQ(m['Content-Type'], 'text/plain; charset="utf-8"')
        EQ(m['Content-Transfer-Encoding'], '8bit')
        # differs in 2.4 and 2.5
        #EQ(m['From'], u'"Foo Bár" <foo@example.com>')
        EQ(m['To'],u'bar@example.com')
        EQ(m.get_text(), u'Zdravíčko')
        m = q[1]
        EQ(m.mailfrom, 'foo@example.com')
        EQ(m.rcpttos, ['baz@example.com'])
        # differs in 2.4 and 2.5
        #EQ(m['From'], u'"Foo Bár" <foo@example.com>')
        #EQ(m['To'],u'"Baz Báz" <baz@example.com>')
        EQ(m.get_text(), u'Zdravíčko')


    @with_ini(INI)
    def test11_OneMailPerRecipients(self):
        EQ = self.assertEqual
        sendmail('test1', one_mail_per_recipient=False)
        q = self.smtpd.messages()
        EQ(len(q), 1)
        m = q[0]
        EQ(m.mailfrom, 'foo@example.com')
        EQ(m.rcpttos, ['bar@example.com', 'baz@example.com'])
        EQ(m['MIME-Version'], '1.0')
        EQ(m['Content-Type'], 'text/plain; charset="utf-8"')
        EQ(m['Content-Transfer-Encoding'], '8bit')
        # differs in 2.4 and 2.5
        #EQ(m['From'], u'"Foo Bár" <foo@example.com>')
        #EQ(m['To'],u'bar@example.com, "Baz Báz" <baz@example.com>')
        EQ(m.get_text(), u'Zdravíčko')



suite = Suite(
    TestSendmailConfig,
    TestMessageComposition,
    TestSendmail,
)

if __name__ == '__main__':
    Main()
