# -*- coding: utf-8 -*-
import random
from datetime import date, datetime

from wm4.api import *
from wm4.schema.api import *
from wm4.web.sessions import TrustedClientData

from adt4.serial import Serial
from adt4.enums import *
from lib4.trees import AbstractTreeNode
from lib4.imports import importObject

from comp4.files.api import ImageBase, FileBase
from comp4.forum import api as forum
from comp4.cms import api as cms
from comp4.trees import api as trees
import comp4.dbgettext.api as dbgettext


from parameters import ParametersFeature, UnicodeParam, FileParam

__all__ = [
    'ShopCategory', 'ShopCategoryParam', 'ShopCategoryRoot',
    'ShopGoodsType', 'ShopGoods',
    'ShopGoodsCategory', 'ShopGoodsTip', 'ShopGoodsRelated',
    'ShopGoodsParam', 'ShopNews',
    'ShopFile', 'ShopImage', 'ShopPhotogalleryImage', 'ShopPhotogalleryRes',
    'ShopManufacturer', 'ShopPerson', 'ShopGoodsPerson',
    'ShopGoodsComment', 'ShopBasket', 'ShopBasketItem',
    'ShopGoodsBookmark', 'ShopCustomer',
    'DeliveryMode', 'PaymentMode', 'DeliveryPayment', 'TotalRangeMapTaxes',
    'InvoiceStatus', 'ShopInvoice', 'ShopInvoiceItem',
    'ShopDiscountTicketBatch', 'ShopDiscountTicket',
    'ShopComp',
]


class ShopCategory(trees.TreeNode):
    pkey = schema.Attribute(Serial)
    name = Attribute(unicode, Required)
    _name = dbgettext.Translations('name')
    mark = Attribute(str)
    template = Attribute(str)
    description = Attribute(unicode)
    position = Attribute(int, Required, Default(0))

    paramDefs = binding.Make(lambda:settings.workspace()['shop'].categoryParams)
    params = ParametersFeature(
        'comp4.shop.entities.ShopCategoryParam', 'category')

    binding.metadata(Unique('mark'))

    children = ReferenceEnd(
            THIS_CLASS, 'parent', ReferenceEndOrder('position', 'name'))

    countAll = binding.Make(lambda self: ~self._workspace(
            '[query.Count(g) for g in ShopGoodsCategory if g.category == self]'
    ))
    count = binding.Make(lambda self: ~self._workspace('''[
            query.Count(gc) for gc in ShopGoodsCategory
                            for g in ShopGoods
                             if gc.category == self
                            and gc.goods == g
                            and g.hidden == False
    ]'''))

    def countAccum(self):
        return sum(n.count for n in self.iterDepthFirst())
    countAccum = binding.Make(countAccum)

    cms = binding.Make(lambda s: cms.getRootNode('shop_category.%s' % s.pkey))

    @setattrs.before()
    def check_description_and_gen_short(self, added, removed):
        if 'description' not in added:
            return
        from utils import tidy_description
        try:
            log = []
            added['description'] = tidy_description(added['description'], log)
        except:
            ee = EntityError()
            ee.add(FeatureSyntaxError('description','invalid XHTML'))
            raise ee

    def news(self):
        ws = self._workspace
        today = date.today()
        qry = ws.parse('''[
            g for g in ShopGoods
              for n in ShopNews
               if g == n.goods
              and n.new_from <= today <= n.new_till
              and g.hidden == False
              and g._oid in [
                gc.goods._oid
                    for gc in ShopGoodsCategory
                    for tc in ShopCategory.treeCache
                     if gc.category.pkey == tc.end
                    and tc.begin == self.pkey
              ]
        ]''')
        qry.limit = query.Const(12)
        qry.offset = query.Const(0)
        lst = list(ws(qry))
        ws['shop'].preloadGoods(lst)
        return lst
    news = binding.Make(news)

    _tips = ReferenceEnd('comp4.shop.entities.ShopGoodsTip', 'category')

    def tips(self):
        tips = self._workspace['shop'].preloadGoods(
            [t.goods for t in self._tips]
        )
        return [g for g in tips if not g.hidden and not g.deleted]
    tips = binding.Make(tips)

    path = binding.Make(
            lambda s: ' / '.join(c.name for c in s.iterParentsRev(True)))

    def goods_query(
        self, subcategories=None, hidden=False, deleted=False, order='name',
    ):
        shop = self._workspace['shop']
        if subcategories is None:
            subcategories = shop.show_goods_from_subcategories
        if subcategories:
            qry = '''
                obj for obj in ShopGoods
                     if obj._oid in [
                        gc.goods._oid
                            for gc in ShopGoodsCategory
                            for tc in ShopCategory.treeCache
                             if gc.category.pkey == tc.end
                            and tc.begin == self.pkey
                    ]
                '''
        else:
            qry = '''
                obj for obj in ShopGoods
                    for gc in ShopGoodsCategory
                     if obj.pkey == gc.goods.pkey
                    and gc.category.pkey == self.pkey
                '''
        qry = qry + ' and obj.deleted == deleted and obj.hidden == hidden'
        return self._workspace.parse('[%s]' % qry)

    all_goods = binding.Make(lambda s: list(s._workspace(s.goods_query())))

    def _template(self):
        while self is not None:
            if self.template is not None:
                return self.template
            self = self.parent
    _template = binding.Make(_template)


class ShopCategoryParam(Entity):
    pkey = Attribute(Serial)
    category = Reference(ShopCategory, Required)
    param = Attribute(unicode, Required)
    value = Attribute(unicode)
    position = Attribute(int, Required, Default(1))

    binding.metadata(PrimaryKey('pkey'))


class ShopCategoryRoot(schema.SchemaComponent, AbstractTreeNode):

    txnAttrs = 'children', 'cats', 'marks',
    parent = None

    def children_and_cats(self):
        self.joinedTxn
        ws = settings.workspace()

        cats = {}
        branches = {}
        marks = {}
        for cat, cnt in ws('''[
            (c, query.Count(g))
                for c  in ShopCategory
                for gc in ShopGoodsCategory
                for g  in ShopGoods
                 if query.LeftJoin(c == gc.category)
                and query.LeftJoin(gc.goods == g and g.hidden == False)
                and query.GroupBy(c)
                and query.OrderBy(c.position, c.name)
        ]'''):
            cats[cat.pkey] = cat
            if cat.mark:
                marks[cat.mark] = cat
            cat.__dict__['count'] = cnt
            branches.setdefault(cat.parent, []).append(cat)
            cat.__dict__['children'] = ()

        for parent, branch in branches.iteritems():
            if parent is None:
                continue
            parent.__dict__['children'] = tuple(branch)
        return dict(
            children   = branches.get(None, ()),
            categories = cats,
            marks      = marks,
        )
    children = categories = marks = binding.MultiMake(children_and_cats)








class ShopGoodsType(object):
    name = None
    title = None
    description = None
    help = None
    controller = None
    view = None
    def __init__(self, **kw):
        self.__dict__.update(kw)
        if self.title is None:
            self.title = self.name



class ShopGoods(Entity):
    pkey = Attribute(Serial)
    typ = Attribute(str, Required)
    typObj = binding.Make(lambda s: s._workspace['shop'].goodsTypes[s.typ])
    name = Attribute(unicode, Required)
    _name = dbgettext.Translations('name')

    short = Attribute(unicode)
    _short = dbgettext.Translations('short', useDefault=False)
    description = Attribute(unicode)
    _description = dbgettext.Translations('description', useDefault=False)

    price = Attribute(float)
    price_common = Attribute(float)
    price_action = Attribute(float)

    not_avail = Attribute(unicode)
    note = Attribute(unicode)

    hidden = Attribute(bool, Required, Default(False))
    deleted = Attribute(bool, Required, Default(False))

    manufacturer = Reference('comp4.shop.entities.ShopManufacturer')
    _persons = ReferenceEnd('comp4.shop.entities.ShopGoodsPerson', 'goods',
                ReferenceEndOrder('position'))

    class Persons:
        def __init__(self, goods):
            self.goods = goods
        def __getattr__(self, attr):
            return [p.person for p in self.goods._persons if p.role == attr]
    persons = property(lambda s: s.Persons(s))

    paramDefs = binding.Make(lambda: settings.workspace()['shop'].goodsParams)
    params = ParametersFeature('comp4.shop.entities.ShopGoodsParam', 'goods')

    binding.metadata(PrimaryKey('pkey'))

    @setattrs.before()
    def check_description_and_gen_short(self, added, removed):
        if 'description' not in added:
            return
        from utils import tidy_description, extract_short_description
        try:
            log = []
            descr = tidy_description(added['description'], log)
        except:
            ee = EntityError()
            ee.add(FeatureSyntaxError('description','invalid XHTML'))
            raise ee
        if 'short' not in removed:
            if 'short' in added:
                short = added['short']
            else:
                short = self.__dict__.get('short')
            if not short:
                added['short'] = extract_short_description(descr)

    orderable = binding.Make(
        lambda s: (not s.not_avail) and (s.price is not None)
    )

    def order_price(self):
        if self.price is None:
            return None
        if self.price_action is not None:
            return self.price_action
        else:
            return self.price
    order_price = binding.Make(order_price)

    def discount(self):
        if self.price_common is None:
            return None
        if self.price is None:
            return None
        if self.price_action is None:
            return self.price_common - self.price
        else:
            return self.price_common - self.price_action
    discount = binding.Make(discount)

    _cats = ReferenceEnd('comp4.shop.entities.ShopGoodsCategory', 'goods')
    _tips = ReferenceEnd('comp4.shop.entities.ShopGoodsTip', 'goods')

    def categories(self):
        psc = self._workspace['shop'].preloadSupercategories
        return tuple(psc(c) for c in self._workspace('''[
            c for c in ShopCategory for x in ShopGoodsCategory
               if c == x.category and x.goods == self
            ]'''))
    categories = binding.Make(categories)

    cms = binding.Make(lambda s: cms.getRootNode('shop_goods.%d' % s._oid))


    photogallery = ReferenceEnd(
        'comp4.shop.entities.ShopPhotogalleryImage','goods',
        ReferenceEndOrder('position'))

    def comments(self):
        branches = {}
        comments = {}
        for com in settings.workspace()('''[
            com for com in ShopGoodsComment
                 if com.goods == self
                and query.OrderBy(com.created)
        ]'''):
            comments[com.pkey] = com
            branches.setdefault(com.parent, []).append(com)
        for par, children in branches.iteritems():
            if par:
                par.__dict__['children'] = children
        for com in comments.itervalues():
            com.__dict__.setdefault('children', ())
        return dict(
            comments = branches.get(None, ()),
            commentDict = comments,
        )
    comments = commentDict = binding.MultiMake(comments)

    def rating(self):
        r = -self._workspace('''[
            query.Avg(c.rating) for c in ShopGoodsComment
                                 if c.goods == self
        ]''')
        if r is None:
            return None
        r = float(r)
        vals = [0.0, 0.5, 1.0, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0, 4.5, 5.0]
        difs = [(abs(r-v), v) for v in vals]
        difs.sort()
        return difs[0][1]
    rating = binding.Make(rating)

    news = ReferenceEnd('comp4.shop.entities.ShopNews', 'goods')
    def current_news(self):
        t = date.today()
        for n in self.news:
            if n.new_from <= t <= n.new_till:
                return n
        return None
    current_news = binding.Make(current_news)

    _related = ReferenceEnd('comp4.shop.entities.ShopGoodsRelated', 'goods')
    def related(self):
        ws = self._workspace
        return ws['shop'].preloadGoods(tuple(ws('''[
                g for g in ShopGoods
                  for r in ShopGoodsRelated
                   if g.pkey == r.related.pkey
                  and r.goods.pkey == self.pkey
                  and g.deleted == False
                  and g.hidden == False
                  and query.OrderBy(g.name)
            ]''')))
    related = binding.Make(related)







class ShopGoodsCategory(Entity):
    goods = Reference(ShopGoods)
    category = Reference(ShopCategory)

    binding.metadata(PrimaryKey('goods', 'category'))


class ShopGoodsTip(Entity):
    goods = Reference(ShopGoods)
    category = Reference(ShopCategory)

    binding.metadata(PrimaryKey('goods', 'category'))

class ShopGoodsRelated(Entity):
    goods = Reference(ShopGoods)
    related = Reference(ShopGoods)

    description = Attribute(unicode)

    binding.metadata(PrimaryKey('goods', 'related'))



class ShopGoodsParam(Entity):
    pkey = Attribute(Serial)
    goods = Reference(ShopGoods, Required)
    param = Attribute(unicode, Required)
    value = Attribute(unicode)
    position = Attribute(int, Required, Default(1))

    binding.metadata(PrimaryKey('pkey'))

class ShopNews(Entity):
    pkey = Attribute(Serial)
    goods = Reference(ShopGoods, Required)
    new_from = Attribute(date, Required)
    new_till = Attribute(date, Required)
    announce = Attribute(unicode)

    binding.metadata(PrimaryKey('pkey'))




class ShopFile(FileBase):
    pkey = Attribute(Serial)
    binding.metadata(PrimaryKey('pkey'))

class ShopImage(ImageBase):
    pkey = Attribute(Serial)
    binding.metadata(PrimaryKey('pkey'))

class ShopPhotogalleryImage(Entity):
    pkey = Attribute(Serial)
    goods = Reference(ShopGoods, Required)
    description = Attribute(unicode)
    position = Attribute(int, Required, Default(10))
    resolutions = ReferenceEnd(
        'comp4.shop.entities.ShopPhotogalleryRes', 'photogallery')

    image = binding.Make(lambda s: s.get('thumb'))

    def get(self, resolution):
        for r in self.resolutions:
            if r.resolution == resolution:
                return r.image
    binding.metadata(PrimaryKey('pkey'))

class ShopPhotogalleryRes(Entity):
    pkey = Attribute(Serial)
    photogallery = Reference(ShopPhotogalleryImage, Required)
    image = Reference(ShopImage, Required)
    resolution = Attribute(str, Required, Default('orig'))

    binding.metadata(PrimaryKey('pkey'))

Workspace.do_remove.before('obj in ShopPhotogalleryImage')
def do_remove_photogallery_image(ws, obj):
    for r in obj.resolutions:
        ws.remove(r)

Workspace.do_remove.before('obj in ShopPhotogalleryRes')
def do_remove_photogallery_res(ws, obj):
    ws.remove(obj.image)





class ShopManufacturer(Entity):
    pkey = Attribute(Serial)
    name = Attribute(unicode, Required)
    image = Reference(ShopImage)
    description = Attribute(unicode)
    url = Attribute(str)
    email = Attribute(str)
    phone = Attribute(unicode)
    address = Attribute(unicode)
    state = Attribute(str)
    cms = binding.Make(lambda s:cms.getRootNode('shop_manufacturer.%s' %s.pkey))
    binding.metadata(PrimaryKey('pkey'))

    def goods_query(self, hidden=False, deleted=False, order='name'):
        qry = '''
            obj for obj in ShopGoods
                 if obj.manufacturer.pkey == self.pkey
                and obj.deleted == deleted
                and obj.hidden == hidden
            '''
        return self._workspace.parse('[%s]' % qry)


class ShopPerson(Entity):
    pkey = Attribute(Serial)
    name = Attribute(unicode, Required)
    image = Reference(ShopImage)
    state = Attribute(str)
    url = Attribute(str)
    description = Attribute(unicode)
    birth_day = Attribute(int)
    birth_month = Attribute(int)
    birth_year = Attribute(int)
    death_day = Attribute(int)
    death_month = Attribute(int)
    death_year = Attribute(int)

    def birth_date(self):
        if not self.birth_year:
            return None
        if not self.birth_month or not self.birth_day:
            return self.birth_year
        return '%d.%d.%d' % (self.birth_day, self.birth_month, self.birth_year)
    birth_date = binding.Make(birth_date)

    def death_date(self):
        if not self.death_year:
            return None
        if not self.death_month or not self.death_day:
            return self.death_year
        return '%d.%d.%d' % (self.death_day, self.death_month, self.death_year)
    death_date = binding.Make(death_date)

    cms = binding.Make(lambda s: cms.getRootNode('shop_person.%s' % s.pkey))

    def goods_query(self, hidden=False, deleted=False, order='name'):
        qry = '''
            obj for obj in ShopGoods
                for gp in 'shop/GoodsPerson'
                 if gp.goods == obj
                and gp.person == self
                and obj.deleted == deleted
                and obj.hidden == hidden
            '''
        return self._workspace.parse('[%s]' % qry)

    binding.metadata(PrimaryKey('pkey'))





class ShopGoodsPerson(Entity):
    pkey = Attribute(Serial)
    goods = Reference(ShopGoods, Required)
    person = Reference(ShopPerson, Required)
    role = Attribute(str, Required)
    position = Attribute(int, Required, Default(0))

    binding.metadata(PrimaryKey('pkey'))


class ShopGoodsComment(Entity, AbstractTreeNode):
    pkey = Attribute(Serial)

    goods = Reference(ShopGoods, Required)
    parent = Reference(THIS_CLASS)
    children = ReferenceEnd(THIS_CLASS, 'parent')

    sender = Attribute(unicode)
    email = Attribute(unicode)
    www = Attribute(unicode)
    subject = Attribute(unicode)
    message = Attribute(unicode, Required)
    #markup = binding.Obtain('message')
    rating = Attribute(int)

    ip = Attribute(str)
    created = Attribute(datetime, defaultDatetime)
    published = Attribute(bool, Required, Default(True))
    deleted = Attribute(bool, Required, Default(False))
    visible = binding.Make(lambda s: s.published and not s.deleted)

    binding.metadata(PrimaryKey('pkey'))






class ShopBasket(binding.Component):

    token = binding.Require()

    workspace = binding.Make(lambda: settings.workspace())

    items = binding.Make(lambda s: list(s.workspace('''[
        i for i in ShopBasketItem if i.token==s.token and query.OrderBy(i.pkey)
    ]''')))

    def total(self):
        if self.items:
            return sum([i.unit_price * i.quantity  for i in self.items])
        else:
            return 0
    total = binding.Make(total)

    pieces = binding.Make(lambda s: sum(i.quantity for i in s.items))

    @classmethod
    def makeElement(klass, ws):
        return klass


class ShopBasketItem(Entity):
    pkey = Attribute(Serial)
    token = Attribute(str, Required)

    goods = Reference(ShopGoods, Required)
    unit  = Attribute(str, Required, Default('piece'))
    quantity = Attribute(float, Required, Default(1))
    unit_price = Attribute(float, Required)

    modified = Attribute(datetime, Required, schema.defaultDatetime)

    binding.metadata(PrimaryKey('pkey'))



class ShopGoodsBookmark(Entity):
    pkey = Attribute(Serial)
    goods = Reference(ShopGoods, Required)
    reg_user = Reference('comp4.registration.entities.RegisteredUser', Required)
    wish = Attribute(bool)
    note = Attribute(unicode)

    binding.metadata(PrimaryKey('pkey'))




class ShopCustomer(Entity):
    pkey = Attribute(Serial)

    reg_user = Reference('comp4.registration.entities.RegisteredUser')

    name = Attribute(unicode, Required)
    email = Attribute(unicode, Required, CheckEmail)
    phone = Attribute(unicode, Required)
    web = Attribute(unicode)

    company = Attribute(unicode)
    ico = Attribute(unicode)
    dic = Attribute(unicode)

    street = Attribute(unicode, Required)
    city = Attribute(unicode, Required)
    zip = Attribute(unicode, Required)
    country = Attribute(unicode, Required)

    dstreet = Attribute(unicode)
    dcity = Attribute(unicode)
    dzip = Attribute(unicode)
    dcountry = Attribute(unicode)

    invoices = QueryFeature('''[
      (inv, query.Count(item), query.Sum(item.unit_price * item.quantity)
                + inv.postage + inv.cod + inv.packaging - inv.discount)
                for inv in ShopInvoice
                for item in ShopInvoiceItem
                 if inv.customer == self
                and query.LeftJoin(inv == item.invoice)
                and query.GroupBy(inv)
                and query.OrderBy(query.Desc(inv.created))
        ]''')

    binding.metadata(PrimaryKey('pkey'))



class _getitem2getattr:
    def __init__(self, obj):
        self.obj = obj
    def __getitem__(self, attr):
        obj = self.obj
        for a in attr.split('.'):
            obj = getattr(obj, a)
        return obj

class DeliveryMode(object):
    key = None
    name = None
    description = None
    result = None
    def __init__(self, **kw):
        self.__dict__.update(kw)
    def result_data(self, invoice):
        if self.result:
            result = self.result % _getitem2getattr(invoice)
        else:
            result = None
        return {'result':result}

class PaymentMode(object):
    key = None
    name = None
    description = None
    result = None
    def __init__(self, **kw):
        self.__dict__.update(kw)
    def result_data(self, invoice):
        if self.result:
            result = self.result % _getitem2getattr(invoice)
        else:
            result = None
        return {'result':result}


class DeliveryPayment(object):
    taxes = lambda basket: {}
    info = None
    def __init__(self, **kw):
        self.__dict__.update(kw)
    def __call__(self, basket):
        d = self.taxes(basket)
        d['info'] = None
        return d


class TotalRangeMapTaxes:

    def __init__(self, *items):
        self.items = list(reversed(items))

    def __call__(self, basket):
        for l, tx in self.items:
            if basket.total >= l:
                return tx
        return {}







class InvoiceStatus(Enum):
    new = enum(u'nová')
    accepted = enum(u'přijatá')
    storno = enum(u'stornovaná')
    settled = enum(u'uhrazená')
    finished = enum(u'dokončená')


class ShopInvoice(Entity):
    pkey = Attribute(Serial)
    customer = Reference(ShopCustomer, Required)

    delivery = Attribute(unicode, Required)
    deliveryDesc = binding.Make(
        lambda s:s._workspace['shop'].delivery_dict[s.delivery])

    payment = Attribute(unicode, Required)
    paymentDesc = binding.Make(
        lambda s:s._workspace['shop'].payment_dict[s.payment])

    postage = Attribute(float, Required, Default(0.0))
    cod  = Attribute(float, Required, Default(0.0))
    packaging = Attribute(float, Required, Default(0.0))

    discount = Attribute(float, Required, Default(0.0))
    discount_ticket = Reference('comp4.shop.entities.ShopDiscountTicket')

    @setattrs.before()
    def check_delivery_payment(self, added, removed):
        if 'delivery' in added or 'payment' in added:
            if 'delivery' in added: d = added['delivery']
            else: d = self.delivery
            if 'payment' in added: p = added['payment']
            else: p = self.payment
            shop = settings.workspace()['shop']
            if (d,p) not in shop.delivery_payment:
                ee = EntityError()
                msg = u'Nepovolená kombinace způsobu platby a dodání %s+%s' % (
                    d, p
                )
                ee.add(FeatureError('delivery', msg))
                ee.add(FeatureError('payment', msg))
                raise ee

    @setattrs.before("'discount_ticket' in added")
    def compute_discount_from_ticket(self, added, removed):
        ee = EntityError()
        ticket = added['discount_ticket']
        if not ticket:
            return
        if not ticket.free:
            ee.add(FeatureSyntaxError('discount_ticket', u'Neplatný kupón'))
        if ee:
            raise ee
        # XXX rounding
        disc = ticket.compute_discount(self.total)
        if 'discount' not in added:
            self.discount += disc
        ticket.invoice = self

    status = Attribute(InvoiceStatus, Required)
    created = Attribute(datetime, Required, schema.defaultDatetime)

    ext_invoice = Attribute(unicode)

    note = Attribute(unicode)

    items = ReferenceEnd(
        'comp4.shop.entities.ShopInvoiceItem','invoice')

    total = binding.Make(
        lambda s: sum([(i.quantity*i.unit_price) for i in s.items]))

    total_total = binding.Make(lambda s: (
        s.total + s.postage + s.cod + s.packaging - s.discount
    ))

    quantity_sum = binding.Make(lambda s: sum(i.quantity for i in s.items))

    binding.metadata(PrimaryKey('pkey'))



class ShopInvoiceItem(Entity):
    pkey = Attribute(Serial)
    invoice = Reference(ShopInvoice, Required)

    goods = Reference(ShopGoods, Required)
    unit  = Attribute(str, Required, Default('piece'))
    quantity = Attribute(int, Required, Default(1))
    unit_price = Attribute(float, Required)

    note = Attribute(unicode)
    available = Attribute(bool, Required, Default(True))

    binding.metadata(PrimaryKey('pkey'))



class ShopDiscountTicket(Entity):
    pkey = Attribute(Serial)
    code = Attribute(str, Required)
    batch = Reference('comp4.shop.entities.ShopDiscountTicketBatch', Required)
    invoice = Reference('comp4.shop.entities.ShopInvoice')
    blocked = Attribute(bool, Required, Default(False))

    def free(self):
        if self.blocked:
            return False
        if self.invoice is not None:
            return False
        return self.batch.valid_from <= date.today() <= self.batch.valid_till
    free = binding.Make(free)

    def compute_discount(self, total):
        if self.batch.absolute is not None:
            if self.batch.absolute > total:
                return total
            else:
                return self.batch.absolute
        else:
            return total * (self.batch.percent / 100.0)

    binding.metadata(PrimaryKey('pkey'))



class ShopDiscountTicketGen(query.Procedure):

    def __init__(self, batch, num):
        self.args = batch, num

@storage.PgSqlConnection.renderExpr.when('expr in ShopDiscountTicketGen')
def render_ts_match(self, expr, relations):
    return 'shop_gen_discount_tickets(%s, %s)' % (
        self.renderExpr(expr.args[0], relations),
        self.renderExpr(expr.args[1], relations),
    )


class ShopDiscountTicketBatch(Entity):
    pkey = Attribute(Serial)
    name = Attribute(unicode, Required)

    valid_from = Attribute(date, Required)
    valid_till = Attribute(date, Required)
    percent = Attribute(float)
    absolute = Attribute(float)

    note = Attribute(unicode)
    created = Attribute(datetime, Required, schema.defaultDatetime)

    tickets = ReferenceEnd('comp4.shop.entities.ShopDiscountTicket', 'batch')

    binding.metadata(PrimaryKey('pkey'))

    @schema.setattrs.after()
    def check_discounts(self, added, removed):
        if self.percent is None and self.absolute is None:
            m = u'Aspoň jedna ze slev musí být vyplněna'
            ee = EntityError()
            ee.add(FeatureError('percent',m))
            ee.add(FeatureError('absolute',m))
            raise ee
        if self.percent is not None and self.absolute is not None:
            m = u'Pouze jedna ze slev může být vyplněna'
            ee = EntityError()
            ee.add(FeatureError('percent',m))
            ee.add(FeatureError('absolute',m))
            raise ee

    count = binding.Make(lambda s: len(s.tickets))
    used = binding.Make(lambda s: len([t for t in s.tickets if t.invoice]))

    def discount_desc(self):
        if self.percent is None:
            return u'%.2fKč' % self.absolute
        else:
            return u'%.2f%%' % self.percent
    discount_desc = binding.Make(discount_desc)

    def gen(self, num):
        self._workspace('ShopDiscountTicketGen(%d,%d)' % (self.pkey, num))







def P(p):
    return binding.Make(lambda: config.Property(p)())


class ShopComp(schema.SchemaComponent):

    terms_of_trade_url = P('comp4.shop.terms_of_trade_url')
    goods_short_lenght = P('comp4.shop.goods.short.lenght')
    show_goods_from_subcategories =P('comp4.shop.show_goods_from_subcategories')

    def goodsTypes(self):
        from lib4.odict import OrderedDict
        types = []
        plugins = config.pluginsFor('comp4.shop.goods_type')
        for name in config.Property('comp4.shop.goods_types')():
            tdef = plugins.get(name, {}).copy()
            factory = importObject(tdef.pop('factory', ShopGoodsType))
            tdef['params'] = tuple(
                self.goodsParams[p] for p in tdef.get('params',())
            )
            types.append((name, factory(name=name, **tdef)))
        return OrderedDict(types)
    goodsTypes = binding.Make(goodsTypes)

    def goodsParams(self):
        params = []
        plugins = config.pluginsFor('comp4.shop.goods_param')
        for name in config.Property('comp4.shop.goods_params')():
            pdef = plugins.get(name, {}).copy()
            factory = importObject(pdef.pop('factory', UnicodeParam))
            params.append((name, factory(name=name, **pdef)))
        return schema.SortedFeatureDict(params)
    goodsParams = binding.Make(goodsParams)

    def categoryParams(self):
        params = []
        plugins = config.pluginsFor('comp4.shop.category_param')
        for name in config.Property('comp4.shop.category_params')():
            pdef = plugins.get(name, {}).copy()
            factory = importObject(pdef.pop('factory', UnicodeParam))
            params.append((name, factory(name=name, **pdef)))
        return schema.SortedFeatureDict(params)
    categoryParams = binding.Make(categoryParams)


    def delivery_modes(self):
        plugs = config.pluginsFor('comp4.shop.delivery')
        dm = []
        for m in config.Property('comp4.shop.delivery')():
            kw = {'visible': True}
            kw.update(plugs.get(m,{}))
            dm.append(DeliveryMode(key=m, **kw))
        return dm
    delivery_modes = binding.Make(delivery_modes)

    delivery_dict = property(
        lambda s: dict((o.key,o) for o in s.delivery_modes))

    def payment_modes(self):
        plugs = config.pluginsFor('comp4.shop.payment')
        pm = []
        for m in config.Property('comp4.shop.payment')():
            kw = {'visible': True}
            kw.update(plugs.get(m,{}))
            pm.append(PaymentMode(key=m, **kw))
        return pm
    payment_modes = binding.Make(payment_modes)

    payment_dict = property(
        lambda s: dict((o.key,o) for o in s.payment_modes))

    def delivery_payment(self):
        plugs = config.pluginsFor('comp4.shop.delivery_payment')
        combinations = {}
        for m in config.Property('comp4.shop.delivery_payment')():
            d,p = m.split('+')
            combinations[(d,p)] = DeliveryPayment(**plugs.get(m,{}))
        return combinations
    delivery_payment = binding.Make(delivery_payment)



    taxes = [
        ('postage', u'Poštovné'),
        ('cod',     u'Dobírkovné'),
    ]

    def get_taxes(self, d, p, basket):
        desc = self.delivery_payment.get((d,p))
        if desc is None:
            return {}
        return desc.taxes(basket)

    person_roles = binding.Make(
        lambda: config.Property('comp4.shop.person.roles')())


    goods_image_resolutions = binding.Make(
        lambda: config.Property('comp4.shop.goods.image.resolutions')())

    goods_image_convertors = binding.Make(lambda s: [
        (r,config.Property('comp4.shop.goods.image.converter.%s' % r)())
        for r in s.goods_image_resolutions
    ])

    def manufacturers(self):
        return self.workspace(
            "[m for m in 'shop/Manufacturer' if query.OrderBy(m.name)]"
        )

    def imageForGoods(self, goods, img_list):
        import lib4.images as img4
        images = []
        self.workspace('''[
            query.Delete(i) for i in ShopPhotogallery
                             if i.goods == goods and i.kind == kind
        ]''')
        for res, conv in self.goods_image_convertors:
            img = conv(orig)
            size = img4.getStorageSize(img)
            fd = img4.getFileDesc(img)
            fd.seek(0,0)
            obj = self.workspace.new(ShopPhotogallery,
                goods        = goods,
                resolution   = res,
                kind         = kind,
                blob         = fd,
                size         = size,
                content_type = img4.getMIME(img),
                filename     = res + '_' + filename,
                width        = img.size[0],
                height       = img.size[1],
            )
            images.append(obj)
        return images

    def preloadSupercategories(self, cat):
        curr = cat
        for c, e in self.workspace('''[
            (c,t.edges)
              for c in ShopCategory
              for t in ShopCategory.treeCache
               if t.end == cat.pkey
              and t.begin == c.pkey
              and query.OrderBy(t.edges)
        ]'''):
            if e == 0:
                assert c is cat
            else:
                curr.__dict__['parent'] = c
                curr = c
        return cat

    def preloadGoods(self, goods_lst):
        if not goods_lst:
            return []
        ws = self.workspace
        goods = [(g._oid, g) for  g in goods_lst]

        ids = [o for o,g in goods if 'name' not in g.__dict__]
        if ids:
            for o in ws('[ o for o in ShopGoods if o.pkey in ids]'):
                pass

        ids = [o for o,g in goods if 'params' not in g.__dict__]
        if ids:
            file_param_defs = [p
                    for (p, pd) in ws['shop'].goodsParams.iteritems()
                     if isinstance(pd, FileParam)]
            files = []
            params = {}
            for o in ws('''[
                o for o in ShopGoodsParam
                   if o.goods.pkey in ids
                  and query.OrderBy(o.position)
            ]'''):
                if o.value and o.param in file_param_defs:
                    try:
                        files.append(int(o.value))
                    except ValueError, TypeError:
                        pass
                params.setdefault(o.goods._oid, []).append(o)
            if files:
                files = list(ws('[f for f in ShopFile if f.pkey in files]'))
                for g in goods_lst:
                    ShopGoods.params.preload(g, params.get(g._oid, ()))

        ids = [o for o,g in goods if 'photogallery' not in g.__dict__]
        if ids:
            images = {}
            for i in ws('''[
                i for i in ShopPhotogalleryImage
                   if i.goods.pkey in ids
                  and query.OrderBy(i.position)
            ]'''):
                images.setdefault(i.goods._oid, []).append(i)
            for g in goods_lst:
                if 'photogallery' not in g.__dict__:
                    g.__dict__['photogallery'] = images.get(g._oid, ())
            images = dict(
                (img._oid, img) for imgs in images.itervalues()
                                for img in imgs)
            image_ids = images.keys()
            if image_ids:
                resolutions = {}
                for r, i in ws('''[
                    (r,i) for r in ShopPhotogalleryRes
                        for i in ShopImage
                         if r.photogallery.pkey in image_ids
                        and r.image == i
                ]'''):
                    r.__dict__.setdefault('image', i)
                    resolutions.setdefault(r.photogallery._oid, []).append(r)
                for oid, img in images.iteritems():
                    img.__dict__['resolutions'] = resolutions.get(oid, ())

        ids = [o for o,g in goods if 'news' not in g.__dict__]
        if ids:
            news = {}
            for o in ws('''[
                o for o in ShopNews
                   if o.goods.pkey in ids
                  and query.OrderBy(o.new_from)
            ]'''):
                news.setdefault(o.goods._oid, []).append(o)
            for g in goods_lst:
                if 'news' not in g.__dict__:
                    g.__dict__['news'] = news.get(g._oid, ())

        return goods_lst




    categories = binding.Make(ShopCategoryRoot)

    def categoryByID(self, id):
        return self.categories.categories.get(id)

    def categoryByMark(self, mark):
        return self.categories.marks.get(mark)



    def bookmarksForUser(self, reg_user):
        return list(self.workspace('''[
            (b,g)
                for b in ShopGoodsBookmark
                for g in ShopGoods
                 if b.reg_user == reg_user
                and b.goods == g
                and query.OrderBy(g.name)
        ]'''))

    def ordersForUser(self, reg_user):
        return list(self.workspace('''[
            (o,c)
                for o in ShopInvoice
                for c in ShopCustomer
                 if c.reg_user == reg_user
                and o.customer == c
                and query.OrderBy(query.Desc(o.created))
        ]'''))


    def order_ns(self, invoice, url):
        from lib4.urls import URL
        return dict(
            url = url,
            server = URL(url).netloc,
            invoice = invoice,
        )

    def sendmail_order_ack(self, invoice, url):
        if not invoice.customer.email:
            return
        from comp4.sendmail.api import sendmail
        params = self.order_ns(invoice, url)
        params['recipients'] = [invoice.customer.email]
        sendmail('shop_order_ack', **params)

    def sendmail_order_adm(self, invoice, url):
        import comp4.sendmail.api as sendmail
        sm = sendmail.sendmail_comp()
        params = sm.params('shop_order_adm', self.order_ns(invoice, url))
        if not params.get('recipients'):
            return
        params['sender'] = invoice.customer.email or params['recipients'][0]
        sm('shop_order_adm', **params)

    def personLetters(self):
        q = '''[query.Distinct(query.Upper(p.name[0:1]))
                       for p in 'shop/Person']'''
        return [l for l, in self.workspace(q)]
    personLetters = storage.TxnAttr(personLetters)

    def manufacturerLetters(self):
        q = '''[query.Distinct(query.Upper(p.name[0:1]))
                       for p in 'shop/Manufacturer']'''
        return [l for l, in self.workspace(q)]
    manufacturerLetters = storage.TxnAttr(manufacturerLetters)

    def newsQuery(self, current=None, limit=None, offset=None):
        if current is None:
            current = date.today()
        if isinstance(current, (tuple,list)):
            start, end = current
            qry = self.workspace.parse('''[
                g for g in 'shop/Goods'
                  for n in 'shop/News'
                   if g == n.goods
                  and g.hidden == False
                  and (
                       n.new_from <= start <= n.new_till or
                       n.new_from <= end <= n.new_till
                      )
                  and query.OrderBy(query.Desc(n.new_from))
            ]''')
        else:
            qry = self.workspace.parse('''[
                g for g in 'shop/Goods'
                  for n in 'shop/News'
                   if g == n.goods
                  and g.hidden == False
                  and n.new_from <= current <= n.new_till
                  and query.OrderBy(query.Desc(n.new_from))
            ]''')
        if limit is not None:
            qry.limit = query.Const(limit)
        if offset is not None:
            qry.offset = query.Const(offset)
        return qry

    def allNews(self, current=None, limit=None, offset=None):
        return list(self.workspace(self.newsQuery(
            current=current, limit=limit, offset=offset
        )))

    def randomNews(self, max=10, current=None):
        all = list(self.workspace(self.newsQuery(current=current)))
        news = []
        for i in range(max):
            try:
                i = random.choice(all)
                news.append(i)
                all.remove(i)
            except:
                break
        return news

