# http://en.wikipedia.org/wiki/Entity-Attribute-Value_model

from wm4.api import *
from wm4.schema.containers import PersistentList, PersistentDict
from lib4.imports import importObject

__all__ = [
    'AbstractParam', 'UnicodeParam', 'HtmlParam', 'IntegerParam', 'SelectParam',
    'BoolParam', 'FileParam', 'ImageParam', 'ReferenceParam',
    'PersistentParamList', 'PersistentParamDict', 'ParametersFeature',
]


class AbstractParam(object):
    name = None
    title = None
    description = None
    help = None
    autoview = True
    autoedit = True
    fulltext = None # weight A/B/C/D

    controller = 'wm4.web.forms.controllers.InputCtrl'
    type = unicode
    lowerBound = 0
    upperBound = 1

    def __init__(self, **kw):
        self.__dict__.update(kw)
        if not self.title:
            self.title = self.name

    def pickle(self, obj, val):
        if val is None:
            return None
        #if val == 'ahoj':
        #    raise schema.FeatureError(self.name, val)
        return unicode(val)

    def unpickle(self, obj, val):
        if val is None:
            return None
        try:
            return self.type(val)
        except:
            raise ValueError('cannot unpickle', val, self.type)

    def view(self, obj, val):
        return val

    def add(self, obj, val):
        raise NotImplementedError
    def update(self, obj, val, old):
        raise NotImplementedError
    def remove(self, obj, val):
        raise NotImplementedError

    def __repr__(self):
        return '%s(name=%r)' % (self.__class__.__name__, self.name)

class UnicodeParam(AbstractParam):
    type = unicode

class HtmlParam(AbstractParam):
    type = unicode
    controller = 'wm4.web.forms.controllers.HtmlCtrl'

class IntegerParam(AbstractParam):
    type = int


class SelectParam(AbstractParam):
    selectOptions = []
    controller = 'comp4.shop.controllers.SelectParamCtrl'

class BoolParam(AbstractParam):
    controller = 'wm4.web.forms.controllers.BoolCtrl'
    def pickle(self, obj, val):
        if val is None:
            return None
        return val and 't' or 'f'
    def unpickle(self, obj, val):
        if val is None:
            return None
        if val == 't' : return True
        else: return False

    def view(self, obj, val):
        if val is True:
            return u'ano'
        elif val is False:
            return u'ne'
        return None


class FileParam(AbstractParam):
    controller = 'comp4.shop.controllers.FileParamCtrl'

    def type(self):
        from entities import ShopFile
        return ShopFile
    type = binding.Make(type)

    def pickle(self, obj, val):
        if val is None:
            return None
        return val.pkey

    def unpickle(self, obj, val):
        from entities import ShopFile
        if val is None:
            return None
        return settings.workspace().get(ShopFile, val)

class ImageParam(AbstractParam):
    controller = 'comp4.shop.controllers.ImageParamCtrl'

    def type(self):
        from entities import ShopImage
        return ShopImage
    type = binding.Make(type)


    def pickle(self, obj, val):
        if val in (NOT_GIVEN, None):
            return None
        return val.pkey

    def unpickle(self, obj, val):
        from entities import ShopImage
        if val is None:
            return None
        return settings.workspace().get(ShopImage, val)



class ReferenceParam(AbstractParam):
    refEntity = None

    def __init__(self, **kw):
        super(ReferenceParam, self).__init__(**kw)
        self.refEntity = importObject(self.refEntity)

    def pickle(self, obj, val):
        if val is None:
            return None
        return val._oid

    def unpickle(self, obj, val):
        if val is None or not val:
            return None
        return settings.workspace().get(self.refEntity, val)




class PersistentParamList(PersistentList):

    _p_dict = None
    _p_param = None

    def _p_register(self, _):
        self._p_dict._p_changed


class PersistentParamDict(PersistentDict):

    __metaclass__ = binding.Activator

    def __init__(self, initdict=None, **kw):
        PersistentDict.__init__(self, initdict)
        self.__dict__.update(kw)

    obj = oldValues = paramDefs = binding.Require()

    def _p_wrap(self, k, v):
        pl = PersistentParamList(v)
        pl._p_dict = self
        return k, pl

    def _p_item_register(self, pl):
        self._p_changed

    def _p_register(self, _):
        self.obj.onChange({}, {})

    inst_attrs = set(['obj', 'oldValues', 'paramDefs'])

    def __getattr__(self, attr, inst_attrs=inst_attrs):
        supattr = super(PersistentParamDict, self).__getattribute__
        if attr in inst_attrs:
            return supattr(attr)
        pdef = supattr('paramDefs').get(attr)
        if pdef is None:
            raise AttributeError(attr)
        vals = self[attr]
        if pdef.upperBound == 1:
            if vals:
                return vals[0]
            else:
                return None
        else:
            return vals

    def __setattr__(self, attr, value):
        pdef = self.paramDefs.get(attr)
        if pdef is None:
            raise AttributeError(attr)
        if pdef.upperBound == 1:
            value = [value]
        self[attr] = value



class ParametersFeature(schema.Feature):
    cardinality = 'dict'
    refPrimitives = ()
    typeRef = None
    typeObj = None
    refAttr = None

    copyAttrs = schema.Feature.copyAttrs + ('typeRef', 'typeObj', 'refAttr')

    def __init__(self, typeRef, refAttr, *metadata, **kw):
        super(ParametersFeature, self).__init__(None, *metadata, **kw)
        self.typeRef = typeRef
        self.refAttr = refAttr

    def activateFeature(self, parent, attr):
        cp = super(ParametersFeature, self).activateFeature(parent, attr)
        cp.typeObj = schema.getTypeObj(cp, cp.typeRef)
        return cp

    def default(self, obj, values={}):
        paramDefs = obj.paramDefs
        values = dict((p,[]) for p in paramDefs.iterkeys())
        pdict = PersistentParamDict(
            values,
            obj       = obj,
            paramDefs = paramDefs,
            oldValues = {},
        )
        return pdict

    def oldValues(self, obj):
        ov = {}
        if obj._workspace is None:
            return ov
        for p in obj._workspace('''[
            p for p in self.typeObj
               if p.%s == obj
              and query.OrderBy(p.position)
        ]''' % self.refAttr):
            ov.setdefault(p.param, []).append(p)
        return ov

    def get(self, obj):
        odct = obj.__dict__
        attr = self.attrName
        pdict = odct.get(attr)
        if pdict is None:
            paramDefs = obj.paramDefs
            values = {}
            oldValues = self.oldValues(obj)
            for p, pd in paramDefs.iteritems():
                unpickle = pd.unpickle
                values[p] = [unpickle(obj,v.value) for v in oldValues.get(p,())]
            pdict = odct[attr] = PersistentParamDict(
                values,
                obj       = obj,
                paramDefs = paramDefs,
                oldValues = oldValues,
            )
        return pdict

    def preload(klass, obj, _oldValues):
        attr = klass.attrName
        if attr in obj.__dict__:
            return
        paramDefs = obj.paramDefs
        values = {}
        oldValues = {}
        for p in _oldValues:
            oldValues.setdefault(p.param, []).append(p)
        for p, pd in paramDefs.iteritems():
            unpickle = pd.unpickle
            values[p] = [unpickle(obj,v.value) for v in oldValues.get(p,())]
        obj.__dict__[attr] = PersistentParamDict(
            values,
            obj       = obj,
            paramDefs = paramDefs,
            oldValues = oldValues,
        )


    def set(self, obj, val):
        attr = self.attrName
        paramDefs = obj.paramDefs
        pdict = obj.__dict__.get(attr)
        if pdict is not None:
            obj.onChange({}, {})
        if isinstance(val, PersistentParamDict):
            obj.__dict__[attr] = val
        elif isinstance(val, dict):
            if pdict is None:
                oldValues = self.oldValues(obj)
            else:
                oldValues = pdict.oldValues
            for p,d in paramDefs.iteritems():
                val.setdefault(p, [])
            obj.__dict__[attr] = PersistentParamDict(
                val,
                obj       = obj,
                paramDefs = paramDefs,
                oldValues = oldValues,
            )
        else:
            raise AttributeError('Invalid params container')

    def delete(self, obj):
        # XXX delete all attributes?
        raise AttributeError('Attribute is readonly')

    def afterUpdate(self, obj):
        pdict = obj.__dict__.get(self.attrName)
        if pdict is None:
            return
        ee = schema.EntityError()
        for name, pdef in pdict.paramDefs.iteritems():
            insts = pdict.oldValues.get(name, [])
            vals = pdict[name]
            obj = pdict.obj
            ws = obj._workspace

            for idx, (inst, val) in enumerate(zip(insts, vals)):
                try:
                    pval = pdef.pickle(pdict, val)
                except schema.SchemaError, se:
                    # XXX add index for upperBound > 1
                    ee.add(se)
                    continue
                # TODO pdef.update
                change = False
                if inst.value != pval:
                    inst.value = pval
                    change = True
                if inst.position != idx:
                    inst.position = idx
                    change = True
                if change:
                    ws.flush(inst)

            li = len(insts)
            lv = len(vals)
            if lv > li:
                for i in range(li, lv):
                    try:
                        pval = pdef.pickle(pdict, vals[i])
                    except schema.SchemaError, se:
                        ee.add(se)
                        continue
                    values = dict(
                        param = pdef.name,
                        value = pval,
                        position = i,
                    )
                    values[self.refAttr] = obj
                    ws.new(self.typeObj, **values)
            elif lv < li:
                for i in range(lv, li):
                    ws.remove(insts[i])

        if ee:
            ne = schema.EntityError()
            ne.add(schema.NestedEntityError(self.attrName, ee))
            raise ne

    afterAdd = afterUpdate
