# -*- coding: utf-8 -*-

from adt4.serial import Serial
from wm4.api import *
from wm4.testing import *
from comp4.simpleparams.api import *
import os

TEST_INI = '''\
# -*- coding: utf-8
[mycomp]
params = ['foo', 'bar', 'baz']

[Plugin mycomp.params:foo]
title = "Foo"

[Plugin mycomp.params:bar]
factory = "comp4.simpleparams.parameters.IntegerParam"
title = u"Barč"
lowerBound = 1
upperBound = 3

[Plugin mycomp.params:baz]
title = u"Báz"


[myapp]
params = ['w','x','y','z']

[Plugin myapp.params:x]
title = 'X'
controller = 'os.execl'

[Plugin myapp.params:y]
factory = "comp4.simpleparams.parameters.IntegerParam"
'''


class TestParamConfig(ConfiguredTestCase):
    
    cfgURLs = (
        'pkgfile:wm4/core.ini',
        'pkgfile:wm4.web/web.ini',
        naming.data_url(TEST_INI),
    )

    def test10_ParametersDefinition(self):
        EQ = self.assertEqual
        pd = parameters_manager()['mycomp.params']
        EQ(set(pd.keys()), set(['foo', 'bar', 'baz']))

        foo = pd['foo']
        EQ(foo.name, 'foo')
        EQ(foo.title, 'Foo')
        EQ(foo.lowerBound, 0)
        EQ(foo.upperBound, 1)

        bar = pd['bar']
        EQ(bar.name, 'bar')
        EQ(bar.title, u'Barč')
        EQ(bar.lowerBound, 1)
        EQ(bar.upperBound, 3)

        pd = parameters_manager()['myapp.params']
        EQ(set(pd.keys()), set(['w','x', 'y', 'z']))

        x = pd['x']
        EQ(x.name, 'x')
        EQ(x.title, 'X')
        EQ(x.lowerBound, 0)
        EQ(x.upperBound, 1)
        EQ(x.controller, 'os.execl')

        y = pd['y']
        EQ(y.name, 'y')
        EQ(y.title, 'y')
        EQ(y.lowerBound, 0)
        EQ(y.upperBound, 1)
        EQ(y.controller, 'wm4.web.forms.controllers.InputCtrl')

    def test12_DefaultParametersDefinition(self):
        pd = parameters_manager()['unknown.params']
        self.assertEqual(pd.keys(), [])

    def test20_ParametersDefinitionCaching(self):
        pd1 = parameters_manager()['mycomp.params']
        pd2 = parameters_manager()['mycomp.params']
        assert pd1 is pd2
    




class Myparam(ParamBase):
    ent = schema.Reference('%s.Myent' % __name__, schema.Required)

class Myent(schema.Entity):
    pkey = schema.Attribute(Serial)
    name = schema.Attribute(str, schema.Required)

    params = SingleParametersFeature('myapp.params', Myparam, 'ent')

    binding.metadata(schema.PrimaryKey('pkey'))





class TestSingleParams(MockWorkspaceCase):
    
    cfgURLs = (
        'pkgfile:wm4/core.ini',
        naming.data_url(TEST_INI),
    )

    def test10_Get(self):
        EQ = self.assertEqual

        self.expects(transaction='begin')
        self.expects(
            sql="SELECT pkey, name FROM myent WHERE pkey = 2",
            res=[[2, 'e2']],
        )
        self.expects(
            sql="SELECT pkey, param, value, ent FROM myparam WHERE ent = 2",
            res=[[21, 'x', 'X', 2],
                 [24, 'y', '201', 2]],
        )
        self.expects(transaction='commit')

        storage.begin()
        g = self.ws.get(Myent, 2)
        self.failUnless(isinstance(g.params, SinglePersistentParamDict))
        EQ(sorted(g.params), ['w','x','y','z'])
        assert g.params['x'] is not g.params['y'] is not g.params['z']
        EQ(g.params['w'], None)
        EQ(g.params.w, None)
        EQ(g.params['x'], 'X')
        EQ(g.params.x, u'X')
        EQ(g.params['y'], 201)
        EQ(g.params.y, 201)
        EQ(g.params['z'], None)
        EQ(g.params.z, None)
        storage.commit()

    def test20_NewWithoutParams(self):
        EQ = self.assertEqual

        self.expects(transaction='begin')
        self.expects(sql="SELECT NEXTVAL('myent_pkey_seq')", res=[[2]])
        self.expects(sql="INSERT INTO myent (pkey, name) VALUES (2, 'e2')")
        self.expects(transaction='commit')

        storage.begin()
        g = self.ws.new(Myent, name='e2')
        EQ(g._oid, 2)
        EQ(g.pkey, 2)
        self.failUnless(isinstance(g.params, SinglePersistentParamDict))
        EQ(g.params['w'], None)
        EQ(g.params.w, None)
        EQ(g.params['x'], None)
        EQ(g.params.x, None)
        EQ(g.params['y'], None)
        EQ(g.params.y, None)
        EQ(g.params['z'], None)
        EQ(g.params.z, None)
        storage.commit()

    def test21_NewWithParamsInConstructor(self):
        EQ = self.assertEqual

        self.expects(transaction='begin')
        self.expects(sql="SELECT NEXTVAL('myent_pkey_seq')", res=[[2]])
        self.expects(sql="INSERT INTO myent (pkey, name) VALUES (2, 'e2')")
        self.expects(sql="SELECT NEXTVAL('myparam_pkey_seq')", res=[[21]])
        self.expects(sql=("INSERT INTO myparam (pkey, param, value, ent) "
                          "VALUES (21, 'x', 'X', 2)"))
        self.expects(sql="SELECT NEXTVAL('myparam_pkey_seq')", res=[[22]])
        self.expects(sql=("INSERT INTO myparam (pkey, param, value, ent) "
                          "VALUES (22, 'y', '201', 2)"))
        self.expects(transaction='commit')

        storage.begin()
        g = self.ws.new(Myent, name='e2', params={'x':'X', 'y':201})
        EQ(g._oid, 2)
        EQ(g.pkey, 2)
        self.failUnless(isinstance(g.params, SinglePersistentParamDict))
        assert g.params['x'] is not g.params['y'] is not g.params['z']
        EQ(g.params['w'], None)
        EQ(g.params.w, None)
        EQ(g.params['x'], 'X')
        EQ(g.params.x, 'X')
        EQ(g.params['y'], 201)
        EQ(g.params.y, 201)
        EQ(g.params['z'], None)
        EQ(g.params.z, None)
        storage.commit()
    

    def test22_NewWithSetitem(self):
        EQ = self.assertEqual

        self.expects(transaction='begin')
        self.expects(sql="SELECT NEXTVAL('myent_pkey_seq')", res=[[2]])
        self.expects(sql="INSERT INTO myent (pkey, name) VALUES (2, 'e2')")
        self.expects(
            sql="SELECT pkey, name FROM myent WHERE pkey = 2 FOR UPDATE",
            res=[[2, 'e2']],
        )
        self.expects(sql="SELECT NEXTVAL('myparam_pkey_seq')", res=[[21]])
        self.expects(sql=("INSERT INTO myparam (pkey, param, value, ent) "
                          "VALUES (21, 'x', 'X', 2)"))
        self.expects(sql="SELECT NEXTVAL('myparam_pkey_seq')", res=[[22]])
        self.expects(sql=("INSERT INTO myparam (pkey, param, value, ent) "
                          "VALUES (22, 'y', '201', 2)"))
        self.expects(transaction='commit')

        storage.begin()
        g = self.ws.new(Myent, name='e2')
        EQ(g._oid, 2)
        EQ(g.pkey, 2)
        EQ(g.params['x'], None)
        EQ(g.params.x, None)
        EQ(g.params['y'], None)
        EQ(g.params.y, None)
        EQ(g.params['z'], None)
        EQ(g.params.z, None)

        g.params.x = 'X'
        g.params.y = 201
        self.failUnless(isinstance(g.params, SinglePersistentParamDict))
        assert g.params['x'] is not g.params['y'] is not g.params['z']
        EQ(g.params['w'], None)
        EQ(g.params.w, None)
        EQ(g.params['x'], 'X')
        EQ(g.params.x, 'X')
        EQ(g.params['y'], 201)
        EQ(g.params.y, 201)
        EQ(g.params['z'], None)
        EQ(g.params.z, None)
        storage.commit()
    
    def test30_SetValue(self):
        EQ = self.assertEqual

        self.expects(transaction='begin')
        self.expects(
            sql="SELECT pkey, name FROM myent WHERE pkey = 2",
            res=[[2, 'e2']],
        )
        self.expects(
            sql="SELECT pkey, param, value, ent FROM myparam WHERE ent = 2",
            res=[[21, 'x', 'X', 2],
                 [22, 'y', '201', 2]],
        )
        self.expects(
            sql="SELECT pkey, name FROM myent WHERE pkey = 2 FOR UPDATE",
            res=[[2, 'e2']],
        )
        self.expects(
            sql=("SELECT pkey, param, value, ent FROM myparam "
                 "WHERE pkey = 21 FOR UPDATE"),
            res=[[21, 'x', 'X', 2]],
        )
        self.expects(sql="UPDATE myparam SET value='XX' WHERE pkey = 21")
        self.expects(sql="DELETE FROM myparam WHERE pkey = 22")
        self.expects(sql="SELECT NEXTVAL('myparam_pkey_seq')", res=[[23]])
        self.expects(sql=("INSERT INTO myparam (pkey, param, value, ent) "
                          "VALUES (23, 'z', 'Z', 2)"))
        self.expects(transaction='commit')

        storage.begin()
        g = self.ws.get(Myent, 2)
        EQ(g.params.w, None)
        EQ(g.params.x, 'X')
        EQ(g.params.y, 201)
        EQ(g.params.z, None)

        g.params.x = 'XX'
        g.params.y = None
        g.params.z = 'Z'

        EQ(g.params.w, None)
        EQ(g.params.x, 'XX')
        EQ(g.params.y, None)
        EQ(g.params.z, 'Z')
        storage.commit()

    def test32_FeatureUpdateByDict(self):
        EQ = self.assertEqual

        self.expects(transaction='begin')
        self.expects(
            sql="SELECT pkey, name FROM myent WHERE pkey = 2",
            res=[[2, 'e2']],
        )
        self.expects(
            sql="SELECT pkey, param, value, ent FROM myparam WHERE ent = 2",
            res=[[20, 'w', 'W', 2],
                 [21, 'x', 'X', 2],
                 [22, 'y', '201', 2]],
        )
        self.expects(
            sql="SELECT pkey, name FROM myent WHERE pkey = 2 FOR UPDATE",
            res=[[2, 'e2']],
        )
        self.expects(sql="DELETE FROM myparam WHERE pkey = 20")
        self.expects(
            sql=("SELECT pkey, param, value, ent FROM myparam "
                 "WHERE pkey = 21 FOR UPDATE"),
            res=[[21, 'x', 'X', 2]],
        )
        self.expects(sql="UPDATE myparam SET value='XX' WHERE pkey = 21")
        self.expects(sql="DELETE FROM myparam WHERE pkey = 22")
        self.expects(sql="SELECT NEXTVAL('myparam_pkey_seq')", res=[[23]])
        self.expects(sql=("INSERT INTO myparam (pkey, param, value, ent) "
                          "VALUES (23, 'z', 'Z', 2)"))
        self.expects(transaction='commit')

        storage.begin()
        g = self.ws.get(Myent, 2)
        EQ(g.params.w, 'W')
        EQ(g.params.x, 'X')
        EQ(g.params.y, 201)
        EQ(g.params.z, None)

        g.params = dict(x='XX', y=None, z='Z')

        EQ(g.params.w, None) # w is deleted
        EQ(g.params.x, 'XX')
        EQ(g.params.y, None)
        EQ(g.params.z, 'Z')
        storage.commit()



suite = Suite(
    TestParamConfig,
    TestSingleParams,
)

if __name__ == '__main__':
    Main()
