# -*- coding: ISO-8859-2 -*-

from unittest import *
from wm4.api import *
from comp4.users.api import *
from comp4.users.tests.common import *



class TestUser(TestBase):
    
    def test10_New(self):
        ws = self.ws
        storage.begin(ws)
        u = ws.User.new(name=u'Jos', login='jose10', email='jose@example.com')
        storage.commit(ws)
        self.assertEqual(schema.OID.get(u), 4)
        assert 4 in ws.User.cache
        storage.begin(ws)
        u = ws.User[4]
        self.assertEqual(schema.OID.get(u), 4)
        self.assertEqual(u.name, u'Jos')
        storage.commit(ws)

    def test11_GetUncommited(self):
        ws = self.ws
        storage.begin(ws)
        u = ws.User.new(name=u'Jos', login='jose11', email='jose@example.com')
        uid = schema.OID.get(u)
        u = ws.User[uid]
        self.assertEqual(u.name, u'Jos')
        self.assertEqual(u.login, u'jose11')
        storage.commit(ws)

    def test20_Update(self):
        ws = self.ws
        storage.begin(ws)
        u = ws.User[3]
        u.name = u'Jos'
        storage.commit(ws)
        assert schema.OID.get(u) == 3
        assert 3 in ws.User.cache
        storage.begin(ws)
        u = ws.User[3]
        assert schema.OID.get(u) == 3
        assert u.name == u'Jos'
        assert u.email == '3@example.com'
        storage.commit(ws)

    def test21_UpdatingOID(self):
        ws = self.ws
        storage.begin(ws)
        u = ws.User[3]
        u.name = u'Jos'
        u.uid = 999
        storage.commit(ws)
        assert schema.OID.get(u) == 999
        assert 999 in ws.User.cache
        assert 3 not in ws.User.cache
        storage.begin(ws)
        u = ws.User[999]
        assert schema.OID.get(u) == 999
        assert u.name == u'Jos'
        assert u.email == '3@example.com'
        storage.commit(ws)

    def test30_Delete(self):
        ws = self.ws
        storage.begin(ws)
        ws.User.remove(ws.User[2])
        assert 2 not in ws.User.cache
        self.assertRaises(KeyError, ws.User.__getitem__, 2)
        storage.commit(ws)
        storage.begin(ws)
        assert 2 not in ws.User.cache
        self.assertRaises(KeyError, ws.User.__getitem__, 2)
        storage.commit(ws)




class TestUserGroupStorage(TestCase):
    
    def setUp(self):
        self.ws = ws = getWS()
        storage.begin(ws)
        for tab in ('user', 'group', 'user_group'):
            ws.sqlCon('DELETE FROM "%s"' % tab)
        for i in range(1,4):
            ws.sqlCon((
              'INSERT INTO "user" (uid,login,name,email) '
              "VALUES (%d,'login%d','name%d','%d@mail.com')"
              ) % (i, i, i, i)
            )
        for i in range(1,3):
            ws.sqlCon(
              '''INSERT INTO "group" (gid,name) VALUES (%d,'group%d')''' % (i,i)
            )
        ws.sqlCon(
            'INSERT INTO "user_group" ("user","group") VALUES (2,2)'
        )
        storage.commit(ws)
    
    def tearDown(self):
        pass

    def test01_New(self):
        ws = self.ws
        storage.begin(ws)
        obj = ws.UserGroup.new(user=ws.User[1], group=ws.Group[1])
        self.assertEqual(schema.OID.get(obj.user), 1)
        self.assertEqual(schema.OID.get(obj.group), 1)
        storage.commit(ws)

    def test10_Get(self):
        ws = self.ws
        storage.begin(ws)
        obj = ws.UserGroup[ws.User[2], ws.Group[2]]
        self.assertEqual(schema.OID.get(obj), (2,2))
        self.assertEqual(schema.OID.get(obj.user), 2)
        self.assertEqual(schema.OID.get(obj.group), 2)
        storage.commit(ws)

    def test11_GetGhosts(self):
        ws = self.ws
        storage.begin(ws)
        obj = ws.UserGroup[
            ws.User.preload({'uid':2}), ws.Group.preload({'gid':2})
        ]
        self.assertEqual(schema.OID.get(obj), (2,2))
        self.assertEqual(schema.OID.get(obj.user), 2)
        self.assertEqual(schema.OID.get(obj.group), 2)
        storage.commit(ws)

    def test14_FindAllUsers(self):
        ws = self.ws
        storage.begin(ws)
        result = list(ws.User.find())
        self.assertEqual(len(result), 3)
        result.sort(lambda x,y: cmp(x.uid, y.uid))
        for i, u in enumerate(result):
            i = i + 1
            self.assertEqual(schema.OID.get(u), i)
            self.assertEqual(u.uid, i)
            self.assertEqual(u.name, u'name%d' % i)
            self.assertEqual(u.login, 'login%d' % i)
            self.assertEqual(u.email, '%d@mail.com' % i)
        storage.commit(ws)

    def test15_FindAllUserGroups(self):
        ws = self.ws
        storage.begin(ws)
        result = list(ws.UserGroup.find())
        self.assertEqual(len(result), 1)
        ug = result[0]
        self.assertEqual(schema.OID.get(ug), (2,2))
        u = ug.user
        self.assertEqual(schema.OID.get(u), 2)
        self.assertEqual(u.uid, 2)
        self.assertEqual(u.name, u'name2')
        self.assertEqual(u.login, 'login2')
        self.assertEqual(u.email, '2@mail.com')
        g = ug.group
        self.assertEqual(schema.OID.get(g), 2)
        self.assertEqual(g.gid, 2)
        self.assertEqual(g.name, u'group2')
        storage.commit(ws)
        
    def test16_UserReferenceEnd(self):
        ws = self.ws
        storage.begin(ws)
        u = ws.User[2]
        _groups = u._groups
        self.assertEqual(len(_groups), 1)
        assert _groups[0] is ws.UserGroup[u,ws.Group[2]]
        u = ws.User[1]
        _groups = u._groups
        self.assertEqual(len(_groups), 0)
        storage.commit(ws)

    def test20_Update(self):
        ws = self.ws
        storage.begin(ws)
        obj = ws.UserGroup[ws.User[2], ws.Group[2]]
        obj.user = ws.User[1]
        storage.commit(ws)
        storage.begin(ws)
        self.assertEqual(schema.OID.get(obj), (1,2))
        self.assertEqual(schema.OID.get(obj.user), 1)
        self.assertEqual(schema.OID.get(obj.group), 2)
        storage.commit(ws)

    def test30_Delete(self):
        ws = self.ws
        storage.begin(ws)
        ws.UserGroup.remove(ws.UserGroup[ws.User[2], ws.Group[2]])
        storage.commit(ws)



def suite():
    return TestSuite((
        makeSuite(TestUser),
        makeSuite(TestUserGroupStorage),
    ))



if __name__ == '__main__':

    main(defaultTest='suite')
