# Thread-safe caching object with file and HTTP implementations
# http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/302997
# http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/252524
# http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/521871
# http://pypi.python.org/pypi/GenericCache

import time
from heapq import *


__all__ = ['LRUCache']


class _LRUNode(object):

    __slots__ = 'key', 'value', 'stamp'

    def __init__(self, key, value, stamp):
        self.key = key
        self.value = value
        self.stamp = stamp

    def __cmp__(self, other):
        return cmp(self.stamp, other.stamp)

    def __repr__(self):
        return '_LRUNode(%r,%r)' % (self.key, self.value)



class LRUCache(object):

    _stamp = time.time

    def __init__(self, size=10):
        self._size = size
        self._cache = {}
        self._lru = []

    def __setitem__(self, key, value):
        node = self._cache.get(key)
        stamp = self._stamp()
        if node is None:
            node = _LRUNode(key, value, stamp)
            self._cache[key] = node
            heappush(self._lru, node)
            self.__shrink()
        else:
            node.stamp = stamp
            node.value = value
            heapify(self._lru)

    def __getitem__(self, key):
        node = self._cache[key]
        node.stamp = self._stamp()
        heapify(self._lru)
        return node.value

    def __delitem__(self, key):
        node = self._cache[key]
        del self._cache[key]
        self._lru.remove(node)
        heapify(self._lru)

    def __shrink(self):
        while len(self._lru) > self._size:
            node = heappop(self._lru)
            del self._cache[node.key]

    def __contains__(self, key):
        return key in self._cache

    def __len__(self):
        return len(self._lru)

    def iterkeys(self, key): return self._cache.iterkeys()
    def itervalues(self, key): return self._cache.itervalues()
    def iteritems(self, key): return self._cache.iteritems()
