#-*- coding: iso-8859-2 -*-
#
# The Common Gateway Interface (CGI) Version 1.1
# http://www.ietf.org/rfc/rfc3875

import os
from lib4.symbols import NOT_GIVEN

__all__ = [
    'sane_environ', 'parse_auth', 'unparse_auth',
    'etag_create', 'http_time', 'is_fast_cgi',
    'encode_qs', 'encode_multipart',
    'TestUpload', 'test_environ',
]


def sane_environ(env):
    '''Modifies enviroment inplace'''

    # Apache sometimes set REDIRECT_* variables that hold true values
    for key, val in env.items():
        if key.startswith('REDIRECT_'):
            env[key[9:]] = val

    # HTTP_X_FORWARDED_FOR = 80.251.240.138
    # HTTP_X_FORWARDED_HOST = www.starobrno.cz
    # HTTP_X_FORWARDED_SERVER = www.starobrno.cz
    forwFor = env.get('HTTP_X_FORWARDED_FOR')
    if forwFor:
        from lib4.validators import checkIP
        if checkIP(forwFor):
            env['HTTP_Y_REMOTE_ADDR'] = env.get('REMOTE_ADDR','')
            env['REMOTE_ADDR'] = forwFor.split(',')[-1].strip()
    forwHost = env.get('HTTP_X_FORWARDED_HOST')
    if forwHost:
        env['HTTP_Y_HTTP_HOST'] = env.get('HTTP_HOST','')
        env['HTTP_HOST'] = forwHost.split(',')[0].strip()
    forwServ = env.get('HTTP_X_FORWARDED_SERVER')
    if forwServ:
        env['HTTP_Y_SERVER_NAME'] = env.get('SERVER_NAME','')
        env['SERVER_NAME'] = forwServ.split(',')[0].strip()

    # If you use Apache directive ScriptAliasMatch use also SetEnv:
    #    ScriptAliasMatch '^.*$'               '/web/Publisher/apache.cgi'
    #    SetEnv           'SCRIPT_ALIAS_MATCH' '/'
    # or
    #    ScriptAliasMatch '^/app/.*$'          '/web/Publisher/apache.cgi'
    #    SetEnv           'SCRIPT_ALIAS_MATCH' '/app/'
    saMatch = env.get('SCRIPT_ALIAS_MATCH')
    if saMatch is not None:
        reqURI = env['REQUEST_URI'].split('?')[0]
        if not reqURI.startswith(saMatch):
            raise ValueError(
                "REQUEST_URI doesn't match SCRIPT_ALIAS_MATCH", reqURI, saMatch)
        if saMatch == '/':
            env['SCRIPT_NAME'] = ''
        else:
            env['SCRIPT_NAME'] = saMatch
        pathInfo = reqURI[len(saMatch):]
        if not pathInfo.startswith('/'):
            pathInfo = '/'+pathInfo
        env['PATH_INFO'] = pathInfo

    # Parse HTTP_AUTHORIZATION into REMOTE_USER and REMOTE_PASSWORD
    if 'HTTP_CGI_AUTHORIZATION' in env:
        env.setdefault('HTTP_AUTHORIZATION', env['HTTP_CGI_AUTHORIZATION'])
    auth = env.get('HTTP_AUTHORIZATION')
    if auth is not None:
        user, pwd = parse_auth(auth)
        if user is not None :
            env.setdefault('REMOTE_USER',user)
        if pwd is not None:
            env.setdefault('REMOTE_PASSWORD',pwd)

def parse_auth(val):
    if not val:
        return None, None
    parts = str(val).split()
    if len(parts) != 2:
        return None, None
    encoded = parts[1].decode('base64')
    if not ':' in encoded:
        return None, None
    return encoded.split(':', 1)

def unparse_auth(login, password):
    return 'Basic ' + ('%s:%s' % (login, password)).strip().encode('base64')



def etag_create(filename, stat=None):
    '''ETag is created from inode-size-mtime'''
    if stat is None:
        stat = os.stat(filename)
    return '"%x-%x-%x"' % (stat.st_ino, stat.st_size, stat.st_mtime)



HTTP_TIME_FMT = '%a, %d %b %Y %H:%M:%S GMT'
HTTP_COOKIE_FMT = '%a, %d-%m-%Y %H:%M:%S GMT'

def http_time(tstamp, fmt=HTTP_TIME_FMT):
     # XXX strftime is locale aware
    import datetime, time
    if isinstance(tstamp, (int,float,long)):
        return time.strftime(fmt,time.gmtime(tstamp))
    elif isinstance(tstamp, datetime.datetime):
        tstamp = datetime.datetime(*(tstamp.utctimetuple()[:6]))
    elif isinstance(tstamp, datetime.date):
        tstamp = datetime.datetime(*(tstamp.timetuple()[:6]))
    else:
        raise ValueError('Unsupported type', tstamp)
    return tstamp.strftime(fmt)

def http_time_parse(tstr):
    raise NotImplementedError


def is_fast_cgi(stdin):
    import socket
    for family in (socket.AF_UNIX, socket.AF_INET):
        try:
            s=socket.fromfd(stdin.fileno(),family,socket.SOCK_STREAM)
            if not s.getsockname():
                # Socket doesn't have an address; it might be a BSD-style
                # socketpair() pipe, as used on Mac OS/X and others
                continue
        except:
            pass
        else:
            return True
    return False





def encode_value(v, enc):
    if isinstance(v, unicode):
        return v.encode(enc)
    else:
        return str(v)

def encode_qs(params, enc='utf-8'):
    from urllib import quote
    qs = []
    for a,v in params.iteritems():
        if isinstance(v, (list,tuple)):
            for lv in v:
                qs.append('%s=%s' % (a,quote(encode_value(lv, enc))))
        else:
            qs.append('%s=%s' % (a,quote(encode_value(v, enc))))
    return '&'.join(qs)

def encode_multipart(params, enc='utf-8'):
    from ClientForm import MimeWriter
    from cStringIO import StringIO
    http_hdrs = []
    data = StringIO()
    mw = MimeWriter(data, http_hdrs)
    mw.startmultipartbody("form-data", add_to_http_hdrs=True, prefix=0)
    for a,v in params.iteritems():
        if isinstance(v,(list,tuple)):
            vlist = v
        else:
            vlist = [v]
        for v in vlist:
            p = mw.nextpart()
            if isinstance(v, TestUpload):
                p.addheader(
                    'Content-disposition',
                    'form-data; name="%s"; filename="%s"' % (a, v.filename),
                    prefix=1
                )
                if v.size is not None:
                    p.addheader('Content-length', str(v.size))
                if v.content_type not in  (None, NOT_GIVEN):
                    f = p.startbody(v.content_type, prefix=0)
                else:
                    f = p.startbody(prefix=0)
                f.write(open(v.path,'rb').read())
            else:
                p.addheader(
                    'Content-disposition',
                    'form-data; name="%s"' % a,
                    prefix=1
                )
                f = p.startbody(prefix=0)
                f.write(encode_value(v, enc))
    mw.lastpart()
    return data.getvalue(), http_hdrs[0][1]


CONTENT_ENCODING_2_MIME = {'gzip': 'application/gzip'}

class TestUpload:

    def __init__(
        self, path, filename=NOT_GIVEN, content_type=NOT_GIVEN, size=NOT_GIVEN
    ):
        self.path = os.path.abspath(os.path.normpath(path))
        if content_type is NOT_GIVEN:
            from content_types import mimeService
            ct, enc = mimeService().guess_type(path)
            if enc:
                content_type = CONTENT_ENCODING_2_MIME.get(enc)
            elif ct:
                content_type = ct
            if not content_type:
                content_type = 'application/octet-stream'
        if size is NOT_GIVEN:
            size = os.stat(path).st_size
        if filename is NOT_GIVEN:
            filename = os.path.split(path)[1]
        self.content_type = content_type
        self.size = size
        self.filename = filename






def test_environ(
    url='/', env={}, get={}, post={}, cookies={}, auth=None, enc='utf-8',
    stdin=None,
):
    from urlparse import urlsplit, urlunsplit
    from StringIO import StringIO
    from urllib import quote

    scheme, hostport, path, query_string, target = urlsplit(url)
    if sum([1 for x in (query_string, get, post, stdin) if x]) > 1:
        raise ValueError(
            'Too much parameters. Only one of "get","post" or "query string" '
            'in url can be specified.'
        )
    if get:
        query_string = encode_qs(get, enc)

    # host and port
    if scheme == 'https':
        port, https = '443', 'on'
    else:
        port, https = '80', 'off'
    if hostport:
        if ':' in hostport:
            host, port = hostport.split(':',1)
        else:
            host = hostport
    else:
        host = 'localhost'

    # accept charset
    if 'utf-8' == enc.lower():
        accept_charset = 'utf-8,*;q=0.7'
    else:
        accept_charset = '%s,utf-8;q=0.7,*;q=0.7' % enc

    # method
    method = env.get('REQUEST_METHOD')
    if method is None:
        if post or stdin is not None:
            method = 'POST'
        else:
            method = 'GET'

    # default enviroment values
    e = dict(
        GATEWAY_INTERFACE   = 'CGI/1.1',
        HTTPS               = https,
        HTTP_ACCEPT_CHARSET = accept_charset,
        HTTP_HOST           = hostport,
        PATH_INFO           = path,
        QUERY_STRING        = query_string,
        REQUEST_METHOD      = method,
        REQUEST_URI         = urlunsplit(('','',path,query_string,'')),
        SCRIPT_NAME         = '/',
        SERVER_NAME         = host,
        SERVER_PORT         = port,
        SERVER_PROTOCOL     = 'HTTP/1.0',
        REMOTE_ADDR         = '127.0.0.1',
    )

    # wsgi defaults
    e.setdefault('wsgi.version', (1,0))
    e.setdefault('wsgi.run_once', 0)
    e.setdefault('wsgi.multithread', 0)
    e.setdefault('wsgi.multiprocess', 0)

    inp = e.setdefault('wsgi.input', StringIO())
    e.setdefault('wsgi.errors', StringIO())
    e.setdefault('wsgi.url_scheme', https == 'on' and 'https' or 'http')

    # post parameters
    # create query string from get parameters
    if method == 'GET':
        if stdin is not None:
            e['QUERY_STRING'] = stdin.strip()

    elif method == 'POST':

        ct = env.get('CONTENT_TYPE','application/x-www-form-urlencoded')
        if stdin is None:
            for v in post.itervalues():
                if isinstance(v, TestUpload):
                    if ct != 'multipart/form-data':
                        if 'CONTENT_TYPE' in env:
                            raise ValueError('Invalid CONTENT_TYPE')
                        else:
                            ct = 'multipart/form-data'
                    break
            if ct == 'application/x-www-form-urlencoded':
                inp.write(encode_qs(post, enc))
            else:
                data, ct = encode_multipart(post, enc)
                inp.write(data)
                if 'CONTENT_TYPE' in env:
                    del env['CONTENT_TYPE']

        else:
            inp.write(stdin)
            if ct.startswith('multipart/form-data'):
                b = stdin.splitlines()[0][2:].strip()
                ct = 'multipart/form-data; boundary=%s' % b
                if 'CONTENT_TYPE' in env:
                    del env['CONTENT_TYPE']
        e['CONTENT_TYPE'] = ct
        e['CONTENT_LENGTH'] = str(inp.tell())
        inp.seek(0,0)

    # authorization info
    if auth:
        login, password = auth
        e['HTTP_AUTHORIZATION'] = unparse_auth(login, password)
        e['AUTH_TYPE'] = 'Basic'

    # cookies
    if cookies:
        import Cookie
        sc = Cookie.SimpleCookie()
        for c,v in cookies.iteritems():
            sc[c] = unicode(v).encode(enc)
        e['HTTP_COOKIE'] = '; '.join([v.OutputString() for v in sc.values()])

    # insert user supplied values into new environment
    e.update(env)
    return e, inp
