# -*- coding: utf-8 -*-

import sys

__all__ = ['interactive_loop', 'debug_memory', 'rusage']


def interactive_loop():
    '''Enters interactive python interpreter loop for debuging.
       When user exit prompt, program can continue further.
    '''
    import ctypes
    libc = ctypes.cdll.LoadLibrary('libc.so.6')
    PyRun_InteractiveLoop = ctypes.pythonapi.PyRun_InteractiveLoop
    stdin = ctypes.c_void_p.in_dll(libc, 'stdin')
    res = PyRun_InteractiveLoop(stdin, "<stdin>")
    if res != 0:
        raise RuntimeError('Cannot run PyRun_InteractiveLoop')


def debug_memory(stream=sys.stdout):
    import gc
    st = {}
    objs = gc.get_objects()
    print >>stream, 'OBJECTS', len(objs)
    for o in objs:
        c = type(o)
        st[c] = st.get(c,0) + 1
    for i,(c,n) in enumerate(reversed(
        list(sorted(st.items(), key=lambda (k,v):v))[-30:]
    )):
        print >>stream, '%03d' % (i+1), n, c
    del objs



def rusage(pids, delay=0.1):
    '''Collects rusage statistics for given PID list. All processes
       must be children of process that call rusage function. Only
       suitable for processes that runs longer time.
    '''
    import os, sys, time
    if hasattr(os, 'wait4'):
        wait4 = os.wait4
    else:
        from wait4 import wait4

    start = time.time()
    pagesize = os.sysconf('SC_PAGESIZE')

    statm = 'size', 'resident', 'share', 'text', 'lib', 'data', 'dt'
    timeu = 'real', 'user', 'sys'
    units = statm + timeu + ('exit','pid')

    rusage = dict(
        (p, dict((u,0) for u in units)) for p in pids
    )

    running = list(pids)
    while running:

        remove = []
        for p in running:
            try:
                s = open('/proc/%d/stat' % p).read().strip().split()
                m = open('/proc/%d/statm' % p).read().strip().split()
            except:
                remove.append(p)
            else:
                if s[2] == 'Z':
                    remove.append(p)
                else:
                    pu = rusage[p]
                    for u,v in zip(statm, m):
                        v = int(v)
                        if v > pu[u]:
                            pu[u] = v

        for p in remove:
            (pid, exit, res) = wait4(p, 0)
            pu = rusage[p]
            pu['pid'] = p
            pu['exit'] = exit
            pu['user'] = res.ru_utime
            pu['sys'] = res.ru_stime
            pu['real'] = time.time() - start
            for u in statm:
                pu[u] = pu[u] * pagesize
            running.remove(p)

        time.sleep(delay)

    avg = dict((u,0) for u in units)
    num = len(rusage)
    for p,u in rusage.items():
        for k in statm + ('sys', 'user'):
            avg[k] += u[k]
        for k in timeu:
            v = u['real']
            if v > avg['real']:
                avg['real'] = v
    avg['pid'] = None
    rusage[None] = avg
    return rusage





if __name__ == '__main__':

    if 0:
        print 'before'
        interactive_loop()
        print 'after'

    if 1:
        import time, os, sys
        from lib4.procutils import ForkCall

        def fork(n, s):
            d = 'a' * n
            time.sleep(s)
            a = 0
            for i in xrange(1000000):
                a = a + 1
            sys.exit(1)

        f1 = ForkCall(fork, 100, 2)
        f2 = ForkCall(fork, 10000000, 1)

        pid1 = os.fork()
        if pid1 == 0:
            a = 0
            for i in xrange(1000000):
                a = a + 1
            sys.exit(2)

        pid2 = os.fork()
        if pid2 == 0:
            for i in xrange(1000):
                time.sleep(0.001)
            sys.exit(3)

        mu = rusage([f1.pid, f2.pid, pid1, pid2])
        for k,v in mu.items():
            print '-----------', k, '-------------'
            for x,y in sorted(v.items()):
                print x,y
