from itertools import *

__all__ = [
    # inherited from itertools
    'chain', 'count', 'cycle', 'dropwhile', 'ifilter', 'ifilterfalse',
    'imap', 'islice', 'izip', 'repeat', 'starmap', 'takewhile',
    # compatibility
    'sorted', 'reversed', 'product', # groupby, tee
    # defined here
    'findFirst', 'findLast', 'Item', 'iterDecorator', 'flatten',
    'pad', 'sequenceToTableVertically', 'sequenceToTableHorizontally',
    'updateSortingAttr', 'unique', 'split_seq'
]

# http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/363051
# http://aspn.activestate.com/ASPN/Mail/Message/python-tutor/2302348
# http://aspn.activestate.com/ASPN/Mail/Message/python-tutor/2302231
# bisimilar_lists
# http://mail.python.org/pipermail/python-dev/2003-October/039445.html


try :
    sorted = sorted
except NameError:
    def sorted(iterable, cmp=None, key=None, reverse=None):
        if key:
            tmp = [(key(elt), elt) for elt in iterable]
        else:
            tmp = list(iterable)
        if cmp:
            tmp.sort(cmp)
        else:
            tmp.sort()
        if key:
            tmp = [b for (a,b) in tmp]
        if reverse:
            tmp.reverse()
        return tmp

try:
    reversed = reversed
except NameError:
    def reversed(iterable):
        tmp = list(iterable)
        tmp.reverse()
        return tmp

try:
    product = product
except NameError:
    def product(*iterables):
        '''Cartesian product of input iterables.

           Equivalent to nested for-loops.
           For example, product(A, B) returns the same as:

            ((x,y) for x in A for y in B).

           The leftmost iterators are in the outermost for-loop, so the output tuples
           cycle in a manner similar to an odometer (with the rightmost element changing
           on every iteration).

           product('ab', range(3)) --> ('a',0) ('a',1) ('a',2) ('b',0) ('b',1) ('b',2)
           product((0,1), (0,1), (0,1)) --> (0,0,0) (0,0,1) (0,1,0) (0,1,1) (1,0,0) ...")
        '''
        if not iterables:
            return
        i = 0
        imax = len(iterables) - 1
        val = [None] * (imax+1)
        iters = [iter(seq) for seq in iterables]
        while i >= 0:
            for val[i] in iters[i]:
                if i == imax:
                    yield tuple(val)
                else:
                    i += 1
                    break
            else:
                iters[i] = iter(iterables[i])
                i -= 1





class itercontext:
    '''An iterator that allows accessing previous and next elements
    '''
    # http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/544296
    def __init__(self,it,countprevious,countcomingnext):
        '''initializes the iterator
        it: the underlying iterator
        countprevious: number of elements to keep after being visited
        countcomingnext: number of elements to look ahead
        '''
        from collections import deque
        self.iter=iter(it)
        # current element
        self.current=None
        self.existcurrent=False
        # list of previous and next elements
        self.comingnext=deque()
        self.previous=deque()
        self.countprevious=countprevious
        try:
            for i in range(countcomingnext):
                self.comingnext.append(self.iter.next())
        except StopIteration:
            pass

    def __iter__(self):
        return self

    def next(self):
        try:
            self.comingnext.append(self.iter.next())
        except StopIteration:
            pass # if the underlying iterator is exhausted, empty comingnext
        if len(self.comingnext)==0:
            raise StopIteration # if no more item ahead, stop
        if self.existcurrent: self.previous.append(self.current)
        self.current=self.comingnext.popleft()
        self.existcurrent=True
        if len(self.previous)>self.countprevious:
            self.previous.popleft()
        return self.current

    def get_coming_next(self):
        return self.comingnext

    def get_previous(self):
        return self.previous

    def list_coming_next(self):
        return list(self.comingnext)

    def list_previous(self):
        return list(self.previous)



def unique(seq, stable=False):
    """unique(seq, stable=False): return a list of the elements in seq
    in arbitrary order, but without duplicates.
    If stable=True it keeps the original element order
    (using slower algorithms).
    """
    # http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/438599
    # Developed from Tim Peters version:
    #   http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/52560

    # Special case of an empty s:
    if not seq: return []

    # if it's a set:
    if isinstance(seq, set):
        return list(seq)

    if stable:
        # Try with a set:
        seqSet = set()
        result = []
        try:
            for e in seq:
                if e not in seqSet:
                    result.append(e)
                    seqSet.add(e)
        except TypeError:
            pass # move on to the next method
        else:
            return result

        # Since you can't hash all elements, use a bisection on sorted elements
        result = []
        sortedElem = []
        from bisect import bisect_left, insort_left
        try:
            for elem in seq:
                pos = bisect_left(sortedElem, elem)
                if pos >= len(sortedElem) or sortedElem[pos] != elem:
                    insort_left(sortedElem, elem)
                    result.append(elem)
        except TypeError:
            pass  # Move on to the next method
        else:
            return result

    else: # Not stable
        # Try using a set first, because it's the fastest and it usually works
        try:
            u = set(seq)
        except TypeError:
            pass # move on to the next method
        else:
            return list(u)

        # Elements can't be hashed, so bring equal items together with a sort
        # and remove them out in a single pass.
        try:
            t = sorted(seq)
        except TypeError:
            pass  # Move on to the next method
        else:
            return [elem for elem,group in groupby(t)]

    # Brute force:
    result = []
    for elem in seq:
        if elem not in result:
            result.append(elem)
    return result


def imerge(*iterables):
    '''Merge multiple sorted inputs into a single sorted output.

    >>> list(imerge([1,3,5,7], [0,2,4,8], [5,10,15,20], [], [25]))
    [0, 1, 2, 3, 4, 5, 5, 7, 8, 10, 15, 20, 25]
    '''
    # http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/491285
    import heapq
    heappop, heappush = heapq.heappop, heapq.heappush
    its = map(iter, iterables)
    h = []
    for it in its:
        try:
            v = it.next()
        except StopIteration:
            continue
        heappush(h, (v, it.next))
    while h:
        v, next = heappop(h)
        yield v
        try:
            v = next()
        except StopIteration:
            continue
        heappush(h, (v, next))



def findFirst(iterable, pred=None, default=None):
    for item in ifilter(pred, iterable):
        return item
    return default


def findLast(iterable, pred=None, default=None, reverse=True):
    if reverse:
        for item in ifilter(pred, reversed(iterable)):
            return item
        return default
    else:
        last = default
        for item in ifilter(pred, iterable):
            last = item
        return last



class Item(object):

    __slots__ = 'item', 'isLast', 'index', 'absIndex', 'itemClass'

    def __init__(self, item, index, absIndex, isLast):
        self.item = item
        self.index = index
        self.absIndex = absIndex
        self.isLast = isLast
        self.itemClass = None

    obj     = property(lambda s: s.item)
    isFirst = property(lambda s: s.index == 0)
    isEven  = property(lambda s: s.index % 2 == 1)
    isOdd   = property(lambda s: s.index % 2 == 0)

    def __getitem__(self, key):
        if key == 0:
            return self
        if key == 1:
            return self.item
        raise IndexError(key)


def iterDecorator(iterable, offset=0, classifier=None, factory=Item):
    iterable = list(iterable)
    last = len(iterable) - 1
    if classifier is None:
        classifier = lambda item: item.isEven and 'even' or 'odd'
    for idx, item in enumerate(iterable):
        d = factory(
            item      = item,
            index     = idx,
            absIndex  = idx+offset,
            isLast    = idx == last,
        )
        d.itemClass = classifier(d)
        yield d




def flatten(iterable):
    for i in iterable:
        if isinstance(i, basestring):
            yield i
        else:
            try:
                it = iter(i)
            except TypeError:
                yield i
            else:
                for j in flatten(it):
                    yield j

def pad(iterable, padding, minLength):
    count = 0
    for item in iterable:
        count += 1
        yield item
    for i in xrange(minLength - count):
        yield padding


def sequenceToTableVertically(iterable, columns, emptyCell=None):
    """Creates table from sequence.

       Table is arranged as follows:

            for row in sequenceToTableVertically(range(7), 3):
                print row

            [0, 3, 5]
            [1, 4, 6]
            [2, None, None]
    """
    ilen = len(iterable)
    if ilen < 1:
        return
    rows = (ilen+columns-1) / columns
    rest = rows * columns - ilen
    # Last row deltas
    last = [0] * columns
    for c in xrange(columns - rest, columns):
        last[c] = 1
    # Prepare offset vector
    o = 0
    offsets = []
    for c in range(columns):
        offsets.append(o)
        o += rows - last[c]
    # Generate all rows except last
    for r in range(rows-1):
        yield [iterable[o+r]  for o in offsets]
    # Treat differently last row
    lastRow = []
    r = rows - 1
    for l, o in zip(last, offsets):
        i = o + r
        if l > 0:
            lastRow.append(emptyCell)
        else:
            lastRow.append(iterable[i])
    yield lastRow


def sequenceToTableVertically2(iterable, columns, emptyCell=None):
    # Similar but produces different results
    # http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/426331
    inp = list(iterable)
    iLen = len(iterable)
    rows = (iLen+columns-1) / columns
    oLen = rows * columns
    if iLen < oLen:
        inp.extend(repeat(emptyCell, oLen-iLen))
    out = []
    for i in range(rows):
        out.append(inp[i::rows])
    return out


def sequenceToTableHorizontally(iterable, columns, emptyCell=None):

    """Creates table from sequence.

       Table is arranged as follows:

            for row in sequenceToTableHorizontally(range(7), 3):
                print row

            [0, 1, 2]
            [3, 4, 5]
            [6, None, None]
    """

    it = chain(iterable, repeat(emptyCell, columns-1))
    for row in izip(*[it for c in range(columns)]):
        yield row









def updateSortingAttr(
    items, start_value=1000, increment=20, min_value=100, key=lambda item: item
):
    '''Update sorting attr on items of particular sequence so attributes
       have values that corresponding to order in sequence.

       This function tries to make such updates so that changes
       of sorting attribute are minimal. IOW it tries to keep
       existing values as possible.
    '''
    l = len(items)
    i = 0
    last = None

    if not l:
        return

    while i < l:

        item = items[i]
        sorting = key(item)
        if sorting is not None:
            if sorting <= last:
                last = last + increment
                yield last, item
            elif sorting <= min_value:
                yield min_value, item
                last = min_value
            else:
                last = sorting
                yield sorting, item
            i += 1
            continue

        next = None
        rlen = 1
        for j in range(i+1,l):
            next = key(items[j])
            if next is not None:
                break
            rlen += 1

        if next is None:
            incr = increment
            if last is None:
                cval = start_value
            else:
                cval = last + increment

        else:
            if last is None:
                cval = next - (rlen*increment)
                if cval < min_value:
                    cval = min_value
                    incr = increment
                else:
                    incr = increment
            else:
                rspace = next - last
                if (rspace-1) < rlen:
                    incr = increment
                else:
                    incr = rspace / (rlen+1)
                    if incr < 1:
                        incr = 1
                    elif incr > increment:
                        incr = increment
                cval = last + incr

        for j in range(i,i+rlen):
            yield cval, items[j]
            last = cval
            cval += incr

        i += rlen

def split_seq(seq, m):
    """Distribute the seq elements in lists in m groups
       according to quasi equitative distribution (decreasing order):
         split_seq(range(13), 4) --> seq = range(13), m=4
         result : [[0, 1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12]]
    """
    n,b,newseq=len(seq),0,[]
    for k in range(m):
        q,r=divmod(n-k,m)
        a, b = b, b + q + (r!=0)
        newseq.append(seq[a:b])
    return newseq

if __name__ == '__main__':

    for x in product('abc', range(2), '1'):
        print x
