#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
knihovna pro praci s mapami a souradnicemi. Zavisi na modulu pyproj.
Pro spravnou funkci S-JTSK souradnic je nutno opatchovat balik "proj":

apt-get source proj

opatchovat src/PJ_krovak.c

--- PJ_krovak.c.proj4   Sun Dec 15 23:31:04 2002
+++ PJ_krovak.c.upr     Mon May  5 10:24:42 2003
@@ -133,8 +133,8 @@
        ro = ro0 * pow(tan(s0 / 2. + s45) , n) / pow(tan(s / 2. + s45) , n)   ;

    /* x and y are reverted! */
-       xy.y = ro * cos(eps) / a;
-       xy.x = ro * sin(eps) / a;
+       xy.y = -ro * cos(eps) / a;
+       xy.x = -ro * sin(eps) / a;

 #ifdef DEBUG
        strcpy(errmess,"a: ");
@@ -196,8 +196,8 @@
 /* Transformation */
    /* revert y, x*/
        xy0=xy.x;
-       xy.x=xy.y;
-       xy.y=xy0;
+       xy.x = -xy.y;
+       xy.y = -xy0;

        ro = sqrt(xy.x * xy.x + xy.y * xy.y);
        eps = atan2(xy.y, xy.x);


fakeroot dpkg-buildpackage


http://www.cdc.noaa.gov/people/jeffrey.s.whitaker/python/pyproj.html
http://mpa.itc.it/radim/jtsk/
'''

import os
import os.path

import pyproj
from PIL import Image
from lib4 import images



__all__ = [
    'SJTSK',
    'MapInfo'
]



class SJTSK(object):
    PROJ_KROVAK = pyproj.Proj(proj='krovak',
                       ellps='bessel',
                       )
    # magicke konstanty pro posunuti vysledku prevodu WGS84->SJTSK
    DELTA_X = 104.
    DELTA_Y = 38.

    @classmethod
    def toLongLat(klass, xy):
        '''konverze souradnic z S-JTSK do WGS84 (ekvivalent:
           cs2cs +proj=krovak +ellps=bessel \
                +towgs84=570.83,85.668,462.843,4.998,1.587,5.262,3.560 \
                +to +proj=latlong +datum=WGS84
        '''
        # problemy s verzema a znamenkama
        if pyproj.__version__ < '1.8':
            return klass.PROJ_KROVAK(
                        abs(xy[0]-klass.DELTA_X), 
                        abs(xy[1]-klass.DELTA_Y),
                        inverse=True)
        else:
            return klass.PROJ_KROVAK(xy[0]-klass.DELTA_X, xy[1]-klass.DELTA_Y,
                                 inverse=True)

    @classmethod
    def fromLongLat(klass, longlat):
        '''konverze souradnic z WGS84 do S-JTSK'''
        x,y = klass.PROJ_KROVAK(longlat[0], longlat[1])
        if x > 0: x *= -1
        if y > 0: y *= -1
        return (x+klass.DELTA_X, y+klass.DELTA_Y)

    @classmethod
    def degToFloat(klass, deg, min, sec):
        f = float(deg)
        f += float(min)/60.
        f += float(sec)/3600.
        return f

    @classmethod
    def floatToDeg(klass, f):
        deg = int(f)
        f -= deg
        min = int(f * 60.)
        f -= min / 60.
        sec = f * 3600.
        return (deg, min, sec)

    @classmethod
    def strToFloat(klass, s):
        deg, _tmp = s.split(u'°')
        deg = int(deg)
        min, _tmp = _tmp.split(u"'")
        min = int(min)
        sec, _tmp = _tmp.split(u'"')
        sec = float(sec)
        return klass.degToFloat(deg, min, sec)

    @classmethod
    def floatToStr(klass, f):
        d,m,s = klass.floatToDeg(f)
        return u'''%s°%s'%.3f"''' % (d,m,s)



class MapInfo(object):
    '''trida shromazdujici vsechny potrebne informace o mape
    '''

    # meritko mapy v textove podobe
    meritko = '1 : 1'

    # korenovy adresar pro vsechny mapy
    tiles_root = None
    # adresar s jednotlivymi dlazdicemi teto mapy
    tiles_directory = None
    # korenove url pro adresar s dlazdicemi
    tiles_url_root = None
    # maska jmen jednotlivych dlazdic
    tile_filename_mask = '%dx%d.png'
    # velikost dlazdice
    tile_w = 500
    tile_h = 500

    # celkova velikost mapy
    map_w = 3000
    map_h = 2500

    # velikost generovane mapy
    out_w = 227
    out_h = 227

    # zamereni mapy - "worldfile"
    worldfile = (14.689401,
                 0.000000,
                 0.000000,
                 -14.690704,
                 -583800.156351,
                 -1169325.694896)


    def __init__(self, **kw):
        for k,v in kw.iteritems():
            setattr(self, k, v)

    def tileFileName(self, x, y):
        return os.path.join(self.tiles_root,
                            self.tiles_directory,
                            self.tile_filename_mask % (x, y))

    def sjtsk2xy(self, sjtsk):
        w = self.worldfile
        row = (w[0]*sjtsk[1]-w[1]*sjtsk[0]+w[1]*w[4]-w[0]*w[5]) / \
                (w[0]*w[3]-w[1]*w[2])
        column = (sjtsk[0]-w[2]*row-w[4])/w[0]
        return (int(round(column)), int(round(row)))

    def xy2sjtsk(self, xy):
        w = self.worldfile
        x = w[0]*xy[0] + w[2]*xy[1] + w[4]
        y = w[3]*xy[1] + w[1]*xy[0] + w[5]
        return (x,y)

    def box(self, x, y):
        '''na vstupu souradnice stredu mapy v pixelech
           vysledkem jsou souradnice leve, prave, horni a dolni hrany
           tez v pixelech
        '''
        out_l = max(0, x - self.out_w/2)
        out_r = out_l + self.out_w
        if out_r > self.map_w:
            out_r = self.map_w
            out_l = out_r - self.out_w

        out_t = max(0, y - self.out_h/2)
        out_b = out_t + self.out_h
        if out_b > self.map_h:
            out_b = self.map_h
            out_t = out_b - self.out_h

        return (out_l, out_t, out_r, out_b)

    def convertCoords(self, point, proj):
        '''zkonvertuje souradnice na souradnice v obrazku'''
        if proj == 'image':
            x,y = point
        elif proj == 'sjtsk':
            x,y = self.sjtsk2xy(point)
        elif proj == 'longlat':
            x,y = self.sjtsk2xy(SJTSK.fromLongLat(point))
        else:
            raise ValueError("Unknown projection")

        return (int(x),int(y))

    def image(self, center, proj='sjtsk'):
        '''na vstupu jsou geograficke souradnice vytouzeneho bodu
           bud v latitude, longitude, nebo v S-JTSK, nebo primo
           v souradnicich obrazku. Druh souradnic je v parametru
           'proj', defaultne se pocita s S-JTSK.
           na vystupu pak obrazek mapy
        '''

        # najit souradnice stredu
        (x,y) = self.convertCoords(center, proj)

        # vysledny obrazek - leva, prava, horni a dolni hrana v
        # souradnicich cele slozene mapy
        (out_l,out_t,out_r,out_b) = self.box(x, y)

        # vytvorim vysledny obrazek
        out_image = Image.new("RGB", (self.out_w, self.out_h))

        # pro vsechny ctverce, ktere zasahuji do vysledneho obrazku
        for ix in range(out_l/self.tile_w, (out_r-1)/self.tile_w + 1):
            for iy in range(out_t/self.tile_h, (out_b-1)/self.tile_h + 1):
                # najit soubor
                try:
                    tile_image = images.loadImage(self.tileFileName(ix, iy))
                except:
                    continue

                # souradnice dlazdice na cele mape
                tile_l = ix * self.tile_w
                tile_t = iy * self.tile_h

                # spocitat souradnice, kam dlazdici vlozit
                paste_l = tile_l - out_l
                paste_t = tile_t - out_t

                # polozit dlazdici
                out_image.paste(tile_image, (paste_l, paste_t))

        return out_image









    def mapCut(self, img_file):
        '''rozreze velky obrazek s mapou na dlazdice pozadovane velikosti
           soubory ulozi do adresare out_directory
           Prostor v obrazku, do ktereho se jiz nevejde cela dlazdice je
           podbarven sede.
        '''
        try:
            os.mkdir(os.path.join(self.tiles_root, self.tiles_directory))
        except:
            return

        image = images.loadImage(img_file)
        (map_w, map_h) = image.size

        num_x = map_w / self.tile_w
        if map_w % self.tile_w > 0:
            num_x += 1
        num_y = map_h / self.tile_h
        if map_h % self.tile_h > 0:
            num_y += 1

        for x in range(num_x):
            for y in range(num_y):
                left = x * self.tile_w
                top = y * self.tile_h
                if left+self.tile_w <= map_w and top+self.tile_h <= map_h:
                    tile = image.crop((left, top,
                                       left+self.tile_w, top+self.tile_h))
                else:
                    tile = Image.new("RGB",
                                     (self.tile_w, self.tile_h),
                                     (170, 170, 170))
                    tile.paste(image.crop((left, top, map_w, map_h)), (0, 0))
                tile.save(self.tileFileName(x, y))
