# Daemon
#   * http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/278731
#   * http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/66012
#   * http://mail.python.org/pipermail/python-list/2001-February/030814.html
#   * Zope3/src/zdaemon
#   * http://www.plope.com/software/supervisor
#   * http://homepages.tesco.net/~J.deBoynePollard/FGA/unix-daemon-design-mistakes-to-avoid.html
#
# Lock
#   * http://pypi.python.org/pypi/locknix/1.0.2
#   * http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/65203
#   * http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/498171
#   * http://www.voidspace.org.uk/python/weblog/arch_d7_2005_11_26.shtml#e148
#   * http://cheeseshop.python.org/pypi/lock_file/
#   * http://cheeseshop.python.org/pypi/zc.lockfile/
#
# Run a function in a separate (forked) process without blocking
# http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/550801
#
# Supervisor
# http://www.plope.com/software/supervisor2/
#
# Shell
# quoting http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/498202
#
# Module to allow Asynchronous subprocess use on Windows and Posix platforms
# http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/440554
#
# Nonblocking
# http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/440554
#
# Enumprocess
# http://cheeseshop.python.org/pypi/enumprocess/0.1

import sys, os, fcntl, errno, signal

__all__ = [
    'ProcessInstanceLockError', 'ProcessInstanceLock',
    'ForkCallError', 'ForkCall', 'daemonize',
]


class ProcessInstanceLockError(Exception):
    pass

class ForkCallError(Exception):
    pass


class ProcessInstanceLock:
    '''Allows only one instance of process/program to run
    '''

    def __init__(self, lockFile, pid=None):
        self.lockFile = lockFile
        if pid is None:
            self.pid = os.getpid()
        else:
            self.pid = pid
        self.lockfd = None

    def acquire(self, blocking=True):
        if self.lockfd is not None:
            raise ProcessInstanceLockError(
                'Lock already acquired', self.lockFile)
        fd = os.open(self.lockFile, os.O_CREAT|os.O_RDWR, 0644)
        if blocking:
            fcntl.lockf(fd, fcntl.LOCK_EX)
            self._writepid(fd, self.pid)
            self.lockfd = fd
            return True
        else:
            try:
                fcntl.lockf(fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
            except IOError, ie:
                if ie.errno in (errno.EAGAIN, errno.EACCES):
                    return False
                else:
                    raise
            else:
                self._writepid(fd, self.pid)
                self.lockfd = fd
                return True

    def savepid(self, pid=None):
        if self.lockfd is None:
            raise ProcessInstanceLockError(
                'Not holding lock', self.lockFile)
        self._writepid(self.lockfd, pid)

    def _writepid(self, fd, pid=None):
        if pid is None:
            pid = self.pid
        if pid is None:
            pid = os.getpid()
        txt = '%d\n' % pid
        os.ftruncate(fd, len(txt))
        os.lseek(fd, 0, 0)
        os.write(fd, txt)
        os.fsync(fd)

    def release(self):
        if self.lockfd is None:
            raise ProcessInstanceLockError(
                'Lock already released', self.lockFile)
        fcntl.lockf(self.lockfd, fcntl.LOCK_UN)
        os.close(self.lockfd)
        self.lockfd = None
        os.remove(self.lockFile)




class ForkCall:

    def __init__(self, func, *args, **kwds):
        import pickle

        pfd, cfd = os.pipe()
        pid = os.fork()
        if pid == 0:
            os.close(pfd)
            try:
                res = func(*args, **kwds)
                os.write(cfd, pickle.dumps(res))
                os.fsync(cfd)
            finally:
                os.close(cfd)
                # XXX sys.exit(...) is catched by test module
                os._exit(0)
        else:
            os.close(cfd)
            fcntl.fcntl(pfd, fcntl.F_SETFL, os.O_NONBLOCK)
            self.pid = pid
            self.fd  = pfd

    def XXXisAlive(self):
        try:
            os.kill(self.pid, 0)
            try:
                w = os.waitpid(self.pid, os.WNOHANG)
                return False
            except:
                return True
        except:
            return False

    def __call__(self):
        try:
            os.kill(self.pid, 0)
        except:
            raise ForkCallError('Child process has died')
        import pickle, select
        fd = self.fd
        data = []
        while True:
            r,_,_ = select.select([fd], [], [])
            if r:
                d = os.read(fd, 1024)
                if not d:
                    break
                else:
                    data.append(d)
        os.close(fd)
        os.waitpid(self.pid, 0)
        try:
            return pickle.loads(''.join(data))
        except:
            raise ForkCallError('Invalid result')




def daemonize():
    # __author__ = "Chad J. Schroeder"
    # __version__ = "$Id: create_daemon.py 77 2004-12-21 14:52:34Z cvs $"
    # http://www.erlenstar.demon.co.uk/unix/faq_2.html#SEC16

    # First fork
    # Fork a child process so the parent can exit.  This will return control
    # to the command line or shell.  This is required so that the new process
    # is guaranteed not to be a process group leader.  We have this guarantee
    # because the process GID of the parent is inherited by the child, but
    # the child gets a new PID, making it impossible for its PID to equal its
    # PGID.
    if os.fork() > 0:
        sys.exit(0)

    # Next we call os.setsid() to become the session leader of this new
    # session.  The process also becomes the process group leader of the
    # new process group.  Since a controlling terminal is associated with a
    # session, and this new session has not yet acquired a controlling
    # terminal our process now has no controlling terminal.  This shouldn't
    # fail, since we're guaranteed that the child is not a process group
    # leader.
    os.setsid()

    # When the first child terminates, all processes in the second child
    # are sent a SIGHUP, so it's ignored.
    signal.signal(signal.SIGHUP, signal.SIG_IGN)

    # Second fork
    # Fork a second child to prevent zombies.  Since the first child is
    # a session leader without a controlling terminal, it's possible for
    # it to acquire one by opening a terminal in the future.  This second
    # fork guarantees that the child is no longer a session leader, thus
    # preventing the daemon from ever acquiring a controlling terminal.
    if os.fork() > 0:
        sys.exit(0)

    # Ensure that the daemon doesn't keep any directory in use.  Failure
    # to do this could make a filesystem unmountable.
    os.chdir('/')
    # Give the child complete control over permissions.
    os.umask(0)

    # Close all open files.  Try the system configuration variable, SC_OPEN_MAX,
    # for the maximum number of open files to close.  If it doesn't exist, use
    # the default value (configurable).
    try:
        maxfd = os.sysconf("SC_OPEN_MAX")
    except (AttributeError, ValueError):
        # default maximum
        maxfd = 256
    for fd in range(maxfd):
        try:
            os.close(fd)
        except OSError:
            pass

    # Redirect the standard file descriptors to /dev/null.
    # Standard file descriptors shouldn't be closed, because
    # newly opened sockets and files could be corrupted
    # by some accidental writes/reads to/from std descriptors 0,1,2.
    os.open("/dev/null", os.O_RDONLY) # standard input (0)
    os.open("/dev/null", os.O_RDWR)   # standard output (1)
    os.open("/dev/null", os.O_RDWR)   # standard error (2)
