# Extending Classes
# http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/412717
#
# Object Proxying
# http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/496741
#
# Finding the value passed for a particular parameter to a function by name
# http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/498090
#
# Implementing the make statement by hacking bytecodes
# http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/498242
#
# http://www.partiallydisassembled.net/inline.py

import sys

__all__ = [
    '_thunk', 'cloneFunction', 
    'code_firstlineno', 'throw_in_file', 'compile_with_lineinfo',
    'superclass', 'supertype',
    'signature', 'values_for_signature', 'mapply',
    'AbstractProxy', 'ObjectProxy', 'LazyProxy', 'CachingLazyProxy',
    'ImportProxy',
]


def _return_const():
    return None
_return_const = _return_const.func_code.co_code

def _thunk(f):

    def _thunk():
        const = None
        return replace_code(f())

    def replace_code(val):
        co = _thunk.func_code
        thunk.func_code = co.__class__(
            co.co_argcount,
            co.co_nlocals,
            co.co_stacksize,
            co.co_flags,
            _return_const,
            (val,),
            co.co_names,
            co.co_varnames,
            co.co_filename,
            co.co_name,
            co.co_firstlineno,
            co.co_lnotab,
            co.co_freevars,
            co.co_cellvars
        )
        return val

    return _thunk


def cloneFunction(f, name, doc, module, frame=None, level=2):
    """Clone a func with reset globals to look like it came from elsewhere"""
    if module is not None:
        __import__(module)
        _globals = sys.modules[module].__dict__
    else:
        _globals = (frame or sys._getframe(level)).f_globals

    # clone the function using the targeted module (or caller's) globals
    from new import function
    f = function(f.func_code, _globals, name, f.func_defaults, f.func_closure)
    f.__doc__  = doc  or f.__doc__
    return f



def code_firstlineno(co, firstlineno, filename=None, name=None, lnotab=None):
    if filename is None:
        filename = co.co_filename
    if name is None:
        name = co.co_name
    if lnotab is None:
        lnotab = co.co_lnotab
    return co.__class__(
        co.co_argcount, co.co_nlocals, co.co_stacksize, co.co_flags,
        co.co_code, co.co_consts, co.co_names, co.co_varnames,
        filename, name, firstlineno, lnotab,
        co.co_freevars, co.co_cellvars
    )


def throw_in_file(exc, method, filename, lineno):
    '''Fictive function that raises given exc in file on lineno'''
    def throw(): raise exc
    throw.__name__ = method
    throw.func_code = code_firstlineno(throw.func_code,
        firstlineno=lineno, filename=filename, name=method, lnotab=''
    )
    throw()


def compile_with_lineinfo(expr, filename, lineno=1, name='expr'):
        try:
            code = compile(expr, filename, 'eval')
        except Exception, e:
            throw_in_file(e, name, filename, lineno)
        return code_firstlineno(code, lineno, name=name, lnotab='')



def superclass(bases, className=None, classDict={}, metaDict={}):
    '''Creates subclass from base classes

        bases           tuple of base classes
        className       name of newly created class
        classDict       add to new class __dict__
        metaDict        add to new metaclass __dict__

        See also:
          http://www.phyast.pitt.edu/~micheles/python/metatype.html
          http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/204197
    '''
    if className is None :
        className = bases[0].__name__

    metaBases = {}
    for i, base in enumerate(bases):
        meta = type(base)
        metaBases.setdefault(meta, i)
        for b in meta.__mro__[1:] :
            metaBases[b] = None

    if not metaDict and len(metaBases) == 1 :
        meta = metaBases.keys()[0]
    else :
        from iterutils import sorted
        metaBases = tuple([
            base for base,i in sorted(metaBases.iteritems(), key=lambda x: x[1])
            if i is not None
        ])
        if not metaDict and len(metaBases) == 1 :
            meta = metaBases[0]
        else :
            meta = type(className + 'Meta', metaBases, metaDict)

    cDict = classDict.copy()
    if meta is not type :
        cDict['__metaclass__'] = meta

    return meta(className, tuple(bases), cDict)



def supertype(supertype,subtype):
    """Workaround for 'super()' not handling metaclasses well

    Note that this will *skip* any classic classes in the MRO!
    """

    mro = iter(subtype.__mro__)

    for cls in mro:
        if cls is supertype:
            for cls in mro:
                if hasattr(cls,'__mro__'):
                    return cls
            break
    raise TypeError("Not sub/supertypes:", supertype, subtype)





def signature(obj):
    '''Returns signature of callable object as tuple of values
      (args, varargs, varkw, defaults)
    '''
    from inspect import getargspec, isfunction, ismethod, isclass

    if isfunction(obj):
        args, varargs, varkw, defaults = getargspec(obj)
    else:
        shift = 0
        if ismethod(obj):
            if obj.im_self is not None:
                shift = 1
        elif isclass(obj):
            if hasattr(obj, '__init__'):
                obj = obj.__init__
            else:
                return ([], None, None, {})
            shift = 1
        elif hasattr(obj, '__call__'):
            return signature(obj.__call__)
        else:
            raise TypeError('Object is not callable', obj)
        args, varargs, varkw, defaults = getargspec(obj)
        args = args[shift:]
    if defaults:
        defaults = dict([(a,v) for a,v in zip(args[-len(defaults):], defaults)])
    else:
        defaults = {}
    return args, varargs, varkw, defaults




def get_positional(arg, default):
    return default

def get_vararg(arg):
    return ()

def get_varkw(arg):
    return {}

_NOT_FOUND = object()

def values_for_signature(
        (args, varargs, varkw, defaults),
        namespace={}, default=None,
        get_positional=get_positional,
        get_vararg = get_vararg,
        get_varkw = get_varkw
):
    arg_vect = []
    for a in args:
        v = namespace.get(a, _NOT_FOUND)
        if v is _NOT_FOUND:
            v = get_positional(a, defaults.get(a, default))
        arg_vect.append(v)
    v = namespace.get(varargs, _NOT_FOUND)
    if v is _NOT_FOUND:
        v = get_vararg(varargs)
    arg_vect.extend(v)

    kw_dict = namespace.get(varkw, _NOT_FOUND)
    if kw_dict is _NOT_FOUND:
        kw_dict = get_varkw(varkw)

    return tuple(arg_vect), kw_dict


def mapply(
        callable, namespace={}, default=None,
        get_positional=get_positional,
        get_vararg = get_vararg,
        get_varkw = get_varkw
):
    sig = signature(callable)
    arg_vect, kw_dict = values_for_signature(
        sig, namespace, default, get_positional, get_vararg, get_varkw
    )
    return callable(*arg_vect, **kw_dict)





class AbstractProxy(object):
    """Delegates all operations (except __subject__ attr) to another object"""
    # from peak.util.context

    __slots__ = ()

    def __call__(self,*args,**kw):
        return self.__subject__(*args,**kw)

    def __getattribute__(self, attr, oga=object.__getattribute__):
        subject = oga(self,'__subject__')
        if attr=='__subject__':
            return subject
        return getattr(subject,attr)

    def __setattr__(self,attr,val, osa=object.__setattr__):
        if attr=='__subject__':
            osa(self,attr,val)
        else:
            setattr(self.__subject__,attr,val)

    def __delattr__(self,attr, oda=object.__delattr__):
        if attr=='__subject__':
            oda(self,attr)
        else:
            delattr(self.__subject__,attr)

    def __nonzero__(self):
        return bool(self.__subject__)

    def __getitem__(self,arg):
        return self.__subject__[arg]

    def __setitem__(self,arg,val):
        self.__subject__[arg] = val

    def __delitem__(self,arg):
        del self.__subject__[arg]

    def __getslice__(self,i,j):
        return self.__subject__[i:j]

    def __setslice__(self,i,j,val):
        self.__subject__[i:j] = val

    def __delslice__(self,i,j):
        del self.__subject__[i:j]

    def __contains__(self,ob):
        return ob in self.__subject__

    for name in 'repr str hash len abs complex int long float iter oct hex'.split():
        exec "def __%s__(self): return %s(self.__subject__)" % (name,name)

    for name in 'cmp','coerce','divmod':
        exec "def __%s__(self,ob): return %s(self.__subject__,ob)" % (name,name)

    for name,op in [
        ('lt','<'), ('gt','>'), ('le','<='), ('ge','>='),
        ('eq','=='), ('ne','!=')
    ]:
        exec "def __%s__(self,ob): return self.__subject__ %s ob" % (name,op)

    for name,op in [('neg','-'), ('pos','+'), ('invert','~')]:
        exec "def __%s__(self): return %s self.__subject__" % (name,op)

    for name, op in [
        ('or','|'),  ('and','&'), ('xor','^'), ('lshift','<<'), ('rshift','>>'),
        ('add','+'), ('sub','-'), ('mul','*'), ('div','/'), ('mod','%'),
        ('truediv','/'), ('floordiv','//')
    ]:
        exec (
            "def __%(name)s__(self,ob):\n"
            "    return self.__subject__ %(op)s ob\n"
            "\n"
            "def __r%(name)s__(self,ob):\n"
            "    return ob %(op)s self.__subject__\n"
            "\n"
            "def __i%(name)s__(self,ob):\n"
            "    self.__subject__ %(op)s=ob\n"
            "    return self\n"
        )  % locals()

    del name, op

    # Oddball signatures

    def __rdivmod__(self,ob):
        return divmod(ob,self.__subject__)

    def __pow__(self,*args):
        return pow(self.__subject__,*args)

    def __ipow__(self,ob):
        self.__subject__ **= ob
        return self

    def __rpow__(self,ob,*modulo):
        return pow(ob, self.__subject__, *modulo)


class ObjectProxy(AbstractProxy):
    """Delegates all operations (except __subject__ attr) to another object"""

    __slots__ = "__subject__"

    def __init__(self,subject):
        self.__subject__ = subject



_osa = object.__setattr__

class LazyProxy(AbstractProxy):
    #__slots__ = ('__proxy_obj__', '__proxy_recipe__')
    __proxy_args__ = None
    __proxy_recipe__ = None
    def __init__(self, recipe, *args):
        _osa(self, '__proxy_args__', args)
        _osa(self, '__proxy_recipe__', recipe)
    def __subject__(self, oga=object.__getattribute__):
        recipe = oga(self, '__proxy_recipe__')
        args = oga(self, '__proxy_args__')
        return recipe(*args)
    __subject__ = property(__subject__)




_noobj = object()

class CachingLazyProxy(LazyProxy):
    __proxy_cache__ = _noobj
    def __subject__(self, oga=object.__getattribute__):
        subj = oga(self, '__proxy_cache__')
        if subj is _noobj:
            recipe = oga(self, '__proxy_recipe__')
            args = oga(self, '__proxy_args__')
            subj = recipe(*args)
            _osa(self, '__proxy_cache__', subj)
        return subj
    __subject__ = property(__subject__)



def ImportProxy(spec):
    from imports import importObject
    return CachingLazyProxy(importObject, spec)
