# http://github.com/jwp/py-postgresql/blob/v1.0/postgresql/cluster.py
import os
import sys
import shutil
import subprocess
import signal
import glob
import time


class TestPG:
    
    driver = 'pgdb'
    name = 'testpg'
    tmpdir = os.getenv('TMP', os.sep + 'tmp')
    pg_config = os.getenv('PG_CONFIG')
    install_prefixes = ['/usr/local/bin', '/usr/bin', '/opt/postgresql*']
    port = 6655

    def __init__(self, **params):
        self.__dict__.update(params)
        if self.pg_config is None:
            for pat in self.install_prefixes:
                for dir in glob.glob(pat):
                    pgc = os.path.join(dir, 'pg_config')
                    if os.path.isfile(pgc):
                        self.pg_config = pgc
                        break
                if self.pg_config is not None:
                    break
        if self.pg_config is None:
            raise OSError('pg_config not found in', self.install_prefixes)
        self.bindir = os.popen('%s --bindir' % self.pg_config).read().strip()
        self.dbdir = os.path.join(self.tmpdir, self.name)
        self.pid = None

    def initdb(self):
        if os.path.isdir(self.dbdir):
            shutil.rmtree(self.dbdir)
        p = subprocess.Popen([
            os.path.join(self.bindir, 'initdb'),
             '-D', self.dbdir,
             '--encoding=UNICODE',
             '--lc-collate=en_US.UTF-8',
             '--lc-ctype=en_US.UTF-8'],
             stdout=subprocess.PIPE,
             stderr=subprocess.PIPE,
        )
        x = p.communicate()
        assert p.returncode == 0

    def start(self):
        #/usr/lib/postgresql/8.3/bin/postgres -p 6655 -D /tmp/tdb -k /tmp/tdb
        if self.pid is not None:
            raise OSError('postgresql is already running')
        exe = os.path.join(self.bindir, 'postgres')
        cmd = [exe, '-p', str(self.port), '-D', self.dbdir, '-k', self.dbdir]
        log = open(os.path.join(self.dbdir, 'postgresql.log'), 'a')
        self.pg = subprocess.Popen(cmd, stdout=log, stderr=log)
        self.pid = self.pg.pid
        # wait for conn
        while True:
            try:
                c = self.connect()
            except:
                time.sleep(0.01)
            else:
                c.close()
                break

    def stop(self):
        if self.pg is None:
            raise OSError('postgresql is not running')
        os.kill(self.pg.pid, signal.SIGINT)
        self.pg.communicate()
        self.pg = None
        self.pid = None
    
    def dburl(self, db):
        return 'pgsql:///%s?rundir=%s&port=%d' % (db, self.dbdir, self.port)

    def createdb(self, db, schema=None):
        exe = os.path.join(self.bindir, 'createdb')
        proc = subprocess.Popen(
            [exe, '-p', str(self.port), '-h', self.dbdir, db],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        out, err = proc.communicate()
        assert proc.returncode == 0, (proc.returncode, out, err)

    def dropdb(self, db):
        exe = os.path.join(self.bindir, 'dropdb')
        proc = subprocess.Popen(
                [exe, '-p', str(self.port), '-h', self.dbdir, db],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE
        )
        out, err = proc.communicate()
        assert proc.returncode == 0, (proc.returncode, out, err)

    def truncate(self, db):
        con = self.connect(db)
        try:
            cur = con.curcor()
            raise NotImplementedError
        finally:
            cur.close()
            con.close()


    def connect(self, db='template1'):
        import pgdb
        return pgdb.connect(host='%s:%d' % (self.dbdir, self.port), database=db)
    
    def __del__(self):
        if self.pg is not None:
            os.kill(self.pid, signal.SIGINT)

if __name__ == '__main__':

    tpg = TestPG()
    print tpg
    #tpg.initdb()
    print tpg.dburl('template1')
    tpg.start()
    tpg.createdb('ahoj')
    #tpg.dropdb('ahoj')
    #tpg.stop()
