# -*- coding: utf-8 -*-
# http://underscode.selidor.net/

import re

__all__ = [
    'sql_escape', 'sql_name', 'bytea_escape', 'bytea_unescape', 'escape_name',
    'KEYWORDS',
]



def sql_escape(s):
    '''Escapes special characters for SQL usage

        >>> sql_escape('a')
        'a'
        >>> sql_escape("a'" )
        "a''"

    '''
    return (
              (
                (
                  (
                    s.replace("\\", "\\\\")
                  ).replace("'", "''")
                ).replace("\011", "\\t")
              ).replace("\012", "\\n")
            ).replace("\015", "\\r" )





_case_change_search = re.compile("([a-z][A-Z])").sub
_name_illegal_chars = re.compile("[^a-z0-9_]+", re.I).sub
_underscore_normalization = re.compile("_+", re.I).sub

def sql_name(name):
    '''Converts class name to more convenient SQL table name

        >> sql_name('foo')
        'foo'
        >> sql_name('Foo')
        'foo'
        >> sql_name('Foo Bar Baz')
        'foo_bar_baz'
        >> sql_name('FooBar   Baz')
        'foo_bar_baz'
        >> sql_name('FooBarBaz')
        'foo_bar_baz'
        >>> sql_name('WWWSite')
        'wwwsite'
        >>> sql_name(unicode('český název', 'utf8'))
        'cesky_nazev'
        >>> sql_name('Predavane udaje, vyjma povinnych')
        'predavane_udaje_vyjma_povinnych'
        >>> sql_name('Foo,;#$Bar')
        'foo_bar'
    '''
    name = unicode(name)

    try:
        name = name.encode('ascii')
    except UnicodeEncodeError:
        from lib4.text import unicode2ascii
        name = unicode2ascii(name)

    name = "_".join(name.split())

    def repl(m):
        x = m.group(0)
        return "%s_%s" % (x[0], x[1])

    t = str(_name_illegal_chars('_', name))
    t = _case_change_search(repl, t).lower()
    return _underscore_normalization('_', t)










_bytea_tooct     = lambda m: "\\\\%03o" % ord(m.group(1))
_bytea_escape    = re.compile("([^\\%s-\\%s\\%s-\\%s\\%s-\\%s])" % (
                        chr(32), chr(38),
                        chr(40), chr(91),
                        chr(93), chr(126)
                    ))

def bytea_escape(data):
    r'''Escapes raw string for using it as bytea data type

        >>> print bytea_escape('0123abc')
        0123abc
        >>> print bytea_escape('\f')
        \\014
        >>> print bytea_escape("'")
        \\047
        >>> print bytea_escape('\\')
        \\134
    '''
    return _bytea_escape.sub(_bytea_tooct, data)


def bytea_unescape(data):
    r'''Unescapes bytea string returned from database

        >>> print bytea_unescape(r'aa\134a\047xxx')
        aa\a'xxx
        >>> print bytea_unescape('3\047\\991')
        3'\991
    '''
    return data.decode('string_escape')



_nonescaped_name = re.compile('^[a-z_][a-z0-9_$]*$').match

def escape_name(n):
    r'''Escapes table or column identifier if needed

        >>> print escape_name('a')
        a
        >>> print escape_name('user')
        "user"
        >>> print escape_name('prodId')
        "prodId"
        >>> print escape_name('prod_01')
        prod_01
        >>> print escape_name('a"b"')
        "a""b"""
        >>> print escape_name('create')
        create

    '''

    if KEYWORDS.get(n.upper(), 'x') == 'r':
        return '"%s"' % n
    elif not _nonescaped_name(n):
        return '"%s"' % n.replace('"', '""')
    else:
        return n

def escape_alias(n):
    if KEYWORDS.get(n.upper(), 'x') == 'r':
        return '"%s"' % n
    else:
        return n


# SQL 92 and Postgres keywords
#
#  'r'   reserved
#  'n'   non-reserved
#  'u'   unknown
#

KEYWORDS = {

   'ABORT':'n', 'ABS':'u', 'ABSOLUTE':'n', 'ACCESS':'n', 'ACTION':'n',
   'ADA':'u', 'ADD':'n', 'ADMIN':'u', 'AFTER':'n', 'AGGREGATE':'n',
   'ALIAS':'u', 'ALL':'r', 'ALLOCATE':'u', 'ALTER':'n', 'ANALYSE':'r',
   'ANALYZE':'r', 'AND':'r', 'ANY':'r', 'ARE':'u', 'ARRAY':'u', 'AS':'r',
   'ASC':'r', 'ASENSITIVE':'u', 'ASSERTION':'u', 'ASSIGNMENT':'u',
   'ASYMMETRIC':'u', 'AT':'n', 'ATOMIC':'u', 'AUTHORIZATION':'n', 'AVG':'u',

   'BACKWARD':'n', 'BEFORE':'n', 'BEGIN':'n', 'BETWEEN':'r', 'BINARY':'r',
   'BIT':'n', 'BITVAR':'u', 'BIT_LENGTH':'u', 'BLOB':'u', 'BOOLEAN':'u',
   'BOTH':'r', 'BREADTH':'u', 'BY':'n',

   'C':'u', 'CACHE':'n', 'CALL':'u', 'CALLED':'u', 'CARDINALITY':'u',
   'CASCADE':'n', 'CASCADED':'u', 'CASE':'r', 'CAST':'r', 'CATALOG':'u',
   'CATALOG_NAME':'u', 'CHAIN':'n', 'CHAR':'n', 'CHARACTER':'n',
   'CHARACTERISTICS':'n', 'CHARACTER_LENGTH':'u', 'CHARACTER_SET_CATALOG':'u',
   'CHARACTER_SET_NAME':'u', 'CHARACTER_SET_SCHEMA':'u', 'CHAR_LENGTH':'u',
   'CHECK':'r', 'CHECKED':'u', 'CHECKPOINT':'n', 'CLASS':'u',
   'CLASS_ORIGIN':'u', 'CLOB':'u', 'CLOSE':'n', 'CLUSTER':'n', 'COALESCE':'n',
   'COBOL':'u', 'COLLATE':'r', 'COLLATION':'u', 'COLLATION_CATALOG':'u',
   'COLLATION_NAME':'u', 'COLLATION_SCHEMA':'u', 'COLUMN':'r',
   'COLUMN_NAME':'u', 'COMMAND_FUNCTION':'u', 'COMMAND_FUNCTION_CODE':'u',
   'COMMENT':'n', 'COMMIT':'n', 'COMMITTED':'n', 'COMPLETION':'u',
   'CONDITION_NUMBER':'u', 'CONNECT':'u', 'CONNECTION':'u',
   'CONNECTION_NAME':'u', 'CONSTRAINT':'r', 'CONSTRAINTS':'n',
   'CONSTRAINT_CATALOG':'u', 'CONSTRAINT_NAME':'u', 'CONSTRAINT_SCHEMA':'u',
   'CONSTRUCTOR':'u', 'CONTAINS':'u', 'CONTINUE':'u', 'CONVERT':'u', 'COPY':'n',
   'CORRESPONDING':'u', 'COUNT':'u', 'CREATE':'n', 'CREATEDB':'n',
   'CREATEUSER':'n', 'CROSS':'r', 'CUBE':'u', 'CURRENT':'u', 'CURRENT_DATE':'r',
   'CURRENT_PATH':'u', 'CURRENT_ROLE':'u', 'CURRENT_TIME':'r',
   'CURRENT_TIMESTAMP':'r', 'CURRENT_USER':'r', 'CURSOR':'n', 'CURSOR_NAME':'u',
   'CYCLE':'n',

   'DATA':'u', 'DATABASE':'n', 'DATE':'u', 'DATETIME_INTERVAL_CODE':'u',
   'DATETIME_INTERVAL_PRECISION':'u', 'DAY':'n', 'DEALLOCATE':'u', 'DEC':'n',
   'DECIMAL':'n', 'DECLARE':'n', 'DEFAULT':'r', 'DEFERRABLE':'r',
   'DEFERRED':'n', 'DEFINED':'u', 'DEFINER':'u', 'DELETE':'n', 'DELIMITERS':'n',
   'DEPTH':'u', 'DEREF':'u', 'DESC':'r', 'DESCRIBE':'u', 'DESCRIPTOR':'u',
   'DESTROY':'u', 'DESTRUCTOR':'u', 'DETERMINISTIC':'u', 'DIAGNOSTICS':'u',
   'DICTIONARY':'u', 'DISCONNECT':'u', 'DISPATCH':'u', 'DISTINCT':'r',
   'DO':'r', 'DOMAIN':'u', 'DOUBLE':'n', 'DROP':'n', 'DYNAMIC':'u',
   'DYNAMIC_FUNCTION':'u', 'DYNAMIC_FUNCTION_CODE':'u',

   'EACH':'n', 'ELSE':'r', 'ENCODING':'n', 'ENCRYPTED':'n', 'END':'r',
   'END-EXEC':'u', 'EQUALS':'u', 'ESCAPE':'n', 'EVERY':'u', 'EXCEPT':'r',
   'EXCEPTION':'u', 'EXCLUSIVE':'n', 'EXEC':'u', 'EXECUTE':'n', 'EXISTING':'u',
   'EXISTS':'n', 'EXPLAIN':'n', 'EXTERNAL':'u', 'EXTRACT':'n',

   'FALSE':'r', 'FETCH':'n', 'FINAL':'u', 'FIRST':'u', 'FLOAT':'n', 'FOR':'r',
   'FORCE':'n', 'FOREIGN':'r', 'FORTRAN':'u', 'FORWARD':'n', 'FOUND':'u',
   'FREE':'u', 'FREEZE':'r', 'FROM':'r', 'FULL':'r', 'FUNCTION':'n',

   'G':'u', 'GENERAL':'u', 'GENERATED':'u', 'GET':'u', 'GLOBAL':'n', 'GO':'u',
   'GOTO':'u', 'GRANT':'n', 'GRANTED':'u', 'GROUP':'r', 'GROUPING':'u',

   'HANDLER':'n', 'HAVING':'r', 'HIERARCHY':'u', 'HOLD':'u', 'HOST':'u',
   'HOUR':'n',

   'IDENTITY':'u', 'IGNORE':'u', 'ILIKE':'r', 'IMMEDIATE':'n',
   'IMPLEMENTATION':'u', 'IN':'r', 'INCREMENT':'n', 'INDEX':'n',
   'INDICATOR':'u', 'INFIX':'u', 'INHERITS':'n', 'INITIALIZE':'u',
   'INITIALLY':'r', 'INNER':'r', 'INOUT':'n', 'INPUT':'u', 'INSENSITIVE':'n',
   'INSERT':'n', 'INSTANCE':'u', 'INSTANTIABLE':'u', 'INSTEAD':'n', 'INT':'u',
   'INTEGER':'u', 'INTERSECT':'r', 'INTERVAL':'n', 'INTO':'r', 'INVOKER':'u',
   'IS':'r', 'ISNULL':'r', 'ISOLATION':'n', 'ITERATE':'u',

   'JOIN':'r',

   'K':'u', 'KEY':'n', 'KEY_MEMBER':'u', 'KEY_TYPE':'u',

   'LANCOMPILER':'n', 'LANGUAGE':'n', 'LARGE':'u', 'LAST':'u', 'LATERAL':'u',
   'LEADING':'r', 'LEFT':'r', 'LENGTH':'u', 'LESS':'u', 'LEVEL':'n', 'LIKE':'r',
   'LIMIT':'r', 'LISTEN':'n', 'LOAD':'n', 'LOCAL':'n', 'LOCALTIME':'u',
   'LOCALTIMESTAMP':'u', 'LOCATION':'n', 'LOCATOR':'u', 'LOCK':'n', 'LOWER':'u',

   'M':'u', 'MAP':'u', 'MATCH':'n', 'MAX':'u', 'MAXVALUE':'n',
   'MESSAGE_LENGTH':'u', 'MESSAGE_OCTET_LENGTH':'u', 'MESSAGE_TEXT':'u',
   'METHOD':'u', 'MIN':'u', 'MINUTE':'n', 'MINVALUE':'n', 'MOD':'u', 'MODE':'n',
   'MODIFIES':'u', 'MODIFY':'u', 'MODULE':'u', 'MONTH':'n', 'MORE':'u',
   'MOVE':'n', 'MUMPS':'u',

   'NAME':'u', 'NAMES':'n', 'NATIONAL':'n', 'NATURAL':'r', 'NCHAR':'n',
   'NCLOB':'u', 'NEW':'r', 'NEXT':'n', 'NO':'n', 'NOCREATEDB':'n',
   'NOCREATEUSER':'n', 'NONE':'n', 'NOT':'r', 'NOTHING':'n', 'NOTIFY':'n',
   'NOTNULL':'r', 'NULL':'r', 'NULLABLE':'u', 'NULLIF':'n', 'NUMBER':'u',
   'NUMERIC':'n',

   'OBJECT':'u', 'OCTET_LENGTH':'u', 'OF':'n', 'OFF':'r', 'OFFSET':'r',
   'OIDS':'n', 'OLD':'r', 'ON':'r', 'ONLY':'r', 'OPEN':'u', 'OPERATION':'u',
   'OPERATOR':'n', 'OPTION':'n', 'OPTIONS':'u', 'OR':'r', 'ORDER':'r',
   'ORDINALITY':'u', 'OUT':'n', 'OUTER':'r', 'OUTPUT':'u', 'OVERLAPS':'r',
   'OVERLAY':'u', 'OVERRIDING':'u', 'OWNER':'n',

   'PAD':'u', 'PARAMETER':'u', 'PARAMETERS':'u', 'PARAMETER_MODE':'u',
   'PARAMETER_NAME':'u', 'PARAMETER_ORDINAL_POSITION':'u',
   'PARAMETER_SPECIFIC_CATALOG':'u', 'PARAMETER_SPECIFIC_NAME':'u',
   'PARAMETER_SPECIFIC_SCHEMA':'u', 'PARTIAL':'n', 'PASCAL':'u', 'PASSWORD':'n',
   'PATH':'n', 'PENDANT':'n', 'PLI':'u', 'POSITION':'n', 'POSTFIX':'u',
   'PRECISION':'n', 'PREFIX':'u', 'PREORDER':'u', 'PREPARE':'u', 'PRESERVE':'u',
   'PRIMARY':'r', 'PRIOR':'n', 'PRIVILEGES':'n', 'PROCEDURAL':'n',
   'PROCEDURE':'n', 'PUBLIC':'r',

   'READ':'n', 'READS':'u', 'REAL':'u', 'RECURSIVE':'u', 'REF':'u',
   'REFERENCES':'r', 'REFERENCING':'u', 'REINDEX':'n', 'RELATIVE':'n',
   'RENAME':'n', 'REPEATABLE':'u', 'REPLACE':'n', 'RESET':'n', 'RESTRICT':'n',
   'RESULT':'u', 'RETURN':'u', 'RETURNED_LENGTH':'u',
   'RETURNED_OCTET_LENGTH':'u', 'RETURNED_SQLSTATE':'u', 'RETURNS':'n',
   'REVOKE':'n', 'RIGHT':'r', 'ROLE':'u', 'ROLLBACK':'n', 'ROLLUP':'u',
   'ROUTINE':'u', 'ROUTINE_CATALOG':'u', 'ROUTINE_NAME':'u',
   'ROUTINE_SCHEMA':'u', 'ROW':'n', 'ROWS':'u', 'ROW_COUNT':'u', 'RULE':'n',

   'SAVEPOINT':'u', 'SCALE':'u', 'SCHEMA':'n', 'SCHEMA_NAME':'u', 'SCOPE':'u',
   'SCROLL':'n', 'SEARCH':'u', 'SECOND':'n', 'SECTION':'u', 'SECURITY':'u',
   'SELECT':'r', 'SELF':'u', 'SENSITIVE':'u', 'SEQUENCE':'n',
   'SERIALIZABLE':'n', 'SERVER_NAME':'u', 'SESSION':'n', 'SESSION_USER':'r',
   'SET':'n', 'SETOF':'n', 'SETS':'u', 'SHARE':'n', 'SHOW':'n', 'SIMILAR':'u',
   'SIMPLE':'u', 'SIZE':'u', 'SMALLINT':'u', 'SOME':'r', 'SOURCE':'u',
   'SPACE':'u', 'SPECIFIC':'u', 'SPECIFICTYPE':'u', 'SPECIFIC_NAME':'u',
   'SQL':'u', 'SQLCODE':'u', 'SQLERROR':'u', 'SQLEXCEPTION':'u', 'SQLSTATE':'u',
   'SQLWARNING':'u', 'START':'n', 'STATE':'u', 'STATEMENT':'n', 'STATIC':'u',
   'STATISTICS':'n', 'STDIN':'n', 'STDOUT':'n', 'STRUCTURE':'u', 'STYLE':'u',
   'SUBCLASS_ORIGIN':'u', 'SUBLIST':'u', 'SUBSTRING':'n', 'SUM':'u',
   'SYMMETRIC':'u', 'SYSID':'n', 'SYSTEM':'u', 'SYSTEM_USER':'u',

   'TABLE':'r', 'TABLE_NAME':'u', 'TEMP':'n', 'TEMPLATE':'n', 'TEMPORARY':'n',
   'TERMINATE':'u', 'THAN':'u', 'THEN':'r', 'TIME':'n', 'TIMESTAMP':'n',
   'TIMEZONE_HOUR':'u', 'TIMEZONE_MINUTE':'u', 'TO':'r', 'TOAST':'n',
   'TRAILING':'r', 'TRANSACTION':'n', 'TRANSACTIONS_COMMITTED':'u',
   'TRANSACTIONS_ROLLED_BACK':'u', 'TRANSACTION_ACTIVE':'u',
   'TRANSFORM':'u', 'TRANSFORMS':'u', 'TRANSLATE':'u', 'TRANSLATION':'u',
   'TREAT':'u', 'TRIGGER':'n', 'TRIGGER_CATALOG':'u', 'TRIGGER_NAME':'u',
   'TRIGGER_SCHEMA':'u', 'TRIM':'n', 'TRUE':'r', 'TRUNCATE':'n',
   'TRUSTED':'n', 'TYPE':'n',

   'UNCOMMITTED':'u', 'UNDER':'u', 'UNENCRYPTED':'n', 'UNION':'r', 'UNIQUE':'r',
   'UNKNOWN':'n', 'UNLISTEN':'n', 'UNNAMED':'u', 'UNNEST':'u', 'UNTIL':'n',
   'UPDATE':'n', 'UPPER':'u', 'USAGE':'u', 'USER':'r',
   'USER_DEFINED_TYPE_CATALOG':'u', 'USER_DEFINED_TYPE_NAME':'u',
   'USER_DEFINED_TYPE_SCHEMA':'u', 'USING':'r',

   'VACUUM':'n', 'VALID':'n', 'VALUE':'u', 'VALUES':'n', 'VARCHAR':'n',
   'VARIABLE':'u', 'VARYING':'n', 'VERBOSE':'r', 'VERSION':'n', 'VIEW':'n',

   'WHEN':'r', 'WHENEVER':'u', 'WHERE':'r', 'WITH':'n', 'WITHOUT':'n',
   'WORK':'n', 'WRITE':'u',

   'YEAR':'n',

   'ZONE':'n',
}
