#include <Python.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <sys/types.h>

#include "date.h"

/* Module itself and module dictionary */
static PyObject *module, *module_dict;

/***********************************************************
 *                                                         *
 *         Forward type declarations                       *
 *                                                         *
 ***********************************************************/
staticforward PyTypeObject DateType;
staticforward PyTypeObject DateTimeType;
staticforward PyTypeObject IntervalType;

/***********************************************************
 *                                                         *
 *            Module instances                             *
 *                                                         *
 ***********************************************************/

/* Exceptions */
static PyObject *DateError;
static PyObject *DateRangeError;

/* Hash */
static PyObject *ZoneMap;

/***********************************************************
 *                                                         *
 *            Some functions                               *
 *                                                         *
 ***********************************************************/

static PyObject *number_not_implemented (PyObject *d)
{
    DateTimeObject  *dt;
    DateObject  *da;
    IntervalObject  *i;

    if(PyDate_Check(d)) {
        da = (DateObject *) d;
        if(da->argument) {
            Py_DECREF(da->argument);
            da->argument = 0;
        }
    } else if(PyDateTime_Check(d)) {
        dt = (DateTimeObject *) d;
        if(dt->argument) {
            Py_DECREF(dt->argument);
            dt->argument = 0;
        }
    } else if(PyInterval_Check(d)) {
        i = (IntervalObject *) d;
        if(i->argument) {
            Py_DECREF(i->argument);
            i->argument = 0;
        }
    }
    PyErr_SetString(PyExc_TypeError,
                    "Operator is not supported for these arguments.");
    return NULL;
}
static PyObject *number_not_implemented1 (PyObject *d)
{
    return number_not_implemented(d);
}

static PyObject *number_not_implemented2 (PyObject *d, PyObject *a1)
{
    return number_not_implemented(d);
}

static PyObject *number_not_implemented3 (PyObject *d, PyObject *a1, PyObject *a2)
{
    return number_not_implemented(d);
}


/***********************************************************
 *                                                         *
 *            Date                                         *
 *                                                         *
 ***********************************************************/

static PyObject * date_Date (PyObject *self, PyObject *args)
{
    DateObject  *no;
    int     d = 1, m = 1, y, r;
    int32_t     sdn;

    if(! PyArg_ParseTuple(args, "i|ii", &y, &m, &d))
        return NULL;
    if((r = date2sdn(y, m, d, &sdn )) != 0) {
        PyErr_SetString(DateRangeError, code2err(err_date2sdn, r) );
        return NULL;
    } 
    if(! (no = PyObject_NEW(DateObject, &DateType))) {
        PyErr_SetString(PyExc_MemoryError, "Out of memory.");
        return NULL;
    }
    no->sdn = sdn;
    no->argument = NULL;
    return (PyObject *) no;
}

static PyObject * date_DateFromString (PyObject *self, PyObject *args)
{
    DateObject      *no;
    DateTimeAttr    dta;
    char            *str, b[128];
    int32_t         sdn, r;

    if(! PyArg_ParseTuple(args, "s", &str))
        return NULL;
    if((r=parsedate(str, &dta)) != 0) {
        snprintf(b, 127, "Invalid Date format: `%s' (%d).", str, r);
        b[127] = '\0';
        PyErr_SetString(DateError, b);
        return NULL;
    }
    if((r = date2sdn(dta.year, dta.month, dta.day, &sdn )) != 0) {
        PyErr_SetString(DateRangeError, code2err(err_date2sdn, r) );
        return NULL;
    } 
    if(! (no = PyObject_NEW(DateObject, &DateType))) {
        PyErr_SetString(PyExc_MemoryError, "Out of memory.");
        return NULL;
    }
    no->sdn = sdn;
    no->argument = NULL;
    return (PyObject *) no;
}

static PyObject * date_DateFromSDN (PyObject *self, PyObject *args)
{
    DateObject  *no;
    int32_t     sdn;

    if(! PyArg_ParseTuple(args, "i", &sdn))
        return NULL;
    if(sdn < SDN_MIN || sdn > SDN_MAX) {
        PyErr_SetString(DateRangeError, "SDN range overflow.");
        return NULL;
    } 
    if(! (no = PyObject_NEW(DateObject, &DateType))) {
        PyErr_SetString(PyExc_MemoryError, "Out of memory.");
        return NULL;
    }
    no->sdn = sdn;
    no->argument = NULL;
    return (PyObject *) no;
}

static PyObject * date_DateFromDateTime (PyObject *self, PyObject *args)
{
    DateObject  *no;
    DateTimeObject  *dt, so;
    PyObject    *po;

    if(! PyArg_ParseTuple(args, "O", &po))
        return NULL;
    if(! PyDateTime_Check(po)) {
        PyErr_SetString(DateError, "Argument must be DateTime instance.");
        return NULL;
    } 
    dt = (DateTimeObject *)po;
    if(gmt2local_fix(dt, &so, NULL) != 0) {
        PyErr_SetString(DateError,
                        "System error in DateFromDateTime:gmt2local_fix.");
        return NULL;
    } 
    if(! (no = PyObject_NEW(DateObject, &DateType))) {
        PyErr_SetString(PyExc_MemoryError, "Out of memory.");
        return NULL;
    }
    no->sdn = so.sdn;
    no->argument = NULL;
    return (PyObject *) no;
}

static PyObject * date_Today (PyObject *self, PyObject *args)
{
    DateObject  *no;
    struct tm   *tm;
    time_t      t;
    int32_t     sdn, r;

    if(! PyArg_ParseTuple(args, ""))
        return NULL;
    t = time(NULL);
    tm = localtime(&t);
    if((r = date2sdn(tm->tm_year+1900, tm->tm_mon + 1, tm->tm_mday, &sdn)) != 0) {
        PyErr_SetString(DateRangeError, code2err(err_date2sdn, r) );
        return NULL;
    } 
    if(! (no = PyObject_NEW(DateObject, &DateType))) {
        PyErr_SetString(PyExc_MemoryError, "Out of memory.");
        return NULL;
    }
    no->sdn = sdn;
    no->argument = NULL;
    return (PyObject *) no;
}

static void date_dealloc (DateObject *d)
{
    if(d) PyMem_DEL(d);
}

static PyObject *date_repr (DateObject *d)
{
    DateTimeAttr    dta;
    char        buf[128];

    date2dta(d, &dta);
    strcpy(buf, "<Date: ");
    d_unix(&dta, &(buf[7]));
    strcat(buf, ">");
    return PyString_FromString(buf);
}

static PyObject *date_str (DateObject *d)
{
    DateTimeAttr    dta;
    char        buf[128];

    date2dta(d, &dta);
    return PyString_FromString(d_unix(&dta, buf));
}

static long date_hash (DateObject *d)
{
    return (long) d->sdn;
}

static int date_cmp(DateObject *o1, DateObject *o2)
{
    if(o1->sdn < o2->sdn) return -1;
    if(o1->sdn == o2->sdn) return 0;
    else return 1;
}

static int date_coerce(PyObject **a1, PyObject **a2)
{
    if(PyDate_Check(*a1)) {
        /* Date + Interval */
        if(PyInterval_Check(*a2)) {
            Py_INCREF(*a1);
            Py_INCREF(*a2);
            return 0;
        }
        /* Date + Datetime => Date will be converted to Datetime */
        if(PyDateTime_Check(*a2)) {
            DateTimeObject  *no = NULL, so, rso;

            so.sdn = ((DateObject *)(*a1))->sdn;
            so.time = 0;
            so.usec = 0;
            if(local2gmt_fix(&so, &rso, NULL) < 0) {
                PyErr_SetString(DateError,
                                "Cannot recompute localtime to GMT.");
                return -1;
            }
            if(! (no = PyObject_NEW(DateTimeObject, &DateTimeType))) {
                PyErr_SetString(PyExc_MemoryError, "Out of memory.");
                return -1;
            }
            no->sdn = rso.sdn;
            no->time = rso.time;
            no->usec = 0;
            no->argument = NULL;
            *a1 = (PyObject *) no;
            Py_INCREF(*a2);
            return 0;
        }
        /* Date + Number => Number will be stored in Date object and
         * Number will be coerced to Date */
        if(PyNumber_Check(*a2)) {
            Py_INCREF(*a2); 
            Py_XDECREF(((DateObject *)(*a1))->argument);
            ((DateObject *)(*a1))->argument = *a2;
            *a2 = *a1;
            Py_INCREF(*a1);
            Py_INCREF(*a2);
            return 0;
        }
    }
    /* Cannot coerce */
    return 1;
}

static PyObject *date_number_add(DateObject *d, PyObject *a)
{
    double      nd;
    DateObject  *no = NULL;
    int32_t     sdn;

    if (PyInterval_Check(a)) {
        if(((IntervalObject *)a)->dir == '-')
            sdn = d->sdn - ((IntervalObject *)a)->days;
        else
            sdn = d->sdn + ((IntervalObject *)a)->days;
        if(sdn < SDN_MIN || sdn > SDN_MAX) {
            PyErr_SetString(DateRangeError, "Date range overflow.");
            return NULL;
        }
        if(! (no = PyObject_NEW(DateObject, &DateType))) {
            PyErr_SetString(PyExc_MemoryError, "Out of memory.");
            return NULL;
        }
        no->sdn = sdn;
        no->argument = 0;
        return (PyObject *) no;
    } else if (PyDate_Check(a)) {
        if (d == (DateObject *)a && ((DateObject *)a)->argument) {
            nd = PyFloat_AsDouble(((DateObject *)a)->argument);
            sdn = d->sdn + (int32_t) nd;
            Py_DECREF(((DateObject *)a)->argument);
            ((DateObject *)a)->argument = 0;
            if(sdn < SDN_MIN || sdn > SDN_MAX) {
                PyErr_SetString(DateRangeError, "Date range overflow.");
                return NULL;
            }
            if(! (no = PyObject_NEW(DateObject, &DateType))) {
                PyErr_SetString(PyExc_MemoryError, "Out of memory.");
                return NULL;
            }
            no->sdn = sdn;
            no->argument = 0;
            return (PyObject *) no;
        } else {
            Py_DECREF(no);
            PyErr_SetString(PyExc_TypeError,
                        "Addintion is not supported on Date values.");
            return (PyObject *) NULL;
        }
    } else {
        Py_DECREF(no);
        PyErr_SetString(PyExc_TypeError,
                    "Addition is not supported for these arguments.");
        return (PyObject *) NULL;
    }
}

static PyObject *date_number_sub(DateObject *d, PyObject *a)
{
    double      nd;
    DateObject  *no = NULL;
    int32_t     sdn;

    if (PyInterval_Check(a)) {
        sdn = d->sdn - ((IntervalObject *)a)->days;
        if(sdn < SDN_MIN || sdn > SDN_MAX) {
            PyErr_SetString(DateRangeError, "Date range overflow.");
            return NULL;
        }
        if(! (no = PyObject_NEW(DateObject, &DateType))) {
            PyErr_SetString(PyExc_MemoryError, "Out of memory.");
            return NULL;
        }
        no->sdn = sdn;
        no->argument = 0;
        return (PyObject *) no;
    } else if (PyDate_Check(a)) {
        if (d == (DateObject *)a && ((DateObject *)a)->argument) {
            nd = PyFloat_AsDouble(((DateObject *)a)->argument);
            sdn = d->sdn - (int32_t) nd;
            Py_DECREF(((DateObject *)a)->argument);
            ((DateObject *)a)->argument = 0;
            if(sdn < SDN_MIN || sdn > SDN_MAX) {
                PyErr_SetString(DateRangeError, "Date range overflow.");
                return NULL;
            }
            if(! (no = PyObject_NEW(DateObject, &DateType))) {
                PyErr_SetString(PyExc_MemoryError, "Out of memory.");
                return NULL;
            }
            no->sdn = sdn;
            no->argument = 0;
            return (PyObject *) no;
        } else {
            return PyInt_FromLong(d->sdn - ((DateObject *)a)->sdn);
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
                    "Subtraction is not supported for these arguments.");
        return (PyObject *) NULL;
    }
}

static int date_number_nonzero (DateObject *d)
{
    return 1;
}

static PyObject *date_number_as_int (DateObject *d)
{
    return PyInt_FromLong(d->sdn );
}

static PyObject *date_number_as_long (DateObject *d)
{
    return PyLong_FromLong(d->sdn   );
}

static PyObject *date_number_as_float (DateObject *d)
{
    return PyFloat_FromDouble((double) d->sdn );
}

static PyNumberMethods date_number_methods = {
    (binaryfunc) date_number_add,              /*nb_add*/
    (binaryfunc) date_number_sub,              /*nb_subtract*/
    (binaryfunc) number_not_implemented2,          /*nb_multiply*/
    (binaryfunc) number_not_implemented2,          /*nb_divide*/
    (binaryfunc) number_not_implemented2,          /*nb_remainder*/
    (binaryfunc) number_not_implemented2,          /*nb_divmod*/
    (ternaryfunc) number_not_implemented3,         /*nb_power*/
    (unaryfunc) number_not_implemented1,           /*nb_negative*/
    (unaryfunc) number_not_implemented1,           /*nb_positive*/
     /* Dzis is NULL checked */
    (unaryfunc) 0,                     /*nb_absolute*/
    (inquiry) date_number_nonzero,             /*nb_nonzero*/
    (unaryfunc) 0,                     /*nb_invert*/
    (binaryfunc) 0,                    /*nb_lshift*/
    (binaryfunc) 0,                    /*nb_rshift*/
    (binaryfunc) 0,                    /*nb_and*/
    (binaryfunc) 0,                    /*nb_xor*/
    (binaryfunc) 0,                    /*nb_or*/
    (coercion) date_coerce,                /*nb_coerce*/
    (unaryfunc) date_number_as_int,            /*nb_int*/
    (unaryfunc) date_number_as_long,           /*nb_long*/
    (unaryfunc) date_number_as_float,          /*nb_float*/
    (unaryfunc) 0,                     /*nb_oct*/
    (unaryfunc) 0,                     /*nb_hex*/
};

static PyObject *date_fmt (DateObject *d, PyObject *args)
{
    char        *fmt, b[512], sb[64];
    int     i, j, l, p;
    DateTimeAttr    dta;

    if(! PyArg_ParseTuple(args, "s", &fmt))
        return NULL;
    if(date2dta(d, &dta) != 0) {
        PyErr_SetString(DateError, "System error in Date.fmt:date2dta.");
        return NULL;
    }
    l = strlen(fmt);
    b[0] = '\0';
    for(i = 0, p = 0, j = 0; i < l; i++) {
        if(p == 1) {
            switch(fmt[i]) {
                case 's' : /* sdn */
                    sprintf(sb, "%d", d->sdn);
                    strcat(b, sb);
                    j += strlen(sb);
                    break;
                case 'Y' : /* year */
                    sprintf(sb, "%d", dta.year);
                    strcat(b, sb);
                    j += strlen(sb);
                    break;
                case 'm' : /* month */
                    sprintf(sb, "%d", dta.month);
                    strcat(b, sb);
                    j += strlen(sb);
                    break;
                case 'B' : /* month_name */
                    strcat(b, dta.month_name);
                    j += strlen(dta.month_name);
                    break;
                case 'b' : /* month_short */
                    strcat(b, dta.month_short);
                    j += strlen(dta.month_short);
                    break;
                case 'd' : /* day */
                    sprintf(sb, "%d", dta.day);
                    strcat(b, sb);
                    j += strlen(sb);
                    break;
                case 'w' : /* wday */
                    sprintf(sb, "%d", dta.wday);
                    strcat(b, sb);
                    j += strlen(sb);
                    break;
                case 'A' : /* wday_name */
                    strcat(b, dta.wday_name);
                    j += strlen(dta.wday_name);
                    break;
                case 'a' : /* wday_short */
                    strcat(b, dta.wday_short);
                    j += strlen(dta.wday_short);
                    break;
                case 'j' : /* yday */
                    sprintf(sb, "%d", dta.yday);
                    strcat(b, sb);
                    j += strlen(sb);
                    break;
                case 'U' : /* week */
                    sprintf(sb, "%d", dta.week);
                    strcat(b, sb);
                    j += strlen(sb);
                    break;
                case 'E' : /* euro */
                    strcat(b, d_euro(&dta, sb));
                    j += strlen(sb);
                    break;
                case 'P' : /* postgres */
                    strcat(b, d_postgres(&dta, sb));
                    j += strlen(sb);
                    break;
                case 'Q' : /* sql */
                    strcat(b, d_sql(&dta, sb));
                    j += strlen(sb);
                    break;
                case 'I' : /* iso */
                    strcat(b, d_iso(&dta, sb));
                    j += strlen(sb);
                    break;
                case 'X' : /* unix */
                    strcat(b, d_unix(&dta, sb));
                    j += strlen(sb);
                    break;
                case 'R' : /* rfc */
                    strcat(b, d_rfc(&dta, sb));
                    j += strlen(sb);
                    break;
                case '%' : /* % */
                    strcat(b, "%");
                    j++;
                    break;
                default :
                    PyErr_SetString(DateError, "Unknown format character.");
                    return NULL;
            }
            p = 0;
        } else if(fmt[i] == '%') {
            p = 1;
        } else {
            b[j] = fmt[i];
            j++;
            b[j] = '\0';
        }
        if(j > 480) {
            PyErr_SetString(DateError, "Maximum length exceeded.");
            return NULL;
        }

    }
    return PyString_FromString(b);
}

static PyObject *date_reduce (DateObject *d, PyObject *args)
{
    PyObject    *tuple=NULL, *pickle=NULL, *construct=NULL;

    if(! PyArg_ParseTuple(args, ""))
        return NULL;
    if(!(construct=PyDict_GetItemString(module_dict,"DateFromSDN")))
        goto error;
    Py_INCREF(construct);
    if(!(tuple = PyTuple_New(2)))
        goto error;
    if(!(pickle = PyTuple_New(1)))
        goto error;

    PyTuple_SetItem(pickle, 0, PyInt_FromLong(d->sdn));
    PyTuple_SetItem(tuple, 0, construct);
    PyTuple_SetItem(tuple, 1, pickle);
    return tuple;

error:
    Py_XDECREF(construct);
    Py_XDECREF(tuple);
    Py_XDECREF(pickle);
    return NULL;
}

static struct PyMethodDef date_methods[] = {
      { "fmt", (PyCFunction) date_fmt, 1 } ,
      { "__reduce__", (PyCFunction) date_reduce, 1 } ,
      { NULL, NULL }
};

static PyObject *date_getattr (DateObject *d, char *name)
{
    DateTimeAttr    dta;
    char        b[128];
    int     r;
    int32_t     sdn;
    DateObject  *no;

    date2dta(d, &dta);
    if(strcmp(name, "sdn") == 0) {
        return PyInt_FromLong(d->sdn);
    } else if(strcmp(name, "year") == 0) {
        return PyInt_FromLong(dta.year);
    } else if(strcmp(name, "month") == 0) {
        return PyInt_FromLong(dta.month);
    } else if(strcmp(name, "month_name") == 0) {
        return PyString_FromString(dta.month_name);
    } else if(strcmp(name, "month_short") == 0) {
        return PyString_FromString(dta.month_short);
    } else if(strcmp(name, "day") == 0) {
        return PyInt_FromLong(dta.day);
    } else if(strcmp(name, "wday") == 0) {
        return PyInt_FromLong(dta.wday);
    } else if(strcmp(name, "wday_name") == 0) {
        return PyString_FromString(dta.wday_name);
    } else if(strcmp(name, "wday_short") == 0) {
        return PyString_FromString(dta.wday_short);
    } else if(strcmp(name, "yday") == 0) {
        return PyInt_FromLong(dta.yday);
    } else if(strcmp(name, "week") == 0) {
        return PyInt_FromLong(dta.week);
    } else if(strcmp(name, "euro") == 0) {
        return PyString_FromString(d_euro(&dta, b));
    } else if(strcmp(name, "postgres") == 0) {
        return PyString_FromString(d_postgres(&dta, b));
    } else if(strcmp(name, "sql") == 0) {
        return PyString_FromString(d_sql(&dta, b));
    } else if(strcmp(name, "iso") == 0) {
        return PyString_FromString(d_iso(&dta, b));
    } else if(strcmp(name, "unix") == 0) {
        return PyString_FromString(d_unix(&dta, b));
    } else if(strcmp(name, "rfc") == 0) {
        return PyString_FromString(d_rfc(&dta, b));
    } else if(strcmp(name, "leap") == 0) {
        return PyInt_FromLong(dta.leap);
    } else if(strcmp(name, "first") == 0) {
        r = date2sdn(dta.year,dta.month, 1, &sdn);
        if(!(no = PyObject_NEW(DateObject, &DateType))) {
            PyErr_SetString(PyExc_MemoryError, "Out of memory.");
            return NULL;
        }
        no->sdn = sdn;
        no->argument = NULL;
        return (PyObject *) no;
    } else if(strcmp(name, "last") == 0) {
        r = date2sdn(dta.year,dta.month, 1, &sdn);
        sdn = sdn + 32;
        sdn2date(sdn, &dta.year, &dta.month, &r);
        r = date2sdn(dta.year,dta.month, 1, &sdn);
        if(!(no = PyObject_NEW(DateObject, &DateType))) {
            PyErr_SetString(PyExc_MemoryError, "Out of memory.");
            return NULL;
        }
        no->sdn = sdn-1;
        no->argument = NULL;
        return (PyObject *) no;
    } else if(strcmp(name, "__members__") == 0) {
        return Py_BuildValue("[ssssssssssssssssss]",\
            "sdn", "year", "month", "month_name",
            "month_short", "day", "yday",\
            "wday", "wday_name", "wday_short", "week",\
            "euro", "postgres", "sql", "iso", "unix", "rfc",\
            "leap", "first", "last", "fmt", "__reduce__");
    } else {
        return Py_FindMethod(date_methods, (PyObject *) d, name);
    }
}

static PyTypeObject DateType = {
    PyObject_HEAD_INIT(&PyType_Type)
    0,
    "Date" ,
    sizeof(DateObject) ,
    0 ,
    (destructor) date_dealloc ,
    (printfunc) 0 ,
    (getattrfunc) date_getattr ,
    (setattrfunc) 0 ,
    (cmpfunc) date_cmp ,
    (reprfunc) date_repr ,
    (PyNumberMethods *) &date_number_methods ,
    (PySequenceMethods *) 0 ,
    (PyMappingMethods *) 0 ,
    (hashfunc) date_hash ,
    (ternaryfunc) 0 ,  /* call */
    (reprfunc) date_str ,
    (getattrofunc) 0 ,
    (setattrofunc) 0 ,
    0 ,
    0 ,
    NULL
};


/***********************************************************
 *                                                         *
 *            DateTime                                     *
 *                                                         *
 ***********************************************************/

static PyObject * datetime_DateTime (PyObject *self, PyObject *args)
{
    int     day = 1, month = 1, year, z, r;
    int     hour = 0, min = 0 , sec = 0, usec = 0;
    char        *zone = NULL;
    DateTimeObject  *no, so;
    DateTimeAttr    dta;

    if(! PyArg_ParseTuple(args, "i|iiiiiiz",
                    &year, &month, &day, &hour, &min, &sec, &usec, &zone))
        return NULL;
    memset(&dta, 0, sizeof(dta));
    dta.year = year;
    dta.month = month;
    dta.day = day;
    dta.hour = hour;
    dta.min = min;
    dta.sec = sec;
    dta.usec = usec;
    if(zone) {
        if(zone[0] == '+' || zone[0] == '-') {
            dta.tztype = TZ_NUMERIC;
            if(strlen(zone) > 7) {
                PyErr_SetString(DateError, "Invalid timezone format.");
                return NULL;
            }
            strncpy(dta.tz.numeric.tz, zone, 7);
        } else {
            dta.tztype = TZ_ZONE;
            z = findzone(zone);
            if(z < 0) {
                PyErr_SetString(DateError, "Unknown timezone.");
                return NULL;
            }
            dta.tz.zone.tz = zone_map[z].tz;
        }
    } else {
        dta.tztype = TZ_NONE;
    }
    if((r=dta2datetime(&dta, &so)) < 0) {
        PyErr_SetString(DateRangeError, code2err(err_dta2datetime, r));
        return NULL;
    }
    if(! (no = PyObject_NEW(DateTimeObject, &DateTimeType))) {
        PyErr_SetString(PyExc_MemoryError, "Out of memory.");
        return NULL;
    }
    no->sdn = so.sdn;
    no->time = so.time;
    no->usec = so.usec;
    no->argument = NULL;
    return (PyObject *) no;
}

static PyObject * datetime_DateTimeFromString (PyObject *self, PyObject *args)
{
    DateTimeObject  *no, so;
    DateTimeAttr    dta;
    char            *str, b[128];
    int             r;

    if(! PyArg_ParseTuple(args, "s", &str))
        return NULL;
    if((r=parsedate(str, &dta)) != 0) {
        snprintf(b, 127, "Invalid DateTime format: `%s' (%d).", str, r);
        b[127] = '\0';
        PyErr_SetString(DateError, b);
        return NULL;
    }
    if((r=dta2datetime(&dta, &so)) < 0) {
        PyErr_SetString(DateRangeError, code2err(err_dta2datetime, r));
        return NULL;
    }
    if(! (no = PyObject_NEW(DateTimeObject, &DateTimeType))) {
        PyErr_SetString(PyExc_MemoryError, "Out of memory.");
        return NULL;
    }
    no->sdn = so.sdn;
    no->time = so.time;
    no->usec = so.usec;
    no->argument = NULL;
    return (PyObject *) no;
}

static PyObject * datetime_DateTimeFromSDNTime (PyObject *self, PyObject *args)
{
    DateTimeObject  *no;
    int32_t     sdn, tm, usec = 0;

    if(! PyArg_ParseTuple(args, "ii|i", &sdn, &tm, &usec))
        return NULL;
    if(sdn < SDN_MIN || sdn > SDN_MAX) {
        PyErr_SetString(DateRangeError, "SDN must be in range 2 - 536838576.");
        return NULL;
    } 
    if(tm < TIME_MIN || tm >= TIME_MAX) {
        PyErr_SetString(DateRangeError, "Time must be in range 0 - 86399.");
        return NULL;
    } 
    if(tm < USEC_MIN || tm > USEC_MAX) {
        PyErr_SetString(DateRangeError,
                    "Microseconds must be in range 0 - 999999.");
        return NULL;
    } 
    if(! (no = PyObject_NEW(DateTimeObject, &DateTimeType))) {
        PyErr_SetString(PyExc_MemoryError, "Out of memory.");
        return NULL;
    }
    no->sdn = sdn;
    no->time =  tm;
    no->usec = usec;
    no->argument = NULL;
    return (PyObject *) no;
}

static PyObject * datetime_DateTimeFromUNIX (PyObject *self, PyObject *args)
{
    DateTimeObject  *no, so;
    DateTimeAttr    dta;
    double      f;
    time_t      t;
    struct tm   *tm;
    int     r;

    if(! PyArg_ParseTuple(args, "d", &f))
        return NULL;
    t = (time_t) f;
    tm = gmtime(&t);
    memset(&dta, 0, sizeof(dta));
    dta.year = tm->tm_year + 1900;
    dta.month = tm->tm_mon + 1;
    dta.day = tm->tm_mday;
    dta.hour = tm->tm_hour;
    dta.min = tm->tm_min;
    dta.sec = tm->tm_sec;
    dta.tztype = TZ_ZONE;
    dta.tz.zone.tz = "GMT";
    if((r=dta2datetime(&dta, &so)) < 0) {
        PyErr_SetString(DateRangeError, code2err(err_dta2datetime, r));
        return NULL;
    }
    if(! (no = PyObject_NEW(DateTimeObject, &DateTimeType))) {
        PyErr_SetString(PyExc_MemoryError, "Out of memory.");
        return NULL;
    }
    no->sdn = so.sdn;
    no->time = so.time;
    no->usec = so.usec;
    no->argument = NULL;
    return (PyObject *) no;
}

static PyObject * datetime_DateTimeFromDate (PyObject *self, PyObject *args)
{
    DateTimeObject  *no, so, rso;
    PyObject    *po;

    if(! PyArg_ParseTuple(args, "O", &po))
        return NULL;
    if(! PyDate_Check(po)) {
        PyErr_SetString(DateError, "Argument must be Date instance.");
        return NULL;
    } 
    so.sdn = ((DateObject *) po)->sdn;
    so.time = 0;
    so.usec = 0;
    if(local2gmt_fix(&so, &rso, NULL) < 0) {
        PyErr_SetString(DateError,
                    "System error in DateFromDateTime:local2gmt_fix.");
        return NULL;
    } 
    if(! (no = PyObject_NEW(DateTimeObject, &DateTimeType))) {
        PyErr_SetString(PyExc_MemoryError, "Out of memory.");
        return NULL;
    }
    no->sdn = rso.sdn;
    no->time = rso.time;
    no->usec = 0;
    no->argument = NULL;
    return (PyObject *) no;
}

static PyObject * datetime_Now (PyObject *self, PyObject *args)
{
    DateTimeObject  *no, so;
    DateTimeAttr    dta;
    struct tm   *tm;
    time_t      t;
    int     r;

    if(! PyArg_ParseTuple(args, ""))
        return NULL;
    t = time(NULL);
    tm = gmtime(&t);
    memset(&dta, 0, sizeof(dta));
    dta.year = tm->tm_year + 1900;
    dta.month = tm->tm_mon + 1;
    dta.day = tm->tm_mday;
    dta.hour = tm->tm_hour;
    dta.min = tm->tm_min;
    dta.sec = tm->tm_sec;
    dta.tztype = TZ_ZONE;
    dta.tz.zone.tz = "GMT";
    if((r=dta2datetime(&dta, &so)) < 0) {
        PyErr_SetString(DateRangeError, code2err(err_dta2datetime, r));
        return NULL;
    }
    if(! (no = PyObject_NEW(DateTimeObject, &DateTimeType))) {
        PyErr_SetString(PyExc_MemoryError, "Out of memory.");
        return NULL;
    }
    no->sdn = so.sdn;
    no->time = so.time;
    no->usec = so.usec;
    no->argument = NULL;
    return (PyObject *) no;
}

static void datetime_dealloc (DateTimeObject *d)
{
    if(d) PyMem_DEL(d);
}

static PyObject *datetime_str (DateTimeObject *d)
{
    DateTimeAttr    dta;
    char        b[128];

    datetime2dta(d, &dta, NULL);
    return PyString_FromString(dt_unix(&dta, b));
}

static PyObject *datetime_repr (DateTimeObject *d)
{
    DateTimeAttr    dta;
    char        b[128];

    datetime2dta(d, &dta, NULL);
    strcpy(b, "<DateTime: ");
    dt_unix(&dta, &(b[11]));
    strcat(b, ">");
    return PyString_FromString(b);
}

static long datetime_hash (DateTimeObject *d)
{
    return (long)d->usec*1000 ^ (long)d->time ^ (long)d->sdn;
}

static int datetime_cmp(DateTimeObject *o1, DateTimeObject *o2)
{
    if(o1->sdn < o2->sdn) return -1;
    if(o1->sdn > o2->sdn) return 1;
    if(o1->time < o2->time) return -1;
    if(o1->time > o2->time) return 1;
    if(o1->usec < o2->usec) return -1;
    if(o1->usec > o2->usec) return 1;
    else return 0;
}

static int datetime_coerce(PyObject **a1, PyObject **a2)
{
    if(PyDateTime_Check(*a1)) {
        /* DateTime  + Interval */
        if(PyInterval_Check(*a2)) {
            Py_INCREF(*a1);
            Py_INCREF(*a2);
            return 0;
        /* DateTime  + Date => Date will be converted to DateTime */
        } else if(PyDate_Check(*a2)) {
            DateTimeObject  *no = NULL, so, rso;

            so.sdn = ((DateObject *)(*a1))->sdn;
            so.time = 0;
            so.usec = 0;
            if(local2gmt_fix(&so, &rso, NULL) < 0) {
                PyErr_SetString(DateError,"Cannot recompute localtime to GMT.");
                return -1;
            }
            if(! (no = PyObject_NEW(DateTimeObject, &DateTimeType))) {
                PyErr_SetString(PyExc_MemoryError, "Out of memory.");
                return -1;
            }
            no->sdn = rso.sdn;
            no->time = rso.time;
            no->usec = 0;
            no->argument = 0;
            Py_INCREF(*a1);
            *a2 = (PyObject *) no;
            return 0;
        /* DateTime + Number => Number will be stored in DateTime
         * object and Number will be coerced to DateTime */
        } else  if(PyNumber_Check(*a2)) {
            Py_INCREF(*a2);
            Py_XDECREF(((DateTimeObject *)(*a1))->argument);
            ((DateTimeObject *)(*a1))->argument = *a2;
            *a2 = *a1;
            Py_INCREF(*a1);
            Py_INCREF(*a2);
            return 0;
        }
    }
    /* Cannot coerce */
    return 1;
}

static PyObject *datetime_number_add(DateTimeObject *d, PyObject *a)
{
    double      nd;
    int     c = 0;
    DateTimeObject  *no = NULL, so;
    IntervalObject  *iv, aiv;

    if (PyInterval_Check(a)) {
        iv = (IntervalObject *) a;
        if(iv->dir == '-') {
            so.usec = d->usec - iv->usec;
            if(so.usec < 0) { c = 1; so.usec += USEC_MAX + 1; }
            else { c = 0; } 
            so.time = d->time - iv->time - c;
            if(so.time < 0) { c = 1; so.time += TIME_MAX + 1; }
            else { c = 0; } 
            so.sdn = d->sdn - iv->days - c;
        } else {
            so.usec = d->usec + iv->usec;
            if(so.usec > USEC_MAX) { c = 1; so.usec -= USEC_MAX + 1; }
            else { c = 0; } 
            so.time = d->time + iv->time + c;
            if(so.time > TIME_MAX) { c = 1; so.time -= TIME_MAX + 1; }
            else { c = 0; } 
            so.sdn = d->sdn + iv->days + c;
        }
        if(so.sdn < SDN_MIN || so.sdn > SDN_MAX) {
            PyErr_SetString(DateRangeError, "Result is out of valid range.");
            return NULL;
        }
        if(! (no = PyObject_NEW(DateTimeObject, &DateTimeType))) {
            PyErr_SetString(PyExc_MemoryError, "Out of memory.");
            return NULL;
        }
        no->sdn = so.sdn;
        no->time = so.time;
        no->usec = so.usec;
        no->argument = 0;
        return (PyObject *) no;
    } else if (PyDateTime_Check(a)) {
        if (d == (DateTimeObject *)a && ((DateTimeObject *)a)->argument) {
            nd = PyFloat_AsDouble(((DateTimeObject *)a)->argument);
            Py_DECREF(((DateTimeObject *)a)->argument);
            ((DateTimeObject *)a)->argument = 0;
            double2interval(nd, &aiv);
            iv = &aiv;
            if(iv->dir == '-') {
                so.usec = d->usec - iv->usec;
                if(so.usec < 0) { c = 1; so.usec += USEC_MAX + 1; }
                else { c = 0; } 
                so.time = d->time - iv->time - c;
                if(so.time < 0) { c = 1; so.time += TIME_MAX + 1; }
                else { c = 0; } 
                so.sdn = d->sdn - iv->days - c;
            } else {
                so.usec = d->usec + iv->usec;
                if(so.usec > USEC_MAX) { c = 1; so.usec -= USEC_MAX + 1; }
                else { c = 0; } 
                so.time = d->time + iv->time + c;
                if(so.time > TIME_MAX) { c = 1; so.time -= TIME_MAX + 1; }
                else { c = 0; } 
                so.sdn = d->sdn + iv->days + c;
            }
            if(so.sdn < SDN_MIN || so.sdn > SDN_MAX) {
                PyErr_SetString(DateRangeError,
                                "Result is out of valid range.");
                return NULL;
            }
            if(! (no = PyObject_NEW(DateTimeObject, &DateTimeType))) {
                PyErr_SetString(PyExc_MemoryError, "Out of memory.");
                return NULL;
            }
            no->sdn = so.sdn;
            no->time = so.time;
            no->usec = so.usec;
            no->argument = 0;
            return (PyObject *) no;
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "Addintion is not supported on DateTime values.");
            return (PyObject *) NULL;
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
                        "Addition is not supported for these arguments.");
        return (PyObject *) NULL;
    }
}

static PyObject *datetime_number_sub(DateTimeObject *d, PyObject *a)
{
    double      nd;
    int     c = 0;
    DateTimeObject  *no = NULL, *ta, so;
    IntervalObject  *iv, aiv, *niv;

    if (PyInterval_Check(a)) {
        /* DateTime - Interval */
        iv = (IntervalObject *) a;
        if(iv->dir == '+') {
            so.usec = d->usec - iv->usec;
            if(so.usec < 0) { c = 1; so.usec += USEC_MAX + 1; }
            else { c = 0; } 
            so.time = d->time - iv->time - c;
            if(so.time < 0) { c = 1; so.time += TIME_MAX + 1; }
            else { c = 0; } 
            so.sdn = d->sdn - iv->days - c;
        } else {
            so.usec = d->usec + iv->usec;
            if(so.usec > USEC_MAX) { c = 1; so.usec -= USEC_MAX + 1; }
            else { c = 0; } 
            so.time = d->time + iv->time + c;
            if(so.time > TIME_MAX) { c = 1; so.time -= TIME_MAX + 1; }
            else { c = 0; } 
            so.sdn = d->sdn + iv->days + c;
        }
        if(so.sdn < SDN_MIN || so.sdn > SDN_MAX) { 
            PyErr_SetString(DateRangeError, "Result is out of valid range.");
            return NULL;
        }
        if(! (no = PyObject_NEW(DateTimeObject, &DateTimeType))) {
            PyErr_SetString(PyExc_MemoryError, "Out of memory.");
            return NULL;
        }
        no->sdn = so.sdn;
        no->time = so.time;
        no->usec = so.usec;
        no->argument = 0;
        return (PyObject *) no;
    } else if (PyDateTime_Check(a)) {
        if (d == (DateTimeObject *)a && ((DateTimeObject *)a)->argument) {
            /* DateTime - Number */
            nd = PyFloat_AsDouble(((DateTimeObject *)a)->argument);
            Py_DECREF(((DateTimeObject *)a)->argument);
            ((DateTimeObject *)a)->argument = 0;
            double2interval(nd, &aiv);
            iv = &aiv;
            if(iv->dir == '+') {
                so.usec = d->usec - iv->usec;
                if(so.usec < 0) { c = 1; so.usec += USEC_MAX + 1; }
                else { c = 0; } 
                so.time = d->time - iv->time - c;
                if(so.time < 0) { c = 1; so.time += TIME_MAX + 1; }
                else { c = 0; } 
                so.sdn = d->sdn - iv->days - c;
            } else {
                so.usec = d->usec + iv->usec;
                if(so.usec > USEC_MAX) { c = 1; so.usec -= USEC_MAX + 1; }
                else { c = 0; } 
                so.time = d->time + iv->time + c;
                if(so.time > TIME_MAX) { c = 1; so.time -= TIME_MAX + 1; }
                else { c = 0; } 
                so.sdn = d->sdn + iv->days + c;
            }
            if(so.sdn < SDN_MIN || so.sdn > SDN_MAX) { 
                PyErr_SetString(DateRangeError, "Result is out of valid range.");
                return NULL;
            }
            if(! (no = PyObject_NEW(DateTimeObject, &DateTimeType))) {
                PyErr_SetString(PyExc_MemoryError, "Out of memory.");
                return NULL;
            }
            no->sdn = so.sdn;
            no->time = so.time;
            no->usec = so.usec;
            no->argument = 0;
            return (PyObject *) no;
        } else {
            /* DateTime - DateTime */
            ta = (DateTimeObject *) a;
            if(datetime_cmp(d, ta) < 0)  {
                DateTimeObject *swp = ta;
                ta = d;
                d = swp;
                aiv.dir = '-';
            } else {
                aiv.dir = '+';
            }
            aiv.usec = d->usec - ta->usec;
            if(aiv.usec < 0) { c = 1; aiv.usec += USEC_MAX + 1; }
            else c = 0;
            aiv.time = d->time - ta->time - c;
            if(aiv.time < 0) { c = 1; aiv.time += TIME_MAX + 1; }
            else c = 0;
            aiv.days = d->sdn - ta->sdn - c;
            if(! (niv = PyObject_NEW(IntervalObject, &IntervalType))) {
                PyErr_SetString(PyExc_MemoryError, "Out of memory.");
                return NULL;
            }
            niv->dir = aiv.dir;
            niv->days = aiv.days;
            niv->time = aiv.time;
            niv->usec = aiv.usec;
            niv->argument = 0;
            return (PyObject *) niv;
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
                        "Subtraction is not supported for these arguments.");
        return (PyObject *) NULL;
    }
}

static int datetime_number_nonzero (DateTimeObject *d)
{
    return 1;
}

static PyNumberMethods datetime_number_methods = {
    (binaryfunc) datetime_number_add,          /*nb_add*/
    (binaryfunc) datetime_number_sub,          /*nb_subtract*/
    (binaryfunc) number_not_implemented2,          /*nb_multiply*/
    (binaryfunc) number_not_implemented2,          /*nb_divide*/
    (binaryfunc) number_not_implemented2,          /*nb_remainder*/
    (binaryfunc) number_not_implemented2,          /*nb_divmod*/
    (ternaryfunc) number_not_implemented3,         /*nb_power*/
    (unaryfunc) number_not_implemented1,           /*nb_negative*/
    (unaryfunc) number_not_implemented1,           /*nb_positive*/
     /* Dzis is NULL checked */
    (unaryfunc) 0,                     /*nb_absolute*/
    (inquiry) datetime_number_nonzero ,        /*nb_nonzero*/
    (unaryfunc) 0,                     /*nb_invert*/
    (binaryfunc) 0,                    /*nb_lshift*/
    (binaryfunc) 0,                    /*nb_rshift*/
    (binaryfunc) 0,                    /*nb_and*/
    (binaryfunc) 0,                    /*nb_xor*/
    (binaryfunc) 0,                    /*nb_or*/
    (coercion) datetime_coerce,            /*nb_coerce*/
    (unaryfunc) 0,                     /*nb_int*/
    (unaryfunc) 0,                     /*nb_long*/
    (unaryfunc) 0,                     /*nb_float*/
    (unaryfunc) 0,                     /*nb_oct*/
    (unaryfunc) 0,                     /*nb_hex*/
};

static PyObject *datetime_trunc (DateTimeObject *d, PyObject *args)
{
    char            *how;
    DateTimeObject  *no;
    int             t;

    if(! PyArg_ParseTuple(args, "s", &how))
        return NULL;
    if(strlen(how) != 1) {
        PyErr_SetString(DateError, "One of characters 'ymdHMS' required.");
        return NULL;
        }
    if(! (no = PyObject_NEW(DateTimeObject, &DateTimeType))) {
        PyErr_SetString(PyExc_MemoryError, "Out of memory.");
        return NULL;
    }
    no->sdn = d->sdn;
    no->time = d->time;
    no->usec = 0;
    no->argument = NULL;
    switch(*how) {
        case 'y' :
        case 'm' :
            PyErr_SetString(PyExc_NotImplementedError, "Not implemented.");
            return NULL;
        case 'd' :
            /* no->time = 0; is not sufficient it works only in GMT domain */
            PyErr_SetString(PyExc_NotImplementedError, "Not implemented.");
            return NULL;
        case 'H' :
            t = d->time;
            no->time = (t/3600) * 3600;
            return (PyObject *) no;
        case 'M' :
            t = d->time;
            no->time = (t/60) * 60;
            return (PyObject *) no;
        case 'S' :
            return (PyObject *) no;
        default :
            PyErr_SetString(DateError, "One of characters 'ymdHMS' required.");
            return NULL;
    }
}



static PyObject *datetime_fmt (DateTimeObject *d, PyObject *args)
{
    char        *fmt, b[512], sb[64], *zone = NULL;
    int     i, j, l, p;
    DateTimeAttr    dta;

    if(! PyArg_ParseTuple(args, "s|z", &fmt, &zone))
        return NULL;
    if((p = datetime2dta(d, &dta, zone) != 0)) {
        PyErr_SetString(DateError, "Bad domain format.");
        return NULL;
    }
    l = strlen(fmt);
    b[0] = '\0';
    for(i = 0, p = 0, j = 0; i < l; i++) {
        if(p == 1) {
            switch(fmt[i]) {
                case 's' : /* sdn */
                    sprintf(sb, "%d", d->sdn);
                    strcat(b, sb);
                    j += strlen(sb);
                    break;
                case 'Y' : /* year */
                    sprintf(sb, "%d", dta.year);
                    strcat(b, sb);
                    j += strlen(sb);
                    break;
                case 'm' : /* month */
                    sprintf(sb, "%d", dta.month);
                    strcat(b, sb);
                    j += strlen(sb);
                    break;
                case 'B' : /* month_name */
                    strcat(b, dta.month_name);
                    j += strlen(dta.month_name);
                    break;
                case 'b' : /* month_short */
                    strcat(b, dta.month_short);
                    j += strlen(dta.month_short);
                    break;
                case 'd' : /* day */
                    sprintf(sb, "%d", dta.day);
                    strcat(b, sb);
                    j += strlen(sb);
                    break;
                case 'w' : /* wday */
                    sprintf(sb, "%d", dta.wday);
                    strcat(b, sb);
                    j += strlen(sb);
                    break;
                case 'A' : /* wday_name */
                    strcat(b, dta.wday_name);
                    j += strlen(dta.wday_name);
                    break;
                case 'a' : /* wday_short */
                    strcat(b, dta.wday_short);
                    j += strlen(dta.wday_short);
                    break;
                case 'j' : /* yday */
                    sprintf(sb, "%d", dta.yday);
                    strcat(b, sb);
                    j += strlen(sb);
                    break;
                case 'U' : /* week */
                    sprintf(sb, "%d", dta.week);
                    strcat(b, sb);
                    j += strlen(sb);
                    break;
                case 'H' : /* hour */
                    sprintf(sb, "%02d", dta.hour);
                    strcat(b, sb);
                    j += strlen(sb);
                    break;
                case 'M' : /* min */
                    sprintf(sb, "%02d", dta.min);
                    strcat(b, sb);
                    j += strlen(sb);
                    break;
                case 'S' : /* sec */
                    sprintf(sb, "%02d", dta.sec);
                    strcat(b, sb);
                    j += strlen(sb);
                    break;
                case 'C' : /* usec */
                    sprintf(sb, "%06d", dta.usec);
                    strcat(b, sb);
                    j += strlen(sb);
                    break;
                case 'E' : /* euro */
                    strcat(b, dt_euro(&dta, sb));
                    j += strlen(sb);
                    break;
                case 'P' : /* postgres */
                    strcat(b, dt_postgres(&dta, sb));
                    j += strlen(sb);
                    break;
                case 'Q' : /* sql */
                    strcat(b, dt_sql(&dta, sb));
                    j += strlen(sb);
                    break;
                case 'I' : /* iso */
                    strcat(b, dt_iso(&dta, sb));
                    j += strlen(sb);
                    break;
                case 'X' : /* unix */
                    strcat(b, dt_unix(&dta, sb));
                    j += strlen(sb);
                    break;
                case 'R' : /* rfc */
                    strcat(b, dt_rfc(&dta, sb));
                    j += strlen(sb);
                    break;
                case '%' : /* % */
                    strcat(b, "%");
                    j++;
                    break;
                default :
                    PyErr_SetString(DateError, "Unknown format character.");
                    return NULL;
            }
            p = 0;
        } else if(fmt[i] == '%') {
            p = 1;
        } else {
            b[j] = fmt[i];
            j++;
            b[j] = '\0';
        }
        if(j > 480) {
            PyErr_SetString(DateError, "Maximum length exceeded.");
            return NULL;
        }

    }
    return PyString_FromString(b);
}

static PyObject *datetime_reduce (DateTimeObject *d, PyObject *args)
{
    PyObject    *tuple=NULL, *pickle=NULL, *construct=NULL;

    if(! PyArg_ParseTuple(args, ""))
        return NULL;
    if(!(construct=PyDict_GetItemString(module_dict,"DateTimeFromSDNTime")))
        goto error;
    Py_INCREF(construct);
    if(!(tuple = PyTuple_New(2)))
        goto error;
    if(!(pickle = PyTuple_New(3)))
        goto error;

    PyTuple_SetItem(pickle, 0, PyInt_FromLong(d->sdn));
    PyTuple_SetItem(pickle, 1, PyInt_FromLong(d->time));
    PyTuple_SetItem(pickle, 2, PyInt_FromLong(d->usec));

    PyTuple_SetItem(tuple, 0, construct);
    PyTuple_SetItem(tuple, 1, pickle);
    return tuple;

error:
    Py_XDECREF(construct);
    Py_XDECREF(tuple);
    Py_XDECREF(pickle);
    return NULL;
}


static struct PyMethodDef datetime_methods[] = {
      { "fmt", (PyCFunction) datetime_fmt, 1 } ,
      { "trunc", (PyCFunction) datetime_trunc, 1 } ,
      { "__reduce__", (PyCFunction) datetime_reduce, 1 } ,
      { NULL, NULL }
};

static PyObject *datetime_getattr (DateTimeObject *d, char *name)
{
    DateTimeAttr    dta;
    char        b[128];

    if(datetime2dta(d, &dta, NULL) < 0) {
        PyErr_SetString(DateError,
                        "System error in getattr:datetime2dta.");
        return NULL;
    }
    if(strcmp(name, "sdn") == 0) {
        return PyInt_FromLong(d->sdn);
    } else if(strcmp(name, "year") == 0) {
        return PyInt_FromLong(dta.year);
    } else if(strcmp(name, "month") == 0) {
        return PyInt_FromLong(dta.month);
    } else if(strcmp(name, "month_name") == 0) {
        return PyString_FromString(dta.month_name);
    } else if(strcmp(name, "month_short") == 0) {
        return PyString_FromString(dta.month_short);
    } else if(strcmp(name, "day") == 0) {
        return PyInt_FromLong(dta.day);
    } else if(strcmp(name, "wday") == 0) {
        return PyInt_FromLong(dta.wday);
    } else if(strcmp(name, "wday_name") == 0) {
        return PyString_FromString(dta.wday_name);
    } else if(strcmp(name, "wday_short") == 0) {
        return PyString_FromString(dta.wday_short);
    } else if(strcmp(name, "yday") == 0) {
        return PyInt_FromLong(dta.yday);
    } else if(strcmp(name, "week") == 0) {
        return PyInt_FromLong(dta.week);
    } else if(strcmp(name, "time") == 0) {
        return PyInt_FromLong(d->time);
    } else if(strcmp(name, "hour") == 0) {
        return PyInt_FromLong(dta.hour);
    } else if(strcmp(name, "min") == 0) {
        return PyInt_FromLong(dta.min);
    } else if(strcmp(name, "sec") == 0) {
        return PyInt_FromLong(dta.sec);
    } else if(strcmp(name, "usec") == 0) {
        return PyInt_FromLong(dta.usec);
    } else if(strcmp(name, "euro") == 0) {
        return PyString_FromString(dt_euro(&dta, b));
    } else if(strcmp(name, "euronotz") == 0) {
        return PyString_FromString(dt_euronotz(&dta, b));
    } else if(strcmp(name, "postgres") == 0) {
        return PyString_FromString(dt_postgres(&dta, b));
    } else if(strcmp(name, "sql") == 0) {
        return PyString_FromString(dt_sql(&dta, b));
    } else if(strcmp(name, "iso") == 0) {
        return PyString_FromString(dt_iso(&dta, b));
    } else if(strcmp(name, "unix") == 0) {
        return PyString_FromString(dt_unix(&dta, b));
    } else if(strcmp(name, "rfc") == 0) {
        return PyString_FromString(dt_rfc(&dta, b));
    } else if(strcmp(name, "leap") == 0) {
        return PyInt_FromLong(dta.leap);
    } else if(strcmp(name, "__members__") == 0) {
        return Py_BuildValue("[sssssssssssssssssssssss]",
            "sdn", "year", "month", "month_name",
            "month_short", "day", "yday",
            "wday", "wday_name", "wday_short", "week",
            "time", "hour", "min", "sec", "usec",
            "euro", "euronotz", "postgres", "sql", "iso", "unix",
            "rfc", "leap");
    } else {
        return Py_FindMethod(datetime_methods, (PyObject *) d, name);
    }
}

static PyTypeObject DateTimeType = {
    PyObject_HEAD_INIT(&PyType_Type)
    0,
    "DateTime" ,
    sizeof(DateTimeObject) ,
    0 ,
    (destructor) datetime_dealloc ,
    (printfunc) 0 ,
    (getattrfunc) datetime_getattr ,
    (setattrfunc) 0 ,
    (cmpfunc) datetime_cmp ,
    (reprfunc) datetime_repr ,
    (PyNumberMethods *) &datetime_number_methods ,
    (PySequenceMethods *) 0 ,
    (PyMappingMethods *) 0 ,
    (hashfunc) datetime_hash ,
    (ternaryfunc) 0 ,
    (reprfunc) datetime_str ,
    (getattrofunc) 0 ,
    (setattrofunc) 0 ,
    0 ,
    0 ,
    NULL
};


/***********************************************************
 *                             *
 *            Interval                 *
 *                             *
 ***********************************************************/

static PyObject * interval_Interval (PyObject *self, PyObject *args)
{
    IntervalObject      *no, so;
    IntervalAttr        ita;
    int         d, h = 0, m = 0, s = 0, us = 0, r;

    if(! PyArg_ParseTuple(args, "i|iiii", &d, &h, &m, &s, &us))
        return NULL;
    ita.dir = '+';
    ita.days = d;
    ita.hour = h;
    ita.min = m;
    ita.sec = s;
    ita.usec = us;
    if((r = ita2interval(&ita, &so)) < 0) {
        PyErr_SetString(DateRangeError, code2err(err_ita2interval, r) );
        return NULL;
    } 
    if(! (no = PyObject_NEW(IntervalObject, &IntervalType))) {
        PyErr_SetString(PyExc_MemoryError, "Out of memory.");
        return NULL;
    }
    no->dir = so.dir;
    no->days = so.days;
    no->time = so.time;
    no->usec = so.usec;
    no->argument = NULL;
    return (PyObject *) no;
}

static PyObject * interval_IntervalFromString (PyObject *self, PyObject *args)
{
    IntervalObject  *no, so;
    char            *str, b[128];
    IntervalAttr    ita;
    int             r;

    if(! PyArg_ParseTuple(args, "z", &str)) 
        return NULL;
    if((r=parseint(str, &ita)) < 0) {
        snprintf(b, 127, "Invalid Interval format '%s' (%d).", str, r);
        b[127] = '\0';
        PyErr_SetString(DateError, b);
        return NULL;
    }
    if((r = ita2interval(&ita, &so)) < 0) {
        PyErr_SetString(DateRangeError, code2err(err_ita2interval, r) );
        return NULL;
    } 
    if(! (no = PyObject_NEW(IntervalObject, &IntervalType))) {
        PyErr_SetString(PyExc_MemoryError, "Out of memory.");
        return NULL;
    }
    no->dir = so.dir;
    no->days = so.days;
    no->time = so.time;
    no->usec = so.usec;
    no->argument = NULL;
    return (PyObject *) no;
}

static PyObject * interval_IntervalFromSec (PyObject *self, PyObject *args)
{
    IntervalObject      *no;
    double          f;
    long int        l;

    if(! PyArg_ParseTuple(args, "d", &f)) 
        return NULL;
    if(! (no = PyObject_NEW(IntervalObject, &IntervalType))) {
        PyErr_SetString(PyExc_MemoryError, "Out of memory.");
        return NULL;
    }
    l = (long) f;
    if(l < 0) {
        no->dir = '-';
        l = -l;
    } else {
        no->dir = '+';
    }
    no->days = l / (24 *3600);
    no->time = l % (24 *3600);
    no->usec = f - l;
    no->argument = NULL;
    return (PyObject *) no;
}

static void interval_dealloc (IntervalObject *r)
{
    if(r) PyMem_DEL(r);
}

static PyObject * interval_repr (IntervalObject *r)
{
    IntervalAttr    ita;
    char        buf[64], b2[64];

    interval2ita(r, &ita);
    sprintf(buf, "<Interval: %c%d:%02d:%02d:%02d", ita.dir, ita.days,
            ita.hour, ita.min, ita.sec);
    if(ita.usec > 0) {
        sprintf(b2, ".%06d", ita.usec);
        strcat(buf, b2);
    }
    strcat(buf, ">");
    return PyString_FromString(buf);
}

static PyObject * interval_str (IntervalObject *r)
{
    IntervalAttr    ita;
    char        buf[64], b2[64];

    interval2ita(r, &ita);
    sprintf(buf, "%c%d:%02d:%02d:%02d", ita.dir, ita.days,
            ita.hour, ita.min, ita.sec);
    if(ita.usec > 0) {
        sprintf(b2, ".%06d", ita.usec);
        strcat(buf, b2);
    }
    return PyString_FromString(buf);
}

static int interval_coerce(PyObject **a1, PyObject **a2)
{
    IntervalObject  *no;
    double      f;
    long        l;
    if(PyInterval_Check(*a1)) {
        if(PyInterval_Check(*a2)) {
            Py_INCREF(*a1);
            Py_INCREF(*a2);
            return 0;
        } else if(PyNumber_Check(*a2)) {
            if(! (no = PyObject_NEW(IntervalObject, &IntervalType))) {
                PyErr_SetString(PyExc_MemoryError, "Out of memory.");
                return -1;
            }
            Py_INCREF(*a1);
            f = PyFloat_AsDouble(*a2);
            l = (long) f;
            if(l < 0) {
                no->dir = '-';
                l = -l;
            } else {
                no->dir = '+';
            }
            no->days = l / (24 *3600);
            no->time = l % (24 *3600);
            no->usec = f - l;
            no->argument = NULL;
            *a2 = (PyObject *) no;
            return 0;
        }
    }
    /* Cannot coerce */
    return 1;
}

static PyObject *interval_number_add(IntervalObject *d, IntervalObject *a)
{
    IntervalObject  *no;
    int     c = 0;

    if (PyInterval_Check(a)) {
        if(! (no = PyObject_NEW(IntervalObject, &IntervalType))) {
            PyErr_SetString(PyExc_MemoryError, "Out of memory.");
            return NULL;
        }
        if(d->dir == a->dir) {
            no->dir = d->dir;
            no->usec = d->usec + a->usec;
            if(no->usec > USEC_MAX) { c = 1; no->usec -= USEC_MAX + 1; }
            else { c = 0; } 
            no->time = d->time + a->time + c;
            if(no->time > TIME_MAX) { c = 1; no->time -= TIME_MAX + 1; }
            else { c = 0; } 
            no->days = d->days + a->days + c;
        } else {
            no->usec = d->usec - a->usec;
            if(no->usec < 0) { c = 1; no->usec += USEC_MAX + 1; }
            else { c = 0; } 
            no->time = d->time - a->time - c;
            if(no->time < 0) { c = 1; no->time += TIME_MAX + 1; }
            else { c = 0; } 
            no->days = d->days - a->days - c;
            if(no->days < 0) {
                /* abs(d) < abs(a) */
                no->days = - no->days;
                if(d->dir == '-')  no->dir = '+';
                else no->dir = '-';
            } else {
                /* abs(d) > abs(a) */
                if(d->dir == '-')  no->dir = '-';
                else no->dir = '+';
            }
        }
        no->argument = 0;
        return (PyObject *) no;
    } else {
        PyErr_SetString(PyExc_TypeError, "Addition is not supported for these arguments.");
        return (PyObject *) NULL;
    }
}

static PyObject *interval_number_sub(IntervalObject *d, IntervalObject *a)
{
    IntervalObject  *no;
    int     c = 0;

    if (PyInterval_Check(a)) {
        if(! (no = PyObject_NEW(IntervalObject, &IntervalType))) {
            PyErr_SetString(PyExc_MemoryError, "Out of memory.");
            return NULL;
        }
        if(d->dir != a->dir) {
            no->dir = d->dir;
            no->usec = d->usec + a->usec;
            if(no->usec > USEC_MAX) { c = 1; no->usec -= USEC_MAX + 1; }
            else { c = 0; } 
            no->time = d->time + a->time + c;
            if(no->time > TIME_MAX) { c = 1; no->time -= TIME_MAX + 1; }
            else { c = 0; } 
            no->days = d->days + a->days + c;
        } else {
            no->usec = d->usec - a->usec;
            if(no->usec < 0) { c = 1; no->usec += USEC_MAX + 1; }
            else { c = 0; } 
            no->time = d->time - a->time - c;
            if(no->time < 0) { c = 1; no->time += TIME_MAX + 1; }
            else { c = 0; } 
            no->days = d->days - a->days - c;
            if(no->days < 0) {
                /* abs(d) < abs(a) */
                no->days = - no->days;
                if(d->dir == '-')  no->dir = '+';
                else no->dir = '-';
            } else {
                /* abs(d) > abs(a) */
                if(d->dir == '-')  no->dir = '-';
                else no->dir = '+';
            }
        }
        no->argument = 0;
        return (PyObject *) no;
    } else {
        PyErr_SetString(PyExc_TypeError, "Subtraction is not supported for these arguments.");
        return (PyObject *) NULL;
    }
}

static PyObject *interval_number_negative(IntervalObject *i)
{
    IntervalObject  *no;

    if(! (no = PyObject_NEW(IntervalObject, &IntervalType))) {
        PyErr_SetString(PyExc_MemoryError, "Out of memory.");
        return NULL;
    }
    no->days = i->days;
    no->time = i->time;
    no->usec = i->usec;
    if(i->dir == '-') no->dir = '+';
    else no->dir = '-';
    no->argument = 0;
    return (PyObject *) no;
}

static int interval_number_nonzero (IntervalObject *d)
{
    return 1;
}

static PyObject *interval_number_as_int (IntervalObject *d)
{
    double  f;
    f = d->days * 86400 + d->time;
    return PyInt_FromLong((long) f );
}

static PyObject *interval_number_as_long (IntervalObject *d)
{
    double  f;
    f = d->days * 86400 + d->time;
    return PyLong_FromLong((long) f );
}

static PyObject *interval_number_as_float (IntervalObject *d)
{
    double  f;
    f = d->days * 86400 + d->time + ((double)d->usec/1000000);
    return PyFloat_FromDouble(f);
}

static PyNumberMethods interval_number_methods = {
    (binaryfunc) interval_number_add,          /*nb_add*/
    (binaryfunc) interval_number_sub,          /*nb_subtract*/
    (binaryfunc) number_not_implemented2,          /*nb_multiply*/
    (binaryfunc) number_not_implemented2,          /*nb_divide*/
    (binaryfunc) number_not_implemented2,          /*nb_remainder*/
    (binaryfunc) number_not_implemented2,          /*nb_divmod*/
    (ternaryfunc) number_not_implemented3,         /*nb_power*/
    (unaryfunc) interval_number_negative,          /*nb_negative*/
    (unaryfunc) number_not_implemented1,           /*nb_positive*/
     /* Dzis is NULL checked */
    (unaryfunc) 0,                     /*nb_absolute*/
    (inquiry) interval_number_nonzero,         /*nb_nonzero*/
    (unaryfunc) 0,                     /*nb_invert*/
    (binaryfunc) 0,                    /*nb_lshift*/
    (binaryfunc) 0,                    /*nb_rshift*/
    (binaryfunc) 0,                    /*nb_and*/
    (binaryfunc) 0,                    /*nb_xor*/
    (binaryfunc) 0,                    /*nb_or*/
    (coercion) interval_coerce,            /*nb_coerce*/
    (unaryfunc) interval_number_as_int,        /*nb_int*/
    (unaryfunc) interval_number_as_long,           /*nb_long*/
    (unaryfunc) interval_number_as_float,          /*nb_float*/
    (unaryfunc) 0,                     /*nb_oct*/
    (unaryfunc) 0,                     /*nb_hex*/
};




static PyObject *interval_reduce (IntervalObject *i, PyObject *args)
{
    PyObject    *tuple=NULL, *pickle=NULL, *construct=NULL;
    double      sec;

    if(! PyArg_ParseTuple(args, ""))
        return NULL;
    if(!(construct=PyDict_GetItemString(module_dict,"IntervalFromSec")))
        goto error;
    Py_INCREF(construct);
    if(!(tuple = PyTuple_New(2)))
        goto error;
    if(!(pickle = PyTuple_New(1)))
        goto error;

    sec = i->days * 86400 + i->time + ((double)i->usec/1000000);
    if(i->dir == '-') sec = -sec;

    PyTuple_SetItem(pickle, 0, PyFloat_FromDouble(sec));
    PyTuple_SetItem(tuple, 0, construct);
    PyTuple_SetItem(tuple, 1, pickle);
    return tuple;

error:
    Py_XDECREF(construct);
    Py_XDECREF(tuple);
    Py_XDECREF(pickle);
    return NULL;
}


static struct PyMethodDef interval_methods[] = {
        { "__reduce__", (PyCFunction) interval_reduce, 1 } ,
        { NULL, NULL }
};      

static PyObject *interval_getattr (IntervalObject *i, char *name)
{
    IntervalAttr    ita;
    char        b[128];
    double      f;

    interval2ita(i, &ita);
    if(strcmp(name, "days") == 0 || strcmp(name, "day") == 0) {
        return PyInt_FromLong(i->days);
    } else if(strcmp(name, "hour") == 0) {
        return PyInt_FromLong(ita.hour);
    } else if(strcmp(name, "min") == 0) {
        return PyInt_FromLong(ita.min);
    } else if(strcmp(name, "sec") == 0) {
        return PyInt_FromLong(ita.sec);
        return PyFloat_FromDouble(f);
    } else if(strcmp(name, "usec") == 0) {
        return PyInt_FromLong(ita.usec);

    } else if(strcmp(name, "asdays") == 0 || strcmp(name, "asday") == 0) {
        f = (double)i->time + ((double)i->days)*86400 + (double)i->usec/1000000.0;
        f /= 86400.0;
        if(i->dir == '-') f = -f;
        return PyFloat_FromDouble(f);
    } else if(strcmp(name, "ashour") == 0) {
        f = (double)i->time + ((double)i->days)*86400 + (double)i->usec/1000000.0;
        f /= 3600.0;
        if(i->dir == '-') f = -f;
        return PyFloat_FromDouble(f);
    } else if(strcmp(name, "asmin") == 0) {
        f = (double)i->time + ((double)i->days)*86400 + (double)i->usec/1000000.0;
        f /= 60.0;
        if(i->dir == '-') f = -f;
        return PyFloat_FromDouble(f);
    } else if(strcmp(name, "assec") == 0) {
        f = (double)i->time + ((double)i->days)*86400 + (double)i->usec/1000000.0;
        if(i->dir == '-') f = -f;
        return PyFloat_FromDouble(f);
    } else if(strcmp(name, "asusec") == 0) {
        f = ((double)i->time + (((double)i->days)*86400))*1000000.0 + (double)i->usec;
        if(i->dir == '-') f = -f;
        return PyFloat_FromDouble(f);
    } else if(strcmp(name, "fmt1") == 0) {
        return PyString_FromString(i_fmt1(&ita, b));
    } else if(strcmp(name, "fmt2") == 0) {
        return PyString_FromString(i_fmt2(&ita, b));
    } else if(strcmp(name, "fmt3") == 0) {
        return PyString_FromString(i_fmt3(&ita, b));
    } else if(strcmp(name, "sql") == 0) {
        return PyString_FromString(i_fmt3(&ita, b));
    } else {
        return Py_FindMethod(interval_methods, (PyObject *) i, name);
    }
}

static int interval_cmp(IntervalObject *o1, IntervalObject *o2)
{
    if(o1->dir != o2->dir) {
        if(o1->dir == '-') return -1;
        else return 1;
    }
    if(o1->dir == '+') {
        if(o1->days < o2->days) return -1;
        if(o1->days > o2->days) return 1;
        if(o1->time < o2->time) return -1;
        if(o1->time > o2->time) return 1;
        if(o1->usec < o2->usec) return -1;
        if(o1->usec > o2->usec) return 1;
    } else {
        if(o1->days < o2->days) return 1;
        if(o1->days > o2->days) return -1;
        if(o1->time < o2->time) return 1;
        if(o1->time > o2->time) return -1;
        if(o1->usec < o2->usec) return 1;
        if(o1->usec > o2->usec) return -1;
    }
    return 0;
}

static PyTypeObject IntervalType = {
    PyObject_HEAD_INIT(&PyType_Type)
    0,
    "Interval" ,
    sizeof(IntervalObject) ,
    0 ,
    (destructor) interval_dealloc ,
    (printfunc) 0 ,
    (getattrfunc) interval_getattr ,
    (setattrfunc) 0 ,
    (cmpfunc) interval_cmp ,
    (reprfunc) interval_repr ,
    (PyNumberMethods *) &interval_number_methods ,
    (PySequenceMethods *) 0 ,
    (PyMappingMethods *) 0 ,
    (hashfunc) 0 ,
    (ternaryfunc) 0 ,
    (reprfunc) interval_str ,
    (getattrofunc) 0 ,
    (setattrofunc) 0 ,
    0 ,
    0 ,
    NULL
};

/***********************************************************
 *                             *
 *            Functions                *
 *                             *
 ***********************************************************/

static PyObject * func_TZGet (PyObject *self, PyObject *args)
{
    PyObject *tup;

    if(! PyArg_ParseTuple(args, ""))
        return NULL;
    tup = PyTuple_New(2);
    PyTuple_SetItem(tup, 0, PyString_FromString(tzname[0]));
    PyTuple_SetItem(tup, 1, PyString_FromString(tzname[1]));
    return tup;
}

static PyObject * func_TZOffset (PyObject *self, PyObject *args)
{
    if(! PyArg_ParseTuple(args, ""))
        return NULL;
    return PyInt_FromLong(timezone / 60);
}

static PyObject * func_IsDST (PyObject *self, PyObject *args)
{
    if(! PyArg_ParseTuple(args, ""))
        return NULL;
    return PyInt_FromLong(daylight);
}

static PyObject * func_IsLeap (PyObject *self, PyObject *args)
{
    int y;
    if(! PyArg_ParseTuple(args, "i", &y)) 
        return NULL;
    if(IS_LEAP_YEAR(y)) return PyInt_FromLong(1L);
    else return PyInt_FromLong(0L);
}

static PyObject * func_Range (PyObject *self, PyObject *args)
{
    int y;
    if(! PyArg_ParseTuple(args, "i", &y)) 
        return NULL;
    if(IS_LEAP_YEAR(y)) return PyInt_FromLong(1L);
    else return PyInt_FromLong(0L);
}

/***********************************************************
 *                             *
 *            Initialization               *
 *                             *
 ***********************************************************/

static struct PyMethodDef module_methods[] = {
    { "Date", date_Date, 1 } ,
    { "DateFromString", date_DateFromString, 1 } ,
    { "DateFromSDN", date_DateFromSDN, 1 } ,
    { "DateFromDateTime", date_DateFromDateTime, 1 } ,
    { "Today", date_Today, 1 } ,
    { "Range", func_Range, 1 } ,
    { "DateTime", datetime_DateTime, 1 } ,
    { "DateTimeFromString", datetime_DateTimeFromString, 1 } ,
    { "DateTimeFromSDNTime", datetime_DateTimeFromSDNTime, 1 } ,
    { "DateTimeFromUNIX", datetime_DateTimeFromUNIX, 1 } ,
    { "DateTimeFromDate", datetime_DateTimeFromDate, 1 } ,
    { "Now", datetime_Now, 1 } ,
    { "Interval", interval_Interval, 1 } ,
    { "IntervalFromString", interval_IntervalFromString, 1 } ,
    { "IntervalFromSec", interval_IntervalFromSec, 1 } ,
    { "TZGet", func_TZGet, 1 } ,
    { "TZOffset", func_TZOffset, 1 } ,
    { "IsDST", func_IsDST, 1 } ,
    { "IsLeap", func_IsLeap, 1 } ,
    { NULL, NULL }
};

static PyObject *BuildZoneMap (void) 
{
    PyObject    *ha, *tu;
    int     i;

    ha = PyDict_New();
    if(! ha) return (PyObject *) NULL;
    for(i = 0; zone_map[i].tz != NULL; i++) {
        tu = PyTuple_New(2);
        if(! tu) return (PyObject *) NULL;
        PyTuple_SetItem(tu, 0, PyInt_FromLong(zone_map[i].tz_offset));
        PyTuple_SetItem(tu, 1, PyInt_FromLong(zone_map[i].dst_offset));
        PyDict_SetItemString(ha, zone_map[i].tz, tu);
    }
    return ha;
}


void initdate (void)
{

    char *unpicklers[] = { "DateFromSDN", "DateTimeFromSDNTime",
                           "IntervalFromSec" };

    PyObject  *copy_reg, *constructor, *unpickler;

    char err[128];

    int  i;

    tzset();
    G_TZ_Offset = - timezone;
    if(daylight == 1) G_TZ_Offset += 3600;

    module = Py_InitModule("date", module_methods);
    module_dict = PyModule_GetDict(module);

    if(PyDict_SetItemString(module_dict, "DateType", (PyObject*)&DateType) != 0)
        Py_FatalError("Cannot add to date dictionary.");
    if(PyDict_SetItemString(module_dict, "DateTimeType",
            (PyObject*) &DateTimeType) != 0)
        Py_FatalError("Cannot add to date dictionary.");
    if(PyDict_SetItemString(module_dict, "IntervalType",
            (PyObject*) &IntervalType) != 0)
        Py_FatalError("Cannot add to date dictionary.");

    DateRangeError = PyString_FromString("date.RangeError");
    if (PyDict_SetItemString(module_dict, "RangeError", DateRangeError) != 0)
        Py_FatalError("Cannot add RangeError to date dictionary.");
    DateError = PyString_FromString("date.Error");
    if (PyDict_SetItemString(module_dict, "Error", DateError) != 0)
        Py_FatalError("Cannot add Error to date dictionary.");

    ZoneMap = BuildZoneMap();
    if (! ZoneMap)
        Py_FatalError("Cannot build variable ZoneMap.");
    if (PyDict_SetItemString(module_dict, "ZoneMap", ZoneMap) != 0) 
        Py_FatalError("Cannot add ZoneMap to date dictionary.");

    if(! (copy_reg = PyImport_ImportModule("copy_reg")) )
        Py_FatalError("Cannot import copy_reg module.");
    if(! (constructor = PyObject_GetAttrString(copy_reg, "constructor")) )
        Py_FatalError("Module copy_reg hasn't attribute `constructor'.");

    for(i=0; i<3; i++) {
        if(! (unpickler = PyDict_GetItemString(module_dict, unpicklers[i])) ) {
            sprintf(err, "Cannot find factory method `%s'.", unpicklers[i]);
            Py_FatalError(err);
        }
        Py_INCREF(unpickler);
        PyObject_CallFunction(constructor, "O", unpickler);
    }
}
