#include <stdio.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <sys/types.h>
#include <Python.h>

#include "date.h"

#define SDN_OFFSET         32045
#define DAYS_PER_5_MONTHS  153
#define DAYS_PER_4_YEARS   1461
#define DAYS_PER_400_YEARS 146097

int IS_LEAP_YEAR(int y)
{
	if(y <= 0) y++;
	return (y) % 4 == 0 && ((y) % 100 != 0 || (y) % 400 == 0);
}


int sdn2date(int32_t sdn,  int32_t *y, int8_t *m, int8_t *d)
{
	int       century;
	int       year;
	int       month;
	int       day;
	long int  temp;
	int       dayOfYear;

	*y = *m = *d = 0;
	if (sdn < SDN_MIN) return -100;
	if(sdn > SDN_MAX) return -101;
	temp = (sdn + SDN_OFFSET) * 4 - 1;

	/* Calculate the century (year/100). */
	century = temp / DAYS_PER_400_YEARS;

	/* Calculate the year and day of year (1 <= dayOfYear <= 366). */
	temp = ((temp % DAYS_PER_400_YEARS) / 4) * 4 + 3;
	year = (century * 100) + (temp / DAYS_PER_4_YEARS);
	dayOfYear = (temp % DAYS_PER_4_YEARS) / 4 + 1;

	/* Calculate the month and day of month. */
	temp = dayOfYear * 5 - 3;
	month = temp / DAYS_PER_5_MONTHS;
	day = (temp % DAYS_PER_5_MONTHS) / 5 + 1;

	/* Convert to the normal beginning of the year. */
	if (month < 10) {
		month += 3;
	} else {
		year += 1;
		month -= 9;
	}
	
	/* Adjust to the B.C./A.D. type numbering. */
	year -= 4800;
	if (year <= 0) year--;
	*y = year;
	*m = month;
	*d = day;
	return 0;
}

int date2sdn(int32_t y, int8_t m, int8_t d, int32_t *sdn)
{
	int year;
	int month;

	/* check for invalid dates */
	if(y == 0 || y < -4714 || y > 1465101) return -100;
	if(m <= 0 || m > 12) return -101;
	if(d <= 0 || d > 31) return -102;
	if(d > month_days[IS_LEAP_YEAR(y)][m]) return -103;

	/* check for dates before SDN 1 (Nov 26, 4714 B.C.) */
	if (y == -4714) {
		if (m < 11) {
			return -104;
		}
		if (m == 11 && d < 26) {
			return -104;
		}
	}
	/* Make year always a positive number. */
	if (y < 0) {
		year = y + 4801;
	} else {
		year = y + 4800;
	}
	/* Adjust the start of the year. */
	if (m > 2) {
		month = m - 3;
	} else {
		month = m + 9;
		year--;
	}
	*sdn = ((year / 100) * DAYS_PER_400_YEARS) / 4
		+ ((year % 100) * DAYS_PER_4_YEARS) / 4
		+ (month * DAYS_PER_5_MONTHS + 2) / 5
		+ d - SDN_OFFSET;
	return 0;
}


int time2HMS(int32_t t, int *H, int *M, int *S)
{
    *H = t/3600;
    t %= 3600;
    *M = t / 60;
    *S = t % 60;
    return 0;
}


int findzone(const char *tzn)
{
	int	z;
       	for(z = 0; zone_map[z].tz; z++)
	       	if(strcasecmp(zone_map[z].tz, tzn) == 0) return z;
	return -1;
}

int datetime2tzindex (int y, int m, int d, int H, int M, int S)
{
       	struct tm       st, *pst;
       	time_t          t;
	char		b[16];

       	if(y < 1971 && y > 2036) {
	       	strncpy(b,  tzname[0], 15);
       	} else {
	       	memset(&st, 0, sizeof(st));
	       	st.tm_year = y - 1900;
	       	st.tm_mon = m - 1;
	       	st.tm_mday = d;
	       	st.tm_hour = H;
	       	st.tm_min = M;
	       	st.tm_sec = S;
	       	t = mktime(&st);
	       	pst = localtime(&t);
	       	strftime(b, 127, "%Z", pst);
       	}
	return findzone(b);
}


int zone2offset(DateTimeObject *f, const char *tz)
{
	int	y, z, off, H, M, S;
	int8_t	m, d;

	if(! tz) {
		sdn2date(f->sdn, &y, &m, &d);
                time2HMS(f->time, &H, &M, &S);
		z = datetime2tzindex(y, m, d, H, M, S);
		if(z < 0) return -1;
		off = zone_map[z].tz_offset + zone_map[z].dst_offset;
	} else {
		int l = strlen(tz);
		if((l == 3 || l == 5) && (tz[0] == '+' || tz[0] == '-')) {
			if(! isdigit(tz[1]) || ! isdigit(tz[2]))
                            return -1;
			off = (D(tz[1]) * 10 + D(tz[2])) * 60;
			if(l == 5) {
				if(! isdigit(tz[3]) || ! isdigit(tz[4]))
                                    return -1;
				off += D(tz[3]) * 10 + D(tz[4]);
			}
			if(tz[0] == '-') off = -off;
		} else {
			z = findzone(tz);
			if(z < 0) return -1;
			off = zone_map[z].tz_offset + zone_map[z].dst_offset;
		}
	}
	return off;
}


int gmt2local_fix (DateTimeObject *f, DateTimeObject *t, const char *tz)
{
	int	off;

	t->sdn = f->sdn;
	t->time = f->time;
	t->usec = f->usec;
	t->argument = f->argument;
	off = zone2offset(f, tz);
	if(off == -1) return -1;
	t->time += off * 60;
	if(t->time < 0) {
		t->time += 24 * 3600;
		t->sdn--;
	} else  if(t->time >= 24 * 3600) {
		t->time -= 24 * 3600;
		t->sdn++;
	}
	return 0;
}


int local2gmt_fix (DateTimeObject *f, DateTimeObject *t, const char *tz)
{
	int	off;

	t->sdn = f->sdn;
	t->time = f->time;
	t->usec = f->usec;
	t->argument = f->argument;
	off = zone2offset(f, tz);
	if(off == -1) return -1;
	t->time -= off * 60;
	if(t->time < 0) {
		t->time += 24 * 3600;
		t->sdn--;
	} else  if(t->time >= 24 * 3600) {
		t->time -= 24 * 3600;
		t->sdn++;
	}
	return 0;
}

int double2interval (double f, IntervalObject *i)
{
       	long  t;
       
	if(f < -(double)2147483647 || f > (double)2147483647) return -1;
       	if(f < 0) t = - (long) f;
       	else t = (long) f;
       	i->days = t / (24 *3600);
       	i->time = t % (24 * 3600);
       	i->usec = f - (long) f;
       	if(f < 0) i->dir = '-';
       	else i->dir = '+';
       	return 0;
}

