#include <Python.h>

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>


/* #include <postgres.h> */
#include <libpq-fe.h>
#include <libpq/libpq-fs.h>



typedef struct {
    PyObject_HEAD
    PGconn	*con;
    char	*log;
    PyObject	*backlog;
} pgint_object;


typedef struct {
    PyObject_HEAD
    PGresult *res;
} pgresult_object;


staticforward PyTypeObject PGIntType;
staticforward PyTypeObject PGResultType;
static PyObject *PGIntError;

#define PyPGInt_Check(op) ((op)->ob_type == &PGIntType)
#define PyPGResult_Check(op) ((op)->ob_type == &PGResultType)

static char *result_code (int code) 
{
    switch(code) {
    case PGRES_EMPTY_QUERY : return "empty_query";
    case PGRES_COMMAND_OK : return "command_ok";
    case PGRES_TUPLES_OK :  return "tuples_ok";
    case PGRES_COPY_OUT : return "copy_out";
    case PGRES_COPY_IN : return "copy_in";
    case PGRES_BAD_RESPONSE : return "bad_response";
    case PGRES_NONFATAL_ERROR : return "nonfatal_error";
    case PGRES_FATAL_ERROR : return "fatal_error";
    default : return "unknown_result";
    }
}

static void pgint_NoticeProcessor (void *arg, const char * message)
{
    pgint_object *self = arg;

    if(self->backlog != NULL) {
	PyObject_CallFunction(self->backlog, "s", message);
    } else {
	fprintf(stderr, "%s", message);
    }
}

static PyObject *pgint_new (PyObject *self, PyObject *args)
{
    pgint_object *no;
    PGconn *con;
    char *host, *db, *log, *port, *opt, *tty, *user, *pass; 
    
    host = db = log = port = opt = tty = user = pass = NULL;
    if(!PyArg_ParseTuple(args, "|zzzzzzzz", &host, &db, &user, &pass, &log, &port, &opt, &tty))
	return NULL;
    con = PQsetdbLogin(host, port, opt, tty, db, user, pass);
    if(PQstatus(con) == CONNECTION_OK) {
	if(! (no = PyObject_NEW(pgint_object, &PGIntType))) {
	    PyErr_SetString(PyExc_MemoryError, "Out of memory.");
	    return NULL;
	}
	no->con = con;
	if(log) no->log = strdup(log);
	else no->log = NULL;
	no->backlog = NULL;
	PQsetNoticeProcessor(no->con, pgint_NoticeProcessor, no);
	return (PyObject *) no;
    } else {
	PyErr_SetString(PGIntError, PQerrorMessage(con));
	PQfinish(con);
	return (PyObject *) NULL;
    }
}

static PyObject *pgint_status(pgint_object *self)
{
    if(PQstatus(self->con) == CONNECTION_OK) {
	return PyInt_FromLong(1L);
    } else {
	Py_INCREF(Py_None);
	return Py_None;
    }
}

static PyObject *pgint_errorMessage(pgint_object *self)
{
    char *s = PQerrorMessage(self->con);

    if(s) return PyString_FromString(s);
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *pgint_reset(pgint_object *self)
{
    PQreset(self->con);
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *pgint_query (pgint_object *self, PyObject *args)
{
    pgresult_object *no;
    PGresult	*res;
    char	*query;
    const char	*es;
    int         nt;
    FILE	*lf = NULL;
    time_t      _t;
    struct tm   *t;

    if(! PyArg_ParseTuple(args, "z", &query)) {
	return NULL;
    }
    res = PQexec(self->con, query);
    if(self->log) lf = fopen(self->log, "a");
    if(res == NULL) {
	if(lf) {
            _t = time(NULL);
            t = localtime(&_t);
	    fprintf(lf, "%d-%02d-%02d %02d:%02d:%02d QUERY(NULL:%s) \"%s\" \n",
                        t->tm_year + 1900, t->tm_mon, t->tm_mday,
                        t->tm_hour, t->tm_min, t->tm_sec,
                        PQerrorMessage(self->con), query);
	    fclose(lf);
	}
	PyErr_SetString(PGIntError, PQerrorMessage(self->con));
	return NULL;
    } else {
	if(lf)
	{
            _t = time(NULL);
            t = localtime(&_t);
	    es = PQresultErrorMessage(res);
	    if(es && es[0]) fprintf(lf, "QUERY(%s:%s) \"%s\"\n",
		      result_code(PQresultStatus(res)),
		      es, query);
	    else {
                if(PQresultStatus(res) == PGRES_TUPLES_OK) {
                    nt = PQntuples(res);
                    fprintf(lf, "%d-%02d-%02d %02d:%02d:%02d "
                                "QUERY(%d records) \"%s\"\n",
                                t->tm_year + 1900, t->tm_mon, t->tm_mday,
                                t->tm_hour, t->tm_min, t->tm_sec,
                                nt, query);
                } else {
                    fprintf(lf, "%d-%02d-%02d %02d:%02d:%02d "
                                "QUERY(%s) \"%s\"\n",
                                t->tm_year + 1900, t->tm_mon, t->tm_mday,
                                t->tm_hour, t->tm_min, t->tm_sec,
		                result_code(PQresultStatus(res)), query);
                }
            }
	    fclose(lf);
	}
    }
    if(! (no = PyObject_NEW(pgresult_object, &PGResultType))) {
	PyErr_SetString(PyExc_MemoryError, "Out of memory.");
	return NULL;
    }
    no->res = res;
    return (PyObject *) no;
}

static PyObject *pgint_db(pgint_object *self)
{
    char    *s = PQdb(self->con);

    if(!s) {
	Py_INCREF(Py_None);
	return Py_None;
    } else {
	return PyString_FromString(s);
    }
}

static PyObject *pgint_user(pgint_object *self)
{
    char    *s = PQuser(self->con);

    if(!s) {
	Py_INCREF(Py_None);
	return Py_None;
    } else {
	return PyString_FromString(s);
    }
}

static PyObject *pgint_host(pgint_object *self)
{
    char    *s = PQhost(self->con);

    if(!s) {
	Py_INCREF(Py_None);
	return Py_None;
    } else {
	return PyString_FromString(s);
    }
}

static PyObject *pgint_pass(pgint_object *self)
{
    char    *s = PQpass(self->con);

    if(!s) {
	Py_INCREF(Py_None);
	return Py_None;
    } else {
	return PyString_FromString(s);
    }
}

static PyObject *pgint_options(pgint_object *self)
{
    char    *s = PQoptions(self->con);

    if(!s) {
	Py_INCREF(Py_None);
	return Py_None;
    } else {
	return PyString_FromString(s);
    }
}

static PyObject *pgint_tty(pgint_object *self)
{
    char    *s = PQtty(self->con);

    if(!s) {
	Py_INCREF(Py_None);
	return Py_None;
    } else {
	return PyString_FromString(s);
    }
}

static PyObject *pgint_port(pgint_object *self)
{
    char    *s = PQport(self->con);

    if(!s) {
	Py_INCREF(Py_None);
	return Py_None;
    } else {
	return PyString_FromString(s);
    }
}

static PyObject *pgint_backendPID(pgint_object *self)
{
    return PyInt_FromLong((long) PQbackendPID(self->con));
}

static PyObject *pgint_setNoticeProcessor(pgint_object *self,	PyObject *args)
{
    PyObject	*obj; 
    
    if(! PyArg_ParseTuple(args, "O", &obj)) {
	return NULL;
    }
    if(self->backlog != NULL) { Py_XDECREF(self->backlog); }
    if(PyFunction_Check(obj)) {
	Py_INCREF(obj);
	self->backlog = obj;
    } else {
	self->backlog = NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
}
  
static PyObject *pgint_notifies(pgint_object *self)
{
    PGnotify	*ntf;
    PyObject	*tup;

    ntf = PQnotifies(self->con);
    if(ntf == NULL) {
	Py_INCREF(Py_None);
	return Py_None;
    }
    if(! (tup = PyTuple_New(2))) {
	PyErr_SetString(PyExc_MemoryError, "Out of memory.");
	free(ntf);
	return NULL;
    }
    PyTuple_SetItem(tup, 0, PyString_FromString(ntf->relname));
    PyTuple_SetItem(tup, 1, PyInt_FromLong((long) ntf->be_pid));
    free(ntf);
    return tup;
}

static struct PyMethodDef pgint_methods[] = {
    { "status", (PyCFunction) pgint_status, 1 } ,
    { "errormessage", (PyCFunction) pgint_errorMessage, 1 } ,
    { "reset", (PyCFunction) pgint_reset, 1 } ,
    { "query", (PyCFunction) pgint_query, 1 } ,
    { "db", (PyCFunction) pgint_db, 1 } ,
    { "user", (PyCFunction) pgint_user, 1 } ,
    { "passwd", (PyCFunction) pgint_pass, 1 } ,
    { "host", (PyCFunction) pgint_host, 1 } ,
    { "tty", (PyCFunction) pgint_tty, 1 } ,
    { "port", (PyCFunction) pgint_port, 1 } ,
    { "options", (PyCFunction) pgint_options, 1 } ,
    { "backendpid", (PyCFunction) pgint_backendPID, 1 } ,
    { "setnoticeprocessor", (PyCFunction) pgint_setNoticeProcessor, 1 } ,
    { "notifies", (PyCFunction) pgint_notifies, 1 } ,
    { NULL, NULL }
};

static PyObject *pgint_getattr (pgint_object *pi, char *name)
{
    return Py_FindMethod(pgint_methods, (PyObject *) pi, name);
}

static PyObject *pgint_dealloc (pgint_object *pi)
{
    if(pi->log) {
	free(pi->log);
    }
    PQfinish(pi->con);
    PyMem_DEL(pi);
    return (PyObject *) NULL;
}

static PyTypeObject PGIntType = {
    PyObject_HEAD_INIT(&PyType_Type)
    0,
    "pgint" ,
    sizeof(pgint_object) ,
    0 ,
    (destructor) pgint_dealloc ,
    (printfunc) 0 ,
    (getattrfunc)  pgint_getattr ,
    (setattrfunc) 0 ,
    (cmpfunc) 0 ,
    (reprfunc) 0 ,
    (PyNumberMethods *) 0 ,
    (PySequenceMethods *) 0 ,
    (PyMappingMethods *) 0 ,
};




/**********
 * RESULT *
 **********/

static PyObject *pgresult_status (pgresult_object *self)
{
    return PyString_FromString(result_code(PQresultStatus(self->res)));
}

static PyObject *pgresult_errorMessage (pgresult_object *self)
{
    const char *es = PQresultErrorMessage(self->res);

    if (es) return PyString_FromString(es);
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *pgresult_ntuples (pgresult_object *self)
{
    if(PQresultStatus(self->res) != PGRES_TUPLES_OK) {
	PyErr_SetString(PGIntError, "Result doesn't hold tuples.");
	return NULL;
    }
    return PyInt_FromLong(PQntuples(self->res));
}

static PyObject *pgresult_nfields (pgresult_object *self)
{
    if(PQresultStatus(self->res) != PGRES_TUPLES_OK) {
	PyErr_SetString(PGIntError, "Result doesn't hold tuples.");
	return NULL;
    }
    return PyInt_FromLong(PQnfields(self->res));
}

static PyObject *pgresult_fname (pgresult_object *self, PyObject *args)
{
    int	i;

    if(! PyArg_ParseTuple(args, "i", &i)) {
	return NULL;
    }
    if(PQresultStatus(self->res) != PGRES_TUPLES_OK) {
	PyErr_SetString(PGIntError, "Result doesn't hold tuples.");
	return NULL;
    }
    if((i >= PQnfields(self->res)) || (i < 0)) {
	PyErr_SetString(PGIntError, "Invalid index.");
	return NULL;
    }
    return PyString_FromString(PQfname(self->res, i));
}

static PyObject *pgresult_fnumber (pgresult_object *self, PyObject *args)
{
    char    *name;

    if(! PyArg_ParseTuple(args, "s", &name)) {
	return NULL;
    }
    if(PQresultStatus(self->res) != PGRES_TUPLES_OK) {
	PyErr_SetString(PGIntError, "Result doesn't hold tuples.");
	return NULL;
    }
    return PyInt_FromLong(PQfnumber(self->res, name));
}

static PyObject *pgresult_ftype (pgresult_object *self, PyObject *args)
{
    int	i;

    if(! PyArg_ParseTuple(args, "i", &i)) {
	return NULL;
    }
    if(PQresultStatus(self->res) != PGRES_TUPLES_OK) {
	PyErr_SetString(PGIntError, "Result doesn't hold tuples.");
	return NULL;
    }
    if((i >= PQnfields(self->res)) || (i < 0)) {
	PyErr_SetString(PGIntError, "Invalid index.");
	return NULL;
    }
    return PyInt_FromLong(PQftype(self->res, i));
}

static PyObject *pgresult_fsize (pgresult_object *self, PyObject *args)
{
    int	i;
    
    if(! PyArg_ParseTuple(args, "i", &i)) {
	return NULL;
    }
    if(PQresultStatus(self->res) != PGRES_TUPLES_OK) {
	PyErr_SetString(PGIntError, "Result doesn't hold tuples.");
	return NULL;
    }
    if((i >= PQnfields(self->res)) || (i < 0)) {
	PyErr_SetString(PGIntError, "Invalid index.");
	return NULL;
    }
    return PyInt_FromLong(PQfsize(self->res, i));
}

static PyObject *pgresult_fmod (pgresult_object *self, PyObject *args)
{
    int	i;

    if(! PyArg_ParseTuple(args, "i", &i)) {
	return NULL;
    }
    if(PQresultStatus(self->res) != PGRES_TUPLES_OK) {
	PyErr_SetString(PGIntError, "Result doesn't hold tuples.");
	return NULL;
    }
    if((i >= PQnfields(self->res)) || (i < 0)) {
	PyErr_SetString(PGIntError, "Invalid index.");
	return NULL;
    }
    return PyInt_FromLong(PQfmod(self->res, i));
}

static PyObject *pgresult_getvalue (pgresult_object *self, PyObject *args)
{
    int	r, f;
    char    *v;
    
    if(! PyArg_ParseTuple(args, "ii", &r, &f)) {
	return NULL;
    }
    if(PQresultStatus(self->res) != PGRES_TUPLES_OK) {
	PyErr_SetString(PGIntError, "Result doesn't hold tuples.");
	return NULL;
    }
    if((f >= PQnfields(self->res)) || (f < 0)) {
	PyErr_SetString(PGIntError, "Invalid field index.");
	return NULL;
    }
    if((r >= PQntuples(self->res)) || (r < 0)) {
	PyErr_SetString(PGIntError, "Inavlid row index.");
	return NULL;
    }
    v = PQgetvalue(self->res, r, f);
    if(v != NULL) {
	return PyString_FromString(v);
    } else {
	Py_INCREF(Py_None);
	return Py_None;
    }
}

static PyObject *pgresult_getlength (pgresult_object *self, PyObject *args)
{
    int	r, f;

    if(! PyArg_ParseTuple(args, "ii", &r, &f)) {
	return NULL;
    }
    if(PQresultStatus(self->res) != PGRES_TUPLES_OK) {
	PyErr_SetString(PGIntError, "Result doesn't hold tuples.");
	return NULL;
    }
    if((f >= PQnfields(self->res)) || (f < 0)) {
	PyErr_SetString(PGIntError, "Invalid field index.");
	return NULL;
    }
    if((r >= PQntuples(self->res)) || (r < 0)) {
	PyErr_SetString(PGIntError, "Inavlid row index.");
	return NULL;
    }
    return PyInt_FromLong(PQgetlength(self->res, r, f));
}

static PyObject *pgresult_getisnull (pgresult_object *self, PyObject *args)
{
    int	r, f;

    if(! PyArg_ParseTuple(args, "ii", &r, &f)) {
	return NULL;
    }
    if(PQresultStatus(self->res) != PGRES_TUPLES_OK) {
	PyErr_SetString(PGIntError, "Result doesn't hold tuples.");
	return NULL;
    }
    if((f >= PQnfields(self->res)) || (f < 0)) {
	PyErr_SetString(PGIntError, "Invalid field index.");
	return NULL;
    }
    if((r >= PQntuples(self->res)) || (r < 0)) {
	PyErr_SetString(PGIntError, "Inavlid row index.");
	return NULL;
    }
    if(PQgetisnull(self->res, r, f) == 1)
	return PyInt_FromLong(1L);
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *pgresult_cmdstatus (pgresult_object *self )
{
    return PyString_FromString(PQcmdStatus(self->res));
}

static PyObject *pgresult_cmdtuples (pgresult_object *self )
{
    const char *t = PQcmdTuples(self->res);
    if(strlen(t) > 0)
	return PyInt_FromLong(atol(t));
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *pgresult_oidstatus (pgresult_object *self )
{
    const char *t = PQoidStatus(self->res);
    if(strlen(t) > 0)
	return PyInt_FromLong(atol(t));
    Py_INCREF(Py_None);
    return Py_None;
}

static struct PyMethodDef pgresult_methods[] = {
    { "status", (PyCFunction) pgresult_status, 1 } ,
    { "errormessage", (PyCFunction) pgresult_errorMessage, 1 } ,
    { "ntuples", (PyCFunction) pgresult_ntuples, 1 } ,
    { "nfields", (PyCFunction) pgresult_nfields, 1 } ,
    { "fname", (PyCFunction) pgresult_fname, 1 } ,
    { "fnumber", (PyCFunction) pgresult_fnumber, 1 } ,
    { "ftype", (PyCFunction) pgresult_ftype, 1 } ,
    { "fsize", (PyCFunction) pgresult_fsize, 1 } ,
    { "fmod", (PyCFunction) pgresult_fmod, 1 } ,
    { "getvalue", (PyCFunction) pgresult_getvalue, 1 } ,
    { "getlength", (PyCFunction) pgresult_getlength, 1 } ,
    { "getisnull", (PyCFunction) pgresult_getisnull, 1 } ,
    { "cmdstatus", (PyCFunction) pgresult_cmdstatus, 1 } ,
    { "cmdtuples", (PyCFunction) pgresult_cmdtuples, 1 } ,
    { "oidstatus", (PyCFunction) pgresult_oidstatus, 1 } ,
    { NULL, NULL }
};

static PyObject *pgresult_getattr (pgresult_object *pr, char *name)
{
    return Py_FindMethod(pgresult_methods, (PyObject *) pr, name);
}

static PyObject *pgresult_dealloc (pgresult_object *pr)
{
    PQclear(pr->res);
    PyMem_DEL(pr);
    return (PyObject *) NULL;
}

static PyTypeObject PGResultType = {
    PyObject_HEAD_INIT(&PyType_Type)
    0,
    "pgresult" ,
    sizeof(pgresult_object) ,
    0 ,
    (destructor) pgresult_dealloc ,
    (printfunc) 0 ,
    (getattrfunc)  pgresult_getattr ,
    (setattrfunc) 0 ,
    (cmpfunc) 0 ,
    (reprfunc) 0 ,
    (PyNumberMethods *) 0 ,
    (PySequenceMethods *) 0 ,
    (PyMappingMethods *) 0 ,
};

/*****************
 * LARGE OBJECTS *
 *****************/

static PyObject *pglo_create (PyObject *self, PyObject *args)
{
    pgint_object    *con;
    int	    mode, result;
    if(! PyArg_ParseTuple(args, "O!i", &PGIntType, &con, &mode)) {
	return NULL;
    }
    result = lo_creat(con->con, mode);
    if(con->log) {
	FILE *lf = fopen(con->log, "a");
	fprintf(lf, "LO_CREATE(%d)\n", result);
	fclose(lf);
    }
    return PyInt_FromLong(result);
}

static PyObject *pglo_open (PyObject *self, PyObject *args)
{
    pgint_object    *con;
    int	    mode, oid, result;
    if(! PyArg_ParseTuple(args, "O!ii", &PGIntType, &con, &oid, &mode)) {
	return NULL;
    }
    result = lo_open(con->con, oid, mode);
    if(con->log) {
	FILE *lf = fopen(con->log, "a");
	fprintf(lf, "LO_OPEN(%d) OID: %d MODE: %d\n", result, oid, mode);
	fclose(lf);
    }
    return PyInt_FromLong(result);
}

static PyObject *pglo_close (PyObject *self, PyObject *args)
{
    pgint_object    *con;
    int	    fd, result;
    if(! PyArg_ParseTuple(args, "O!i", &PGIntType, &con, &fd)) {
	return NULL;
    }
    result = lo_close(con->con, fd);
    if(con->log) {
	FILE *lf = fopen(con->log, "a");
	fprintf(lf, "LO_CLOSE(%d) FD: %d\n", result, fd);
	fclose(lf);
    }
    return PyInt_FromLong(result);
}

static PyObject *pglo_read (PyObject *self, PyObject *args)
{
    pgint_object    *con;
    int	    fd, len, r;
    PyObject	*buf;

    if(! PyArg_ParseTuple(args, "O!ii", &PGIntType, &con, &fd, &len)) {
	return NULL;
    }
    buf = PyString_FromStringAndSize((char *) NULL, len);
    r = lo_read(con->con, fd, PyString_AS_STRING((PyStringObject *) buf), len);
    if(con->log) {
	FILE *lf = fopen(con->log, "a");
	fprintf(lf, "LO_READ(%d) FD: %d\n", r, fd);
	fclose(lf);
    }
    _PyString_Resize(&buf, r);
    return buf;
}

static PyObject *pglo_write (PyObject *self, PyObject *args)
{
    pgint_object    *con;
    int	    fd, len, result;
    char	*buf;

    if(! PyArg_ParseTuple(args, "O!is#", &PGIntType, &con, &fd, &buf, &len)) {
	return NULL;
    }
    result = lo_write(con->con, fd, buf, len);
    if(con->log) {
	FILE *lf = fopen(con->log, "a");
	fprintf(lf, "LO_WRITE(%d) FD: %d LEN: %d\n", result, fd, len);
	fclose(lf);
    }
    return PyInt_FromLong(result);
}

static PyObject *pglo_unlink (PyObject *self, PyObject *args)
{
    pgint_object    *con;
    int	    oid, result;

    if(! PyArg_ParseTuple(args, "O!i", &PGIntType, &con, &oid)) {
	return NULL;
    }
    result = lo_unlink(con->con, oid);
    if(con->log) {
	FILE *lf = fopen(con->log, "a");
	fprintf(lf, "LO_UNLINK(%d) OID: %d\n", result, oid);
	fclose(lf);
    }
    return PyInt_FromLong(result);
}

static PyObject *pglo_lseek (PyObject *self, PyObject *args)
{
    pgint_object    *con;
    int	    fd, whence, off, result;

    if(! PyArg_ParseTuple(args, "O!iii", &PGIntType, &con, &fd, &off, &whence)) {
	return NULL;
    }
    result = lo_lseek(con->con, fd, off, whence);
    if(con->log) {
	FILE *lf = fopen(con->log, "a");
	fprintf(lf, "LO_LSEEK(%d) FD: %d OFF: %d, WHEN: %d\n",
	    result, fd, off, whence);
	fclose(lf);
    }
    return PyInt_FromLong(result);
}

static PyObject *pglo_tell (PyObject *self, PyObject *args)
{
    pgint_object    *con;
    int	    fd, result;

    if(! PyArg_ParseTuple(args, "O!i", &PGIntType, &con, &fd)) {
	return NULL;
    }
    result = lo_tell(con->con, fd);
    if(con->log) {
	FILE *lf = fopen(con->log, "a");
	fprintf(lf, "LO_TELL(%d) FD: %d\n", result, fd);
	fclose(lf);
    }
    return PyInt_FromLong(result);
}

static PyObject *pglo_import (PyObject *self, PyObject *args)
{
    pgint_object    *con;
    char	*path;
    int	    result;

    if(! PyArg_ParseTuple(args, "O!s", &PGIntType, &con, &path)) {
	return NULL;
    }
    result = lo_import(con->con, path);
    if(con->log) {
	FILE *lf = fopen(con->log, "a");
	fprintf(lf, "LO_IMPORT(%d) PATH: %s\n", result, path);
	fclose(lf);
    }
    return PyInt_FromLong(result);
}

static PyObject *pglo_export (PyObject *self, PyObject *args)
{
    pgint_object    *con;
    int	    oid, result;
    char	*path;

    if(! PyArg_ParseTuple(args, "O!is", &PGIntType, &con, &oid, &path)) {
	return NULL;
    }
    result = lo_export(con->con, oid, path);
    if(con->log) {
	FILE *lf = fopen(con->log, "a");
	fprintf(lf, "LO_EXPORT(%d) OID: %d PATH: %s\n", result, oid, path);
	fclose(lf);
    }
    return PyInt_FromLong(result);
}


static struct PyMethodDef pgint_functions[] = {
    { "new", (PyCFunction) pgint_new, 1 } ,
    { "lo_create", (PyCFunction) pglo_create, 1 } ,
    { "lo_open", (PyCFunction) pglo_open, 1 } ,
    { "lo_close", (PyCFunction) pglo_close, 1 } ,
    { "lo_read", (PyCFunction) pglo_read, 1 } ,
    { "lo_write", (PyCFunction) pglo_write, 1 } ,
    { "lo_unlink", (PyCFunction) pglo_unlink, 1 } ,
    { "lo_lseek", (PyCFunction) pglo_lseek, 1 } ,
    { "lo_tell", (PyCFunction) pglo_tell, 1 } ,
    { "lo_import", (PyCFunction) pglo_import, 1 } ,
    { "lo_export", (PyCFunction) pglo_export, 1 } ,
    { NULL, NULL }
};

void initpgint (void)
{
    PyObject *module, *dict;

    module = Py_InitModule("pgint", pgint_functions);
    dict = PyModule_GetDict(module); 
    PyDict_SetItemString(dict, "SEEK_SET", PyInt_FromLong(SEEK_SET));
    PyDict_SetItemString(dict, "SEEK_CUR", PyInt_FromLong(SEEK_CUR));
    PyDict_SetItemString(dict, "SEEK_END", PyInt_FromLong(SEEK_END));
    PyDict_SetItemString(dict, "INV_READ", PyInt_FromLong(INV_READ));
    PyDict_SetItemString(dict, "INV_WRITE", PyInt_FromLong(INV_WRITE));
    PyDict_SetItemString(dict, "PGIntType", (PyObject*) &PGIntType);
    PyDict_SetItemString(dict, "PGResultType", (PyObject*) &PGResultType);
    PGIntError = PyString_FromString("pgint.error");
    PyDict_SetItemString(dict, "error", PGIntError);
}
