from os.path import join, isdir, isfile, exists, abspath
from logger import getLogger
import stat, mimetypes, imp, os, types


__all__ = [
    'Traversable', 'HTTPHandler', 'TraverseFS', 'TraverseDirectory',
    'TraverseApplication', 'TraverseModule', 'TraverseBuiltin',
    'TraverseFile',
]


log = getLogger('wm3.web')


class Traversable(object):

    def traversePath(self, req, path):
        raise NotImplementedError


class HTTPHandler(object):

    def handleHTTP(self, req):
        raise NotImplementedError






class TraverseFS(Traversable, HTTPHandler):

    def __init__(self, absPath):
        self.absPath = abspath(absPath)



class TraverseDirectory(TraverseFS):

    def __init__(self, absPath):
        super(TraverseDirectory, self).__init__(absPath)
        self.pathCache = {}

    def traversePath(self, req, path):
        if not self.absPath.startswith(req.conf.app_dir):
            log.warn(
                'path is not within configured app_dir: %r %r',
                self.absPath, req.conf.app_dir
            )
            return None

        if not path:
            absPath = join(self.absPath, 'index.py')
            req.traversal_path.append('index.py')
            req.url = req.url + '/'
        else:
            name = path[0]
            absPath = join(self.absPath, name)
            if exists(absPath):
                path.pop(0)
            else:
                if name == 'index.py':
                    log.warn(
                        'file index.py not found in directory: %r', self.absPath
                    )
                    return None
                else:
                    absPath = join(self.absPath, 'index.py')
                    req.traversal_path.append('index.py')
                    req.url = req.url + '/'

        obj = self.pathCache.get(absPath)
        if obj is not None:
            log.debug('cached instance found: %r %r', absPath, obj)
            return obj.traversePath(req, path)

        if isfile(absPath):
            # XXX Umoznit konfiguraci handleru podle pripony
            if absPath.endswith('.py'):
                obj = TraverseModule(absPath)
            elif absPath.endswith('.tal') or absPath.endswith('.zpt'):
                raise NotImplementedError
            else:
                obj = TraverseFile(absPath)
        elif isdir(absPath):
            obj = TraverseDirectory(absPath)
        else:
            log.warn('file %r not found', self.absPath)
            return None
        log.debug('new instance created: %r %r', absPath, obj)
        self.pathCache[absPath] = obj
        return obj.traversePath(req, path)






class TraverseApplication(TraverseDirectory):

    popupInstance = None

    def traversePath(self, req, path):
        if path and path[0] == '__popup__':

            if self.popupInstance is None:
                import wm3.popup
                absPath = wm3.popup.__file__
                self.popupInstance = TraverseBuiltin(
                    absPath, moduleInstance=wm3.popup, moduleName='wm3.popup',
                )
                log.debug('newly created wm3.popup: %r', self.popupInstance)
            else:
                log.debug('cached wm3.popup: %r', self.popupInstance)

            path.pop(0)
            return self.popupInstance.traversePath(req, path)

        return super(TraverseApplication, self).traversePath(req,path)






module_counter = 0

class TraverseModule(TraverseFS):

    def __init__(self, absPath, moduleInstance=None, moduleName=None):
        global module_counter
        super(TraverseModule, self).__init__(absPath)
        absPath = self.absPath

        if moduleName is None:
            moduleName = 'wm3module%03d' % module_counter
            module_counter = module_counter + 1
        self.moduleName = moduleName

        if moduleInstance is None:
            moduleInstance = imp.load_source(moduleName, absPath, open(absPath))
        self.moduleInstance = moduleInstance

        self.fileStat = os.stat(self.absPath)

    def traversePath(self, req, path):
        if not self.absPath.startswith(req.conf.app_dir):
            log.warn(
                'path is not within configured app_dir: %r %r',
                self.absPath, req.conf.app_dir
            )
            return None
        if path:
            req.args = list(path)
        return self

    def handleHTTP(self, req):
        methodName = 'P_' + req.kw.get('me', 'default')
        req.file = self.absPath
        req.object_name = methodName
        req.object = obj = getattr(self.moduleInstance, methodName)

        if type(obj) is types.FunctionType:
            lineNo = obj.func_code.co_firstlineno
        elif isinstance(obj, object):
            obj = obj.__call__
            lineNo = obj.im_func.func_code.co_firstlineno
        elif isinstance(obj, HTTPHandler):
            obj = obj.handleHTTP
            lineNo = obj.im_func.func_code.co_firstlineno
        else:
            raise NotImplementedError('unsupported type', obj, type(obj))

        req.object_id = '%d:%d:%d' % (
            self.fileStat[stat.ST_DEV], self.fileStat[stat.ST_INO], lineNo,
        )
        req.object(req)


class TraverseBuiltin(TraverseModule):

    def traversePath(self, req, path):
        if path:
            req.args = list(path)
        return self



class TraverseFile(TraverseFS):

    mime_initialized = None

    def traversePath(self, req, path):
        if not self.absPath.startswith(req.conf.app_dir):
            log.warn(
                'path is not within configured app_dir: %r %r',
                self.absPath, req.conf.app_dir
            )
            return None
        if path:
            log.warn(
                'traversing file does not make sense: %r %r', self.absPath, path
            )
            return None
        else:
            return self

    def handleHTTP(self, req):
        absPath = self.absPath

        # Kontrola na nepovolene pripony
        for e in req.conf.denied_extensions:
            if absPath.endswith(e):
                req.status = None
                return None

        # Nalezeni mime-type a encoding
        req.header('Content-Length', os.stat(absPath)[stat.ST_SIZE])

        klass = self.__class__
        if not klass.mime_initialized:
            mimetypes.init(req.conf.mime_types_config)
            klass.mime_initialized

        mt, ex = mimetypes.guess_type(absPath)
        if mt: req.header('Content-Type', mt)
        if ex: req.header('Content-Encoding', ex)

        # Doplneni hlavicek z konfigurace
        for e,v in req.conf.extension_handlers.items():
            if absPath.endswith(e):
                req.headers = req.headers + v.headers

        # Odesli data
        f = open(absPath)
        BL = 65536
        while 1:
            b = f.read(BL)
            req.write(b)
            if len(b) < BL:
                break
        f.close()
