from unittest import *
from wm4.core import *
from wm4.binding.api import *
from wm4.exceptions import *
from itertools import count

next = count().next

class CM(ClassMetadata):
    pass

class AM(AttributeMetadata):
    pass

class C(object):
    metadata(
        CM(lambda s,f: (s,f,next())),
        a = AM(lambda s,f: (s,f,next())),
        b = AM(lambda s,f: (s,f,next())),
    )

class D(C):
    pass

class E(D):
    pass


class TestClassMetadata(TestCase):

    def test01_Metaclass(self):
        class M1(ClassMetadata): pass
        class M2(ClassMetadata): pass
        class M3(M2): pass
        assert M1._registry is not M2._registry is not M3._registry

    def test10_ValueCaching(self):
        s,f,v1 = CM.of(C)
        s,f,v2 = CM.of(C)
        self.assertEqual(s, C)
        self.assertEqual(f, C)
        self.assertEqual(v1, v2)

    def test12_Inheritance(self):
        cs,cf,cv1 = CM.of(C)
        cs,cf,cv2 = CM.of(C)
        ds,df,dv1 = CM.of(D)
        ds,df,dv2 = CM.of(D)

        self.assertEqual(cs, C)
        self.assertEqual(cf, C)
        self.assertEqual(cv1, cv2)

        self.assertEqual(ds, D)
        self.assertEqual(df, C)
        self.assertEqual(dv1, dv2)

        self.assertNotEqual(cv1, dv1)

    def test15_Wildcard(self):

        class WCM(ClassMetadata):
            pass

        WCM(lambda s,f: (s,f,next())).register('*')

        cs,cf,cv1 = WCM.of(C)
        cs,cf,cv2 = WCM.of(C)
        ds,df,dv1 = WCM.of(D)
        ds,df,dv2 = WCM.of(D)
        at,af,av1 = WCM.of('*')
        at,af,av2 = WCM.of('*')

        self.assertEqual(at, '*')
        self.assertEqual(af, '*')
        self.assertEqual(av1, av2)

        self.assertEqual(cs, C)
        self.assertEqual(cf, '*')
        self.assertEqual(cv1, cv2)

        self.assertEqual(ds, D)
        self.assertEqual(df, '*')
        self.assertEqual(dv1, dv2)

        self.assertNotEqual(av1, cv1)
        self.assertNotEqual(av1, dv1)
        self.assertNotEqual(cv1, dv1)

    def test30_SettingConflict(self):
        self.assertRaises(SettingConflict, declareClassMetadata, C, CM(42))
        self.assertRaises(AlreadyRead, declareClassMetadata, D, CM(42))

    def test32_RedefinitionInSubclass(self):
        class E(D):
            metadata(CM(42))
        self.assertEqual(CM.of(E), 42)

    def test33_RedefinitionOfMiddleClass(self):

        class FCM(ClassMetadata):
            pass

        FCM('foo').register(C)

        self.assertEqual(FCM.of(C), 'foo')
        self.assertEqual(FCM.of(E), 'foo')
        self.assertRaises(AlreadyRead, declareClassMetadata, D, FCM(42))

    def test40_NotFound(self):
        self.assertRaises(exceptions.MetadataNotFound, CM.of, object)
        self.assertEqual(CM.of(object, default='foo'), 'foo')




class TestAttributeMetadata(TestCase):

    def test10_ValueCaching(self):
        s,f,v1 = AM.of(C, 'a')
        s,f,v2 = AM.of(C, 'a')
        self.assertEqual(s, (C,'a'))
        self.assertEqual(f, (C,'a'))
        self.assertEqual(v1, v2)

    def test12_Inheritance(self):
        cs,cf,cv1 = AM.of(C,'a')
        cs,cf,cv2 = AM.of(C,'a')
        ds,df,dv1 = AM.of(D,'a')
        ds,df,dv2 = AM.of(D,'a')

        self.assertEqual(cs, (C,'a'))
        self.assertEqual(cf, (C,'a'))
        self.assertEqual(cv1, cv2)

        self.assertEqual(ds, (D,'a'))
        self.assertEqual(df, (C,'a'))
        self.assertEqual(dv1, dv2)

        self.assertNotEqual(cv1, dv1)

    def test13_AttributeIsolation(self):
        cs,cf,cv = AM.of(C,'a')
        bs,bf,bv = AM.of(C,'b')

        self.assertEqual(cs, (C,'a'))
        self.assertEqual(cf, (C,'a'))
        self.assertEqual(bs, (C,'b'))
        self.assertEqual(bf, (C,'b'))

        self.assertNotEqual(cv, bv)

    def test15_ClassWildcard(self):
        class WAM(AttributeMetadata):
            pass
        next = count().next
        WAM(lambda s,f: (s,f,next())).register('*','a')
        WAM(lambda s,f: (s,f,next()+40)).register(E,'a')

        self.assertEqual(WAM.of(C,'a'), ((C,'a'),('*','a'),0))
        self.assertEqual(WAM.of(C,'a'), ((C,'a'),('*','a'),0))
        self.assertEqual(WAM.of(D,'a'), ((D,'a'),('*','a'),1))
        self.assertEqual(WAM.of('*','a'), (('*','a'),('*','a'),2))
        self.assertEqual(WAM.of(E,'a'), ((E,'a'),(E,'a'),43))
        self.assertRaises(MetadataNotFound, WAM.of, '*', '*')

    def test17_AttributeWildcard(self):
        class WAM(AttributeMetadata):
            pass
        next = count().next
        WAM(lambda s,f: (s,f,next())).register(object,'*')
        WAM(lambda s,f: (s,f,next()+40)).register(E,'*')

        self.assertEqual(WAM.of(C,'a'), ((C,'a'),(object,'*'),0))
        self.assertEqual(WAM.of(C,'a'), ((C,'a'),(object,'*'),0))
        self.assertEqual(WAM.of(C,'b'), ((C,'b'),(object,'*'),1))
        self.assertEqual(WAM.of(D,'a'), ((D,'a'),(object,'*'),2))
        self.assertEqual(WAM.of(D,'a'), ((D,'a'),(object,'*'),2))
        self.assertEqual(WAM.of(E,'a'), ((E,'a'),(E,'*'),43))
        self.assertEqual(WAM.of(E,'a'), ((E,'a'),(E,'*'),43))
        self.assertEqual(WAM.of(object,'*'), ((object,'*'),(object,'*'),4))
        self.assertEqual(WAM.of(E,'*'), ((E,'*'),(E,'*'),45))
        self.assertEqual(WAM.of(C,'*'), ((C,'*'),(object,'*'),6))
        self.assertRaises(MetadataNotFound, WAM.of, '*', '*')

    def test19_FullWildcard(self):
        class WAM(AttributeMetadata):
            pass
        next = count().next
        WAM(lambda s,f: (s,f,next())).register('*','*')
        WAM(lambda s,f: (s,f,next()+40)).register(E,'*')

        self.assertEqual(WAM.of(C,'a'), ((C,'a'),('*','*'),0))
        self.assertEqual(WAM.of(C,'a'), ((C,'a'),('*','*'),0))
        self.assertEqual(WAM.of(C,'b'), ((C,'b'),('*','*'),1))
        self.assertEqual(WAM.of(D,'a'), ((D,'a'),('*','*'),2))
        self.assertEqual(WAM.of(E,'a'), ((E,'a'),(E,'*'),43))
        self.assertEqual(WAM.of('*','a'), (('*','a'),('*','*'),4))
        self.assertEqual(WAM.of('*','*'), (('*','*'),('*','*'),5))


    def test30_SettingConflict(self):
        self.assertRaises(SettingConflict, declareAttribute, C, 'a', AM(42))
        self.assertRaises(AlreadyRead, declareAttribute, D, 'a', AM(42))

    def test32_RedefinitionInSubclass(self):
        class E(D):
            metadata(a=AM(42))
        self.assertEqual(AM.of(E,'a'), 42)

    def test33_RedefinitionOfMiddleClass(self):

        class FAM(AttributeMetadata):
            pass

        FAM('foo').register(C,'a')

        self.assertEqual(FAM.of(C,'a'), 'foo')
        self.assertEqual(FAM.of(E,'a'), 'foo')
        self.assertRaises(AlreadyRead, declareAttribute, D, 'a', FAM(42))

    def test40_NotFound(self):
        self.assertRaises(exceptions.MetadataNotFound, AM.of, object, 'a')
        self.assertEqual(AM.of(object,'a',default='foo'), 'foo')





def suite():

    return TestSuite((
        makeSuite(TestClassMetadata),
        makeSuite(TestAttributeMetadata),
    ))

if __name__ == '__main__':
    main(defaultTest='suite')
