#include <Python.h>
#include "structmember.h"

/* Thread local context */

static PyObject *ctx_key, *frozen, *not_given, *not_found;

static PyObject* _read_ctx(PyObject *self, PyObject *args)
{
    PyObject *tdict = PyThreadState_GetDict();
    PyObject *ctx = PyDict_GetItem(tdict, ctx_key);
        
    if (ctx == NULL) {
        ctx = PyDict_New();
        if (ctx == NULL)
            return NULL;
        if (PyDict_SetItem(tdict, ctx_key, ctx) != 0)
            return NULL;
    } else {
        Py_INCREF(ctx);
    }
    return ctx;
}

static PyObject* _write_ctx(PyObject *self, PyObject *args)
{
    PyObject *tdict = PyThreadState_GetDict();
    PyObject *ctx = PyDict_GetItem(tdict, ctx_key);

    if (ctx == NULL) {
        ctx = PyDict_New();
        if (ctx == NULL)
            return NULL;
        if (PyDict_SetItem(tdict, ctx_key, ctx) != 0)
            return NULL;
    } else {
        if (PyDict_GetItem(ctx, frozen) == NULL) {
            Py_INCREF(ctx);
        } else {
            ctx = PyDict_Copy(ctx);
            if (ctx == NULL)
                return NULL;
            if (PyDict_SetItem(tdict, ctx_key, ctx) != 0)
                return NULL;
            if (PyDict_DelItem(ctx, frozen) != 0)
                return NULL;
        }
    }
    return ctx;
}

static PyObject* snapshot(PyObject *self, PyObject *args)
{
    PyObject *ctx = _read_ctx(NULL, NULL);
    if (PyDict_SetItem(ctx, frozen, Py_True) != 0)
        return NULL;
    return ctx;
}

static PyObject* swap(PyObject *self, PyObject *args)
{
    PyObject *tdict = PyThreadState_GetDict();
    PyObject *old = _read_ctx(NULL, NULL);
    PyObject *new;

    if (!PyArg_ParseTuple(args, "O", &new))
        return NULL;
    if (PyDict_SetItem(tdict, ctx_key, new) != 0)
        return NULL;
    return old;
}

static PyObject* new_ctx(PyObject *self, PyObject *args)
{
    PyObject *ctx = PyDict_New();
    if (PyDict_SetItem(ctx, frozen, Py_True) != 0)
        return NULL;
    return ctx;
}



/* Setting */

typedef struct {
    PyObject_HEAD
    PyObject    *name;
    PyObject    *doc;
    PyObject    *dflt;
} Setting;


static PyObject *
Setting_new(PyTypeObject *type, PyObject *args, PyObject *kw)
{
    Setting *self = NULL;
    PyObject *name = NULL;
    PyObject *doc = NULL;
    PyObject *dflt = NULL;
    static char *argnames[] = {"name", "doc", "default",0};

    if(!PyArg_ParseTupleAndKeywords(args,kw,"|OOO",argnames,&name,&doc,&dflt))
        return 0;
    
    self = (Setting *) (type->tp_alloc(type, 0));
    if (self != NULL) {
        if (name == NULL) name = Py_None;
        if (doc == NULL) doc = Py_None;
        if (dflt == NULL) dflt = Py_None;
        Py_INCREF(name);
        Py_INCREF(doc);
        Py_INCREF(dflt);
        self->name = name;
        self->doc = doc;
        self->dflt = dflt;
    }
    return (PyObject *) self;
}

static PyObject *
Setting_call(Setting *self, PyObject *args, PyObject *kw)
{
    PyObject    *value = NULL;
    static char *argnames[] = {"value", 0};

    if(!PyArg_ParseTupleAndKeywords(args, kw, "|O", argnames, &value))
        return 0;
    
    if (value == NULL){
        printf("C1\n");
    } else {
        printf("C2\n");
    }
    Py_INCREF(Py_None);
    return Py_None;
}

static int
Setting_traverse(Setting *self, visitproc visit, void *arg)
{
	Py_VISIT(self->name);
	Py_VISIT(self->doc);
	Py_VISIT(self->dflt);
	return 0;
}

static void
Setting_dealloc(Setting *self)
{
    Py_XDECREF(self->name);
    Py_XDECREF(self->doc);
    Py_XDECREF(self->dflt);
}



static PyMemberDef Setting_members[] = {
	{"name", T_OBJECT, offsetof(Setting, name), READONLY,
	         PyDoc_STR("Setting name")},
	{"doc", T_OBJECT, offsetof(Setting, doc), READONLY,
	         PyDoc_STR("Setting doc")},
	{"default", T_OBJECT, offsetof(Setting, dflt), READONLY,
	         PyDoc_STR("Setting default")},
	{0}
};

static PyMethodDef Setting_methods[] = {
    /* {"demo",	(PyCFunction)Setting_demo,	METH_VARARGS,
 	PyDoc_STR("demo() -> None")}, */
    {NULL, NULL}		/* sentinel */
};


static PyTypeObject SettingType = {
	/* The ob_type field must be initialized in the module init function
	 * to be portable to Windows without using C++. */
	PyObject_HEAD_INIT(NULL)
	0,					/*ob_size*/
	"_speedups.Setting",			/*tp_name*/
	sizeof(Setting),			/*tp_basicsize*/
	0,					/*tp_itemsize*/
	(destructor)Setting_dealloc, 		/*tp_dealloc*/
	0,					/*tp_print*/
	0,                      		/*tp_getattr*/
	0,                      		/*tp_setattr*/
	0,					/*tp_compare*/
	0,					/*tp_repr*/
	0,					/*tp_as_number*/
	0,					/*tp_as_sequence*/
	0,					/*tp_as_mapping*/
	0,					/*tp_hash*/
        (ternaryfunc)Setting_call, 		/*tp_call*/
        0,                      		/*tp_str*/
        0,                      		/*tp_getattro*/
        0,                      		/*tp_setattro*/
        0,                      		/*tp_as_buffer*/
        Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, /*tp_flags*/
        0,                      		/*tp_doc*/
        (traverseproc)Setting_traverse, 	/*tp_traverse*/
        0,                      		/*tp_clear*/
        0,                      		/*tp_richcompare*/
        0,                      		/*tp_weaklistoffset*/
        0,                      		/*tp_iter*/
        0,                      		/*tp_iternext*/
        Setting_methods,        		/*tp_methods*/
        Setting_members,        		/*tp_members*/
        0,                      		/*tp_getset*/
        0,                      		/*tp_base*/
        0,                      		/*tp_dict*/
        0,                      		/*tp_descr_get*/
        0,                      		/*tp_descr_set*/
        0,                      		/*tp_dictoffset*/
        0,                      		/*tp_init*/
        0,                      		/*tp_alloc*/
        Setting_new,            		/*tp_new*/
        0,                      		/*tp_free*/
        0,                      		/*tp_is_gc*/
};








static PyMethodDef speedups_methods[] = {
	{"_read_ctx",  _read_ctx,  METH_NOARGS},
	{"_write_ctx", _write_ctx, METH_NOARGS},
	{"snapshot",   snapshot,   METH_NOARGS},
	{"swap",       swap,       METH_VARARGS},
	{"new_ctx",    new_ctx,    METH_NOARGS},
	{NULL, NULL}
};


PyMODINIT_FUNC
init_speedups(void)
{
    PyObject *module, *symbols, *tmp;

    module = Py_InitModule("_speedups", speedups_methods);
    if (module == NULL)
        return;

    /* string constans */
    frozen = PyString_FromString("__frozen__");
    if (frozen == NULL)
        return;
    ctx_key = PyString_FromString("wm4.contexts");
    if (ctx_key == NULL)
        return;

    /* from lib4.symbols import NOT_GIVEN, NOT_FOUND */
    tmp = PyString_FromString("lib4.symbols");
    symbols = PyImport_Import(tmp);
    not_given = PyObject_GetAttrString(symbols, "NOT_GIVEN");
    not_found = PyObject_GetAttrString(symbols, "NOT_FOUND");
    Py_DECREF(tmp);

    /* Types */
    if (PyType_Ready(&SettingType) < 0)
        return;
    PyModule_AddObject(module, "Setting", (PyObject *)&SettingType);
}
