from lib4.symbols import *
from lib4.runtime import cloneFunction, ImportProxy
from lib4.imports import importObject, importString
from protocols.advice import getMRO
from wm4.exceptions import SettingConflict, PropertyNotFound
from wm4.utils.property_rules import *
from types import FunctionType
from thread import get_ident
import protocols, re, sys

from interfaces import *


__all__ = [
    'with_', 'enter_',
    'new_ctx', 'snapshot', 'swap', 'StateContextManager','ValueContextManager',
    'Global', 'Config', 'Setting', 'ResourceArea', 'Resource', 'factoryFor',
    'prefix_suffix',
    'property_name',
    'AbstractProperty', 'Property', 'FlatProperty', 'TypeProperty',
    'ImportProxy',
    # from property_rules
    'IRule', 'Value', 'Rule', 'ExpressionRule', 'ImportRule',
    'Registry', 'pluginsFor', 'PluginsFor', 'Plugin',
]

def with_(ctx, func, *args, **kw):
    inp = ctx.__enter__()
    try:
        retval = func(*args, **kw)
    except:
        ctx.__exit__(*sys.exc_info())
        raise
    else:
        ctx.__exit__(None, None, None)
        return retval

def enter_(ctx):
    old = snapshot()
    ctx.__enter__()
    def exit_(et=None, ei=None, tb=None):
        ctx.__exit__(et, ei, tb)
        return swap(old)
    return exit_

contexts = {}

def _read_ctx():
    tid = get_ident()
    try:
        return contexts[tid]
    except:
        d = contexts[tid] = {}
        return d

def _write_ctx():
    tid = get_ident()
    try:
        d = contexts[tid]
    except KeyError:
        d = contexts[tid] = {}
    if '__frozen__' in d:
        d = contexts[tid] = d.copy()
        del d['__frozen__']
    return d

def new_ctx():
    """Return a new, empty thread state"""
    return {'__frozen__':True}

def snapshot():
    """Return a snapshot of the all of this thread's current state globals

    This returns an object that can be passed into ``context.swap()`` to
    restore all context globals to what they were at this point in time.
    A copy-on-write strategy is used, so that snapshots can always be taken
    in constant time, and no extra memory is used if multiple snapshots are
    taken during a period where no changes have occurred.
    """
    ctx = _read_ctx()
    ctx['__frozen__'] = True
    return ctx

def swap(state):
    """Set this thread's state to `snapshot`, returning a "before" snapshot

    `snapshot` should be a value previously returned by either the
    ``snapshot()`` function or by ``swap()``.  Before the thread state
    is set, a new snapshot is taken and returned.

    The passed-in snapshot is not modified by this routine, so you can reuse
    it as many times as you want to restore that particular state.
    """
    old = snapshot()
    contexts[get_ident()] = state
    return old




class StateContextManager(object):

    __slots__ = 'state', 'old'

    def __init__(self, state):
        self.state = state
        self.old = None

    def __enter__(self):
        assert self.old is None
        self.old = swap(self.state)

    def __exit__(self, et, ei, tb):
        assert self.old is not None
        swap(self.old)
        self.old = None


class ValueContextManager(object):

    __slots__ = 'key', 'new', 'old'

    def __init__(self, key, new):
        self.key = key
        self.new = new
        self.old = NOT_GIVEN

    def __enter__(self):
        self.old = self.key()
        _write_ctx()[self.key] = self.new

    def __exit__(self, *exc_info):
        _write_ctx()[self.key] = self.old


def Global(name='global', default=None, doc=None, module=None):

    _not_given = NOT_GIVEN
    _rctx = _read_ctx
    _vcm = ValueContextManager

    def f(value = _not_given):
        if value is _not_given:
            return _rctx().get(f, default)
        else:
            return _vcm(f, value)
    f = cloneFunction(f, name, doc, module)
    return f



class ValueCell(object):
    __slots__ = 'factory', 'value',

    def __init__(self, factory):
        self.factory = factory
        self.value = NOT_OBJECT

    def __call__(self, l, f):
        if self.value is NOT_OBJECT:
            self.value = self.factory(l,f)
        return self.value

AlreadyReadCell = ValueCell(lambda l,f: NOT_FOUND)


class Config(object):

    current = staticmethod(lambda: None)

    def __init__(self, parent=None, ignoreConflicts=False):
        if parent is None:
            parent = self.current()
        if parent is None:
            parent = getattr(self,'root',None)
        self.parent = parent
        self.data = {}
        self.ignoreConflicts = ignoreConflicts

    def __enter__(self):
        key = self.current
        self.old = key()
        _write_ctx()[key] = self

    def __exit__(self, *exc_info):
        _write_ctx()[self.current] = self.old

    def __getitem__(self,key):
        try:
            return self.data[key]
        except KeyError:
            try:
                self[key] = value = self.parent[key]
            except TypeError:
                raise AssertionError(
                    "Invalid Setting object or Config() hierarchy"
                )
            return value

    def __setitem__(self, key, val):
        old = self.data.setdefault(key, val)
        if old is not val and old != val:
            if self.ignoreConflicts:
                self.data[key] = val
            else:
                raise SettingConflict(
                    "a different value for is already defined", key, old, val
                )

    def iterParents(self):
        cfg = self
        while cfg:
            yield cfg
            cfg = cfg.parent

Config.root = Config()
Config.current = staticmethod(Global(
    'current_config', default=Config.root, doc='Current configuration scope'
))






def Setting(name="setting", default=None, doc=None, module=None, scope=Config):

    _cfg = scope.current
    _not_given = NOT_GIVEN

    def setting(arg=_not_given):
        '''context.Setting without docstring'''
        if arg is _not_given:
            return _cfg()[setting]
        else:
            _cfg()[setting] = arg

    setting = cloneFunction(setting,name,doc,module)
    Config.root[setting] = default
    return setting















class ResourceArea(object):

    current = staticmethod(Global('current_resource_area', None))

    get_config = staticmethod(Config.current)

    def __init__(self, config=None):
        if config is None:
            config = Config.current()
        self.config = config
        self.cache = {}

    def __enter__(self):
        key = self.current
        self.old = key()
        _write_ctx()[key] = self

    def __exit__(self, *exc_info):
        _write_ctx()[self.current] = self.old

    def __getitem__(self, key):
        res = self.cache.get(key, NOT_FOUND)
        if res is NOT_FOUND:
            cfg = self.config
            factory = cfg[key]
            res = self.cache[key] = with_(cfg, factory)
        return res

protocols.declareAdapter(
    protocols.NO_ADAPTER_NEEDED,
    provides = [IResourceFactory], forTypes = [FunctionType, type]
)

# XXX
protocols.declareAdapter(
    ImportProxy, provides=[IResourceFactory], forTypes=[basestring]
)





def Resource(
    name="resource", factory=None, doc=None, module=None, scope=ResourceArea
):
    _not_given = NOT_GIVEN
    _area = scope.current
    _cfg = scope.get_config

    factory = IResourceFactory(factory, None)

    def resource(fact=_not_given):
        if fact is _not_given:
            area = _area()
            try:
                return area[resource]
            except TypeError:
                if area is None:
                    raise RuntimeError(
                        "Cannot be used outside of ResourceArea",
                        (name,factory,doc,module), area
                    )
                raise
        else:
            _cfg()[resource] = fact

    resource = cloneFunction(resource, name, doc, module=None)
    Config.root[resource] = factory
    return resource


def factoryFor(res, factory=NOT_GIVEN):
    if factory is NOT_GIVEN:
        return Config.current()[res]
    else:
        Config.current()[res] = IResourceFactory(factory)






#from _context import BaseProperty

class ConfigCell(object):

    __slots__ = 'rule', 'value'

    def __init__(self, rule, value=NOT_OBJECT):
        self.rule = rule
        self.value = value

    def __iter__(self):
        return iter((self.rule, self.value))


_alreadyRead = ConfigCell(NOT_OBJECT)


class BaseProperty(object):
    __slots__ = ()

    def lookupRule(self, cfg, rKey, keys):
        data = cfg.data
        for fKey in keys:
            rule, value = data.setdefault(fKey, (NOT_OBJECT, NOT_OBJECT))
            if rule is not NOT_OBJECT:
                return rule, fKey
        if cfg.parent:
            return self.lookupRule(cfg.parent, rKey, keys)
        else:
            return NOT_OBJECT, NOT_OBJECT

    def __call__(self, value=NOT_GIVEN, default=NOT_GIVEN):
        cfg = self.get_config()
        data = cfg.data

        reg_key = self.regKey()
        if value is NOT_GIVEN:

            lookup_keys = self.lookupKeys()

            rule, val = data.setdefault(
                reg_key, (NOT_OBJECT, NOT_OBJECT)
            )
            if val is not NOT_OBJECT:
                if val is NOT_FOUND:
                    if default is NOT_GIVEN:
                        raise PropertyNotFound(reg_key)
                    else:
                        return default
                return val

            fkey = reg_key
            if rule is NOT_OBJECT:
                rule, fkey = self.lookupRule(cfg, reg_key, lookup_keys)

            if rule is NOT_OBJECT:
                val = NOT_FOUND
            else:
                val = self.makeValue(rule, reg_key, fkey)

            data[reg_key] = (data[reg_key][0], val)

            if val is NOT_FOUND:
                if default is NOT_GIVEN:
                    raise PropertyNotFound(reg_key)
                else:
                    return default
            else:
                return val
        else:
            cfg[reg_key] = (self.rule_interface(value), NOT_OBJECT)



class AbstractProperty(BaseProperty):

    protocols.advise(instancesProvide=[IProperty])

    __slots__ = 'value', 'lookup_keys', 'reg_key'

    get_config = staticmethod(Config.current)

    rule_interface = IRule

    def setdefault(self, default):
        Config.root[self.regKey()] = (IRule(default), NOT_OBJECT)

    def regKey(self):
        return (self.__class__, self.value)

    def lookupKeys(self):
        raise NotImplementedError

    def makeValue(self, rule, reg, key):
        return rule(reg[1], key[1])

    def __cmp__(self, other):
        return cmp(self.regKey(), other)

    def __hash__(self):
        return hash(self.regKey())

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.value)





def prefix_suffix(lookProp, foundProp):
    pref = foundProp
    if not foundProp.endswith('*'):
        return pref, ''
    pref = pref[:-1]
    assert lookProp.startswith(pref), '%s x %s' % (lookProp, pref)
    if pref:
        return pref, lookProp[len(pref):]
    else:
        return '', lookProp







pm = re.compile(r"^(\*|([-+!:_a-z0-9]+(\.[-+!:_a-z0-9]+)*(\.\*)?))$",re.I).match

class property_name(str):

    def __new__(klass, value):
        if not pm(value):
            raise ValueError('Invalid Property name', value)
        return str.__new__(klass, value)


class Property(AbstractProperty):
    """Name of a configuration property, usable as a configuration key"""

    __slots__ = ()

    def __new__(klass, value, default=NOT_GIVEN):
        if not pm(value):
            raise ValueError('Invalid Property value', value)
        self = AbstractProperty.__new__(klass)
        self.value = value
        if default is not NOT_GIVEN:
            self.setdefault(default)
        return self

    def asPrefix(self):
        p = self.value
        if p.endswith('*'):
            p=p[:-1]
        if p and not p.endswith('.'):
            p=p+'.'
        return p

    def lookupKeys(self):
        val = self.value
        keys = [val]
        if val.endswith('*'):
            return keys
        name = val
        while '.' in name:
            name = name[:name.rindex('.')]
            keys.append(name+'.*')
        keys.append('*')
        return keys

    def regKey(self):
        return self.value

    def makeValue(self, rule, reg, key):
        return rule(reg, key)







fpm = re.compile( r"^(\*|[-+!:@._a-z0-9]+)$", re.I ).match

class FlatProperty(AbstractProperty):

    __slots__ = ()

    def __new__(klass, value, default=NOT_GIVEN):
        #if not fpm(value):
        #    raise ValueError('Invalid FlatProperty value', value)
        self = AbstractProperty.__new__(klass)
        self.value = value
        if default is not NOT_GIVEN:
            self.setdefault(default)
        return self

    def lookupKeys(self):
        val = self.value
        klass = self.__class__
        if val == '*':
            return (klass, val),
        else:
            return (klass, val), (klass, '*')





class TypeProperty(AbstractProperty):

    __slots__ = ()

    _registry = {}

    def __new__(klass, value, default=NOT_GIVEN):
        reg = (klass, value)
        self = klass._registry.get(reg)
        if self is None:
            self = klass._registry[reg] = AbstractProperty.__new__(klass)
            self.value = value
            if value == '*':
                self.lookup_keys = (klass, '*'),
            else:
                self.lookup_keys = tuple(
                    (klass,base) for base in getMRO(value))+ ((klass,'*'),)
        if default is not NOT_GIVEN:
            self.setdefault(default)
        return self

    def lookupKeys(self):
        return self.lookup_keys


class Registry(object):

    __slots__ = 'name',


    class __metaclass__(type):

        def __init__(cls, name, bases, cdict):
            cls._instances = {}

    get_config = staticmethod(Config.current)

    def __new__(klass, name):
        self = klass._instances.get(name)
        if self is None:
            self = object.__new__(klass)
            self.name = name
            klass._instances[name] = self
        return self

    def set(self, attr, val):
        self.setFactory(attr, lambda: val)

    def setFactory(self, attr, fact):
        currentCfg = self.get_config()
        regCell = currentCfg.data.get(self)
        if regCell is None:
            regCell = currentCfg.data[self] = ConfigCell({})
        else:
            if regCell.value is not NOT_OBJECT:
                raise SettingConflict('Already read')
        regCell.rule[attr] = fact


    def __call__(self):
        currentCfg = self.get_config()
        regCell = currentCfg.data.get(self)
        if regCell is None:
            regCell = currentCfg.data[self] = ConfigCell({})
        elif regCell.value is not NOT_OBJECT:
            return regCell.value

        values = regCell.value = {}

        for cfg in currentCfg.iterParents():
            regCell = cfg.data.get(self)
            if regCell is None:
                continue
            for a,f in regCell.rule.iteritems():
                if a not in values:
                    values[a] = f()
        return values


def pluginsFor(comp):
    return dict(
        (k, Plugin('%s:%s' % (comp, k))())
        for k,v in PluginsFor(comp)().items()
    )

class PluginsFor(Registry):
    pass


class Plugin(Registry):
    pass
