# -*- coding: ISO-8859-2 -*-

from unittest import *
from wm4.config.api import *
from wm4.exceptions import SettingConflict, PropertyNotFound
from wm4.utils.property_rules import *
from lib4.runtime import ImportProxy
from itertools import count


class TestGlobals(TestCase):

    def test01_VariableName(self):
        x = Global('foo')
        self.assertEqual(x.__name__, 'foo')

    def test10_DefaultValue(self):
        x = Global('x')
        self.assertEqual(x(), None)
        y = Global('y', 42)
        self.assertEqual(y(), 42)

    def test20_AsContextManager(self):
        x = Global('x',1)
        self.assertEqual(x(), 1)
        def g():
            self.assertEqual(x(), 3)
        def f():
            self.assertEqual(x(), 2)
            with_(x(3), g)
            self.assertEqual(x(), 2)
        with_(x(2), f)
        self.assertEqual(x(), 1)

    def test30_PushWithExplicitPop(self):
        x = Global('x',1)
        self.assertEqual(x(), 1)
        def mkpush():
            self.assertEqual(x(), 1)
            c = x(2)
            c.__enter__()
            self.assertEqual(x(), 2)
            c.__exit__(None,None,None)
        mkpush()
        self.assertEqual(x(), 1)



class TestConfig(TestCase):

    def test30_ConfigAsContextManager(self):
        p = Property('x', 42)
        c1 = Config.current()
        c2 = Config(c1)
        self.assertEqual(p(), 42)
        with_(c2, p, 43)
        self.assertEqual(with_(c2, p), 43)
        self.assertEqual(p(), 42)


class TestSettings(TestCase):

    def test10_DefaultValue(self):
        s = Setting('s')
        self.assertEqual(s(), None)
        s = Setting('s', 42)
        self.assertEqual(s(), 42)

    def test12_InvalidMultipleSetting(self):
        s = Setting('s', 1)
        self.assertEqual(s(), 1)
        s(1) # OK
        self.assertRaises(SettingConflict, s, 2)

    def test14_InheritingValueFromParentConfig(self):
        s = Setting('s', 1)
        cfg = Config()
        r = with_(cfg, lambda: (Config.current(), s()))
        self.assertEqual(r, (cfg,1))

    def test16_NewValuesWithinChildConfig(self):
        s = Setting('s', 1)
        cfg = Config()
        cur = Config.current

        def f(val=2):
            self.assertEqual(cur(), cfg)
            s(val)
            self.assertEqual(s(), val)

        with_(cfg, f)
        self.assertEqual(s(), 1)

    def test18_InvalidMultipleSettingInChildConfig(self):
        s = Setting('s', 1)
        cfg = Config()
        cur = Config.current

        def f(val=2):
            self.assertEqual(cur(), cfg)
            s(val)
            self.assertEqual(s(), val)
            return val

        self.assertEqual(with_(cfg, f), 2)
        self.assertRaises(SettingConflict, with_, cfg, f, 3)
        self.assertEqual(with_(cfg, f), 2)
        self.assertEqual(s(), 1)





class TestResources(TestCase):

    def test10_NoDefaultValue(self):
        r = Resource('r', dict)
        self.assertRaises(RuntimeError, r)

    def test12_NoDefaultValue(self):
        r = Resource('r', dict)
        a = ResourceArea()
        d1 = with_(a, r)
        self.failUnless(isinstance(d1,dict))
        d2 = with_(a, r)
        self.failUnless(d1 is d2)

    def test14_DefaultFactory(self):
        r = Resource('r', dict)
        self.failUnless(factoryFor(r) is dict)

    def test16_ReplacingFactory(self):
        r = Resource('r', dict)
        cfg = Config()
        with_(cfg, factoryFor, r, list)
        self.failUnless(with_(cfg, factoryFor, r) is list)
        self.failUnless(factoryFor(r) is dict)

        a1 = ResourceArea()
        r1 = with_(a1, r)
        self.assertEqual(r1, {})
        a2 = ResourceArea(cfg)
        r2 = with_(a2, r)
        self.assertEqual(r2, [])

    def test20_StringAsResourceFactory(self):
        r = Resource('r','__builtin__.dict')
        d = with_(ResourceArea(), r)
        self.assertEqual(d, {})

        r = Resource('r',ImportProxy('__builtin__.dict'))
        d = with_(ResourceArea(), r)
        self.assertEqual(d, {})






class TestProperties(TestCase):

    def setUp(self):
        self.cfg = Config()
        self.exit_ = enter_(self.cfg)

    def tearDown(self):
        self.exit_()

    def test01_Syntax(self):
        self.assertRaises(ValueError, Property, '')
        self.assertRaises(ValueError, Property, '*.b')
        self.assertRaises(ValueError, Property, 'a..b')
        self.assertRaises(ValueError, Property, 'a.*.b')
        self.assertRaises(ValueError, Property, '.a.b')
        self.failUnless(Property('*'))
        self.failUnless(Property('a'))
        self.failUnless(Property('a.b'))

    def test02_DefaultValueRegistration(self):
        p = Property('x.y.z', 43)
        p(42)
        self.assertEqual(self.cfg.parent.data['x.y.z'][0].value, 43)
        self.assertEqual(self.cfg.data['x.y.z'][0].value, 42)

    def test03_Comparison(self):
        p1 = Property('x.y')
        p2 = Property('x.y')
        p3 = Property('x.z')
        self.failIf(p1 is p2)
        self.assertEqual(p1, p2)
        self.assertNotEqual(p1, p3)

    def test10_ExplicitValue(self):
        p = Property('a')
        p(1)
        self.assertEqual(p(), 1)
        self.assertRaises(SettingConflict, Property('a'), Value(1))
        self.assertRaises(PropertyNotFound, Property('b'))
        self.assertRaises(PropertyNotFound, Property('*'))

    def test11_DefaultValue(self):
        p = Property('a.*')
        p(1)
        self.assertEqual(Property('a.a')(), 1)
        self.assertEqual(Property('a.b')(), 1)
        self.assertRaises(PropertyNotFound, Property('*'))
        self.assertRaises(PropertyNotFound, Property('a'))
        self.assertRaises(PropertyNotFound, Property('aa'))

    def test13_PrefixWildcard(self):
        p = Property('a.*')
        p(lambda l,f: prefix_suffix(l,f)[1])
        self.assertEqual(Property('a.*')(), '*')
        self.assertEqual(Property('a.a')(), 'a')
        self.assertEqual(Property('a.b')(), 'b')
        self.assertEqual(Property('a.b.c')(), 'b.c')
        self.assertRaises(PropertyNotFound, Property('*'))
        self.assertRaises(PropertyNotFound, Property('aa'))
        self.assertRaises(PropertyNotFound, Property('b'))
        self.assertRaises(PropertyNotFound, Property('a.b.*'))

    def test14_RootWildcard(self):
        p = Property('*')
        p(lambda l,f: prefix_suffix(l,f)[1])
        self.assertEqual(Property('*')(), '*')
        self.assertEqual(Property('a')(), 'a')
        self.assertEqual(Property('a.b')(), 'a.b')
        self.assertEqual(Property('a.b.c')(), 'a.b.c')
        self.assertRaises(PropertyNotFound, Property('a.*'))

    def test15_SettingAfterNotFound(self):
        p = Property('a')
        p(1)
        self.assertEqual(p(), 1)
        self.assertRaises(SettingConflict, Property('a'), Value(1))
        self.assertRaises(PropertyNotFound, Property('b'))
        self.assertRaises(SettingConflict, Property('b'), Value(1))

    def test20_RuleInheriting(self):
        p = Property('a.*')
        p(lambda r,f: prefix_suffix(r,f)[1])

        cfg = Config()
        exit_ = enter_(cfg)
        try:
            self.assertEqual(Property('a.*')(), '*')
            self.assertEqual(Property('a.b')(), 'b')
            self.assertRaises(PropertyNotFound, Property('b'))
        finally:
            exit_()

        self.assertEqual(Property('a.*')(), '*')
        self.assertEqual(Property('a.b')(), 'b')
        self.assertRaises(SettingConflict, Property('b'), Value(1))

        p = Property('a.c')
        p(lambda r,f: 'C(%s)'% r)
        self.assertEqual(Property('a.c')(), 'C(a.c)')

        exit_ = enter_(cfg)
        try:
            self.assertEqual(Property('a.*')(), '*')
            self.assertEqual(Property('a.b')(), 'b')
            self.assertRaises(PropertyNotFound, Property('b'))
            self.assertEqual(Property('a.c')(), 'C(a.c)')
        finally:
            exit_()

    def test30_ValueCaching(self):
        from itertools import count
        cfg1 = Config()
        cfg2 = Config(cfg1)
        cfg3 = Config(cfg1)
        counter = count().next
        p = Property('c.*')
        p(lambda l,f: 'c.%d' % counter())
        exit_ = enter_(cfg2)
        try:
            self.assertEqual(Property('c.a')(), 'c.0')
            self.assertEqual(Property('c.b')(), 'c.1')
        finally:
            exit_()
        exit_ = enter_(cfg1)
        try:
            self.assertEqual(Property('c.a')(), 'c.2')
            self.assertEqual(Property('c.b')(), 'c.3')
        finally:
            exit_()
        exit_ = enter_(cfg3)
        try:
            self.assertEqual(Property('c.a')(), 'c.4')
            self.assertEqual(Property('c.b')(), 'c.5')
        finally:
            exit_()
        exit_ = enter_(cfg1)
        try:
            self.assertEqual(Property('c.a')(), 'c.2')
            self.assertEqual(Property('c.b')(), 'c.3')
        finally:
            exit_()
        exit_ = enter_(cfg2)
        try:
            self.assertEqual(Property('c.a')(), 'c.0')
            self.assertEqual(Property('c.b')(), 'c.1')
        finally:
            exit_()
        exit_ = enter_(cfg3)
        try:
            self.assertEqual(Property('c.a')(), 'c.4')
            self.assertEqual(Property('c.b')(), 'c.5')
        finally:
            exit_()

    def test31_ChildConfigDoesntAffectParent(self):
        p = Property('x.*')
        class X:
            def __init__(self):
                pass
        p(lambda r,f: X())
        c2 = Config()
        p11 = p()
        p21 = with_(c2, p)
        p12 = p()
        p22 = with_(c2, p)
        assert p11 is p12
        assert p21 is p22
        assert p11 is not p21

    def test40_Comparison(self):
        self.assertEqual(Property('a'), Property('a'))
        self.assertNotEqual(Property('a'), Property('b'))



class TestFlatProperties(TestCase):

    def setUp(self):
        self.cfg = Config()
        self.exit_ = enter_(self.cfg)

    def tearDown(self):
        self.exit_()

    def XXXtest01_Syntax(self):
        self.assertRaises(ValueError, FlatProperty, '')
        self.assertRaises(ValueError, FlatProperty, 'a.*')
        self.assertRaises(ValueError, FlatProperty, '*.b')
        self.assertRaises(ValueError, FlatProperty, 'a.*.b')
        self.failUnless(FlatProperty('*'))
        self.failUnless(FlatProperty('a'))
        self.failUnless(FlatProperty('a.b'))
        self.failUnless(FlatProperty('a...b'))
        self.failUnless(FlatProperty('a@s:b'))

    def test02_Keys(self):
        p = FlatProperty('a.b')
        self.assertEqual(p.regKey(), (FlatProperty,'a.b'))
        self.assertEqual(p.lookupKeys(),
            ((FlatProperty,'a.b'), (FlatProperty, '*'))
        )
        self.assertEqual(repr(p), "FlatProperty('a.b')")
        self.assertEqual(str(p), "FlatProperty('a.b')")

    def test03_WildcardKeys(self):
        p = FlatProperty('*')
        self.assertEqual(p.regKey(), (FlatProperty,'*'))
        self.assertEqual(p.lookupKeys(),((FlatProperty,'*'),))

    def test04_DefaultValueRegistration(self):
        p = FlatProperty('x.y.z', 43)
        p(42)
        self.assertEqual(
            self.cfg.parent.data[(FlatProperty,'x.y.z')][0].value,43)
        self.assertEqual(
            self.cfg.data[(FlatProperty,'x.y.z')][0].value, 42)

    def test10_ExplicitValue(self):
        p = FlatProperty('a')
        p(1)
        self.assertEqual(p(), 1)
        self.assertRaises(SettingConflict, FlatProperty('a'), Value(1))
        self.assertRaises(PropertyNotFound, FlatProperty('b'))
        self.assertRaises(PropertyNotFound, FlatProperty('*'))

    def test11_DefaultValue(self):
        p = FlatProperty('*')
        p(1)
        self.assertEqual(FlatProperty('a')(), 1)
        self.assertEqual(FlatProperty('b')(), 1)
        self.assertEqual(FlatProperty('a.b')(), 1)

    def test15_SettingAfterNotFound(self):
        p = FlatProperty('a')
        p(1)
        self.assertEqual(p(), 1)
        self.assertRaises(SettingConflict, FlatProperty('a'), Value(1))
        self.assertRaises(PropertyNotFound, FlatProperty('b'))
        self.assertRaises(SettingConflict, FlatProperty('b'), Value(1))

    def test20_RuleInheriting(self):
        p = FlatProperty('*')
        p(lambda r,f: '%s:%s' % (r,f))

        cfg = Config()
        exit_ = enter_(cfg)
        try:
            self.assertEqual(FlatProperty('*')(), '*:*')
            self.assertEqual(FlatProperty('a.b')(), 'a.b:*')
        finally:
            exit_()

        self.assertEqual(FlatProperty('*')(), '*:*')
        self.assertEqual(FlatProperty('a.b')(), 'a.b:*')
        self.assertRaises(SettingConflict, FlatProperty('a.b'), Value(1))

        p = FlatProperty('a.c')
        p(lambda r,f: 'C(%s:%s)' % (r,f))
        self.assertEqual(FlatProperty('a.c')(), 'C(a.c:a.c)')

        exit_ = enter_(cfg)
        try:
            self.assertEqual(FlatProperty('*')(), '*:*')
            self.assertEqual(FlatProperty('a')(), 'a:*')
            self.assertEqual(FlatProperty('a.b')(), 'a.b:*')
            self.assertEqual(FlatProperty('a.c')(), 'C(a.c:a.c)')
        finally:
            exit_()

    def test30_ValueCaching(self):
        from itertools import count
        cfg1 = Config()
        cfg2 = Config(cfg1)
        counter = count().next
        p = FlatProperty('*')
        p(lambda l,f: 'c.%d' % counter())
        exit_ = enter_(cfg2)
        try:
            self.assertEqual(FlatProperty('c.a')(), 'c.0')
            self.assertEqual(FlatProperty('c.b')(), 'c.1')
        finally:
            exit_()
        exit_ = enter_(cfg1)
        try:
            self.assertEqual(FlatProperty('c.a')(), 'c.2')
            self.assertEqual(FlatProperty('c.b')(), 'c.3')
        finally:
            exit_()
        exit_ = enter_(cfg2)
        try:
            self.assertEqual(FlatProperty('c.a')(), 'c.0')
            self.assertEqual(FlatProperty('c.b')(), 'c.1')
        finally:
            exit_()
        exit_ = enter_(cfg1)
        try:
            self.assertEqual(FlatProperty('c.a')(), 'c.2')
            self.assertEqual(FlatProperty('c.b')(), 'c.3')
        finally:
            exit_()



class CA: pass
class CB(CA): pass

class NA(object): pass
class NB(NA): pass
class NC(NA): pass

class TestTypeProperties(TestCase):

    def setUp(self):
        self.cfg = Config()
        self.exit_ = enter_(self.cfg)

    def tearDown(self):
        self.exit_()

    def test01_WildcardKeys(self):
        p = TypeProperty('*')
        self.assertEqual(p.regKey(), (TypeProperty,'*'))
        self.assertEqual(p.lookupKeys(), ((TypeProperty, '*'),))

    def test02_ClassicalClassKeys(self):
        p = TypeProperty(CB)
        self.assertEqual(p.regKey(), (TypeProperty,CB))
        self.assertEqual(p.lookupKeys(),(
            (TypeProperty,CB), (TypeProperty,CA), (TypeProperty,'*')
        ))

    def test03_NewClassKeys(self):
        p = TypeProperty(NB)
        self.assertEqual(p.regKey(), (TypeProperty,NB))
        self.assertEqual(p.lookupKeys(),(
            (TypeProperty,NB), (TypeProperty,NA),
            (TypeProperty,object), (TypeProperty,'*')
        ))

    def test04_DefaultValueRegistration(self):
        p = TypeProperty(NB, 43)
        p(42)
        k = p.regKey()
        self.assertEqual(self.cfg.parent.data[k][0].value, 43)
        self.assertEqual(self.cfg.data[k][0].value, 42)

    def test05_Comparison(self):
        p1 = TypeProperty(int)
        p2 = TypeProperty(int)
        p3 = TypeProperty(object)
        self.failUnless(p1 is p2) # instances are cached
        self.assertEqual(p1, p2)
        self.assertNotEqual(p1, p3)

    def test10_ExplicitValue(self):
        p = TypeProperty(NB)
        p(1)
        self.assertEqual(p(), 1)
        self.assertRaises(SettingConflict, p, Value(1))
        self.assertRaises(PropertyNotFound, TypeProperty(int))
        self.assertRaises(PropertyNotFound, TypeProperty('*'))

    def test11_DefaultValue(self):
        p = TypeProperty('*')
        p(1)
        self.assertEqual(TypeProperty(NA)(), 1)
        self.assertEqual(TypeProperty(NB)(), 1)
        self.assertEqual(TypeProperty(CA)(), 1)
        self.assertEqual(TypeProperty(CB)(), 1)

    def test12_BaseLookup(self):
        class MyProp(TypeProperty): pass
        MyProp(CA)('C')
        MyProp(NA)('N')
        self.assertEqual(MyProp(CA)(), 'C')
        self.assertEqual(MyProp(CB)(), 'C')
        self.assertEqual(MyProp(NA)(), 'N')
        self.assertEqual(MyProp(NB)(), 'N')

    def test15_SettingAfterNotFound(self):
        class MyProp(TypeProperty): pass
        p = MyProp(CA)
        p(1)
        self.assertEqual(p(), 1)
        self.assertRaises(SettingConflict, p, Value(1))
        self.assertRaises(PropertyNotFound, MyProp(NA))
        self.assertRaises(SettingConflict, MyProp(NA), Value(1))

    def test31_ChildConfigDoesntAffectParent(self):
        class MyProp(TypeProperty): pass
        class X:
            def __init__(self):
                pass
        p = MyProp(CA)
        p(lambda r,f: X())
        c2 = Config()
        p11 = p()
        p21 = with_(c2, p)
        p12 = p()
        p22 = with_(c2, p)
        assert p11 is p12
        assert p21 is p22
        assert p11 is not p21




class MyReg(Registry):
    pass

class MyReg2(MyReg):
    pass

class TestRegistries(TestCase):

    def setUp(self):
        self.cfg = Config()
        self.exit_ = enter_(self.cfg)

    def tearDown(self):
        self.exit_()

    def test02_InstanceCaching(self):
        i1 = MyReg('foo')
        i2 = MyReg('foo')
        i3 = MyReg2('foo')
        assert i1 is i2
        assert i1 is not i3

    def test10_Empty(self):
        r = MyReg('empty')()
        self.assertEqual(r, {})

    def test20_SetValue(self):
        r = MyReg('foo')
        r.set('a', 'A')
        r.set('b', 42)
        self.assertEqual(r(), {'a':'A', 'b':42})

    def test21_SetValueAfterRead(self):
        r = MyReg('foo')
        r.set('a', 'A')
        self.assertEqual(r(), {'a':'A'})
        self.assertRaises(SettingConflict, r.set, 'b', 42)

    def test30_SetFactory(self):
        r = MyReg('foo')
        r.setFactory('a', lambda:'A')
        r.setFactory('b', lambda:42)
        self.assertEqual(r(), {'a':'A', 'b':42})

    def test31_SetFactoryAfterRead(self):
        r = MyReg('foo')
        r.setFactory('a', lambda:'A')
        self.assertEqual(r(), {'a':'A'})
        self.assertRaises(SettingConflict, r.setFactory, 'b', 42)

    def test40_ValuesAreCached(self):
        next = count().next
        r = MyReg('foo')
        r.setFactory('a', next)
        self.assertEqual(r(), {'a': 0})
        self.assertEqual(r(), {'a': 0})

    def test50_InheritingAttributes(self):
        r = MyReg('bar')
        r.setFactory('a', lambda:'A')
        self.assertEqual(r(), {'a':'A'})

        exit = enter_(Config())
        r.setFactory('a', lambda:'AA')
        r.setFactory('b', lambda: 42)
        self.assertEqual(r(), {'a':'AA', 'b':42})
        exit()

        self.assertEqual(r(), {'a':'A'})






def suite():
    return TestSuite((
        makeSuite(TestGlobals),
        makeSuite(TestConfig),
        makeSuite(TestSettings),
        makeSuite(TestResources),
        makeSuite(TestProperties),
        makeSuite(TestFlatProperties),
        makeSuite(TestTypeProperties),
        makeSuite(TestRegistries),
    ))


if __name__ == '__main__':
    main(defaultTest='suite')
