# -*- coding: ISO-8859-2 -*-

from unittest import *
from wm4.config.api import *
from wm4.exceptions import PropertyNotFound
from wm4.config.ini_files import ConfigReader

test_resource = Resource('test_resource', lambda:'DEFAULT_RESOURCE')
test_setting  = Setting('test_setting', 'DEFAULT_SETTING')

class TestConfigReader(TestCase):

    def setUp(self):
        self.state = snapshot()
        loadConfig('pkgfile:wm4/core.ini')

    def tearDown(self):
        swap(self.state)

    def parse(self, lines):
        cr = ConfigReader()
        cr.readString(lines)

    def test10_SectionlessProperties(self):
        self.parse(
            'a = "1"\n'
            'a.b = "1.2"\n'
            'a.b.c = "1.2.3"\n'
            'a.* = "1.%s" % config.prefix_suffix(property,foundProperty)[1]\n'
        )
        self.assertEqual(Property('a')(), "1")
        self.assertEqual(Property('a.b')(), "1.2")
        self.assertEqual(Property('a.b.c')(), "1.2.3")
        self.assertEqual(Property('a.c')(), "1.c")
        self.assertEqual(Property('a.c.d')(), "1.c.d")
        self.assertEqual(Property('a.*')(), "1.*")
        self.assertRaises(PropertyNotFound, Property('b'))

    def test20_ImportOnDemand(self):
        self.parse(
            '[Import On Demand]\n'
            'a = "xmlrpclib"\n'
            'b = importString("xmlrpclib")\n'
            'c = lambda: "foo"\n'
            '[e]\n'
            'a = a\n'
            'b = b\n'
            'c = c\n'
        )
        import xmlrpclib
        self.assertEqual(Property('e.a')(), xmlrpclib)
        self.assertEqual(Property('e.b')(), xmlrpclib)
        self.assertEqual(Property('e.c')()(), "foo")

    def test30_ResourceFactories(self):
        self.parse((
            '[Resource Factories]\n'
            '%s.test_resource = lambda: "CONFIGURED_RESOURCE"\n'
        ) % __name__)
        area = ResourceArea()
        exit_ = enter_(area)
        try:
            self.assertEqual(test_resource(), "CONFIGURED_RESOURCE")
        finally:
            exit_()

    def test40_ResourceFactories(self):
        self.parse((
            '[Global Settings]\n'
            '%s.test_setting = "CONFIGURED_SETTING"\n'
        ) % __name__)
        self.assertEqual(test_setting(), "CONFIGURED_SETTING")


    def test50_LoadSettingsFromMapping(self):
        self.parse(
            '[p.q]\n'
            'a = 3\n'
            'c = 4\n'
            '[Load Settings From]\n'
            "mapping = {'p.q.a':1, 'p.q.b':2}\n"
        )
        self.assertEqual(Property('p.q.a')(), 1)
        self.assertEqual(Property('p.q.b')(), 2)
        self.assertEqual(Property('p.q.c')(), 4)

    def test51_LoadSettingsFromMapping(self):
        self.parse(
            '[Load Settings From]\n'
            "mapping = {'p.q.a':1, 'p.q.b':2}\n"
            '[p.q]\n'
            'a = 3\n'
            'c = 4\n'
        )
        self.assertEqual(Property('p.q.a')(), 3)
        self.assertEqual(Property('p.q.b')(), 2)
        self.assertEqual(Property('p.q.c')(), 4)

    def test55_LoadSettingsFromMapping(self):
        self.parse(
            '[p.q]\n'
            'a = 3\n'
            'c = 4\n'
            '[Load Settings From]\n'
            "mapping = {'p.q.a':1, 'p.q.b':2}\n"
        )
        self.assertEqual(Property('p.q.a')(), 1)
        self.assertEqual(Property('p.q.b')(), 2)
        self.assertEqual(Property('p.q.c')(), 4)

    def test56_LoadSettingsFromFile(self):
        self.parse(
            '[Load Settings From]\n'
            "file = 'pkgfile:wm4.config.tests/include.ini'\n"
            '[p.q]\n'
            'a = 3\n'
            'c = 4\n'
        )
        import xmlrpclib
        self.assertEqual(Property('p.q.a')(), 3)
        self.assertEqual(Property('p.q.b')(), 2)
        self.assertEqual(Property('p.q.c')(), 4)
        self.assertEqual(Property('p.q.p')(), xmlrpclib)

    def test57_LoadSettingsFromFile(self):
        self.parse(
            '[Load Settings From]\n'
            "file = naming.lookup('pkgfile:wm4.config.tests/include.ini')\n"
            '[p.q]\n'
            'p = p2\n'
        )
        self.assertEqual(Property('p.q.a')(), 1)
        self.assertEqual(Property('p.q.b')(), 2)
        self.assertRaises(NameError, Property('p.q.p'))

    def test60_PluginsFor(self):
        self.parse(
            '[Plugin mycomp:P1]\n'
            'a = 42\n'
            'b = "foo"\n'
            '[Plugin mycomp:P2]\n'
            'a = 43\n'
            'c = "boo"\n'
            '[Plugin your.app:delta your.app:theta]\n'
            'b = "baz"\n'
            'c = 2.0\n'
            '[Plugin your.app:theta]\n'
            'c = 3.0\n'
            'd = 42\n'
        )
        self.assertEqual(
            pluginsFor('mycomp'),
            {'P1': dict(a=42, b='foo'), 'P2': dict(a=43, c='boo')}
        )
        self.assertEqual(
            pluginsFor('your.app'),
            {'delta':dict(b='baz', c=2.0), 'theta':dict(b='baz', c=3.0, d=42)}
        )
        exit = enter_(Config())
        try:
            self.parse(
                '[Plugin mycomp:P1]\n'
                'a = 84\n'
                '[Plugin mycomp:P3]\n'
                'c = "C"\n'
            )
            self.assertEqual(
                pluginsFor('mycomp'),
                {'P1': dict(a=84, b='foo'),
                 'P2': dict(a=43, c='boo'),
                 'P3': dict(c='C'),
                }
            )
        finally:
            exit()
        self.assertEqual(
            pluginsFor('mycomp'),
            {'P1': dict(a=42, b='foo'), 'P2': dict(a=43, c='boo')}
        )

def suite():
    return TestSuite((
        makeSuite(TestConfigReader),
    ))


if __name__ == '__main__':
    main(defaultTest='suite')
