# -*- coding: ISO-8859-2 -*-

from unittest import *
from wm4.api import *
from wm4.events.api import *

class ChangeRegistry:

    def __init__(self):
        self.changes = []

    def __call__(self, evt):
        self.changes.append(evt)


LIST_INIT = range(10)
LIST_INIT_A = [(i,i) for i in LIST_INIT]

class TestObservableList(TestCase):

    def prepare(self):
        l,c = ObservableList(), ChangeRegistry()
        ListChanged.subscribe(l,c)
        return l,c

    def check(self, reg, changes):
        self.assertEqual(len(reg.changes), len(changes))
        for evt, (added, removed) in zip(reg.changes, changes):
            self.assertEqual(evt.added, added)
            self.assertEqual(evt.removed, removed)

    def test01_Append(self):
        l,c = self.prepare()
        l.append('A')
        self.assertEqual(len(l), 1)
        self.check(c, [([(0,'A')],[])])

    def test02_Extend(self):
        l,c = self.prepare()
        l.extend(LIST_INIT)
        self.assertEqual(list(l), range(10))
        self.check(c, [(LIST_INIT_A,[])])

    def test03_Insert(self):
        l,c = self.prepare()
        l.extend(LIST_INIT)
        l.insert(-2, 'A')
        l.insert(2, 'B')
        self.check(c, [(LIST_INIT_A,[]), ([(8,'A')],[]), ([(2,'B')],[])])
        self.assertEqual(
            list(l), [0, 1,'B', 2, 3, 4, 5, 6, 7,'A', 8, 9]
        )

    def test04_Getitem(self):
        l,c = self.prepare()
        l.extend(LIST_INIT)
        self.assertEqual(l[0], 0)
        self.assertEqual(l[1], 1)
        self.assertEqual(l[2], 2)
        self.check(c, [(LIST_INIT_A,[])])

    def test05_Getslice(self):
        l,c = self.prepare()
        l.extend(LIST_INIT)
        self.assertEqual(l[1::2], [1,3,5,7,9])
        self.check(c, [(LIST_INIT_A,[])])

    def test10_Setitem(self):
        l,c = self.prepare()
        l.extend(LIST_INIT)
        l[1] = 'X'
        self.assertEqual(list(l), [0,'X',2,3,4,5,6,7,8,9])
        self.check(c, [(LIST_INIT_A,[]), ([(1,'X')], [(1,1)])])

    def test11_SetitemWithNegativeIndex(self):
        l,c = self.prepare()
        l.extend(LIST_INIT)
        l[-2] = 'X'
        l[-1] = 'Y'
        self.assertEqual(list(l), [0,1,2,3,4,5,6,7,'X','Y'])
        self.check(c, [
            (LIST_INIT_A,[]),
            ([(8,'X')], [(8,8)]),
            ([(9,'Y')], [(9,9)]),
        ])

    def test12_SetSameValue(self):
        l,c = self.prepare()
        l.extend(LIST_INIT)
        l[1] = 1
        self.assertEqual(list(l), [0,1,2,3,4,5,6,7,8,9])
        self.check(c, [(LIST_INIT_A,[])])

    def test20_Setslice(self):
        l,c = self.prepare()
        l.extend(LIST_INIT)
        l[4:-2] = ['C',5,'E','F']
        self.assertEqual(list(l), [0,1,2,3,'C',5,'E','F',8,9])
        self.check(c, [
            (LIST_INIT_A,[]),
            ([(4,'C'),(6,'E'),(7,'F')],[(4,4),(6,6),(7,7)])
        ])

    def test21_SetsliceWithShrink(self):
        l,c = self.prepare()
        l.extend(LIST_INIT)
        l[4:-2] = ['C',5,'E']
        self.assertEqual(list(l), [0,1,2,3,'C',5,'E',8,9])
        self.check(c, [
            (LIST_INIT_A,[]),
            ([(4,'C'),(6,'E'),(7,8),(8,9)],
             [(4,4),(6,6),(7,7),(8,8),(9,9)])
        ])

    def test22_SetsliceWithExpand(self):
        l,c = self.prepare()
        l.extend(LIST_INIT)
        l[4:-4] = ['C',5,'E']
        self.assertEqual(list(l), [0,1,2,3,'C',5,'E',6,7,8,9])
        self.check(c, [
            (LIST_INIT_A,[]),
            ([(4,'C'),(6,'E'),(7,6),(8,7),(9,8),(10,9)],
             [(4,4),(6,6),(7,7),(8,8),(9,9)])
        ])

    def test23_SetsliceWithStride(self):
        l,c = self.prepare()
        l.extend(LIST_INIT)
        l[0::3] = ['A','B',6,'D']
        self.assertEqual(list(l), ['A',1,2,'B',4,5,6,7,8,'D'])
        self.check(c, [
            (LIST_INIT_A,[]),
            ([(0,'A'),(3,'B'),(9,'D')],[(0,0),(3,3),(9,9)])
        ])

    def test30_Delitem(self):
        l,c = self.prepare()
        l.extend(LIST_INIT)
        del l[7]
        self.assertEqual(len(l), 9)
        self.check(c, [
            (LIST_INIT_A,[]),
            ([(7,8),(8,9)], [(7, 7), (8, 8), (9, 9)]),
        ])

    def test31_DelitemWithNegativeIndex(self):
        l,c = self.prepare()
        l.extend(LIST_INIT)
        del l[-1]
        self.assertEqual(list(l), [0,1,2,3,4,5,6,7,8])
        self.check(c, [(LIST_INIT_A,[]), ([],[(9,9)])])

    def test35_Delslice(self):
        l,c = self.prepare()
        l.extend(LIST_INIT)
        del l[1:3]
        self.assertEqual(list(l), [0,3,4,5,6,7,8,9])
        self.check(c, [
            (LIST_INIT_A,[]),
            ([(1,3),(2,4),(3,5),(4,6),(5,7),(6,8),(7,9)],
             [(1,1),(2,2),(3,3),(4,4),(5,5),(6,6),(7,7),(8,8),(9,9)]),
        ])

    def test36_DelsliceWithStride(self):
        l,c = self.prepare()
        l.extend(LIST_INIT)
        del l[0::3]
        self.assertEqual(list(l), [1,2,4,5,7,8])
        self.check(c, [
            (LIST_INIT_A,[]),
            ([(0,1),(1,2),(2,4),(3,5),(4,7),(5,8)],
             [(0,0),(1,1),(2,2),(3,3),(4,4),(5,5),(6,6),(7,7),(8,8),(9,9)]),
        ])




DICT_INIT = dict(a=1, b=2, c=3)

class TestObservableDict(TestCase):

    def prepare(self):
        d,c = ObservableDict(), ChangeRegistry()
        DictChanged.subscribe(d,c)
        return d,c

    def check(self, reg, changes):
        self.assertEqual(len(reg.changes), len(changes))
        for evt, (added, removed) in zip(reg.changes, changes):
            self.assertEqual(evt.added, added)
            self.assertEqual(evt.removed, removed)

    def test01_Update(self):
        d,c = self.prepare()
        d.update(DICT_INIT)
        self.assertEqual(len(d), len(DICT_INIT))
        self.check(c, [(DICT_INIT,{})])

    def test02_InvalidGetitem(self):
        d,c = self.prepare()
        try: d[1]
        except: pass
        self.assertEqual(len(d), 0)
        self.check(c, [])

    def test03_Getitem(self):
        d,c = self.prepare()
        d.update(DICT_INIT)
        self.assertEqual(d['a'], 1)
        self.assertEqual(d['b'], 2)
        self.check(c, [(DICT_INIT,{})])

    def test04_ClearSetitem(self):
        d,c = self.prepare()
        d[1] = 'A'
        d[2] = 'B'
        self.assertEqual(len(d), 2)
        self.check(c, [({1:'A'},{}), ({2:'B'},{})])

    def test05_ReplacingSetitem(self):
        d,c = self.prepare()
        d.update(DICT_INIT)
        d['a'] = 'A'
        d['b'] = 'B'
        self.assertEqual(len(d), 3)
        self.check(c, [(DICT_INIT,{}),({'a':'A'},{'a':1}),({'b':'B'},{'b':2})])

    def test06_Delitem(self):
        d,c = self.prepare()
        d.update(DICT_INIT)
        del d['a']
        del d['b']
        self.assertEqual(len(d), 1)
        self.check(c, [(DICT_INIT,{}), ({}, {'a':1}), ({}, {'b':2})])


def suite():
    return TestSuite((
        makeSuite(TestObservableList),
        makeSuite(TestObservableDict),
    ))


if __name__ == '__main__':
    main(defaultTest='suite')
