from protocols import advise, Interface, Attribute
import copy

__all__ = [
    'ITranslatable', 'ITranslationService',
    'MessageIDFactory', 'MessageID', 'EncapsulatedMessageID', 'typeName',
    'monthName', 'monthNameShort', 'dayName',
]


class ITranslationService(Interface) :

    def translate(
        msgid, domain=None, mapping=None, context=None, lang=None, default=None
    ):
        """Translates one message"""

    def typeName(cls, lang=None, default=None):
        """Returns local type name for given class"""


class ITranslatable(Interface) :
    domain = Attribute('Translation domain')
    mapping = Attribute('Substitution mapping')
    default = Attribute('Default value')




class MessageID(unicode) :
    """Message ID.

    This is a string used as a message ID.  It has a domain attribute that is
    its source domain, and a default attribute that is its default text to
    display when there is no translation.  domain may be None meaning there is
    no translation domain.  default may also be None, in which case the
    message id itself implicitly serves as the default text.

    MessageID objects also have a mapping attribute which must be set after
    construction of the object.  This is used when translating and
    substituting variables.
    """

    advise(instancesProvide=[ITranslatable])

    __slots__ = ('domain', 'default', 'mapping')

    def __new__(cls, ustr, domain=None, default=None, mapping={}):
        self = unicode.__new__(cls, ustr)
        self.domain = domain
        if default is None:
            self.default = ustr
        else:
            self.default = unicode(default)
        self.mapping = mapping
        return self

    def __getstate__(self):
        return unicode(self), self.domain, self.default, self.mapping

    def __setstate__(self, (ustr, domain, default, mapping)):
        super(MessageID, self).__init__(ustr)
        self.domain = domain
        if default is None:
            self.default = ustr
        else:
            self.default = default
        self.mapping = mapping

    def __mod__(self, mapping) :
        r = copy.copy(self)
        r.mapping = mapping
        return r




class MessageIDFactory :

    def __init__(self, domain) :
        self.domain = domain

    def __call__(self, uni, default=None) :
        return MessageID(uni, self.domain, default)






class EncapsulatedMessageID(unicode):

    __slots__ = ('translations')

    def __new__(cls, ustr, translations={}):
        self = unicode.__new__(cls, ustr)
        self.translations = translations
        return self

    def translate(self, lang):
        t = self.translations.get(lang)
        if t is None:
            return unicode(self)
        else:
            return unicode(t)



def typeName(cls, lang=None, default=None):
    if default is None:
        default = cls.__name__
    return MessageID(
            u'TypeName of %s.%s' % (cls.__module__, cls.__name__),
            domain = 'wm4',
            default = unicode(default),
    )


monthArray = [None,
    u'January', u'February', u'March', u'April', u'May', u'June',
    u'July', u'August', 'September', 'October', 'November', 'December'
]

def monthName(mnum, lang=None, default=None):
    eng = monthArray[mnum]
    if default is None:
        default = eng
    return MessageID(eng, domain='wm4', default=eng)

monthArrayShort = [None,
    u'Jan', u'Feb', u'Mar', u'Apr', u'May', u'Jun',
    u'Jul', u'Aug', u'Sep', u'Oct', u'Nov', u'Dec'
]

def monthNameShort(mnum, lang=None, default=None):
    eng = monthArrayShort[mnum]
    if default is None:
        default = eng
    return MessageID(eng.lower(), domain='wm4', default=eng)

dayArray = [u'Mon', u'Tue', u'Wed', u'Thu', u'Fri', u'Sat', u'Sun']

def dayName(mnum, lang=None, default=None):
    eng = dayArray[mnum]
    if default is None:
        default = eng
    return MessageID(eng, domain='wm4', default=eng)
