from wm4.core import *
from lib4.imports import importString
from api import MessageID, ITranslationService, typeName
import re, os, gettext

__all__ = [
    'TRANSLATOR', 'NullTranslator', 'GettextTranslator', 'TranslationService'
]


def TRANSLATOR(domain, default=NOT_GIVEN):
    return config.Property('wm4.i18n.translator.%s' % domain, default)

def default_translator(property, foundProperty):
    prefix, suffix = config.prefix_suffix(property, foundProperty)
    return GettextTranslator.forPackage(suffix)

TRANSLATOR('*', default_translator)




_cre = re.compile(r'\$(?:([_A-Za-z][-\w]*)|\{([_A-Za-z][-\w]*)\})')




class NullTranslator:

    def translate(
        self, ts, text, domain=None, mapping=None, lang=None, default=None
    ):
        if default is not None:
            text = default

        def _repl(m) :
            val = mapping[m.group(m.lastindex)]
            if isinstance(val, MessageID):
                val = ts.translate(val, lang=lang)
            return unicode(val)

        return _cre.sub(_repl, text)




class GettextFallback:
    def ugettext(self,text):
        return NOT_FOUND
_fallback = GettextFallback()


class GettextTranslator:

    def __init__(self, catalogDir):
        self.dir = catalogDir
        self.cache = {}

    def forPackage(klass, package, subdir='messages'):
        try:
            mod = importString(package)
        except:
            return NullTranslator()
        dir = os.path.split(mod.__file__)[0]
        if subdir:
            dir = os.path.join(dir,subdir)
        if os.path.isdir(dir):
            return klass(dir)
        else:
            return NullTranslator()
    forPackage = classmethod(forPackage)


    def translate(
        self, ts, text, domain=None, mapping=None, lang=None, default=None
    ):

        def _repl(m) :
            val = mapping[m.group(m.lastindex)]
            if isinstance(val, MessageID):
                val = ts.translate(val, lang=lang)
            return unicode(val)

        if default is None:
            default = text

        if not lang:
            return _cre.sub(_repl, default)

        catalog = None
        if lang in self.cache:
            catalog = self.cache[lang]
        else:
            register = []
            mo = os.path.join(self.dir, lang+'.mo')
            if os.path.exists(mo):
                register.append(lang)
            elif len(lang) > 2:
                mo = os.path.join(self.dir, lang[:2]+'.mo')
                if os.path.exists(mo):
                    register.append(lang[:2])
                else:
                    mo = None
            else:
                mo = None
            if mo is not None:
                catalog = gettext.GNUTranslations(fp=open(mo))
                catalog.add_fallback(_fallback)
                for lang in register:
                    self.cache[lang] = catalog
            else:
                self.cache[lang] = None

        if catalog is None:
            return _cre.sub(_repl, default)

        msg = catalog.ugettext(text)
        if msg is NOT_FOUND:
            return _cre.sub(_repl, default)
        else:
            return _cre.sub(_repl, msg)

    def __repr__(self):
        return '<GettextTranslator %r>' % self.dir







class TranslationService(binding.Component):

    protocols.advise(instancesProvide=[ITranslationService])

    #translators = binding.Make(config.Namespace('peak.i18n.translator'))

    def translate(
        self, msgid, domain=None, mapping=None, lang=None, default=None
    ):
        lang = {
            'cz' : 'cs',
        }.get(lang,lang)

        # MessageID attributes override arguments
        if isinstance(msgid, MessageID) :
            domain = msgid.domain
            mapping = msgid.mapping
            default = msgid.default

        if not domain:
            if default is not None:
                return default
            else:
                return msgid

        translator = TRANSLATOR(domain)() #self.translators.get(domain)

        return translator.translate(
            self,msgid,domain=domain,mapping=mapping,default=default,lang=lang
        )

    def typeName(self, cls, lang=None, default=None):
        return typeName(cls,lang=lang, default=default)
