from wm4.api import *
import re, os, sys
from lib4.imports import importString
from urlparse import urlsplit, urlunsplit
from urllib import quote, unquote

from urls import *
from interfaces import *

__all__ = [
    'URLFactory', 'FileFactory', 'DataFactory',
    'OpenableURL', 'FileURL', 'PkgFileURL', 'DataURL',
]


class URLFactory(object):

    protocols.advise(instancesProvide=[IStreamFactory])

    def __init__(self, address):
        self.address = address

    def open(self, mode='rb'):
        if mode != 'rb':
            raise ValueError("'rb' mode required")
        from urllib2 import urlopen
        return urlopen(self.address)

    def exists(self):
        from urllib2 import urlopen, HTTPError
        try:
            urlopen(self.address).next()
        except (HTTPError, IOError, StopIteration):
            return False
        else:
            return True


class FileFactory(object):

    protocols.advise(instancesProvide=[IStreamFactory])

    def __init__(self, address):
        self.address = address

    def open(self, mode='rb'):
        return open(self.address, mode)

    def exists(self):
        return os.path.exists(self.address)



class DataFactory(object):
    protocols.advise(instancesProvide=[IStreamFactory])

    def __init__(self, data):
        self.data = data

    def open(self, mode='rb'):
        from cStringIO import StringIO
        return StringIO(self.data)



netloc_parser = re.compile(
    '^((?P<user>[^/:@]+)(:(?P<password>[^/:@]+))?@)?'
    '((?P<host>%s)(:(?P<port>%s))?)?$' % (DNSNAME_RE, PORT_RE), re.I).match


class OpenableURL(AbstractURL):

    slots = 'user', 'password', 'host', 'port', 'path', 'query', 'fragment'

    SCHEMES = set(['http', 'https', 'ftp'])

    @classmethod
    def parse(klass, scheme, data):
        if scheme not in klass.SCHEMES:
            raise exceptions.InvalidName('Invalid scheme',scheme,klass)
        user = password = host = port = None
        _, netloc, path, query, fragment = urlsplit(scheme+':'+data)
        if netloc:
            m = netloc_parser(netloc)
            if m:
                gd = m.groupdict()
                user = gd['user']
                password = gd['password']
                host = gd['host']
                port = gd['port']
            else:
                raise exceptions.InvalidName('Syntax error in URL',scheme,data)
        return klass(
            scheme, user=user, password=password, host=host, port=port,
            path=path or None, query=query or None, fragment=fragment or None
        )


    def __str__(self):
        netloc = ''
        if self.user:
            netloc += self.user
            if self.password:
                netloc += ':' + self.password
            netloc += '@'
        if self.host:
            netloc += self.host
            if self.port:
                netloc += ':' + self.port
        return urlunsplit((
            self.scheme, netloc,
            self.path or '', self.query or '', self.fragment or ''
        ))

    def getObject(self):
        return URLFactory(address=str(self))




class FileURL(OpenableURL):

    SCHEMES = set(['file'])

    def getObject(self):
        from urllib import url2pathname
        return FileFactory(address=url2pathname(self.path))

    @classmethod
    def fromFilename(klass, path):
        m = URLMATCH(path)
        if m and (m.end() > 2 or sys.platform <> 'win32'):
            return klass.parse(m.group(1), path[m.end():])
        else:
            from os.path import abspath, normcase
            from urllib import pathname2url
            return klass.parse('file', pathname2url(normcase(abspath(path))))





class PkgFileURL(AbstractURL):

    slots = 'package', 'path'

    MATCH = re.compile(
        '^(?P<package>[a-z][_a-z0-9]*(\\.[a-z][_a-z0-9]*)*)/(?P<path>.*)$', re.I
    ).match

    SCHEMES = set(['pkgfile'])

    def getObject(self):
        package = importString(self.package)
        filename = os.path.join(
            os.path.dirname(getattr(package,'__file__','')),
            *self.path.split('/')
        )
        return FileFactory(address=filename)

    def __str__(self):
        return 'pkgfile:%s/%s' % (self.package, self.path)


class DataURL(AbstractURL):

    slots = 'content_type', 'parameters', 'data'

    SCHEMES = set(['data'])

    #MATCH = re.compile('(?P<content_type>[^;]*)(;(?P<parameters>[^,]+))*,(?P<data>.*)', re.MULTILINE).match

    @classmethod
    def parse(klass, scheme, data):
        if scheme not in klass.SCHEMES:
            raise exceptions.InvalidName('Invalid scheme', scheme, klass)
        s = data.split(',', 1)
        if len(s) == 1:
            return klass(scheme, content_type=None, parameters=(), data=s[0])
        p, d = s
        if p:
            p = p.split(';')
            return klass(
                scheme, content_type=p[0], parameters=tuple(p[1:]), data=d
            )

        else:
            return klass(scheme, content_type=None, parameters=(), data=d)

    def getObject(self):
        if 'base64' in self.parameters:
            data = self.data.decode('base64')
        else:
            data = unquote(self.data)
        return DataFactory(data)

    def __str__(self):
        u = 'data:'
        if self.content_type:
            u += self.content_type
        if self.parameters:
            u += ';' + ';'.join(self.parameters)
        u += ',' + self.data
        return u
