# -*- coding: ISO-8859-2 -*-

from unittest import *
from wm4.api import *
from wm4.naming.openable import *
from urllib import quote, pathname2url
import os, sys



class TestFileURL(TestCase):

    def test10_FromFilename(self):
        def norm(p):
            n = pathname2url(
                os.path.normcase(os.path.abspath(os.path.normpath(p))))
            if n.startswith('///'): n = n[2:]
            return n

        f = FileURL.fromFilename('/a')
        self.failUnless(isinstance(f, FileURL))
        n = norm('/a')
        self.assertEqual(f.path, n)

        f = FileURL.fromFilename('./a')
        self.failUnless(isinstance(f, FileURL))
        n = norm(os.path.join(os.getcwd(),'a'))
        self.assertEqual(f.path, n)

        self.assertRaises(
            exceptions.InvalidName, FileURL.fromFilename, 'http://www.dat.cz')


    def test15_Parsing(self):
        for u in ('file:/a', 'file:///a'):
            f = naming.toName(u)
            self.assertEqual(f.scheme, 'file')
            self.assertEqual(f.user, None)
            self.assertEqual(f.password, None)
            self.assertEqual(f.host, None)
            self.assertEqual(f.port, None)
            self.assertEqual(f.path, '/a')
            self.assertEqual(f.query, None)
            self.assertEqual(f.fragment, None)

    def test20_Lookup(self):
        u = 'file://%s/__init__.py' % (
            pathname2url(os.path.dirname(os.path.abspath(__file__)))
        )
        o = naming.lookup(u)
        self.failUnless(isinstance(o, FileFactory))
        self.assertEqual(o.open().read(), '# empty\n')


class TestPkgFileURL(TestCase):

    def test10_Parsing(self):
        f = naming.toName('pkgfile:wm4.naming.tests/test_urls.py')
        self.assertEqual(f.scheme, 'pkgfile')
        self.assertEqual(f.package, 'wm4.naming.tests')
        self.assertEqual(f.path, 'test_urls.py')
        self.assertEqual(str(f), 'pkgfile:wm4.naming.tests/test_urls.py')

    def test12_InvalidPackage(self):
        self.assertRaises(exceptions.InvalidName, naming.toName, 'pkgfile:xxx')

    def test20_Lookup(self):
        o = naming.lookup('pkgfile:wm4.naming.tests/__init__.py')
        self.failUnless(isinstance(o, FileFactory))
        self.assertEqual(o.open().read(), '# empty\n')


class TestOpenableURL(TestCase):

    def test10_Parsing(self):
        f = naming.toName('http://www.dat.cz')
        self.assertEqual(f.scheme, 'http')
        self.assertEqual(f.user, None)
        self.assertEqual(f.password, None)
        self.assertEqual(f.host, 'www.dat.cz')
        self.assertEqual(f.port, None)
        self.assertEqual(f.path, None)
        self.assertEqual(f.query, None)
        self.assertEqual(f.fragment, None)

    def test11_Parsing(self):
        f = naming.toName('http://foo@www.dat.cz/')
        self.assertEqual(f.scheme, 'http')
        self.assertEqual(f.user, 'foo')
        self.assertEqual(f.password, None)
        self.assertEqual(f.host, 'www.dat.cz')
        self.assertEqual(f.port, None)
        self.assertEqual(f.path, '/')
        self.assertEqual(f.query, None)
        self.assertEqual(f.fragment, None)

    def test12_Parsing(self):
        f = naming.toName('http://foo:bar@www.dat.cz/?a=1#FR')
        self.assertEqual(f.scheme, 'http')
        self.assertEqual(f.user, 'foo')
        self.assertEqual(f.password, 'bar')
        self.assertEqual(f.host, 'www.dat.cz')
        self.assertEqual(f.port, None)
        self.assertEqual(f.path, '/')
        self.assertEqual(f.query, 'a=1')
        self.assertEqual(f.fragment, 'FR')

    def test21_Lookup(self):
        u = 'http://www.dat.cz/'
        o = naming.lookup(u)
        self.failUnless(isinstance(o, URLFactory))


DATA = ''.join(chr(i) for i in range(256))

class TestDataURL(TestCase):

    def test10_Simple(self):
        u = 'data:,foo'
        f = naming.toName(u)
        self.assertEqual(f.scheme, 'data')
        self.assertEqual(f.content_type, None)
        self.assertEqual(f.parameters, ())
        self.assertEqual(f.data, 'foo')
        self.assertEqual(str(f), u)
        o = naming.lookup(u)
        self.assertEqual(o.open().read(), 'foo')

    def test11_Quoted(self):
        u = 'data:,' + quote(DATA)
        f = naming.toName(u)
        self.assertEqual(f.scheme, 'data')
        self.assertEqual(f.content_type, None)
        self.assertEqual(f.parameters, ())
        self.assertEqual(f.data, quote(DATA))
        self.assertEqual(str(f), u)
        o = naming.lookup(u)
        self.assertEqual(o.open().read(), DATA)

    def test12_ContentType(self):
        u = 'data:text/plain,foo'
        f = naming.toName(u)
        self.assertEqual(f.scheme, 'data')
        self.assertEqual(f.content_type, 'text/plain')
        self.assertEqual(f.parameters, ())
        self.assertEqual(f.data, 'foo')
        self.assertEqual(str(f), u)
        o = naming.lookup(u)
        self.assertEqual(o.open().read(), 'foo')

    def test13_Encoded(self):
        u = 'data:text/plain;base64,' + DATA.encode('base64')
        f = naming.toName(u)
        self.assertEqual(f.scheme, 'data')
        self.assertEqual(f.content_type, 'text/plain')
        self.assertEqual(f.parameters, ('base64',))
        self.assertEqual(f.data, DATA.encode('base64'))
        self.assertEqual(str(f), u)
        o = naming.lookup(u)
        self.assertEqual(o.open().read(), DATA)

    def test20_NoComma(self):
        u = 'data:foo'
        f = naming.toName(u)
        self.assertEqual(f.scheme, 'data')
        self.assertEqual(f.content_type, None)
        self.assertEqual(f.parameters, ())
        self.assertEqual(f.data, 'foo')
        self.assertEqual(str(f), 'data:,foo')
        o = naming.lookup(u)
        self.assertEqual(o.open().read(), 'foo')

    def test21_MultiComma(self):
        u = 'data:,foo,bar'
        f = naming.toName(u)
        self.assertEqual(f.scheme, 'data')
        self.assertEqual(f.content_type, None)
        self.assertEqual(f.parameters, ())
        self.assertEqual(f.data, 'foo,bar')
        self.assertEqual(str(f), u)
        o = naming.lookup(u)
        self.assertEqual(o.open().read(), 'foo,bar')





class TestMisc(TestCase):

    def test10_Import(self):
        o = naming.lookup('import:sys.stderr')
        self.assertEqual(o, sys.stderr)
        o = naming.lookup('import:sys:stderr')
        self.assertEqual(o, sys.stderr)

    def test40_MockDB(self):
        from wm4.storage.mockdb import MockSqlConnection
        o = naming.lookup('mockdb:')
        assert isinstance(o, MockSqlConnection)

    def test50_Sqlite(self):
        from wm4.storage.connection import SqliteConnection
        o = naming.lookup('sqlite:/tmp/test.db')
        assert isinstance(o, SqliteConnection)
        # in memory sql database
        o = naming.lookup('sqlite:')
        assert isinstance(o, SqliteConnection)


def suite():
    return TestSuite((
        makeSuite(TestFileURL),
        makeSuite(TestPkgFileURL),
        makeSuite(TestOpenableURL),
        makeSuite(TestDataURL),
        makeSuite(TestMisc),
    ))


if __name__ == '__main__':
    main(defaultTest='suite')
