from wm4.core import *
from lib4.imports import importString as imps
import re

from interfaces import IAddress

__all__ = [
    'SCHEME_PARSER', 'PACKAGE_RE', 'DNSNAME_RE', 'PORT_RE',
    'URLMATCH', 'AbstractURL', 'parseURL', 'toName', 'lookup',
    'data_url',
]



def SCHEME_PARSER(scheme, default=NOT_GIVEN):
    return config.Property('wm4.naming.scheme_parser.%s' % scheme, default)


PACKAGE_RE = '([._a-z0-9]+)'
DNSNAME_RE = '([-.a-z0-9]+)'
PORT_RE    = '\d+'
URLMATCH   = re.compile('([-+.a-z0-9]+):',re.I).match



class AbstractName(object):
    # XXX not implemented

    protocols.advise(instancesProvide=[IAddress])


class AbstractURL(AbstractName):

    slots = ()

    def __init__(self, scheme, **values):
        self.scheme = scheme
        values = tuple((a, values.get(a))  for a in self.slots)
        for a,v in values:
            setattr(self,a,v)
        self._hashcmp = (self.__class__, scheme, values)

    MATCH = None

    SCHEMES = set([])

    @classmethod
    def parse(klass, scheme, data):
        if scheme not in klass.SCHEMES:
            raise exceptions.InvalidName('Invalid scheme', scheme, klass)
        m = klass.MATCH(data)
        if not m:
            raise exceptions.InvalidName('Syntax error in URL', scheme, data)
        return klass(scheme, **m.groupdict())

    def __hash__(self):
        return hash(self._hashcmp)

    def __eq__(self, other):
        if isinstance(other, AbstractURL):
            return self._hashcmp == other._hashcmp
        else:
            return False

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        klass = self.__class__
        return '%s(%s)' % (klass.__name__, ', '.join(
            '%s=%r' % (a,getattr(self,a)) for a in klass.slots
        ))


def makeURL(scheme, data):
    factory = SCHEME_PARSER(scheme)()
    return factory(scheme, data)


def parseURL(url):
    if isinstance(url, AbstractURL):
        return url
    elif isinstance(url, basestring):
        m = URLMATCH(url)
        if m:
            return makeURL(m.group(1), url[m.end():])
        else:
            raise exceptions.InvalidName('URL without scheme', url)
    else:
        raise ValueError('Unsupported URL type', url)



def toName(name):
    if isinstance(name, AbstractName):
        return name
    elif isinstance(name, basestring):
        m = URLMATCH(name)
        if m:
            return makeURL(m.group(1), name[m.end():])
        else:
            raise NotImplementedError
    else:
        raise ValueError('Unsupported Name type', name)


def lookup(name):
    n = toName(name)
    return n.getObject()



class ImportURL(AbstractURL):

    slots = 'importSpec',

    MATCH = re.compile(
        '^(?P<importSpec>[a-z][a-z0-9]*([.:][a-z][_a-z0-9]*)*)$', re.I
    ).match

    SCHEMES = set(['import'])

    def getObject(self):
        return imps(self.importSpec)



SCHEME_PARSER(
    'import', lambda r,f: ImportURL.parse)
SCHEME_PARSER(
    'pkgfile', lambda r,f: imps('wm4.naming.openable.PkgFileURL').parse)

SCHEME_PARSER(
    'file', lambda r,f: imps('wm4.naming.openable.FileURL').parse)
SCHEME_PARSER(
    'http', lambda r,f: imps('wm4.naming.openable.OpenableURL').parse)
SCHEME_PARSER(
    'https', lambda r,f: imps('wm4.naming.openable.OpenableURL').parse)
SCHEME_PARSER(
    'ftp', lambda r,f: imps('wm4.naming.openable.OpenableURL').parse)
SCHEME_PARSER(
    'data', lambda r,f: imps('wm4.naming.openable.DataURL').parse)

SCHEME_PARSER(
    'logger', lambda r,f: imps('wm4.running.logs.LoggerURL').parse)

SCHEME_PARSER(
    'pgsql', lambda r,f: imps('wm4.storage.connection.PgSqlURL').parse)
SCHEME_PARSER(
    'psycopg', lambda r,f: imps('wm4.storage.connection.PsycopgSqlURL').parse)
SCHEME_PARSER(
    'mssql', lambda r,f: imps('wm4.storage.connection.MSSqlURL').parse)
SCHEME_PARSER(
    'sqlite', lambda r,f: imps('wm4.storage.connection.SqliteURL').parse)
SCHEME_PARSER(
    'firebird', lambda r,f: imps('wm4.storage.connection.FirebirdURL').parse)
SCHEME_PARSER(
    'mockdb', lambda r,f: imps('wm4.storage.mockdb.MockSqlURL').parse)
SCHEME_PARSER(
    'ldap', lambda r,f: imps('wm4.storage.ldapconn.LDAPURL').parse)
SCHEME_PARSER(
    'ldaps', lambda r,f: imps('wm4.storage.ldapconn.LDAPURL').parse)
SCHEME_PARSER(
    'ldapi', lambda r,f: imps('wm4.storage.ldapconn.LDAPURL').parse)


def data_url(data):
    from urllib import quote
    return 'data:,' + quote(data)
