from wm4.api import *
from lib4.ast_builder import *

from expressions import *
from interfaces import *
from queries import *
from variables import *


__all__ = ['parse', 'QueryError', 'QueryBuilder']



def parse(expr, ctx, *namespaces):
    if isinstance(expr, query):
        return parse_expr(expr.source, QueryBuilder(ctx, *expr.namespaces))
    else:
        return parse_expr(expr, QueryBuilder(ctx, *namespaces))


class QueryError(Exception):
    pass



class QueryBuilder(object):
    '''Arguments:
            con
                database connection
            namespaces
                typicaly: local_dict, global_dict, __builtins__
                or:       frame.f_locals, frame.f_globals, frame.f_builtins
    '''

    simplify_comparisons = True

    def __init__(self, con, *namespaces):
        self.con = con
        self.namespaces = list(namespaces)
        self.transform = con.transform

    def Name(self,name):
        for ns in self.namespaces:
            value = ns.get(name,self)
            if value is self:
                continue
            if isinstance(value, Abstract):
                return value
            else:
                return Const(value)
        raise NameError(name)

    Const = Const

    _cmp_ops = {
        '>':GT, '>=':GE, '<':LT, '<=':LE, '<>':NE, '!=':NE, '==':EQ,
        'in':IN, 'not in':NOT_IN, 'is':IS, 'is not':IS_NOT
    }
    def Compare(self, left, ((op,right),)):
        klass = self._cmp_ops[op]
        left = build(self, left)
        right = build(self, right)
        if isinstance(left, Const) and isinstance(right, Const):
            return self.transform(Const(klass.func(left.value, right.value)))
        return self.transform(klass(left, right))

    def mkBinOp(klass):
        def method(self,left,right):
            left = build(self, left)
            right = build(self, right)
            if isinstance(left, Const) and isinstance(right, Const):
                return self.transform(Const(klass.func(left.value,right.value)))

            return self.transform(klass(left, right))
        return method

    Add        = mkBinOp(ADD)
    Sub        = mkBinOp(SUB)
    Mul        = mkBinOp(MUL)
    Div        = mkBinOp(DIV)
    Mod        = mkBinOp(MOD)
    FloorDiv   = mkBinOp(FLOOR)

    Power      = mkBinOp(POW)
    LeftShift  = mkBinOp(LSHIFT)
    RightShift = mkBinOp(RSHIFT)


    def multiOp(klass):
        def method(self,items):
            res = build(self, items[0])
            for item in items[1:]:
                tmp = build(self,item)
                if isinstance(res, Const) and isinstance(tmp, Const):
                    res = Const(klass.func(res.value, tmp.value))
                else:
                    res = klass(res, tmp)
            return self.transform(res)
        return method

    Bitand = multiOp(BAND)
    Bitor  = multiOp(BOR)
    Bitxor = multiOp(BXOR)

    def unaryOp(klass):
        def method(self,expr):
            tmp = build(self, expr)
            if isinstance(tmp, Const):
                return Const(klass.func(tmp.value))
            else:
                return self.transform(klass(tmp))
        return method

    UnaryPlus  = unaryOp(POS)
    UnaryMinus = unaryOp(NEG)
    Invert     = unaryOp(INV)
    Not        = unaryOp(NOT)

    def And(self, args):
        al = []
        for a in args:
            a = build(self, a)
            if isinstance(a, Const):
                if not a.value:
                    return a
            else:
                al.append(a)
        if len(al) == 1:
            return al[0]
        elif len(al) > 1:
            return self.transform(AND(*al))
        else:
            return Const(True)

    def Or(self, args):
        al = []
        for a in args:
            a = build(self, a)
            if isinstance(a, Const):
                if a.value:
                    return a
            else:
                al.append(a)
        if len(al) == 1:
            return al[0]
        elif al > 1:
            return self.transform(OR(*al))
        else:
            return Const(False)

    def Getattr(self, expr, name):
        obj = build(self,expr)
        if isinstance(expr, Const):
            return Const(getattr(obj.value, name))
        else:
            return self.transform(Getattr(obj,name))

    def Tuple(self, items):
        args = [build(self,a) for a in items]
        for a in args:
            if not isinstance(a, Const):
                return self.transform(Tuple(*args))
        return Const(tuple(a.value for a in args))

    def List(self, items):
        args = [build(self,a) for a in items]
        for a in args:
            if not isinstance(a, Const):
                return self.transform(List(*args))
        return Const([a.value for a in args])

    def Dict(self, items):
        items = [(build(self,k), build(self,v)) for k,v in items]
        for k,v in items:
            if not isinstance(k,Const) or not isinstance(v,Const):
                return self.transform(Dict(*(Tuple(k,v) for k,v in items)))
        else:
            return Const(dict((k.value, v.value) for k,v in items))

    def Subscript(self, obj, idx):
        obj = build(self, obj)
        idx = build(self, idx)
        if isinstance(idx, SliceObj):
            return self.transform(Slice(obj, *idx.args))
        else:
            return self.transform(Getitem(obj, idx))

    def Slice(self, start, stop):
        # XXX FIXME noninteger slices ???
        start = build(self, start)
        stop = build(self, stop)
        return SliceObj(start, stop, Const(None))

    def Sliceobj(self, start, stop, stride):
        start = build(self,start)
        stop = build(self,stop)
        stride = build(self,stride)
        return SliceObj(start, stop, stride)


    def CallFunc(self, funcExpr, args, kw, star, dstar):
        func = build(self,funcExpr)
        if not isinstance(func, Const):
            raise NotImplementedError('nonconst functions')
        if star:
            star = build(self,star)
            if not isinstance(star, Const):
                raise NotImplementedError('nonconst star argument')
        else:
            star = NoExpr
        if dstar:
            dstar = build(self,dstar)
            if not isinstance(dstar, Const):
                raise NotImplementedError('nonconst dstar argument')
        else:
            dstar = NoExpr
        func = func.value
        args = tuple(build(self, a) for a in args)
        kw   = dict((build(self,k).value, build(self,v)) for k,v in kw)

        if isinstance(func, type) and issubclass(func, Abstract):
            if star:
                args = args + tuple(star.value)
            if dstar:
                kw.update(dstar.value)
            return self.transform(func(*args, **kw))

        else:
            var = [a for a in args if not isinstance(a, Const)] or [
                   v for v in kw.itervalues() if not isinstance(v, Const)]
            if var:
                if star:
                    args = args + tuple(Const(a) for a in star.value)
                if dstar:
                    for k,v in dstar.value.iteritems():
                        kw[k] = Const(v)
                return self.transform(Call(func,args,kw))

            else:
                args = tuple(a.value for a in args)
                if star:
                    args = args + tuple(star.value)
                kw = dict((k,v.value) for k,v in kw.iteritems())
                if dstar:
                    kw.update(dstar.value)
                return Const(func(*args, **kw))

    def unpackAssignment(self, var, rel, counter):
        #if isinstance(rel, (SubselectVar, SelectAsQueryable)):
        if isinstance(var, Tuple):
            raise NotImplementedError(
                'tuple unpacking not supported', var, rel)
            if isinstance(rel, SelectAsQueryable):
                assert isinstance(rel.sel.expr, Tuple)
                assert len(var.args) == len(rel.sel.expr.args)
                ns = {}
                for a,e in zip(var.args, rel.sel.expr.args):
                    v,n = self.unpackAssignment(a, e, counter)
                    ns.update(n)
                return 'A%d' % counter(), ns
            if isinstance(rel, Tuple):
                assert len(var.args) == len(rel.args)
                ns = {}
                for a,e in zip(var.args, rel.args):
                    v,n = self.unpackAssignment(a, e, counter)
                    ns.update(n)
                return 'A%d' % counter(), ns
            else:
                raise NotImplementedError(
                    'tuple unpacking not supported', var, rel)
        else:
            var = var.name
            return var, {var:rel.alias(var)}

    def forAssignment(self, assign, rel):
        oldns = self.namespaces
        self.namespaces = [AssignNS()]
        from itertools import count
        try:
            var = build(self, assign)
            return self.unpackAssignment(var, rel, count().next)
        finally:
            self.namespaces = oldns


    def ListComp(self, expr, quals):
        # Prepare assignment namespace
        assignments = {}
        self.namespaces.insert(0, assignments)
        try:
            for_items = []
            for (assign, for_item, cond) in quals:
                rel = build(self, for_item)
                if isinstance(rel, Const):
                    rel = IQueryable(rel.value)
                else:
                    rel = IQueryable(rel)
                var, ns = self.forAssignment(assign, rel)
                assignments.update(ns)
                for_items.append((
                    var, rel.alias(var), cond and build(self,cond) or None
                ))
            expr = build(self, expr)

            # update
            if expr.__class__ is Update:
                if len(for_items) != 1:
                    raise NotImplementedError('%d for items' % len(for_items))
                assign, for_item, cond = for_items[0]
                assert expr.rel == for_item
                return self.transform(
                    UpdateQuery(expr.rel, expr.values, cond or NoExpr))

            # delete
            elif expr.__class__ is Delete:
                if len(for_items) != 1:
                    raise NotImplementedError('%d for items' % len(for_items))
                assign, for_item, cond = for_items[0]
                assert expr.rel == for_item
                return self.transform(
                    DeleteQuery(expr.rel, cond or NoExpr))

            # select
            else:
                return self.transform(BaseSelect(expr, for_items))

        finally:
            self.namespaces.pop(0)


class AssignNS:

    def get(self, name, defalt=None):
        return Assign(name)

class Assign(Abstract):

    __slots__ = ('name',)

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return 'A(%s)' % self.name
