from wm4.core import *
import operator, re
from itertools import chain
from interfaces import IExpression

__all__ = [
    'in_', 'not_in', 'log_and', 'log_or', 'brackets',
    'Abstract', 'Const',
    'ExprBase', 'UnOp', 'BinOp', 'ArithOp', 'CompOp',
    'POS','NEG', 'INV', 'NOT',
    'ADD','SUB','MUL','DIV','FLOOR','MOD', 'POW',
    'BAND','BOR','BXOR', 'INV', 'LSHIFT','RSHIFT',
    'LT','LE','EQ','NE','GE','GT', 'IN', 'NOT_IN', 'IS', 'IS_NOT',
    'Like', 'ILike',
    'AND', 'OR',
    'Getattr', 'Tuple', 'List', 'Dict', 'NoExpr', 'Call',
    'Cast', 'SliceObj', 'Getitem', 'Slice', 'Case',
]


# functions that are missing in module `operators'

def in_(a1, a2):
    return a1 in a2

def not_in(a1, a2):
    return a1 not in a2

def log_and(*args):
    for arg in args:
        if not arg:
            break
    return arg

def log_or(*args):
    for arg in args:
        if arg:
            break
    return arg


safe_expr = re.compile(r'^[-_.$a-z0-9\'\"]+(\([^()]*\))?$', re.I).match

def brackets(expr):
    '''enclose expression in brackets only if needed'''
    e = str(expr)
    if safe_expr(e):
        return e
    elif e.startswith('(') and e.endswith(')'):
        return e
    else:
        return '(%s)' % e





class Abstract(object):
    '''Base class for query expressions'''

    __metaclass__ = binding.Activator

    protocols.advise(instancesProvide=[IExpression])

    def conjuncts(self):
        return self,

    def disjuncts(self):
        return self,

    def alias(self, alias):
        return self

    typ = object

    relvars = {}

    aggregates = set()

    namespace = {}

    def __eq__(self, other):
        if isinstance(other, Abstract):
            return self.hashcmp == other.hashcmp
        else:
            return False

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self.hashcmp)

    def joins(self, builder):
        return ()


class Const(Abstract):
    '''Constant value'''

    protocols.advise(instancesProvide=[IExpression], asAdapterForTypes=[object])

    def __init__(self, value):
        self.value = value

    hashcmp = binding.Make(lambda s: (s.__class__, s.value))

    typ = binding.Make(lambda s: s.value.__class__)

    def __repr__(self):
        return repr(self.value)



class ExprBase(Abstract):
    '''Base class for expressions that consist of function/operator
       and sequence of child subexpressions. Attribute 'func' is class
       attribute, not instance attribute.
    '''

    func = None
    repr = None

    def __init__(self, *args):
        self.args = args

    hashcmp = binding.Make(
        lambda s:(s.__class__, tuple(a.hashcmp for a in s.args))
    )

    @classmethod
    def mk(klass, name, **cdict):
        return type(klass)(name, (klass,), cdict)

    def joins(self, builder):
        for a in self.args:
            for j in a.joins(builder):
                yield j

    def relvars(self):
        rv = filter(None, (a.relvars for a in self.args))
        if len(rv) == 1:
            return rv[0]
        elif rv:
            c = rv[0].copy()
            for d in rv[1:]:
                c.update(d)
            return c
        else:
            return {}
    relvars = binding.Make(relvars)


    aggregates = binding.Make(lambda s: set(
        chain(*(a.aggregates for a in s.args))
    ))


class UnOp(ExprBase):

    def __init__(self, arg):
        super(UnOp, self).__init__(arg)

    def __repr__(self):
        return '%s %s' % (self.repr, brackets(self.args[0]))

POS = UnOp.mk('POS', repr='+', func=operator.pos)
NEG = UnOp.mk('NEG', repr='-', func=operator.neg)
INV = UnOp.mk('INV', repr='~', func=operator.inv)
NOT = UnOp.mk('NOT', repr='not', func=operator.not_)

class BinOp(ExprBase):

    commut = False

    left  = binding.Make(lambda s: s.args[0])
    right = binding.Make(lambda s: s.args[1])

    def __init__(self, left, right):
        super(BinOp, self).__init__(left, right)

    def __repr__(self):
        l, r = self.args
        return '%s %s %s' % (brackets(l), self.repr, brackets(r))

    def hashcmp(self):
        if self.commut:
            return (self.__class__, tuple(sorted(a.hashcmp for a in self.args)))
        else:
            return (self.__class__, tuple(a.hashcmp for a in self.args))
    hashcmp = binding.Make(hashcmp)


class ArithOp(BinOp):
    pass

ADD = ArithOp.mk('ADD',repr='+',func=operator.add, commut=True)
SUB = ArithOp.mk('SUB',repr='-',func=operator.sub)
MUL = ArithOp.mk('MUL',repr='*',func=operator.mul, commut=True)
DIV = ArithOp.mk('DIV',repr='/',func=operator.div)
MOD = ArithOp.mk('MOD',repr='%',func=operator.mod)
FLOOR = ArithOp.mk('FLOOR',repr='//',func=operator.floordiv)

BAND = ArithOp.mk('BAND',repr='&',func=operator.and_, commut=True)
BOR = ArithOp.mk('BOR',repr='|',func=operator.or_, commut=True)
BXOR = ArithOp.mk('BXOR',repr='^',func=operator.xor, commut=True)

POW = ArithOp.mk('POW',repr='**',func=pow)

LSHIFT = ArithOp.mk('LSHIFT',repr='<<',func=operator.lshift)
RSHIFT = ArithOp.mk('RSHIFT',repr='>>',func=operator.rshift)


class CompOp(BinOp):
    pass

LT = CompOp.mk('LT', repr='<', func=operator.lt)
LE = CompOp.mk('LE', repr='<=', func=operator.le)
EQ = CompOp.mk('EQ', repr='==', func=operator.eq, commut=True)
NE = CompOp.mk('NE', repr='!=', func=operator.ne, commut=True)
GE = CompOp.mk('GE', repr='>=', func=operator.ge)
GT = CompOp.mk('GT', repr='>', func=operator.gt)

Like = CompOp.mk('Like', repr='LIKE')
ILike = CompOp.mk('ILike', repr='ILIKE')

IN = CompOp.mk('IN', repr='in', func=staticmethod(in_))
NOT_IN = CompOp.mk('NOT_IN', repr='not in', func=staticmethod(not_in))
IS = CompOp.mk('IS', repr='is', func=operator.is_, commut=True)
IS_NOT = CompOp.mk('IS_NOT', repr='is not', func=operator.is_not, commut=True)


class AND(ExprBase):

    repr = 'and'
    func = staticmethod(log_and)
    commut = True

    def __new__(klass, *args):
        if not args:
            raise TypeError('no valid argument for', klass, args)
        if len(args) == 1:
            return args[0]
        else:
            return ExprBase.__new__(klass)

    def __init__(self, *args):
        self.args = tuple(
            x for x in  chain(*(a.conjuncts() for a in args))
            if not isinstance(x, Const)
        )

    def conjuncts(self):
        return self.args

    def __repr__(self):
        j = ' %s ' % self.repr
        return j.join(brackets(a) for a in self.args)

class OR(AND):

    repr = 'or'
    func = staticmethod(log_or)

    def __init__(self, *args):
        self.args = tuple(
            x for x in  chain(*(a.disjuncts() for a in args))
            if not isinstance(x, Const)
        )

    def conjuncts(self):
        return self,

    def disjuncts(self):
        return self.args





class Getattr(Abstract):

    def __init__(self, obj, attr):
        self.obj = obj
        self.attr = attr

    hashcmp = binding.Make(lambda s:(s.__class__, s.obj.hashcmp, s.attr))

    relvars = aggregates = joins = binding.Delegate('obj')

    def __repr__(self):
        return '%s.%s' % (brackets(self.obj), self.attr)





class Tuple(ExprBase):

    func = tuple

    def __repr__(self):
        return '(%s)' % ''.join(repr(a)+',' for a in self.args)

    def namespace(self):
        ns = {}
        for a in self.args:
            ns.update(a.namespace)
        return ns
    namespace = binding.Make(namespace)

class List(Tuple):

    func = list

    def __repr__(self):
        return '[%s]' % ','.join(repr(a) for a in self.args)

    namespace = {}

class Dict(Tuple):

    func = dict
    commut = True

    def __repr__(self):
        return '{%s}' % ','.join(repr(a) for a in self.args)

    hashcmp = binding.Make(
        lambda s:(s.__class__,tuple(sorted(a.hashcmp for a in s.args)))
    )

    namespace = {}

class _NoExpr(Abstract):
    '''Empty expression that is always False'''

    def __repr__(self):
        return 'NoExpr'

    def __nonzero__(self):
        return False

_NoExpr.hashcmp = _NoExpr
NoExpr = _NoExpr()



class Call(ExprBase):

    def __init__(self, func, args=(), kw={}):
        self.func = func
        self.args = args
        self.kw = kw

    hashcmp = binding.Make(lambda s: (
            s.__class__, s.func,
            tuple(a.hashcmp for a in s.args),
            tuple((k,v.hashcmp) for k,v in s.kw.iteritems())
    ))

    def __repr__(self):
        a = []
        if self.args:
            a.append(','.join(repr(a) for a in self.args))
        for k,v in self.kw.iteritems():
            a.append('%s=%s' % (k, repr(v)))
        return '%s(%s)' % (self.func, ','.join(a))



class Cast(Abstract):

    def __init__(self, obj, typ):
        self.obj = obj
        self.typ = typ

    hashcmp = binding.Make(lambda s:(s.__class__, s.obj.hashcmp, s.typ))

    relvars = aggregates = joins = binding.Delegate('obj')

    def __repr__(self):
        return '%s::%s' % (brackets(self.obj), self.typ)



class SliceObj(ExprBase):

    def __init__(self, start, stop, stride):
        self.args = start, stop, stride

    def __repr__(self):
        return '%r:%r:%r' % self.args


class Getitem(Abstract):

    def __init__(self, obj, idx):
        self.obj = obj
        self.idx = idx

    hashcmp = binding.Make(lambda s:(s.__class__, s.obj.hashcmp, s.idx.hashcmp))

    relvars = aggregates = joins = typ = binding.Delegate('obj')

    def __repr__(self):
        return '%s[%s]' % (brackets(self.obj), brackets(self.idx))



class Slice(Abstract):

    def __init__(self, obj, start, stop, stride):
        self.obj = obj
        self.start = start
        self.stop = stop
        self.stride = stride

    hashcmp = binding.Make(lambda s:(
                s.__class__, s.obj.hashcmp,
                s.start.hashcmp, s.stop.hashcmp, s.stride.hashcmp
    ))

    relvars = aggregates = joins = typ = binding.Delegate('obj')

    def __repr__(self):
        return '%s[%s:%s:%s]' % (
            brackets(self.obj),
            brackets(self.start),
            brackets(self.stop),
            brackets(self.stride)
        )


class Case(ExprBase):

    def __init__(self, *conds, **kwds):
        self.conds = []
        args = []
        for ce in conds:
            if not isinstance(ce, Tuple):
                raise SyntaxError('tuple required')
            c, e = ce.args
            self.conds.append((c,e))
            args.append(ce)
        self.default = kwds.get('default')
        args.append(self.default)
        super(Case, self).__init__(*args)
