'''Server side procedures'''

from wm4.core import *
from expressions import Abstract, ExprBase, NoExpr
from itertools import chain


__all__ = [
    'Aggregate', 'Count', 'Min', 'Max', 'Avg', 'Sum',
    'Procedure', 'Length', 'DatePart', 'Substr',
    'Lower', 'Upper',
    'SeqFunc', 'NextVal', 'CurVal', 'SetVal',
    'Coalesce', 'Random', 'Least', 'Greatest',
]



# Aggregates

class Aggregate(Abstract):

    def __init__(self, expr):
        self.expr = expr

    hashcmp = binding.Make(lambda s: (s.__class__, s.expr.hashcmp))

    relvars = joins = binding.Delegate('expr')

    aggregates = binding.Make(lambda s: set((s,)))


class Count(Aggregate):

    typ = int

    def __init__(self, expr=NoExpr):
        self.expr = expr

class TypedAggregate(Aggregate):
    typ = binding.Make(lambda s: s.expr.typ)

class Min(TypedAggregate): pass
class Max(TypedAggregate): pass
class Sum(TypedAggregate): pass

class Avg(Aggregate):
    typ = float




# Procedure base

class Procedure(Abstract):

    hashcmp = binding.Make(
            lambda s:(s.__class__, tuple(a.hashcmp for a in s.args))
    )

    def relvars(self):
        rv = filter(None, (a.relvars for a in self.args))
        if len(rv) == 1:
            return rv[0]
        elif rv:
            c = rv[0].copy()
            for d in rv[1:]:
                c.update(d)
            return c
        else:
            return {}
    relvars = binding.Make(relvars)

    def joins(self, builder):
        for a in self.args:
            for j in a.joins(builder):
                yield j

    aggregates = binding.Make(
            lambda s: set(chain(*(a.aggregates for a in s.args)))
    )


class Length(Procedure):
    typ = int
    def __init__(self, expr):
        self.args = expr,
    def __repr__(self):
        return 'Length(%s)' % self.args[0]

class DatePart(Procedure):

    def __init__(self, obj, part):
        self.args = obj, part


class Substr(Procedure):
    typ = binding.Make(lambda s: s.args[0].typ)

    def __init__(self, expr, start, stop):
        self.args = expr, start, stop

    def __repr__(self):
        return 'Substr(%s, %s, %s)' % self.args

class Lower(Procedure):
    typ = binding.Make(lambda s: s.args[0].typ)

    def __init__(self, expr):
        self.args = expr,

class Upper(Lower):
    pass

# Sequences

class SeqFunc(Procedure):
    pass

class NextVal(SeqFunc):

    def __init__(self, seq):
        self.args = seq,

class CurVal(SeqFunc):

    def __init__(self, seq):
        self.args = seq,

class SetVal(SeqFunc):

    def __init__(self, seq, val):
        self.args = seq, val



class Coalesce(Procedure):
    def __init__(self, *args):
        self.args = args
    def __repr__(self):
        return 'Coalesce(%s)' % repr(self.args)

class Least(Procedure):
    def __init__(self, *args):
        self.args = args
    def __repr__(self):
        return 'Least(%s)' % repr(self.args)

class Greatest(Procedure):
    def __init__(self, *args):
        self.args = args
    def __repr__(self):
        return 'Greatest(%s)' % repr(self.args)


class Random(Procedure):
    def __init__(self):
        self.args = ()
    def __repr__(self):
        return 'Random()'
