from wm4.api import *
import sys

from expressions import *
from interfaces import *
from procedures import Count

__all__ = [
    'query', 'Query', 'QueryModifier', 'As',
    'BaseSelect', 'Select', 'Insert', 'Update', 'UpdateQuery',
    'Delete', 'DeleteQuery', 'Distinct',
    'Join', 'LeftJoin', 'RightJoin',
    'SharedLock', 'ExclusiveLock',
    'OrderBy', 'Asc', 'Desc', 'NullsLast',
    'GroupBy', 'Having',
]

class query:

    protocols.advise(instancesProvide=[IQuerySrc])

    def __init__(self, src, namespaces=None):
        self.source = src
        if namespaces is None:
            frm = sys._getframe(1)
            namespaces = frm.f_locals, frm.f_globals, frm.f_builtins
        self.namespaces = namespaces


class Query(Abstract):

    protocols.advise(instancesProvide=[IQuery])


class QueryModifier(Abstract):
    pass


class As(Abstract):

    def __init__(self, expr, alias):
        self.expr = expr
        if not isinstance(alias, Const):
            raise ValueError('As requires string constant for alias')
        alias = alias.value
        if not isinstance(alias, basestring):
            raise ValueError('As requires string constant for alias')
        self.alias = alias

    hashcmp = binding.Make(lambda s: (s.__class__, s.expr.hashcmp, s.alias))

    namespace = binding.Make(lambda s: {s.alias:s.expr})


class BaseSelect(Query):

    def __init__(self, expr, for_items=()):
        self.expr = expr
        self.for_items = for_items

    def hashcmp(self):
        return (self.__class__, self.expr.hashcmp, tuple(
                    (a, i.hashcmp, c and c.hashcmp or None)
                    for a,i,c in self.for_items
        ))
    hashcmp = binding.Make(hashcmp)

    def buildSelect(self, builder):
        for_items = self.for_items
        lock = None
        expr = self.expr
        distinct = False
        if isinstance(expr, SharedLock):
            lock = expr.__class__
            expr = expr.expr
        elif isinstance(expr, Distinct):
            distinct = True
            expr = Tuple(*expr.args)

        implicitGroup = None
        if expr.aggregates:
            if isinstance(expr, Tuple):
                gb = [o for o in expr.args if not o.aggregates]
                if gb:
                    implicitGroup = GroupBy(*gb)

        where   = []
        orderBy = []
        groupBy = []
        having  = []

        aidx = dict((v,i) for i,(v,_,_) in enumerate(for_items))
        joins = {}

        for (var, for_item, cond) in for_items:
            joins[var] = []
            if not cond:
                continue
            for j in cond.conjuncts():
                if isinstance(j, Join):
                    # XXX join condition must refer only to two relations
                    #     j.relvars should return exactly two consecutive
                    #     relations
                    r = sorted(
                        (aidx.get(v,-1),v)
                        for v in set(v for v in j.relvars)
                    )[-1][1]
                    assert r in joins
                    joins[r].append(j)
                elif isinstance(j, OrderBy):
                    orderBy.extend(j.args)
                elif isinstance(j, GroupBy):
                    groupBy.extend(j.args)
                elif isinstance(j, Having):
                    having.append(j.expr)
                elif isinstance(j, QueryModifier):
                    raise NotImplementedError
                elif j.aggregates:
                    having.append(j)
                else:
                    where.append(j)

        relations = []
        for (var, for_item, cond) in for_items:
            j = joins[var]
            if not j:
                relations.append((var, for_item, None, None))
            elif len(j) > 1:
                raise TypeError("More than one join for", var, j)
            else:
                relations.append((var, for_item, j[0], None))

        where   = where and AND(*where) or NoExpr
        groupBy = groupBy and GroupBy(*groupBy) or implicitGroup or NoExpr
        orderBy = orderBy and OrderBy(*orderBy) or NoExpr
        having  = having and AND(*having) or NoExpr

        # add generated relvars
        assignments = dict((v,f) for v,f,_ in for_items)
        for obj in (expr, where, groupBy, orderBy, having):
            for a,r,j,k in obj.joins(builder):
                if a not in assignments:
                    assignments[a] = r
                    relations.append((a,r,j,k))
        return Select(
            expr, relations,
            where   = where,
            groupBy = groupBy,
            orderBy = orderBy,
            having  = having,
            lock = lock,
            distinct = distinct,
        )



class Select(BaseSelect):

    protocols.advise(
        instancesProvide = [IQuery],
        asAdapterForProtocols = [IExpression],
    )

    def __init__(
        self, expr, relations=(),
        where=NoExpr, groupBy=NoExpr, having=NoExpr,
        orderBy=NoExpr, limit=NoExpr, offset=NoExpr,
        lock=None, distinct=False,
    ):
        self.expr = expr
        self.relations = relations

        self.where = where
        self.orderBy = orderBy
        self.groupBy = groupBy
        self.having = having

        self.limit = limit
        self.offset = offset
        self.lock = lock
        self.distinct = distinct

    def copy(self):
        return self.__class__(
            self.expr, self.relations,
            self.where, self.groupBy, self.having,
            self.orderBy, self.limit, self.offset,
            self.lock, self.distinct,
        )

    def hashcmp(self):
        return (
            self.__class__, self.expr.hashcmp,
            tuple(
                (a, i.hashcmp, c and c.hashcmp or None, k)
                for a,i,c,k in self.relations
            ),
            self.where and self.where.hashcmp or None,
            self.groupBy and self.groupBy.hashcmp or None,
            self.having and self.having.hashcmp or None,
            self.orderBy and self.orderBy.hashcmp or None,
            self.limit, self.offset, self.lock, self.distinct,
        )
    hashcmp = binding.Make(hashcmp)

    def namespace(self):
        ns = dict((a,i) for a,i,c,k in self.relations)
        ns.update(self.expr.namespace)
        return ns
    namespace = binding.Make(namespace)

    def getExpr(self, path):
        ps = path.split('.')
        expr = self.namespace[ps[0]]
        rest = ps[1:]
        if rest:
            con = settings.storage_connection()
            for a in rest:
                expr = con.transform(Getattr(expr,a))
        return expr

    def parseSubexpr(self, src, frame=None):
        if frame is None:
            frame = sys._getframe(1)
        ns = [self.namespace, frame.f_locals, frame.f_globals, frame.f_builtins]
        con = settings.storage_connection()
        from builder import parse
        return parse(src, con, *ns)

    def allConditions(self):
        conds = []
        for c in (self.where, self.having, self.groupBy, self.orderBy):
            if c: conds.append(c)
        return conds

    def withExpr(self, con, expr):
        # XXX check expr relvars
        for_items = [(v,i,c) for v,i,c,k in self.relations if k is None]
        conds = self.allConditions()
        lv, li, lc = for_items[-1]
        if lc:
            conds.append(lc)
        if conds:
            conds = AND(*conds)
        else:
            conds = None
        for_items[-1] = (lv, li, conds)
        return con.transform(BaseSelect(expr, for_items))

    def withCondition(self, con, expr):
        # XXX check expr relvars
        for_items = [(v,i,c) for v,i,c,k in self.relations if k is None]
        conds = self.allConditions()
        lv, li, lc = for_items[-1]
        if lc:
            conds.append(lc)
        conds.append(expr)
        if conds:
            conds = AND(*conds)
        else:
            conds = None
        for_items[-1] = (lv, li, conds)
        sel = self.expr
        if self.lock:
            sel = self.lock(sel)
        q = con.transform(BaseSelect(sel, for_items))
        q.distinct = self.distinct
        return q

    def asCount(self, con):
        if self.offset or self.limit:
            raise NotImplementedError(
                'query with offset or limit')
        if self.groupBy or self.having or self.distinct:
            from builder import parse
            cp = '[Count() for a in self]'
            return parse(cp, con, {'self':self}, globals())
        else:
            cp = self.withExpr(con, Count())
            cp.orderBy = NoExpr
        return cp

    def __repr__(self):
        r = ['<SELECT ', repr(self.expr)]
        if self.relations:
            r.append(' FROM')
            for v,f,c,k in self.relations:
                if c:
                    r.append(' %r ON %s' % (f, repr(c)))
                else:
                    r.append(' %r' % f)
        for k,e in [
            (' WHERE ', self.where),   (' GROUP ', self.groupBy),
            (' HAVING ', self.having), (' ORDER ', self.orderBy),
            (' LOCK ', self.lock),
        ]:
            if e:
                r.append(k)
                r.append(repr(e))
        r.append('>')
        return ''.join(r)













class Insert(Query):

    def __init__(self, relation, **values):
        self.rel = IExpression(relation)
        self.values = dict((a,IExpression(v)) for a,v in values.iteritems())

    hashcmp = binding.Make(lambda s:(
            s.__class__, s.rel.hashcmp,
            tuple(sorted((a,v.hashcmp) for a,v in s.values.iteritems())),
    ))

    def __repr__(self):
        return 'Insert(%r, %s)' % (
            self.rel,
            ', '.join('%s=%r' % (a,e) for a,e in self.values.iteritems()),
        )


class Update(QueryModifier):
    def __init__(self, rel, **values):
        self.rel    = rel
        self.values = values

class UpdateQuery(Query):

    def __init__(self, rel, values, where=NoExpr):
        self.rel = rel
        self.values = values
        self.where = where

    def hashcmp(self):
        return (
            self.__class__, self.rel.hashcmp,
            tuple(sorted((a,v.hashcmp) for a,v in self.values.iteritems())),
            self.where.hashcmp,
        )
    hashcmp = binding.Make(hashcmp)

    def __repr__(self):
        return '<UPDATE %r SET %s WHERE %s>' % (
            self.rel,
            ','.join('%s=%r' % (k,v) for k,v in self.values.iteritems()),
            self.where
        )



class Delete(QueryModifier):
    def __init__(self, rel):
        self.rel = rel

class DeleteQuery(Query):

    def __init__(self, rel, where=NoExpr):
        self.rel = rel
        self.where = where

    hashcmp = binding.Make(lambda s: (
                s.__class__, s.rel.hashcmp, s.where and s.where.hashcmp or None
    ))




class Distinct(QueryModifier):

    def __init__(self, *args):
        self.args = args

    hashcmp = binding.Make(
                lambda s: (s.__class__,tuple(a.hashcmp for a in self.args)))




class Join(QueryModifier):

    def __init__(self, cond):
        self.cond = cond

    hashcmp = binding.Make(lambda s: (s.__class__, s.cond.hashcmp))

    relvars = joins = binding.Delegate('cond')

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, repr(self.cond))

class LeftJoin(Join):
    pass

class RightJoin(Join):
    pass




class SharedLock(QueryModifier):

    def __init__(self, expr):
        self.expr = expr

    hashcmp = binding.Make(lambda s: (s.__class__, s.expr.hashcmp))

    relvars = joins = binding.Delegate('expr')

class ExclusiveLock(SharedLock):
    pass



class OrderBy(ExprBase, QueryModifier):
    pass


class Ordering(QueryModifier):

    def __init__(self, expr):
        self.expr = expr

    hashcmp = binding.Make(lambda s: (s.__class__, s.expr.hashcmp))
    relvars = joins = binding.Delegate('expr')

class Asc(Ordering):
    repr = '<'

class Desc(Ordering):
    repr = '>'

class NullsLast(Ordering):
    repr = 'NULLS LAST'


class GroupBy(ExprBase, QueryModifier):
    pass

class Having(QueryModifier):
    def __init__(self, expr):
        self.expr = expr

    hashcmp = binding.Make(lambda s: (s.__class__, s.expr.hashcmp))

    relvars = joins = binding.Delegate('expr')
