# -*- coding: ISO-8859-2 -*-

from unittest import *
import sys
from wm4.api import *
from wm4.query.api import *
from wm4.query.builder import *

from app import *

class TestBase(TestCase):

    def setUp(self):
        self.state = config.snapshot()
        config.loadConfig()
        settings.storage_url('mockdb:')
        self.con = settings.storage_connection()

    def tearDown(self):
        config.swap(self.state)

class TestTypeConvertors(TestBase):

    def F(self, typ, val):
        conv = self.con.convertorFor(typ)
        return conv.fromSql(val)

    def T(self, typ, val):
        conv = self.con.convertorFor(typ)
        return conv.toSql(val)

    def test10_String(self):
        self.assertEqual(self.T(str, None), "NULL")
        self.assertEqual(self.T(str, 'foo'), "'foo'")
        self.assertEqual(self.F(str, 'foo'), "foo")
        self.assertEqual(self.F(str, None), None)

    def test12_Unicode(self):
        self.assertEqual(self.T(unicode, None), "NULL")
        self.assertEqual(self.T(unicode, u'etina'), "'\xc4\x8de\xc5\xa1tina'")
        self.assertEqual(self.F(unicode, '\xc4\x8de\xc5\xa1tina'), u"etina")
        self.assertEqual(self.F(unicode, None), None)

    def test20_Integer(self):
        self.assertEqual(self.T(int, None), "NULL")
        self.assertEqual(self.T(int, 42), "42")
        self.assertEqual(self.F(int, '42'), 42)
        self.assertEqual(self.F(int, 42), 42)
        self.assertEqual(self.F(int, None), None)

    def test22_MyInteger(self):
        class myint(int): pass
        self.assertEqual(self.T(myint, None), "NULL")
        self.assertEqual(self.T(myint, myint(42)), "42")
        v = self.F(myint, '42')
        self.assertEqual(v, 42)
        assert isinstance(v, myint), type(v)
        self.assertEqual(self.F(myint, 42), 42)
        self.assertEqual(self.F(myint, None), None)

    def test24_Float(self):
        self.assertEqual(self.T(float, None), "NULL")
        self.assertEqual(self.T(float, 42.1), "42.1")
        self.assertEqual(self.F(float, '42.1'), 42.1)
        self.assertEqual(self.F(float, 42.1), 42.1)
        self.assertEqual(self.F(float, None), None)

    def test30_Bool(self):
        self.assertEqual(self.T(bool, None), "NULL")
        self.assertEqual(self.T(bool, True), "'t'")
        self.assertEqual(self.T(bool, False), "'f'")
        self.assertEqual(self.T(bool, 1), "'t'")
        self.assertEqual(self.T(bool, 0), "'f'")
        self.assertEqual(self.F(bool, 't'), True)
        self.assertEqual(self.F(bool, 'true'), True)
        self.assertEqual(self.F(bool, True), True)
        self.assertEqual(self.F(bool, 'f'), False)
        self.assertEqual(self.F(bool, None), None)



class TestRowConvertors(TestBase):

    def C(self, expr):
        f = sys._getframe(1)
        qry = parse(expr, self.con, f.f_locals, f.f_globals, f.f_builtins)
        qry = IQuery(qry)
        return self.con.row_convertor(qry.expr)

    def test10_Expression(self):
        c = self.C('[u.uid for u in User]')
        self.assertEqual(c(iter([42])), 42)

    def test20_Tuple(self):
        c = self.C('[(u.uid,u.login) for u in User]')
        self.assertEqual(c(iter([42,'john'])), (42,'john'))

    def test30_RelVar(self):
        c = self.C('[u for u in User]')
        o = c(iter([42,'John','john', None]))
        assert isinstance(o, User)
        self.assertEqual(o.uid, 42)
        self.assertEqual(o.name, u'John')
        self.assertEqual(o.login, 'john')
        self.assertEqual(o.mtime, None)




class TestRendering(TestBase):

    def p(self, expr, up=1):
        f = sys._getframe(up)
        return parse(
            expr, self.con, f.f_locals, f.f_globals, f.f_builtins
        )

    def q(self, expr):
        return IQuery(self.p(expr, up=2))

    def Q(self, expr):
        f = sys._getframe(1)
        qry = IQuery(
            parse(expr, self.con, f.f_locals, f.f_globals, f.f_builtins)
        )
        return self.con.render(qry)

    def test10_SimpleScalar(self):
        self.assertEqual(
            self.Q('[u.uid for u in User]'),
            'SELECT uid FROM "user"'
        )

    def test11_SimpleEntity(self):
        self.assertEqual(
            self.Q('[u for u in User]'),
            'SELECT uid, name, login, mtime FROM "user"'
        )

    def test12_SimpleWithTypeRef(self):
        uref = 'wm4.query.tests.app.User'
        self.assertEqual(
            self.Q('[u.uid for u in schema.TypeRef(uref)()]'),
            'SELECT uid FROM "user"'
        )

    def test13_SelectTuple(self):
        self.assertEqual(
            self.Q('[(u.uid+2, u.login) for u in User]'),
            'SELECT uid + 2, login FROM "user"'
        )

    def test14_SelectNextVal(self):
        self.assertEqual(
            self.Q('NextVal("aaa_seq")'),
            "SELECT NEXTVAL('aaa_seq')",
        )

    def test20_WhereNotNull(self):
        self.assertEqual(
            self.Q('[u for u in User if u.login != None]'),
            'SELECT uid, name, login, mtime FROM "user" WHERE login NOTNULL'
        )

    def test20_WhereLike(self):
        self.assertEqual(
            self.Q('[u for u in User if Like(u.login,"b_")]'),
            'SELECT uid, name, login, mtime FROM "user" WHERE login LIKE \'b_\''
        )
        self.assertEqual(
            self.Q('[u for u in User if not Like(u.login,"b_")]'),
            'SELECT uid, name, login, mtime FROM "user" WHERE NOT '
            '(login LIKE \'b_\')'
        )

    def test20_WhereNeg(self):
        self.assertEqual(
            self.Q('[u for u in User if -u.uid == -10]'),
            'SELECT uid, name, login, mtime FROM "user" WHERE (- uid) = -10'
        )

    def test21_InSubselect(self):
        self.assertEqual(
            self.Q('[u for u in User if u in [g.user for g in UserGroup]]'),
            'SELECT uid, name, login, mtime FROM "user" '
            'WHERE uid IN (SELECT "user" FROM user_group)'
        )

    def test22_ComplexCondition(self):
        self.assertEqual(
            self.Q('[u for u in User if (u.uid+1)*2 == 42]'),
            'SELECT uid, name, login, mtime FROM "user" '
            "WHERE ((uid + 1) * 2) = 42"
        )

    def test23_Case(self):
        self.assertEqual(
            self.Q('[Case((u.uid == 1, 42), (u.uid > 1, 43), default=41) for u in User]'),
            'SELECT CASE WHEN uid = 1 THEN 42 WHEN uid > 1 THEN 43 ELSE 41 END FROM "user"'
        )
        self.assertEqual(
            self.Q('[Case((u.uid == 1, 42), default=None) for u in User]'),
            'SELECT CASE WHEN uid = 1 THEN 42 ELSE NULL END FROM "user"'
        )


    def test30_CrossJoin(self):
        self.assertEqual(
            self.Q('[u for u in User for ug in UserGroup]'),
            'SELECT u.uid, u.name, u.login, u.mtime '
            'FROM "user" AS u CROSS JOIN user_group AS ug'
        )

    def test31_NaturalJoin(self):
        self.assertEqual(
            self.Q('''[
                u for u in User for ug in UserGroup if Join(u==ug.user)
            ]'''),
            'SELECT u.uid, u.name, u.login, u.mtime '
            'FROM "user" AS u '
            'INNER JOIN user_group AS ug ON (u.uid = ug."user")'
        )

    def test32_LeftJoin(self):
        self.assertEqual(
            self.Q('''[
                u for u in User for ug in UserGroup if LeftJoin(u==ug.user)
            ]'''),
            'SELECT u.uid, u.name, u.login, u.mtime '
            'FROM "user" AS u '
            'LEFT JOIN user_group AS ug ON (u.uid = ug."user")'
        )

    def test39_Lock(self):
        self.assertEqual(
            self.Q('[ExclusiveLock(u) for u in User]'),
            'SELECT uid, name, login, mtime FROM "user" FOR UPDATE'
        )
        self.assertEqual(
            self.Q('[SharedLock(u) for u in User]'),
            'SELECT uid, name, login, mtime FROM "user" FOR SHARE'
        )

    def test3F_Distinct(self):
        self.assertEqual(
            self.Q('[Distinct(u.name) for u in User]'),
            'SELECT DISTINCT name FROM "user"'
        )
        self.assertEqual(
            self.Q('[Distinct(u.login, u.name) for u in User]'),
            'SELECT DISTINCT login, name FROM "user"'
        )

    def test3G_DistinctAsCount(self):
        q = self.q('[Distinct(u.name) for u in User]').asCount(self.con)
        self.assertEqual(
            self.con.render(q),
            'SELECT COUNT(*) FROM (SELECT DISTINCT name FROM "user") AS a (s0)'
        )

    def test40_OrderBy(self):
        self.assertEqual(
            self.Q('[u for u in User if OrderBy(u.login,Desc(u.uid))]'),
            'SELECT uid, name, login, mtime '
            'FROM "user" ORDER BY login, uid DESC'
        )

    def test50_GroupByWithCountAll(self):
        self.assertEqual(
            self.Q('[(u.uid,Count()) for u in User]'),
            'SELECT uid, COUNT(*) FROM "user" GROUP BY uid'
        )

    def test51_GroupByWithCountEnt(self):
        self.assertEqual(
            self.Q('[(u.uid,Count(u)) for u in User]'),
            'SELECT uid, COUNT(*) FROM "user" GROUP BY uid'
        )

    def test52_GroupByWithCountVal(self):
        self.assertEqual(
            self.Q('[(u.login,Count(u.uid)) for u in User]'),
            'SELECT login, COUNT(uid) FROM "user" GROUP BY login'
        )

    def test53_GroupByWithGeneralAggregate(self):
        self.assertEqual(
            self.Q('[(u.login,Max(u.uid)) for u in User]'),
            'SELECT login, MAX(uid) FROM "user" GROUP BY login'
        )

    def test54_GroupByAll(self):
        self.assertEqual(
            self.Q('[(u,Max(u.uid)) for u in User]'),
            'SELECT uid, name, login, mtime, MAX(uid) FROM "user" '
            'GROUP BY uid, name, login, mtime'
        )

    def test55_ImplicitHaving(self):
        self.assertEqual(
            self.Q('[(u,Max(u.uid)) for u in User if Max(u.uid)>100]'),
            'SELECT uid, name, login, mtime, MAX(uid) '
            'FROM "user" '
            'GROUP BY uid, name, login, mtime '
            "HAVING MAX(uid) > 100"
        )

    def test56_ExplicitHaving(self):
        self.assertEqual(
            self.Q('[(u,Max(u.uid)) for u in User if Having(Max(u.uid)>100)]'),
            'SELECT uid, name, login, mtime, MAX(uid) '
            'FROM "user" '
            'GROUP BY uid, name, login, mtime '
            "HAVING MAX(uid) > 100"
        )

    def test57_AggregatesAsCount(self):
        q = self.q(
            '[(u,Max(u.uid)) for u in User if Having(Max(u.uid)>100)]'
        ).asCount(self.con)
        self.assertEqual(
            self.con.render(q),
            'SELECT COUNT(*) FROM ('
                'SELECT uid, name, login, mtime, MAX(uid) '
                'FROM "user" '
                'GROUP BY uid, name, login, mtime '
                "HAVING MAX(uid) > 100"
            ') AS a (s0, s1)'
        )

    def test60_UpdateAll(self):
        self.assertEqual(
            self.Q('[Update(u, uid=u.uid+1) for u in User]'),
            'UPDATE "user" SET uid=uid + 1',
        )

    def test61_UpdateWhere(self):
        self.assertEqual(
            self.Q('[Update(u, uid=u.uid+1) for u in User if u.uid < 10]'),
            'UPDATE "user" SET uid=uid + 1 WHERE uid < 10',
        )

    def test70_DeleteAll(self):
        self.assertEqual(
            self.Q('[Delete(u) for u in User]'),
            'DELETE FROM "user"',
        )

    def test71_DeleteWhere(self):
        self.assertEqual(
            self.Q('[Delete(u) for u in User if u.uid == 0]'),
            'DELETE FROM "user" WHERE uid = 0',
        )

    def test80_Insert(self):
        self.assertEqual(
            self.Q("Insert(Group, gid=1, name='foo')"),
            'INSERT INTO "group" (gid, name) VALUES (1, \'foo\')'
        )

    def test90_Subselect(self):
        self.assertEqual(
            self.Q('[a for a in [u for u in User]]'),
            'SELECT a.uid, a.name, a.login, a.mtime '
            'FROM (SELECT uid, name, login, mtime FROM "user") '
            'AS a (uid, name, login, mtime)'
        )

    def test91_SubselectWithSelectExpr(self):
        self.assertEqual(
            self.Q('[a for a in [(u.login,Max(u.uid)) for u in User]]'),
            'SELECT a.s0, a.s1 '
            'FROM (SELECT login, MAX(uid) FROM "user" GROUP BY login) '
            'AS a (s0, s1)'
        )

    def test92_SubselectWithLeftJoin(self):
        self.assertEqual(
            self.Q('''[a for a in [
                u for u in User for ug in UserGroup if LeftJoin(u==ug.user)
            ]]'''),
            'SELECT a.uid, a.name, a.login, a.mtime '
            'FROM ('
                'SELECT u.uid, u.name, u.login, u.mtime '
                'FROM "user" AS u '
                'LEFT JOIN user_group AS ug ON (u.uid = ug."user")'
            ') AS a (uid, name, login, mtime)'
        )

    def test93_SubselectWithAggregation(self):
        self.assertEqual(
            self.Q('''[Count() for a in [
                (u,Max(u.uid)) for u in User if Having(Max(u.uid)>100)
            ]]'''),
            'SELECT COUNT(*) FROM ('
                'SELECT uid, name, login, mtime, MAX(uid) '
                'FROM "user" '
                'GROUP BY uid, name, login, mtime '
                'HAVING MAX(uid) > 100'
            ') AS a (s0, s1)'
        )

    def XXXtest94_SubselectTupleUnpack(self):
        self.assertEqual(
            self.Q('''[
                (l,m) for l,m in [
                    (u.login, Max(u.uid)) for u in User
                ]]'''),
            'SELECT a.s0, a.s1 '
            'FROM (SELECT login, MAX(uid) FROM "user" GROUP BY login) '
            'AS a (s0, s1)'
        )
    def XXXtest_Join(self):
        pass
        # Tohle se renderuje spatne XXX
        #
        #   [(ug, ug.user, ug.group) for ug in UserGroup]
        #
        # SELECT
        #   ug."user", ug."group", "ug.user".uid, "ug.user".login,
        #   "ug.user".nick, "ug.user".password, "ug.user".name,
        #   "ug.user".email, "ug.user".phone, "ug.user".mobil,
        #   "ug.user".workplace, "ug.user".occupation, "ug.user".note,
        #   "ug.user".kind, "ug.user".active, "ug.user".hidden,
        #
        #   "ug.group".gid, "ug.group".name, "ug.group".description,
        #   "ug.group".kind, "ug.group".active, "ug.group".hidden
        #
        # FROM user_group AS ug
        #   LEFT JOIN "user" AS "ug.user" ON (ug."user" = ug."user")
        #   LEFT JOIN "group" AS "ug.group" ON (ug."group" = ug."group")


def suite():
    return TestSuite((
        makeSuite(TestTypeConvertors),
        makeSuite(TestRowConvertors),
        makeSuite(TestRendering),
    ))

if __name__ == '__main__':
    main(defaultTest='suite')
