from wm4.core import *
import sys, logging, logging.handlers, thread, re
from logging import (
    FATAL, CRITICAL, ERROR, WARNING, WARN, INFO, DEBUG, _levelNames,
    Formatter
)

__all__ = [
    'File', 'Stream', 'Memory', 'Syslog', 'Sendmail', 'Null', 'Request',
    'FATAL', 'CRITICAL', 'ERROR', 'WARNING', 'WARN', 'INFO', 'DEBUG',
    'getLogger', 'logging_service', 'LoggingService', 'LoggerURL',

]


# TODO reload, restart, ...
# see http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/474089
# http://www.simplistix.co.uk/software/python/mailinglogger
#
# Custom formatter with extra fields
# http://maxischenko.in.ua/blog/entries/132/python-logging-hack-custom-formatters/




def HANDLER(logger, default=NOT_GIVEN):
    return config.Property('wm4.logs.%s' % logger, default=default)

HANDLER('*', lambda reg,found: Stream())






class File(logging.FileHandler):

    def __init__(self, filename, level=WARNING, formatter=None):
        logging.FileHandler.__init__(self, filename)
        self.setLevel(level)
        if formatter is not None:
            self.setFormatter(formatter)


class Stream(logging.StreamHandler):

    def __init__(self, stream=sys.stderr, level=WARNING, formatter=None):
        logging.StreamHandler.__init__(self, stream)
        self.setLevel(level)
        if formatter is not None:
            self.setFormatter(formatter)


class Memory(logging.Handler):

    capacity = 0
    buffer = []
    clock = thread.allocate_lock()

    def __init__(self, level=WARNING, formatter=None):
        logging.Handler.__init__(self, level=level)
        if formatter is not None:
            self.setFormatter(formatter)

    def _getLog (klass, level=logging.DEBUG) :
        klass.clock.acquire()
        try :
            b = []
            for r,m in klass.buffer :
                if r.levelno >= level :
                    b.append(m)
            return b
        finally :
            klass.clock.release()
    getLog = classmethod(_getLog)

    def _clearLog (klass) :
        klass.clock.acquire()
        try :
            klass.buffer[:] = []
        finally :
            klass.clock.release()
    clearLog = classmethod(_clearLog)

    def _setCapacity (klass, cap) :
        klass.capacity = cap
    setCapacity = classmethod(_setCapacity)

    def emit(self, r):
        t = self.format(r)
        b = self.__class__.buffer
        if self.capacity and len(b) >= self.capacity :
            del b[0]
        b.append((r, t))
        r.exc_info = None

LOGREC_ARGS = logging.LogRecord.__init__.im_func.func_code.co_argcount

class Request(logging.Handler):

    def __init__(
        self, level=DEBUG, emitLevel=ERROR, emitHandler=None, accumulate=True,
        formatter=None
    ):
        logging.Handler.__init__(self, level=level)
        if formatter is not None:
            self.setFormatter(formatter)
        from wm4.web.interfaces import request
        self.get_req = request
        self.emitLevel = emitLevel
        self.emitHandler = emitHandler
        self.accumulate = accumulate

    def emit(self, r):
        req = self.get_req()
        if r.levelno < self.emitLevel:
            if req is not None:
                req.req_log.append(r)
        else:
            if self.emitHandler is not None:
                if req is not None:
                    req.req_log.append(r)
                    if self.accumulate:
                        t = '\n'.join(
                            self.format(o).rstrip() for o in req.req_log
                        )
                        if LOGREC_ARGS == 8:
                            n = logging.LogRecord(
                                r.name, r.levelno, r.pathname, r.lineno,
                                t, (), None)
                        else:
                            n = logging.LogRecord(
                                r.name, r.levelno, r.pathname, r.lineno,
                                t, (), None, None)
                        self.emitHandler.emit(n)
                    else:
                        for o in req.req_log:
                            self.emitHandler.emit(o)
                    req.req_log[:] = []
                else:
                    self.emitHandler(r)

SYSLOG_UDP_PORT = 514
LOG_USER        = 1

class Syslog(logging.handlers.SysLogHandler):

    def __init__(
        self, address=('localhost', SYSLOG_UDP_PORT), facility=LOG_USER,
        level=WARNING, formatter=None
    ):
        logging.handlers.SysLogHandler.__init__(self, address, facility)
        self.setLevel(level)
        if formatter is not None:
            self.setFormatter(formatter)




# Enhanced http://cheeseshop.python.org/pypi/mailinglogger/3.0.0

class Sendmail(logging.handlers.SMTPHandler):

    def __init__(
       self, toaddr, subject, mailhost='localhost',
       fromaddr=None, level=WARNING, formatter=None
    ):
        if fromaddr is None:
            import socket, pwd
            fromaddr = '%s@%s' % (
                pwd.getpwuid(os.getuid()).pw_name,
                socket.getfqdn()
            )
        if isinstance(toaddr, basestring):
            toaddr = toaddr,
        logging.handlers.SMTPHandler.__init__(
            self, mailhost, fromaddr, toaddr, subject
        )
        self.setLevel(level)
        if formatter is not None:
            self.setFormatter(formatter)






class Null(logging.Handler):

    def emit(self, r):
        pass







class LoggingService(binding.Component):

    _loggers = binding.Make(dict)

    def getLogger(self, name):
        logger = self._loggers.get(name)
        if logger is None:
            logger = self._loggers[name] = logging.Logger(name)
            handlers = HANDLER(name)()
            if not isinstance(handlers, (tuple,list)):
                handlers = handlers,
            el = ERROR
            for h in handlers:
                el = min(el, h.level)
                logger.addHandler(h)
            logger.setLevel(el)
        return logger


logging_service = config.Resource(
    'logging_service', 'wm4.running.logs.LoggingService')



def getLogger(name):
    return logging_service().getLogger(name)


class LoggerURL(naming.AbstractURL):

    slots = 'logger',

    MATCH = re.compile(
        '^(?P<logger>[a-z][a-z0-9]*(\\.[a-z][a-z0-9]*)*)$', re.I
    ).match

    SCHEMES = set(['logger'])

    def __str__(self):
        return '%s:%s' % (self.scheme, self.logger)

    def getObject(self):
        return getLogger(self.logger)
