from wm4.core import *
import errno, os, sys, signal

__all__ = [
    'spawnFunc', 'ChildProcess', 'ProcessTemplate', 'SlaveInterpreter',
    'WorkerProcess', 'ProcessPool',
]


def spawnFunc(func, *args, **kwds):
    pid = os.fork()
    if pid == 0:
        ret = func(*args, **kwds)
        ret = int(ret)
        os._exit(ret) # sys.exit(ret) ???
    else:
        return ChildProcess(pid=pid)



class ChildProcess:
    pid = None

    isRunning = True
    isStopped = False

    exitStatus = None
    stoppedBecause = None
    exitedBecause = None

    def __init__(self, pid):
        self.pid = pid

    def __repr__(self):
        return '<ChildProcess pid=%s>' % self.pid

    def sendSignal(self, signum=1):
        if isinstance(signum, basestring):
            signum = getattr(signal, signum)
        os.kill(self.pid, signum)

    def checkStatus(self, block=False):
        if not self.isRunning:
            return
        flags = os.WCONTINUED | os.WUNTRACED
        if not block:
            flags |= os.WNOHANG
        try:
            p, s = os.waitpid(self.pid, flags)
        except OSError,v:
            #if v.args[0]==errno.ECHILD:
            # child was collected by someone else
            if v.args[0]==errno.EINTR:
                self.checkStatus()
            else:
                raise
        else:
            if p==self.pid:
                self._setStatus(s)

    def _setStatus(self,status=None):
        self.stoppedBecause = None
        self.isStopped = os.WIFSTOPPED(status)
        if os.WIFEXITED(status):
            self.exitStatus = os.WEXITSTATUS(status)
        if self.isStopped:
            self.stoppedBecause = os.WSTOPSIG(status)
        if os.WIFSIGNALED(status):
            self.exitedBecause = os.WTERMSIG(status)
        self.isRunning = self.exitedBecause is None and self.exitStatus is None




class ProcessTemplate(object):

    # default factory for process proxy
    proxyClass = ChildProcess

    # pipes for communication between parent and child process
    pipes  = ()

    def spawn(self):
        # Pipes for communication with child
        parentPipes = {}
        childPipes = {}
        for name in self.pipes:
            parentPipes[name], childPipes[name] = self._mkPipe()

        pid = os.fork()
        if pid:
            # Parent process
            for f in childPipes.values():
                f.close()
            del childPipes
            return self._makeProxy(pid, parentPipes)

        else:
            # Child process
            for f in parentPipes.values():
                f.close()
            del parentPipes
            for name, stream in childPipes.items():
                setattr(self, name, stream)
            self.run()

    def _mkPipe(self):
        import socket
        r, w = socket.socketpair(socket.AF_UNIX, socket.SOCK_STREAM)
        return r.makefile('r+'), w.makefile('r+')

    def _makeProxy(self, pid, pipes):
        proxy = self.proxyClass(pid=pid)
        for name, stream in pipes.items():
            setattr(proxy, name, stream)
        return proxy

    def run(self):
        raise NotImplementedError







class SlaveInterpreter(ProcessTemplate):

    pipes = 'commands',

    def __init__(self, namespace=None):
        if namespace is None:
            frame = sys._getframe(1)
            namespace = frame.f_globals.copy()
            namespace.update(frame.f_locals)
            namespace['self'] = self
        self.namespace = namespace

    def _makeProxy(self, pid, pipes):
        import pickle
        proxy = super(SlaveInterpreter, self)._makeProxy(pid, pipes)

        def quit():
            proxy.commands.write('quit 0\n')
            try:
                proxy.commands.flush()
            except IOError:
                # flush on broken pipe
                pass
            proxy.checkStatus(block=True)
        proxy.quit = quit

        def readresult():
            status, datalen = proxy.commands.readline().strip().split()
            datalen = int(datalen)
            if datalen == 0:
                data = None
            else:
                data = proxy.commands.read(datalen)
            if status == 'result':
                if data is None:
                    return None
                else:
                    return pickle.loads(data)
            elif status == 'exception':
                e = pickle.loads(data)
                raise e
            else:
                raise RuntimeError('unknown result', status)

        def eval(e):
            proxy.commands.write('eval %d\n%s' % (len(e), e))
            proxy.commands.flush()
            return readresult()
        proxy.eval = eval

        def suite(s):
            proxy.commands.write('exec %d\n%s' % (len(s),s))
            proxy.commands.flush()
            return readresult()
        proxy.suite = suite

        return proxy

    def run(self):
        import pickle
        while True:
            _cmd = self.commands.readline()
            if not _cmd:
                os._exit(0)

            cmd, srclen = _cmd.strip().split()
            srclen = int(srclen)
            if srclen:
                src = self.commands.read(srclen)

            if cmd == 'quit':
                os._exit(0)

            elif cmd == 'eval':
                try:
                    r = eval(src, self.namespace)
                except:
                    _,e,_ = sys.exc_info()
                    r = pickle.dumps(e)
                    self.commands.write('exception %d\n' % len(r))
                    self.commands.write(r)
                    self.commands.flush()
                else:
                    r = pickle.dumps(r)
                    self.commands.write('result %d\n' % len(r))
                    self.commands.write(r)
                    self.commands.flush()

            elif cmd == 'exec':
                try:
                    exec src in self.namespace
                except:
                    _,e,_ = sys.exc_info()
                    r = pickle.dumps(e)
                    self.commands.write('exception %d\n' % len(r))
                    self.commands.write(r)
                    self.commands.flush()
                else:
                    self.commands.write('result 0\n')
                    self.commands.flush()




class WorkerProcess(ProcessTemplate):

    pipes = 'commands',

    def run(self):
        pass
        # main loop

    def work(self):
        pass



class ProcessPool(binding.Component):

    template = binding.Require()

    busy = binding.Make(list)
    spare = binding.Make(list)

    start_procs = 0
    min_spare_procs = 0
    max_spare_procs = 0
    max_procs = sys.maxint

    def start(self):
        pass

    def stop(self):
        pass

    def checkStatus(self):
        for p in self.busy[:]:
            p.checkStatus()
            if not p.isRunning:
                self.busy.remove(p)

        for p in self.spare[:]:
            p.checkStatus()
            if not p.isRunning:
                self.spare.remove(p)

    def get(self):
        self.checkStatus()
        if self.spare:
            p = self.spare.pop()
        else:
            p = self.template().spawn()
        self.busy.append(p)
        return p
