from unittest import *
from wm4.running.process import *
import os, sys, signal, time, select

class TestChildProcess(TestCase):

    def setUp(self):
        self.children = []

    def tearDown(self):
        for pid in self.children:
            try:
                os.kill(pid, 9)
            except:
                pass

    def spawnFunc(self, func, *args, **kwds):
        p = spawnFunc(func, *args, **kwds)
        self.children.append(p.pid)
        return p

    def sleep(self, t=0.1):
        select.select([],[],[], t)

    def test10_Exited(self):
        p = self.spawnFunc(lambda: 3)
        p.checkStatus(True)
        self.assertEqual(p.isRunning, False)
        self.assertEqual(p.isStopped, False)
        self.assertEqual(p.stoppedBecause, None)
        self.assertEqual(p.exitedBecause, None)
        self.assertEqual(p.exitStatus, 3)

    def test12_Killed(self):
        def f():
            while True:
                pass
        p = self.spawnFunc(f)
        p.sendSignal('SIGKILL')
        p.checkStatus(True)
        self.assertEqual(p.isRunning, False)
        self.assertEqual(p.isStopped, False)
        self.assertEqual(p.stoppedBecause, None)
        self.assertEqual(p.exitedBecause, 9)
        self.assertEqual(p.exitStatus, None)

    def test14_Stopped(self):
        def f():
            while True:
                pass
        p = self.spawnFunc(f)
        p.sendSignal('SIGSTOP')
        self.sleep()
        p.checkStatus()
        self.assertEqual(p.isRunning, True)
        self.assertEqual(p.isStopped, True)
        self.assertEqual(p.stoppedBecause, 19)
        self.assertEqual(p.exitedBecause, None)
        self.assertEqual(p.exitStatus, None)


    def test20_Complex(self):
        def f():
            while True:
                pass
        p = self.spawnFunc(f)
        self.sleep()
        p.checkStatus()

        self.assertEqual(p.isRunning, True)
        self.assertEqual(p.isStopped, False)
        self.assertEqual(p.stoppedBecause, None)
        self.assertEqual(p.exitedBecause, None)
        self.assertEqual(p.exitStatus, None)

        p.sendSignal('SIGSTOP')
        self.sleep()
        p.checkStatus()

        self.assertEqual(p.isRunning, True)
        self.assertEqual(p.isStopped, True)
        self.assertEqual(p.stoppedBecause, 19)
        self.assertEqual(p.exitedBecause, None)
        self.assertEqual(p.exitStatus, None)

        p.sendSignal('SIGCONT')
        self.sleep()
        p.checkStatus()

        self.assertEqual(p.isRunning, True)
        self.assertEqual(p.isStopped, False)
        self.assertEqual(p.stoppedBecause, None)
        self.assertEqual(p.exitedBecause, None)
        self.assertEqual(p.exitStatus, None)

        p.sendSignal('SIGKILL')
        p.checkStatus(True)

        self.assertEqual(p.isRunning, False)
        self.assertEqual(p.isStopped, False)
        self.assertEqual(p.stoppedBecause, None)
        self.assertEqual(p.exitedBecause, 9)
        self.assertEqual(p.exitStatus, None)





class TestSlaveInterpreter(TestCase):

    def test10_Eval(self):
        def foo(): return 42
        p = SlaveInterpreter().spawn()
        self.assertEqual(p.eval('"foo"'), 'foo') # constant
        self.assertEqual(p.eval('signal.SIGKILL'), 9) # global
        self.assertEqual(p.eval('foo()'), 42) # local
        p.quit(); self.assertEqual(p.isRunning, False)

    def test12_EvalWithError(self):
        p = SlaveInterpreter().spawn()
        self.assertRaises(ZeroDivisionError, p.eval, '1/0')
        p.quit(); self.assertEqual(p.isRunning, False)

    def test20_Suite(self):
        p = SlaveInterpreter().spawn()
        p.suite('a = 10\n')
        p.suite('def bar(): return a * 2\n')
        self.assertEqual(p.eval('bar()'), 20)
        p.quit(); self.assertEqual(p.isRunning, False)

    def test22_SuiteWithError(self):
        p = SlaveInterpreter().spawn()
        self.assertRaises(NameError, p.suite, 'a = b\n')
        p.quit(); self.assertEqual(p.isRunning, False)



class TestProcessPool(TestCase):

    def test(self):
        pp = ProcessPool(template=SlaveInterpreter)
        si1 = pp.get()
        si2 = pp.get()
        for i in range(2):
            pp.get()
        print si1, si2
        print pp.busy, pp.spare
        si1.quit()
        pp.checkStatus()
        import time
        b = time.time()
        print b
        for i in xrange(1000):
            pp.checkStatus()
        e = time.time()
        print e, e-b
        print pp.busy, pp.spare



def suite():
    return TestSuite((
        makeSuite(TestChildProcess),
        makeSuite(TestSlaveInterpreter),
        #makeSuite(TestProcessPool),
    ))

if __name__ == '__main__':
    main(defaultTest='suite')
