from wm4.api import *
from dispatch.functions import AbstractGeneric
from dispatch.strategy import ordered_signatures, default
import re

from metadata import PrimaryKey
from errors import *

__all__ = [
    'ConstraintRegistry',
    'setattrs', 'on_insert', 'on_update', 'on_delete', 'on_commit',
    'Required', 'RequiredFeatures', 'ReadOnly',
    'NewValueConstraint', 'Syntax', 'MinLength', 'MaxLength', 'CheckEmail',
]

_ = i18n.MessageIDFactory('wm4')




class ConstraintRegistry(AbstractGeneric):

    def combine(self, items):
        constraints = {'before':[], 'after':[]}
        for group in ordered_signatures(items):
            for sig,(when,cons) in group:
                constraints[when].insert(0,cons)
        return lambda *a: constraints

    def before(self, cond=default, priority=100):
        return self._decorate(cond, 'before')

    def after(self, cond=default, priority=100):
        return self._decorate(cond, 'after')

    def check_before(self, obj, added, removed):
        ee = EntityError()
        for c in self(obj, added, removed)['before']:
            try:
                c(obj, added, removed)
            except FeatureError, fe:
                ee.add(fe)
        if ee:
            raise ee

    def check_after(self, obj, added, removed):
        ee = EntityError()
        for c in self(obj, added, removed)['after']:
            c(obj, added, removed)
            try:
                c(obj, added, removed)
            except FeatureError, fe:
                ee.add(fe)
        if ee:
            raise ee

def setattrs(self, added, removed): pass
setattrs = ConstraintRegistry(setattrs).delegate

def on_insert(workspace, klass): pass
on_insert = ConstraintRegistry(on_insert).delegate

def on_update(workspace, klass): pass
on_update = ConstraintRegistry(on_update).delegate

def on_delete(workspace, klass): pass
on_delete = ConstraintRegistry(on_delete).delegate

def on_commit(workspace, klasses): pass
on_commit = ConstraintRegistry(on_commit).delegate




# Required

class _Required(binding.AttributeMetadata): pass
Required = _Required(True)

class RequiredFeatures(binding.ClassMetadata):
    pass

def requiredFeatures(klass, fCls):
    return set(a for a in klass.features.keys() if _Required.of(klass,a,None))|\
           set(a for a in PrimaryKey.of(klass,default=()))

RequiredFeatures(requiredFeatures).register('*')


@setattrs.before(priority=20)
def checkRequired(self, added, removed):
    for a in RequiredFeatures.of(self.__class__):
        if a in added and added[a] is None:
            raise RequiredError(a)




# ReadOnly

class ReadOnlyConstraint(binding.AttributeMetadata):
    pass

ReadOnly = ReadOnlyConstraint(True)

[binding.declareAttribute.when(ReadOnlyConstraint)]
def checkReadOnly(klass, attr, constraint):
    constraint.register(klass, attr)
    @setattrs.before('self in klass and attr in added and attr in removed')
    def readonly_attribute(self, added, removed):
        raise AttributeError('Attribute is read only', klass, attr)




# Base class for value constraints


class NewValueConstraint:

    def __call__(self, obj, attr, val):
        raise NotImplementedError


[binding.declareAttribute.when(NewValueConstraint)]
def checkSyntax(klass, attr, constraint):
    @setattrs.before('self in klass and attr in added')
    def syntax(self, added, removed):
        val = added[attr]
        constraint(self, attr, val)


# Syntax

class Syntax(NewValueConstraint):

    def __init__(self, regex, msg=None):
        if isinstance(regex, basestring):
            regex = re.compile(regex)
        self.regex = regex
        self.msg = msg

    def __call__(self, obj, attr, val):
        if val is None:
            return
        m = self.regex.match(val)
        if not m:
            raise FeatureSyntaxError(attr, value=val)
        if m.group(0) != val:
            raise FeatureSyntaxError(attr, value=val)


class MinLength(NewValueConstraint):

    def __init__(self, len, msg=None):
        self.len = len
        self.msg = msg

    def __call__(self, obj, attr, val):
        if val is None:
            return
        if len(val) < self.len:
            raise MinLengthError(attr, value=val, len=self.len)

class MaxLength(MinLength):

    def __call__(self, obj, attr, val):
        if val is None:
            return
        if len(val) < self.len:
            raise MaxLengthError(attr, value=val, len=self.len)



class CheckEmail(NewValueConstraint):

    def __call__(self, obj, attr, val):
        if val is None:
            return
        from lib4.validators import checkEmail
        if not checkEmail(val):
            raise FeatureSyntaxError(attr, _(u'Invalid email address'))


CheckEmail = CheckEmail()
