from wm4.api import *

from interfaces import *
from metadata import *
from constraints import *
from utils import SortedFeatureDict
from errors import *

__all__ = [
    'TypeConfig', 'StructuredTypeConfig',
    'StructuredTypeMeta', 'StructuredType', 'Immutable',
    'EntityMeta', 'Entity', 'ViewMeta', 'View',
]


class TypeConfig(config.TypeProperty):
    '''TypeConfig for given StructuredType'''



class StructuredTypeConfig(binding.Component):
    '''Type properties that are configuration dependent'''

    typeObj = binding.Require()
    parent = None
    parents = binding.Make(lambda s: (s,))

    def features(self):
        klass = self.typeObj

        featDict = {}
        for a,d in klass.__all_descriptors__.iteritems():
            d = IFeature(d, None)
            if d is not None:
                featDict[a] = d.activateFeature(self, a)

        featSorting = {}
        for bnum, base in enumerate(klass.__bases__):
            for i, (a,_) in enumerate(getattr(base,'features',())):
                if a in featDict:
                    featSorting.setdefault(a,(bnum,i))

        cnum = len(klass.__bases__) + 1
        cDesc = klass.__class_descriptors__
        for a,f in cDesc.iteritems():
            if a in featDict:
                featSorting.setdefault(a, (cnum,f.ordering))

        return SortedFeatureDict(
            (a,featDict[a]) for a,_ in
            sorted(featSorting.items(), key=lambda i: (i[1], i[0]))
        )
    features = binding.Make(features)

    def __getitem__(self, key):
        return self.features[key]

    def refFeatures(self):
        feats = self.features
        return SortedFeatureDict(
            (a,feats[a]) for a in PrimaryKey.of(self.typeObj)
        )
    refFeatures = binding.Make(refFeatures)

    def primitives(self):
        return [p for a,d in self.features for p in d.refPrimitives]
    primitives = binding.Make(primitives)

    def refPrimitives(self):
        return [p for a,d in self.refFeatures for p in d.refPrimitives]
    refPrimitives = binding.Make(refPrimitives)

    def valueFromObj(self, obj, relativeTo=None):
        return obj

    def oidFromObj(self):
        refPrimitives = self.refPrimitives
        if len(refPrimitives) == 1:
            vfo = refPrimitives[0].valueFromObj
            def oidFromObj(obj):
                return vfo(obj)
        else:
            def oidFromObj(obj):
                oid = tuple(p.valueFromObj(obj) for p in refPrimitives)
                if None in oid:
                    return None
                return oid
        return oidFromObj
    oidFromObj = binding.Make(oidFromObj)

    def oidToAttrs(self, oid):
        return dict(
            (a,f.valueFromOID(oid)) for a,f in self.refFeatures
        )

    def oidToPrimitives(self, oid):
        return dict(
            (p.storageAttr, p.valueFromOID(oid)) for p in self.refPrimitives
        )

    def oidFromInput(self):
        oidMap = self.refFeatures
        inpLen = len(oidMap)

        if inpLen == 1:
            fMap = oidMap.values()[0]
            if len(fMap.refPrimitives) == 1:
                pMap = fMap.refPrimitives[0]
                def oidFromInput(val):
                    return pMap.valueFromObj(fMap.normalize(val), fMap)
            else:
                def oidFromInput(val):
                    val = fMap.normalize(val)
                    return tuple(
                        p.valueFromObj(val,fMap) for p in fMap.refPrimitives
                    )
        else:
            def oidFromInput(val):
                if not isinstance(val, tuple):
                    raise ValueError('OID should be tuple', self.typeObj, val)
                if len(val) != inpLen:
                    raise ValueError('Invalid OID length', self.typeObj, val)
                return tuple(
                    pMap.valueFromObj(fMap.normalize(v),fMap)
                    for (_,fMap), v in zip(oidMap, val)
                    for pMap in fMap.refPrimitives
                )
        return oidFromInput
    oidFromInput = binding.Make(oidFromInput)







class StructuredTypeMeta(binding.ActiveClass):
    '''Class that has features and activates IClonedDescriptors'''

    __metaclass__ = binding.Activator

    protocols.advise(instancesProvide=[IStructuredType, ISchemaElementFactory])

    features = property(lambda s: TypeConfig(s)().features)

    unique_features = binding.Make(list)

    def makeElement(klass, workspace):
        return klass




class StructuredType(object):

    __metaclass__ = StructuredTypeMeta

    def __init__(self, **values):
        values = self.checkAttrs(values)
        values = self.defaultValues(values)
        values = self.normalizeValues(values)

        setattrs.check_before(self, values, {})
        for a,v in values.iteritems():
            object.__setattr__(self,a,v)
        setattrs.check_after(self, values, {})

    def checkAttrs(self, values):
        klass = self.__class__
        for a,v in values.iteritems():
            if not hasattr(klass,a):
                raise AttributeError(klass, a)
        return values

    def normalizeValues(self, values):
        normalized = {}
        fDict = self.__class__.features
        ee = EntityError()
        for attr, val in values.iteritems():
            if attr in fDict:
                f = fDict[attr]
                try:
                    val = f.normalize(val)
                except:
                    ee[attr] = NormalizationError(
                        attr, value=val, type_name=i18n.typeName(f.typeObj))
            normalized[attr] = val
        if ee:
            raise ee
        return normalized

    def defaultValues(self, values):
        if values is None:
            values = {}
        for a,f in self.__class__.features:
            if a not in values:
                v = f.default(self, values)
                if v is NOT_GIVEN:
                    continue
                else:
                    values[a] = v
        return values

    def __setattr__(self, attr, value):
        return self._setattrs(**{attr:value})

    def _setattrs(self, **values):
        klass = self.__class__

        values = self.checkAttrs(values)
        values = self.normalizeValues(values)

        odict = self.__dict__
        removed = dict((a,odict[a]) for a in values if a in odict)

        self.onChange(values, removed)

        setattrs.check_before(self, values, removed)
        for a,v in values.iteritems():
            object.__setattr__(self,a,v)
        setattrs.check_after(self, values, removed)

    def onChange(self, added, removed):
        pass

    def __getstate__(self):
        tdesc = TypeConfig(self.__class__)()
        return dict(
            (k,v) for a,f in tdesc.features for k,v in f.toState(self)
        )

    def __setstate__(self, state):
        tDesc = TypeConfig(self.__class__)()
        iDict = self.__dict__
        for a,f in tDesc.features:
            if a not in iDict:
                v = f.fromState(state)
                if v is not NOT_GIVEN:
                    iDict[a] = v

TypeConfig(StructuredType, lambda r,f: StructuredTypeConfig(typeObj=r))









class Immutable(StructuredType):

    def __init__(self, *args, **values):
        klass = self.__class__
        fNames = klass.features.keys()
        fLen = len(fNames)
        if args and values:
            raise ValueError('use arguments or keywords; not both')
        elif args:
            if len(args) > fLen:
                raise ValueError('maximal lenght exceeded', fLen)
            values = dict((a,v) for v,a in zip(args, fNames))

        super(Immutable, self).__init__(**values)

    def __setattr__(self, attr, value):
        raise AttributeError('Immutable is read only', self.__class__, attr)

    def __iter__(self):
        for a,f in self.__class__.features:
            yield getattr(self, a)





class EntityMeta(StructuredTypeMeta):

    def allocate(klass, workspace=None):
        return object.__new__(klass, _workspace=workspace)



class Entity(StructuredType):

    __metaclass__ = EntityMeta

    def get_oid(self):
        dct = self.__dict__
        oid = dct.get('_oid', NOT_FOUND)
        if oid is NOT_FOUND:
            oid = dct['_oid'] = TypeConfig(self.__class__)().oidFromObj(self)
        return oid

    def set_oid(self, value):
        raise AttributeError('read-only attribute', self, '_oid', value)

    _oid = property(get_oid, set_oid)

    def _updateOid(self):
        oid = TypeConfig(self.__class__)().oidFromObj(self)
        self.__dict__['_oid'] = oid
        return oid

    _workspace = property(lambda o: o.__dict__.get('_workspace'))

    nonTxnAttrs = set(('_oid', '_workspace'))

    def _finishTransaction(self, ws, ts, commited):
        nta = self.nonTxnAttrs
        dct = self.__dict__
        for a in dct.keys():
            if a not in nta:
                del dct[a]

    def _preload(self, obj):
        ws = self._workspace
        for a,f in self.__class__.features:
            f.on_preload(self, ws, obj)

    def onChange(self, added, removed):
        ws = self._workspace
        if ws is None:
            return
        if self.__dict__.get('__lock__') is not None:
            return
        lock = self.__dict__['__lock__'] = ws.do_load(
            self.__class__, self._oid, preload=False, lock=True
        )
        self._preload(lock)
        ws.on_change(self)

    def loadDeactivatedFeature(self, feat):
        ws = self._workspace
        if ws and '_oid' in self.__dict__:
            ws.do_load(self.__class__, self._oid)



    def beforeAdd(self):
        for attr, feature in self.__class__.features:
            feature.beforeAdd(self)
    def afterAdd(self):
        for attr, feature in self.__class__.features:
            feature.afterAdd(self)

    def beforeUpdate(self):
        for attr, feature in self.__class__.features:
            feature.beforeUpdate(self)
    def genChanges(self):
        return dict(
            (a,v) for attr,feature in self.__class__.features
                  for a,v in feature.genChanges(self)
        )
    def afterUpdate(self):
        for attr, feature in self.__class__.features:
            feature.afterUpdate(self)

    def beforeRemove(self):
        for attr, feature in self.__class__.features:
            feature.beforeRemove(self)
    def afterRemove(self):
        for attr, feature in self.__class__.features:
            feature.afterRemove(self)



# default values

def defaultComputeDefault(search, found):
    ent, attr = search
    feat = ent.features.get(attr)
    if feat is None:
        return lambda o,a,v: None
    return ComputeDefaultForType.of(feat.typeObj)
ComputeDefault(defaultComputeDefault).register(StructuredType, '*')


# Binding for Entity instances

@binding.getParentComponent.when(Entity)
def get_parent_of_entity(component):
    return component.__dict__.get('_workspace')

@binding.getComponentName.when(Entity)
def get_entity_name(component):
    return str(component.__dict__.get('_oid'))






class ViewMeta(EntityMeta):

    query = binding.Require()

    def __new__(klass, name, bases, cdict):
        return EntityMeta


class View(Entity):

    __metaclass__ = ViewMeta

    _implementationClass = binding.Require()

    def __getattr__(self, attr):
        raise NotImplementedError

    def __setattr__(self, attr, value):
        # XXX proxy
        raise NotImplementedError
