from wm4.api import *
from itertools import count
import sys

from interfaces import *
from metadata import PrimaryKey, ReferenceEndOrder, Normalize, ComputeDefault
from utils import SortedFeatureDict

__all__ = [
    'getTypeObj',
    'Feature', 'One', 'Attribute', 'Reference', 'QueryFeature', 'ReferenceEnd',
]


next_serial = count().next


class Feature(object):

    __metaclass__ = binding.Activator

    protocols.advise(instancesProvide=[binding.IActiveDescriptor, IFeature])

    # IActiveDescriptor
    attrName = None
    offerAs = ()
    uponAssembly = False
    metadata = None

    def activateInClass(self, klass, attr):
        binding.declareAttribute(klass, attr, self.metadata)
        cp = self.copy()
        cp.attrName = attr
        setattr(klass, attr, cp)
        return cp

    # IFeature
    ordering = None
    cardinality = None
    parent = None
    parents = binding.Make(lambda s:s.parent.parents+(s,), suggestParent=False)
    path = binding.Make(lambda s: '.'.join(p.attrName for p in s.parents[1:]))

    def activateFeature(self, parent, attr):
        cp = self.copy()
        cp.parent = parent
        cp.attrName = attr
        return cp

    def __init__(self, *metadata, **kw):
        self.metadata = metadata
        if not 'ordering' in kw:
            self.ordering = next_serial()
        binding.initAttrs(self, kw.iteritems())

    copyAttrs = 'attrName', 'ordering', 'parent'

    def copy(self, **attrs):
        copy = object.__new__(self.__class__)
        for a in self.copyAttrs:
            setattr(copy, a, getattr(self,a))
        for a,v in attrs.iteritems():
            setattr(copy, a, v)
        return copy

    def __get__(self, obj, typ):
        self = typ.features[self.attrName]
        if obj is None :
            return self
        return self.get(obj)

    def __set__(self, obj, value):
        obj.__class__.features[self.attrName].set(obj, value)

    def __delete__(self, obj):
        obj.__class__.features[self.attrName].delete(obj)

    def get(self, obj):
        raise NotImplementedError(obj, self.attrName)

    def set(self, obj, value):
        raise NotImplementedError(obj, self.attrName, value)

    def delete(self, obj):
        raise NotImplementedError(obj, self.attrName)

    def normalize(self, value):
        return value

    def default(self, obj, values={}):
        return NOT_GIVEN

    def on_add(self, obj, ws):
        pass

    def on_remove(self, obj, ws):
        pass

    def on_preload(self, obj, ws, pobj):
        pass

    def toState(self, obj):
        return ()

    def fromState(self, state):
        return NOT_FOUND

    def beforeAdd(self, obj): pass
    def afterAdd(self, obj): pass

    def beforeUpdate(self, obj): pass
    def genChanges(self, obj): return ()
    def afterUpdate(self, obj): pass

    def beforeRemove(self, obj): pass
    def afterRemove(self, obj): pass



def getTypeObj(feat, typeRef):
    if typeRef is THIS_CLASS:
        return feat.parent.typeObj
    else:
        return TypeRef(typeRef)()





class One(Feature):

    protocols.advise(instancesProvide=[ITypeDescriptor])

    cardinality = 'single'
    typeRef = None
    typeObj = None

    copyAttrs = Feature.copyAttrs + ('typeRef', 'typeObj')

    def activateFeature(self, parent, attr):
        cp = super(One, self).activateFeature(parent, attr)
        cp.typeObj = getTypeObj(cp, cp.typeRef)
        return cp

    def __init__(self, typeRef, *metadata, **kw):
        super(One,self).__init__(*metadata, **kw)
        self.typeRef = typeRef

    def get(self, obj):
        attr = self.attrName
        val = obj.__dict__.get(attr, NOT_FOUND)
        if val is NOT_FOUND :
            obj.loadDeactivatedFeature(self)
            try :
                val = obj.__dict__[attr]
            except KeyError:
                raise AttributeError(
                    'Lazy load failed', obj.__class__, attr,
                    obj.__dict__.get('_oid'),
                    obj.__dict__.get('_workspace'),
                )
        return val

    def set(self, obj, value):
        obj.__dict__[self.attrName] = value

    def delete(self, obj):
        raise TypeError("Features don't support delete", self)

    normalize = binding.Make(
        lambda s: Normalize.of(s.typeObj), suggestParent=False
    )

    def default(self, obj, values={}):
        cd = ComputeDefault.of(obj.__class__, self.attrName)
        return cd(obj, self.attrName, values)

    primitives = binding.Require()

    oidPart = binding.Make(lambda s: s.attrName in s.parent.refFeatures)

    def rootOidPart(self):
        parent = self.parent
        if isinstance(parent, One):
            return parent.rootOidPart and self.oidPart
        else:
            return self.oidPart
    rootOidPart = binding.Make(rootOidPart)

    def valueFromObj(self, obj, relativeTo=None):
        if obj is None:
            return None
        if self is relativeTo:
            return obj
        else:
            o = self.parent.valueFromObj(obj, relativeTo)
            if o is None:
                return None
            else:
                return getattr(o, self.attrName)

    def genChanges(self, obj):
        attr = self.attrName
        if not attr in obj.__dict__:
            return
        new = self.get(obj)
        old = self.get(obj.__dict__['__lock__'])
        if old != new:
            yield attr, new




class Attribute(One):

    def activateFeature(self, parent, attr):
        cp = super(Attribute, self).activateFeature(parent, attr)
        if issubclass(cp.typeObj, schema.Entity):
            raise TypeError(
                'Entity subclass used in Attribute feature',self,cp.typeObj)
        return cp

    refPrimitives = binding.Make(lambda s: [s])

    def storageAttr(self):
        root = self.parents[0]
        rootFeat = self.parents[1]
        if rootFeat is self:
            return self.attrName
        if len(rootFeat.refPrimitives) > 1:
            return self.path.replace('.', '_')
        else:
            return rootFeat.attrName
    storageAttr = binding.Make(storageAttr)

    def toState(self, obj):
        yield self.storageAttr, self.valueFromObj(obj)

    def fromState(self, state):
        return state.get(self.storageAttr, NOT_FOUND)

    def valueFromOID(self):
        if not self.rootOidPart:
            raise TypeError('%r is not part of root OID' % self)
        root = self.parents[0]
        if len(root.refPrimitives) == 1:
            def valueFromOID(oid):
                return oid
        else:
            idx = root.refPrimitives.index(self)
            def valueFromOID(oid):
                return oid[idx]
        return valueFromOID
    valueFromOID = binding.Make(valueFromOID)

    def on_preload(self, obj, ws, pobj):
        attr = self.attrName
        if attr not in obj.__dict__ and attr in pobj.__dict__:
            obj.__dict__[attr] = pobj.__dict__[attr]


class Reference(One):

    def activateFeature(self, parent, attr):
        cp = super(Reference, self).activateFeature(parent, attr)
        if not issubclass(cp.typeObj, schema.Entity):
            raise TypeError(
                'Entity subclass required in Reference feature',self,cp.typeObj)
        return cp

    def features(self):
        return SortedFeatureDict(
            (a,f.copy(parent=self)) for a,f in self.typeObj.features
        )
    features = binding.Make(features)

    def refFeatures(self):
        feats = self.features
        return SortedFeatureDict(
            (a,feats[a]) for a in PrimaryKey.of(self.typeObj)
        )
    refFeatures = binding.Make(refFeatures)

    def __getitem__(self, key):
        return self.features[key]

    def refPrimitives(self):
        return [p for a,d in self.refFeatures for p in d.refPrimitives]
    refPrimitives = binding.Make(refPrimitives)

    def toState(self, obj):
        for a,f in self.refFeatures:
            for k,v in f.toState(obj):
                yield k,v

    def fromState(self, state):
        obj = self.typeObj.allocate()
        for a,f in self.refFeatures:
            v = f.fromState(state)
            if v is NOT_FOUND:
                return NOT_FOUND
            obj.__dict__[a] = v
        return obj

    def valueFromOID(self):
        if not self.rootOidPart:
            raise TypeError('%r is not part of root OID' % self)
        map = [(a,f.valueFromOID) for a,f in self.refFeatures]
        def valueFromOID(oid):
            values = {}
            for a,g in map:
                v = g(oid)
                if v is None:
                    return None
                values[a] = v
            obj = self.typeObj.allocate()
            for a,v in values.iteritems():
                obj.__dict__[a] = v
            return obj
        return valueFromOID
    valueFromOID = binding.Make(valueFromOID)

    def on_add(self, obj, ws):
        val = self.get(obj)
        if val is not None:
            ws.add(val)

    def on_preload(self, obj, ws, pobj):
        attr = self.attrName
        if attr not in obj.__dict__:
            if attr in pobj.__dict__:
                val = obj.__dict__[attr] = pobj.__dict__[attr]
            else:
                return
        else:
            val = obj.__dict__[attr]
        if val is not None and val._workspace is not ws:
            val = ws.preload(val)

    def genChanges(self, obj):
        attr = self.attrName
        if not attr in obj.__dict__:
            return
        new = self.get(obj)
        if new is None:
            newoid = None
        else:
            newoid = new._oid
        old = self.get(obj.__dict__['__lock__'])
        if old is None:
            oldoid = None
        else:
            oldoid = old._oid
        if oldoid != newoid:
            yield attr, new









class QueryFeature(Feature):

    cardinality = 'set'
    query = None
    gdict = None

    copyAttrs = Feature.copyAttrs + ('query', 'gdict')

    def __init__(self, query, *metadata, **kw):
        super(QueryFeature, self).__init__(*metadata, **kw)
        self.query = query
        gdict = kw.get('gdict')
        if gdict is None:
            gdict = sys._getframe(1).f_globals
        self.gdict = gdict

    def get(self, obj):
        ns = [{'obj':obj, 'self':obj}, self.gdict, sys._getframe().f_builtins]
        return self.perform(obj, self.query, ns)

    def perform(self, obj, qry, ns):
        ws = obj.__dict__.get('_workspace')
        if not ws:
            return ()
        else:
            if self.attrName not in obj.__dict__:
                obj.__dict__[self.attrName] = tuple(ws(qry, namespaces=ns))
            return obj.__dict__[self.attrName]

    refPrimitives = ()


class ReferenceEnd(QueryFeature):

    typeRef = None
    typeObj = None
    refAttr = None

    copyAttrs = Feature.copyAttrs + ('typeRef', 'typeObj', 'refAttr')

    def activateFeature(self, parent, attr):
        cp = super(ReferenceEnd, self).activateFeature(parent, attr)
        cp.typeObj = getTypeObj(cp, cp.typeRef)
        return cp

    def __init__(self, typeRef, refAttr, *metadata, **kw):
        super(ReferenceEnd, self).__init__(None, *metadata, **kw)
        self.typeRef = typeRef
        self.refAttr = refAttr

    def get(self, obj):
        ns = [{'o':obj, 'f':self}, globals(), sys._getframe().f_builtins]
        ord = ReferenceEndOrder.of(obj.__class__, self.attrName, None)
        if ord:
            qry = '[e for e in f.typeObj if e.%s==o and query.OrderBy(%s)]'%(
                self.refAttr, ', '.join('e.%s' % o for o in ord)
            )
        else:
            qry = '[e for e in f.typeObj if e.%s==o]' % self.refAttr
        return self.perform(obj, qry, ns)



class List(Feature):

    cardinality = 'list'
    typeRef = None
    typeObj = None
    indexFeature = None

    def __init__(self, typeRef, refAttr, indexFeature='idx', *metadata, **kw):
        super(List, self).__init__(*metadata, **kw)
        self.typeRef = indexFeature
        self.indexFeature = indexFeature

    copyAttrs = Feature.copyAttrs + (
        'typeRef', 'typeObj', 'refAttr', 'indexFeature'
    )

    def activateFeature(self, parent, attr):
        cp = super(List, self).activateFeature(parent, attr)
        cp.typeObj = getTypeObj(cp, cp.typeRef)
        return cp

    def get(self, obj):
        attr = self.attrName
        val = obj.__dict__.get(attr, NOT_FOUND)
        if val is NOT_FOUND :
            val = obj.__dict__[attr] = ObservableList()
        return val

    def set(self, obj, val):
        raise NotImplementedError

    def delete(self, obj):
        raise TypeError("Features don't support delete", self)
