from wm4.core import *
from protocols import Interface, Attribute
from lib4.symbols import Symbol
from lib4.imports import importString
from wm4.config.context import FlatProperty, AbstractProperty
import sys


__all__ = [
    'THIS_CLASS', 'TypeRef',
    'ISchemaElementFactory',
    'ITypeDescriptor', 'IFeature', 'IStructuredType',
    'IManagedClass', 'IManagedEntity', 'IWorkspace', 'IWorkspaceSPI',
]

THIS_CLASS   = Symbol("THIS_CLASS", __name__)

class TypeRef(FlatProperty):

    __slots__ = ()

    _registry = {}

    def __new__(klass, value, default=NOT_GIVEN):
        self = klass._registry.get(value)

        if self is None:

            ivalue = value

            if isinstance(ivalue, TypeRef):
                self = value

            elif not isinstance(value, basestring):
                module = value.__module__
                name = value.__name__

                # check that class is reachable
                if not module in sys.modules:
                    raise TypeError('Cannot find module', value, module)
                m = sys.modules[module]
                if getattr(m,name,None) is None:
                    raise TypeError(
                        'Cannot find type %s in module %s' % (name,module),value
                    )
                self = AbstractProperty.__new__(klass)
                self.value = '%s.%s' % (module, name)

            else:
                self = AbstractProperty.__new__(klass)
                self.value = value

            klass._registry[ivalue] = self

        if default is not NOT_GIVEN:
            self.setdefault(default)

        return self


TypeRef('*', lambda r,f: importString(r))








class ISchemaElementFactory(Interface):

    def makeElement(workspace, path):
        pass



class ITypeDescriptor(Interface):

    typeObj = Attribute('Python type subclass')

    features = Attribute('Sorted dictionary of all features')

    primitives = Attribute('Sequence of primitive (or atomic) features')




class IStructuredType(Interface):
    '''Structured (non-primitive) type'''

    features = Attribute('Sorted dictionary of all features')

    def allocate(workspace=None):
        '''Creates new entity within given workspace'''






class IFeature(Interface):

    ordering = Attribute('Value for feature ordering')

    cardinality = Attribute('single, set')

    parent = Attribute('parent ITypeDescriptor')

    def get(obj):
        '''Returns value of this feature in object obj'''

    def set(obj, value):
        '''Set value of this feature in object obj'''

    def delete(obj):
        '''Delete value of this feature in object obj'''

    def onAdd(obj, ws):
        '''Called when obj is added to workspace ws'''

    def onRemove(obj, ws):
        '''Called when obj is removed from workspace ws'''

    def onPreload(obj, ws, pobj):
        '''Called on preload'''











class IManagedClass(Interface):

    def _oidFromState(state):
        pass


class IManagedEntity(Interface):
    '''Object that delegates its state operation to some IWorkspace
    '''

    _workspace = Attribute('IWorkspace or None')

    _oid = Attribute('Unique entity identifier within one class-workspace')

    def _updateOid():
        '''Updates _oid attributes on demand

           Attribute _oid doesn't reflect changes in features automaticaly,
           i.e. if someone sets feature, that is part of primary key,
           _oid keeps old value. This method updates _oid inplace
           and returns it.
        '''

    def _preload(state):
        '''Method is called when trusted state data are retrived from
           persistent storage. Object is given chance to fill missing
           attributes.
        '''

    def _finishTransaction(ws, ts, commited):
        '''Called on transaction commit or abort
           Default implementation deactivates instance, i.e. __dict__
           is cleared except attributes enumerated in nonTxnAttrs.
        '''





class IWorkspace(Interface):
    '''Container for IManagedEntity instances

       Interface has only domentation purpose.
    '''

    def allocate(klass, oid):
        '''Alocates new ghost object

           Returns new instance without __init__ called. Sets _workspace
           and _oid attributes. Object state can be later obtained
           and set by following code:

                obj = ws.allocate(klass, oid)
                state = workspace.do_load(klass,oid)
                obj._preload(state)
        '''

    def get(klass, oid, default=NOT_GIVEN):
        '''Returns IManagedEntity with given oid of given klass

           Raises KeyError if object is not found and default is NOT_GIVEN.
           Object is immediately retrieved from storage if IWorkspace
           represents persistent storage.
        '''

    def add(obj):
        '''Add object to this workspace

           Object must be instance of IManagedEntity and its _oid must
           be unique within this workspace. If entity is already in this
           workspace (obj._workspace is self) this method is a no-op.
           If entity is in another workspace, method raises exception.
        '''

    def flush(obj):
        '''Updates state of changed object in persistent storage.

           This method is called automaticaly on transaction commit
           for every changed entity i.e. entity that called `onChange(obj)'
           method within transaction.
        '''

    def remove(obj):
        '''Remove object from workspace.

           The object must belong to this workspace (obj._workspace is self).
           Attribute _workspace is set to None after removal. Database
           operation is executed immediately.
        '''

    def __call__(qry, namespaces=None, depth=1, preload=True):
        '''Execute query

           Query can be string or IQuery instance.

           Whena namespaces argument is None, namespaces (globals, locals
           and bulitins) are obtained from stack frame with given depth
           (default depth=1).

           I preload is False, returned objects are transient, i.e. they
           are not bound to this workspace.
        '''

class IWorkspaceSPI(Interface):
    '''"Backend" workspace functionality

       Interface has only domentation purpose.
    '''

    def do_load(klass, oid, preload=False, lock=False):
        '''Returns persistent state for given klass-oid

           It is mainly called from ghost object that need fill missing
           attributes or attributes deactivated after transaction (lazy load).
        '''

    def do_preload(klass, obj):
        '''....'''

    def do_add(obj):
        '''Performs INSERT in persistent storage'''

    def on_change(obj):
        '''Method should be called from obj when obj changes its state.

           Default implementation locks object in persistent storage
           and register object as dirty. Changes are flushed on transaction
           commit with flush(obj) method.
        '''

    def do_update(obj):
        '''Performs UPDATE in persistent storage'''

    def do_remove(obj):
        '''Performs DELETE in persistent storage'''
