# -*- coding: ISO-8859-2 -*-

from wm4.testing import *
from wm4.api import *
from wm4.schema.containers import *

class TestPersistentList(TestCase):

    def setUp(self):
        self.change = None

    def register(self, pl):
        assert self.change is None
        self.change = list(pl)
        return self.change

    def test01_Empty(self):
        pl = PersistentList()
        pl._p_register = self.register
        self.assertEqual(pl, [])
        self.assertEqual(self.change, None)

    def test10_Reading(self):
        pl = PersistentList((1,2))
        pl._p_register = self.register
        self.assertEqual(pl, [1,2])
        self.assertEqual(pl[0], 1)
        self.assertEqual(self.change, None)

    def test20_Setitem(self):
        pl = PersistentList((1,2))
        pl._p_register = self.register
        pl[1] = 3
        self.assertEqual(pl, [1,3])
        self.assertEqual(pl._p_changed, [1,2])
        self.assertEqual(self.change, [1,2])

    def test21_Append(self):
        pl = PersistentList((1,2))
        pl._p_register = self.register
        pl.append(3)
        self.assertEqual(pl, [1,2,3])
        self.assertEqual(self.change, [1,2])

    def test22_Pop(self):
        pl = PersistentList((1,2))
        pl._p_register = self.register
        pl.pop()
        self.assertEqual(pl, [1])
        self.assertEqual(self.change, [1,2])

    def test90_Multichange(self):
        pl = PersistentList((1,2))
        pl._p_register = self.register
        pl[0] = 'a'
        pl[1] = 'b'
        self.assertEqual(pl, ['a','b'])
        self.assertEqual(self.change, [1,2])


class TestPersistentDict(TestCase):

    def setUp(self):
        self.change = None

    def register(self, pd):
        assert self.change is None
        self.change = dict(pd)
        return self.change

    def test01_Empty(self):
        pd = PersistentDict()
        pd._p_register = self.register
        self.assertEqual(pd, {})
        self.assertEqual(self.change, None)

    def test10_Reading(self):
        pd = PersistentDict({1:'a', 2:'b'})
        pd._p_register = self.register
        self.assertEqual(pd, {1:'a', 2:'b'})
        self.assertEqual(pd[1], 'a')
        self.assertEqual(self.change, None)

    def test20_Setitem(self):
        pd = PersistentDict({1:'a', 2:'b'})
        pd._p_register = self.register
        pd[1] = 'A'
        self.assertEqual(self.change, {1:'a', 2:'b'})
        self.assertEqual(pd._p_changed, {1:'a', 2:'b'})
        self.assertEqual(pd, {1:'A', 2:'b'})

    def test21_Delitem(self):
        pd = PersistentDict({1:'a', 2:'b'})
        pd._p_register = self.register
        del pd[2]
        self.assertEqual(self.change, {1:'a', 2:'b'})
        self.assertEqual(pd, {1:'a'})

    def test22_Clear(self):
        pd = PersistentDict({1:'a', 2:'b'})
        pd._p_register = self.register
        pd.clear()
        self.assertEqual(self.change, {1:'a', 2:'b'})
        self.assertEqual(pd, {})

    def test23_Update(self):
        pd = PersistentDict({1:'a', 2:'b'})
        pd._p_register = self.register
        pd.update({1:'A', 3:'C'})
        self.assertEqual(self.change, {1:'a', 2:'b'})
        self.assertEqual(pd, {1:'A', 2:'b', 3:'C'})

    def test24_SetdefaultNonexisting(self):
        pd = PersistentDict({1:'a', 2:'b'})
        pd._p_register = self.register
        pd.setdefault(3, 'C')
        self.assertEqual(self.change, {1:'a', 2:'b'})
        self.assertEqual(pd, {1:'a', 2:'b', 3:'C'})

    def test25_SetdefaultExisting(self):
        pd = PersistentDict({1:'a', 2:'b'})
        pd._p_register = self.register
        pd.setdefault(1, 'A')
        self.assertEqual(self.change, None)
        self.assertEqual(pd, {1:'a', 2:'b'})

    def test90_Multichange(self):
        pd = PersistentDict({1:'a', 2:'b'})
        pd._p_register = self.register
        pd[1] = 'A'
        pd[2] = 'B'
        pd.setdefault(3, 'C')
        self.assertEqual(self.change, {1:'a', 2:'b'})
        self.assertEqual(pd, {1:'A', 2:'B', 3:'C'})



class TestPersistentDictOfLists(TestCase):

    def setUp(self):
        self.change = None

    def register(self, pd):
        assert self.change is None
        self.change = dict((k,list(v)) for k,v in pd.items())
        return self.change

    def test01_Empty(self):
        pd = PersistentDictOfLists()
        pd._p_register = self.register
        self.assertEqual(pd, {})
        self.assertEqual(self.change, None)

    def test10_Reading(self):
        pd = PersistentDictOfLists(dict(a=[0,1], b=[]))
        pd._p_register = self.register
        self.assertEqual(pd, dict(a=[0,1], b=[]))
        self.assertEqual(pd['a'][1], 1)
        self.failUnless(isinstance(pd['a'], PersistentList))
        self.failUnless(isinstance(pd['b'], PersistentList))
        self.assertEqual(self.change, None)

    def test20_Setitem(self):
        pd = PersistentDictOfLists(dict(a=[0,1], b=[]))
        pd._p_register = self.register
        pd['a'][1] = 2
        pd['c'] = [42]
        self.failUnless(isinstance(pd['c'], PersistentList))
        self.assertEqual(pd, dict(a=[0,2], b=[], c=[42]))
        self.assertEqual(self.change, dict(a=[0,1], b=[]))

    def test21_Append(self):
        pd = PersistentDictOfLists(dict(a=[0,1], b=[]))
        pd._p_register = self.register
        pd['a'].append(2)
        pd['b'].append('a')
        self.assertEqual(pd, dict(a=[0,1,2], b=['a']))
        self.assertEqual(self.change, dict(a=[0,1], b=[]))

    def test22_Pop(self):
        pd = PersistentDictOfLists(dict(a=[0,1], b=[]))
        pd._p_register = self.register
        pd['a'].pop()
        self.assertEqual(pd, dict(a=[0], b=[]))
        self.assertEqual(self.change, dict(a=[0,1], b=[]))

    def test30_InitWithIter(self):
        pd = PersistentDictOfLists(dict(a=[0,1], b=[]).iteritems())
        pd._p_register = self.register
        pd['a'][1] = 2
        pd['c'] = [42]
        self.failUnless(isinstance(pd['a'], PersistentList))
        self.failUnless(isinstance(pd['b'], PersistentList))
        self.failUnless(isinstance(pd['c'], PersistentList))
        self.assertEqual(pd, dict(a=[0,2], b=[], c=[42]))
        self.assertEqual(self.change, dict(a=[0,1], b=[]))

    def test31_InitWithKeywords(self):
        pd = PersistentDictOfLists(a=[0,1], b=[])
        pd._p_register = self.register
        pd['a'][1] = 2
        pd['c'] = [42]
        self.failUnless(isinstance(pd['a'], PersistentList))
        self.failUnless(isinstance(pd['b'], PersistentList))
        self.failUnless(isinstance(pd['c'], PersistentList))
        self.assertEqual(pd, dict(a=[0,2], b=[], c=[42]))
        self.assertEqual(self.change, dict(a=[0,1], b=[]))

    def test40_Clean(self):
        pd = PersistentDictOfLists(a=[0,1], b=[])
        pd._p_register = self.register
        pd['a'][1] = 2
        pd['c'] = [42]
        self.assertEqual(pd, dict(a=[0,2], b=[], c=[42]))
        self.assertEqual(pd._p_changed, dict(a=[0,1], b=[]))
        pd._p_clean()
        assert '_p_changed' not in pd.__dict__
        assert '_p_changed' not in pd['a'].__dict__
        assert '_p_changed' not in pd['b'].__dict__
        assert '_p_changed' not in pd['c'].__dict__
        self.change = None

        pd['a'][1] = 3
        self.assertEqual(pd, dict(a=[0,3], b=[], c=[42]))
        self.assertEqual(pd._p_changed, dict(a=[0,2], b=[], c=[42]))
        assert '_p_changed' in pd.__dict__
        assert '_p_changed' in pd['a'].__dict__
        assert '_p_changed' not in pd['b'].__dict__
        assert '_p_changed' not in pd['c'].__dict__
        pd._p_clean()
        assert '_p_changed' not in pd.__dict__
        assert '_p_changed' not in pd['a'].__dict__
        assert '_p_changed' not in pd['b'].__dict__
        assert '_p_changed' not in pd['c'].__dict__


suite = Suite(
    TestPersistentList,
    TestPersistentDict,
    TestPersistentDictOfLists,
)

if __name__ == '__main__':
    Main()
