# -*- coding: ISO-8859-2 -*-

from unittest import *
from wm4.api import *
from wm4.schema.api import *
from wm4.schema.tests.utils import *
from wm4.query.tests.app import *
from datetime import date
import sys

CN = binding.ComponentName
GCP = binding.getComponentPath
MN = GCP(sys.modules[__name__])



class S(Entity): pass

class E1(Entity):
    a = b = Attribute(int)

class E2(E1):
    b = Attribute(unicode)


class TestTypeConfig(TestCase):

    def test10_TypeConfigInstance(self):
        etc = TypeConfig(E1)()
        ftc = TypeConfig(E2)()
        self.assertNotEqual(etc, ftc)

        cfg = config.Config()
        etc2 = config.with_(cfg, TypeConfig(E1))
        self.assertNotEqual(etc, etc2)

    def test12_FeatureRegistry(self):
        f1 = E1.features
        cfg = config.Config()
        f2 = config.with_(cfg, getattr, E1, 'features')
        self.assertNotEqual(f1, f2)
        self.assertNotEqual(f1, E2.features)

    def test13_FeatureRegistryInstance(self):
        f1 = E1.features
        f2 = E1.features
        assert f1 is f2

    def test15_Features(self):
        self.assertEqual(E1.a, E1.features['a'])
        self.assertEqual(E2.a, E2.features['a'])
        self.assertNotEqual(E1.a, E2.a)
        cfg = config.Config()
        a2 = config.with_(cfg, getattr, E1, 'a')
        self.assertNotEqual(E1.a, a2)

    def test16_FeatureParent(self):
        etc = TypeConfig(E1)()
        self.assertEqual(E1.a.parent, etc)
        self.assertEqual(E1.__dict__['a'].parent, None)



class TestTypeClass(TestCase):

    def test10_Features(self):
        class E(Entity):
            a = b = Attribute(int)
        self.failUnless('a' in E.features)
        self.failUnless('b' in E.features)
        self.assertEqual(E.features.keys(), ['a','b'])
        self.assertEqual(E.features.values(), [E.a, E.b])
        self.assertEqual(E.features.items(), (('a',E.a), ('b',E.b)))
        self.assertEqual(E.features['a'], E.a)
        self.assertEqual(tuple(E.features), E.features.items())

        self.assertEqual(E.a.attrName, 'a')
        self.assertEqual(E.a.typeObj, int)
        self.assertEqual(E.b.attrName, 'b')
        self.assertEqual(E.b.typeObj, int)
        assert E.a is E.a

    def test11_InheritedFeatures(self):
        class E(Entity):
            a = b = Attribute(int)
        class F(E):
            c = Attribute(str)
        self.assertEqual(E.features.keys(), ['a','b'])
        self.assertEqual(F.features.keys(), ['a','b','c'])
        self.assertEqual(E.a.attrName, 'a')
        self.assertEqual(E.a.typeObj, int)
        self.assertEqual(F.a.attrName, 'a')
        self.assertEqual(F.a.typeObj, int)
        self.assertEqual(F.c.attrName, 'c')
        self.assertEqual(F.c.typeObj, str)

    def test30_TypeAsComponent(self):
        self.assertEqual(GCP(S), CN(MN.parts+['S']))







class TestFeatureSorting(TestCase):

    def test10_SimpleEntity(self):
        class E(Entity):
            b = Attribute(int)
            c = Attribute(int)
            a = Attribute(int)
        self.assertEqual(E.features.keys(), ['b','c','a'])

    def test11_LexicograficSorting(self):
        class E(Entity):
            b = c = a = Attribute(int)
        self.assertEqual(E.features.keys(), ['a','b','c'])

    def test13_ReferenceSorting(self):
        class E(Entity):
            b = Attribute(int)
            c = Reference(S)
            a = Attribute(int)
        self.assertEqual(E.features.keys(), ['b','c','a'])

    def test20_SubclassSorting(self):
        class E(Entity):
            b = Attribute(int)
            a = Attribute(int)
        class F(E):
            c = Attribute(int)
        self.assertEqual(E.features.keys(), ['b','a'])
        self.assertEqual(F.features.keys(), ['b','a','c'])

    def test21_SortingWithMultipleInheritance(self):
        class A(Entity):
            a = Attribute(int)
        class B(Entity):
            b = Attribute(int)
        class C(A,B):
            c = Attribute(int)
        self.assertEqual(C.features.keys(), ['a','b','c'])
        class D(B,A):
            c = Attribute(int)
        self.assertEqual(D.features.keys(), ['b','a','c'])

    def test22_ReplacingFeatureInSubclass(self):
        class E(Entity):
            b = Attribute(int)
            a = Attribute(int)
        class F(E):
            c = Attribute(int)
            a = Attribute(unicode)
        self.assertEqual(F.features.keys(), ['b','a','c'])

    def test23_ReplacingFeatureWithNonFeatureInSubclass(self):
        class E(Entity):
            b = Attribute(int)
            a = Attribute(int)
        class F(E):
            c = Attribute(int)
            a = binding.Make(dict)
        self.assertEqual(F.features.keys(), ['b','c'])






class TestState(TestCase):

    def test10_GetAttributeState(self):
        g = Group(gid=100, name='Foo')
        self.assertEqual(g.__getstate__(), {'gid':100, 'name':'Foo'})

        u = User(uid=1, login='bob')
        self.assertEqual(u.__getstate__(), {
            'uid':1, 'login':'bob', 'name':None, 'mtime':None,
        })

    def test15_GetReferenceState(self):
        ug = UserGroup(
                user  = User(uid=1,login='bar'),
                group = Group(gid=100, name='Foo')
        )
        self.assertEqual(ug.__getstate__(), {'user':1, 'group':100})
        ugs = UGStat(ug=ug, term=date(2004,1,1))
        self.assertEqual(ugs.__getstate__(), {
            'ug_user_uid':1, 'ug_group_gid':100,
            'term':date(2004,1,1), 'value':None,
        })

    def test30_UpdateWithEmptyState(self):
        g = Group(gid=100, name='Foo')
        g.__setstate__({})
        self.assertEqual(g.gid, 100)
        self.assertEqual(g.name, 'Foo')

    def test31_SetstateDoesntReplaceExistingFeatures(self):
        g = Group(gid=100, name='Foo')
        g.__setstate__({'gid':200})
        self.assertEqual(g.gid, 100)
        self.assertEqual(g.name, 'Foo')

    def test32_SetstateUpdatesMissingFeatures(self):
        g = Group(gid=100, name='Foo')
        del g.__dict__['gid']
        g.__setstate__({'gid':200})
        self.assertEqual(g.gid, 200)
        self.assertEqual(g.name, 'Foo')

    def test90_PickleAttribute(self):
        import pickle
        g1 = Group(gid=100, name='Foo')
        p = pickle.dumps(g1)
        g2 = pickle.loads(p)
        self.assertEqual(g2.gid, 100)
        self.assertEqual(g2.name, 'Foo')

    def test95_PickleReference(self):
        import pickle
        ug1 = UserGroup(
                user  = User(uid=1,login='bar'),
                group = Group(gid=100, name='Foo'),
        )
        p = pickle.dumps(ug1)
        ug2 = pickle.loads(p)

        self.failUnless(isinstance(ug2.user, User))
        self.assertEqual(ug2.user.uid, 1)
        self.failIf('login' in ug2.user.__dict__)

        self.failUnless(isinstance(ug2.group, Group))
        self.assertEqual(ug2.group.gid, 100)
        self.failIf('name' in ug2.group.__dict__)















class ST(Immutable):
    a = Attribute(int)
    s = Attribute(str)
    f = Attribute(float, Default(2.1))

class TestImmutables(TestCase):

    def test10_Features(self):
        self.assertEqual(ST.features.keys(), ['a','s','f'])
        self.assertEqual(ST.a.attrName, 'a')

    def test20_ImmutableAsComponent(self):
        self.assertEqual(GCP(ST), CN(MN.parts+['ST']))

    def test30_Instances(self):
        st = ST(a=42, s='foo', f=3.14)
        self.assertEqual(st.a, 42)
        self.assertEqual(st.s, 'foo')
        self.assertEqual(st.f, 3.14)

    def test31_Normalization(self):
        st = ST(a='42', f='3.14')
        self.assertEqual(st.a, 42)
        self.assertEqual(st.s, None)
        self.assertEqual(st.f, 3.14)

    def test32_DefaultValues(self):
        st = ST()
        self.assertEqual(st.a, None)
        self.assertEqual(st.s, None)
        self.assertEqual(st.f, 2.1)

    def test33_InvalidAttributes(self):
        self.assertRaises(AttributeError, ST, x='foo')

    def test34_FromTuple(self):
        st = ST(42, 'foo', 3.14)
        self.assertEqual(st.a, 42)
        self.assertEqual(st.s, 'foo')
        self.assertEqual(st.f, 3.14)

    def test35_ToTuple(self):
        st = ST(42, 'foo', 3.14)
        self.assertEqual(tuple(st), (42,'foo',3.14))
        self.assertEqual(tuple(ST()), (None,None,2.1))

    def test40_ReadOnly(self):
        st = ST(a=42, s='foo')
        self.assertRaises(AttributeError, setattr, st, 'a', 1)
        self.assertRaises(AttributeError, setattr, st, 'x', 1)




class R(Entity):
    i = Attribute(int)

class E(Entity):
    i = Attribute(int)
    s = Attribute(str)
    r = Reference(R)

class TestEntityCreation(TestCase):

    def test10_SimpleInstance(self):
        e = E(i=1, s='42')
        self.assertEqual(e.i, 1)
        self.assertEqual(e.s, '42')

    def test11_EntityParentComponent(self):
        e = E()
        assert binding.getParentComponent(e) is None
        e = E()
        e.__dict__['_workspace'] = 'foo'
        assert binding.getParentComponent(e) == 'foo'

    def test12_InvalidAttribute(self):
        self.assertRaises(AttributeError, E, n=1)

    def test13_ReadOnlyAttribute(self):
        self.assertRaises(AttributeError, E, _oid=1)

    def test20_InstanceWithReference(self):
        e = E()
        self.assertEqual(e.r, None)
        e = E(r=R())
        assert isinstance(e.r, R)

    def test30_PrimitiveNormalization(self):
        assertEntityError(self,[('i',NormalizationError)], E, i='a')

    def test31_ReferenceNormalization(self):
        assertEntityError(self,[('r',NormalizationError)], E, r='a')

    def test40_DefaultValue(self):
        class E(Entity) :
            i = Attribute(int, Default(42))
        e = E()
        self.assertEqual(e.i, 42)

    def test41_DefaultValueNormalization(self):
        class E(Entity) :
            i = Attribute(int, Default('42'))
        e = E()
        self.assertEqual(e.i, 42)

    def test42_ComputeDefault(self):
        class E(Entity) :
            i = Attribute(int, ComputeDefault(lambda s,f:(lambda o,a,v:42)))
        e = E()
        self.assertEqual(e.i, 42)

    def test43_ComputeDefaultResultNormalization(self):
        class E(Entity) :
            i = Attribute(int, ComputeDefault((lambda s,f:(lambda o,a,v:'42'))))
        e = E()
        self.assertEqual(e.i, 42)




class TestAttributeUpdating(TestCase):

    def test10_SetattrNonexistingAttr(self):
        class E(Entity) : i = Attribute(int)
        e = E()
        self.assertEqual(e.i, None)
        self.assertRaises(AttributeError, setattr, e, 'x', 1)

    def test11_UpdateNonexistingAttr(self):
        class E(Entity) : i = Attribute(int)
        e = E()
        self.assertEqual(e.i, None)
        self.assertRaises(AttributeError, e._setattrs, x=1)

    def test13_ReadOnlyAttribute(self):
        e = E()
        self.assertRaises(AttributeError, setattr, e, '_oid', 1)
        self.assertRaises(AttributeError, delattr, e, '_oid')

    def test20_SetattrNormalization(self):
        class E(Entity):
            a = Attribute(int)
        e = E(a=1)
        self.assertEqual(e.a, 1)
        assertEntityError(
            self, [('a',NormalizationError)], e.__setattr__, 'a', 'x'
        )

    def test21_UpdateNormalization(self):
        class E(Entity):
            a = Attribute(int)
        e = E(a=1)
        self.assertEqual(e.a, 1)
        assertEntityError(self,[('a',NormalizationError)], e._setattrs, a='x')




class TestEntityInstances(TestCase):

    def test10_BindingAttributes(self):
        class E(Entity):
            d = binding.Make(dict)
            x = binding.Obtain('d')
        e = E()
        self.assertEqual(e.d, {})
        assert e.x is e.d



class X(Entity):
    a = b = c = Attribute(int)
    binding.metadata(PrimaryKey('a'))

class V(View):
    query = '[(x.a, x.b) for x in X]'


class TestViews(MockWorkspaceCase):

    def test(self):
        ws = settings.workspace()
        q = ws.parse(V.query)
        print V.query, q
        print dir(q)
        print q.expr
        print q.namespace
        print q.expr.namespace
        print q.relvars
        print
        a = q.expr.args[0]
        print a
        print dir(a)
        print a.namespace
        print a.pmap
        print (a.var, a.typ)



def suite():
    return TestSuite((
        makeSuite(TestTypeConfig),
        makeSuite(TestTypeClass),
        makeSuite(TestFeatureSorting),
        makeSuite(TestState),
        makeSuite(TestImmutables),
        makeSuite(TestEntityCreation),
        makeSuite(TestAttributeUpdating),
        makeSuite(TestEntityInstances),
        #makeSuite(TestViews),
    ))


if __name__ == '__main__':
    main(defaultTest='suite')
