# -*- coding: ISO-8859-2 -*-

from unittest import *
from wm4.api import *
from wm4.schema.api import *

from wm4.query.tests.app import *
from utils import *
from datetime import datetime
from lib4.timeutils import Datetime


class TestWorkspaceEnvironment(TestCase):

    def setUp(self):
        self.old = config.snapshot()
        config.loadConfig()

    def tearDown(self):
        ts = settings.transaction_service()
        if ts.isActive():
            ts.abort()
        config.swap(self.old)

    def test01_Pretest(self):
        ws = settings.workspace()
        assert isinstance(ws, Workspace)

    def test02_TransactionComponent(self):
        ws = settings.workspace()
        self.assertRaises(exceptions.OutsideTransaction,getattr,ws,'joinedTxn')

    def test03_InsindeTransaction(self):
        ws = settings.workspace()
        self.assertEqual(ws.inTransaction, False)

        storage.begin()
        self.assertEqual(ws.inTransaction, False)
        self.assertEqual(ws.joinedTxn, settings.transaction_service())
        self.assertEqual(ws.inTransaction, True)

        storage.commit()
        self.assertEqual(ws.inTransaction, False)

        storage.begin()
        self.assertEqual(ws.inTransaction, False)
        self.assertEqual(ws.joinedTxn, settings.transaction_service())
        self.assertEqual(ws.inTransaction, True)

        storage.abort()

    def test04_QueryOutsideTransaction(self):
        ws = settings.workspace()
        self.assertRaises(exceptions.OutsideTransaction, ws, 'qry')

    def test10_TypeConfig(self):
        c0 = config.Config.current()
        config.loadConfig()
        c1 = config.Config.current()
        pm1 = TypeConfig(User)()
        config.loadConfig()
        c2 = config.Config.current()
        pm2 = TypeConfig(User)()
        assert c1 is not c0
        assert c2 is not c0
        assert c1 is not c2
        assert pm1 is not pm2

    # XXX
    # def test10_WorkspaceIsolation(self):
    # def test20_TwoIndependentWorkspaces(self):







class C1(SchemaComponent):
    pass
class E1(Entity):
    pass
class M2(SchemaComponent):
    pass
class C2(SchemaComponent):
    pass
class E21(Entity):
    pass
class E22(Entity):
    pass

SCHEMA_ELEMENTS = '''\
[Schema Elements]
c1    = "%(mod)s.C1"
c1/e1 = "%(mod)s.E1"

[Schema Elements m2]
c2 = "%(mod)s.C2"
c2/E21 = importString("%(mod)s.E21")
c2/E22 = "%(mod)s.E22"
''' % {'mod':__name__}

class TestSchemaElements(TestCase):

    def setUp(self):
        self.state = config.snapshot()
        config.loadConfig('pkgfile:wm4/core.ini')

    def tearDown(self):
        config.swap(self.state)

    @with_ini(SCHEMA_ELEMENTS)
    def test10_ElementsLookupAndCreation(self):
        ws = settings.workspace()
        c1 = ws['c1']
        assert isinstance(c1, C1)
        assert c1.workspace is ws
        e1 = ws['c1/e1']
        assert e1 is E1

        m2c2 = ws['m2/c2']
        assert isinstance(m2c2, C2)
        assert m2c2.workspace is ws
        m2c2e21 = ws['m2/c2/E21']
        assert m2c2e21 is E21
        m2c2e22 = ws['m2/c2/E22']
        assert m2c2e22 is E22

    @with_ini(SCHEMA_ELEMENTS)
    def test12_ElementsCaching(self):
        ws = settings.workspace()
        for p in('c1', '/c1/e1', 'm2/c2','m2/c2/E21', 'm2/c2/E22'):
            i1 = ws[p]
            i2 = ws[p]
            assert i1 is i2, 'Element %r is not cached' % p

    @with_ini(SCHEMA_ELEMENTS)
    def test24_TraversalErrors(self):
        ws = settings.workspace()
        self.assertRaises(ElementNotFound, ws.getElement, 'foo')

    @with_ini(SCHEMA_ELEMENTS)
    def test20_GetPath(self):
        ws = settings.workspace()
        self.assertEqual(ws.getPath(E1), 'c1/e1')
        self.assertEqual(ws.getPath(E21), 'm2/c2/E21')
        self.assertEqual(ws.getPath(E22), 'm2/c2/E22')
        c1 = ws['c1']
        self.assertEqual(ws.getPath(c1.__class__), 'c1')
        self.assertRaises(ElementNotFound, ws.getPath, 'c1')
        self.assertRaises(ElementNotFound, ws.getPath, object)





class TestEditingContext(TestCase):

    def setUp(self):
        self.old = config.snapshot()
        config.loadConfig()
        self.ws = settings.workspace()
        self.ts = settings.transaction_service()

    def tearDown(self):
        if self.ts.isActive():
            self.ts.abort()
        config.swap(self.old)

    def test10_ContextForOutsideTransaction(self):
        self.assertRaises(
            exceptions.OutsideTransaction, self.ws.contextFor, User
        )

    def test11_ContextIsolationAndPersistence(self):
        storage.begin()
        uctx = self.ws.contextFor(User)
        gctx = self.ws.contextFor(Group)
        self.failIf(uctx is gctx)
        uctx2 = self.ws.contextFor(User)
        self.failUnless(uctx is uctx2)
        storage.commit()

    def test12_ContextAsTransactionComponent(self):

        class FakeEnt:
            data = (None,None,None)
            def _finishTransaction(self, ws, ts, commited):
                self.data = (ws, ts, commited)

        e = FakeEnt()
        storage.begin()
        ctx = self.ws.contextFor(FakeEnt)
        ctx.cache[1] = e
        storage.commit()
        self.assertEqual(e.data, (self.ws, self.ts, True))

        # context has not joined transaction
        e.data = (None, None, None)
        storage.begin()
        storage.abort()
        self.assertEqual(e.data, (None, None, None))

        # context has joined transaction with commit
        e.data = (None, None, None)
        storage.begin()
        self.ws.contextFor(FakeEnt)
        storage.abort()
        self.assertEqual(e.data, (self.ws, self.ts, False))

        # context has joined transaction with abort
        e.data = (None, None, None)
        storage.begin()
        self.ws.contextFor(FakeEnt)
        storage.commit()
        self.assertEqual(e.data, (self.ws, self.ts, True))







class TestPreload(MockWorkspaceCase):

    def test10_PreloadUnmanagedInstance(self):
        storage.begin()
        u = User(uid=1, login='baz', name='Baz')
        pu = self.ws.preload(u)
        self.failUnless(u is pu)
        self.assertEqual(u._oid, 1)
        self.failUnless(u._workspace is self.ws)
        storage.commit()

    def test11_PreloadSameManagedInstance(self):
        storage.begin()
        u = self.ws.preload(User(uid=1, login='baz', name='Baz'))
        pu = self.ws.preload(u)
        self.failUnless(u is pu)
        self.assertEqual(u._oid, 1)
        self.failUnless(u._workspace is self.ws)
        storage.commit()

    def test12_PreloadOtherUnmanagedInstance(self):
        storage.begin()
        u = self.ws.preload(User(uid=1, login='baz', name='Baz'))
        u2 = User(uid=1, login='bar', name='Bar')
        self.failIf(u is u2)
        pu = self.ws.preload(u2)
        self.failUnless(u is pu)
        self.assertEqual(u._oid, 1)
        self.assertEqual(u.uid, 1)
        self.assertEqual(u.login, 'baz')
        self.assertEqual(u.name, 'Baz')
        self.failUnless(u._workspace is self.ws)
        storage.commit()

    def test13_PreloadMissingAttribute(self):
        storage.begin()
        u = self.ws.preload(User(uid=1, login='baz', name='Baz'))
        del u.__dict__['login']
        u2 = User(uid=1, login='bar', name='Bar')
        pu = self.ws.preload(u2)
        self.failUnless(u is pu)
        self.assertEqual(u._oid, 1)
        self.assertEqual(u.uid, 1)
        self.assertEqual(u.login, 'bar')
        self.assertEqual(u.name, 'Baz')
        self.failUnless(u._workspace is self.ws)
        storage.commit()

    def test20_PreloadReferences(self):
        storage.begin()
        ug = UserGroup(
            user  = User(uid=1, login='baz', name='Baz'),
            group = Group(gid=20, name='bar'),
        )
        pug = self.ws.preload(ug)
        self.failUnless(ug is pug)
        self.failUnless(ug._workspace is self.ws)
        self.failUnless(ug.user is pug.user)
        self.failUnless(ug.user._workspace is self.ws)
        self.failUnless(ug.group is pug.group)
        self.failUnless(ug.group._workspace is self.ws)
        storage.commit()

    def test21_PreloadNullReferences(self):
        storage.begin()
        f = Folder(
            name = 'root',
        )
        pf = self.ws.preload(f)
        self.failUnless(f is pf)
        self.failUnless(f._workspace is self.ws)
        self.failUnless(f.parent is None)
        storage.commit()








class TestAddEntity(MockWorkspaceCase):

    def test10_AddEntityOutsideTransaction(self):
        o = User(uid=1, login='foo')
        self.assertRaises(exceptions.OutsideTransaction, self.ws.add, o)

    def test12_AddAddedEntityIsNoop(self):
        storage.begin()
        o = User(uid=1, login='foo')
        o.__dict__['_workspace'] = self.ws
        self.ws.add(o)
        storage.abort()

    def test12_AddEntityFromOtherWorkspace(self):
        storage.begin()
        o = User(uid=1, login='foo')
        o.__dict__['_workspace'] = self
        self.assertRaises(ValueError, self.ws.add, o)
        storage.abort()

    def test13_AddEntities(self):
        self.expects(transaction='begin')
        self.expects(sql=(
                'INSERT INTO "user" (uid, name, login, mtime) '
                'VALUES (1, NULL, \'foo\', NULL)'
        ))
        self.expects(sql=(
                'INSERT INTO "user" (uid, name, login, mtime) '
                'VALUES (2, NULL, \'bar\', NULL)'
        ))
        self.expects(transaction='commit')

        storage.begin()

        o1 = self.ws.add(User(uid=1, login='foo'))
        self.assertEqual(o1._workspace, self.ws)
        self.assertEqual(o1._oid, 1)
        self.assertEqual(o1.uid, 1)
        self.assertEqual(o1.login, 'foo')

        o2 = self.ws.add(User(uid=2, login='bar'))
        self.assertEqual(o2._workspace, self.ws)
        self.assertEqual(o2._oid, 2)
        self.assertEqual(o2.uid, 2)
        self.assertEqual(o2.login, 'bar')

        storage.commit()

        assert 1 in self.ws.contexts[User].cache
        assert 2 in self.ws.contexts[User].cache

    def test14_AddDuplicateEntity(self):
        self.expects(transaction='begin')
        self.expects(sql=(
                'INSERT INTO "user" (uid, name, login, mtime) '
                'VALUES (1, NULL, \'foo\', NULL)'
        ))
        self.expects(transaction='abort')

        storage.begin()
        o = self.ws.add(User(uid=1, login='foo'))
        self.assertRaises(ValueError, self.ws.add, User(uid=1, login='bar'))
        storage.abort()

    def test15_DeactivationAfterCommit(self):
        self.expects(transaction='begin')
        self.expects(sql=(
                'INSERT INTO "user" (uid, name, login, mtime) '
                'VALUES (1, NULL, \'foo\', NULL)'
        ))
        self.expects(transaction='commit')

        storage.begin()
        o = self.ws.add(User(uid=1, login='foo'))
        storage.commit()

        # non transactional attributes
        self.assertEqual(o._workspace, self.ws)
        self.assertEqual(o._oid, 1)
        assert 1 in self.ws.contexts[User].cache

        # transactional attributes are deactivated
        assert 'uid' not in o.__dict__
        assert 'login' not in o.__dict__
        assert 'name' not in o.__dict__

    def test16_CacheAfterAbort(self):
        self.expects(transaction='begin')
        self.expects(sql=(
                'INSERT INTO "user" (uid, name, login, mtime) '
                'VALUES (1, NULL, \'foo\', NULL)'
        ))
        self.expects(transaction='abort')

        storage.begin()
        o = self.ws.add(User(uid=1, login='foo'))
        storage.abort()
        assert 1 not in self.ws.contexts[User].cache
        assert '_workspace' not in o.__dict__
        assert 'uid' not in o.__dict__
        assert 'login' not in o.__dict__
        assert 'name' not in o.__dict__

    def test20_AddEntityWithReferences(self):
        self.expects(transaction='begin')
        self.expects(sql=(
                'INSERT INTO user_group ("user", "group") '
                "VALUES (2, 200)"
        ))
        self.expects(transaction='commit')

        storage.begin()

        u = self.ws.preload(User(uid=2, login='boo'))
        g = self.ws.preload(Group(gid=200))

        ug = self.ws.add(UserGroup(user=u, group=g))

        assert isinstance(ug, UserGroup)
        assert ug.user is u and ug.group is g
        assert (2,200) in self.ws.contexts[UserGroup].cache
        assert 2 in self.ws.contexts[User].cache
        assert 200 in self.ws.contexts[Group].cache
        assert ug._workspace == self.ws
        assert u._workspace == self.ws and g._workspace == self.ws

        storage.commit()

    def test21_AddEntityWithTransientReferences(self):
        self.expects(transaction='begin')
        self.expects(sql=(
                'INSERT INTO "user" (uid, name, login, mtime) '
                "VALUES (2, NULL, 'boo', NULL)"
        ))
        self.expects(sql=(
                'INSERT INTO "group" (gid, name) '
                "VALUES (200, NULL)"
        ))
        self.expects(sql=(
                'INSERT INTO user_group ("user", "group") '
                "VALUES (2, 200)"
        ))
        self.expects(transaction='commit')

        storage.begin()
        u = User(uid=2, login='boo')
        g = Group(gid=200)
        ug = self.ws.add(UserGroup(user=u, group=g))
        assert isinstance(ug, UserGroup)
        assert ug.user is u and ug.group is g
        assert (2,200) in self.ws.contexts[UserGroup].cache
        assert 2 in self.ws.contexts[User].cache
        assert 200 in self.ws.contexts[Group].cache
        assert ug._workspace == self.ws
        assert u._workspace == self.ws and g._workspace == self.ws
        storage.commit()

    def test22_AddEntityWithReferencesFromOtherWorkspace(self):
        storage.begin()
        u = User(uid=2, login='boo')
        u.__dict__['_workspace'] = 'OTHERWS'
        g = Group(gid=200)
        g.__dict__['_workspace'] = 'OTHERWS'
        self.assertRaises(
            ValueError, self.ws.add, UserGroup(user=u, group=g)
        )
        storage.abort()

    def test23_AddEntityNullReferences(self):
        self.expects(transaction='begin')
        self.expects(sql=(
            "INSERT INTO folder (name, parent) VALUES ('root', NULL)"
        ))
        self.expects(transaction='commit')

        storage.begin()
        f = self.ws.add(Folder(name='root', parent=None))
        self.assertEqual(f._workspace, self.ws)
        self.assertEqual(f.name, 'root')
        self.assertEqual(f.parent, None)
        storage.commit()


    def test30_ComputeNextValForSerial(self):
        self.expects(transaction='begin')
        self.expects(sql=("SELECT NEXTVAL('se_s_seq')"), res=[[32]])
        self.expects(sql=('INSERT INTO se (s) VALUES (32)'))
        self.expects(transaction='commit')

        storage.begin()
        o = self.ws.new(SE)
        storage.commit()

    def test40_AddAfter(self):
        pass





from adt4.serial import Serial
class SE(Entity):
    s = Attribute(Serial, Required)
    binding.metadata(PrimaryKey('s'))




class TestGetEntity(MockWorkspaceCase):

    cfgURLs = ('pkgfile:wm4/core.ini',)

    def test01_GetOutsideTransaction(self):
        self.assertRaises(exceptions.OutsideTransaction, self.ws.get, User, 1)

    def test10_GetUncachedEntity(self):
        self.expects(transaction='begin')
        self.expects(
            sql= 'SELECT uid, name, login, mtime FROM "user" WHERE uid = 1',
            res = [[1,'Foo','foo',None]]
        )
        self.expects(transaction='commit')

        storage.begin()
        o = self.ws.get(User, 1)
        assert isinstance(o, User), repr(o)
        self.assertEqual(o._oid, 1)
        self.assertEqual(o._workspace, self.ws)
        self.assertEqual(o.uid, 1)
        self.assertEqual(o.name, 'Foo')
        self.assertEqual(o.login, 'foo')
        self.assertEqual(o.mtime, None)
        assert 1 in self.ws.contexts[User].cache
        storage.commit()

        # object is deactivated upon commit
        self.assertEqual(o._oid, 1)
        self.assertEqual(o._workspace, self.ws)
        assert 'uid' not in o.__dict__
        assert 'login' not in o.__dict__
        assert 1 in self.ws.contexts[User].cache

    def test12_GetCachedEntity(self):
        self.expects(transaction='begin')
        self.expects(
            sql= 'SELECT uid, name, login, mtime FROM "user" WHERE uid = 1',
            res = [[1,'Foo','foo',None]]
        )
        self.expects(transaction='commit')
        storage.begin()
        o1 = self.ws.get(User, 1)
        o2 = self.ws.get(User, 1)
        assert o1 is o2
        storage.commit()

    def test14_GetDeactivatedEntity(self):
        self.expects(transaction='begin')
        self.expects(
            sql= 'SELECT uid, name, login, mtime FROM "user" WHERE uid = 1',
            res = [[1,'Foo','foo',None]]
        )
        self.expects(transaction='commit')
        storage.begin()
        o1 = self.ws.get(User, 1)
        storage.commit()
        storage.begin()
        o2 = self.ws.get(User, 1)
        assert o1 is o2
        storage.commit()

    def test15_GetDeactivatedEntityAfterCommit(self):
        self.expects(transaction='begin')
        self.expects(
            sql= 'SELECT uid, name, login, mtime FROM "user" WHERE uid = 1',
            res = [[1,'Foo','foo',None]]
        )
        self.expects(transaction='commit')
        self.expects(transaction='begin')
        self.expects(
            sql= 'SELECT uid, name, login, mtime FROM "user" WHERE uid = 1',
            res = [[1,'Foo','foo',None]]
        )
        self.expects(transaction='commit')

        storage.begin()
        o1 = self.ws.get(User, 1)
        storage.commit()
        assert 1 in self.ws.contexts[User].cache

        storage.begin()
        o2 = self.ws.get(User, 1)
        assert o1 is o2
        assert o2.name == 'Foo'
        storage.commit()

    def test16_GetDeactivatedEntityAfterAbort(self):
        self.expects(transaction='begin')
        self.expects(
            sql= 'SELECT uid, name, login, mtime FROM "user" WHERE uid = 1',
            res = [[1,'Foo','foo',None]]
        )
        self.expects(transaction='abort')
        self.expects(transaction='begin')
        self.expects(
            sql= 'SELECT uid, name, login, mtime FROM "user" WHERE uid = 1',
            res = [[1,'Foo','foo',None]]
        )
        self.expects(transaction='commit')

        storage.begin()
        o1 = self.ws.get(User, 1)
        storage.abort()
        assert 1 in self.ws.contexts[User].cache

        storage.begin()
        o2 = self.ws.get(User, 1)
        assert o1 is o2
        assert o2.name == 'Foo'
        storage.commit()

    def test17_GetNonexistentOid(self):
        self.expects(transaction='begin')
        self.expects(
            sql= 'SELECT uid, name, login, mtime FROM "user" WHERE uid = 1',
            res = [],
        )
        self.expects(transaction='commit')
        storage.begin()
        self.assertRaises(KeyError, self.ws.get, User, 1)
        storage.commit()

    def test18_GetNonexistentOidWithDefault(self):
        self.expects(transaction='begin')
        self.expects(
            sql= 'SELECT uid, name, login, mtime FROM "user" WHERE uid = 1',
            res = [],
        )
        self.expects(transaction='commit')
        storage.begin()
        self.assertEqual(self.ws.get(User, 1, default='foo'), 'foo')
        storage.commit()

    def test20_ReferencePreloading(self):
        self.expects(transaction='begin')
        self.expects(
            sql= 'SELECT "user", "group" FROM user_group ' +\
                 'WHERE ("group" = 200) AND ("user" = 2)',
            res = [[2,200]]
        )
        self.expects(transaction='commit')
        storage.begin()
        u = self.ws.preload(User(uid=2, login='boo'))
        assert u._oid == 2
        g = self.ws.preload(Group(gid=200))
        assert g._oid == 200
        ug = self.ws.get(UserGroup, (u,g))
        assert isinstance(ug, UserGroup)
        assert (2,200) in self.ws.contexts[UserGroup].cache
        assert 2 in self.ws.contexts[User].cache
        assert 200 in self.ws.contexts[Group].cache
        assert ug.user is u
        assert ug.group is g
        storage.commit()

    def test30_CacheWeakness(self):
        self.expects(transaction='begin')
        self.expects(
            sql= 'SELECT uid, name, login, mtime FROM "user" WHERE uid = 1',
            res = [[1,'Foo','foo',None]]
        )
        self.expects(transaction='commit')
        storage.begin()
        o = self.ws.get(User, 1)
        storage.commit()
        cache = self.ws.contexts[User].cache
        assert 1 in cache
        del o
        assert 1 not in cache

    @with_ini(
        '[Schema Elements]\n'
        'testusers = "wm4.schema.workspace.SchemaComponent"\n'
        'testusers/User = "wm4.query.tests.app.User"\n')
    def test40_GetWithSchemaPath(self):
        self.expects(transaction='begin')
        self.expects(
            sql= 'SELECT uid, name, login, mtime FROM "user" WHERE uid = 1',
            res = [[1,'Foo','foo',None]]
        )
        self.expects(transaction='commit')
        storage.begin()
        o = self.ws.get('testusers/User', 1)
        storage.commit()




class TestQueries(MockWorkspaceCase):

    def test10_FindAll(self):
        self.expects(transaction='begin')
        self.expects(
            sql= 'SELECT uid, name, login, mtime FROM "user"',
            res = [
                [1, 'Bob','bob','2004-12-31'],
                [2, u'K'.encode('utf-8'),'kun',None]
            ]
        )
        self.expects(transaction='commit')
        storage.begin()
        res = list(self.ws.find(User))
        self.assertEqual(len(res), 2)
        self.assertEqual(res[0].uid, 1)
        self.assertEqual(res[0].mtime, Datetime(2004,12,31))
        self.assertEqual(res[1].uid, 2)
        self.assertEqual(res[1].name, u'K')
        storage.commit()
        assert 1 in self.ws.contexts[User].cache
        assert 2 in self.ws.contexts[User].cache

    def test11_FindSome(self):
        self.expects(transaction='begin')
        self.expects(
            sql= ('SELECT uid, name, login, mtime FROM "user" '
                  'WHERE name = (\'%s\')') % u'K'.encode('utf-8'),
            res = [[2, u'K'.encode('utf-8'),'kun',None]]
        )
        self.expects(transaction='commit')
        storage.begin()
        res = list(self.ws.find(User, name=u'K'))
        self.assertEqual(len(res), 1)
        self.assertEqual(res[0].uid, 2)
        self.assertEqual(res[0].name, u'K')
        storage.commit()
        assert 2 in self.ws.contexts[User].cache

    def test20_SelectAll(self):
        self.expects(transaction='begin')
        self.expects(
            sql= 'SELECT uid, name, login, mtime FROM "user"',
            res = [
                [1, 'Bob','bob','2004-12-31'],
                [2, u'K'.encode('utf-8'),'kun',None]
            ]
        )
        self.expects(transaction='commit')
        storage.begin()
        res = list(self.ws('[u for u in User]'))
        self.assertEqual(len(res), 2)
        self.assertEqual(res[0].uid, 1)
        self.assertEqual(res[0].mtime, Datetime(2004,12,31))
        self.assertEqual(res[1].uid, 2)
        self.assertEqual(res[1].name, u'K')
        storage.commit()
        assert 1 in self.ws.contexts[User].cache
        assert 2 in self.ws.contexts[User].cache

    def test21_SelectSome(self):
        k = u'K'
        self.expects(transaction='begin')
        self.expects(
            sql= ('SELECT uid, name, login, mtime FROM "user" '
                  'WHERE name = (\'%s\')') % k.encode('utf-8'),
            res = [[2, k.encode('utf-8'),'kun',None]]
        )
        self.expects(transaction='commit')
        storage.begin()
        res = list(self.ws('[u for u in User if u.name == k]'))
        self.assertEqual(len(res), 1)
        self.assertEqual(res[0].uid, 2)
        self.assertEqual(res[0].name, k)
        storage.commit()
        assert 2 in self.ws.contexts[User].cache

    def test22_CacheWeakness(self):
        self.expects(transaction='begin')
        self.expects(
            sql= 'SELECT uid, name, login, mtime FROM "user"',
            res = [
                [1, 'Bob','bob','2004-12-31'],
                [2, u'K'.encode('utf-8'),'kun',None]
            ]
        )
        self.expects(transaction='commit')
        storage.begin()
        for o in self.ws('[u for u in User]'):
            pass
        del o
        storage.commit()
        assert 1 not in self.ws.contexts[User].cache
        assert 2 not in self.ws.contexts[User].cache

    def test23_PreloadExisting(self):
        self.expects(transaction='begin')
        self.expects(
            sql= 'SELECT uid, name, login, mtime FROM "user"',
            res = [
                [1, 'Bob','bob','2004-12-31'],
                [2, u'K'.encode('utf-8'),'kun',None]
            ]
        )
        self.expects(transaction='commit')
        storage.begin()
        u = self.ws.preload(User(uid=1,login='bob'))
        ul = [o for o in self.ws('[u for u in User]')]
        assert u is ul[0]
        storage.commit()

    def test24_PreloadWithTuple(self):
        self.expects(transaction='begin')
        self.expects(
            sql= 'SELECT login, uid, name, login, mtime FROM "user"',
            res = [
                ['bob', 1, 'Bob','bob','2004-12-31'],
                ['kun', 2, u'K'.encode('utf-8'),'kun',None]
            ]
        )
        self.expects(transaction='commit')
        storage.begin()
        ul = [o for o in self.ws('[(u.login,u) for u in User]')]
        self.assertEqual(ul[0][0], 'bob')
        self.failUnless(isinstance(ul[0][1], User))
        self.failUnless(ul[0][1]._workspace is self.ws)
        self.assertEqual(ul[1][0], 'kun')
        self.failUnless(isinstance(ul[1][1], User))
        self.failUnless(ul[1][1]._workspace is self.ws)
        storage.commit()

    def test30_PreloadingReferences(self):
        self.expects(transaction='begin')
        self.expects(
            sql= 'SELECT name, parent FROM folder',
            res = [
                ['f2', 'f1'],
                ['f0', None],
                ['f1', 'f0'],
            ]
        )
        self.expects(transaction='commit')
        storage.begin()
        f2, f0, f1 = tuple(self.ws('[f for f in Folder]'))
        self.assertEqual(f0.parent, None)
        self.assertEqual(f0.name, 'f0')
        self.assertEqual(f1.parent, f0)
        self.assertEqual(f1.name, 'f1')
        self.assertEqual(f2.parent, f1)
        self.assertEqual(f2.name, 'f2')
        self.failUnless(f0._workspace is self.ws)
        self.failUnless(f1._workspace is self.ws)
        self.failUnless(f2._workspace is self.ws)
        storage.commit()










class TestUpdate(MockWorkspaceCase):

    def test01_UpdateOutsideTransaction(self):
        u = User(uid=1, login='foo', name='Foo')
        u.__dict__['_workspace'] = self.ws
        self.assertRaises(
            exceptions.OutsideTransaction, setattr, u, 'name', 'Boo'
        )

    def test10_UpdateFreshInstance(self):
        self.expects(transaction='begin')
        self.expects(
            sql= ('SELECT uid, name, login, mtime FROM "user" '
                  'WHERE uid = 1 FOR UPDATE'),
            res = [[1,'Foo','foo',None]]
        )
        self.expects(
            sql= 'UPDATE "user" SET name=\'Boo\' WHERE uid = 1',
        )
        self.expects(transaction='commit')
        storage.begin()
        u = self.ws.preload(User(uid=1, login='foo', name='Foo'))
        u.name = 'Boo'
        storage.commit()
        assert not self.ws.contexts[User].dirty
        assert not self.ws.contexts[User].saved

    def test11_UpdateWithNoChange(self):
        self.expects(transaction='begin')
        self.expects(
            sql= ('SELECT uid, name, login, mtime FROM "user" '
                  'WHERE uid = 1 FOR UPDATE'),
            res = [[1,'Foo','foo',None]]
        )
        self.expects(transaction='commit')
        storage.begin()
        u = self.ws.preload(User(uid=1, login='foo', name='Foo'))
        u.name = 'Foo'
        storage.commit()
        assert not self.ws.contexts[User].dirty
        assert not self.ws.contexts[User].saved

    def test12_UpdateOid(self):
        self.expects(transaction='begin')
        self.expects(
            sql= ('SELECT uid, name, login, mtime FROM "user" '
                  'WHERE uid = 1 FOR UPDATE'),
            res = [[1,'Foo','foo',None]]
        )
        self.expects(
            sql= 'UPDATE "user" SET uid=2 WHERE uid = 1',
        )
        self.expects(transaction='commit')
        storage.begin()
        u = self.ws.preload(User(uid=1, login='foo', name='Foo'))
        u.uid = 2
        storage.commit()
        assert 1 not in self.ws.contexts[User].cache
        assert 2 in self.ws.contexts[User].cache
        assert u._oid == 2

    def test13_UpdateReference(self):
        self.expects(transaction='begin')
        self.expects(
            sql= ('SELECT "user", "group" FROM user_group '
                  'WHERE ("group" = 10) AND ("user" = 1)'),
            res = [[1,10]]
        )
        self.expects(
            sql= ('SELECT "user", "group" FROM user_group '
                  'WHERE ("group" = 10) AND ("user" = 1) FOR UPDATE'),
            res = [[1,10]]
        )
        self.expects(
            sql= ('UPDATE user_group SET "group"=20 '
                  'WHERE ("group" = 10) AND ("user" = 1)'),
        )
        self.expects(transaction='commit')
        storage.begin()
        u1 = self.ws.preload(User(uid=1, login='baz', name='Baz'))
        g10 = self.ws.preload(Group(gid=10, name='foo'))
        g20 = self.ws.preload(Group(gid=20, name='bar'))
        ug = self.ws.get(UserGroup, (u1,g10))
        assert ug.user is u1
        assert ug.group is g10
        ug.group = g20
        storage.commit()








class TestRemove(MockWorkspaceCase):

    def test01_RemoveOutsideTransaction(self):
        u = User(uid=1, login='foo', name='Foo')
        u.__dict__['_workspace'] = self.ws
        self.assertRaises(
            exceptions.OutsideTransaction, self.ws.remove, u
        )

    def test10_Delete(self):
        self.expects(transaction='begin')
        self.expects(sql= 'DELETE FROM "user" WHERE uid = 1')
        self.expects(transaction='commit')
        storage.begin()
        u = self.ws.preload(User(uid=1, login='foo', name='Foo'))
        self.ws.remove(u)
        storage.commit()
        assert not self.ws.contexts[User].dirty
        assert not self.ws.contexts[User].saved
        assert 1 not in self.ws.contexts[User].cache







class TestQueryFeature(MockWorkspaceCase):

    def test01_QueryFeature(self):
        self.expects(transaction='begin')
        self.expects(
            sql= ('SELECT g.gid, g.name '
                  'FROM user_group AS ug '
                  'CROSS JOIN "group" AS g '
                  'WHERE (ug."user" = 1) AND (ug."group" = g.gid)'),
            res = [[10,'Foo'], [20,'Bar']]
        )
        self.expects(transaction='commit')

        storage.begin()
        u = self.ws.preload(User(uid=1, login='foo', name='Foo'))

        u.groups
        u.groups

        gcache = self.ws.contexts[Group].cache
        assert 10 in gcache
        assert 20 in gcache

        self.assertEqual(gcache[10].gid, 10)
        self.assertEqual(gcache[10].name, 'Foo')
        self.assertEqual(gcache[20].gid, 20)
        self.assertEqual(gcache[20].name, 'Bar')

        storage.commit()
        assert 'groups' not in u.__dict__
        assert 10 not in gcache
        assert 20 not in gcache

    def test10_ReferenceEnd(self):
        self.expects(transaction='begin')
        self.expects(
            sql= ('SELECT "user", "group" FROM user_group '
                  'WHERE "user" = 1 ORDER BY "group"'),
            res = [[1,10], [1,20]]
        )
        self.expects(transaction='commit')

        storage.begin()
        u = self.ws.preload(User(uid=1, login='foo', name='Foo'))
        u.ugroups
        u.ugroups

        ugcache = self.ws.contexts[UserGroup].cache
        assert (1,10) in ugcache
        assert (1,20) in ugcache
        gcache = self.ws.contexts[Group].cache
        assert 10 in gcache
        assert 20 in gcache

        self.assertEqual(ugcache[(1,10)].user.uid, 1)
        self.assertEqual(ugcache[(1,10)].group.gid, 10)
        self.assertEqual(ugcache[(1,20)].user.uid, 1)
        self.assertEqual(ugcache[(1,20)].group.gid, 20)

        storage.commit()

        assert 'ugroups' not in u.__dict__
        assert (1,10) not in ugcache
        assert (1,20) not in ugcache
        assert 10 not in gcache
        assert 20 not in gcache

    def test23_PreloadExistingReference(self):
        self.expects(transaction='begin')
        self.expects(
            sql= 'SELECT "user", "group" FROM user_group',
            res = [[1,10], [1,20]]
        )
        self.expects(transaction='commit')

        storage.begin()
        u1 = self.ws.preload(User(uid=1,login='bob'))
        ul = [o for o in self.ws('[ug for ug in UserGroup]')]
        assert ul[0].user is u1
        storage.commit()






def before(evt):
    evt.obj._workspace('"before"')

def after(evt):
    evt.obj._workspace('"after"')

class TestWorkspaceEvents(MockWorkspaceCase):

    cfgURLs = ('pkgfile:wm4/core.ini',)

    @with_ini((
        '[Event Handler wm4.schema.workspace.BeforeAdd]\n'
        'wm4.query.tests.app.User = "%(mod)s.before"\n'
        '[Event Handler wm4.schema.workspace.AfterAdd]\n'
        'wm4.query.tests.app.User = "%(mod)s.after"\n'
    ) % dict(mod=__name__))
    def test10_AddTrigger(self):
        self.expects(transaction='begin')
        self.expects(sql="SELECT 'before'", res=[['before']])
        self.expects(sql=(
                'INSERT INTO "user" (uid, name, login, mtime) '
                'VALUES (1, NULL, \'foo\', NULL)'
        ))
        self.expects(sql="SELECT 'after'", res=[['after']])
        self.expects(transaction='commit')

        storage.begin()
        o1 = self.ws.add(User(uid=1, login='foo'))
        self.assertEqual(o1._oid, 1)
        self.assertEqual(o1.uid, 1)
        self.assertEqual(o1.login, 'foo')
        storage.commit()

    @with_ini((
        '[Event Handler wm4.schema.workspace.BeforeUpdate]\n'
        'wm4.query.tests.app.User = "%(mod)s.before"\n'
        '[Event Handler wm4.schema.workspace.AfterUpdate]\n'
        'wm4.query.tests.app.User = "%(mod)s.after"\n'
    ) % dict(mod=__name__))
    def test20_UpdateTrigger(self):
        self.expects(transaction='begin')
        self.expects(
            sql= ('SELECT uid, name, login, mtime FROM "user" '
                  'WHERE uid = 1 FOR UPDATE'),
            res = [[1,'Foo','foo',None]])
        self.expects(sql="SELECT 'before'", res=[['before']])
        self.expects(
            sql= 'UPDATE "user" SET name=\'Boo\' WHERE uid = 1',)
        self.expects(sql="SELECT 'after'", res=[['after']])
        self.expects(transaction='commit')

        storage.begin()
        u = self.ws.preload(User(uid=1, login='foo', name='Foo'))
        u.name = 'Boo'
        storage.commit()

    @with_ini((
        '[Event Handler wm4.schema.workspace.BeforeRemove]\n'
        'wm4.query.tests.app.User = "%(mod)s.before"\n'
        '[Event Handler wm4.schema.workspace.AfterRemove]\n'
        'wm4.query.tests.app.User = "%(mod)s.after"\n'
    ) % dict(mod=__name__))
    def test30_RemoveTrigger(self):
        self.expects(transaction='begin')
        self.expects(sql="SELECT 'before'", res=[['before']])
        self.expects(sql= 'DELETE FROM "user" WHERE uid = 1')
        self.expects(sql="SELECT 'after'", res=[['after']])
        self.expects(transaction='commit')

        storage.begin()
        u = self.ws.preload(User(uid=1, login='foo', name='Foo'))
        self.ws.remove(u)
        storage.commit()


def suite():
    return TestSuite((
        makeSuite(TestWorkspaceEnvironment),
        makeSuite(TestSchemaElements),
        makeSuite(TestEditingContext),
        makeSuite(TestPreload),
        makeSuite(TestAddEntity),
        makeSuite(TestGetEntity),
        makeSuite(TestQueries),
        makeSuite(TestUpdate),
        makeSuite(TestRemove),
        makeSuite(TestQueryFeature),
        makeSuite(TestWorkspaceEvents),
    ))


if __name__ == '__main__':
    main(defaultTest='suite')
