from wm4.api import *
from lib4.imports import importString as imps, importObject
from wm4.storage.transactions import TransactionComponent
import sys, weakref
from posixpath import join, normpath

from elements import *
from features import *
from constraints import *
from interfaces import ISchemaElementFactory

__all__ = [
    'WorkspaceEvent', 'BeforeAdd', 'AfterAdd', 'BeforeUpdate',
    'AfterUpdate', 'BeforeRemove', 'AfterRemove',
    'ElementNotFound',
    'ElementRegistry', 'SchemaComponentMeta', 'SchemaComponent',
    'EditingContext', 'DefaultEditingContext', 'Workspace',
]


class WorkspaceEvent(events.Event):
    __slots__ = 'obj',

class BeforeAdd(WorkspaceEvent): pass
class AfterAdd(WorkspaceEvent): pass
class BeforeUpdate(WorkspaceEvent): pass
class AfterUpdate(WorkspaceEvent): pass
class BeforeRemove(WorkspaceEvent): pass
class AfterRemove(WorkspaceEvent): pass


class ElementNotFound(LookupError):
    '''Schema element was not found'''


ElementRegistry = config.Registry('wm4.schema.workspace.ElementRegistry')



def schema_elements(parser, section, name, value, lineInfo):
    section = section.split()[2:]
    if section:
        name = '/'.join(section) + '/' + name
    if name.startswith('/'):
        name = name[1:]
    ElementRegistry.setFactory(
        name, lambda: importObject(eval(value, parser.globalDict)))





class SchemaComponentMeta(type(TransactionComponent)):

    protocols.advise(instancesProvide=[ISchemaElementFactory])

    def makeElement(klass, workspace):
        return klass(workspace=workspace)



class SchemaComponent(TransactionComponent):

    __metaclass__ = SchemaComponentMeta

    workspace = binding.Require()























class EditingContext(config.TypeProperty):
    pass

class DefaultEditingContext(TransactionComponent):
    '''Per class context for editing instances in workspace

       Context factory can be configured via EditingContextFactory property.
       Parent of context is workspace. Use Workspace.contextFor method
       for obtaining context. It ensures that context joins current
       transaction.
    '''

    workspace = binding.Make(lambda: settings.workspace())

    cache = binding.Make('weakref.WeakValueDictionary')

    dirty = saved = inserted = binding.Make(dict)

    def readyToVote(self, ts):
        # flush dirty objects
        res = True
        for ob in self.dirty.values():
            self.workspace.flush(ob)
            res = False
        return res

    def voteForCommit(self, ts):
        # make sure that readyToVote was successful
        assert not self.dirty

    def commitTransaction(self, ts):
        # self.dirty should be clear, all objects are cached
        self.saved.clear()
        self.inserted.clear()

    def abortTransaction(self, ts):
        # deactivate dirty and saved objects
        for reg in (self.dirty, self.saved):
            for ob in reg.values():
                ob._finishTransaction(self.workspace, ts, False)
            reg.clear()
        # newly inserted objects should be removed from cache
        for ob in self.inserted.values():
            ob._finishTransaction(self.workspace, ts, False)
            del ob.__dict__['_workspace']
            del self.cache[ob._oid]
        self.inserted.clear()

    def finishTransaction(self, ts, commited):
        # deactivate all cached objects
        # XXX some object are possibly deactivated twice
        for pk, ob in self.cache.items():
            ob._finishTransaction(self.workspace, ts, commited)
        super(DefaultEditingContext, self).finishTransaction(ts, commited)


EditingContext('*', lambda r,f: DefaultEditingContext())






class Workspace(TransactionComponent):

    connection = binding.Make(lambda: settings.storage_connection())

    contexts = binding.Make(dict)

    def contextFor(self, klass):
        ctx = self.contexts.get(klass)
        if ctx is None:
            ctx = self.contexts[klass] = EditingContext(klass)()
        ctx.joinedTxn
        return ctx

    # IWorkspace implementation

    @dispatch.generic()
    def get(self, klass, key, default=NOT_GIVEN):
        '''Get instance of given klass with given oid'''

    @get.when('klass in type and issubclass(klass, Entity)')
    def get_entity(self, klass, key, default=NOT_GIVEN):
        self.joinedTxn
        ectx = self.contextFor(klass)
        tdesc = TypeConfig(klass)()
        try:
            oid = tdesc.oidFromInput(key)
        except:
            raise KeyError('Invalid key', klass, key)
        cache = ectx.cache
        if oid in cache:
            return cache[oid]
        else:
            try:
                return ~self.find(klass, **tdesc.oidToAttrs(oid)) # XXX
            except exceptions.TooFewResults:
                if default is NOT_GIVEN:
                    raise KeyError('Record not found', klass, oid)
                return default

    @get.when('klass in str')
    def get_by_string(self, klass, key, default=NOT_GIVEN):
        klass = self[klass]
        return self.get(klass, key, default)


    def new(self, __klass, **values):
        return self.add(__klass(**values))
        obj = object.__new__(__klass)
        obj.__dict__['_workspace'] = self
        obj.__init__(**values)
        ectx = self.contextFor(__klass)
        oid = obj._oid
        if oid in ectx.cache:
            raise ValueError('OID already exists', obj, oid)

        return self.do_add(obj)

    def add(self, obj):
        self.joinedTxn
        klass = obj.__class__
        ectx = self.contextFor(klass)

        ws = obj._workspace
        if ws is self:
            return obj
        if ws is not None:
            raise ValueError('object already in another workspace', obj, ws)

        oid = obj._oid
        if oid in ectx.cache:
            raise ValueError('OID already exists', obj, oid)

        return self.do_add(obj)

    def flush_obj(self, ectx, obj):
        if not id(obj) in ectx.dirty:
            return
        old = obj._oid
        self.do_update(obj)
        new = obj._updateOid()
        if old != new:
            obj.__dict__['_oid'] = new
            del ectx.cache[old]
            ectx.cache[new] = obj
        del ectx.dirty[id(obj)]
        ectx.saved[id(obj)] = obj

    def flush(self, obj=None):
        if obj is None:
            for ectx in self.contexts.values():
                for obj in ectx.dirty:
                    self.flush_obj(ectx, obj)
        else:
            # XXX refactor to IMangagedEntity: Bad Entity dependance
            ectx = self.contextFor(obj.__class__)
            self.flush_obj(ectx, obj)


    def remove(self, obj):
        self.joinedTxn
        klass = obj.__class__
        ectx = self.contextFor(klass)

        self.do_remove(obj)
        k = id(obj)
        for d in (ectx.saved, ectx.inserted, ectx.dirty):
            if k in d:
                del d[k]
        del ectx.cache[obj._oid]


    # Querying

    _parse = binding.Make(lambda s: imps('wm4.query.builder.parse'))

    def parse(self, qry, namespaces=None, depth=1):
        if isinstance(qry, query.query):
            return self._parse(qry, self.connection)
        if not isinstance(qry, basestring):
            return query.IQuery(qry)
        if namespaces is None:
            frame = sys._getframe(depth)
            namespaces = frame.f_locals, frame.f_globals, frame.f_builtins
            del frame
        return self._parse(qry, self.connection, *namespaces)

    def __call__(self, qry, namespaces=None, depth=1, preload=True):
        self.joinedTxn
        expr = self.parse(qry, namespaces=namespaces, depth=depth+1)
        return self.connection(expr, ws=preload and self or None)

    def find(self, _klass, **cond):
        if cond:
            q = '[e for e in _klass if %s]' % (' and '.join([
                'e.%s == cond.__getitem__("%s")' % (a,a) for a in cond
            ]))
        else:
            q = '[e for e in _klass]'
        return self(q)

    def preload(self, pobj):
        klass = pobj.__class__
        ectx = self.contextFor(klass)

        oid = pobj._oid
        obj = ectx.cache.get(oid)

        if obj is None:
            pobj.__dict__['_workspace'] = self
            obj = ectx.cache[oid] = pobj

        assert obj._workspace is self
        obj._preload(pobj)
        return obj


    # IWorkspaceSPI implementation

    @dispatch.generic()
    def do_load(self, klass, oid, preload=True, lock=False):
        '''Finds instance with given oid'''

    @do_load.when('issubclass(klass,Entity)')
    def do_load(self, klass, oid, preload=True, lock=False):
        tdesc = TypeConfig(klass)()
        attrs = tdesc.oidToAttrs(oid)
        if lock:
            q = '[query.ExclusiveLock(e) for e in klass if %s]'
        else:
            q = '[e for e in klass if %s]'
        q = q % ' and '.join([
            'e.%s == attrs.__getitem__("%s")' % (a,a) for a in attrs
        ])
        for o in self(q, preload=preload):
            return o


    def configured_triggers(self):
        '''Triggers configured in .ini files'''
        triggers = {}
        for (trig, when, klass, name), handler in workspace_triggers().items():
            triggers.setdefault(
                imps(klass), {}
            ).setdefault(
                trig, {}
            ).setdefault(
                when, {}
            )[name] = handler(name, name)
        return triggers
    configured_triggers = binding.Make(configured_triggers)





    @dispatch.generic()
    def do_add(self, obj):
        '''Insert object into workspace'''

    @do_add.around('obj in Entity')
    def do_add_entity_around(next_method, self, obj):
        klass = obj.__class__
        ectx = self.contextFor(klass)

        # Set _workspace and _oid attribute
        oid = obj._oid
        obj.__dict__['_workspace'] = self
        ectx.cache[oid] = obj

        # Notify all features that entity is added to workspace
        for a,f in klass.features:
            f.on_add(obj, self)

        # Do insert
        try:
            BeforeAdd(klass, obj=obj)
            ret = next_method(self, obj)
            AfterAdd(klass, obj=obj)
        except:
            del obj.__dict__['_workspace']
            del ectx.cache[oid]
            raise
        else:
            ectx.inserted[id(obj)] = obj
        return ret

    @do_add.when('obj in Entity')
    def do_add_entity(self, obj):
        obj.beforeAdd()
        klass  = obj.__class__
        dct  = obj.__dict__
        values = dict(
            (a, dct[a]) for a,d in klass.features
                         if a in dct and d.cardinality == 'single'
        )
        self(query.Insert(klass, **values))
        obj.afterAdd()
        return obj



    @dispatch.generic()
    def on_change(self, obj):
        pass

    @on_change.when('obj in Entity')
    def on_change(self, obj):
        klass = obj.__class__
        ectx = self.contextFor(klass)
        ectx.dirty[id(obj)] = obj
        assert obj._oid in ectx.cache, 'object is not cached obj=%r, oid=%r' % (
            obj, obj._oid
        )




    @dispatch.generic()
    def do_update(self, obj):
        '''Update object in workspace'''

    @do_update.around('obj in Entity')
    def do_update_around(next_method, self, obj):
        klass = obj.__class__
        BeforeUpdate(klass, obj=obj)
        ret = next_method(self, obj)
        AfterUpdate(klass, obj=obj)

    @do_update.when('obj in Entity')
    def do_update_entity(self, obj):
        #lock = obj.__dict__.pop('__lock__')
        obj.beforeUpdate()
        changes = obj.genChanges()
        if changes:
            klass = obj.__class__
            tdesc = TypeConfig(klass)()
            attrs = tdesc.oidToAttrs(obj._oid)
            q = '[query.Update(e, %s) for e in klass if %s]' % (
                ', '.join(
                    '%s=changes.__getitem__("%s")' % (a,a) for a in changes
                ),
                ' and '.join(
                    'e.%s==attrs.__getitem__("%s")' % (a,a) for a in attrs
                )
            )
            self(q)
        obj.afterUpdate()
        obj.__dict__.pop('__lock__')


    @dispatch.generic()
    def do_remove(self, obj):
        '''Remove object from workspace'''

    @do_remove.around('obj in Entity')
    def do_remove_entity_around(next_method, self, obj):
        klass = obj.__class__

        # Notify all features that entity is removed to workspace
        for a,f in klass.features:
            f.on_remove(obj, self)

        BeforeRemove(klass, obj=obj)
        # Do insert
        ret = next_method(self, obj)
        AfterRemove(klass, obj=obj)
        del obj.__dict__['_workspace']
        return ret

    @do_remove.when('obj in Entity')
    def do_remove_entity(self, obj):
        obj.beforeRemove()
        klass = obj.__class__
        tdesc = TypeConfig(klass)()
        attrs = tdesc.oidToAttrs(obj._oid)
        q = '[query.Delete(e) for e in klass if %s]' % (
            ' and '.join('e.%s==attrs.__getitem__("%s")' % (a,a) for a in attrs)
        )
        self(q)
        obj.afterRemove()





    _elementsCache = binding.Make(dict)
    _pathToFactory = binding.Make(lambda s: ElementRegistry())
    _factoryToPath = binding.Make(lambda s: dict(
            (v,k) for k,v in s._pathToFactory.iteritems()
    ))

    def getPath(self, fact):
        path = self._factoryToPath.get(fact, NOT_FOUND)
        if path is NOT_FOUND:
            raise ElementNotFound(fact)
        return path

    def getElement(self, path):
        path = normpath(path)
        while path.startswith('/'):
            path = path[1:]
        cache = self._elementsCache
        elem = cache.get(path, NOT_FOUND)
        if elem is NOT_FOUND:
            fact = self._pathToFactory.get(path, NOT_FOUND)
            if fact is NOT_FOUND:
                raise ElementNotFound(path)
            elem = cache[path] = fact.makeElement(self)
        return elem

    __getitem__ = getElement
