from wm4.api import *
from dispatch.strategy import default
from datetime import date, datetime
from sql4.utils import *
from wm4.query.api import *

from type_descriptors import *

__all__ = ['SqlTableName', 'AbstractConnection']


class SqlTableName(config.TypeProperty):
    '''Table name for given Entity'''


SqlTableName('*', lambda r,f: sql_name(r.__name__))





class AbstractConnection(binding.Component):
    '''Base class for SQL connections'''

    encoding = 'utf-8'

    def convertorFor(self, typ):
        return SqlTypeDescriptor(typ)()

    def sqlTable(self, klass):
        return SqlTableName(klass)()


    # Query instance transformation

    @dispatch.generic()
    def transform(self, expr):
        '''Transforms every built expression for convenience of some connection
        '''
    @transform.when(default)
    def default_transformation(self, expr):
        return expr

    @transform.when(
        'expr in BaseSelect')
    def transform_base_select(self, expr):
        return expr.buildSelect(self)

    @transform.when(
        'expr in Getattr and expr.obj in CompoundVar')
        #'and expr.attr in expr.obj.pmap.attrs')
    def transform_getattr_on_compound_var(self, expr):
        obj = expr.obj
        atr = expr.attr
        if atr == '_oid':
            return self.transform(OidVar(obj))
        map = obj.pmap[atr]
        if isinstance(map, (schema.StructuredTypeConfig, schema.Reference)):
            map = schema.TypeConfig(map.typeObj)()
            return self.transform(RelVar(atr, map, obj))
        else:
            return self.transform(AttrVar(atr, map, obj))

    @transform.when(
        'expr in Variable and len(expr.parents) > 2 and '
        'expr.parents[-2] in RelVar and '
        'expr.var in expr.parents[-2].pmap.refFeatures')
    def transform_relvar_to_pkeyvar(self, expr):
        # replace RelVar with PKeyVar if there isn't need for whole record
        p1, p2, p3 = expr.parents[-3:]
        return p3.__class__(p3.var, p3.pmap, PKeyVar(p2.var, p2.pmap, p1))

    @transform.when(
      'expr in Getattr and expr.obj in Const')
    def transform_getattr_on_const(self, expr):
        return Const(getattr(expr.obj.value, expr.attr))

    @transform.when(
      'expr in Getattr and (expr.obj.typ == date or expr.obj.typ == datetime)')
    def transform_getattr_on_date(self, expr):
        if not hasattr(expr.obj.typ, expr.attr):
            raise AttributeError(expr.obj.typ, expr.attr)
        return self.transform(DatePart(expr.obj, Const(expr.attr)))

    @transform.when('expr in OidVar and expr.expr in Variable')
    def transform_oid_of_variable(self, expr):
        # Unpack oid to primitives
        ref = []
        for pf in expr.expr.pmap.refPrimitives:
            obj = expr.expr
            for p in pf.parents[1:]:
                obj = self.transform(Getattr(obj,p.attrName))
            ref.append(obj)
        if len(ref) == 1:
            return ref[0]
        else:
            return self.transform(Tuple(*ref))

    @transform.when('expr in OidVar and expr.expr in Const')
    def transform_oid_of_constant(self, expr):
        # Unpack oid to primitives
        obj = expr.expr.value
        pmap = schema.TypeConfig(obj.__class__)()
        ref = []
        for pf in pmap.refPrimitives:
            val = obj
            for p in pf.parents[1:]:
                val = getattr(val, p.attrName)
            ref.append(Const(val))
        if len(ref) == 1:
            return ref[0]
        else:
            return self.transform(Tuple(*ref))

    @transform.when(
      '(expr in EQ or expr in NE or expr in IS or expr in IS_NOT) and '
      'expr.left.typ in schema.EntityMeta and '
      'expr.right.typ in schema.EntityMeta')
    def transform_rv_comparison(self, expr):
        l = expr.left
        r = expr.right
        if l.typ != r.typ:
            return Const(False)
        OP = expr.__class__
        return self.transform(
            OP(self.transform(OidVar(l)), self.transform(OidVar(r)))
        )
        cond = []
        for a,f in l.pmap.refFeatures:
            cond.append(self.transform(
                OP(self.transform(Getattr(l,a)), self.transform(Getattr(r,a)))
            ))
        return AND(*cond)

    @transform.when(
      'expr in CompOp and expr.left in Tuple and expr.right in Tuple')
    def transform_tuple_comparison(self, expr):
        op = expr.__class__
        l, r = expr.left, expr.right
        if len(l.args) != len(r.args):
            raise TypeError(
                'Cannot compare tuples with different size', l, r)
        return AND(*(
            self.transform(op(la, ra)) for la, ra in zip(l.args, r.args)
        ))

    @transform.when(
      '(expr in IN or expr in NOT_IN) and '
      'expr.left.typ in schema.EntityMeta and expr.right in Select')
    def transform_in_subquery(self, expr):
        lTyp = expr.left.typ
        rTyp = expr.right.expr.typ
        if lTyp != rTyp:
            if expr.__class__ == IN:
                return Const(False)
            else:
                return Const(True)
        left = self.transform(OidVar(expr.left))
        select = expr.right
        sexpr = self.transform(OidVar(expr.right.expr))
        return self.transform(expr.__class__(
            left, select.withExpr(self, sexpr)
        ))

    @transform.when(
      'expr in Call and expr.func == len and '
      'issubclass(expr.args[0].typ,basestring)')
    def transform_len_func(self, expr):
        assert len(expr.args) == 1
        assert not expr.kw
        return self.transform(Length(expr.args[0]))

    @transform.when(
      'expr in Call and expr.func in type')
    def transform_len_func(self, expr):
        assert len(expr.args) == 1
        assert not expr.kw
        assert not issubclass(expr.func, schema.Entity)
        return self.transform(Cast(expr.args[0], Const(expr.func)))

    @transform.when(
        'expr in Getitem and '
        'issubclass(expr.typ, basestring) and '
        'issubclass(expr.idx.typ, int)')
    def transform_getitem_to_substr(self, expr):
        idx = expr.idx
        if isinstance(idx, Const):
            idx = idx.value
            if idx < 0:
                return Substr(expr.obj,
                    SUB(Length(expr.obj), Const(abs(idx)-1)), Const(1))
            else:
                return Substr(expr.obj,
                    Const(idx+1), Const(1))
        else:
            raise NotImplementedError

    @transform.when(
        'expr in Slice and issubclass(expr.typ, basestring)')
    def transform_getslice(self, expr):
        start, stop, stride = expr.start, expr.stop, expr.stride
        if not isinstance(stride, Const) or stride.value not in (1,None):
            raise NotImplementedError('unsupported stride', stride)

        if not isinstance(start, Const):
            raise NotImplementedError('nonconst start', start)
        if not isinstance(stop, Const):
            raise NotImplementedError('nonconst stop', stop)
        start, stop = start.value, stop.value

        if start is None and stop is None:
            return expr.obj
        elif stop is None:
            if start >= 0:
                return Substr(expr.obj, Const(start+1), Const(None))
            else:
                f = abs(start)-1
                if f:
                    return Substr(
                        expr.obj, SUB(Length(expr.obj),Const(f)), Const(None))
                else:
                    return Substr(expr.obj, Length(expr.obj), Const(None))
        else:
            if start is None:
                start = 0
            if start >= 0:
                if stop >= 0:
                    return Substr(expr.obj, Const(start+1), Const(stop-start))
                else:
                    ldiff = start + abs(stop)
                    return Substr(expr.obj,
                        Const(start+1), SUB(Length(expr.obj),Const(ldiff)))
            else:
                raise NotImplementedError


    # Query rendering

    @dispatch.generic()
    def render(self, expr):
        '''Generic function for converting expression to SQL query'''

    def valuesForEdit(self, relvar, vals, relations=None):
        if relations is None:
            relations = [(None, relvar, None, None)]
        rMap = relvar.pmap
        typ = relvar.typ

        edit = []
        for a,f in typ.features:
            e = vals.get(a, NOT_FOUND)
            if e is NOT_FOUND:
                continue
            fMap = rMap[a]
            if issubclass(fMap.typeObj, schema.Entity):
                if not isinstance(e, Const):
                    raise NotImplementedError('non const for structured type')
                val = e.value
                for pMap in fMap.refPrimitives:
                    pa = pMap.storageAttr
                    pv = pMap.valueFromObj(val, fMap)
                    pv = self.convertorFor(pMap.typeObj).toSql(pv)
                    edit.append((pa, pv))
            else:
                edit.append((fMap.attrName, self.renderExpr(e,relations)))
        return edit

    @render.when('expr in Insert and expr.rel in Const')
    def render_insert(self, expr):
        relvar = IQueryable(expr.rel.value).alias('ins')
        relations = [(None, relvar, None, None)]

        ins = self.valuesForEdit(relvar, expr.values, relations)

        q = [ 'INSERT INTO ', self.renderFrom(relvar,relations), ' (',
              ', '.join(escape_name(a) for a,_ in ins),
              ') VALUES (',
              ', '.join(v for _,v in ins),
              ')',
        ]
        return ''.join(q)

    @render.when('expr in UpdateQuery')
    def render_update(self, expr):
        relations = [(None, expr.rel, None, None)]
        upd = self.valuesForEdit(expr.rel, expr.values, relations)
        q = [ 'UPDATE ', self.renderFrom(expr.rel,relations), ' SET ',
              ', '.join('%s=%s' % (escape_name(a),v) for a,v in upd)
            ]
        if expr.where:
            q.append(' WHERE %s' % self.renderExpr(expr.where,relations))
        return ''.join(q)

    @render.when('expr in DeleteQuery')
    def render_delete(self, expr):
        relations = [(None, expr.rel, None, None)]
        q = ['DELETE FROM ', self.renderFrom(expr.rel,relations)]
        if expr.where:
            q.append(' WHERE %s' % self.renderExpr(expr.where,relations))
        return ''.join(q)

    @render.when('expr in Select')
    def render_select(self, expr):
        rels = expr.relations
        sel = self.renderExpr(expr.expr,rels).strip()
        if sel.startswith('(') and sel.endswith(')'):
            sel = sel[1:-1].strip()
        if expr.distinct:
            q = ['SELECT DISTINCT ', sel]
        else:
            q = ['SELECT ', sel]
        if expr.relations:
            q.append(' FROM ')
        for x,(a,i,c,_) in enumerate(expr.relations):
            f = self.renderFrom(i,rels)
            if x == 0:
                assert not c, repr(c)
                q.append(f)
            else:
                if isinstance(c, LeftJoin):
                    q.append(' LEFT JOIN ')
                    c = c.cond
                elif isinstance(c, RightJoin):
                    q.append(' RIGHT JOIN ')
                    c = c.cond
                elif isinstance(c, Join):
                    q.append(' INNER JOIN ')
                    c = c.cond
                elif not c:
                    q.append(' CROSS JOIN ')
                else:
                    raise NotImplementedError('unsupported join', c)
                q.append(f)
                if c:
                    q.append(' ON (%s)' % self.renderExpr(c,rels))
        if expr.where:
            q.append(' WHERE %s' % self.renderExpr(expr.where,rels))
        if expr.groupBy:
            q.append(' GROUP BY ')
            q.append(', '.join(
                self.renderExpr(o,rels) for o in expr.groupBy.args
            ))
        if expr.having:
            q.append(' HAVING %s' % self.renderExpr(expr.having, rels))
        if expr.orderBy:
            q.append(' ORDER BY ')
            q.append(', '.join(
                self.renderExpr(o,rels) for o in expr.orderBy.args
            ))
        if expr.limit:
            q.append(' LIMIT %s' % self.renderExpr(expr.limit, rels))
        if expr.offset:
            q.append(' OFFSET %s' % self.renderExpr(expr.offset, rels))
        if expr.lock is ExclusiveLock:
            q.append(' FOR UPDATE')
        elif expr.lock is SharedLock:
            q.append(' FOR SHARE')
        return ''.join(q)




    def alias(self, var):
        return escape_name(var.storage_node.path)
        return escape_name(
            '.'.join(p.var for p in var.storage_node.parents)
        )


    @dispatch.generic()
    def renderFrom(self, expr, relations):
        pass

    @renderFrom.when('expr in RelVar')
    def render_from_rel_var(self, expr, relations):
        tab = SqlTableName(expr.typ)()
        if len(relations) < 2:
            return escape_name(tab)
        else:
            return '%s AS %s' % (escape_name(tab), self.alias(expr))

    @renderFrom.when('expr in SubselectVar')
    def render_from_subselect(self, expr, relations):
        s = expr.sel
        a = escape_name(expr.var)
        e = s.expr
        if isinstance(e, Tuple):
            return '(%s) AS %s (%s)' % (
                self.render(s), a,
                ', '.join('s%d' % i for i,e in enumerate(e.args))
            )
        else:
            return '(%s) AS %s (%s)' % (
                self.render(s), a,
                ', '.join(p.storageAttr for p in e.pmap.primitives)
            )

    @dispatch.generic()
    def renderExpr(self, expr, relations):
        pass

    @renderExpr.when('expr in AttrVar')
    def render_attr_as_expr(self, expr, relations):
        sn = expr.storage_node
        ex = expr
        p = []
        while ex != sn:
            p.insert(0, ex.var)
            ex = ex.parent
        mp = ex.pmap
        for a in p:
            mp = mp[a]
        if len(relations) < 2:
            return escape_name(mp.storageAttr)
        else:
            return '%s.%s' % (
                self.alias(expr.parent),escape_name(mp.storageAttr)
            )

    @renderExpr.when('expr in RelVar')
    def render_rel_as_expr(self, expr, relations):
        smap = expr.pmap.primitives
        if len(relations) < 2:
            return ', '.join(escape_name(p.storageAttr) for p in smap)
        else:
            a = self.alias(expr)
            return ', '.join(
                '%s.%s' % (a,escape_name(p.storageAttr)) for p in smap
            )

    @renderExpr.when('expr in SubselectVar')
    def render_rel_as_expr(self, expr, relations):
        s = expr.sel
        a = escape_name(expr.var)
        e = s.expr
        if isinstance(e, Tuple):
            return ', '.join(
                '%s.s%d' % (a,i) for i,e in enumerate(e.args))
        else:
            return ', '.join(
                '%s.%s' % (a,p.storageAttr) for p in e.pmap.primitives)

    @renderExpr.when('expr in Const')
    def render_const_expr(self, expr, relations):
        conv = self.convertorFor(expr.typ)
        return conv.toSql(expr.value)

    @renderExpr.when('expr in As')
    def render_aliased_expr(self, expr, relations):
        return '%s AS %s' % (
            self.renderExpr(expr.expr, relations),
            escape_name(expr.alias)
        )

    @renderExpr.when('expr in Tuple')
    def render_tuple_expr(self, expr, relations):
        return '(%s)' % ', '.join(
            self.renderExpr(a, relations) for a in expr.args
        )

    op_map = {'==':'=', '!=':'<>', 'in':'IN', 'not in':'NOT IN'}

    @renderExpr.when('expr in BinOp')
    def render_bin_op_expr(self, expr, relations):
        left = brackets(self.renderExpr(expr.left, relations))
        right = brackets(self.renderExpr(expr.right, relations))
        op = expr.repr
        op = self.op_map.get(op,op)
        return '%s %s %s' % (left, op, right)

    @renderExpr.when('expr in UnOp')
    def render_un_op_expr(self, expr, relations):
        arg = brackets(self.renderExpr(expr.args[0], relations))
        return '%s %s' % (expr.repr.upper(), arg)

    @renderExpr.when('expr in AND or expr in OR')
    def render_logop(self, expr, relations):
        op = ' %s ' % expr.__class__.__name__
        return op.join(
            #brackets(self.renderExpr(a, relations)) for a in expr.args
            '(%s)' % self.renderExpr(a, relations) for a in expr.args
        )

    @renderExpr.when('expr in EQ or expr in IS')
    def render_eq_expr(self, expr, relations):
        left = brackets(self.renderExpr(expr.left, relations))
        right = brackets(self.renderExpr(expr.right, relations))
        if right == 'NULL':
            return '%s ISNULL' % left
        else:
            return '%s = %s' % (left, right)

    @renderExpr.when('expr in NE or expr in IS_NOT')
    def render_ne_expr(self, expr, relations):
        left = brackets(self.renderExpr(expr.left, relations))
        right = brackets(self.renderExpr(expr.right, relations))
        if right == 'NULL':
            return '%s NOTNULL' % left
        else:
            return '%s != %s' % (left, right)

    @renderExpr.when('(expr in IN or expr in NOT_IN) and expr.right in Const')
    def render_in_constant_sequence(self, expr, relations):
        seq = expr.right.value
        assert isinstance(seq, (list, tuple, set))
        assert len(seq) > 0
        left = brackets(self.renderExpr(expr.left, relations))
        conv = self.convertorFor(type(seq[0]))
        right = ','.join([conv.toSql(v) for v in seq])
        return '%s %s (%s)' % (left, expr.__class__.__name__, right)

    @renderExpr.when('expr in Select')
    def render_subselect(self, expr, relations):
        return '(%s)' % self.render(expr)

    @renderExpr.when('expr in Desc')
    def render_desc_expr(self, expr, relations):
        return '%s DESC' % self.renderExpr(expr.expr, relations)

    @renderExpr.when('expr in Asc')
    def render_asc_expr(self, expr, relations):
        return '%s ASC' % self.renderExpr(expr.expr, relations)

    @renderExpr.when('expr in NullsLast')
    def render_desc_expr(self, expr, relations):
        return '%s NULLS LAST' % self.renderExpr(expr.expr, relations)

    @renderExpr.when('expr in Count')
    def render_count_aggregate(self, expr, relations):
        arg = expr.expr
        if not arg:
            return 'COUNT(*)'
        if isinstance(arg, CompoundVar):
            if len(relations) == 1:
                return 'COUNT(*)'
            else:
                return 'COUNT(%s)' % self.alias(arg)
        else:
            return 'COUNT(%s)' % self.renderExpr(arg, relations)

    @renderExpr.when('expr in Aggregate')
    def render_aggregate(self, expr, relations):
        return '%s(%s)' % (
            expr.__class__.__name__.upper(),
            self.renderExpr(expr.expr, relations)
        )

    @renderExpr.when('expr in SeqFunc')
    def render_seq_func(self, expr, relations):
        return '%s(%s)' % (
            expr.__class__.__name__.upper(),
            ','.join(self.renderExpr(a,relations) for a in expr.args)
        )

    @renderExpr.when('expr in Substr')
    def render_substr(self, expr, relations):
        str, start, stop = expr.args
        if isinstance(stop, Const) and stop.value is None:
            return 'SUBSTRING(%s FROM %s)' % (
                self.renderExpr(str, relations),
                self.renderExpr(start, relations),
            )
        else:
            return 'SUBSTRING(%s FROM %s FOR %s)' % (
                self.renderExpr(str, relations),
                self.renderExpr(start, relations),
                self.renderExpr(stop, relations),
            )


    @renderExpr.when('expr in Lower')
    def render_lower(self, expr, relations):
        return '%s(%s)' % (
            expr.__class__.__name__.upper(),
            self.renderExpr(expr.args[0], relations),
        )

    @renderExpr.when('expr in Coalesce')
    def render_coalesce(self, expr, relations):
        return 'COALESCE(%s)' % ','.join(self.renderExpr(a, relations) for a in expr.args)

    @renderExpr.when('expr in Least')
    def render_least(self, expr, relations):
        return 'LEAST(%s)' % ','.join(self.renderExpr(a, relations) for a in expr.args)

    @renderExpr.when('expr in Greatest')
    def render_greatest(self, expr, relations):
        return 'GREATEST(%s)' % ','.join(self.renderExpr(a, relations) for a in expr.args)

    @renderExpr.when('expr in Case')
    def render_case(self, expr, relations):
        s = ['CASE']
        for c,e in expr.conds:
            s.append('WHEN')
            s.append(self.renderExpr(c, relations))
            s.append('THEN')
            s.append(self.renderExpr(e, relations))
        if expr.default is not None:
            s.append('ELSE')
            s.append(self.renderExpr(expr.default, relations))
        s.append('END')
        return ' '.join(s)

    @renderExpr.when('expr in Cast and expr.typ in Const')
    def render_cast(self, expr, relations):
        typ = expr.typ.value
        if issubclass(typ, basestring):
            cast = 'TEXT'
        else:
            # XXX SQL type name can be different
            cast = typ.__name__
        return 'CAST(%s AS %s)' % (
            self.renderExpr(expr.obj, relations), cast
        )

    @renderExpr.when('expr in DatePart and expr.args[1] in Const')
    def render_date_part(self, expr, relations):
        return 'EXTRACT(%s FROM %s)' % (
            expr.args[1].value,
            self.renderExpr(expr.args[0], relations),
        )

    @renderExpr.when('expr in Random')
    def render_random(self, expr, relations):
        return 'random()'



    # Row convertor

    @dispatch.generic()
    def row_convertor(self, expr, workspace=None):
        '''Method gets Select.expr as expr argument and creates
           convertor that converts tuples returned from storage.'''


    @row_convertor.when(default)
    def default_row_convertor(self, expr, workspace=None):
        fromSql = self.convertorFor(expr.typ).fromSql
        def conv(rowIter):
            return fromSql(rowIter.next())
        return conv

    @row_convertor.when('expr in BaseSelect')
    def select_row_convertor(self, expr, workspace=None):
        expr_conv = self.row_convertor(expr.expr, workspace)
        def conv(row):
            return expr_conv(iter(row))
        return conv

    @row_convertor.when('expr in Tuple')
    def tuple_row_convertor(self, expr, workspace=None):
        item_convertors = [self.row_convertor(a,workspace) for a in expr.args]
        def conv(rowIter):
            return tuple(
                ic(rowIter) for ic in item_convertors
            )
        return conv

    # XXX move to workspace

    def _convertor_for_features(self, klass, features, workspace):
        ac = [
            (attr, self.row_convertor(fMap,workspace))
            for (attr,fMap) in features if fMap.cardinality == 'single'
        ]
        if workspace:
            def conv(rowIter):
                state = dict((a,c(rowIter)) for a,c in ac)
                obj = self.preload(klass, state)
                if obj is not None:
                    return workspace.preload(obj)
                else:
                    return None
        else:
            def conv(rowIter):
                state = dict((a,c(rowIter)) for a,c in ac)
                return self.preload(klass, state)
        return conv

    def preload(self, klass, state):
        obj = klass.allocate()
        for a,v in state.iteritems():
            obj.__dict__[a] = v
        if obj._oid is None: # XXX
            return None
        return obj

    @row_convertor.when('expr in schema.Attribute')
    def primitive_map_convertor(self, expr, workspace=None):
        fromSql = self.convertorFor(expr.typeObj).fromSql
        def conv(rowIter):
            return fromSql(rowIter.next())
        return conv

    @row_convertor.when('expr in schema.Reference')
    def primitive_map_convertor(self, expr, workspace=None):
        return self._convertor_for_features(
            expr.typeObj, expr.refFeatures, workspace
        )

    @row_convertor.when('expr in AttrVar')
    def entity_row_convertor(self, expr, workspace=None):
        return self.row_convertor(
            expr.pmap, workspace
        )

    @row_convertor.when('expr in RelVar')
    def entity_row_convertor(self, expr, workspace=None):
        return self._convertor_for_features(
            expr.typ, expr.pmap.features, workspace
        )

    @row_convertor.when('expr in Cast and expr.typ in Const')
    def casted_column_covertor(self, expr, workspace=None):
        typ = expr.typ.value
        def conv(rowIter):
            return schema.normalize(rowIter.next(), typ)
        return conv
