from wm4.api import *
from weakref import WeakValueDictionary
import re
import traceback

from transactions import TransactionComponent
from base import AbstractConnection

from interfaces import *


__all__ = [
    'DatabaseURL', 'Cursor', 'DBAPIConnection',
    'PgSqlURL', 'PgSqlConnection',
    'SqliteURL', 'SqliteConnection',
    'MSSqlURL', 'MSSqlConnection',
    'FirebirdURL', 'FirebirdConnection',
]




def connectionFromURL():
    return naming.lookup(settings.storage_url())





class DatabaseURL(naming.AbstractURL):

    scheme = None
    slots = 'user', 'password', 'host', 'database', 'port', 'params'

    MATCH = re.compile(
        '^//'
        '((?P<user>[^/:@]+)(:(?P<password>[^/:@]+))?@)?'
        '((?P<host>%s)(:(?P<port>%s))?)?'
        '/(?P<database>[^?]+)(?P<params>.*)$' % (
            naming.DNSNAME_RE, naming.PORT_RE),
        re.I).match

    def __str__(self):
        u = '%s://' % self.scheme
        if self.user:
            u += self.user
            if self.password:
                u+= ':' + self.password
            u+= '@'
        if self.host:
            u+= self.host
            if self.port:
                u+= ':' + self.port
        return u + '/' + self.database + (self.params or '')




class Cursor(object):

    __slots__ = 'cursor', 'convertor', '__weakref__'

    def __init__(self, cursor, convertor=None):
        self.cursor = cursor
        if convertor is None:
            convertor = lambda row:row
        self.convertor = convertor

    description = property(lambda s: s.cursor.description)

    def __iter__(self):
        fetch = self.cursor.fetchone
        conv = self.convertor
        while True :
            row = fetch()
            if row is None :
                break
            yield conv(row)

    def justOne(self):
        i = iter(self)
        try :
            row = i.next()
        except StopIteration :
            raise exceptions.TooFewResults
        try :
            row = i.next()
        except StopIteration :
            return row
        else :
            raise exceptions.TooManyResults
    __invert__ = justOne


    def oneOrNone(self):
        i = iter(self)
        try :
            row = i.next()
        except StopIteration :
            return None
        try :
            row = i.next()
        except StopIteration :
            return row
        else :
            raise exceptions.TooManyResults
    __neg__ = oneOrNone

    def __len__(self):
        return self.cursor.rowcount

    def close(self):
        self.cursor.close()



class DBAPIConnection(TransactionComponent, AbstractConnection):

    protocols.advise(instancesProvide=[ISQLConnection])

    encoding = 'utf-8'
    address  = binding.Require('Connection URL')
    dbcon    = binding.Make(lambda s: s.open(s.address))
    logger   = binding.Make(lambda: naming.lookup('logger:sql'))
    cursors  = binding.Make(WeakValueDictionary)


    def closeCursors(self):
        for c in self.cursors.values():
            c.close()

    # storage.ITransactionParticipant

    def onJoinTxn(self, txnService):
        self.logger.debug("BEGIN: -----------------------------------")
        super(DBAPIConnection, self).onJoinTxn(txnService)

    def voteForCommit(self, txnService):
        self.closeCursors()

    def abortTransaction(self, txnService):
        self.logger.debug("ABORT: -----------------------------------")
        self.closeCursors()
        self.dbcon.rollback()

    def commitTransaction(self, txnService):
        self.logger.debug("COMMIT: ----------------------------------")
        self.dbcon.commit()

    def __call__(self, sql, conv=None, ws=None):
        if not isinstance(sql, basestring):
            sql = query.IQuery(sql)
            if conv is None and isinstance(sql, query.BaseSelect):
                conv = self.row_convertor(sql, ws)
            sql = self.render(sql)
        self.joinedTxn
        if 0:
            tb = ''.join(traceback.format_stack()).splitlines()
            tb.insert(0, '')
            self.logger.debug('\nSQLTB: '.join(tb))
        if isinstance(sql, unicode):
            sql = sql.encode('utf8')
        self.logger.debug('QUERY: '+ str(sql)[:2048])
        cur = self.dbcon.cursor()
        cur.execute(sql)
        cur = Cursor(cur, conv)
        self.cursors[id(cur)] = cur
        return cur







class PgSqlURL(DatabaseURL):

    SCHEMES = set(['pgsql'])

    def getObject(self):
        return PgSqlConnection(address=self)


class PgSqlConnection(DBAPIConnection):
    
    def version(self):
        return tuple(
            map(int, (~self('SELECT VERSION()'))[0].split()[1].split('.'))
        )
    version = binding.Make(version)

    def keywords(self, addr):
        kwds = {}
        if addr.host:
            kwds['host'] = addr.host
        if addr.port:
            kwds['port'] = addr.port
        if addr.params:
            from cgi import parse_qsl
            for n, v in parse_qsl(addr.params[1:]):
                if n == 'rundir':
                    kwds['host'] = v
                elif n == 'port':
                    kwds['port'] = v
        if addr.database:
            kwds['database'] = addr.database
        if addr.user:
            kwds['user'] = addr.user
        if addr.password:
            kwds['password'] = addr.password
        return kwds

    def open(self, addr):
        from sql4.postgresql import connect
        kwds = self.keywords(addr)
        port = kwds.pop('port', None)
        if port is not None and kwds.get('host'):
            kwds['host'] = kwds['host'] + ':' + port
        return connect(**kwds)


class PsycopgSqlURL(DatabaseURL):

    SCHEMES = set(['psycopg'])

    def getObject(self):
        return PsycopgSqlConnection(address=self)


class PsycopgSqlConnection(PgSqlConnection):
    
    def open(self, addr):
        kwds = {}
        if addr.host:
            if addr.host and addr.port:
                kwds['host'] = addr.host + ':' + addr.port
            else:
                kwds['host'] = addr.host
        if addr.database:
            kwds['database'] = addr.database
        if addr.user:
            kwds['user'] = addr.user
        if addr.password:
            kwds['password'] = addr.password
        return connect(**kwds)

    def open(self, addr):
        from psycopg2 import connect
        return connect(**self.keywords(addr))




class SqliteURL(naming.AbstractURL):

    scheme = None
    slots = 'filename',

    MATCH = re.compile('^(?P<filename>.*)$').match

    SCHEMES = set(['sqlite'])

    def __str__(self):
        return '%s:%s' % (self.scheme, self.filename)

    def getObject(self):
        return SqliteConnection(address=self)


class SqliteConnection(DBAPIConnection):

    def open(self, addr):
        import sqlite
        return sqlite.connect(addr.filename or ':memory:')



class MSSqlURL(DatabaseURL):

    SCHEMES = set(['mssql'])

    def getObject(self):
        return MSSqlConnection(address=self)


class MSSqlConnection(DBAPIConnection):

    def open(self, addr):
        import pymssql
        if addr.host and addr.port:
            host = addr.host + ':' + addr.port
        else:
            host = addr.host
        return pymssql.connect(
            host=host, database=addr.database,
            user=addr.user, password=addr.password
        )



class FirebirdURL(DatabaseURL):

    SCHEMES = set(['firebird'])

    def getObject(self):
        return FirebirdConnection(address=self)


class FirebirdConnection(DBAPIConnection):

    def open(self, addr):
        import kinterbasdb
        if addr.host and addr.port:
            host = addr.host + '/' + addr.port
        else:
            host = addr.host
        return kinterbasdb.connect(
            database = addr.database.replace('|', ':'),
            user = addr.user,
            password = addr.password,
            host = host,
            charset = 'UNICODE'
        )
