"""Transaction and Persistence tests"""

from unittest import *
from wm4.api import *
from wm4.storage import mockdb


class TestTypeConvertors(TestCase):

    def setUp(self):
        self.old = config.snapshot()
        config.loadConfig()
        settings.storage_url('mockdb:')
        self.con = settings.storage_connection()
        self.ts = settings.transaction_service()

    def tearDown(self):
        if self.ts.isActive():
            self.ts.abort()
        config.swap(self.old)
        self.con.verify()
        
    def test_string_to_sql(self):
        from wm4.storage.type_descriptors import SQLDefault
        c = SQLDefault(str)
        self.assertEqual(c.toSql(None), "NULL")
        self.assertEqual(c.toSql('ahoj'), "'ahoj'")
        self.assertEqual(c.toSql('ahoj\n'), "E'ahoj\\n'")

    def test_unicode_to_sql(self):
        from wm4.storage.type_descriptors import SQLUnicode
        c = SQLUnicode(unicode)
        self.assertEqual(c.toSql(None), "NULL")
        self.assertEqual(c.toSql(u'ahoj'), "'ahoj'")
        self.assertEqual(c.toSql(u'ahoj\n'), "E'ahoj\\n'")


class TestMockSqlConnection(TestCase):

    def setUp(self):
        self.old = config.snapshot()
        config.loadConfig()
        settings.storage_url('mockdb:')
        self.con = settings.storage_connection()
        self.ts = settings.transaction_service()

    def tearDown(self):
        if self.ts.isActive():
            self.ts.abort()
        config.swap(self.old)
        self.con.verify()

    def test01_Pretest(self):
        assert isinstance(self.con, mockdb.MockSqlConnection), repr(self.con)

    def test10_NoRows(self):
        self.con.expects(transaction='begin')
        self.con.expects(sql='SELECT 1', res=[])
        self.con.expects(transaction='commit')
        self.ts.begin()
        r = list(self.con('SELECT 1'))
        self.ts.commit()
        self.assertEqual(r, [])

    def test11_ManyRows(self):
        self.con.expects(transaction='begin')
        self.con.expects(sql='SELECT * FROM tab', res=[[1,'foo'],[2,'bar']])
        self.con.expects(transaction='commit')
        self.ts.begin()
        r = list(self.con('SELECT * FROM tab'))
        self.ts.commit()
        self.assertEqual(r, [[1,'foo'],[2,'bar']])

    def test12_GetOne(self):
        self.con.expects(transaction='begin')
        self.con.expects(sql='SELECT * FROM tab', res=[[1,'foo']])
        self.con.expects(transaction='commit')
        self.ts.begin()
        r = ~self.con('SELECT * FROM tab')
        self.ts.commit()
        self.assertEqual(r, [1,'foo'])

    def test13_GetOneWhileMany(self):
        self.con.expects(transaction='begin')
        self.con.expects(sql='SELECT * FROM tab', res=[[1,'foo'],[2,'bar']])
        self.con.expects(transaction='abort')
        self.ts.begin()
        r = self.con('SELECT * FROM tab')
        self.assertRaises(exceptions.TooManyResults, r.justOne)
        self.ts.abort()

    def test14_GetOneWhileNoRow(self):
        self.con.expects(transaction='begin')
        self.con.expects(sql='SELECT * FROM tab', res=[])
        self.con.expects(transaction='abort')
        self.ts.begin()
        r = self.con('SELECT * FROM tab')
        self.assertRaises(exceptions.TooFewResults, r.justOne)
        self.ts.abort()



def suite():
    return TestSuite((
        makeSuite(TestMockSqlConnection),
        makeSuite(TestTypeConvertors),
    ))


if __name__ == '__main__':
    main(defaultTest='suite')
