"""Transaction and Persistence tests"""

from unittest import TestCase, makeSuite, TestSuite, main
from wm4.utils.MiniTable import Table
from wm4.api import *

class LogParticipant(storage.AbstractParticipant):

    def readyToVote(self, txnService):
        self.log.append(("readyToVote", txnService))
        return True

    def voteForCommit(self, txnService):
        self.log.append(("voteForCommit", txnService))

    def commitTransaction(self, txnService):
        self.log.append(("commit", txnService))

    def abortTransaction(self, txnService):
        self.log.append(("abort", txnService))

    def finishTransaction(self, txnService, committed):
        self.log.append(("finish", txnService, committed))

class UnreadyParticipant(LogParticipant):

    def readyToVote(self, txnService):
        self.log.append(("readyToVote", txnService))
        return False

class ProcrastinatingParticipant(LogParticipant):

    status = 0

    def readyToVote(self, txnService):
        self.log.append(("readyToVote", txnService))
        old = self.status
        self.status += 1
        return old





class TestTxnState(TestCase):

    def setUp(self):
        self.ts = storage.TransactionService()
        self.p  = LogParticipant()
        self.log = self.p.log = []

    def test1_Outside(self):
        ts = self.ts
        for op in ts.commit, ts.abort, ts.fail:
            self.assertRaises(exceptions.OutsideTransaction, op)
        self.assertRaises(exceptions.OutsideTransaction, ts.join, self.p)
        assert self.log == []

    def test2_Fail(self):
        ts = self.ts
        ts.begin()
        ts.fail()
        self.assertRaises(exceptions.BrokenTransaction, ts.join, self.p)
        self.assertRaises(exceptions.BrokenTransaction, ts.commit)
        ts.abort()
        assert self.log == []

    def test3_StampAndActive(self):
        ts = self.ts
        assert not ts.isActive()
        assert ts.getTimestamp() is None
        for fini in ts.commit, ts.abort:
            ts.begin()
            assert ts.isActive()
            from time import time
            assert abs(time() - ts.getTimestamp()) < 5
            fini()
            assert not ts.isActive()
            assert ts.getTimestamp() is None
        assert self.log ==[]

    def test4_SimpleUse(self):
        ts = self.ts
        ts.begin()
        ts.join(self.p)
        ts.commit()
        ts.begin()
        ts.join(self.p)
        ts.abort()
        assert self.log == [
            ('readyToVote',ts), ('voteForCommit',ts), ('commit',ts),
            ('finish',ts,True), ('abort',ts), ('finish',ts,False),
        ]




class TestVoting(TestCase):

    def setUp(self):
        self.ts = storage.TransactionService()
        self.p_u  = UnreadyParticipant()
        self.p_p  = ProcrastinatingParticipant()
        self.p_n  = LogParticipant()
        self.log = self.p_u.log = self.p_p.log = self.p_n.log = []

    def test1_Unready(self):
        ts = self.ts
        ts.begin()
        ts.join(self.p_u)
        ts.join(self.p_p)
        ts.join(self.p_n)
        self.assertRaises(exceptions.NotReadyError, ts.commit)
        # just a lot of ready-to-vote attempts
        assert self.log == [('readyToVote',ts)]*len(self.log)

    def test2_Mixed(self):
        ts = self.ts
        ts.begin()
        ts.join(self.p_p)
        ts.join(self.p_n)
        ts.commit()
        # 2 participants * 1 retry for first, rest are all * 2 participants
        assert self.log == \
            [('readyToVote',ts)]   * 4 + \
            [('voteForCommit',ts)] * 2 + \
            [('commit',ts)]        * 2 + \
            [('finish',ts,True)]   * 2

    def test3_Normal(self):
        ts = self.ts
        ts.begin()
        ts.join(self.p_n)
        ts.commit()
        assert self.log == [
            ('readyToVote',ts),
            ('voteForCommit',ts),
            ('commit',ts),
            ('finish',ts,True),
        ]







class TxnTable(storage.TransactionComponent):

    stableState = ()
    colNames = ()

    table = binding.Make(lambda self: Table(self.colNames, self.stableState))

    def dump(self):
        colNames = self.colNames
        return [row.dump(colNames) for row in self.table]

    def commitTransaction(self, txnService):
        self.stableState = self.dump()

    def abortTransaction(self, txnService):
        self._delBinding('table')

    def DELETE(self, whereItems):
        self.joinedTxn
        self.table.DELETE(whereItems)

    def INSERT(self, items):
        self.joinedTxn
        self.table.INSERT(items)

    def INSERT_ROWS(self, rowList):
        self.joinedTxn
        self.table.INSERT_ROWS(self.colNames, rowList)

    def SELECT(self, whereItems=()):
        return self.table.SELECT(whereItems)

    def SET(self, whereItems, setItems):
        self.joinedTxn
        self.table.SET(whereItems, setItems)

    def __iter__(self):
        return iter(self.table)


class SampleTable(TxnTable):
    colNames = 'a', 'b'


class TestTable(TestCase):

    def setUp(self):
        self.state = config.snapshot()
        config.loadConfig()
        self.table = SampleTable()

    def tearDown(self):
        if settings.transaction_service().isActive():
            storage.abort()
        config.swap(self.state)

    def test1_NoChangeOutsideTxn(self):
        self.assertRaises(exceptions.OutsideTransaction,
            self.table.INSERT, dict(a=1,b=2)
        )

    def test2_Rollback(self):
        assert self.table.dump()==[]
        storage.begin()
        self.table.INSERT(dict(a=1,b=2))
        assert self.table.dump()==[(1,2)]
        storage.abort()
        assert self.table.dump()==[]

    def test3_Commit(self):
        assert self.table.dump()==[]
        storage.begin()
        self.table.INSERT(dict(a=1,b=2))
        assert self.table.dump()==[(1,2)]
        storage.commit()
        assert self.table.dump()==[(1,2)]
        storage.begin()
        self.table.INSERT(dict(a=3,b=4))
        assert self.table.dump()==[(1,2),(3,4)]
        storage.abort()
        assert self.table.dump()==[(1,2)]





def suite():
    return TestSuite((
        makeSuite(TestTxnState),
        makeSuite(TestVoting),
        makeSuite(TestTable),
    ))


if __name__ == '__main__':
    main(defaultTest='suite')
