#!/usr/bin/env python

import threading, Queue

__all__ = ['SerializedThread']

class SerializedThread(threading.Thread):

    STOP = object()
    ERROR = object()

    def __init__(self):
        threading.Thread.__init__(self)
        self._task_q = Queue.Queue()
        self._result_q = Queue.Queue()
        self._error_occured = False

    def run(self):
        while True:
            func, args, kwds = self._task_q.get()
            if func is self.STOP:
                self._result_q.put(self.STOP)
                return
            try:
                result = func(*args, **kwds)
            except:
                self._result_q.put(self.ERROR)
                raise
            else:
                # don't hold references to external objects
                self._result_q.put(result)
                del func, args, kwds, result

    def do(self, func, *args, **kwds):
        if self._error_occured:
            raise RuntimeError('stopped by exception')
        if not self.isAlive():
            self.start()
        self._task_q.put((func, args, kwds))
        res = self._result_q.get()
        if res is self.ERROR:
            self._error_occured = True
        return res

    def join(self):
        if not self._error_occured:
            if self.isAlive():
                self.do(self.STOP)
        threading.Thread.join(self)

if __name__ == '__main__':

    st = SerializedThread()
    other = SerializedThread()

    def add_two(a):
        print 'add_two(%s) in %s' % (a, threading.currentThread())
        return a + 2

    def long_run():
        import time
        print 'sleeping', threading.currentThread()
        time.sleep(1)
        return (42, 'foo')

    def raising():
        1/0

    r = st.do(add_two, 3)
    print r
    r = st.do(add_two, 4)
    print r
    r = other.do(add_two, 42)
    print r
    r = st.do(long_run)
    print r
    r = st.do(add_two, 1)
    print r
    r = st.do(raising)
    print r
    st.join()
    st.join()
    other.join()
