'''Mock SMTP server for unittesting

Usage:

    class TestMail(TestCase):

        def setUp(self):
            # fork mock SMTP server on localhost:2525
            self.smtpd = MockSmtpServer(port=2525).spawn()

        def tearDown(self):
            self.smtpd.exit()

        def test(self):

            # open connection to mock SMTP server
            # same as smtplib.SMTP('localhost',2525)
            con = self.smtpd.connection()

            # send any number of emails
            con.sendmail(
                'foo@example.cz', ['bar@example.com'],
                'Subject: test\n\nTEST BODY\n',
            )

            # check SMTP queue
            msgs = self.smtpd.messages()
            self.assertEqual(len(msgs), 1)

            # check sent mail
            msg = msgs[1]
            self.assertEqual(msg.mailfrom, 'foo@example.cz')
            self.assertEqual(msg.rcpttos, ['bar@example.com'])
            self.failUnless('TEST' in msg.data)

            # clear() method removes queued messages
            self.assertEqual(len(self.smtpd.messages()), 1)
            self.smtpd.clear()
            self.assertEqual(len(self.smtpd.messages()), 0)

'''


import sys, os, smtplib, smtpd, asyncore, pickle, signal
from wm4.running.process import *


class Message:
    peer = None
    mailfrom = None
    rcpttos = None
    data = None

    def __init__(self, **kw):
        self.__dict__.update(kw)

    def _msg(self):
        msg = self.__dict__.get('msg')
        if msg is None:
            from email import message_from_string
            msg = self.__dict__['msg'] = message_from_string(self.data)
        return msg
    msg = property(_msg)

    def get_raw(self, hdr):
        return self.msg.get(hdr)

    def get(self, hdr):
        if sys.version_info >= (2,5):
            from email.header import decode_header
        else:
            from email.Header import decode_header
        val = self.get_raw(hdr)
        if not val:
            return val
        return ''.join(
            unicode(p,e or 'ascii') for p,e in decode_header(val)
        )

    __getitem__ = get

    def get_content_charset(self):
        return self.msg.get_content_charset()

    def get_payload(self):
        return self.msg.get_payload()

    def get_text(self):
        return unicode(self.get_payload(), self.get_content_charset())


class TestSMTPChannel(smtpd.SMTPChannel):

    def smtp_TEST(self, arg):
        server = self._SMTPChannel__server
        if arg == 'exit':
            os._exit(0)
        elif arg == 'messages':
            data = pickle.dumps(server.trap)
            self.push('250 %d' %  len(data))
            server.ctrl.write(data)
            server.ctrl.flush()
        elif arg == 'clear':
            l = len(server.trap)
            server.trap[:] = []
            self.push('250 %d' %  l)
        else:
            self.push('501 unknown TEST command %r' % arg)



class TestSMTPServer(smtpd.SMTPServer):

    def __init__(self, ctrl, laddr=('localhost',2525), raddr=('localhost',25)):
        smtpd.SMTPServer.__init__(self, laddr, raddr)
        self.ctrl = ctrl
        self.trap = []

    def handle_accept(self):
        conn, addr = self.accept()
        print >> smtpd.DEBUGSTREAM, 'Incoming connection from %s' % repr(addr)
        channel = TestSMTPChannel(self, conn, addr)

    def process_message(self, peer, mailfrom, rcpttos, data):
        msg = Message(peer=peer, mailfrom=mailfrom, rcpttos=rcpttos, data=data)
        self.trap.append(msg)





class MockSmtpServer(ProcessTemplate):

    def __init__(self, port=2525):
        self.port = port

    pipes = ('msgs',)

    def run(self):
        server = TestSMTPServer(self.msgs, laddr=('localhost', self.port))
        self.msgs.write('ready\n')
        self.msgs.flush()
        while True:
            asyncore.loop()

    def _makeProxy(self, pid, pipes):

        proxy = super(MockSmtpServer, self)._makeProxy(pid, pipes)

        def connection():
            return smtplib.SMTP('localhost', self.port)
        proxy.connection = connection

        def messages():
            c, l = proxy._con.docmd('TEST', 'messages')
            data = proxy.msgs.read(int(l))
            return pickle.loads(data)
        proxy.messages = messages

        def clear():
            c, l = proxy._con.docmd('TEST', 'clear')
            return int(l)
        proxy.clear = clear

        def exit():
            proxy.checkStatus()
            if proxy.isRunning:
                try:
                    proxy._con.docmd('TEST', 'exit')
                except:
                    proxy.sendSignal(9)
            proxy.checkStatus(block=True)
        proxy.exit = exit


        try:
            ready = proxy.msgs.readline()
            assert ready.strip() == 'ready'
            proxy._con = connection()
            return proxy
        except:
            proxy.sendSignal(9)
            return None
