'''Mock XMLRPC server for unittesting

Usage:

    class TestMyService(TestCase):

        def setUp(self):
            self.xmlrpc = MockXMLRPCServer().spawn()

        def tearDown(self):
            if self.xmlrpc is not None:
                try:
                    self.xmlrpc.verify()
                finally:
                    self.xmlrpc.exit()

        def test(self):
            x = self.xmlrpc
            p = self.xmlrpc.proxy

            # two expectations
            # (<func name>, <arg tuple>, <return value>)
            x.expects('methodA', ('pi', 3.14), 'foo')
            x.expects('methodB', (42,), 'bar')

            # two rpc calls
            self.assertEqual(p.methodA('pi', 3.14), 'foo')
            self.assertEqual(p.methodB(42), 'bar')
'''

import xmlrpclib as lx
import SimpleXMLRPCServer as sx
import socket, os
from wm4.running.process import *


__all__ = ['MockXMLRPCSocketServer', 'MockXMLRPCServer']


class MockXMLRPCSocketServer(sx.SimpleXMLRPCServer):

    allow_reuse_address = True

    def __init__(self, addr=("localhost", 8787), expects=()):
        self._expects = list(expects)
        sx.SimpleXMLRPCServer.__init__(self, addr, logRequests=False)

    def expects(self, method, args, result, exception):
        self._expects.append((method, tuple(args), result, exception))
        return 1

    def _dispatch(self, method, params):
        if method == 'mock_expects':
            self.expects(*params)
            return 1
        elif method == 'mock_exit':
            self.server_close()
            os._exit(0)
        elif method == 'mock_verify':
            assert len(self._expects) == 0, self._expects
            return 1
        else:
            m, a, r, e = self._expects.pop(0)
            assert m == method, '%r x %r' % (m, method)
            assert a == params, '%r x %r' % (a, params)
            if e is not None:
                raise e
            return r


class MockXMLRPCServer(ProcessTemplate):

    def __init__(self, addr=("localhost", 8787)):
        self.addr = addr

    pipes = 'ctrl',

    def run(self):
        try:
            server = MockXMLRPCSocketServer(self.addr)
        except:
            os._exit(1)
        self.ctrl.write('ready\n')
        self.ctrl.flush()
        server.serve_forever()

    def _makeProxy(self, pid, pipes):
        proxy = super(MockXMLRPCServer, self)._makeProxy(pid, pipes)

        def expects(method, args, result=0, exception=None):
            return proxy.proxy.mock_expects(method, args, result, exception)
        proxy.expects = expects

        def verify():
            return proxy.proxy.mock_verify()
        proxy.verify = verify

        def exit():
            proxy.checkStatus()
            try:
                proxy.proxy.mock_exit()
            except:
                proxy.sendSignal(9)
            proxy.checkStatus(block=True)
        proxy.exit = exit

        proxy.proxy = lambda: lx.ServerProxy("http://%s:%s/" % self.addr)

        try:
            ready = proxy.ctrl.readline()
            assert ready.strip() == 'ready', 'RECEIVED %r' % ready
            proxy.proxy = lx.ServerProxy(
                "http://%s:%s/" % self.addr, allow_none=True)
        except:
            proxy.sendSignal(9)
            proxy.checkStatus(block=True)
            raise

        return proxy



def test():
    s = MockXMLRPCServer().spawn()
    try:
        p = s.proxy
        s.expects('foo', (1,'a'), 'bar')
        s.expects('foo', (2,'a'), 'bar')
        s.expects('foo', (3,'a'), 'bar')
        s.expects('time', (), 11121)
        print p.foo(1,'a')
        print p.foo(2,'a')
        print p.foo(3,'a')
        print p.time()
    finally:
        s.exit()


if __name__ == '__main__':
    test()
