# http://lukearno.com/projects/barrel/
# http://www.root.cz/clanky/bezpecne-prihlasovani-uzivatelu/
# http://www.berenddeboer.net/rest/authentication.html#permanent-logout

from wm4.api import *
from dispatch.strategy import default
from lib4.cgi import parse_auth
from lib4.urls import URL
from lib4.imports import importObject
import time

from interfaces import IUser, IGroup, app_secret, request
from environ import appURL, requestURI, requestURL
from sessions import TrustedClientData
from cookies import getCookie, setCookie, delCookie


__all__ = [
    'getUser', 'getAuthenticator', 'AuthenticationToken',
    'BaseAuthenticationMiddleware', 'HTTPAuthenticationMiddleware',
    'cookieLoginHandler', 'cookieLogoutHandler',
    'CookieAuthenticationMiddleware', 'TKTAuthenticationMiddleware',
]

_ = i18n.MessageIDFactory('wm4')


AUTHENTICATORS = 'windmill.authenticators'
DEFAULT_AUTH_DOMAIN = 'remote_user'


def getUser(environ, auth_domain=DEFAULT_AUTH_DOMAIN):
    return environ[AUTHENTICATORS][auth_domain].getUser(environ)


def getAuthenticator(environ, auth_domain=DEFAULT_AUTH_DOMAIN):
    return environ[AUTHENTICATORS][auth_domain]




class User:
    protocols.advise(instancesProvide=[IUser])
    uid = None
    name = None
    login = None
    groups = ()
    def __init__(self, **attrs):
        self.__dict__.update(attrs)
    def __repr__(self):
        return '<User %s in %s>' % (
            self.login,
            [g.name for g in self.groups]
        )

class Group:
    protocols.advise(instancesProvide=[IGroup])
    gid = None
    name = None
    def __init__(self, **attrs):
        self.__dict__.update(attrs)




class AuthenticationToken(TrustedClientData):
    attributes = [('username','U'), ('timestamp', 'd'), ('ttl', 'd')]
    digest_attributes = ('secret', 'remote_addr', 'browser')

    def __repr__(self):
        ts = time.strftime('%Y-%m-%dT%H:%M:%S',
                    time.localtime(self.timestamp))
        exp = time.strftime('%Y-%m-%dT%H:%M:%S',
                    time.localtime(self.timestamp + self.ttl))
        return 'AuthenticationToken(user=%s@%s created=%s expires=%s ttl=%d)' %(
                self.username, self.remote_addr, ts, exp, self.ttl)








class BaseAuthenticationMiddleware(binding.Component):
    on_ports = [443]
    auth_domain = DEFAULT_AUTH_DOMAIN
    password_encrypted = False
    secure_only = False
    find_user = lambda self, environ, username: None

    _find_user = binding.Make(lambda s: importObject(s.find_user))

    def __init__(self, application, **params):
        self.application = application
        self.__dict__.update(params)
        self.logger = logs.getLogger('web')

    def __call__(self, environ, start_response):
        environ.setdefault(AUTHENTICATORS, {})[self.auth_domain] = self
        exit_cu = config.enter_(settings.current_user_factory(
            lambda: getUser(environ)
        ))
        try:
            return self.application(environ, start_response)
        finally:
            exit_cu()

    def mustAuthenticate(self, environ):
        port = int(environ.get('SERVER_PORT', '80'))
        scheme = environ.get('wsgi.url_scheme','http')
        for p in self.on_ports:
            if p in ('*','all'):
                return True
            elif p in ('http', 'https'):
                if scheme == p:
                    return true
            else:
                if port == int(p):
                    return True
        return False

    def validateUsernameAndPassword(self, environ, username, password):
        if not username:
            return None
        user = self._find_user(self, environ, username)
        if user is None:
            self.logger.info('username %r not found' % username)
            return None
        # XXX encrypted passwords
        enc = self.password_encrypted
        if enc == 'crypt':
            import crypt
            password = crypt.crypt(password, user.password)
        elif not enc:
            pass
        else:
            raise ValueError('Unknown password encryption method', enc)

        if user.password != password:
            self.logger.info('invalid password for %r' % username)
            return None

        return user


class HTTPAuthenticationMiddleware(BaseAuthenticationMiddleware):

    realm = _(u'Enter your username and password')


    def authenticationFailed(self, environ, username=None, detail=None):
        raise web.Unauthorized(realm=self.realm, auth_domain=self.auth_domain)

    def getUsernameAndPassword(self, environ):
        if 'HTTP_CGI_AUTHORIZATION' in environ:
            environ.setdefault(
                'HTTP_AUTHORIZATION', environ['HTTP_CGI_AUTHORIZATION']
            )
        auth = environ.get('HTTP_AUTHORIZATION')
        if auth is None:
            return None, None
        return parse_auth(auth)

    def getUser(self, environ):
        if self.auth_domain in environ:
            return environ[self.auth_domain]
        if not self.mustAuthenticate(environ):
            environ[self.auth_domain] = None
            return None
        username, password = self.getUsernameAndPassword(environ)
        user = self.validateUsernameAndPassword(environ, username, password)
        if user is None:
            return self.authenticationFailed(environ, username)
        environ[self.auth_domain] = user
        return user








from wm4.web.routes import Route, Const, Signature
class cookieAuthenticationSitemap(Route):
    virtual = True
    route_name = Const('cookieAuthentication')
    class login(Route):
        match = 'login'
        action = Const('login')
        template = Const('wm4.web.bases/public_login.wt')
        handler = Const('wm4.web.authentication.cookieLoginHandler')
        signature = Signature('route_name', 'action')
    class logout(Route):
        match = 'logout'
        action = Const('logout')
        handler = Const('wm4.web.authentication.cookieLogoutHandler')
        signature = Signature('route_name', 'action')



from wm4.web.forms.api import *
class cookieLoginForm(FormCtrl):
    ctrlAllowedMethods = set([FormMethod.post])

    class ref(InputCtrl):
        ctrlVisibility = "hidden"
        def defaultInput(self):
            return self.ctrlReq.environ.get('HTTP_REFERER')

    class username(InputCtrl):
        ctrlRequired = True

    class password(PasswordCtrl):
        ctrlRequired = True

    class remember(InputCtrl):
        ctrlType = int
        def defaultOutput(self):
            return 0
        def parseValue(self, txt):
            try:
                return int(txt)
            except:
                return 0


def cookieLoginHandler(req):
    authenticator = getAuthenticator(
        req.environ, req.params.get('auth_domain', DEFAULT_AUTH_DOMAIN))

    if authenticator.secure_only:
        if not req.requestURI.startswith('https://'):
            loc = req.requestURI.replace('http://', 'https://')
            raise web.SeeOther(location=str(loc))

    form = cookieLoginForm(ctrlReq=req)
    res = form()
    message = ''

    if res:
        username = form['username'].newValue
        password = form['password'].newValue
        remember = form['remember'].newValue or 0
        try:
            user = authenticator.doLogin(
                    req, username, password, remember=remember)
        except Exception:
            user = None
            raise
        if user is None:
            message = _(u'Incorrect login or password')

    if not res or user is None:
        template = req.getResource(req.params['template'])
        return template.handleHTTP(req, 
                    authenticator=authenticator,
                    form=form,
                    message=message,
                    )
    loc = form['ref'].newValue
    if loc in (None, NOT_GIVEN, ''):
        loc = req.appURL
    raise web.SeeOther(location=str(loc))

def cookieLogoutHandler(req):
    authenticator = getAuthenticator(
        req.environ, req.params.get('auth_domain', DEFAULT_AUTH_DOMAIN))
    authenticator.doLogout(req)
    loc = req.getValue('ref')
    if not loc:
        loc = req.environ.get('HTTP_REFERER')
    if not loc:
        loc = req.appURL
    raise web.SeeOther(location=str(loc))


class CookieAuthenticationMiddleware(BaseAuthenticationMiddleware):
    cookie_domain = '/'

    login_url = '/login'
    login_handler = binding.Constant(cookieLoginHandler)
    login_template = 'wm4.web.bases/login.wt'

    logout_url = '/logout'
    logout_handler = binding.Constant(cookieLogoutHandler)

    cookie_name = binding.Make(lambda s: s.auth_domain)

    secret = binding.Make(lambda: app_secret())

    time = staticmethod(time.time)

    timeout = 3600
    remember_timeout = 3600 * 24 * 5

    def secure_cookie(self, environ):
        return environ['wsgi.url_scheme'] == 'https'

    def normalize_url(self, idict, attr):
        u = getattr(self, attr[1:])
        if u.startswith('/'):
            return u
        else:
            return '/'+u
    _login_url = _logout_url = binding.Make(normalize_url)

    def mustAuthenticate(self, environ):
        must = super(CookieAuthenticationMiddleware, self).mustAuthenticate(environ)
        if must and requestURL(environ).path == self.login_url:
            return False
        return must

    def authenticationFailed(self, environ, username=None, detail=None):
        login = self._login_url
        app = appURL(environ)
        if app.endswith('/'):
            location = app + login[1:]
        else:
            location = app + login
        location = URL(location, params={'ref':requestURI(environ)})
        raise web.SeeOther(location=str(location), headers=[
            ('X-Authentication-Domain', self.auth_domain)
        ])

    def __call__(self, environ, start_response):
        environ.setdefault(AUTHENTICATORS, {})[self.auth_domain] = self
        path_info = environ.get('PATH_INFO')
        try:
            return self.application(environ, start_response)
        except web.Unauthorized, e:
            if e.auth_domain == self.auth_domain:
                self.authenticationFailed(environ)
            else:
                raise

    def getUser(self, environ):
        if self.auth_domain in environ:
            return environ[self.auth_domain]
        token = self.getToken(environ)
        if token is None:
            if self.mustAuthenticate(environ):
                return self.authenticationFailed(environ)
            return None
        username = token.username
        user = self._find_user(self, environ, username)
        if user is None:
            if self.mustAuthenticate(environ):
                return self.authenticationFailed(
                    environ, username, detail='Login not found'
                )
            return None
        # check that token is not staled
        time = self.time()
        if (token.timestamp + token.ttl) < time:
            if self.mustAuthenticate(environ):
                return self.authenticationFailed(
                    environ, username, detail='Login timeouted'
                )
            return None
        # update timestamp in token
        token = self.createToken(
                environ, user, ttl=token.ttl, timestamp=time)
        self.setToken(environ, token)
        environ[self.auth_domain] = user
        return user

    def createToken(self, environ, user, ttl, timestamp=None):
        if timestamp is None:
            timestamp = self.time()
        return AuthenticationToken(
            username = user.login,
            timestamp = timestamp,
            ttl = ttl,
            secret = self.secret,
            remote_addr = environ['REMOTE_ADDR'],
            browser = environ.get('HTTP_USER_AGENT', 'unknown')
        )

    def getToken(self, environ):
        try:
            return AuthenticationToken.unpack(
                getCookie(environ, self.cookie_name),
                secret = self.secret,
                remote_addr = environ['REMOTE_ADDR'],
                browser = environ.get('HTTP_USER_AGENT', 'unknown')
            )
        except:
            return None

    def setToken(self, environ, token):
        p = environ.get('SCRIPT_NAME') or '/'
        s = self.secure_cookie(environ)
        environ.setdefault(
            'windmill.authentication_tokens', {})[self.cookie_name] = token
        setCookie(environ, self.cookie_name, str(token), path=p, secure=s)

    def deleteToken(self, environ):
        p = environ.get('SCRIPT_NAME') or '/'
        s = self.secure_cookie(environ)
        delCookie(environ, self.cookie_name, path=p, secure=s)

    def doLogin(self, req, username, password, remember=0):
        environ  = req.environ
        user = self.validateUsernameAndPassword(environ, username, password)
        if user is None:
            return None
        if remember:
            ttl = self.remember_timeout
        else:
            ttl = self.timeout
        token = self.createToken(environ, user, ttl)
        self.setToken(environ, token)
        environ[self.auth_domain] = user
        return user

    def doLogout(self, req):
        self.deleteToken(req.environ)





class TKTAuthenticationMiddleware(CookieAuthenticationMiddleware):

    cookie_name = binding.Constant('auth_tkt')

    def getUser(self, environ):
        login = environ.get("REMOTE_USER")
        if not login:
            return None
        user = User(
            login = login,
            name = login,
            groups = [Group(name=t) for t in
                environ.get('REMOTE_USER_TOKENS','').split(',') if t],
        )
        return user

    def createToken(self, environ, user, ttl, timestamp=None):
        if timestamp is None:
            timestamp = self.time()
        try:
            from hashlib import md5
        except ImportError:
            from md5 import md5
        user_data = str(ttl)
        tokens = ','.join([g.name for g in user.groups])
        ip_chars = ''.join(map(chr, map(int,environ['REMOTE_ADDR'].split('.'))))
        t = int(timestamp)
        ts = ((t & 0xff000000) >> 24,
              (t & 0xff0000) >> 16,
              (t & 0xff00) >> 8,
              t & 0xff)
        ts_chars = ''.join(map(chr, ts))

        digest0 = md5(
            ip_chars + ts_chars + self.secret + user.login + '\0' +
            tokens + '\0' + user_data
        ).hexdigest()
        digest = md5(digest0 + self.secret).hexdigest()
        cval = '%s%08x%s!' % (digest, t, user.login)
        if tokens:
            cval += tokens + '!'
        cval += user_data
        return cval

    def cookie(self):
        c = Cookie.SimpleCookie()
        c[self.cookie_name] = self.cookie_value().encode('base64').strip()
        c[self.cookie_name]['path'] = '/'
        if self.secure:
            c[self.cookie_name]['secure'] = 'true'
        return c
