# http://lukearno.com/projects/barrel/
# http://www.root.cz/clanky/bezpecne-prihlasovani-uzivatelu/

from wm4.api import *
from dispatch.strategy import default

from interfaces import IUser, IGroup, permission_checker
from authentication import DEFAULT_AUTH_DOMAIN, getUser

__all__ = [
    'Permission', '_And', '_Or', '_Not', 'Anybody', 'Nobody', 'Authenticated',
    'UserName', 'MemberOf', 'GroupName', 'RemoteAddr', 'ListPerm',
    'Denial', 'PermissionChecker', 'AuthorizationMiddleware',
]


_ = i18n.MessageIDFactory('wm4')

class Permission(object):

    def __and__(self, other):
        return _And(self, other)

    def __or__(self, other):
        return _Or(self, other)

    def __invert__(self):
        return _Not(self)

    def __eq__(self, other):
        return (
            self.__class__ == other.__class__ and self.__dict__== other.__dict__
        )

    def __ne__(self, other):
        return not self.__eq__(other)




class _Not(Permission):

    def __new__(klass, perm):
        if isinstance(perm, _Not):
            return perm.perm
        elif isinstance(perm, Nobody):
            return _anybody
        elif isinstance(perm, Anybody):
            return _nobody
        return Permission.__new__(klass)

    def __init__(self, perm):
        self.perm = perm

    def __repr__(self):
        return '~(%s)' % repr(self.perm)

class _And(Permission):

    def __new__(klass, *perms):
        if len(perms) == 1:
            return perms[0]
        return Permission.__new__(klass)

    def __init__(self, *perms):
        self.perms = perms

    def __repr__(self):
        return ' & '.join('(%r)' % p for p in self.perms)

class _Or(_And):

    def __repr__(self):
        return ' | '.join('(%r)' % p for p in self.perms)





class Anybody(Permission):
    
    def __repr__(self):
        return 'any()'

_anybody = Anybody()

class Nobody(Permission):

    def __repr__(self):
        return 'nobody()'

_nobody = Nobody()




class Authenticated(Permission):

    def __init__(self, auth_domain=DEFAULT_AUTH_DOMAIN, throw=None):
        self.auth_domain = auth_domain
        self.throw = throw

    def __repr__(self):
        return 'auth()'





class UserName(Permission):

    def __init__(self, login, auth_domain=DEFAULT_AUTH_DOMAIN):
        self.login = login
        self.auth_domain = auth_domain

    def __repr__(self):
        return 'user(%r)' % self.login



class MemberOf(Permission):

    def __init__(self, group, auth_domain=DEFAULT_AUTH_DOMAIN):
        self.group = group
        self.auth_domain = auth_domain

    def __repr__(self):
        return 'group(%r)' % self.group

GroupName = MemberOf


class RemoteAddr(Permission):
    '''Checks REMOTE_ADDR agains given IP range'''

    def __init__(self, ip):
        from adt4.ip import IP
        self.ip = IP(ip)

    def __repr__(self):
        return "ip('%s')" % self.ip



class ListPerm(Permission):
    
    def __init__(self, *perms, **kw):
        default = kw.get('default', None)
        if default is not None:
            assert isinstance(default, (Nobody, Anybody, Authenticated))
        self.default = default
        self.users = {}
        self.groups = {}
        for p in perms:
            self.add_perm(p)
    
    def add_perm(self, perm):
        if isinstance(perm, web._Not):
            perm = perm.perm
            if isinstance(perm, web.UserName):
                self.users[perm.login] = False
            elif isinstance(perm, web.MemberOf):
                self.groups[perm.group] = False
            else:
                raise NotImplementedError(perm)
        elif isinstance(perm, (Nobody, Anybody, Authenticated)):
            self.default = perm
        elif isinstance(perm, web.UserName):
            self.users[perm.login] = True
        elif isinstance(perm, web.MemberOf):
            self.groups[perm.group] = True
        else:
            raise NotImplementedError(perm)
        
    def copy(self, other=None):
        cp = self.__class__()
        cp.default = self.default
        cp.users = self.users.copy()
        cp.groups = self.groups.copy()
        if other is not None:
            if other.default is not None:
                cp.default = other.default
            cp.users.update(other.users)
            cp.groups.update(other.groups)
        return cp

    def perms(self):
        p = []
        for u,a in self.users.items():
            if a:
                p.append(UserName(u))
            else:
                p.append(_Not(UserName(u)))
        for g,a in self.groups.items():
            if a:
                p.append(MemberOf(g))
            else:
                p.append(_Not(MemberOf(g)))
        return p
    perms = property(perms)

    def __repr__(self):
        p = list(repr(p) for p in self.perms)
        if self.default is not None:
            p.append('default=%s' % repr(self.default))
        return 'list(%s)' % ', '.join(p)




class Denial(object):
    '''Object representing denial of access'''

    __slots__ = 'message',

    def __init__(self, message=_(u'Permission denied')):
        self.message = message

    def __nonzero__(self): return False
    __len__ = __nonzero__

    def __str__(self): return str(self.message)
    def __unicode__(self): return unicode(self.message)
    def __repr__(self): return 'Denial(%r)' % self.message





class PermissionChecker:

    globals = {
        'any': Anybody,
        'nobody': Nobody,
        'auth': Authenticated,
        'user': UserName,
        'group': MemberOf,
        'auth': Authenticated,
        'ip': RemoteAddr,
        'list': ListPerm,
    }

    def parse(self, expr):
        if not expr or not expr.strip():
            return None
        return eval(expr, self.globals.copy())
    
    @dispatch.generic()
    def allows(self, perm, environ):
        '''Generic function for checking permissions'''

    @allows.when(default)
    def deny(self, perm, environ):
        return Denial()

    @allows.when('perm == None or perm == NOT_GIVEN')
    def allow_none(self, perm, environ):
        return True

    @allows.when('perm in Anybody')
    def allow_nobody(self, perm, environ):
        return True

    @allows.when('perm in Nobody')
    def allow_nobody(self, perm, environ):
        return Denial()

    @allows.when('perm in _Not')
    def allow_not(self, perm, environ):
        if self.allows(perm.perm, environ):
            return Denial()
        else:
            return True

    @allows.when('perm in _And')
    def allow_and(self, perm, environ):
        for p in perm.perms:
            if not self.allows(p, environ):
                return Denial()
        return True

    @allows.when('perm in _Or')
    def allow_or(self, perm, environ):
        for p in perm.perms:
            if self.allows(p, environ):
                return True
        return Denial()

    @allows.when('perm in Authenticated')
    def allow_authenticated(self, perm, environ):
        if IUser(getUser(environ, perm.auth_domain), None):
            return True
        else:
            if perm.throw:
                raise web.Unauthorized(auth_domain=perm.auth_domain)
            else:
                return Denial(_(u"Permission denied for anonymous users"))

    @allows.when('perm in UserName')
    def allow_user_name(self, perm, environ):
        user = IUser(getUser(environ, perm.auth_domain), None)
        if user is None:
            return Denial(_(u"Permission denied for anonymous users"))
        if user.login == perm.login:
            return True
        return Denial()

    @allows.when('perm in MemberOf')
    def allow_member_of(self, perm, environ):
        user = IUser(getUser(environ, perm.auth_domain), None)
        group = perm.group
        if user is None:
            return Denial(_(u"Permission denied for anonymous users"))
        for g in user.groups:
            if g.name == group:
                return True
        msg = _(u"You are not member of group $group") % dict(group=group)
        return Denial(msg)

    @allows.when('perm in RemoteAddr')
    def allow_remote_addr(self, perm, environ):
        from adt4.ip import IP
        remote = IP(environ['REMOTE_ADDR'])
        if remote in perm.ip:
            return True
        return Denial()

    @allows.when('perm in ListPerm')
    def allow_list(self, perm, environ):
        user = IUser(web.getUser(environ), None)
        if user is None:
            default = perm.default
            if default is None:
                default = _nobody
            return self.allows(default, environ)

        groups = set(g.name for g in user.groups)
        allow = perm.users.get(user.login)
        if allow is True:
            return True
        elif allow is False:
            return Denial('login %r not allowed' % user.login)

        allowed_once = False
        for g in groups:
            allow = perm.groups.get(g)
            if allow is True:
                allowed_once = True
            elif allow is False:
                return Denial('group %r not allowed' % g)
        if allowed_once:
            return True
        else:
            default = perm.default
            if default is None:
                default = _nobody
            return self.allows(default, environ)





class AuthorizationMiddleware:


    def __init__(self, application, **params):
        self.application = application
        self.checker = permission_checker()
        self.__dict__.update(params)

    def getPermission(self, environ):
        route = environ['windmill.route']
        return getattr(route, 'permission', None)

    def __call__(self, environ, start_response):
        from errors import NotAllowed
        perm = self.getPermission(environ)
        res = self.checker.allows(perm, environ)
        if not res:
            raise NotAllowed(repr(perm), repr(res))
        return self.application(environ, start_response)



class AuthorizationRedirectMiddleware(AuthorizationMiddleware):

    def __call__(self, environ, start_response):
        perm = self.getPermission(environ)
        res = self.checker.allows(perm, environ)
        if not res:
            if not web.getUser(environ):
                req = web.request()
                u = req.get_url('cookieAuthentication', 'login') % dict(
                        ref=req.requestURL)
                raise web.SeeOther(u)
            else:
                from errors import NotAllowed
                raise NotAllowed(repr(perm), repr(res))
        return self.application(environ, start_response)


