# -*- coding: utf-8 -*-

import time
import os
import re

from wm4.api import *
from wm4.web.cookies import getCookie

class CachingMiddleware:

    ttl = 100
    key_prefix = ''
    disabled_patterns = []

    def __init__(self, application, **params):
        self.application = application
        self.__dict__.update(params)
        self.logger = logs.getLogger('web')
        self.cache = {}

    def get(self, key, default=None):
        return self.cache.get(key, default)

    def set(self, key, value):
        self.cache[key] = value

    def delete(self, key):
        if key in self.cache:
            del self.cache[key]

    def check_cacheable(self, environ):
        pi = environ.get('PATH_INFO', '')
        for p in self.disabled_patterns:
            if re.match(p, pi):
                return False
        if environ.get('REQUEST_METHOD','GET').upper() not in ('HEAD','GET'):
            return False
        if environ.get('wsgi.url_scheme','http') != 'http':
            return False
        if environ.get('QUERY_STRING'):
            return False
        if environ.get('HTTP_AUTHORIZATION'):
            return False
        if getCookie(environ, 'remote_user'):
            return False
        if getCookie(environ, 'registered_user'):
            return False
        return True

    def __call__(self, environ, start_response):
        if not self.check_cacheable(environ):
            return self.application(environ, start_response)

        now = time.time()
        pi = self.key_prefix + environ.get('SERVER_NAME', '') + environ.get('REQUEST_URI','')
        if 'HTTP_CACHE_CONTROL' in environ:
            self.delete(pi)
            self.logger.debug('no cache: reload')

        cstat, chead, cbody, ctime = self.get(pi, (None,None,None,None))
        if ctime is None or (now - ctime) > self.ttl:
            if ctime is None:
                self.logger.debug('no cache: not found')
            else:
                self.logger.debug('no cache: stale')
            response = []
            def start_response_detector(status, response_headers,exc_info=None):
                response.append((status, response_headers, exc_info))
                return start_response(status, response_headers, exc_info)
            cbody = self.application(environ, start_response_detector)
            cstat, chead, cexc = response[0]
            if cexc is None and cstat.startswith('200'):
                self.set(pi, (cstat, chead, cbody, now))
            else:
                self.logger.debug('no cache: not saving %s', cstat)
        else:
            self.logger.debug('cache: found')
            start_response(cstat, chead)
        return cbody

class MemcacheCachingMiddleware(CachingMiddleware):
    server = '127.0.0.1'
    port = 11211

    def __init__(self, application, **params):
        import memcache
        self.application = application
        self.__dict__.update(params)
        self.logger = logs.getLogger('web')
        self.cache = memcache.Client(['%s:%s' % (self.server, self.port)])

    def get(self, key, default=None):
        self.logger.debug('cache: get(%s)' % key)
        try:
            val = self.cache.get(key)
        except:
            val = None
        if val is None:
            return default
        return val

    def set(self, key, value):
        try:
            self.cache.set(key, value, self.ttl)
        except:
            pass

    def delete(self, key):
        try:
            self.cache.delete(key)
        except:
            pass
