from Cookie import SimpleCookie
from lib4.symbols import NOT_FOUND
from lib4.cgi import http_time, HTTP_COOKIE_FMT

__all__ = ['_getCookies', 'getCookie', 'setCookie', 'delCookie']

KEY = 'windmill.cookies'


def _getCookies(env):
    cookies = env.get(KEY)
    if cookies is None:
        cookies = env[KEY] = Cookies(env.get('HTTP_COOKIE',''))
    return cookies

class Cookies(SimpleCookie):
    '''Cookie jar that keeps status of newly set cookies'''

    encoding = 'utf-8' # XXX

    def __setitem__(self, key, value):
        SimpleCookie.__setitem__(self, key,value)
        self[key].isNew = True

    def load(self, rawdata):
        SimpleCookie.load(self, rawdata)
        for v in self.values():
            if not hasattr(v, 'isNew'):
                v.isNew = False

    def headers(self):
        for a in self.keys():
            v = self[a]
            if v.isNew:
                yield str(v.output(header='').strip())

    def getCookie(self, key, default=None):
        val = self.get(key, NOT_FOUND)
        if val is NOT_FOUND:
            return default
        else:
            return unicode(val.value, self.encoding)

    def setCookie(self, key, value, **kw):
        enc = self.encoding
        self[key] = unicode(value).encode(enc)
        for k,v in kw.items():
            if k == 'path':
                path = unicode(v).encode(enc)
                if not path.startswith('/'):
                    raise ValueError('Cookie path should start with slash', v)
                self[key]['path'] = path
            elif k == 'expires':
                self[key]['expires'] = http_time(v, HTTP_COOKIE_FMT)
            elif k in ('max_age', 'max-age'):
                self[key]['max-age'] = str(int(v))
            elif k == 'secure':
                if v:
                    self[key]['secure'] = 'secure'
            else:
                self[key][k] = unicode(v).encode(enc)

    def delCookie(self, key, **kw):
        '''Same as expiration in the past'''
        kw['expires'] = 1
        self.setCookie(key, '', **kw)



def getCookie(env, key, default=None):
    return _getCookies(env).getCookie(key, default)

def setCookie(env, key, value, **kw):
    return _getCookies(env).setCookie(key, value, **kw)

def delCookie(env, key, **kw):
    return _getCookies(env).delCookie(key, **kw)


class CookieMiddleware:

    callbacks = []

    def __init__(self, application, **params):
        self.application = application
        self.__dict__.update(params)

    def __call__(self, environ, start_response):
        response = []

        def _start_response(status, response_headers, exc_info=None):
            response.append(status)
            response.append(response_headers)
            response.append(exc_info)

        if self.callbacks:
            cookies = _getCookies(environ)
            for cb in self.callbacks:
                cb(environ, cookies)

        body = self.application(environ, _start_response)

        status, headers, exc_info = response
        if KEY in environ:
            headers = list(headers)
            for c in environ[KEY].headers():
                if len(c) > 4096:
                    raise ValueError('Cookie too large', c[:30] + '...')
                headers.append(('Set-Cookie', c))

        start_response(status, headers, exc_info)
        return body
