# http://blog.ianbicking.org/2008/05/16/pdb-in-the-browser/

from wm4.api import *
import sys, os

__all__ = [
 'SeeOther', 'TemporaryRedirect', 'NotModified',
 'BadRequest', 'Unauthorized','NotAllowed','NotFound', 'MethodNotAllowed',
 'WebException',
 'HTTPExceptionMiddleware', 'ErrorMiddleware',
]


TEMPLATE = """\
<html>\r
  <head><title>%(title)s</title></head>\r
  <body>\r
    <h1>%(title)s</h1>\r
    <p>%(body)s</p>\r
    <hr noshade>\r
    <div align="right">%(server)s</div>\r
  </body>\r
</html>\r
"""

class WebException(Exception):
    '''Internal server error'''
    code = 500
    title = 'Internal Server Error'
    logLevel = logs.ERROR
    headers = []
    detail = None
    msg = None
    args = ()

    def __init__(self, *args, **kw):
        klass = self.__class__
        self.args = args
        for a,v in kw.items():
            if not hasattr(klass, a):
                raise AttributeError('Class %r has no attribute %r' % (klass,a))
            setattr(self, a, v)

    def makeHeaders(self, envrion, traceback):
        headers = list(self.headers or ())
        headers.append(('Content-type', 'text/plain'))
        return headers

    def makeBody(self, environ, traceback):
        return ['%d %s\r\n' % (self.code, self.title)]

    def handleException(self, environ, traceback):
        return '%d %s' % (self.code, self.title), \
               self.makeHeaders(environ,traceback), \
               self.makeBody(environ,traceback)



class SeeOther(WebException):
    '''Location should be specified as first argument (location,)'''
    code = 303
    title = 'See Other'
    logLevel  = logs.DEBUG
    location = None

    def __init__(self, location, **kw):
        WebException.__init__(self, **kw)
        self.location = location

    def makeHeaders(self, envrion, traceback):
        headers = list(self.headers or ())
        headers.append(('Location', str(self.location)))
        return headers

    #def handleException(self, environ, traceback):
    #    status = '%d %s' % (self.code, self.title)
    #    return status, self.makeHeaders(environ, traceback), self.makeBody

class TemporaryRedirect(SeeOther):
    code = 307
    title = 'Temporary Redirect'

class NotModified(WebException):
    code = 304
    title = 'Not Modified'
    logLevel  = logs.DEBUG

    def makeHeaders(self, envrion, traceback):
        return self.headers or []

    def makeBody(self, environ, traceback):
        return []

class BadRequest(WebException):
    code = 400
    title = 'Bad Request'
    logLevel  = logs.ERROR
    msg = None

class Unauthorized(WebException):
    code = 401
    title = 'Unauthorized'
    logLevel  = logs.DEBUG
    realm = u'Enter your username and password'
    auth_domain = None

    def makeHeaders(self, envrion, traceback):
        realm = unicode(self.realm or self.__class__.realm).encode('utf-8')
        headers = list(self.headers or ())
        headers.append(('Content-type', 'text/plain'))
        headers.append(('WWW-Authenticate', 'Basic realm="%s"' % realm))
        if self.auth_domain:
            headers.append(('X-Authentication-Domain', str(self.auth_domain)))
        return headers

class NotAllowed(WebException):
    code = 403
    title = 'Forbidden'
    logLevel = logs.DEBUG

class NotFound(WebException):
    code = 404
    title = 'Not Found'
    logLevel = logs.DEBUG

class MethodNotAllowed(WebException):
    code = 405
    title = 'Method Not Allowed'
    logLevel  = logs.WARNING



http_status_map = {
    303 : SeeOther,
    304 : NotModified,
    307 : TemporaryRedirect,
    400 : BadRequest,
    401 : Unauthorized,
    403 : NotAllowed,
    404 : NotFound,
    405 : MethodNotAllowed,
    500 : WebException,
}



class HTTPExceptionMiddleware:

    throw_errors = False
    log_errors = True
    template403 = 'wm4.web.bases/errors.wt'
    template404 = 'wm4.web.bases/errors.wt'
    template500 = 'wm4.web.bases/errors.wt'

    def __init__(self, application, **params):
        self.application = application
        self.__dict__.update(params)
        self.logger = logs.getLogger('web')

    def __call__(self, environ, start_response):
        import traceback

        response = []
        def start_response_detector(status, response_headers, exc_info=None):
            response.append((status, response_headers, exc_info))
            return start_response(status, response_headers, exc_info)

        try:
            return self.application(environ, start_response_detector)
        except WebException, exc:
            exc_info = sys.exc_info()
            # XXX docasne logovani
            #logger = logs.getLogger('web')
            #logger.error('exception:', exc_info=exc_info)
            self.logger.log(
                exc_info[0].logLevel, 'web_exception', exc_info=exc_info)
            try:
                if exc.code >= 500:
                    storage.abort()
                handler = getattr(self, 'handle%s' % exc.code, None)
                template = getattr(self, 'template%s' % exc.code, None)
                if handler:
                    s,h,b = handler(exc, environ, exc_info[2])
                elif template:
                    s,h,b = exc.handleException(environ, exc_info[2])
                    req = web.Request(environ=environ)
                    tmpl = web.getResource(template)
                    html = web.renderTemplate(
                        tmpl, req, error=exc, traceback=traceback)
                    html = html.encode('utf-8')
                    for k,v in [
                        ('Content-type', 'text/html; charset=utf-8'),
                        ('Content-Length', str(len(html))),
                    ]:
                        for i in reversed(range(len(h))):
                            if h[i][0].lower() == k.lower():
                                del h[i]
                        h.append((k,v))
                    b = [html]
                else:
                    s,h,b = exc.handleException(environ, exc_info[2])
                start_response(s, h)
                return b
            finally:
                exc_info = None


class HTTPExceptionMiddlewareRedirect(HTTPExceptionMiddleware):

    def handle403(self, exc, environ, traceback):
        req = web.Request(environ=environ)
        location = req.get_url('cookieAuthentication', 'login') % dict(
                        ref=req.requestURL)
        #location = URL(os.path.join(req.appURL, 'login'),
        #               params=dict(ref=req.requestURL))
        body = ''
        headers = [('Location', str(location))]
        return '303 SeeOther', headers, body

class HTTPExceptionMiddlewareSlash(HTTPExceptionMiddleware):

    def handle404(self, exc, environ, traceback):
        from lib4.urls import URL
        exc_info = sys.exc_info()
        req = web.Request(environ=environ)
        url = req.requestURL
        if not url.endswith('/'):
            orig = req.requestURI
            location = URL(url+'/', params=orig.params, target=orig.target)
            return '303 SeeOther', [('Location', str(location))], ''
        else:
            template = getattr(self, 'template%s' % exc.code, None)
            s,h,b = exc.handleException(environ, exc_info[2])
            if template:
                req = web.Request(environ=environ)
                tmpl = web.getResource(template)
                html = web.renderTemplate(
                    tmpl, req, error=exc, traceback=traceback)
                html = html.encode('utf-8')
                for k,v in [
                    ('Content-type', 'text/html; charset=utf-8'),
                    ('Content-Length', str(len(html))),
                ]:
                    for i in reversed(range(len(h))):
                        if h[i][0].lower() == k.lower():
                            del h[i]
                    h.append((k,v))
                b = [html]
            return s,h,b

class ErrorMiddleware:

    throw_errors = False
    log_errors = True
    show_details = False
    logger = None
    exit_on_error = True

    template = 'wm4.web.bases/errors.wt'

    def __init__(self, application, **params):
        self.application = application
        self.__dict__.update(params)
        self.logger = logs.getLogger('web')

    def __call__(self, environ, start_response):

        response = []
        def start_response_detector(status, response_headers, exc_info=None):
            response.append((status, response_headers, exc_info))
            return start_response(status, response_headers, exc_info)

        if self.exit_on_error:
            environ['wsgi.exit'] = True
        try:
            body = self.application(environ, start_response_detector)
            return body
        except KeyboardInterrupt:
            os._exit(0)
        except:
            if self.throw_errors:
                raise
            exc_info = sys.exc_info()
            try:
                s,h,b = self.handleException(environ, exc_info)
                start_response(s, h, exc_info)
                return b
            finally:
                exc_info = None

    def handleException(self, environ, exc_info):
        import traceback, pprint

        status = '500 Internal Server Error'
        try:
            env = environ.copy()
            if 'REMOTE_PASSWORD' in env:
                del env['REMOTE_PASSWORD']
            message = \
                pprint.PrettyPrinter(indent=2).pformat(env) +\
                '\n-----------------------------\n' +\
                ''.join(traceback.format_exception(*exc_info))
        except:
            message = 'Error in error handler :-)'
        try:
            self.logger.error(message)
        except:
            pass
        if self.show_details:
            return status, [('Content-type', 'text/plain')], [message]
        else:
            exc = WebException()
            req = web.Request(environ=environ)
            tmpl = web.getResource(self.template)
            html = web.renderTemplate(tmpl, req, error=exc, traceback=None)
            html = html.encode('utf-8')
            return status, [
                ('Content-type', 'text/html; charset=utf-8'),
                ('Content-Length', str(len(html))),
            ], [html]
