# XXX property -> cell.Value/cell.Input

from wm4.api import *
from adt4.enums import *
from lib4.imports import importObject
from itertools import chain
from wm4.utils.files import IFileInfo, IImageInfo
from datetime import date, datetime
import os

from wm4.web.sessions import TrustedClientData
from wm4.web.metadata import *
from wm4.web.interfaces import (
    app_secret, upload_manager, IFormElement, IFormInput, IFormAction,
    FORM_WIDGET,
)

_ = i18n.MessageIDFactory('wm4')


__all__ = [
    'FormMethod',
    'OnClickEvent', 'OnChangeEvent', 'OnAddEvent', 'OnRemoveEvent',
    'OnErrorEvent', 'event_handler',
    'CSValue', 'ClientState',
    'AbstractCtrlMeta', 'AbstractCtrl',
    'BaseActionCtrl', 'ButtonCtrl', 'SubmitCtrl',
    'DependentSpanCtrl',
    'ValueCtrl', 'InputCtrl', 'HiddenCtrl',
    'PasswordCtrl', 'ValidatingPasswordCtrl',
    'DateCtrl', 'DatetimeCtrl',
    'CheckBoxCtrl', 'TextAreaCtrl', 'RestCtrl', 'HtmlCtrl',
    'FileUploadCtrl', 'ImageUploadCtrl',
    'Option', 'OptGroup', 'SelectCtrl', 'BoolCtrl', 'EnumCtrl',
    'MultiValueCtrl', 'MultiSelectCtrl',
    'MultiCheckboxCtrl', 'HiearchicalMultiCheckboxCtrl',
    'EntityReferenceCtrl',
    'ContainerCtrl', 'InstanceCtrl', 'EntityCtrl', 'SelectOrAddEntityCtrl',
    'PopupCtrl',
    'RepeatCtrl', 'SequenceItemCtrl', 'SequenceCtrl', 'FormCtrl',
    'makeFeatureCtrl', 'makeFormCtrl',
]

class FormMethod(Enum):
    get   = enum(('get', 'application/x-www-form-urlencoded'))
    post  = enum(('post', 'application/x-www-form-urlencoded'))
    multi = enum(('post', 'multipart/form-data'))



class OnClickEvent(events.Event):
    __slots__ = ()

class OnChangeEvent(events.Event):
    __slots__ = ('new', 'old')

class OnAddEvent(events.Event):
    __slots__ = ('item',)

class OnRemoveEvent(events.Event):
    __slots__ = ('item',)

class OnErrorEvent(events.Event):
    __slots__ = ('error',)




def event_handler(event, sender=None, hold=False):
    if sender is None:
        sender = '.'
    def decorator(func):
        attr = '__event_handler__'
        if hasattr(func, attr):
            eh = func.__event_handler__
        else:
            eh = func.__event_handler__ = []
        eh.append((event, sender, hold))
        return func
    return decorator




class CSValue(TrustedClientData):
    attributes = (('val','S'),)
    digest_attributes = ('secret',)

class ClientState(object):

    protocols.advise(instancesProvide=[binding.IActiveDescriptor])

    def __init__(
        self, typ=unicode, default=lambda s,o,a:NOT_GIVEN,
        secure=False, attr=None
    ):
        self.typ = typ
        self.default = default
        self.attr = attr
        self.secure = secure

    def activateInClass(self, klass, attr):
        if self.attr != attr:
            self = self.__class__(self.typ, self.default, self.secure, attr)
            setattr(klass, attr, self)
        return self

    def eid(self, obj):
        e = obj.ctrlId
        a = self.attr
        return e and e + '$' + a or a

    def render(self, val):
        return unicode(val)

    def ctrlState(self, obj):
        eid = self.eid(obj)
        val = self.__get__(obj, NOT_GIVEN)
        if val is not NOT_GIVEN:
            val = self.render(val)
            if self.secure:
                val = str(CSValue(val=val.encode('utf-8'), secret=app_secret()))
        return eid, val

    def __get__(self, obj, typ):
        if obj is None:
            return self
        else:
            var = self.eid(obj)
            odct = obj.__dict__
            key = self
            if  key in odct:
                return odct[key]
            else:
                val = obj.ctrlReq.getValue(var, NOT_FOUND)
                if val is NOT_FOUND:
                    val = self.default(self, obj, self.attr)
                if val is not NOT_GIVEN:
                    if self.secure:
                        val = unicode(
                            CSValue.unpack(val, secret=app_secret()).val,'utf-8'
                        )
                    val = self.typ(val)
                odct[key] = val
                return val

    def __set__(self, obj, value):
        obj.__dict__[self] = self.typ(value)

    def __delete__(self, obj):
        obj.__dict__[self] = NOT_GIVEN





class AbstractCtrlMeta(binding.Activator):

    __metaclass__ = binding.Activator

    def form_elements_and_handlers(klass, cdict, attr):
        features = {}
        handlers = []
        state = {}
        for attr in dir(klass):
            if attr in (
                '__ctrl_features__', '__event_handlers__','__state_features__'
            ):
                continue
            desc = getattr(klass, attr)
            if hasattr(desc, '__event_handler__'):
                handlers.append((attr, desc.__event_handler__))
            if attr not in klass.ctrlIgnoredFeatures:
                if isinstance(desc, AbstractCtrlMeta):
                    features[attr] = desc
                elif isinstance(desc, ClientState):
                    state[attr] = desc
        return {'__ctrl_features__':features,
                '__event_handlers__':handlers,
                '__state_features__':state}

    __ctrl_features__ = \
    __event_handlers__ = \
    __state_features__ = binding.MultiMake(form_elements_and_handlers)


class AbstractCtrl(object):

    __metaclass__ = AbstractCtrlMeta

    protocols.advise(instancesProvide=[IFormElement])

    # Controller binding and component hierarchy

    ctrlIgnoredFeatures = set()

    ctrlName = None
    ctrlParent = None
    _ctrlChildren = binding.Require()
    _ctrlParents = property(lambda s:(
        s.ctrlParent is None and (s,) or s.ctrlParent._ctrlParents+(s,)))
    ctrlRoot = property(
        lambda s: s._ctrlParents[0])
    def ctrlId(self):
        if self._ctrlParents[0].ctrlName is None:
            return '$'.join(e.ctrlName for e in self._ctrlParents[1:]) or None
        else:
            return '$'.join(e.ctrlName for e in self._ctrlParents) or None
    ctrlId = property(ctrlId)

    # Controller hierarchy tree traversal

    def ctrlChildren(self):
        return self._ctrlChildren.itervalues()

    def ctrlIter(self):
        yield self
        last = self
        for node in self.ctrlIter():
            for child in node.ctrlChildren():
                yield child
                last = child
            if last is node:
                return

    def __getitem__(self, path):
        # absolute or relative path
        if path.startswith('/'):
            elem = self.ctrlRoot
            path = path[1:]
        else:
            elem = self
        # traverse path
        for p in path.split('/'):
            if p in ('', '.'):
                continue
            elif p == '..':
                elem = elem.ctrlParent
            else:
                try:
                    elem = elem._ctrlChildren[p]
                except KeyError:
                    raise KeyError(p, 'of', self.ctrlId)
        return elem


    # Description

    ctrlLabel = binding.Make(lambda s: s.ctrlName.capitalize())
    ctrlHelp = None
    ctrlDescription = None

    # Values and errors

    ctrlError = None
    ctrlType = object
    ctrlRequired = False
    oldValue = NOT_GIVEN
    newValue = binding.Make(lambda s: s.oldValue)

    # Controller behaviour

    def ctrlVisibility(self):
        if self.ctrlParent is None: return 'visible'
        else: return self.ctrlParent.ctrlVisibility
    ctrlVisibility = binding.Make(ctrlVisibility)

    def ctrlReadOnly(self):
        if self.ctrlParent is None: return False
        else: return self.ctrlParent.ctrlReadOnly
    ctrlReadOnly = binding.Make(ctrlReadOnly)

    ctrlWidget = None

    def ctrlAllowedMethods(self):
        fiter = iter(self.ctrlChildren())
        try:
            am = fiter.next().ctrlAllowedMethods
        except StopIteration:
            return set(FormMethod.values())
        for f in fiter:
            am = am & f.ctrlAllowedMethods
        return am
    ctrlAllowedMethods = binding.Make(ctrlAllowedMethods)


    ctrlReq = binding.Require()

    def __init__(
        self, ctrlReq, ctrlName=None, ctrlParent=None, oldValue=NOT_GIVEN, **kw
    ):
        self.ctrlInit(ctrlReq, ctrlName, ctrlParent, oldValue, kw)
        self._ctrlChildren = self._ctrlSetupChildren()

    def ctrlInit(self, ctrlReq, ctrlName, ctrlParent, oldValue, kw):
        self.ctrlReq = ctrlReq
        self.ctrlName = ctrlName
        self.ctrlParent = ctrlParent
        if oldValue is not NOT_GIVEN:
            self.oldValue = oldValue
        binding.initAttrs(self, kw.items())

    def _ctrlSetupChildren(self):
        features = {}
        data = self.oldValue
        for attr, featCls in self.__class__.__ctrl_features__.items():
            if data is NOT_GIVEN:
                cdata = NOT_GIVEN
            else:
                cdata = getattr(data, attr, NOT_GIVEN)
            feat = featCls(self.ctrlReq, attr, self, oldValue=cdata)
            features[feat.ctrlName] = feat
        return features

    def ctrlSetupHandlers(self):
        for attr, ehl in self.__class__.__event_handlers__:
            handler = getattr(self, attr)
            for evt, path, hold in ehl:
                sender = self[path]
                evt.subscribe(sender, handler, hold)

    formAction = binding.Delegate('ctrlRoot')

    def ctrlGetAction(self):
        pass

    def ctrlOnInit(self):
        for attr in self.__class__.__state_features__:
            getattr(self, attr)

    featureCommiting = {'*':'before'}

    def _commitBeforeAfter(self):
        fc = self.featureCommiting; cl = []
        dc = fc.get('*', 'none')
        _commitBefore = []; _commitAfter = []
        for a,f in self._ctrlChildren.iteritems():
            c = fc.get(a, dc)
            if c == 'before':
                _commitBefore.append(a)
            elif c == 'after':
                _commitAfter.append(a)
        return locals()
    _commitBefore = _commitAfter = binding.MultiMake(_commitBeforeAfter)

    def commitBefore(self):
        res = True
        for a in self._commitBefore:
            if not self[a].commitChanges():
                res = False
        return res

    def commitSelf(self):
        return True

    def commitChanges(self):
        if not self.commitBefore():
            return False
        if not self.commitSelf():
            return False
        if self.ctrlError:
            return False
        if not self.commitAfter():
            return False
        return True

    def commitAfter(self):
        res = True
        for a in self._commitAfter:
            if not self[a].commitChanges():
                res = False
        return res

    def ctrlState(self):
        for a,d in self.__class__.__state_features__.iteritems():
            i, v = d.ctrlState(self)
            if v is not NOT_GIVEN:
                yield i,v
        for child in self.ctrlChildren():
            for x in child.ctrlState():
                yield x




@binding.getParentComponent.when(AbstractCtrl)
def get_parent_of_AbstractCtrl(component):
    return component.ctrlParent

@binding.getComponentName.when(AbstractCtrl)
def get_name_of_AbstractCtrl(component):
    return component.ctrlName





class BaseActionCtrl(AbstractCtrl):

    protocols.advise(instancesProvide=[IFormAction])

    jsOnClick = None

    def ctrlGetAction(self):
        if self.ctrlId in self.ctrlReq:
            return self

class ButtonCtrl(BaseActionCtrl):
    ctrlWidget = FORM_WIDGET('button')

class SubmitCtrl(ButtonCtrl):
    ctrlWidget = FORM_WIDGET('submit')
    ctrlLabel = binding.Constant(_(u'Save'))



class DependentSpanCtrl(AbstractCtrl):
    ctrlWidget = FORM_WIDGET('dependent_span')
    dependsOn = '..'
    defaultText = ''
    spanId = binding.Make(lambda s: s.ctrlReq.get_new_id())


class ValueCtrl(AbstractCtrl):

    protocols.advise(instancesProvide=[IFormInput])

    ctrlWidget = FORM_WIDGET('input.text')
    ctrlType = unicode

    emptyIsNone = True
    stripInput = staticmethod(unicode.strip)

    def ctrlOnInit(self):
        # Load ClientState
        super(ValueCtrl,self).ctrlOnInit()
        # ctrlId can change if the element is moved somewhere
        # or its name is changed. User input must be taken
        # before any action/click.
        self.inputText

    def renderValue(self, val):
        return web.render(self.ctrlReq, val)

    def parseValue(self, txt):
        typ = self.ctrlType
        try:
            return schema.normalize(txt, typ)
        except:
            raise schema.NormalizationError(
                self.ctrlName, value=txt, type_name=i18n.typeName(typ)
            )

    def defaultInput(self):
        return None

    def defaultOutput(self):
        return self.oldValue

    def getInputText(self):
        fid = self.ctrlId
        req = self.ctrlReq
        if self.formAction or fid in req:
            txt = req.getValue(fid, u'')
            txt = self.stripInput(txt)
            return txt
        else:
            if self.oldValue is NOT_GIVEN:
                return self.renderValue(self.defaultInput())
            else:
                return self.renderValue(self.oldValue)
    inputText = binding.Make(getInputText)

    def commitSelf(self):
        if not self.ctrlId in self.ctrlReq:
            self.newValue = self.defaultOutput()
        else:
            if not self.inputText and self.ctrlVisibility == 'hidden':
                self.newValue = NOT_GIVEN
            elif not self.inputText and self.emptyIsNone:
                if self.ctrlRequired:
                    self.ctrlError = schema.RequiredError('')
                    self.newValue = NOT_GIVEN
                    return False
                else:
                    self.newValue = None
            else:
                try:
                    self.newValue = self.parseValue(self.inputText)
                except schema.SchemaError, e:
                    self.ctrlError = e
                    self.newValue = NOT_GIVEN
                    OnErrorEvent(self, error=e)
                    return False
        if self.oldValue != self.newValue:
            OnChangeEvent(self, old=self.oldValue, new=self.newValue)
        return True

class InputCtrl(ValueCtrl):
    maxlength = None

class HiddenCtrl(ValueCtrl):
    ctrlVisibility = 'hidden'

class PasswordCtrl(InputCtrl):
    ctrlWidget = FORM_WIDGET('input.password')


class ValidatingPasswordCtrl(PasswordCtrl):

    class validation(PasswordCtrl):
        def defaultInput(self):
            return self['..'].inputText

    def parseValue(self, txt):
        typ = self.ctrlType
        val1 = unicode(txt)
        val2 = unicode(self['validation'].inputText)
        if val1 != val2:
            raise schema.FeatureError(
                self.ctrlName, msgId=_(u'Passwords differ'))
        return val1



class DateCtrl(ValueCtrl):
    ctrlType = date
    ctrlWidget = FORM_WIDGET('input.date')
    dateFormat = 'dd.mm.yyyy' # XXX locales

class DatetimeCtrl(ValueCtrl):
    ctrlType = datetime
    ctrlWidget = FORM_WIDGET('input.datetime')
    dateFormat = 'dd.mm.yyyy' # XXX locales


class CheckBoxCtrl(InputCtrl):
    # XXX parsing
    ctrlType = bool



class TextAreaCtrl(ValueCtrl):
    rows = None
    cols = None
    stripInput = staticmethod(unicode)
    ctrlWidget = FORM_WIDGET('input.area')



class FileUploadCtrl(AbstractCtrl):

    protocols.advise(instancesProvide=[IFormInput, IFileInfo])

    ctrlRequired = False
    ctrlAllowedMethods = binding.Constant(set((FormMethod.multi,)))
    ctrlWidget = FORM_WIDGET('input.file')

    filename = ClientState()
    content_type = ClientState()
    size = ClientState(int)
    path = ClientState(secure=True)
    deleted = ClientState(int, default=lambda s,o,a:0)

    sizeLimit = 0
    allowedMimeTypes = ('*/*',)
    download = None

    class deleteCurrent(ButtonCtrl):
        ctrlLabel = binding.Constant(_(u'Remove'))

    @event_handler(OnClickEvent, 'deleteCurrent')
    def on_delete(self, evt):
        del self.filename
        del self.size
        del self.path
        del self.content_type
        self.deleted = 1

    def ctrlOnInit(self):
        super(FileUploadCtrl,self).ctrlOnInit()
        if not self.formAction and self.oldValue not in (NOT_GIVEN, None):
            old = IFileInfo(self.oldValue)
            self.filename = old.filename
            self.content_type = old.content_type
            self.size = old.size
        self.inputText

    mgr = binding.Make(lambda: importObject(upload_manager())())

    # XXX file convertor
    # XXX sizeLimit
    # XXX allowed mime types
    def getInputText(self):
        if not self.formAction:
            return ''
        val = self.ctrlReq.getUpload(self.ctrlId)
        if val is not None:
            self.filename = val.filename
            self.content_type = val.content_type
            self.size = val.size
            self.path = val.path
            self.deleted = 0
        return ''
    inputText = binding.Make(getInputText)

    def makeFile(self):
        return self.ctrlType(
            filename=self.filename, path=self.path, size=self.size,
            content_type=self.content_type
        )

    def commitSelf(self):
        if not self.ctrlId in self.ctrlReq:
            # keep old value
            return True
        if self.oldValue is NOT_GIVEN:
            if self.deleted or self.path is NOT_GIVEN:
                nv = None
            elif self.path and not self.size and self.emptyIsNone:
                nv = None
            else:
                nv = 'new'
        else:
            if self.deleted:
                nv = None
            elif self.path is NOT_GIVEN:
                nv = 'old'
            elif self.path and not self.size and self.emptyIsNone:
                nv = None
            else:
                nv = 'new'
        if nv == 'old':
            return True
        elif nv == 'new':
            try:
                self.newValue = self.makeFile()
            except schema.SchemaError, e:
                self.newValue = NOT_GIVEN
                self.ctrlError = e
                OnErrorEvent(self, error=e)
                return False
        else:
            assert nv == None
            if self.ctrlRequired:
                self.ctrlError = schema.RequiredError('')
                self.newValue = NOT_GIVEN
                return False
            else:
                self.newValue = None
        if self.oldValue != self.newValue:
            OnChangeEvent(self, old=self.oldValue, new=self.newValue)
        return True

class ImageUploadCtrl(FileUploadCtrl):

    protocols.advise(instancesProvide=[IImageInfo])

    ctrlWidget = FORM_WIDGET('input.image')

    width = ClientState(int)
    height = ClientState(int)
    converter = None

    # XXX max size

    def ctrlOnInit(self):
        super(FileUploadCtrl,self).ctrlOnInit() # FileUploadCtrl
        if not self.formAction and self.oldValue not in (NOT_GIVEN, None):
            old = IImageInfo(self.oldValue)
            self.filename = old.filename
            self.content_type = old.content_type
            self.size = old.size
            self.width = old.width
            self.heigh = old.height
        self.inputText

    # XXX conversions
    def getInputText(self):
        if not self.formAction:
            return ''
        val = self.ctrlReq.getUpload(self.ctrlId)
        if val is not None:
            from lib4.images import loadImage, getMIME, getStorageSize
            try:
                img = loadImage(val.path) # XXX may raise error
            except:
                self.path = val.path
                return ''
            if self.converter:
                _img = img
                img = self.converter(_img)
                if _img is not img:
                    img.save(val.path, img.format)
            self.filename = val.filename
            self.content_type = getMIME(img)
            self.path = val.path
            self.deleted = 0
            self.width, self.height = img.size
            self.size = getStorageSize(img)
        return ''
    inputText = binding.Make(getInputText)

    def makeFile(self):
        return self.ctrlType(
            filename=self.filename, path=self.path, size=self.size,
            content_type=self.content_type, width=self.width, height=self.height
        )




class RestCtrl(TextAreaCtrl):
    ctrlWidget = FORM_WIDGET('input.rest')

    # ReST parameters
    initial_header_level = 2
    report_level = 2
    halt_level = 6
    # XXX TODO parsing

    class upload(FileUploadCtrl):
        class ctrlType:
            def __init__(self, **kw):
                self.__dict__.update(kw)

    @event_handler(OnChangeEvent, 'upload')
    def on_upload(self, evt):
        new = evt.sender.newValue
        if new in (None, NOT_GIVEN):
            return
        self.inputText = unicode(open(new.path).read(), 'utf-8')

class HtmlCtrl(TextAreaCtrl):
    ctrlWidget = FORM_WIDGET('input.html.fck')
    toolbarSet = 'Default'






class Option:

    def __init__(self, key, label, value, selected=False):
        self.key = key
        self.label = label
        self.value = value
        self.selected = selected

    def __eq__(self, other):
        return self.key == other.key \
           and self.label == other.label \
           and self.value == other.value \
           and self.selected == other.selected

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return 'Option(%r,%r,%r,%r)' % (
            self.key,self.label,self.value,self.selected)

class OptGroup(list):

    def __new__(self, label, opts):
        self = list.__new__(opts)
        self.label = label
        return self

class SelectCtrl(ValueCtrl):

    ctrlWidget = FORM_WIDGET('input.select')

    noneLabel = ''

    def defaultOptions(self):
        if not self.ctrlRequired:
            return [('', self.noneLabel, None)]
        return []
    defaultOptions = binding.Make(defaultOptions)

    selectOptions = binding.Require('list of (key,label,value) tuples')

    allOptions = binding.Make(lambda s: [
        (unicode(k), l, v)
        for k,l,v in chain(s.defaultOptions, s.selectOptions)
    ])

    def options(self):
        inp = self.inputText
        options = []
        for k,l,v in self.allOptions:
            options.append(Option(key=k, label=l, value=v, selected=(k==inp)))
        return options
    options = binding.Make(options)

    def flatten(self, opts):
        for o in opts:
            if isinstance(o, OptGroup):
                for x in self.flatten(o):
                    yield o
            yield o

    def renderValue(self, val):
        for k,l,v in self.allOptions:
            if v == val:
                return k
        if not val:
            return ''
        # XXX Should we raise this on unknown old value?
        #raise schema.FeatureSyntaxError(self.ctrlName, value=val)

    def parseValue(self, txt):
        for k,l,v in self.allOptions:
            if unicode(k) == txt:
                return v
        if txt:
            raise schema.FeatureSyntaxError(self.ctrlName, value=txt)



class MultiValueCtrl(ValueCtrl):

    # XXX minLength, maxLength

    def parseValue(self, inp):
        typ = self.ctrlType
        out = []
        for txt in inp:
            try:
                out.append(schema.normalize(txt, typ))
            except:
                raise schema.NormalizationError(
                    self.ctrlName, value=txt, type_name=i18n.typeName(typ)
                )
        return out

    def defaultInput(self):
        return []

    def defaultOutput(self):
        return self.oldValue

    def getInputText(self):
        fid = self.ctrlId
        req = self.ctrlReq
        if self.formAction or fid in req:
            return req.getList(fid)
        else:
            if self.oldValue is NOT_GIVEN:
                return [self.renderValue(inp) for inp in self.defaultInput()]
            else:
                return self.oldValue
    inputText = binding.Make(getInputText)

    def commitSelf(self):
        if not self.ctrlId in self.ctrlReq:
            self.newValue = self.defaultOutput()
        else:
            if not self.inputText and self.ctrlVisibility == 'hidden':
                self.newValue = NOT_GIVEN
            elif not self.inputText and self.emptyIsNone:
                if self.ctrlRequired:
                    self.ctrlError = schema.RequiredError('')
                    self.newValue = NOT_GIVEN
                    return False
                else:
                    self.newValue = []
            else:
                try:
                    self.newValue = self.parseValue(self.inputText)
                except schema.SchemaError, e:
                    self.ctrlError = e
                    self.newValue = NOT_GIVEN
                    OnErrorEvent(self, error=e)
                    return False
        if self.oldValue != self.newValue:
            OnChangeEvent(self, old=self.oldValue, new=self.newValue)
        return True



class MultiSelectCtrl(MultiValueCtrl):

    ctrlWidget = FORM_WIDGET('input.multiselect')
    selectOptions = binding.Require('list of (key,label,value) tuples')

    allOptions = binding.Make(lambda s: [
        (unicode(k), l, v) for k,l,v in s.selectOptions
    ])

    def options(self):
        inp = self.inputText
        options = []
        for k,l,v in self.allOptions:
            options.append(Option(key=k, label=l, value=v, selected=(k in inp)))
        return options
    options = binding.Make(options)

    def flatten(self, opts):
        for o in opts:
            if isinstance(o, OptGroup):
                for x in self.flatten(o):
                    yield o
            yield o

    def parseValue(self, inp):
        typ = self.ctrlType
        out = []
        for txt in inp:
            for k,l,v in self.allOptions:
                if unicode(k) == txt:
                    out.append(v)
                    break
            else:
                raise schema.NormalizationError(
                    self.ctrlName, value=txt, type_name=i18n.typeName(typ)
                )
        return out


class MultiCheckboxCtrl(MultiSelectCtrl):
    ctrlWidget = FORM_WIDGET('input.multicheckbox')


class HiearchicalMultiCheckboxCtrl(MultiSelectCtrl):
    ctrlWidget = FORM_WIDGET('input.hiearchicalmulticheckbox')
    includeRoot = False
    valuesRoot = binding.Require('AbstractTreeNode, root of values tree')
    valuesGetKey = binding.Require('Function returning key from treenode')
    valuesGetLabel = binding.Require('Function returning label from treenode')

    def selectOptions(self):
        return [(self.valuesGetKey(n), self.valuesGetLabel(n), n) for n in self.valuesRoot.iterDepthFirst(include_self=self.includeRoot)]
    selectOptions = binding.Make(selectOptions)



class BoolCtrl(SelectCtrl):
    ctrlWidget = FORM_WIDGET('input.bool')

    ctrlType = bool

    selectOptions = [('yes',_(u'Yes'),True), ('no',_(u'No'),False)]


class EnumCtrl(SelectCtrl):

    def selectOptions(self):
        return [(e.name,e.value,e) for e in self.ctrlType.values()]
    selectOptions = binding.Make(selectOptions)









class EntityReferenceCtrl(ValueCtrl):

    def renderValue(self, val):
        if val is None:
            return u''
        return web.render(self.ctrlReq, val._oid)

    _workspace = binding.Make(lambda: settings.workspace())

    def parseValue(self, txt):
        if not txt:
            return None
        try:
            return self._workspace.get(self.ctrlType, txt)
        except:
            raise schema.FeatureError(
                self.ctrlName,msgId=_('Entity not found. Key=$key.'),key=txt)













class ContainerCtrl(AbstractCtrl):

    protocols.advise(instancesProvide=[IFormInput])

    def commitSelf(self):
        raise NotImplementedError



class InstanceCtrl(ContainerCtrl):

    oldFeatures = newFeatures = binding.Make(dict)

    def ctrlSetupHandlers(self):
        super(InstanceCtrl,self).ctrlSetupHandlers()
        for child in self._ctrlChildren.values():
            if IFormInput(child, None) is None:
                continue
            OnChangeEvent.subscribe(child, self.formCollectChanges)

    def formCollectChanges(self, evt):
        if evt.old is not NOT_GIVEN:
            self.oldFeatures[evt.sender.ctrlName] = evt.old
        if evt.new is not NOT_GIVEN:
            self.newFeatures[evt.sender.ctrlName] = evt.new

    def addEntity(self):
        return self.ctrlType(**self.newFeatures)

    def editEntity(self):
        obj = self.oldValue
        for a in self.newFeatures:
            v = self[a].newValue
            setattr(obj, a, v)
        for a in self.oldFeatures:
            if a not in self.newFeatures:
                delattr(obj, a)
        return obj

    def commitSelf(self):
        if not self.oldFeatures and not self.newFeatures:
            return True
        try:
            if self.oldValue is NOT_GIVEN:
                self.newValue = self.addEntity()
            else:
                self.newValue = self.editEntity()
        except schema.SchemaError, e:
            self.ctrlError = e
            self.newValue = NOT_GIVEN
            OnErrorEvent(self, error=e)
            return False
        OnChangeEvent(self, old=self.oldValue, new=self.newValue)
        return True


class EntityCtrl(InstanceCtrl):

    _workspace = binding.Make(lambda: settings.workspace())

    def addEntity(self):
        obj = self._workspace.new(self.ctrlType, **self.newFeatures)
        self._workspace.flush(obj)
        return obj

    def editEntity(self):
        obj = self.oldValue
        obj._setattrs(**self.newFeatures)
        self._workspace.flush(obj)
        return obj



NEW_RECORD = object()

class SelectOrAddEntityCtrl(EntityCtrl):

    selectOptions = binding.Require()

    class selectOrAdd(SelectCtrl):

        oldValue = binding.Make(lambda s: s.ctrlParent.oldValue)
        selectOptions = binding.Make(lambda s: s.ctrlParent.selectOptions)

        def defaultOptions(self):
            opts = []
            if not self.ctrlParent.ctrlRequired:
                opts.append(('', '', None))
            opts.append(('__new__', _(u'-- new record --'), NEW_RECORD))
            return opts
        defaultOptions = binding.Make(defaultOptions)

    def ctrlSetupHandlers(self):
        super(InstanceCtrl,self).ctrlSetupHandlers()
        for child in self._ctrlChildren.values():
            if child.ctrlName == 'selectOrAdd':
                continue
            if IFormInput(child, None) is None:
                continue
            OnChangeEvent.subscribe(child, self.formCollectChanges)

    def commitSelf(self):
        selVal = self['selectOrAdd'].newValue
        if selVal == NEW_RECORD:
            if not self.oldFeatures and not self.newFeatures:
                return True
            try:
                self.newValue = self.addEntity()
            except schema.SchemaError, e:
                self.ctrlError = e
                self.newValue = NOT_GIVEN
                OnErrorEvent(self, error=e)
                return False
        else:
            self.newValue = selVal
        if self.newValue != self.oldValue:
            OnChangeEvent(self, old=self.oldValue, new=self.newValue)
        return True



class PopupCtrl(AbstractCtrl):

    ctrlWidget = FORM_WIDGET('popup')

    popupMap = {}
    popupUrl = '#'
    popupSize = (400, 200)

    popId = binding.Make(
        lambda s: 'WM4-POPUP-ID-'+ s.ctrlId.replace('$','-').upper()
    )

    popVar = binding.Make(
        lambda s: s.popId.replace('-','_').lower()
    )











class RepeatCtrl(ContainerCtrl):
    ctrlItems = binding.Require()



class SequenceItemCtrl(AbstractCtrl):

    itemState = ClientState(typ=int)

    class removeItem(ButtonCtrl):
        ctrlLabel = binding.Constant(_(u'Remove'))

    class moveUp(ButtonCtrl):
        ctrlLabel = binding.Constant(_(u'Up'))

    @event_handler(OnClickEvent, 'moveUp')
    def onMoveUp(self, evt):
        self['moveOffset'].inputText = '-1'
        self['moveOffset'].newValue = -1
        OnClickEvent(self['moveItem'])

    class moveDown(ButtonCtrl):
        ctrlLabel = binding.Constant(_(u'Down'))

    @event_handler(OnClickEvent, 'moveDown')
    def onMoveDown(self, evt):
        self['moveOffset'].inputText = '1'
        self['moveOffset'].newValue = 1
        OnClickEvent(self['moveItem'])

    class moveItem(ButtonCtrl):
        pass

    class moveOffset(InputCtrl):
        ctrlType = int

    itemPosition = property(lambda s: int(s.ctrlName))

    def prevItem(self):
        pos = self.itemPosition
        if pos > 0:
            return self.ctrlParent[str(pos-1)]
        else:
            return None
    prevItem = property(prevItem)

    def nextItem(self):
        pos = self.itemPosition
        if pos < len(self.ctrlParent.ctrlItems):
            return self.ctrlParent[str(pos+1)]
        else:
            return None
    nextItem = property(nextItem)



class SequenceCtrl(RepeatCtrl):

    ctrlIgnoredFeatures = set(['itemClass'])

    itemCount = ClientState(typ=int)
    itemClass = None
    ctrlType = list

    # XXX TODO
    minLength = binding.Make(lambda s: s.ctrlRequired and 1 or 0)
    maxLength = None

    def _ctrlSetupChildren(self):
        features = dict(
            (e.ctrlName, e) for e in self.ctrlItems
        )
        for e in features.itervalues():
            OnClickEvent.subscribe(e['removeItem'], self.onRemoveItem)
            OnClickEvent.subscribe(e['moveItem'], self._onMoveItem)
        #features.update(items)
        features.update(super(SequenceCtrl,self)._ctrlSetupChildren())
        return features

    def makeItem(self, idx, oldValue=NOT_GIVEN):
        idx = str(idx)
        item = self.itemClass(self.ctrlReq, idx, self, oldValue=oldValue)
        assert isinstance(item, SequenceItemCtrl)
        return item

    def makeDefaultItems(self):
        return []

    def makeCtrlItems(self):
        items = []
        if self.itemCount is NOT_GIVEN:
            container = self.oldValue
            if container is NOT_GIVEN:
                for idx, elem in enumerate(self.makeDefaultItems()):
                    elem.itemState = -1
                    items.append(elem)
                self.itemCount = len(items)
            else:
                for idx, item in enumerate(container):
                    elem = self.makeItem(idx, item)
                    elem.itemState = idx
                    items.append(elem)
                self.itemCount = len(container)
        else:
            container = self.oldValue
            if container is NOT_GIVEN:
                container = []
            eid = self.ctrlId
            for idx in range(self.itemCount):
                # XXX ???
                st = int(self.ctrlReq.getValue(
                    '%s$%d$itemState' % (eid, idx), default='-1'
                ))
                if st == -1:
                    elem = self.makeItem(idx, NOT_GIVEN)
                else:
                    elem = self.makeItem(idx, container[st])
                items.append(elem)
        return items
    ctrlItems = binding.Make(lambda s: s.makeCtrlItems())

    class addItemCount(InputCtrl):
        ctrlType = int
        def defaultInput(self):
            return '1'
        def defaultOutput(self):
            return 1

    class addItem(ButtonCtrl):
        ctrlLabel = binding.Constant(_(u'Add'))

    @event_handler(OnClickEvent, 'addItem')
    def _onAddItem(self, evt):
        self.onAddItem(evt)

    def onAddItem(self, evt):
        try:
            ic = int(self['addItemCount'].inputText)
        except:
            ic = 1
        for i in range(ic):
            elem = self.addNewItem()
            elem.itemState = -1
            self.ctrlItems.append(elem)
            self._ctrlChildren[elem.ctrlName] = elem
            self.itemCount += 1
            OnAddEvent(self, item=elem)

    def addNewItem(self):
        return self.makeItem(self.itemCount)

    def onRemoveItem(self, evt):
        idx = int(evt.sender.ctrlParent.ctrlName)
        self.itemCount -= 1
        items = self.ctrlItems
        OnRemoveEvent(self, item=items[idx])
        del items[idx]
        for i,e in enumerate(items):
            e.itemState
            e.ctrlName = str(i)
        self._ctrlChildren = self._ctrlSetupChildren()

    def _onMoveItem(self, evt):
        idx = int(evt.sender.ctrlParent.ctrlName)
        self.onMoveItem(evt, idx, evt.sender['../moveOffset'].newValue)

    def onMoveItem(self, evt, idx, offset):
        items = self.ctrlItems
        if offset < 0:
            items.insert(idx+offset, items[idx])
            del items[idx+1]
        else:
            item = items[idx]
            del items[idx]
            items.insert(idx+offset, item)
        for i,e in enumerate(items):
            e.itemState
            e.ctrlName = str(i)
        self._ctrlChildren = self._ctrlSetupChildren()

    def updateContainer(self, removed):
        return [i.newValue for i in self.ctrlItems]

    def commitSelf(self):
        old = self.oldValue
        if old is NOT_GIVEN:
            old = []
        old = range(len(old))
        new = [i.itemState for i in self.ctrlItems]
        if old == new:
            return True
        removed = [i for i in old if i not in new]
        try:
            self.newValue = self.updateContainer(removed)
        except schema.SchemaError, e:
            self.ctrlError = e
            self.newValue = NOT_GIVEN
            OnErrorEvent(self, error=e)
            return False
        OnChangeEvent(self, old=self.oldValue, new=self.newValue)
        return True











class FormCtrl(AbstractCtrl):

    class doSubmit(SubmitCtrl):
        pass

    def formAction(self):
        action = None
        for child in self.ctrlIter():
            act = child.ctrlGetAction()
            if act is not None:
                if action is not None:
                    raise web.BadRequest(
                        'More than one action', action.ctrlId, act.ctrlId
                    )
                action = act
        return action
    formAction = binding.Make(formAction)

    def formId(self):
        klass = self.ctrlType
        i = klass.__module__ + '.' + klass.__name__
        return i.upper().replace('.','-').replace('_','-')
    formId = binding.Make(formId)

    formURL = binding.Make(lambda s: s.ctrlReq.requestURL)

    def formErrors(self):
        ee = schema.EntityError()
        for e in self.ctrlIter():
            if e.ctrlError is None:
                continue
            a = (e.ctrlId or '').replace('$', '.')
            if isinstance(e.ctrlError, schema.EntityError):
                ee[a] = schema.NestedEntityError(a, e.ctrlError)
            else:
                ee[a] = e.ctrlError
        return ee
    formErrors = binding.Make(formErrors)

    def formOnInit(self):
        for child in self.ctrlIter():
            child.ctrlOnInit()

    def formSetupHandlers(self):
        for child in self.ctrlIter():
            child.ctrlSetupHandlers()

    def __call__(self):
        self.formAction
        self.formOnInit()
        self.formSetupHandlers()

        # XXX may we can only check that request method is POST
        if self.formAction is not None:
            OnClickEvent(self.formAction)
        else:
            return False

        if self.formAction is self['doSubmit']:
            if not self.commitChanges():
                return False

        if self.formErrors:
            return False
        return self.formAction is self['doSubmit']


def makeFeatureCtrl(entCls, attr):
    desc = entCls.features[attr]
    featClass = FeatureController.of(entCls, attr, default=None)
    typ = desc.typeObj
    if featClass is None:
        # Implicit feature controller for given type
        if featClass is None:
            featClass = FeatureControllerForType.of(typ)
    featClass = importObject(featClass)
    featDict = dict(
        __module__ = entCls.__module__,
        ctrlType = typ,
        ctrlRequired = schema.Required.of(entCls, attr, False),
    )
    if schema.ReadOnly.of(entCls, attr, default=False):
        featDict['ctrlReadOnly'] = True
    if Hidden.of(entCls, attr, default=False):
        featDict['ctrlVisibility'] = 'hidden'
    widget = Widget.of(entCls, attr, NOT_FOUND)
    if widget is not NOT_FOUND:
        featDict['ctrlWidget'] = FORM_WIDGET(widget)

    return type(featClass)(attr, (featClass,), featDict)


def makeFormCtrl(entCls, baseFormCls=EntityCtrl):
    '''Generates Form subclass for given schema.Entity

       Many parameters are resolved from entCls metadata.
    '''
    module = entCls.__module__
    existing = baseFormCls.__ctrl_features__
    addDict = {}
    fakeObj = object.__new__(entCls)

    for attr, desc in entCls.features:
        if attr in existing:
            continue
        if desc.cardinality != 'single':
            continue
        addDict[attr] = makeFeatureCtrl(entCls, attr)

    addDict['__module__'] = module
    addDict['ctrlType'] = entCls
    return type(baseFormCls)(
        entCls.__name__+'Form', (baseFormCls,FormCtrl), addDict
    )
