from wm4.api import *
from lib4.imports import importObject
from lib4.runtime import signature
import re

from publishing import DEFAULT_METHODS, PublishingMiddleware


__all__ = [
    'TemplateHandler',  'with_template', 'render_template', 'expose',
    'Handler', 'XmlHandler', 'ListingHandler', 'EntityHandler', 'BaseHandler',
]

def TemplateHandler(req):
    tmpl = req.getResource(req.params['template'])
    return tmpl.handleHTTP(req)


def with_template(template=None):
    def decorate(ctrl):
        def handler(req):
            ns = req.apply(ctrl)
            if template is None:
                _template = req.params['template']
            else:
                _template = template
            tmpl = req.getResource(_template)
            return tmpl.handleHTTP(req, **ns)
        return handler
    return decorate


def render_template(req, template=None, **ns):
    if template is None:
        template = req.params['template']
    if isinstance(template, basestring):
        template = req.getResource(template)
    return template.handleHTTP(req, **ns)


def expose(template=None):
    def decorate(method):
        sig = signature(method)
        assert not sig[1] and not sig[2] and not sig[3]
        args = sig[0][1:]
        def act(self):
            req = self.req
            av = []
            for a in args:
                if a == 'req':
                    av.append(req)
                else:
                    av.append(getattr(self, a))
            ns = method(self, *av)
            if template is None:
                _template = req.params['template']
            else:
                _template = template
            tmpl = req.getResource(_template)
            return tmpl.handleHTTP(req, **ns)
        return act
    return decorate


class HandlerMeta(type(binding.Component)):
    pass


class Handler(binding.Component):

    __metaclass__ = HandlerMeta

    req = action = binding.Require(suggestParent=False)

    workspace = binding.Make(lambda: settings.workspace(), suggestParent=False)

    @classmethod
    def handle(klass, req):
        action = str(req.params['action'])
        self = klass(req=req, action=action)
        meth = getattr(self, 'do_%s_%s' % (action, req.method), None)
        if meth is None:
            meth = getattr(self, 'do_%s' % action)
            allowed = getattr(meth, 'allowed_methods', DEFAULT_METHODS)
            if req.method not in allowed:
                raise web.MethodNotAllowed(req.method)
        return req.apply(meth)

    def render(self, **kw):
        req = self.req
        path = req.params['template']
        tmpl = req.getResource(path)
        if tmpl is None:
            raise ValueError('cannot find template', path)
        return tmpl.handleHTTP(req, **kw)

    def render_json(self, data, indent=None, content_type='application/json'):
        try:
            import json
        except ImportError:
            import simplejson as json
        output = json.dumps(data, indent=indent)
        resp = self.req.response
        resp.addheader('Content-type', content_type)
        resp.addheader('Content-Length', str(len(output)))
        resp.body = [output]
        return resp




class XmlHandler(Handler):
    def render(self, mime, **kw):
        req = self.req
        tmpl = req.getResource(req.params['template'])
        resp = req.response
        enc = req.output_encoding
        output = tmpl.render(req, **kw).encode(enc)
        resp.addheader('Content-type', '%s;charset=%s' % (
                                    mime, req.output_encoding))
        resp.addheader('Content-Length', str(len(output)))
        resp.body = [output]
        return resp

    def do_view(self, req, mime):
        return self.render(mime)



@PublishingMiddleware.handleHTTP.when('handler in HandlerMeta')
def handle_handler_class(publisher, req, handler):
    return handler.handle(req)



class ListingHandler(Handler):

    entity = binding.Require(suggestParent=False)

    listPagerFactory = 'wm4.web.lists.PagerState'

    listStateFactory  = 'wm4.web.lists.QueryListState'

    def listQuery(self):
        # usually [obj for obj in Ent]
        raise NotImplementedError

    listFilterFactory = None

    def listFilter(self):
        if self.listFilterFactory:
            filter = self.listFilterFactory(ctrlReq=self.req)
            filter()
            return filter
        else:
            return None
    listFilter = binding.Make(listFilter, suggestParent=False)

    def listFilterConditions(self, filter, qry):
        return ()

    listPageLimit = binding.Make(lambda s: int(s.req.params.get('limit',20)))

    def listNS(self, **kw):
        return kw

    def listData(self):
        req = self.req
        ws = self.workspace
        qry = self.listQuery()

        filter = self.listFilter
        if filter is not None:
            cond = self.listFilterConditions(filter, qry)
            if len(cond):
                _cond = []
                for c in cond:
                    if isinstance(c, basestring):
                        c = qry.parseSubexpr(c)
                    _cond.append(c)
                cond = query.AND(*_cond)
                qry = qry.withCondition(ws.connection, cond)

        limit = self.listPageLimit
        count = ~ws(qry.asCount(self.workspace.connection))
        pager = importObject(self.listPagerFactory)(
                    req=req, itemLimit=limit, itemCount=count,
        )
        if pager:
            offset = 0
            if limit:
                offset = pager[pager._current].first
            if limit:
                qry.limit = query.Const(limit)
                qry.offset = query.Const(offset)
            items = tuple(ws(qry))
        else:
            offset = 0
            items = ()

        listing = importObject(self.listStateFactory)(
            req=req, items=items, count=count, offset=offset, query=qry,
        )
        return dict(
            mode='list', obj=self.entity, pager=pager, listing=listing,
            list_filter=filter,
        )

    def do_list(self):
        tns = self.listData()
        return self.template.handleHTTP(self.req, **self.listNS(**tns))

    def template(self):
        return self.req.getResource(self.req.params['template'])
    template = binding.Make(template)






class EntityHandler(ListingHandler):
    '''Provides list, add, view, edit and delete handlers for some schema.Entity
    '''
    entity = binding.Require(suggestParent=False)

    def listQuery(self):
        return self.workspace.parse('[obj for obj in self.entity]')

    def formCtrl(self):
        from wm4.web.forms.controllers import makeFormCtrl
        return makeFormCtrl(self.entity)
    formCtrl = binding.Make(formCtrl, suggestParent=False)

    oidParam = 'oid'

    def getObject(self):
        oid = self.req.params[self.oidParam]
        try:
            obj = self.workspace.get(self.entity, oid)
        except KeyError:
            raise web.NotFound('entity not found', self.entity, oid)
        return obj

    obj = binding.Make(getObject)




    def viewNS(self, **kw):
        return kw

    def viewForm(self):
        return self.formCtrl(
            ctrlReq=self.req, oldValue=self.obj, ctrlReadOnly=True
        )

    def do_view(self):
        form = self.viewForm(); form()
        return self.render(**self.viewNS(form=form, obj=self.obj, mode='view'))





    def addNS(self, **kw):
        return kw

    def addForm(self):
        return self.formCtrl(ctrlReq=self.req)

    def addResult(self, obj):
        raise web.SeeOther(self.req.get_url(obj))

    def do_add(self):
        form = self.addForm()
        if form():
            storage.commit()
            storage.begin()
            return self.addResult(form.newValue)
        else:
            if form.formErrors:
                storage.abort()
                storage.begin()
            return self.render(**self.addNS(form=form, obj=None, mode='add'))





    def editNS(self, **kw):
        return kw

    def editResult(self, obj):
        raise web.SeeOther(self.req.get_url(obj))

    def editForm(self):
        return self.formCtrl(ctrlReq=self.req, oldValue=self.obj)

    def do_edit(self):
        obj = self.obj
        form = self.editForm()
        if form():
            storage.commit()
            storage.begin()
            return self.editResult(form.newValue)
        else:
            if form.formErrors:
                storage.abort()
                storage.begin()
            return self.render(**self.editNS(form=form, obj=obj, mode='edit'))





    def deleteResult(self, obj):
        raise web.SeeOther(self.req.get_url(self.entity))

    def do_delete(self):
        req = self.req
        ws = self.workspace
        back = req.get_url(self.obj)
        try:
            ws.remove(self.obj)
            storage.commit()
            storage.begin()
        except schema.SchemaError, se:
            if isinstance(se, schema.EntityError):
                error = se
            else:
                error = schema.EntityError()
                error.add(se)
            tmpl = req.getResource(req.params['entity_error_template'])
            return tmpl.handleHTTP(req, error=error, back=back)
        else:
            return self.deleteResult(self.obj)




class BaseHandler(Handler):

    def do_list(self, req, clsObj, controller, action):
        lst = importObject(controller)(ctrlReq=req)
        lst()
        return self.render(listing=lst, action=action)

    do_popup = do_list

    def do_view(self, req, clsObj, obj, controller, action):
        form = importObject(controller)(
            ctrlReq=req, oldValue=obj, ctrlReadOnly=True
        )
        form()
        return self.render(
            form=form, clsObj=clsObj, obj=obj, action=action
        )



    def addResult(self, req, obj):
        raise web.SeeOther(req.get_url(obj))

    def do_add(self, req, clsObj, controller, action):
        form = importObject(controller)(ctrlReq=req)
        if form():
            storage.commit()
            storage.begin()
            return self.addResult(req, form.newValue)
        else:
            if form.formErrors:
                storage.abort()
                storage.begin()
            return self.render(
                form=form, clsObj=clsObj, obj=None, action=action
            )




    def editResult(self, req, obj):
        raise web.SeeOther(req.get_url(obj))

    def do_edit(self, req, clsObj, obj, controller, action):
        form = importObject(controller)(ctrlReq=req, oldValue=obj)
        if form():
            storage.commit()
            storage.begin()
            return self.editResult(req, form.newValue)
        else:
            if form.formErrors:
                storage.abort()
                storage.begin()
            return self.render(
                form=form, clsObj=clsObj, obj=obj, action=action
            )



    def deleteResult(self, req, cls):
        raise web.SeeOther(req.get_url(cls))

    def do_delete(self, req, obj, clsObj, entity_error_template):
        ws = self.workspace
        back = req.get_url(obj)
        try:
            ws.remove(obj)
            storage.commit()
            storage.begin()
        except schema.SchemaError, se:
            if isinstance(se, schema.EntityError):
                error = se
            else:
                error = schema.EntityError()
                error.add(se)
            tmpl = req.getResource(entity_error_template)
            return tmpl.handleHTTP(req, error=error, back=back)
        else:
            return self.deleteResult(req, clsObj)
