from itertools import chain

from lib4.iterutils import iterDecorator

from xml4 import xhtml as xh
from xml4.templates import Domlet
from xml4.const import *
from xml4.elements import Text
from xml4.filters import merge_text

from wm4.api import *


__all__ = [
    'HtmlForm', 'HtmlFormField', 'HtmlInput', 'HtmlSelect', 'HtmlHierSelect',
]


FORM_DATA = 'FORM_DATA'

# TODO
# * autofill as json-path that traverses FORM_DATA for getting value
#   a[1].b -> getattr((getattr(form_data, 'a')[1], 'b')
#   a{foo}.b -> getattr((getattr(form_data, 'a')['foo'], 'b')


class HtmlForm(Domlet):

    def expand(self, state, data=NOT_GIVEN):
        if data is not NOT_GIVEN:
            state = state.copy()
            state[FORM_DATA] = data
        return Domlet.expand(self, state)


class HtmlFormField(Domlet):
    
    def get_value(self, state, autofill=None, value=NOT_GIVEN):
        if autofill is None:
            return NOT_FOUND
        if value is not NOT_GIVEN:
            return value
        data = state.get(FORM_DATA, NOT_FOUND)
        if data is not NOT_FOUND:
            if hasattr(data, '__getitem__'):
                value = data[autofill]
            else:
                value = getattr(data, autofill, NOT_FOUND)
            return value
        req = state.get('req')
        if req is not None:
            return req.getValue(autofill, default=NOT_FOUND)
        return NOT_FOUND



class HtmlInput(HtmlFormField):

    def expand(self, state, autofill=None, value=NOT_GIVEN):
        tag = self._tag
        attrs = self._attrs.expand(state)
        if autofill is not None:
            attrs.setdefault('name', autofill)
            value = self.get_value(state, autofill, value)
            if value is NOT_FOUND:
                pass # don't touch original value
            elif value is None:
                attrs.set('value', u'')
            else:
                attrs.set('value', unicode(value))
        yield START_TAG, (tag, attrs)
        yield END_TAG, (tag, attrs)


class HtmlTextarea(HtmlFormField):

    def expand(self, state, autofill=None, value=NOT_GIVEN):
        txt = u''
        for e in self._children:
            if isinstance(e, Text):
                txt += unicode(e)
        tag = self._tag
        attrs = self._attrs.expand(state)
        if autofill is not None:
            attrs.setdefault('name', autofill)
            value = self.get_value(state, autofill, value)
            if value is NOT_FOUND:
                pass # don't touch original value
            elif value is None:
                txt = u''
            else:
                txt = unicode(value)
        yield START_TAG, (tag, attrs)
        yield TEXT, txt
        yield END_TAG, (tag, attrs)


class HtmlInputDate(HtmlFormField):

    def expand(self, state, autofill=None, value=NOT_GIVEN):
        req = state['req']
        for res in (
            'wm4.web.static/jquery/jquery.js',
            'wm4.web.static/jquery/ui.datepicker.js',
            'wm4.web.static/jquery/ui.datepicker-cs.js',
            'wm4.web.static/jquery/ui.datepicker.css',
        ):
            web.getResource(res).render(req)
        tag = self._tag
        attrs = self._attrs.expand(state)
        id = attrs.get('id')
        if id is None:
            id = 'INPUT-%d' % req.get_new_id()
            attrs.set('id', id)
        attrs.setdefault('size', '10')
        if autofill is not None:
            attrs.setdefault('name', autofill)
            value = self.get_value(state, autofill, value)
            if value is NOT_FOUND:
                pass # don't touch original value
            elif value is None:
                attrs.set('value', unicode(''))
            else:
                attrs.set('value', unicode(value))
        yield START_TAG, (tag, attrs)
        yield END_TAG, (tag, attrs)
        for e in xh.script(
            "$(document).ready(function(){$('#%s').datepicker({});});" % id,
            type="text/javascript",
        ).expand(state):
            yield e


class HtmlSelect(HtmlFormField):

    def expand(self, state, autofill=None, value=NOT_GIVEN, options=None):
        tag = self._tag
        attrs = self._attrs.expand(state)
        selected = NOT_FOUND
        if autofill is not None:
            attrs.setdefault('name', autofill)
            selected = self.get_value(state, autofill, value)

        yield START_TAG, (tag, attrs)

        if selected is not NOT_FOUND:
                selected = unicode(selected)

        if options is None:

            if selected is NOT_FOUND:
                for elem in self._children:
                    for e in elem.expand(state):
                        yield e
            else:
                last_opt = None
                iter = merge_text(chain(
                        *(e.expand(state) for e in self._children)
                ))
                for evt, val in iter:
                    if evt is START_TAG and val[0] == 'option':
                        dv = val[1].get('value')
                        if dv is None:
                            last_opt = val
                        else:
                            if dv == selected:
                                val[1].set('selected', 'selected')
                            else:
                                val[1].pop('selected', None)
                            yield evt, val
                            last_opt = None
                    elif evt is TEXT and last_opt is not None:
                        if val == selected:
                            last_opt[1].set('selected', 'selected')
                        else:
                            last_opt[1].pop('selected', None)
                        yield START_TAG, last_opt
                        yield evt, val
                        last_opt = None
                    else:
                        yield evt, val
                        last_opt = None
        else:
            for opt in options:
                if isinstance(opt, (tuple,list)):
                    val = unicode(opt[0])
                    txt = unicode(opt[1])
                else:
                    val = txt = unicode(opt)
                if val == selected:
                    oelem = xh.option(txt, value=val, selected='selected')
                else:
                    oelem = xh.option(txt, value=val)
                for e in oelem.expand(state):
                    yield e

        yield END_TAG, (tag, attrs)




class HtmlHierSelect(HtmlFormField):

    def expand(
        self, state, autofill, roots, value=NOT_GIVEN,
        columns=1, css='treeview-famfamfam', label='---',
        width = 600, height = 400
    ):
        req = state['req']
        attrs = self._attrs.expand(state)
        id = attrs.get('id')
        if id is None:
            id = 'INPUT-%d' % req.get_new_id()

        for e in xh.select(
            xh.option(label, value=''),
            name=autofill,
            id=id,
        ).expand(state):
            yield e

        def branch(children, expanded):
            if expanded:
                ul = xh.ul()
            else:
                ul = xh.ul(style='display:none;')
            for item, node in iterDecorator(children):
                ul.append(render_item(node, item.isLast))
            return ul

        def render_item(node, last=False):
            expanded = False
            lnk = xh.a(node.nazev, href=node.pkey)
            if expanded:
                hit = "hitarea collapsable-hitarea"
                if last:
                    if node.children:
                        li = xh.li(
                            xh.div(_class=hit + " lastCollapsable-hitarea"),
                            lnk, branch(node.children),
                            _class='collapsable lastCollapsable',
                        )
                    else:
                        li = xh.li(lnk, _class='last')
                else:
                    if node.children:
                        li = xh.li(
                            xh.div(_class=hit),
                            lnk, branch(node.children),
                            _class='collapsable',
                        )
                    else:
                        li = xh.li(lnk)
            else:
                hit = "hitarea expandable-hitarea"
                if last:
                    if node.children:
                        li = xh.li(
                            xh.div(_class=hit + " lastExpandable-hitarea"),
                            lnk, branch(node.children, expanded=expanded),
                            _class='expandable lastExpandable',
                        )
                    else:
                        li = xh.li(lnk, _class='last')
                else:
                    if node.children:
                        li = xh.li(
                            xh.div(_class=hit),
                            lnk, branch(node.children, expanded=expanded),
                            _class='expandable',
                        )
                    else:
                        li = xh.li(lnk)
            return li

        tr = xh.tr()
        ll = len(roots)
        cs = ((ll - 1) / columns) + 1
        for i in range(columns):
            ul = xh.ul(_class=css)
            for item, node in iterDecorator(roots[i*cs:i*cs+cs]):
                ul.append(render_item(node, item.isLast))
            if i > 0:
                tr.append(xh.td(
                    ul, style="vertical-align:top; padding-left:30px;"))
            else:
                tr.append(xh.td(
                    ul, style="vertical-align:top;"))
        for e in xh.div(
            xh.div(
                xh.a('smazat hodnotu', id="%s-RESET" % id), 
                style="text-align:right;",
            ),
            xh.table(tr),
            id=id + '-POPUP',
            style="display:none; position:absolute; z-index:9999; border:solid black 1px; padding:5px; overflow:auto; width:%(width)dpx; height:%(height)dpx; background-color:white;" % locals(),
        ).expand(state):
            yield e

        js = xh.script(xh.Literal('''
            $(document).ready(function(){
                $('#%(id)s-POPUP ul.%(css)s').Treeview({
                    prerendered:true,
                    unique:true
                });
                $('#%(id)s-RESET').click(function(e) {
                    var o = $('#%(id)s option');
                    o.val('');
                    o.text('%(label)s');
                    $('#%(id)s-POPUP').hide();
                });
                $('#%(id)s').click(function(e) {
                    e.cancelBubble = true;
                    e.returnValue = false;
                    if (e.stopPropagation) {e.stopPropagation();}
                    if (e.preventDefault) {e.preventDefault();}
                    $(this).blur();
                    var pop = $('#%(id)s-POPUP');
                    var pop_w = %(width)d;
                    var pop_h = %(height)d;
                    var sel = $(this);
                    var sel_off = sel.offset();
                    var left = sel_off.left - ((pop_w - sel.width()) / 2);
                    if(left < 10) {left = 10;}
                    var top = sel_off.top + this.offsetHeight;
                    pop.css({top: top, left: left}).toggle();
                    return false;
                });
                $('#%(id)s-POPUP ul.%(css)s a').click(function(e) {
                    e.cancelBubble = true;
                    e.returnValue = false;
                    if (e.stopPropagation) {e.stopPropagation();}
                    if (e.preventDefault) {e.preventDefault();}
                    var s = $(this);
                    var t = '';
                    s.parents('li').each(function(){
                        var x = $('>a', this).text();
                        if (t) { t = x + ' / ' + t; } 
                        else { t = x; }
                    })
                    var o = $('#%(id)s option');
                    o.val(s.attr('href'));
                    o.text(t);
                    $('#%(id)s-POPUP').hide();
                    return false;
                });
            });''' % locals()),
            type="text/javascript",
        )
        for e in js.expand(state):
            yield e




class HtmlPopup(HtmlFormField):

    def expand(
        self, state, autofill=None, value=NOT_GIVEN, text=NOT_GIVEN,
        width=400, height=300, onshow=None, onhide=None, onset=None
    ):
        req = state['req']
        attrs = self._attrs.expand(state)
        id = attrs.get('id')
        if id is None:
            id = 'INPUT-%d' % req.get_new_id()
        for res in [
            'wm4.web.static/jquery/jquery.js',
            'wm4.web.static/nyroModal/js/jquery.nyroModal.js',
            'wm4.web.static/nyroModal/styles/nyroModal.css',
        ]:
            web.getResource(res).render(req)
        markup = xh.span(
            xh.select(
                xh.option('---', value=""),
                name=autofill,
                id=id,
            ),
            xh.div(
                self.child_expander(),
                id="%s-POPUP" % id,
                style="display:none; width:%dpx; height:%dpx" % (width, height),
            ),
            xh.script('''
                $(document).ready(function() {
                    $('#%(id)s').click(function(e) {
                        e.cancelBubble = true;
                        e.returnValue = false;
                        if (e.stopPropagation) {e.stopPropagation();}
                        if (e.preventDefault) {e.preventDefault();}
                        $(this).blur();
                        $.nyroModalManual({url:'#%(id)s-POPUP'});
                        return false;
                    });
                });
            ''' % {'id':id}),
            type="text/javascript",
        )
        for e in markup.expand(state):
            yield e
