from wm4.api import *
from lib4.imports import importString, importObject
from lib4.cgi import etag_create, http_time
from lib4.timeutils import Datetime
from lib4.content_types import *
from xml4.utils import escape
from xml4 import xhtml as xh
from wm4.utils.files import IFileInfo, IImageInfo
import os, os.path

from interfaces import resource_manager, RES_FACTORY, filenameAsProperty
from errors import MethodNotAllowed, NotModified

__all__ = [
    'Resource', 'ImageResource', 'HeaderResourceContainer', 'HeaderResource',
    'JavascriptResource', 'CSSResource', 'getResource', 'ResourceManager',
    'ResourceMiddleware',
]

class AbstractResource(binding.Component):
    pass


class Resource(AbstractResource):
    '''Base class for all resources. This class implements wrapper
       for static resources.
    '''

    protocols.advise(instancesProvide=[IFileInfo])

    filename = module = modpath = binding.Require()

    size  = binding.Make(lambda s: s.stat.st_size)
    modified = binding.Make(lambda s: Datetime(s.stat.st_mtime))
    content_type = binding.Make(
        lambda s: guessContentType(s.filename, path=s.filename))

    stat  = binding.Make(lambda s: os.stat(s.filename))
    etag  = binding.Make(lambda s: etag_create(s.filename, s.stat))
    lastm = binding.Make(lambda s: http_time(s.modified))

    blocksize = 16384

    def handleHTTP(self, req):
        environ = req.environ
        resp    = req.response
        method  = environ.get('REQUEST_METHOD','GET')

        if method not in ('GET','HEAD'):
            raise MethodNotAllowed(method)

        # X-Sendfile support
        #  SetEnv "HTTP_X_SENDFILE" "X-Sendfile"
        if environ.get('HTTP_X_SENDFILE','').lower() == 'x-sendfile':
            resp.addheader('Content-Type', self.content_type)
            resp.addheader('X-Sendfile', self.filename)
            resp.body = ()
            return resp

        # http://www.xml.com/pub/a/2006/02/01/doing-http-caching-right-introducing-httplib2.html
        # http://svn.zope.org/z3c.conditionalviews/trunk/
        #
        # Modern browsers send informations about cached files:
        # HTTP_IF_MODIFIED_SINCE = Sun, 21 Nov 2004 14:35:21 GMT
        # HTTP_IF_MODIFIED_SINCE = Wed, 10 May 2006 06:00:00 GMT; length=10862
        # HTTP_IF_NONE_MATCH     = "17156-3b0-438b26aa"
        # XXX compare as datetime not as strings
        try:
            lastm = Datetime(environ.get('HTTP_IF_MODIFIED_SINCE',self))
        except:
            lastm = NOT_FOUND
        etag = environ.get('HTTP_IF_NONE_MATCH')

        resp.addheader('Content-Length', str(self.size))
        resp.addheader('Content-Type', self.content_type)
        resp.addheader('ETag', self.etag)
        resp.addheader('Last-Modified', self.lastm)

        if lastm==self.lastm and etag==self.etag:
            resp.status = NotModified.status
            resp.body = ()
            return resp

        if method == 'HEAD':
            resp.body = ()
        else:
            stream = open(self.filename, 'rb')
            def file_iter():
                size = self.blocksize
                while 1:
                    data = stream.read(size)
                    if not data: break
                    yield data
                stream.close()
            resp.body = file_iter()
        return resp




class RenderableResource(Resource):

    def base_url(self):
        base_url=config.Property(
                'wm4.web.resources.base_url.%s' % filenameAsProperty(
                    self.filename.split(os.sep)[-1]))()
        return base_url
    base_url = binding.Make(base_url)

    def render(self, req):
        raise NotImplementedError

    def url(self):
        return '%s/%s/%s' % (self.base_url, self.module, self.modpath)
    url = binding.Make(url)


class ImageResource(RenderableResource):

    protocols.advise(instancesProvide=[IImageInfo])

    def width_height(self):
        from lib4.images import loadImage
        width, height = loadImage(self.filename).size
        return locals()
    width = height = binding.MultiMake(width_height)

    def render(self, req):
        return xh.img(src=self.url, width=self.width, height=self.height)





class HeaderResourceContainer(object):
    # http://cheeseshop.python.org/pypi/zif.headincludes/

    def __init__(self):
        self.resource_set = set()
        self.resources = []

    def add(self, res):
        if res not in self.resource_set:
            self.resource_set.add(res)
            self.resources.append(res)

    def __iter__(self):
        # CSS first
        for res in self.resources:
            if not isinstance(res, CSSResource): continue
            yield res
        # others
        for res in self.resources:
            if isinstance(res, CSSResource): continue
            yield res


class HeaderResource(RenderableResource):

    def render(self, req):
        req.params[HeaderResourceContainer].add(self)

    def include(self, req):
        raise NotImplementedError


class JavascriptResource(HeaderResource):

    def include(self, req):
        return u'<script type="text/javascript" src="%s"></script>\n' % self.url

class CSSResource(HeaderResource):

    def include(self, req):
        return (u'<link href="%s" rel="stylesheet" media="all" '
                       'type="text/css" />\n') % self.url






def findPackage(pkg):
    """Find containing package (module w/'__path__') for module named 'pkg'"""
    mod = importString(pkg)
    while not hasattr(mod,'__path__') and '.' in pkg:
        pkg = '.'.join(pkg.split('.')[:-1])
        mod = importString(pkg)
    return pkg, mod


def getResource(path):
    return resource_manager().getResource(path)





class ResourceManager(binding.Component):

    cache = binding.Make(dict)

    def getResource(self, name, default=None):
        # cache with limited size
        cache = self.cache
        if len(cache) > 300:
            cache.clear()
        res = cache.get(name)

        # negative hit cached
        if res is NOT_FOUND:
            return default
        elif res:
            return res

        module, modpath = name.split('/',1)
        try:
            pkg, mod = findPackage(module)
        except ImportError:
            cache[name] = NOT_FOUND
            return default

        dirname = os.path.dirname(mod.__file__)
        filename = os.path.join(dirname, *(modpath.split('/')))

        if not os.path.isfile(filename):
            cache[name] = NOT_FOUND
            return default

        fact = RES_FACTORY(filename.split(os.sep)[-1])(default=None)
        if fact is None:
            cache[name] = NOT_FOUND
            return default

        res = cache[name] = fact(
            self, name, module=module, modpath=modpath, filename=filename
        )
        return res



class ResourceMiddleware(binding.Component):

    def __init__(self, application, **params):
        self.application = application
        self.__dict__.update(params)

    def __call__(self, environ, start_response):
        #environ['windmill.resource_middleware'] = self
        # XXX
        # path_info = environ.get('PATH_INFO')
        # if path_info.startswith(self.prefix):
        #     if not <is allowed>:
        #         raise NotAllowed()
        #     req = web.getRequest(environ)
        #     self.getResource(....).handleHTTP()
        return self.application(environ, start_response)
