'''Implements traversal mechanism'''

from wm4.api import *
from dispatch import functions, strategy, predicates, ast_builder
from new import instancemethod
from lib4.urls import urlPathJoin
from lib4.symbols import NOT_FOUND
from lib4.runtime import CachingLazyProxy
import sys, re, operator


__all__ = [
    'SitemapError', 'SitemapSyntaxError', 'RoutePattern',
    'ConstraintCheckingGenericFunction', 'Signature', 'MultiSignature',
    'Acquire', 'RouteParam', 'Param', 'Const', 'Lazy',
    'RouteMeta', 'Route', 'SitemapMiddleware',
]

class SitemapError(Exception):
    pass

class SitemapSyntaxError(Exception):
    pass



NORMALIZE_URL = re.compile(r'//+').sub
SLASH_SPLIT = re.compile(r'(/+)').split
ROUTE_SPLIT = re.compile(r'((?<!\\)(?::|\*)\(?[a-z_][a-z0-9_]*\)?)',re.I).split

LEXEM_DELIM = 'delim'
LEXEM_STATIC = 'static'
LEXEM_SLASH = 'slash'
LEXEM_PARAM = 'param'


class RoutePattern(object):

    def __init__(self, pattern_list):

        def lexer():
            yield LEXEM_DELIM, 0
            for i, pat in enumerate(pattern_list):
                if i > 0:
                    yield LEXEM_DELIM, i
                if not pat:
                    yield LEXEM_STATIC, pat
                else:
                    for j, p in enumerate(ROUTE_SPLIT(pat)):
                        if j % 2 == 0:
                            if p:
                                p = p.replace('\\:', ':')
                                for t in SLASH_SPLIT(p):
                                    if t:
                                        if t.startswith('/'):
                                            yield LEXEM_SLASH, t
                                        else:
                                            yield LEXEM_STATIC, t
                        else:
                            t = p[0]
                            if p[1] == '(' or p[-1] == ')':
                                assert p[1] == '(' and p[-1] == ')'
                                yield LEXEM_PARAM, (t, p[2:-1], i)
                            else:
                                yield LEXEM_PARAM, (t, p[1:], i)

        variables = self.variables = set()
        fmt = []
        regex = ['^']
        hcmp = [[]]

        for kind, arg in lexer():
            if kind is LEXEM_DELIM:
                if arg > 0:
                    regex.append(')')
                regex.append('(?P<r%d>' % arg)

            elif kind is LEXEM_STATIC:
                fmt.append(arg.replace('%','%%'))
                regex.append(re.escape(arg))
                htop = hcmp[-1].append
                htop(0)
                htop(arg)

            elif kind is LEXEM_SLASH:
                fmt.append('/')
                regex.append('/')
                hcmp[-1].append(9)
                hcmp.append([])

            elif kind is LEXEM_PARAM:
                wild, param, part = arg
                var = 'p%d_%s' % (part, param)
                fmt.append('%%(%s)s' % var)
                htop = hcmp[-1].append
                if wild == ':':
                    regex.append('(?P<%s>[^/]+)' % var)
                    htop(1)
                    htop(param)
                elif wild == '*':
                    regex.append('(?P<%s>.*)' % var)
                    htop(2)
                    htop(param)
                else:
                    raise SitemapSyntaxError(
                        'Unknown wildcard', wild, pattern_list)
                if var in variables:
                    raise SitemapSyntaxError(
                        'More than one parameter named', var, pattern_list)
                variables.add(var)

        regex.append(')$')
        hcmp[-1].append(9)

        self.fmt = ''.join(fmt)
        self.regex = ''.join(regex)
        self.match = re.compile(self.regex).match
        self.hash_cmp = tuple(tuple(p) for p in hcmp)
        self.pattern_list = pattern_list

    def __cmp__(self, other):
        return cmp(self.hash_cmp, other.hash_cmp)

    def __repr__(self):
        return 'RoutePattern(%s)' % repr(''.join(self.pattern_list))

    def __str__(self):
        return ''.join(self.pattern_list)





def skip_throwing(funcs):
    def first_ok(*a):
        i = iter(funcs)
        for f in i:
            try:
                v = f(*a)
            except SitemapError:
                continue
            else:
                # the rest must raise SitemapError
                for g in i:
                    try:
                        w = g(*a)
                    except SitemapError:
                        pass
                    else:
                        raise dispatch.AmbiguousMethod(
                            'ARGUMENTS', a, 'SIGNATURES', f, g, 'RESULTS', v, w
                        )
                        #raise dispatch.AmbiguousMethod(a, f, g, v.partition, w.partition)
                return v
    return first_ok


class ConstraintCheckingGenericFunction(functions.GenericFunction):

    def combine(self,cases):
        for group in strategy.ordered_signatures(cases):
            if len(group) == 1:
                return group[0][1]
            else:
                funcs = [f for s,f in group]
                return skip_throwing(funcs)
        else:
            return dispatch.NoApplicableMethods()



class MultiSignature:

    inherit = False

    def __init__(self, sigs):
        self.sigs = sigs

    def __or__(self, other):
        if isinstance(other, MultiSignature):
            return MultiSignature(self.sigs + other.sigs)
        elif isinstance(other, Signature):
            return MultiSignature(self.sigs+(other,))
        else:
            return TypeError('invalid operand', other)

    def __iter__(self):
        return iter(self.sigs)

    def __repr__(self):
        return 'MultiSignature(%s)' % '|'.join(repr(s) for s in self.sigs)

    def predicates(self, route):
        for sig in self.sigs:
            for p in sig.predicates(route):
                yield p

class Signature:

    def __init__(self, *args, **kw):
        self.args = args
        self.when = kw.get('when')
        self.inherit = kw.get('inherit', False)

    def __or__(self, other):
        if isinstance(other, Signature):
            return MultiSignature((self, other))
        else:
            return TypeError('invalid operand', other)

    def __iter__(self):
        yield self

    def __repr__(self):
        if self.when:
            return 'Signature(%s,when=%s)' % (','.join(self.args), self.when)
        else:
            return 'Signature(%s)' % ','.join(self.args)

    def predicates(self, route):
        # ``alen == N''
        predicate = strategy.Signature([
          (strategy.Argument(pos=0), strategy.Inequality('=',len(self.args)))
        ])
        namespace = {}
        for i, a in enumerate(self.args):
            # ``argv[i] in param.typ''
            param = getattr(route, a)
            if isinstance(param, Acquire):
                param = param.get(route)
            arg = predicates.Call(
              operator.getitem, strategy.Argument(pos=1), predicates.Const(i)
            )
            namespace[a] = arg
            predicate = predicate & strategy.Signature([
                    (arg, strategy.ClassCriterion(param.typ))
            ])
            if isinstance(param, Const):
                # ``argv[i] == param.val
                predicate = predicate & strategy.Signature([
                    (arg, strategy.Inequality('=', param.value))
                ])
        if self.when:
            # when
            builder = predicates.CriteriaBuilder(
                namespace, sys.modules[route.__module__].__dict__
            )
            predicate = predicate & ast_builder.parse_expr(self.when, builder)
        return (self, predicate),







class AbstractParam(object):
    depth = 0
    attr  = None




class Acquire(AbstractParam):
    '''Proxy descriptor that gets/sets/deletes attribute on parent route
    '''

    PARENT_ATTR_RE = re.compile(r'^(\.\./)*[a-zA-Z0-9_]+$').match

    def __init__(self, spec):
        if isinstance(spec, tuple):
            self.attr, self.depth = spec
        else:
            if not self.PARENT_ATTR_RE(spec):
                raise ValueError('Invalid path', spec)
            x = spec.split('/')
            self.attr, self.depth = x[-1], len(x)-1
        self.range = range(self.depth)

    def __get__(self, obj, typ):
        attr = self.attr
        depth = self.depth
        if obj is None:
            return self
        else:
            for i in self.range:
                obj = obj.parentRoute
            return getattr(obj, attr)

    def get(self, typ):
        typ = typ.__all_parents__[-(self.depth+1)]
        return getattr(typ, self.attr)

    def __set__(self, obj, value):
        for i in self.range:
            obj = obj.parentRoute
        setattr(obj, self.attr, value)

    def __delete__(self, obj):
        for i in self.range:
            obj = obj.parentRoute
        delattr(obj, self.attr)

    def __repr__(self):
        return 'Acquire(%r,%r)' % (self.attr, self.depth)








NotFound = object()
Lock = object()



class RouteParam(AbstractParam):

    protocols.advise(instancesProvide=[binding.IActiveDescriptor])

    depth = 0
    inherit = True
    typ = unicode
    attr = None
    recipe = None
    equalTo = None
    acquire = None
    virtual = False

    def __init__(
        self, recipe=None, typ=object, inherit=True, equalTo=None, virtual=False
    ):
        self.recipe = recipe
        self.typ = typ
        self.inherit = inherit
        if equalTo is not None:
            self.equalTo = equalTo
            self.acquire = Acquire(equalTo)
        self.virtual = virtual

    def activateInClass(self, klass, attr):
        assert self.attr is None, (klass, self.attr, attr)
        copy = object.__new__(self.__class__)
        copy.__dict__.update(self.__dict__)
        copy.attr = attr
        setattr(klass, attr, copy)
        return copy

    def copy(self):
        cp = object.__new__(self.__class__)
        cp.__dict__.update(self.__dict__)
        cp.attr = None
        return cp

    def __get__(self, obj, typ):
        if obj is None:
            return self
        attr = self.attr
        odct = obj.__dict__
        val = odct.get(attr, NotFound)
        if val is NotFound:
            if self.recipe:
                odct[attr] = Lock
                val = self.recipe(obj)
                odct[attr] = val
                if self.acquire:
                    self.acquire.__set__(obj, val)
                return val
            else:
                raise AttributeError(obj, attr)
        elif val is Lock:
            if self.acquire is None:
                raise RuntimeError('Recursive attempt', obj, attr)
            else:
                odct[attr] = aval = self.acquire.__get__(obj, typ)
                rval = self.recipe(obj)
                if aval != rval:
                    raise SitemapError()
                return aval
        else:
            return val

    def __set__(self, obj, val):
        if not isinstance(val, self.typ):
            try:
                val = schema.normalize(val, self.typ)
            except:
                raise TypeError('Cast error', obj, val, self.typ)
        attr = self.attr
        odct = obj.__dict__
        old = odct.get(attr, NotFound)
        if old is not NotFound and val != old:
            raise SitemapError('Value conflict', obj.__class__,attr,old,val)
        odct[attr] = val
        if self.acquire:
            self.acquire.__set__(obj, val)

    def __delete__(self, obj):
        raise AttributeError(
            'Attribute deletion not supported', obj.__class__, self.attr
        )


class Param(RouteParam):

    def __repr__(self):
        return 'Param(%r)' % self.attr



class Const(RouteParam):

    def __init__(self, value, **kw):
        self.value = value
        kw['recipe'] = lambda obj: value
        kw.setdefault('typ', value.__class__)
        super(Const, self).__init__(**kw)

    def __set__(self, obj, val):
        if val != self.value:
            raise SitemapError(
                'Value conflict', obj.__class__, self.attr, self.value, val)

    def __repr__(self):
        return 'Const(%s)' % repr(self.value)


class Lazy(Param):

    def __init__(
        self, recipe=None, typ=object, inherit=True, equalTo=None, virtual=False
    ):
        def lazy(obj):
            return CachingLazyProxy(recipe, obj)
        super(Lazy, self).__init__(lazy, typ, inherit, equalTo, virtual)









class RouteMeta(binding.ActiveClass):

    __metaclass__ = binding.Activator

    __constraints__ = binding.Make(set)

    def __all_constraints__(klass):
        ac = set()
        for base in klass.__mro__:
            if isinstance(base, RouteMeta):
                ac = ac | base.__constraints__
        return ac
    __all_constraints__ = binding.Make(__all_constraints__)

    def __all_parents__(klass):
        parent = klass.__parent__[0]
        if isinstance(parent, RouteMeta):
            return parent.__all_parents__ + (klass,)
        else:
            return (klass,)
    __all_parents__ = binding.Make(__all_parents__)

    __root__ = binding.Make(lambda c: c.__all_parents__[0])

    __level__ = binding.Make(lambda c: len(c.__all_parents__)-1)

    def _copy_with_parent(klass, parent, name):
        cdict = {}
        cdict['__parent__'] = parent,
        cdict['__module__'] = klass.__module__
        for a,d in parent.__params__.iteritems():
            if not d.inherit:
                continue
            if not hasattr(klass, a):
                if d.virtual:
                    cdict[a] = d.copy()
                else:
                    cdict[a] = Acquire((a, 1))
        for a,d in parent.__acquisitions__.iteritems():
            if not hasattr(klass, a) and a not in cdict:
                cdict[a] = Acquire((d.attr, d.depth+1))
        return type(klass)(name, (klass,), cdict)

    def __cpau(klass):
        __child_routes__ = {}
        __params__ = {}
        __acquisitions__ = {}
        __url_for__ = []
        # we need scan also metaclass because that can have classAttr's
        # and these must be computed now
        for attr in [a for a in dir(type(klass))] + dir(klass):
            if attr.startswith('__'):
                continue
            desc = getattr(klass, attr)
            if isinstance(desc, RouteMeta):
                __child_routes__[attr] = desc
            elif isinstance(desc, RouteParam):
                if desc.attr is None:
                    # activate param that was computed via classAttr
                    desc = desc.activateInClass(klass, attr)
                __params__[attr] = desc
            elif isinstance(desc, Acquire):
                __acquisitions__[attr] = desc
            elif hasattr(desc, 'url_for'):
                __url_for__.append(desc)
        return locals()

    __child_routes__ = \
    __params__ = \
    __acquisitions__ = \
    __url_for__ = binding.MultiMake(__cpau)

    def __children__(klass):
        return dict(
            (a, r._copy_with_parent(klass, a))
            for a,r in klass.__child_routes__.iteritems()
        )
    __children__ = binding.Make(__children__)

    def _child_route(klass, path):
        cls = klass
        for p in path.split('/'):
            cls = cls.__children__[p]
        return cls

    def __pattern__(klass):
        pats = [p.match for p in klass.__all_parents__]
        if not pats[0].startswith('/'):
            pats[0] = '/'+pats[0]
        return RoutePattern(pats)
    __pattern__ = binding.Make(__pattern__)

    def _iter(klass):
        yield klass
        for c in klass.__children__.itervalues():
            for x in c._iter():
                yield x

    def __traversal_table__(klass):
        tt = [
            (r.__pattern__, -len(r.__all_parents__), r)
            for r in klass._iter() if not r.virtual
        ]
        tt.sort()
        return [(p,r) for p,_,r in tt]
    __traversal_table__ = binding.Make(__traversal_table__)

    def fromGroupDict(klass, gdct, pathInfo):
        def make_path(prev, idx):
            key = 'r%d' % idx
            gdct[key] = new = prev + gdct[key]
            return new
        reduce(make_path, range(len(klass.__all_parents__)), '')
        obj = klass({}, gdct)
        return obj

    def traverse(klass, pathInfo):
        if not pathInfo.startswith('/'):
            pathInfo = '/' + pathInfo
        for m, r in klass.__traversal_table__:
            x = m.match(pathInfo)
            if x:
                return r.fromGroupDict(x.groupdict(), pathInfo)
        return None

    def __signatures__(klass):
        if klass.signature is None:
            return []
        return [p for s in klass.signature for p in s.predicates(klass)]
    __signatures__ = binding.Make(__signatures__)

    def __route_for__(klass):

        def route_for((routeCls, sig), ac, av):
            attrs = dict(zip(sig.args, av))
            return routeCls(attrs)

        no_route = lambda ac, av: None
        func = ConstraintCheckingGenericFunction(no_route)
        func.addMethod(strategy.default, no_route)
        seen = set()

        def add_slash(gen):
            def wrapped(al, av):
                u = gen(al, av)
                if not u.startswith('/'):
                    u = '/' + u
                return u
            return wrapped

        for route in klass._iter():

            for sig, pred in route.__signatures__:
                func.addMethod(
                    pred, instancemethod(route_for, (route,sig), tuple)
                )

            for urlf in route.__url_for__:
                if urlf in seen:
                    continue
                seen.add(urlf)
                for pred, gen in urlf.url_for:
                    gen = add_slash(gen)
                    f = lambda al,av, gen=gen: Route(dict(pathInfo=gen(al, av)))
                    func.addMethod(pred, f)

        return func
    __route_for__ = binding.Make(__route_for__)

    def route_for(klass, *args):
        obj = klass.__route_for__(len(args), args)
        return obj

    def get_url(klass, *args):
        r = klass.__route_for__(len(args), args)
        if r is None:
            return None
        return r.pathInfo


class Route(object):

    __metaclass__ = RouteMeta

    match = ''
    signature = None
    virtual = False

    def __init__(self, params={}, groupDict={}, check=True):
        klass = self.__class__
        # force param activation
        klass.__params__
        level = klass.__level__

        # set explicit params
        for a,v in params.items():
            d = getattr(klass, a, NOT_FOUND)
            if d is NOT_FOUND:
                raise AttributeError('Route %r has no attribute %r' % (klass,a))
            if getattr(d, 'depth', 0) == 0:
                params.pop(a)
                setattr(self, a, v)

        # set explicit group params
        pp = 'p%d_' % level
        rp = 'r%d' % level
        for a,v in groupDict.items():
            if a == rp:
                self.pathInfo = v
                groupDict.pop(a)
            if a.startswith(pp):
                p = a[len(pp):]
                d = getattr(klass, p, NOT_FOUND)
                if d is NOT_FOUND:
                    raise AttributeError(
                        'Route %r has no attribute %r' % (klass,p))
                groupDict.pop(a)
                setattr(self, p, v)

        # create immediately parent route
        try:
            pcls = klass.__all_parents__[-2]
        except IndexError:
            pcls = None
        if pcls:
            self.parentRoute = pcls(params, groupDict, False)
        else:
            self.parentRoute = None

        if check:
            self.get_params(False)

    def get_params(self, delayed=True):
        klass = self.__class__
        params = {}
        obj = self
        while obj:
            cls = obj.__class__
            for attr in cls.__params__:
                desc = getattr(cls, attr)
                val = getattr(obj, attr)
                if obj is self:
                    params[attr] = val
            obj = obj.parentRoute

        asort = klass.__acquisitions__.items()
        asort.sort(key=lambda x:x[1].depth)

        for a,d in asort:
            params[a] = v = getattr(self,a)

        return params

    def pathInfo(self):
        klass = self.__class__
        pattern = klass.__pattern__
        plen = len(klass.__all_parents__)
        dct = {}
        for p in pattern.variables:
            l, a = p.split('_')
            l = int(l[1:])
            obj = self
            for x in xrange(plen - l - 1):
                obj = obj.parentRoute
            dct[p] = getattr(obj, a)
        return pattern.fmt % dct
    pathInfo = binding.Make(pathInfo)

    def __div__(self, path):
        pathInfo = urlPathJoin(self.pathInfo, path)
        return self.__class__.__root__.traverse(pathInfo)

    def __repr__(self):
        id = object.__repr__(self).split()[3][:-1]
        return '<%s route at %s>' % (
            '.'.join(c.__name__ for c in self.__class__.__all_parents__), id
        )







class SitemapMiddleware:

    sitemap = None

    def __init__(self, application, **params):
        self.application = application
        self.__dict__.update(params)

    def __call__(self, environ, start_response):
        from errors import NotFound

        pathInfo = environ['PATH_INFO']
        while pathInfo.startswith('//'):
            pathInfo = pathInfo[1:]
        try:
            route = self.sitemap.traverse(pathInfo)
        except SitemapError:
            raise NotFound(pathInfo)
        if route is None:
            raise NotFound(pathInfo)
        environ['windmill.sitemap'] = self.sitemap
        environ['windmill.route'] = route
        return self.application(environ, start_response)
