# http://svn.colorstudy.com/home/ianb/scarecrow_session_interface.py
# From: Ian Bicking <ianb@colorstudy.com>
# Subject: [Web-SIG] Session interface, v2
# Date: Wed, 17 Aug 2005 22:21:41 -0500
# XXX: locking
# XXX: hmac?

from wm4.api import *
import time, struct, cPickle, os, random, sys
try:
    from hashlib import sha1
except ImportError:
    from sha import new as sha1
from lib4.imports import importObject, importString

from interfaces import app_secret
from cookies import getCookie, setCookie


__all__ = [
    'TrustedClientDataError', 'TrustedClientData',
    'SecureValue', 'SessionID', 'Session', 'getSession', 'SessionMiddleware',
]


class TrustedClientDataError(ValueError):
    '''Raised when unpack TrustedClientData that cannot be verified'''


class TrustedClientData:
    '''Signed data that client cannot change

       Instances are used for generating a verifying sensitive data. Data
       are digitaly signed so server has certainity that client hasn't
       changed values. Data are not encrypted.
    '''

    attributes = ()
    digest_attributes = ()

    def __init__(self, **data):
        self.__dict__.update(data)

    def pack(self):
        buf = []; app = buf.append
        for a, p in self.attributes:
            v = getattr(self, a)
            if p == 'S':
                app(struct.pack('B', len(v)))
                app(v)
            elif p == 'U':
                v = v.encode('utf-8')
                app(struct.pack('B', len(v)))
                app(v)
            else:
                app(struct.pack(p, v))
        pdata = ''.join(buf)
        dattr = cPickle.dumps(tuple(
            getattr(self,a,None) for a in self.digest_attributes
        ))
        m = sha1()
        m.update(pdata)
        m.update(dattr)
        pdata += m.digest()
        return pdata.encode('base64').translate(B64TRANS).rstrip()

    def __str__(self):
        return self.pack()

    @classmethod
    def unpack(klass, packed, **digest):
        packed = str(packed).translate(B64UNTRANS).decode('base64')
        ptr = 0
        data = {}
        for a, p in klass.attributes:
            if p == 'S':
                s = ord(packed[ptr])
                ptr += 1
                data[a] = packed[ptr:ptr+s]
                ptr += s
            elif p == 'U':
                s = ord(packed[ptr])
                ptr += 1
                data[a] = unicode(packed[ptr:ptr+s], 'utf-8')
                ptr += s
            else:
                s = struct.calcsize(p)
                data[a] = struct.unpack(p, packed[ptr:ptr+s])[0]
                ptr += s
        dattr = cPickle.dumps(tuple(
            digest.get(a,None) for a in klass.digest_attributes
        ))
        m = sha1()
        m.update(packed[:ptr])
        m.update(dattr)
        if m.digest() != packed[ptr:]:
            raise TrustedClientDataError('data digest verification failed')
        obj = klass(**data)
        obj.__dict__.update(digest)
        return obj

B64TRANS = [chr(_x) for _x in range(256)]
B64TRANS[ord('+')] = '-'
B64TRANS[ord('/')] = '_'
B64TRANS = ''.join(B64TRANS)

B64UNTRANS = [chr(_x) for _x in range(256)]
B64UNTRANS[ord('-')] = '+'
B64UNTRANS[ord('_')] = '/'
B64UNTRANS = ''.join(B64UNTRANS)






class SecureValue(TrustedClientData):
    attributes = [('value','U')]
    digest_attributes = ('secret',)


class SessionID(TrustedClientData):
    '''Instance that holds trusted session ID

       ID depends on remote address and browser.
    '''
    attributes = [('timestamp','d'), ('rnd', 'H')]
    digest_attributes = ('secret', 'remote_addr', 'browser')




class Session:

    def __init__(self, id, ctime, atime, data=None):
        self.id = id
        self.ctime = ctime
        self.atime = atime
        if data is None:
            self.data = {}
        else:
            self.data = data

    def get(self, key, default=None):
        return self.data.get(key, default)

    def pop(self, key, *args):
        return self.data.pop(key, *args)

    def items(self):
        return self.data.items()

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, value):
        self.data[key] = value

    def __delitem__(self, key):
        del self.data[key]

    def commit(self):
        raise NotImplementedError

    def rollback(self):
        raise NotImplementedError

    def __repr__(self):
        return 'Session(%s, %s, %s, %s)' % (
            self.id, self.ctime, self.atime, self.data)





# See http://www.infrae.com/products/tramline

class UploadManager(binding.Component):

    if sys.platform == 'win32':
        upload_dir = 'C:\\var\\tmp\\wm4.upload'
    else:
        upload_dir = '/var/tmp/wm4.upload.%(uid)s'

    def _upload_dir(self):
        d = self.upload_dir % {'uid':os.getuid()}
        if not os.path.isdir(d):
            os.makedirs(d)
            os.chmod(d, 0700)
        return d
    _upload_dir = binding.Make(_upload_dir)

    _mktemp = binding.Make(lambda: importString('tempfile').mktemp)

    def mktemp(self, suffix=''):
        return self._mktemp(suffix=suffix, dir=self._upload_dir)





ENVKEY = 'windmill.session_middleware'
SESSKEY = 'windmill.session'

def getSession(environ):
    return environ[ENVKEY].getSession(environ)



class SessionMiddleware(binding.Component):
    '''Session component that stores session in temporary files'''

    session_id_factory = SessionID
    session_factory = Session

    _fact = binding.Make(lambda s: importObject(s.session_factory))
    _id_fact = binding.Make(lambda s: importObject(s.session_id_factory))

    if sys.platform == 'win32':
        session_dir = 'C:\\var\\tmp\\wm4.session'
    else:
        session_dir = '/var/tmp/wm4.session.%(uid)s'
    secret = None
    cookie_name = 'wm4sess'
    secret = binding.Make(lambda: app_secret())

    getCookie = staticmethod(getCookie)
    setCookie = staticmethod(setCookie)
    time      = staticmethod(time.time)

    lazy_cookie = True

    def __init__(self, application, **params):
        self.application = application
        self.__dict__.update(params)

    def __call__(self, environ, start_response):
        environ[ENVKEY] = self

        if not self.lazy_cookie:

            id = self.getCookie(environ, self.cookie_name)
            if id is not None:
                try:
                    id = self.verifyID(environ, id)
                except:
                    id = None
            if id is None:
                newID = self.createID(environ)
                path = environ.get('SCRIPT_NAME') or '/'
                self.setCookie(environ, self.cookie_name, str(newID), path=path)

        try:
            body = self.application(environ, start_response)
        finally:
            sess = environ.get(SESSKEY, NOT_FOUND)
            if sess is not NOT_FOUND:
                self.saveSession(environ, sess)
        return body

    def _getDir(self):
        d = self.session_dir % {'uid':os.getuid()}
        if not os.path.isdir(d):
            os.makedirs(d)
            os.chmod(d, 0700)
        return d

    def _sessionFile(self, id):
        return os.path.join(self._getDir(), str(id))

    def createID(self, environ):
        return self._id_fact(
            timestamp = self.time(),
            rnd = random.randint(0, 65535),
            secret = self.secret,
            remote_addr = environ.get('REMOTE_ADDR'),
            browser = environ.get('HTTP_USER_AGENT'),
        )

    def verifyID(self, environ, id):
        if isinstance(id, basestring):
            return self._id_fact.unpack(
                id,
                secret = self.secret,
                remote_addr = environ.get('REMOTE_ADDR'),
                browser = environ.get('HTTP_USER_AGENT'),
            )
        elif isinstance(id, self._id_fact):
            return id
        else:
            raise ValueError('invalid session id type', id)

    def getSession(self, environ):
        sess = environ.get(SESSKEY, NOT_FOUND)
        if sess is not NOT_FOUND:
            return sess
        id = self.getCookie(environ, self.cookie_name)
        if id is not None:
            try:
                id = self.verifyID(environ, id)
            except:
                id = None
        if id is None:
            sess = self.createSession(environ)
            path = environ.get('SCRIPT_NAME') or '/'
            self.setCookie(environ, self.cookie_name, str(sess.id), path=path)
        else:
            sess = self.loadSession(environ, id)
        environ[SESSKEY] = sess
        return sess

    def createSession(self, environ, id=None):
        if id is None:
            id = self.createID(environ)
        t = id.timestamp
        return self._fact(id=id, ctime=t, atime=t, data={})

    def loadSession(self, environ, id):
        f = self._sessionFile(id)
        try:
            c, a, d = cPickle.load(open(f,'rb'))
        except:
            try: os.unlink(f)
            except: pass
            t = id.timestamp
            c, a, d = t, t, {}
        return self._fact(id=id, ctime=c, atime=a, data=d)

    def saveSession(self, environ, session):
        session.atime = self.time()
        fd = open(self._sessionFile(session.id),'wb')
        cPickle.dump((session.ctime, session.atime, session.data), fd)
        fd.close()

    def deleteSession(self, environ, id):
        id = self.verifyID(environ, id)
        f = self._sessionFile(id)
        if os.path.isfile(f):
            os.unlink(f)

    def touch(self, environ, id, stamp=None):
        if stamp is None:
            stamp = self.time()
        id = self.verifyID(environ, id)
        fname = self._sessionFile(id)
        ct, at, data = cPickle.load(open(fname,'rb'))
        cPickle.dump((ct, stamp, data), open(fname,'wb'))

    def deleteExpiredSessions(self, older_than):
        d = self._getDir()
        for n in os.listdir(d):
            f = os.path.join(d, n)
            c,a,d = cPickle.load(open(f,'rb'))
            if a < older_than:
                os.unlink(f)
