from wm4.api import *
from dispatch import strategy, predicates, ast_builder
import sys, inspect, operator

from routes import *
from authorization import GroupName
from navigation import Link, actionLinks, DELIMITER
from errors import NotFound


__all__ = [
    'ActionRoute', 'ClassRoute', 'ListRoute', 'InstanceRoute', 'ViewRoute',
    'findEntity', 'EntitySitemap', 'mainMenuParam', 'SitemapWithMainMenu',
    'url_for', 'Location',
]

_ = i18n.MessageIDFactory('wm4')


class ActionRoute(Route):
    '''Typical route that do some action.

       Matching PATH_INFO part, action and template name are constructed
       from class name.
    '''
    match = binding.classAttr(binding.Make(lambda c: c.__name__))
    action = binding.classAttr(binding.Make(lambda c: Const(c.__name__)))

class ClassRoute(ActionRoute):
    signature = Signature('clsObj','action')

class ListRoute(ClassRoute):
    match = binding.classAttr(binding.Constant(''))
    signature = Signature('clsObj') | Signature('clsObj','action')

class InstanceRoute(ActionRoute):
    signature = Signature('obj','action')

class ViewRoute(InstanceRoute):
    match = binding.classAttr(binding.Constant(''))
    signature = Signature('obj') | Signature('obj','action')







def findEntity(route, clsObj, key):
    try:
        o = settings.workspace().get(clsObj, key, default=None)
    except KeyError:
        raise SitemapError('Object not found', clsObj, key)
    if o is None:
        raise SitemapError('Object not found', clsObj, key)
        #raise NotFound(clsObj, key)
    return o


class EntitySitemap(Route):
    virtual = True
    clsObj = None
    tmplmod = Param(lambda s: '.'.join(s.clsObj.__module__.split('.')[:-1]))

    viewPerm = Const(None)
    editPerm = Const(GroupName('admin'))

    workspace = Lazy(lambda s: settings.workspace())

    entity_error_template = Const('wm4.web.bases/entity_error.wt')

    class cls(Route):
        virtual = True
        actionLinks = Lazy(lambda s: actionLinks(s, '', './add'))

        class list(ListRoute):
            permission = Param(lambda s: s.viewPerm, typ=object)
            template = Param(lambda s: '%s/%s.wt' % (s.tmplmod, s.action))
            limit = Const(20)
            asAction = Lazy(Link.fromRoute)

        class add(ClassRoute):
            permission = Param(lambda s: s.editPerm, typ=object)
            template = Param(lambda s: '%s/form.wt' % s.tmplmod)
            asAction = Lazy(Link.fromRoute)

    class obj(Route):
        virtual = True
        match = ':oid/'
        oid = Param(lambda s: s.obj._oid)

        def obj(klass):
            for p in klass.__all_parents__:
                if not hasattr(p, 'clsObj'):
                    continue
                p = p.clsObj
                if not isinstance(p, Const):
                    continue
                clsObj = p.value
            return Param(lambda s: findEntity(s, clsObj, s.oid), typ=clsObj)
        obj = binding.classAttr(binding.Make(obj))

        asAction = Lazy(Link.fromRoute, virtual=True)
        actionLinks = Lazy(lambda s: actionLinks(
            s, '', './edit', './delete', DELIMITER, '../', '../add',
        ))

        class view(ViewRoute):
            permission = Param(lambda s: s.viewPerm)
            template = Param(lambda s: '%s/%s.wt' % (s.tmplmod, s.action))

        class edit(InstanceRoute):
            permission = Param(lambda s: s.editPerm)
            template = Param(lambda s: '%s/form.wt' % s.tmplmod)

        class delete(InstanceRoute):
            permission = Param(lambda s: s.editPerm)
            def asAction(self):
                return Link.fromRoute(self,
                    confirm=_(u'Are you sure you want to delete this record?'))
            asAction = Lazy(asAction)



def mainMenuParam(self):
    from xml4 import xhtml as xh
    s = xh.select(
        onchange="if (this.value!='menu') window.location.href=this.value;")
    for r in self.mainMenuItems:
        l = r.asMainMenu
        v = r.pathInfo
        if l == self.asMainMenu:
            s.append(xh.option(l, value=v, selected='selected'))
        else:
            s.append(xh.option(l, value=v))
        s.append('\n')
    return xh.form(s, style="display: inline;")

class SitemapWithMainMenu(EntitySitemap):
    '''Requires definition of mainMenuItems on parent route

       class CompSitemap(Route):

            mainMenuItems = Lazy(lambda s: [s/p for p in ['comp1/','comp2/']])

            class comp1(SitemapWithMainMenu):
                match = 'comp1/'
                ...

            class comp2(SitemapWithMainMenu):
                match = 'comp2/'
                asMainMenu = Const(u'Component 2')
                ...
    '''

    asMainMenu = Param(lambda s: s.__class__.__name__)

    mainMenu = Const(mainMenuParam)

    class cls(EntitySitemap.cls):
        actionLinks = Lazy(lambda s: actionLinks(
                    s, '', './add', DELIMITER, s.mainMenu(s)))

    class obj(EntitySitemap.obj):
        actionLinks = Lazy(lambda s: actionLinks(
                    s, '', './edit', './delete', DELIMITER,
                    '../', '../add', DELIMITER, s.mainMenu(s)))






def url_for(when, wrap=False):
    '''Hand made or external URLs

        class MyApp(Route):

            @url_for('obj in Foo')
            def url_for_foo(obj):
                return '/foo/%d' % obj.foo_id

            @url_for('obj in Bar')
            def url_for_foo(obj, action='view'):
                return '/foo/%d/%s' % (obj.foo_id, action)
    '''
    def callback(func):
        args, star, dstar, defaults = inspect.getargspec(func)
        alen = len(args)
        if star or dstar:
            raise TypeError('star or dstar arguments not supported')
        namespace = dict(
            (a, predicates.Call(
                operator.getitem, strategy.Argument(pos=1), predicates.Const(i)
            )) for i,a in enumerate(args)
        )
        frame = sys._getframe(2)
        builder = predicates.CriteriaBuilder(
            namespace, frame.f_locals, frame.f_globals, frame.f_builtins
        )

        _alen = strategy.Signature([
            (strategy.Argument(pos=0), strategy.Inequality('=',alen))
        ])
        _when = ast_builder.parse_expr(when, builder)
        pred = _alen & _when

        signatures = [(pred, lambda al,av: func(*av))]
        for i in range(len(defaults or ())):
            i += 1
            _alen = strategy.Signature([
                (strategy.Argument(pos=0), strategy.Inequality('=',alen-i))
            ])
            pred = _alen & _when
            signatures.append((
                pred, lambda al,av,i=i:func(*(av + tuple(defaults[-i:])))
            ))

        func.url_for = signatures
        return func

    return callback



class Location(binding.Component):
    '''For backward compatibility'''

    defaults = {}
    prefix = ''

    def __init__(self, **defaults):
        super(Location,self).__init__(defaults=defaults)

    route_list = binding.Make(list)

    def route(self, pattern, handler, permission=NOT_GIVEN, **params):
        route = (pattern, handler, permission, params)
        for i,r in enumerate(self.route_list):
            if r[0] == pattern:
                self.route_list[i] = route
                return
        self.route_list.append(route)

    url_for_list = binding.Make(list)

    def url_for(self, when, gen):
        uf = url_for(when)(gen)
        self.url_for_list.append(uf)

    def copy(self):
        loc = Location()
        loc.defaults = self.defaults.copy()
        loc.route_list = self.route_list[:]
        loc.url_for_list = self.url_for_list[:]
        return loc

    def makeRoute(self):
        cdict = dict((a,Const(v)) for a,v in self.defaults.iteritems())
        meta = type(Route)

        for i,(m,h,r,p) in enumerate(self.route_list):
            name = '_route_%d' % i
            rdict = dict(
                match = m,
                handler = Const(h),
                permission = Const(r),
            )
            for a,v in p.iteritems():
                rdict[a] = Const(v)
            for v in RoutePattern([m]).variables:
                rdict[v[3:]] = Param()
            cdict[name] = meta(name, (Route,), rdict)

        for i,u in enumerate(self.url_for_list):
            cdict['url_for_%d' % i] = u

        return meta('location', (Route,), cdict)

    _route = binding.Make(lambda s: s.makeRoute())

    def get_url(self, *args):
        return self._route.get_url(*args)
