/*

Depends on:

    - x_core.js
      xGetElementById, xParent

    - x_event.js
      xEvent, xAddEventListener, xRemoveEventListener, xPreventDefault

Usage:
    
    
    Source elements S1 and S2 can be dragged and dropped on target elements
    T1, T2 and T3.

        ...
        <div id='S1'>Drag source 1</div>
        <div id='S2'>Drag source 2</div>
        ...
        <span id='T1'>Drop target 1</span>
        <span id='T2'>Drop target 2</span>
        <span id='T3'>Drop target 3</span>
        ...
    
    Instance of wmDragAndDrop manages DnD process.

        dnd = new wmDragAndDrop(myOnDropHandler);

        window.onload = function() {

            // Draggable elements are found by IDs.
            // Second argument for registration function is value
            // that receives onDrop handler as second argument (srcArg).
            dnd.enableSource('S1', 's1a');
            dnd.enableSource('S2', 's2a'); 

            // Drop targets are found by IDs.
            // Second argument for registration function is value
            // that receives onDrop handler as fourth argument (tgtArg).
            dnd.enableTarget('T1', 't1a');
            dnd.enableTarget('T2', 't2a');
            dnd.enableTarget('T3', 't3a');
        }

    Drop handler receives source element (that is dragged, i.e. some <div>)
    and target element (i.e. some <span>) together with args specified by
    registration routines.

        def myOnDropHandler(src, srcArg, tgt, tgtArg) {
            alert('Element ' + srcElem + ' was dropped on ' + tgtElem +
                  ' with args (' + srcArg + ' -> ' + tgtArg + ')'
            );
        }

*/


function wmDragAndDrop(onDrop) {
    var self = this;
    this.sourceElements = new Array();
    this.targetElements = new Array();
    this.source = null;
    this.targetCursor = 'crosshair';
    this.notAllowedCursor = 'not-allowed';

    var body=document.documentElement ? document.documentElement:document.body;

    this.onDrop = function(srcElem, srcArg, tgtElem, tgtArg) {
        alert('Element ' + srcElem + ' was dropped on ' + tgtElem +
              ' with args (' + srcArg + ' -> ' + tgtArg + ')'
        );
    }

    if(onDrop) this.onDrop = onDrop;

    this.enableSource = function(src, srcArg) {
        var e = xGetElementById(src);
        self.sourceElements.push(new Array(e, srcArg));
        xAddEventListener(e,'mousedown', self.onMouseDown, false);
        e.wmDragSource = true;
    }
    
    this.enableTarget = function(tgt, tgtArg) {
        var e = xGetElementById(tgt);
        self.targetElements.push(new Array(e, tgtArg));
        e.wmDropTarget = true;
    }
    
    this.onMouseDown = function(evt) {
        var s = (new xEvent(evt)).target;
        while(s && !s.wmDragSource) {
            s = xParent(s, true);
        }
        if(s) {
            xPreventDefault(evt);
            self.source = s;
            var te = self.targetElements;
            for(var i=0; i<te.length; i++) {
                te[i][0].style.cursor = self.targetCursor;
            }
            xAddEventListener(document,'mouseup', self.onMouseUp, false);
            xAddEventListener(document,'mousemove', self.onMouseMove, false);
            body.style.cursor = self.notAllowedCursor;
        }
    }

    this.dragFinish = function() {
        var te = self.targetElements;
        xRemoveEventListener(document,'mouseup', self.onMouseUp, false);
        xRemoveEventListener(document,'mousemove', self.onMouseMove, false);
        for(var i=0; i<te.length; i++) {
            te[i][0].style.cursor = 'auto';
        }
        self.src = null;
        body.style.cursor = 'auto';
    }

    this.onMouseMove = function(evt) {
        xPreventDefault(evt);
    }

    this.onMouseUp = function(evt) {
        var tgt = (new xEvent(evt)).target;
        while(tgt && !tgt.wmDropTarget) {
            tgt = xParent(tgt, true);
        }

        self.dragFinish();
        if(tgt) {
            xPreventDefault(evt);

            var src = self.source;
            var se = self.sourceElements;
            var te = self.targetElements;
            var srcArg = null;
            var tgtArg = null;
            for(var i=0; i<se.length; i++) {
                if(se[i][0] == src) srcArg=se[i][1];
            }
            for(var i=0; i<te.length; i++) {
                if(te[i][0] == tgt) tgtArg=te[i][1];
            }
            self.onDrop(src, srcArg, tgt, tgtArg);
        }
    }
}
