# -*- coding: ISO-8859-2 -*-

from unittest import *
from wm4.api import *
from wm4.web.authentication import *
from Cookie import SimpleCookie
import time

from utils import *


class User(Record):
    pass


_users = dict(
    ian = User(login='ian', password='foo'),
    bob = User(login='bob', password='bar'),
)



def find_user(middleware, environ, username):
    return _users.get(username)




class TestHTTPAuthenticationMiddleware(WebCase):

    def setUp(self):
        WebCase.setUp(self)
        self.parse_ini((
            '[WSGI Middleware authentication]\n'
            'find_user = "%s.find_user"\n'
        ) % (__name__,))

    def authapp(self, app):
        app = self.mkapp(
            'environment', 'errors', 'transaction', 'cookies',
            'http_exceptions', 'session', 'authentication', funcMiddleware(app)
        )
        return app

    def test10_AuthenticationNotNeeded(self):
        _u = []
        def app(env, start):
            _u.append(getUser(env))
            start('200 OK', [])
            return ['OK']

        s,h,b = self.handle(self.authapp(app))
        self.assertEqual(s, '200 OK')
        self.assertEqual(h, [])
        self.assertEqual(b, 'OK')
        self.assertEqual(_u[0], None)

    def test12_Unauthorized(self):
        def app(env, start):
            getUser(env)
            start('200 OK', [])
            return ['OK']

        s,h,b = self.handle(self.authapp(app), self.env('https://localhost'))
        self.assertEqual(s, '401 Unauthorized')
        self.assertEqual(h, [
         ('Content-type', 'text/plain'),
         ('WWW-Authenticate', 'Basic realm="Enter your username and password"'),
         ('X-Authentication-Domain', 'remote_user'),
        ])
        self.assertEqual(b, '401 Unauthorized\r\n')

    def test14_BadPassword(self):
        def app(env, start):
            getUser(env)
            start('200 OK', [])
            return ['OK']

        s,h,b = self.handle(
            self.authapp(app),
            self.env('https://localhost', auth=('ian','xxx'))
        )
        self.assertEqual(s, '401 Unauthorized')
        self.assertEqual(h, [
         ('Content-type', 'text/plain'),
         ('WWW-Authenticate', 'Basic realm="Enter your username and password"'),
         ('X-Authentication-Domain', 'remote_user'),
        ])
        self.assertEqual(b, '401 Unauthorized\r\n')

    def test16_BadLogin(self):
        def app(env, start):
            getUser(env)
            start('200 OK', [])
            return ['OK']

        s,h,b = self.handle(
            self.authapp(app),
            self.env('https://localhost', auth=('xxx','xxx'))
        )
        self.assertEqual(s, '401 Unauthorized')
        self.assertEqual(h, [
         ('Content-type', 'text/plain'),
         ('WWW-Authenticate', 'Basic realm="Enter your username and password"'),
         ('X-Authentication-Domain', 'remote_user'),
        ])
        self.assertEqual(b, '401 Unauthorized\r\n')

    def test17_EmptyLogin(self):
        def app(env, start):
            getUser(env)
            start('200 OK', [])
            return ['OK']

        s,h,b = self.handle(
            self.authapp(app),
            self.env('https://localhost', auth=('','xxx'))
        )
        self.assertEqual(s, '401 Unauthorized')
        self.assertEqual(h, [
         ('Content-type', 'text/plain'),
         ('WWW-Authenticate', 'Basic realm="Enter your username and password"'),
         ('X-Authentication-Domain', 'remote_user'),
        ])
        self.assertEqual(b, '401 Unauthorized\r\n')

    def test20_GoodPassword(self):
        _u = []
        def app(env, start):
            _u.append(getUser(env))
            start('200 OK', [])
            return ['OK']

        s,h,b = self.handle(
            self.authapp(app),
            self.env('https://localhost', auth=('ian','foo'))
        )

        self.assertEqual(s, '200 OK')
        self.assertEqual(h, [])
        self.assertEqual(b, 'OK')
        self.assertEqual(_u[0], _users['ian'])






_time = int(time.time())
def time():
    global _time
    _time += 1
    return _time


class TestCookieAuthenticationMiddleware(WebCase):

    domain = 'app.registered_user'

    def setUp(self):
        WebCase.setUp(self)
        self.parse_ini((
          '[WSGI Middleware authentication]\n'
          'factory = "wm4.web.authentication.CookieAuthenticationMiddleware"\n'
          'find_user = "%s.find_user"\n'
          'time = importString("%s.time")\n'
          'on_ports = [80, 443]\n'
          'auth_domain = "%s"\n'
        ) % (__name__, __name__, self.domain))

    def authapp(self, app):
        app = self.mkapp(
            'environment', 'errors', 'transaction', 'cookies',
            'http_exceptions', 'session', 'authentication', funcMiddleware(app)
        )
        return app

    def test01_Repr(self):
        token = AuthenticationToken(
            username = 'ian',
            timestamp = _time,
            ttl = 1000,
            secret = web.app_secret(),
            remote_addr = '127.0.0.1',
            browser = 'unknown',
        )
        repr(token)

    def test10_AuthenticationNotNeeded(self):
        _u = []
        def app(env, start):
            _u.append(getUser(env, self.domain))
            start('200 OK', [])
            return ['OK']
        s,h,b = self.handle(self.authapp(app), self.env('http://localhost:42'))
        self.assertEqual(s, '200 OK')
        self.assertEqual(h, [])
        self.assertEqual(b, 'OK')
        self.assertEqual(_u[0], None)

    def test20_Unauthorized(self):
        def app(env, start):
            getUser(env, self.domain)
            start('200 OK', [])
            return ['OK']
        s,h,b = self.handle(self.authapp(app))
        self.assertEqual(s, '303 See Other')
        self.assertEqual(h, [
         ('X-Authentication-Domain', 'app.registered_user'),
         ('Location', 'http://localhost/login?ref=http%3A//localhost/'),
        ])
        self.assertEqual(b, '303 See Other\r\n')

    def test21_AuthorizationOnDemandInAnotherDomain(self):
        def app(env, start):
            raise web.Unauthorized(auth_domain='app.other_domain')
        s,h,b = self.handle(self.authapp(app), self.env('http://localhost:42'))
        self.assertEqual(s, '401 Unauthorized')
        self.assertEqual(h, [
         ('Content-type', 'text/plain'),
         ('WWW-Authenticate', 'Basic realm="Enter your username and password"'),
         ('X-Authentication-Domain', 'app.other_domain'),
        ])
        self.assertEqual(b, '401 Unauthorized\r\n')

    def test22_AuthorizationOnDemand(self):
        def app(env, start):
            raise web.Unauthorized(auth_domain=self.domain)
        s,h,b = self.handle(self.authapp(app), self.env('http://localhost:42'))
        self.assertEqual(s, '303 See Other')
        self.assertEqual(h, [
         ('X-Authentication-Domain', 'app.registered_user'),
         ('Location','http://localhost:42/login?ref=http%3A//localhost%3A42/'),
        ])
        self.assertEqual(b, '303 See Other\r\n')

    def test30_AuthenticatedWithToken(self):
        _u = []
        def app(env, start):
            _u.append(getUser(env, self.domain))
            start('200 OK', [])
            return ['OK']
        token = AuthenticationToken(
            username = 'ian',
            timestamp = time(),
            ttl = 1000,
            secret = web.app_secret(),
            remote_addr = '127.0.0.1',
            browser = 'unknown',
        )
        s,h,b = self.handle(
            self.authapp(app),
            self.env('http://localhost', cookies={self.domain: str(token)}),
        )
        # timestamp is updated in token and new cookie is set
        new = AuthenticationToken(
            username = 'ian',
            timestamp = token.timestamp + 1,
            ttl = 1000,
            secret = web.app_secret(),
            remote_addr = '127.0.0.1',
            browser = 'unknown',
        )
        self.assertEqual(s, '200 OK')
        c = SimpleCookie(dict(h)['Set-Cookie']).get('app.registered_user')
        self.assertEqual(c.value, str(new))
        self.assertEqual(b, 'OK')
        self.assertEqual(_u[0], _users['ian'])

    def test32_AttemptWithInvalidToken(self):
        _u = []
        def app(env, start):
            _u.append(getUser(env, self.domain))
            start('200 OK', [])
            return ['OK']
        token = AuthenticationToken(
            username = 'ian',
            timestamp = time(),
            ttl = 1000,
            secret = web.app_secret(),
            remote_addr = '127.0.0.1',
            browser = 'Godzilla',
        )
        s,h,b = self.handle(
            self.authapp(app),
            self.env('http://localhost', cookies={self.domain: str(token)}),
        )
        self.assertEqual(s, '303 See Other')
        self.assertEqual(h, [
         ('X-Authentication-Domain', 'app.registered_user'),
         ('Location','http://localhost/login?ref=http%3A//localhost/'),
        ])
        self.assertEqual(b, '303 See Other\r\n')
        self.assertEqual(_u, [])

    def test33_AttemptWithInvalidLogin(self):
        _u = []
        def app(env, start):
            _u.append(getUser(env, self.domain))
            start('200 OK', [])
            return ['OK']
        token = AuthenticationToken(
            username = 'foo',
            timestamp = time(),
            ttl = 1000,
            secret = web.app_secret(),
            remote_addr = '127.0.0.1',
            browser = 'unknown',
        )
        s,h,b = self.handle(
            self.authapp(app),
            self.env('http://localhost', cookies={self.domain: str(token)}),
        )
        self.assertEqual(s, '303 See Other')
        self.assertEqual(h, [
         ('X-Authentication-Domain', 'app.registered_user'),
         ('Location','http://localhost/login?ref=http%3A//localhost/'),
        ])
        self.assertEqual(b, '303 See Other\r\n')
        self.assertEqual(_u, [])

    def test34_AuthenticatedWithTimeoutedToken(self):
        _u = []
        def app(env, start):
            _u.append(getUser(env, self.domain))
            start('200 OK', [])
            return ['OK']
        token = AuthenticationToken(
            username = 'ian',
            timestamp = time() - 10000,
            ttl = 1000,
            secret = web.app_secret(),
            remote_addr = '127.0.0.1',
            browser = 'unknown',
        )
        s,h,b = self.handle(
            self.authapp(app),
            self.env('http://localhost', cookies={self.domain: str(token)}),
        )
        self.assertEqual(s, '303 See Other')
        self.assertEqual(h, [
         ('X-Authentication-Domain', 'app.registered_user'),
         ('Location','http://localhost/login?ref=http%3A//localhost/'),
        ])
        self.assertEqual(b, '303 See Other\r\n')
        self.assertEqual(_u, [])

    def test35_AuthenticatedWithTimeoutedTokenButRememberOption(self):
        _u = []
        def app(env, start):
            _u.append(getUser(env, self.domain))
            start('200 OK', [])
            return ['OK']
        token = AuthenticationToken(
            username = 'ian',
            timestamp = time() - 100000,
            ttl = 200000,
            secret = web.app_secret(),
            remote_addr = '127.0.0.1',
            browser = 'unknown',
        )
        s,h,b = self.handle(
            self.authapp(app),
            self.env('http://localhost', cookies={self.domain: str(token)}),
        )
        # timestamp is updated in token and new cookie is set
        new = AuthenticationToken(
            username = 'ian',
            timestamp = _time,
            ttl = 200000,
            secret = web.app_secret(),
            remote_addr = '127.0.0.1',
            browser = 'unknown',
        )
        self.assertEqual(s, '200 OK')
        c = SimpleCookie(dict(h)['Set-Cookie']).get('app.registered_user')
        self.assertEqual(c.value, str(new))
        self.assertEqual(_u[0], _users['ian'])

    def test38_AuthenticatedViaHTTPS(self):
        _u = []
        def app(env, start):
            _u.append(getUser(env, self.domain))
            start('200 OK', [])
            return ['OK']
        token = AuthenticationToken(
            username = 'ian',
            timestamp = time(),
            ttl = 1000,
            secret = web.app_secret(),
            remote_addr = '127.0.0.1',
            browser = 'unknown',
        )
        s,h,b = self.handle(
            self.authapp(app),
            self.env('https://localhost', cookies={self.domain: str(token)}),
        )
        # timestamp is updated in token and new cookie is set
        new = AuthenticationToken(
            username = 'ian',
            timestamp = token.timestamp + 1,
            ttl = 1000,
            secret = web.app_secret(),
            remote_addr = '127.0.0.1',
            browser = 'unknown',
        )
        self.assertEqual(s, '200 OK')
        c = SimpleCookie(dict(h)['Set-Cookie']).get('app.registered_user')
        self.assertEqual(c.value, str(new))
        self.assertEqual(b, 'OK')
        self.assertEqual(_u[0], _users['ian'])

    def XXXtest40_BuiltinLoginHandler(self):
        def app(env, start):
            start('200 OK', [])
            return ['FOO']
        s,h,b = self.handle(
            self.authapp(app),
            self.env('http://localhost/login',
                     post = {'username':'ian', 'password':'foo'}),
        )
        self.assertEqual(s, '303 See Other')
        self.assertEqual(len(h), 2)
        self.assertEqual(h[0], ('Location','http://localhost/'))
        self.assertEqual(h[1][0], 'Set-Cookie')
        token = AuthenticationToken.unpack(
            h[1][1].split('=')[1] + '=',
            secret = web.app_secret(),
            remote_addr = '127.0.0.1',
            browser = 'unknown',
        )
        self.assertEqual(token.username, 'ian')

    def test42_BuiltinLoginHandlerWithInvalidPassword(self):
        def app(env, start):
            start('200 OK', [])
            return ['FOO']
        s,h,b = self.handle(
            self.authapp(app),
            self.env('http://localhost/login',
                     post = {'username':'ian', 'password':'xxx'}),
        )
        self.assertEqual(s, '200 OK')
        # XXX self.assertEqual(b, '303 See Other')

    def XXXtest50_BuiltinLogoutToAppURL(self):
        def app(env, start):
            start('200 OK', [])
            return ['OK']
        token = str(AuthenticationToken(
            username = 'ian',
            timestamp = time(),
            secret = web.app_secret(),
            remote_addr = '127.0.0.1',
            browser = 'unknown',
        ))
        s,h,b = self.handle(
            self.authapp(app),
            self.env('http://localhost/logout', cookies={self.domain:token}),
        )
        self.assertEqual(s, '303 See Other')
        self.assertEqual(h, [
         ('Location','http://localhost/'),
         ('Set-Cookie',
          'app.registered_user=; expires=Thu, 01-01-1970 00:00:01 GMT; Path=/;')
        ])

    def XXXtest52_BuiltinLogoutToReferer(self):
        def app(env, start):
            start('200 OK', [])
            return ['OK']
        token = str(AuthenticationToken(
            username = 'ian',
            timestamp = time(),
            secret = web.app_secret(),
            remote_addr = '127.0.0.1',
            browser = 'unknown',
        ))
        s,h,b = self.handle(
            self.authapp(app),
            self.env('http://localhost/logout',
                     env={'HTTP_REFERER': 'http://localhost/bar'},
                     cookies={self.domain:token}),
        )
        self.assertEqual(s, '303 See Other')
        self.assertEqual(h, [
         ('Location','http://localhost/bar'),
         ('Set-Cookie',
          'app.registered_user=; expires=Thu, 01-01-1970 00:00:01 GMT; Path=/;')
        ])

    def XXXtest54_BuiltinLogoutToParam(self):
        def app(env, start):
            start('200 OK', [])
            return ['OK']
        token = str(AuthenticationToken(
            username = 'ian',
            timestamp = time(),
            secret = web.app_secret(),
            remote_addr = '127.0.0.1',
            browser = 'unknown',
        ))
        s,h,b = self.handle(
            self.authapp(app),
            self.env('http://localhost/logout?ref=http%3A//localhost/foo',
                     env={'HTTP_REFERER': 'http://localhost/bar'},
                     cookies={self.domain:token}),
        )
        self.assertEqual(s, '303 See Other')
        self.assertEqual(h, [
         ('Location','http://localhost/foo'),
         ('Set-Cookie',
          'app.registered_user=; expires=Thu, 01-01-1970 00:00:01 GMT; Path=/;')
        ])



def suite():
    return TestSuite((
        makeSuite(TestHTTPAuthenticationMiddleware),
        makeSuite(TestCookieAuthenticationMiddleware),
    ))


if __name__ == '__main__':
    main(defaultTest='suite')
