# -*- coding: ISO-8859-2 -*-

from unittest import *
from wm4.api import *

from utils import *


class TestErrors(WebCase):

    def setUp(self):
        WebCase.setUp(self)
        self.mh = logs.Memory(level=logs.DEBUG)
        self.mh.flush()
        self.tc = LoggingTransactionComponent()
        config.Property('wm4.logs.web')(config.Value(self.mh))
        config.Property('wm4.logs.web')()._clearLog()

    def errapp(self, exc, *args, **kw):
        def bad_app(env, start_response):
            self.tc.joinedTxn
            raise exc(*args, **kw)
        app = self.mkapp(
            'environment', 'errors', 'transaction', 'cookies',
            'http_exceptions', funcMiddleware(bad_app)
        )
        return app

    def test10_ErrorCatching(self):
        app = self.errapp(ZeroDivisionError)
        s,h,b = self.handle(app, self.env())
        self.assertEqual(s, '500 Internal Server Error')
        self.assertEqual(h, [
            ('Content-type', 'text/html; charset=utf-8'),
            ('Content-Length', '1467'),
        ])
        self.failUnless('ERROR 500' in b, b)
        lb = self.mh._getLog()
        self.assertEqual(len(lb), 1)
        assert 'ZeroDivisionError' in lb[0]
        self.assertEqual(self.tc.log, ['begin', 'abort'])

    @with_ini('[WSGI Middleware errors]\nthrow_errors = True\n')
    def test12_ErrorRaising(self):
        app = self.errapp(KeyError, 42)
        self.assertRaises(KeyError, self.handle, app, self.env())
        self.assertEqual(len(self.mh._getLog()), 0)
        self.assertEqual(self.tc.log, ['begin', 'abort'])


    def test13_ErrorAfterStartResponse(self):
        def handler(env, start_response):
            self.tc.joinedTxn
            start_response('200 OK', [('Content-type', 'image/gif')])
            raise AttributeError('foo')

        app = self.mkapp(
            'errors', 'transaction', 'http_exceptions', funcMiddleware(handler)
        )
        s,h,b = self.handle(app, self.env())

        self.assertEqual(s, '500 Internal Server Error')
        self.assertEqual(h, [
            ('Content-type', 'text/html; charset=utf-8'),
            ('Content-Length', '1467'),
        ])
        self.failUnless('ERROR 500' in b, b)
        lb = self.mh._getLog()
        assert 'AttributeError' in lb[0]
        self.assertEqual(self.tc.log, ['begin', 'abort'])

    def test303_SeeOther(self):
        app = self.errapp(web.SeeOther, location='http://foo.cz')
        s,h,b = self.handle(app, self.env())
        self.assertEqual(s, '303 See Other')
        self.assertEqual(h, [('Location', 'http://foo.cz')])
        self.assertEqual(b, '303 See Other\r\n')
        self.assertEqual(self.tc.log, ['begin', 'commit'])

    def test304_NotModified(self):
        app = self.errapp(web.NotModified)
        s,h,b = self.handle(app, self.env())
        self.assertEqual(s, '304 Not Modified')
        self.assertEqual(h, [])
        self.assertEqual(b, '')
        self.assertEqual(self.tc.log, ['begin', 'commit'])

    def test304_NotModifiedWithHeaders(self):
        app = self.errapp(web.NotModified, headers=[('ETag','foo')])
        s,h,b = self.handle(app, self.env())
        self.assertEqual(s, '304 Not Modified')
        self.assertEqual(h, [('ETag','foo')])
        self.assertEqual(b, '')
        self.assertEqual(self.tc.log, ['begin', 'commit'])

    def test307_TemporaryRedirect(self):
        app = self.errapp(web.TemporaryRedirect, location='http://foo.cz')
        s,h,b = self.handle(app, self.env())
        self.assertEqual(s, '307 Temporary Redirect')
        self.assertEqual(h, [('Location', 'http://foo.cz')])
        self.assertEqual(b, '307 Temporary Redirect\r\n')
        self.assertEqual(self.tc.log, ['begin', 'commit'])

    def test401_Unauthorized(self):
        app = self.errapp(web.Unauthorized, realm='please, login')
        s,h,b = self.handle(app, self.env())
        self.assertEqual(s, '401 Unauthorized')
        self.assertEqual(h, [
            ('Content-type', 'text/plain'),
            ('WWW-Authenticate', 'Basic realm="please, login"'),
        ])
        self.assertEqual(b, '401 Unauthorized\r\n')
        self.assertEqual(self.tc.log, ['begin', 'commit'])

    def test403_NotAllowed(self):
        app = self.errapp(web.NotAllowed)
        s,h,b = self.handle(app, self.env())
        self.assertEqual(s, '403 Forbidden')
        #self.assertEqual(h, [('Content-type', 'text/plain')])
        #self.assertEqual(b, '403 Forbidden\r\n')
        self.assertEqual(self.tc.log, ['begin', 'commit'])

    def test404_NotFound(self):
        app = self.errapp(web.NotFound)
        s,h,b = self.handle(app, self.env())
        self.assertEqual(s, '404 Not Found')
        #self.assertEqual(h, [('Content-type', 'text/plain')])
        #self.assertEqual(b, '404 Not Found\r\n')
        self.assertEqual(self.tc.log, ['begin', 'commit'])

    def handle404(self, environ, traceback):
        return '404', [('Foo', '42')], ['Foo', 'Bar']

    @with_ini('[WSGI Middleware http_exceptions]\nhandle404 = handle404\n',
        handle404=handle404)
    def test404_NotFoundWithCustomHandler(self):
        app = self.errapp(web.NotFound)
        s,h,b = self.handle(app, self.env())
        self.assertEqual(s, '404')
        self.assertEqual(h, [('Foo', '42')])
        self.assertEqual(b, 'FooBar')
        self.assertEqual(self.tc.log, ['begin', 'commit'])

    @with_ini(
        '[WSGI Middleware http_exceptions]\n'
        'template404 = "wm4.web.tests.res/404.wt"\n')
    def test404_NotFoundWithCustomTemplate(self):
        app = self.errapp(web.NotFound)
        s,h,b = self.handle(app, self.env())
        self.assertEqual(s, '404 Not Found')
        self.assertEqual(h, [
            ('Content-type', 'text/html; charset=utf-8'),
            ('Content-Length', '283')
        ])
        assert b.startswith('\n<!DOCTYPE html')
        assert b.endswith('</html>\n')
        self.assertEqual(self.tc.log, ['begin', 'commit'])

    def test500_InternalServerError(self):
        app = self.errapp(web.WebException)
        s,h,b = self.handle(app, self.env())
        self.assertEqual(s, '500 Internal Server Error')
        self.assertEqual(h, [
            ('Content-type', 'text/html; charset=utf-8'),
            ('Content-Length', '1467'),
        ])
        self.failUnless('ERROR 500' in b, b)
        # exceptions with code >= 500 abort transactions
        self.assertEqual(self.tc.log, ['begin', 'abort'])

def suite():
    return TestSuite((
        makeSuite(TestErrors),
    ))


if __name__ == '__main__':
    main(defaultTest='suite')
