# -*- coding: utf-8 -*-

# Making in-process XML-RPC calls in Python (with wsgi_intercept)
# http://www.advogato.org/person/titus/diary.html?start=218

from unittest import *
from wm4.api import *
from wm4.web.rpc import *
from utils import *
import xmlrpclib


class RPC(RpcHandler):

    @rpc()
    def add(self, a, b):
        return a + b

    @rpc()
    def echo(self, arg):
        return arg

    def nonrpc(self):
        pass

class HIER(RpcHandler):

    class h1(RPC):

        @rpc()
        def echo(self, arg):
            return 'echo:%s' % arg

        @rpc(permission=web.Authenticated())
        def guarded(self):
            return 'GUARDED'

    class h2(RPC):
        permission = web.Nobody()

    @rpc()
    def foo(self):
        return 'FOO'

    @rpc(permission=web.Authenticated())
    def double(self, arg):
        return arg * 2

class GUARDED(RPC):
    permission = web.Authenticated()


class TestRpcDispatcher(TestCase):

    def test10_SimpleDispatch(self):
        d = RPC(req=None)
        m = d._rpc_dispatch('add')
        self.assertEqual(m, d.add)
        m = d._rpc_dispatch('echo')
        self.assertEqual(m, d.echo)
        self.assertEqual(m.permission, None)

    def test12_HierDispatch(self):
        d = HIER(req=None)
        m = d._rpc_dispatch('h1.add')
        self.assertEqual(m, d.h1.add)
        m = d._rpc_dispatch('h1.echo')
        self.assertEqual(m, d.h1.echo)
        m = d._rpc_dispatch('foo')
        self.assertEqual(m, d.foo)


class SM(web.Route):

    class xml_rpc(web.Route):
        match = 'xml_rpc/'
        handler = web.Const(HIER)

class SMG(web.Route):

    class xml_rpc(web.Route):
        match = 'xml_rpc/'
        handler = web.Const(GUARDED)

INI = """\
[wm4.logs]
* = logs.Null()

[WSGI Middleware sitemap]
sitemap = importString('%s.%%s')
""" % __name__


class TestXmlRpc(WebCase):

    def xenv(self, method, params):
        req = self.req(url = 'http://localhost/xml_rpc/')
        env = req.environ
        xml = xmlrpclib.dumps(params, method)
        env['REQUEST_METHOD'] = 'POST'
        env['CONTENT_TYPE'] = 'text/xml'
        env['CONTENT_LENGTH'] = len(xml)
        env['wsgi.input'].write(xml)
        env['wsgi.input'].seek(0, 0)
        return env

    def xhandle(self, method, *params):
        env = self.xenv(method, params)
        app = web.makeApp()
        return self.handle(app, env)

    @with_ini(INI % 'SM')
    def test01_BadMethod(self):
        env = self.xenv('h1.add', (20, 22))
        env['REQUEST_METHOD'] = 'GET'
        app = web.makeApp()
        s,h,b = self.handle(app, env)
        self.assertEqual(s, '405 Method Not Allowed')

    @with_ini(INI % 'SM')
    def test02_BadContentType(self):
        env = self.xenv('h1.add', (20, 22))
        env['CONTENT_TYPE'] = 'text/plain'
        app = web.makeApp()
        s,h,b = self.handle(app, env)
        self.assertEqual(s, '405 Method Not Allowed')

    @with_ini(INI % 'SM')
    def test03_CorruptedData(self):
        env = self.xenv('h1.add', (20, 22))
        env['wsgi.input'].seek(0, 0)
        env['wsgi.input'].write('>>>>>>>>>>>>>>>>>')
        env['wsgi.input'].seek(0, 0)
        app = web.makeApp()
        s,h,b = self.handle(app, env)
        self.assertEqual(s, '400 Bad Request')

    @with_ini(INI % 'SM')
    def test10_Add(self):
        s,h,b = self.xhandle('h1.add', 20, 22)
        self.assertEqual(s, '200 OK')
        self.assertEqual(sorted(h), [
            ('Content-Length', str(len(b))),
            ('Content-Type', 'text/xml'),
        ])
        r = xmlrpclib.loads(b)[0]
        self.assertEqual(r, (42,))

    @with_ini(INI % 'SM')
    def test11_Echo(self):
        s,h,b = self.xhandle('h1.echo', u'čeština')
        self.assertEqual(s, '200 OK')
        r = xmlrpclib.loads(b)[0]
        self.assertEqual(r, (u'echo:čeština',))

    @with_ini(INI % 'SM')
    def test20_Fault(self):
        s,h,b = self.xhandle('h1.add', 20)
        self.assertEqual(s, '200 OK')
        self.assertEqual(sorted(h), [
            ('Content-Length', str(len(b))),
            ('Content-Type', 'text/xml'),
        ])
        try:
            xmlrpclib.loads(b)
        except xmlrpclib.Fault, e:
            e = str(e)
            self.failUnless(
                'add() takes exactly 3 arguments (2 given)' in e
            )
        else:
            self.fail('exception not raised')

    @with_ini(INI % 'SM')
    def test30_MethodNotFound(self):
        s,h,b = self.xhandle('bar', 20)
        self.assertEqual(s, '200 OK')
        self.assertEqual(sorted(h), [
            ('Content-Length', str(len(b))),
            ('Content-Type', 'text/xml'),
        ])
        try:
            xmlrpclib.loads(b)
        except xmlrpclib.Fault, e:
            e = str(e)
            self.failUnless('wm4.web.rpc.RpcError' in e)
            self.failUnless('method not found' in e)
            self.failUnless('bar' in e)
        else:
            self.fail('exception not raised')

    @with_ini(INI % 'SM')
    def test31_NonRpcMethod(self):
        s,h,b = self.xhandle('h1.nonrpc', 20)
        self.assertEqual(s, '200 OK')
        self.assertEqual(sorted(h), [
            ('Content-Length', str(len(b))),
            ('Content-Type', 'text/xml'),
        ])
        try:
            xmlrpclib.loads(b)
        except xmlrpclib.Fault, e:
            e = str(e)
            self.failUnless('wm4.web.rpc.RpcError' in e)
            self.failUnless('private method' in e)
            self.failUnless('h1.nonrpc' in e)
        else:
            self.fail('exception not raised')

    @with_ini(INI % 'SM')
    def test35_Permission(self):
        s,h,b = self.xhandle('double', 20)
        self.assertEqual(s, '403 Forbidden')
        s,h,b = self.xhandle('h1.guarded')
        self.assertEqual(s, '403 Forbidden')
        s,h,b = self.xhandle('h2.echo')
        self.assertEqual(s, '403 Forbidden')

    @with_ini(INI % 'SMG')
    def test36_ClassPermission(self):
        s,h,b = self.xhandle('echo', 20)
        self.assertEqual(s, '403 Forbidden')

def suite():
    return TestSuite((
        makeSuite(TestRpcDispatcher),
        makeSuite(TestXmlRpc),
    ))


if __name__ == '__main__':
    main(defaultTest='suite')
