# -*- coding: utf-8 -*-

from unittest import *
from wm4.api import *
from wm4.web.sessions import *
import time, os, cPickle

from utils import *



class TestTrustedClientData(WebCase):

    def test10_Verified(self):
        ts = 123123123.456
        ts = time.time()
        p = SessionID(
            timestamp = ts, rnd=2**16-1,
            secret='passwd', remote_addr='1.2.3.4', browser='Netscape/1.0',
        ).pack()
        b = SessionID.unpack(p,
            secret='passwd', remote_addr='1.2.3.4', browser='Netscape/1.0',
        )
        self.assertEqual(b.timestamp, ts)
        self.assertEqual(b.rnd, 2**16-1)
        self.assertEqual(b.secret, 'passwd')
        self.assertEqual(b.remote_addr, '1.2.3.4')
        self.assertEqual(b.browser, 'Netscape/1.0')

    def test12_VerifiedWithDefaultValues(self):
        p = SessionID(
            timestamp = 123123123.456, rnd=44123,
            secret='passwd',
        ).pack()
        b = SessionID.unpack(p,
            secret='passwd',
        )
        self.assertEqual(b.timestamp, 123123123.456)
        self.assertEqual(b.rnd, 44123)
        self.assertEqual(b.secret, 'passwd')

    def test20_DiggestAttrChange(self):
        p = SessionID(
            timestamp = 123123123.456, rnd=42,
            secret='passwd', remote_addr='1.1.1.1', browser='Netscape/1.0',
        ).pack()
        self.assertRaises(
            TrustedClientDataError,
            SessionID.unpack, p,
            secret='passwd', remote_addr='1.1.1.4', browser='Netscape/1.0',
        )

    def test22_DataChange(self):
        p = SessionID(
            timestamp = 123123123.456, rnd=42,
            secret='passwd', remote_addr='1.1.1.1', browser='Netscape/1.0',
        ).pack()
        p = 'x' + p[1:]
        self.assertRaises(
            TrustedClientDataError,
            SessionID.unpack, p,
            secret='passwd', remote_addr='1.1.1.1', browser='Netscape/1.0',
        )

    def test30_Strings(self):
        class X(TrustedClientData):
            attributes = [('s', 'S'), ('u', 'U'), ('e', 'i')]
        x = X(s='čeština', u=u'čeština', e=42)
        y = X.unpack(x.pack())
        assert x.s == y.s == 'čeština'
        assert x.u == y.u == u'čeština'
        assert x.e == y.e == 42

    def test31_Bytes(self):
        class X(TrustedClientData):
            attributes = [('u', 'U'), ('b', 'b'), ('B', 'B')]
        x = X(u=u'čeština', b=-100, B=100)
        y = X.unpack(x.pack())
        assert x.u == y.u == u'čeština'
        assert x.b == y.b == -100
        assert x.B == y.B == 100





_time = int(time.time())


class TSessionMiddleware(SessionMiddleware):

    secret = 'TESTPWD'

    def time(self):
        global _time
        _time += 1
        return _time


class TestSessionMiddleware(WebCase):

    sdir = '/tmp/ts'

    def setUp(self):
        WebCase.setUp(self)
        self.parse_ini((
            '[WSGI Middleware session]\n'
            'factory = "%s.TSessionMiddleware"\n'
            'session_dir = "%s"\n'
        ) % (__name__, self.sdir))

    def tearDown(self):
        WebCase.tearDown(self)
        if not os.path.isdir(self.sdir):
            return
        for f in os.listdir(self.sdir):
            f = os.path.join(self.sdir, f)
            os.unlink(f)
        os.rmdir(self.sdir)

    def sessapp(self, app):
        app = self.mkapp(
            'environment', 'errors', 'transaction', 'cookies',
            'http_exceptions', 'session', funcMiddleware(app),
        )
        return app

    def mkid(self):
        browser = 'no-browser/1.0'
        remote = '127.0.0.42'
        return SessionID(
                timestamp = _time,
                rnd = os.getpid(),
                remote_addr = remote,
                browser = browser,
                secret = TSessionMiddleware.secret,
        )

    def test10_NewUnmodifiedSessionIsNotStored(self):
        _sess = []
        def new_session(env, start):
            sess = getSession(env)
            start('200 OK', [])
            _sess.append(sess)
            return ['OK']
        app = self.sessapp(new_session)
        s,h,b = self.handle(app)
        sess = _sess[0]
        id = str(sess.id)
        self.assertEqual(s, '200 OK')
        self.assertEqual(len(h), 1)
        self.assertEqual(
            strip_semicol(h)[0],
            ('Set-Cookie', 'wm4sess=%s; Path=/' % id)
        )
        assert os.path.isfile(os.path.join(self.sdir, id))
        t = sess.id.timestamp
        self.assertEqual(sess.ctime, t)
        self.assertEqual(sess.atime, t+1)
        self.assertEqual(sess.data, {})

    def test12_NewModifiedSessionIsNotStored(self):
        _sess = []
        def new_session(env, start):
            sess = getSession(env)
            sess['foo'] = 42
            start('200 OK', [])
            _sess.append(sess)
            return ['OK']
        app = self.sessapp(new_session)
        s,h,b = self.handle(app)
        sess = _sess[0]
        id = str(sess.id)
        self.assertEqual(s, '200 OK')
        self.assertEqual(len(h), 1)
        self.assertEqual(
            strip_semicol(h)[0],
            ('Set-Cookie', 'wm4sess=%s; Path=/' % id)
        )
        assert os.path.isfile(os.path.join(self.sdir, id))
        t = sess.id.timestamp
        self.assertEqual(sess.ctime, t)
        self.assertEqual(sess.atime, t+1)
        self.assertEqual(sess.data, {'foo':42})

    def test16_LoadNonexistingSession(self):
        id = self.mkid()

        _sess = []
        def get_session(env, start):
            sess = getSession(env)
            start('200 OK', [])
            _sess.append(sess)
            return [str(sess.get('foo','bar'))]

        e = self.env(
            cookies={'wm4sess': str(id)},
            env={'REMOTE_ADDR':id.remote_addr, 'HTTP_USER_AGENT':id.browser}
        )
        s,h,b = self.handle(self.sessapp(get_session), e)

        sess = _sess[0]
        self.assertEqual(s, '200 OK')
        # cookie is valid so we don't set it again
        self.assertEqual(len(h), 0)
        self.assertEqual(b, 'bar')

        assert os.path.isfile(os.path.join(self.sdir, str(id)))
        t = id.timestamp
        self.assertEqual(sess.ctime, t)
        self.assertEqual(sess.atime, t+1)
        self.assertEqual(sess.data, {})

    def test18_LoadInvalidSessionID(self):
        id = self.mkid()

        _sess = []
        def get_session(env, start):
            sess = getSession(env)
            start('200 OK', [])
            _sess.append(sess)
            return [str(sess.get('foo','bar'))]

        e = self.env(
            cookies={'wm4sess': str(id)},
            env={'REMOTE_ADDR':id.remote_addr, 'HTTP_USER_AGENT':'ie'}
        )
        s,h,b = self.handle(self.sessapp(get_session), e)

        sess = _sess[0]
        self.assertNotEqual(sess.id.timestamp, id.timestamp)
        self.assertEqual(s, '200 OK')
        # cookie was invalid so we must set it
        self.assertEqual(len(h), 1)
        self.assertEqual(b, 'bar')

        assert os.path.isfile(os.path.join(self.sdir, str(sess.id)))
        t = sess.id.timestamp
        self.assertEqual(sess.ctime, t)
        self.assertEqual(sess.atime, t+1)
        self.assertEqual(sess.data, {})

    def test20_UpdateSessionExisting(self):
        _id = []
        def set_session(env, start):
            sess = getSession(env)
            sess['foo'] = 42
            _id.append(sess.id)
            start('200 OK', [])
            return []
        self.handle(self.sessapp(set_session))
        id = _id[0]

        assert os.path.isfile(os.path.join(self.sdir, str(id)))

        _sess = []
        def get_session(env, start):
            sess = getSession(env)
            _sess.append(sess)
            start('200 OK', [])
            return [str(sess.get('foo'))]

        s,h,b = self.handle(
            self.sessapp(get_session),
            self.env(cookies={'wm4sess': str(id)})
        )

        sess = _sess[0]
        t = id.timestamp
        self.assertEqual(sess.id.timestamp, t)
        self.assertEqual(sess.ctime, t)
        self.assertEqual(sess.atime, t+2)
        self.assertEqual(b, '42')




class TestSessionManager(WebCase):

    def test10_CreateSession(self):
        sess = self.mgr.create_session(self._req)
        self.assertEqual(sess.id.timestamp, self.mgr._time)
        self.assertEqual(sess.ctime, self.mgr._time)
        self.assertEqual(sess.atime, self.mgr._time)

    def test20_LoadSession(self):
        sess = self.mgr.create_session(self._req)
        sess['foo'] = 42
        t = sess.ctime
        self.mgr.store.save_session(self.mgr, sess)

        sess = self.mgr.load_session(self._req, sess.id)
        self.assertEqual(sess.id.timestamp, t)
        self.assertEqual(sess.ctime, t)
        self.assertEqual(sess.atime, t+1)
        self.assertEqual(sess.data, {'foo':42})

    def test30_ModifySession(self):
        sess = self.mgr.create_session(self._req)
        sess['foo'] = 42
        t = sess.ctime
        self.mgr.store.save_session(self.mgr, sess)

        sess = self.mgr.load_session(self._req, sess.id)
        sess['bar'] = 'BAR'
        self.mgr.save_session(sess)

        self.assertEqual(sess.id.timestamp, t)
        self.assertEqual(sess.ctime, t)
        self.assertEqual(sess.atime, t+1)
        self.assertEqual(sess.data, {'foo':42, 'bar':'BAR'})

        sess = self.mgr.load_session(self._req, sess.id)

        self.assertEqual(sess.id.timestamp, t)
        self.assertEqual(sess.ctime, t)
        self.assertEqual(sess.atime, t+3)
        self.assertEqual(sess.data, {'foo':42, 'bar':'BAR'})




def suite():
    return TestSuite((
        makeSuite(TestTrustedClientData),
        makeSuite(TestSessionMiddleware),
    ))


if __name__ == '__main__':
    main(defaultTest='suite')
