from utils import escape, quoteattr, attr_list
from const import *

__all__ = [
    'Node', 'Fragment', 'DocType', 'XmlDecl', 'ProcessingInstruction',
    'Text', 'Literal', 'Comment', 'Element', 'Document', 'ElementFactory',
    'xml',
]


class Node(object):

    __slots__ = ()

    def expand(self, state):
        raise NotImplementedError

    def __len__(self):
        return 0

    def __getstate__(self):
        state = {}
        for a in self.__class__.__slots__:
            state[a] = getattr(self, a)
        return state

    def __setstate__(self, state):
        for a in self.__class__.__slots__:
            setattr(self, a, state[a])



class Fragment(Node):

    __slots__ = ('_children',)

    def __init__(self, content):
        self._children = []
        self.extend(content)

    def wrap(self, node):
        if isinstance(node, Node):
            return node
        else:
            return Text(node)

    def __len__(self):
        return len(self._children)

    def __iter__(self):
        return iter(self._children)

    def __getitem__(self, i):
        return self._children[i]

    def __setitem__(self, i, node):
        self._children[i] = self.wrap(node)

    def __delitem__(self, i):
        del self._children[i]

    def append(self, node):
        self._children.append(self.wrap(node))

    def insert(self, i, node):
        self._children.insert(i, self.wrap(node))

    def remove(self, node):
        self._children.remove(node)

    def extend(self, nodes):
        wrap = self.wrap
        for node in nodes:
            self._children.append(wrap(node))

    def replace(self, old, new):
        i = self._children.index(old)
        self._children[i] = self.wrap(new)

    def expand(self, state=None):
        for elem in self:
            for e in elem.expand(state):
                yield e

    def __getstate__(self):
        return {'_children' : self[:]}

    def __setstate__(self, state):
        self._children = []
        self.extend(state['_children'])
    def __unicode__(self):
        from serialization import xml_serializer
        return u''.join(xml_serializer(self.expand(None)))

    def __str__(self):
        return str(unicode(self))



class XmlDecl(Node):

    __slots__ = ('version', 'encoding', 'standalone')

    def __init__(self, version, encoding, standalone):
        self.version = version
        self.encoding = encoding
        self.standalone = standalone

    def expand(self, state):
        yield XML_DECL, (self.version, self.encoding, self.standalone)


class DocType(Node):

    __slots__ = ('name', 'sysid', 'pubid', 'has_internal_subset')

    def __init__(self, name, sysid, pubid, has_internal_subset=0):
        self.name = name
        self.sysid = sysid
        self.pubid = pubid
        self.has_internal_subset = has_internal_subset

    def expand(self, state):
        yield DOC_TYPE, (
            self.name, self.sysid, self.pubid, self.has_internal_subset
        )

    def __str__(self):
        return '<!DOCTYPE %s PUBLIC %s %s>' % (
            self.name, quoteattr(self.pubid), quoteattr(self.sysid)
        )

class ProcessingInstruction(Node):

    __slots__ = ('target', 'data')

    def __init__(self, target, data):
        self.target = target
        self.data = data

    def expand(self, state):
        yield PI, (self.target, self.data)


class Text(Node, unicode):

    __slots__ = ()

    def expand(self, state):
        yield TEXT, unicode(self)


class Literal(Node, unicode):

    __slots__ = ()

    def expand(self, state):
        yield LITERAL, unicode(self)


class Comment(Literal):

    __slots__ = ()

    def expand(self, state):
        yield COMMENT, unicode(self)



class Element(Fragment):

    __slots__ = ('tag', '_attrs')

    def __init__(self, tag, attrs, content):
        self.tag = unicode(tag)
        self._attrs = attr_list(attrs)
        self._children = []
        self.extend(content)

    def get(self, attr, default=None):
        return self._attrs.get(attr, default)

    def set(self, attr, value):
        self._attrs.set(attr, value)

    def setdefault(self, attr, value):
        self._attrs.setdefault(attr, value)

    def hasattr(self, attr):
        return self._attrs.hasattr(attr)

    def pop(self, attr, default=None):
        return self._attrs.pop(attr, default)

    def items(self):
        return iter(self._attrs)

    def expand(self, state=None):
        tag = self.tag
        attrs = self._attrs
        yield START_TAG, (tag, attrs)
        for elem in self:
            for e in elem.expand(state):
                yield e
        yield END_TAG, (tag, attrs)

    def __getstate__(self):
        return {
            'tag'       : self.tag,
            '_attrs'    : list(self.items()),
            '_children' : self[:],
        }

    def __setstate__(self, state):
        self.tag = state['tag']
        self._attrs = attr_list(state['_attrs'])
        self._children = []
        self.extend(state['_children'])


class Document(Fragment):

    __slots__ = ()




class ElementFactory:

    __slots__ = ('tag', 'factory')

    def __init__(self,
        tag, elem_factory=Element, doc_factory=Document, attr_map={}
    ):
        self.tag = tag
        self.elem_factory = elem_factory
        self.doc_factory = doc_factory
        self.attr_map = attr_map

    def __call__(self, attrd={}, *content, **attrkw):
        if isinstance(attrd, dict):
            attr_iterators = [attrd.items(), attrkw.items()]
        elif isinstance(attrd, attr_list):
            attr_iterators = [attrd, attrkw.items()]
        elif isinstance(attrd, (list,tuple)):
            attr_iterators = [attrd, attrkw.items()]
        else:
            attr_iterators = [attrkw.items()]
            content = (attrd,) + content
        amap = self.attr_map.get
        attrs = [(amap(a,a),v) for ai in attr_iterators for a,v in ai]
        tag = self.tag
        if tag is None:
            assert not attrs
            return self.doc_factory(content)
        else:
            return self.elem_factory(self.tag, attrs, content)

    def __getattr__(self, attr):
        return self.__class__(attr, self.elem_factory)

xml = ElementFactory(None)
