# -*- coding: iso-8859-2 -*-
from xml.parsers import expat
from itertools import izip

from elements import *


__all__ = ['SimpleXMLParser', 'XMLParser']


class SimpleXMLParser(object):
    '''Simple XML parser withou namespace processing'''

    expat_params = (
        'buffer_text', 'returns_unicode',
        'ordered_attributes', 'specified_attributes',
        'namespace_prefixes',
    )

    buffer_text = True
    returns_unicode = True
    ordered_attributes = True
    specified_attributes = True
    namespace_prefixes = False
    namespace_separator = None

    url = '<xml source>'

    def __init__(self, parser_params={}):
        self.__dict__.update(parser_params)

    def parseFile(self, filename, encoding=None):
        self.url = filename
        parser = self.createParser(encoding)
        parser.ParseFile(open(filename))
        return self.Finish()

    def parse(self, source, encoding=None):
        self.url = '<xml source>'
        parser = self.createParser(encoding)
        parser.Parse(source, True)
        return self.Finish()

    def createParser(self, encoding=None):
        #from peak.util import pyexpat as expat
        self.parser = parser = expat.ParserCreate(
            encoding, self.namespace_separator
        )
        self.setupData()
        self.setupParams()
        self.setupCallbacks()
        return parser

    def setupData(self):
        data = dict(children=[])
        self.stack = [data]
        return data

    def setupParams(self):
        parser = self.parser
        for p in self.expat_params:
            setattr(parser, p, getattr(self,p))

    def setupCallbacks(self):
        parser = self.parser
        parser.XmlDeclHandler = self.XmlDecl
        parser.StartDoctypeDeclHandler = self.StartDoctype
        parser.StartElementHandler = self.StartElement
        parser.EndElementHandler = self.EndElement
        parser.CharacterDataHandler = self.CharacterData
        parser.DefaultHandler = self.Default
        parser.CommentHandler = self.Comment
        parser.ProcessingInstructionHandler = self.ProcessingInstruction

    def pushData(self, tag, raw_attrs):
        ai = iter(raw_attrs)
        data = dict(children=[], tag=tag, attrs=[(a,v) for a,v in izip(ai,ai)])
        self.stack.append(data)
        return data

    def popData(self):
        return self.stack.pop()

    def Finish(self):
        stack = self.stack
        assert len(stack) == 1
        return Document(stack[-1]['children'])

    def Omit(self, *a):
        pass

    def XmlDecl(self, version, encoding, standalone):
        self.stack[-1]['children'].append(XmlDecl(
            version, encoding, standalone
        ))

    def StartDoctype(self, name, sysid, pubid, has_internal_subset):
        self.stack[-1]['children'].append(DocType(
            name, sysid, pubid, has_internal_subset
        ))

    def ProcessingInstruction(self, target, data):
        self.stack[-1]['children'].append(ProcessingInstruction(target,data))

    def err(self, msg, kind=SyntaxError, lineno=None, filename=None):
        """Generate an error with current parse location"""
        p = self.parser
        e = kind()
        e.msg = msg
        if filename is None:
            e.filename = self.url
        else:
            e.filename = filename
        e.text = p.GetInputContext().rstrip('\0x00')
        if lineno is None:
            e.lineno = p.CurrentLineNumber
        else:
            e.lineno = lineno
        e.offset = p.CurrentColumnNumber
        e.args = msg, (e.filename,e.lineno,e.offset,e.text)
        raise e

    def validateAttrs(self, data, required=(), optional=(), one_of=()):
        attrs = dict(data['attrs'])
        all = set(required+optional+one_of)
        for a in attrs:
            if a not in all:
                self.err('Invalid attribute %r for tag %r' % (a,data['tag']))
        for a in required:
            if not a in attrs:
                self.err('Missing %r attribute in tag %r' % (a,data['tag']))
        if one_of and len([a for a in attrs if a in one_of]) != 1:
            self.err('Missing one of attributes %s' % repr(one_of))

    def validateParent(self, data, *tags):
        parent = self.stack[-2].get('tag')
        for tag in tags:
            if tag == parent:
                return parent
        self.err('Invalid parent %r element for element %r. Required: %r' % (
            parent, data['tag'], tags
        ))

    def StartElement(self, tag, raw_attrs):
        self.pushData(tag, raw_attrs)

    def EndElement(self, tag):
        data = self.popData()
        self.stack[-1]['children'].append(
            Element(data['tag'], data['attrs'], data['children'])
        )

    def CharacterData(self, text):
        self.stack[-1]['children'].append(Text(text))

    def Default(self, text):
        self.stack[-1]['children'].append(Literal(text))

    def Comment(self, text):
        self.stack[-1]['children'].append(Comment(text))




class XMLParser(SimpleXMLParser):
    '''Namespace aware XML parser'''

    def __init__(self, parser_params={}, known_namespaces={}):
        super(XMLParser,self).__init__(parser_params)
        self.known_namespaces = known_namespaces

    def setupData(self):
        data = super(XMLParser,self).setupData()
        self.namespaces = dict([
            (p,[n]) for p,n in self.known_namespaces.iteritems()
        ])
        return data

    def pushData(self, tag, raw_attrs):
        data = super(XMLParser,self).pushData(tag, raw_attrs)
        namespaces = self.namespaces
        data['ns_decl'] = ns_decl = {}
        ns_attrs = []
        for attr, value in data['attrs']:
            if ':' in attr:
                pref, name = attr.split(':',1)
                if pref == 'xmlns':
                    ns_decl[name] = value
                    namespaces.setdefault(name, []).append(value)
                ns_attrs.append((attr, pref, name, value))
            elif attr == 'xmlns':
                ns_decl[''] = value
                namespaces.setdefault('', []).append(value)
                ns_attrs.append((attr, '', attr, value))
            else:
                ns_attrs.append((attr, '', attr, value))
        if ':' in tag:
            pref, tag = tag.split(':',1)
        else:
            pref = ''
        if pref in namespaces:
            ns = namespaces[pref][-1]
        else:
            ns = None
        data['ns_tag'] = ns, tag
        data['ns_attrs'] = [
            (q, p in namespaces and namespaces[p][-1] or None, a, v)
            for q,p,a,v in ns_attrs
        ]
        return data

    def popData(self):
        data = super(XMLParser,self).popData()
        namespaces = self.namespaces
        for pref in data['ns_decl']:
            namespaces[pref].pop()
        return data
