from utils import escape, quoteattr
from const import *
from lib4.text import ufTemplate

__all__ = ['XMLSerializer', 'PrettyPrinter', 'pretty_print', 'XHTMLSerializer']


class ALLOW_EMPTY:
    def __contains__(self, key):
        return True
ALLOW_EMPTY = ALLOW_EMPTY()


class XMLSerializer(object):
    '''Serializer for XML elements

       It expects unicode data yielded by expand() element method.
    '''

    detectEmptyTags = True

    def __init__(self, emptyEnabled=ALLOW_EMPTY, keepXmlDecl=True):
        self.emptyEnabled = emptyEnabled
        self.keepXmlDecl = keepXmlDecl

    def __call__(self, stream):
        mmap = {
            XML_DECL    : self.xml_decl,
            DOC_TYPE    : self.doc_type,
            PI          : self.pi,
            TEXT        : self.text,
            LITERAL     : self.literal,
            DEFER       : self.defer,
            COMMENT     : self.comment,
            START_TAG   : self.start,
            END_TAG     : self.end,
        }

        data = ufTemplate()
        write = data.extend

        if self.detectEmptyTags:
            opened = None
            emptyEnabled = self.emptyEnabled
            for e,a in stream:
                meth = mmap[e]
                if e is START_TAG:
                    if opened:
                        write(self.start(opened))
                    opened = a
                elif e is END_TAG:
                    if opened:
                        if a[0] in emptyEnabled:
                            write(self.empty(a))
                        else:
                            write(self.start(opened))
                            write(self.end(a))
                    else:
                        write(meth(a))
                    opened = None
                else:
                    u = meth(a)
                    if not u:
                        continue
                    if opened:
                        write(self.start(opened))
                        opened = None
                    write(u)
            assert opened is None
        else:
            for e,a in stream:
                write(mmap[e](a))
        return data()

    def attrs(self, attrs):
        alist = []
        for a,v in attrs:
            alist.append(u' %s=' % a)
            if isinstance(v,basestring):
                alist.append(quoteattr(v))
            else:
                alist.append(v)
        return alist

    def start(self, (tag, attrs)):
        if attrs:
            return [u'<%s' % tag] + self.attrs(attrs) + [u'>']
        else:
            return [u'<%s>' % tag]

    def end(self, (tag, attrs)):
        return [u'</%s>' % tag]

    def empty(self, (tag, attrs)):
        if attrs:
            return [u'<%s' % tag] + self.attrs(attrs) + [u' />']
        else:
            return [u'<%s />' % tag]

    def text(self, txt):
        if isinstance(txt, basestring) and not txt:
            return []
        else:
            return [escape(unicode(txt))]

    def literal(self, txt):
        if isinstance(txt, basestring) and not txt:
            return []
        else:
            return [unicode(txt)]

    def defer(self, cb):
        return [cb]

    def xml_decl(self, (version, encoding, standalone)):
        out = []
        if self.keepXmlDecl:
            out.append(u'<?xml version=%s' % quoteattr(version))
            if encoding:
                out.append(u' encoding=' + quoteattr(encoding))
            if standalone:
                out.append(u' standalone="yes"')
            out.append(u'?>')
        return out

    def doc_type(self, (name, sysid, pubid, has_internal_subset)):
        assert has_internal_subset == 0 # not implemented
        return [u'<!DOCTYPE %s PUBLIC %s %s>' % (
            name, quoteattr(pubid), quoteattr(sysid)
        )]

    def pi(self, (target, data)):
        return [u'<?%s %s?>' % (target, data)]

    def comment(self, txt):
        return [u'<!-- %s -->' % txt]


xml_serializer = XMLSerializer()



class PrettyPrinter(XMLSerializer):

    def __init__(self, emptyEnabled=ALLOW_EMPTY, keepXmlDecl=True):
        XMLSerializer.__init__(self, emptyEnabled, keepXmlDecl)
        self.level = 0

    def start(self, arg):
        stream = XMLSerializer.start(self, arg)
        stream.insert(0, '\n' + '  ' * self.level)
        self.level += 1
        return stream

    def end(self, arg):
        self.level -= 1
        stream = XMLSerializer.end(self, arg)
        stream.insert(0, '\n' + '  ' * self.level)
        return stream

    def empty(self, arg):
        return ['\n' + '  ' * self.level] + XMLSerializer.empty(self, arg)

pretty_print = PrettyPrinter()






EMPTYENABLED = set(('area','br','hr','input','img','link','meta','option'))

class XHTMLSerializer(XMLSerializer):

    def __init__(self, emptyEnabled=EMPTYENABLED, keepXmlDecl=False):
        super(XHTMLSerializer, self).__init__(emptyEnabled, keepXmlDecl)

xhtml_serializer = XHTMLSerializer()







NOCLOSE = dict.fromkeys((
    'area', 'base', 'basefont', 'br', 'col', 'frame', 'hr', 'img',
    'input', 'isindex', 'link', 'meta', 'param'
))

NOESCAPE = dict.fromkeys(('style', 'script'))

BOOL_ATTRS = dict.fromkeys(('selected', 'checked', 'readonly', 'XXX'))

class HTML4Serializer(XMLSerializer):

    def __init__(self, emptyEnabled={}, keepXmlDecl=False):
        super(XHTMLSerializer, self).__init__(emptyEnabled, keepXmlDecl)

    def empty(self, tag, attrs):
        if attrs:
            return u'<%s%s>' % (tag, self.attrs(attrs))
        else:
            return u'<%s>' % tag
