#-*- coding: iso-8859-2 -*-
from unittest import *
from xml4 import templates, const
from xml4.serialization import XHTMLSerializer
from xml4 import xhtml as xh
from xml4.definitions import *
from lib4.imports import importString
import sys, os


MYNS = 'http://foo.com'

THIS_DIR = os.path.abspath(os.path.dirname(__file__))

def findDomlet(state, ns, name):
    fact = state.get((ns,name))
    if fact is None:
        fact = importString(name)
    return fact


class TestBase(TestCase):

    def parse(self, src, **tmpl_namespace):
        parser = templates.TemplateParser(
            known_namespaces={'my':MYNS, 'py':NS_XML4_TMPL},
            domlet_namespaces={MYNS:findDomlet},
            tmpl_namespace=tmpl_namespace
        )
        return parser.parse(src)

    def parseFile(self, filename, **tmpl_namespace):
        parser = templates.TemplateParser(tmpl_namespace=tmpl_namespace)
        return parser.parseFile(filename)

    def serialize(self, tmpl, **locals):
        state = templates.State(ldict=locals)
        return u''.join(
            XHTMLSerializer()(tmpl.expand(state))
        )

    def check(self, src, result, tmpl_namespace={}, **locals):
        tmpl = self.parse(src, **tmpl_namespace)
        res = self.serialize(tmpl, **locals)
        self.assertEqual(res, result)



class TestTemplateParsing(TestBase):

    def test01_Static(self):
        self.check('<html />', '<html></html>')

    def test02_DoctypePreserving(self):
        s = ('<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" '
                      '"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">\n'
             '<html></html>')
        self.check(s, s)

    def test03_XmlDeclStripping(self):
        s = '<?xml version="1.0"?>\n<html />'
        self.check(s, '\n<html></html>')

    def test04_InvalidConstructs(self):
        self.assertRaises(SyntaxError,self.check,'<p py:xx=""/>', '')

    def test10_DomletNamespaces(self):
        src = '<div><p my:%s.Signature="12"/></div>' % __name__
        self.check(src, '<div>12,42</div>')






class TestProcessingInstructions(TestBase):

    def test01_ProcessingInstruction(self):
        self.check('<?python x="X"?><p>$x</p>', '<p>X</p>')

    def test04_MultilineProcessingInstruction(self):
        import sys
        self.check(
            '<?py \n'
            '  import sys\n'
            '  x=len(sys.path)\n'
            '?>'
            '<p>$x</p>',
            '<p>%d</p>' % len(sys.path))

    def test10_TemplateLevelPI(self):
        t = self.parse(
            '<?py \n'
            '   from itertools import count\n'
            '   counter = count().next\n'
            '   tc = counter()\n'
            '?>'
            '<doc>'
              '<?py lc = counter() ?>'
              '${tc}:${lc}'
            '</doc>'
        )
        # Template tmpl_namespace contains 'count', 'counter' and 'tc'.
        #self.assertEqual(len(t.tmpl_namespace), 3, repr(t.tmpl_namespace))
        old_counter = t.tmpl_namespace['counter']
        # Template level PI is ignored
        self.assertEqual(len(t), 1)
        # Local level PI is not
        self.assertEqual(len(t[0]), 4)
        # tc is computed only once during template parsing
        r1 = self.serialize(t)
        self.assertEqual(r1, '<doc>0:1</doc>')
        r2 = self.serialize(t)
        self.assertEqual(r2, '<doc>0:2</doc>')
        # Template level object are not recreated on serialize
        assert t.tmpl_namespace['counter'] is old_counter
        old_counter()
        self.assertEqual(self.serialize(t), '<doc>0:4</doc>')

    def test11_LocalLevelVariables(self):
        t = self.parse(
            '<?py tv = 1 ?>'
            '<doc>'
              '${tv}'
              '<?py tv = 2 ?>'
              '<p>'
                '<?py tv = 3 ?>'
                '${tv}'
                '<?py tv = 4 ?>'
                '${tv}'
              '</p>'
              '${tv}'
            '</doc>'
        )
        self.assertEqual(self.serialize(t), '<doc>1<p>34</p>2</doc>')
        self.assertEqual(self.serialize(t), '<doc>1<p>34</p>2</doc>')
        self.assertEqual(t.tmpl_namespace['tv'], 1) # Template level 'tv'

    def test20_Indentation(self):
        t = self.parse(
            '<?py \n'
            '   def x():\n'
            '       return 42\n'
            '?>'
            '<doc>${x()}</doc>'
        )
        self.assertEqual(self.serialize(t), '<doc>42</doc>')

    def test21_Indentation(self):
        t = self.parse(
            '<?py \n'
            '   def x():\n'
            '       return 42\n'
            '   if True:\n'
            '      q = 1\n'
            '   else:\n'
            '      q = 2\n'
            '?>'
            '<doc>$q</doc>'
        )
        self.assertEqual(self.serialize(t), '<doc>1</doc>')

    def test30_GlobalVars(self):
        self.check(
            '<?py d=4?><doc><?py def x(a): return a*d?>${x(3)}</doc>',
            '<doc>12</doc>'
        )

    def test31_FuncDefaults(self):
        self.check(
          '<d><?py d=4\ndef x(a=d): return a?>${x(2)},${x()}</d>',
          '<d>2,4</d>'
        )

    #def test35_LocalVars(self):
    #    self.check(
    #        '<doc><?py d=4?><p><?py def x(a): return a*d?>${x(3)}</p></doc>',
    #        '<doc><p>12</p></doc>'
    #    )





class TestInterpolation(TestBase):

    def test10_LongInterpolation(self):
        self.check('<p>b${x}e</p>', '<p>bXe</p>', x='X')

    def test11_LongInterpolation(self):
        self.check('<p>b${21*2}e</p>', '<p>b42e</p>')

    def test12_NoneSubstitution(self):
        self.check('<p>b${None}e</p>', '<p>be</p>')

    def test13_ShortInterpolation(self):
        self.check('<p>b $x a</p>', '<p>b X a</p>', x='X')

    def test14_ShortInterpolation(self):
        class X: pass
        x = X(); x.y = X(); x.y.z = '42'
        self.check('<p>b $x.y.z a</p>', '<p>b 42 a</p>', x=x)

    def test16_InterpolationEscaping(self):
        self.check('<p>$x</p>', '<p>&lt;X&amp;&gt;"</p>', x='<X&>"')

    def test17_SubstitutingXML(self):
        self.check('<div>$x</div>','<div><p>&nbsp;</p></div>',x=xh.p(xh.nbsp))

    def test20_StaticAttributes(self):
        self.check('<p x="1" z="3" y="42" />', '<p x="1" z="3" y="42"></p>')

    def test21_StaticAttributes(self):
        self.check(
            '<p x="1" z="3"><?python x=22?><q y="42" /></p>',
            '<p x="1" z="3"><q y="42"></q></p>'
        )

    def test22_AttributeInterpolation(self):
        self.check('<a href="${21*2}"/>', '<a href="42"></a>')
        self.check('<a href="${x}"/>', '<a></a>', x=None)
        self.check('<a href="${x}"/>', '<a href="42"></a>', x=42)
        self.check('<a href="b${x}e"/>', '<a href="be"></a>', x=None)

    def test23_DynamicAttributes(self):
        self.check('<a py:attrs="dict(href=21*2)"/>', '<a href="42"></a>')
        self.check('<a py:attrs="dict(href=None)"/>', '<a></a>')
        self.check('<a py:attrs="dict(a=1, b=2)"/>', '<a a="1" b="2"></a>')

    def test24_StaticAttributeReplacing(self):
        self.check('<a x="1" py:attrs="dict(x=2)"/>', '<a x="2"></a>')
        self.check('<a x="1" py:attrs="dict(x=None)"/>', '<a></a>')
        self.check('<a x="1" py:attrs="dict(x=None)"/>', '<a></a>')

    def test25_AttributeEscaping(self):
        self.check('<p x="$x"/>', '<p x="&lt;X&amp;&quot;"></p>', x='<X&"')
        self.check('<p x="i${x}i"/>', '<p x="i&lt;X&amp;&quot;i"></p>',x='<X&"')
        self.check(
            '<p py:attrs="dict(x=x)"/>','<p x="&lt;&amp;&quot;"></p>',x='<&"')

    def test26_AttributeXMLContent(self):
        m = xh.p()
        self.check('<p x="$x"/>', '<p x="&lt;p&gt;&lt;/p&gt;"></p>', x=m)

    def test27_NoneValueInAttribute(self):
        self.check('<p x="i${None}i"/>', '<p x="ii"></p>')

    def test28_StrippingInterpolatedAttribute(self):
        self.check('<p x="${42}" py:strip=""/>', '')

    def test30_Content(self):
        self.check('<p py:content="12"/>', '<p>12</p>')
        self.check('''<p py:content="''"/>''', '<p></p>')
        self.check('<p py:content="x"/>', '<p>&lt;X&amp;&gt;"</p>', x='<X&>"')

    def test31_Replace(self):
        self.check('<p py:replace="12"/>', '12')
        self.check('''<p py:replace="''"/>''', '')
        self.check('<p py:replace="x"/>', '&lt;X&amp;&gt;"', x='<X&>"')

    def test32_XMLContent(self):
        m = xh.p(xh.nbsp)
        self.check('<div py:replace="x" />','<p>&nbsp;</p>',x=m)
        self.check('<div py:content="x" />','<div><p>&nbsp;</p></div>',x=m)






class TestControllSturctures(TestBase):

    def test10_Strip(self):
        self.check('<p py:strip="">a</p>', 'a')
        self.check('<p py:strip="True">a</p>', 'a')
        self.check('<p py:strip="False">a</p>', '<p>a</p>')

    def test20_If(self):
        self.check('<p py:if="True">a</p>', '<p>a</p>')
        self.check('<p py:strip="" py:if="True">a</p>', 'a')

    def test21_IfNot(self):
        self.check('<p py:if="False">a</p>', '')
        self.check('<p py:strip="" py:if="False">a</p>', '')

    def test30_For(self):
        self.check(
            '<p py:for="i in range(1,4)">${i}</p>',
            '<p>1</p><p>2</p><p>3</p>'
        )

    def test31_For(self):
        self.check(
            '''<p py:for="i,(l,u) in enumerate([('a','A'),('b','B')])">'''
            '''${i}<l py:content="l"/><u py:content="u"/></p>''',
            '<p>0<l>a</l><u>A</u></p>'
            '<p>1<l>b</l><u>B</u></p>'
        )

    def test50_TemplateLevelDef(self):
        t = self.parse(
            '<?py greeting="Hello World"?>'
            '<a>'
              '<b py:def="hello()">'
                 '<c py:replace="greeting"/>'
              '</b>'
              '<d py:replace="hello()"/>'
            '</a>'
        )
        assert 'greeting' in t.tmpl_namespace
        assert 'hello' in t.tmpl_namespace
        self.assertEqual(self.serialize(t), '<a><b>Hello World</b></a>')

    def test51_TemplateLevelIsVisibleBeforeDef(self):
        t = self.parse(
            '<?py greeting="Hello World"?>'
            '<a>'
              '<d py:replace="hello()"/>'
              '<b py:def="hello()">'
                 '<c py:replace="greeting"/>'
              '</b>'
            '</a>'
        )
        assert 'greeting' in t.tmpl_namespace
        assert 'hello' in t.tmpl_namespace
        self.assertEqual(self.serialize(t), '<a><b>Hello World</b></a>')

    def test52_TemplateLevelDefVariableBinding(self):
        # name is freevar in hello
        self.check(
            '<?py name="Bob"?>'
            '<a>'
              '<b py:def="hello()">'
                 '<?python greeting="Hello "+name ?>'
                 '<c py:replace="greeting"/>'
              '</b>'
              '<?py name="Ian"?>'
              '<d py:replace="hello()"/>'
            '</a>',
            '<a><b>Hello Ian</b></a>'
        )

    def test54_DefArguments(self):
        self.check(
            '<a>'
              '<?py name="Bob"?>'
              '<b py:def="hello(name)">'
                 '<?python greeting="Hello "+name ?>'
                 '<c py:replace="greeting"/>'
              '</b>'
              '<?py name="Ian"?>'
              '<d py:replace="hello(name)"/>'
            '</a>',
            '<a><b>Hello Ian</b></a>'
        )

    def test55_DefArgumentTupleUnpacking(self):
        self.check(
            '<a>'
              '<b py:def="hello(a,(b,c))">$a,$b,$c</b>'
              '<?py point=(1,2)?>'
              '<d py:replace="hello(42,point)"/>'
            '</a>',
            '<a><b>42,1,2</b></a>'
        )

    def test56_NestedDef(self):
        self.check(
            '<a>'
              '<b py:def="foo(a)">'
                '<c py:def="bar(b)">$a,$b</c>'
                '${bar(a+1)}'
              '</b>'
              '<d py:replace="foo(1)"/>'
            '</a>',
            '<a><b><c>1,2</c></b></a>'
        )

    def test57_NestedDefWithPI(self):
        self.check(
            '<a>'
              '<b py:def="foo(a)">'
                '<c py:def="bar(b)"><?py x=a+b?>$x</c>'
                '${bar(a+1)}'
              '</b>'
              '<d py:replace="foo(1)"/>'
            '</a>',
            '<a><b><c>3</c></b></a>'
        )

    def test60_LocalDef(self):
        t = self.parse(
            '<?py greeting="Hello World"?>'
            '<a>'
              '<b>'
                 '<c py:def="hello()">'
                    '<d py:replace="greeting"/>'
                 '</c>'
                 '<e py:replace="hello()"/>'
              '</b>'
            '</a>'
        )
        assert 'greeting' in t.tmpl_namespace
        # hello is not template level definition
        assert 'hello' not in t.tmpl_namespace
        self.assertEqual(self.serialize(t),'<a><b><c>Hello World</c></b></a>')

    def test61_LocalDefNotVisibleFromOutside(self):
        t = self.parse(
            '<?py greeting="Hello World"?>'
            '<a>'
              '<b>'
                 '<c py:def="hello()">'
                    '<d py:replace="greeting"/>'
                 '</c>'
              '</b>'
              '<e py:replace="hello()"/>'
            '</a>'
        )
        self.assertRaises(NameError, self.serialize, t)

    def test62_LocalDefNotVisibleBeforeDef(self):
        t = self.parse(
            '<?py greeting="Hello World"?>'
            '<a>'
              '<b>'
                 '<e py:replace="hello()"/>'
                 '<c py:def="hello()">'
                    '<d py:replace="greeting"/>'
                 '</c>'
              '</b>'
            '</a>'
        )
        self.assertRaises(NameError, self.serialize, t)

    def test70_WithBinding(self):
        self.check(
            '''<p>$n $t<b py:with="n=2; t='bar'">$n $t</b>$n $t</p>''',
            '<p>1 foo<b>2 bar</b>1 foo</p>',
            n=1, t='foo'
        )









class TestInheritance(TestBase):

    def test01_Single(self):
        macros = self.parse(
            '<macros>'
              '<a py:def="a(A)" py:content="A"/>'
              '<b py:def="b(B)" py:content="B"/>'
            '</macros>'
        )
        self.check(
            '<html py:extends="macros">'
              '<B py:def="b(B)" py:content="B"/>'
              '<p py:content="a(1)"/>'
              '<p py:content="b(2)"/>'
            '</html>',
            '<html>'
              '<p><a>1</a></p>'
              '<p><B>2</B></p>'
            '</html>',
            tmpl_namespace = {'macros':macros},
        )

    def test05_MoreThanOneChildren(self):
        supermacros = self.parse(
            '<macros>'
              '<a1 py:def="a(A)" py:content="A"/>'
              '<b1 py:def="b(A)" py:content="B"/>'
            '</macros>'
        )
        macros = self.parse(
            '<macros>'
              '<a2 py:def="a(A)" py:content="A"/>'
              '<c2 py:def="c(C)" py:content="C"/>'
            '</macros>'
        )
        self.check(
            '<html py:extends="macros,supermacros">'
              '<b3 py:def="b(B)" py:content="B"/>'
              '<p py:content="a(1)"/>'
              '<p py:content="b(2)"/>'
              '<p py:content="c(2)"/>'
            '</html>',
            '<html>'
              '<p><a1>1</a1></p>'
              '<p><b3>2</b3></p>'
              '<p><c2>2</c2></p>'
            '</html>',
            tmpl_namespace = {'macros':macros, 'supermacros':supermacros},
        )

    def test05_InheritingTemplateWithInheritance(self):
        supermacros = self.parse(
            '<macros>'
              '<a1 py:def="a(A)" py:content="A"/>'
              '<b1 py:def="b(A)" py:content="B"/>'
            '</macros>'
        )
        macros = self.parse(
            '<macros py:extends="macros">'
              '<a2 py:def="a(A)" py:content="A"/>'
              '<c2 py:def="c(C)" py:content="C"/>'
            '</macros>',
            macros = supermacros,
        )
        self.check(
            '<html py:extends="macros">'
              '<b3 py:def="b(B)" py:content="B"/>'
              '<p py:content="a(1)"/>'
              '<p py:content="b(2)"/>'
              '<p py:content="c(2)"/>'
            '</html>',
            '<html>'
              '<p><a2>1</a2></p>'
              '<p><b3>2</b3></p>'
              '<p><c2>2</c2></p>'
            '</html>',
            tmpl_namespace = {'macros':macros}
        )

    def test10_StateReplacesInherited(self):
        macros = self.parse(
            '<macros>'
              '<a2 py:def="a(A)" py:content="A"/>'
            '</macros>',
        )
        tmpl = self.parse(
            '<html py:extends="macros">'
              '<p py:content="a"/>'
            '</html>',
            macros = macros,
        )
        html = self.serialize(tmpl, a=42)
        self.assertEqual(html, '<html><p>42</p></html>')








class TestLayout(TestBase):

    def test20_Layout(self):
        layout = self.parse(
            '<html>'
              '<head><title py:content="title()">TITLE</title></head>'
              '<body><div py:replace="body()">BODY</div></body>'
            '</html>')
        self.check(
            '<doc py:layout="lout">'
              '<x py:def="title()" py:strip="">Test</x>'
              '<y py:def="body()" py:strip="">Body <r>R</r></y>'
            '</doc>',
            '<html>'
              '<head><title>Test</title></head>'
              '<body>Body <r>R</r></body>'
            '</html>',
            lout = layout,
        )

    def test21_Layout(self):
        layout = self.parse(
            '<?py x = 12 ?>'
            '<html>'
              '<head><title py:content="title()">TITLE</title></head>'
              '<body><div py:replace="body()">BODY</div></body>'
            '</html>')
        self.check(
            '<?py title = lambda: u"Test"; x=42 ?>'
            '<doc py:layout="lout">'
              '<y py:def="body()" py:strip="">Body <x>$x</x></y>'
            '</doc>',
            '<html>'
              '<head><title>Test</title></head>'
              '<body>Body <x>42</x></body>'
            '</html>',
            lout = layout,
        )

    def test22_LayoutWithInheritance(self):
        layout = self.parse(
            '<html>'
              '<head><title py:content="title()">TITLE</title></head>'
              '<body><div py:replace="body()">BODY</div></body>'
            '</html>')
        macros = self.parse(
            '<m>'
              '<t py:def="title()" py:strip="">Test</t>'
              '<b py:def="body()" py:strip="">Nobody</b>'
              '<r py:def="f()">R</r>'
            '</m>'
        )
        self.check(
            '<doc py:extends="macros" '
            '     py:layout="lout">'
              '<y py:def="body()" py:strip="">Body <f py:replace="f()"/></y>'
            '</doc>',
            '<html>'
              '<head><title>Test</title></head>'
              '<body>Body <r>R</r></body>'
            '</html>',
            tmpl_namespace = {'macros':macros},
            lout = layout,
        )

    def test23_LayoutIgnoresPageDocElements(self):
        layout = self.parse(
            '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" '
            '          "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">'
            '<html>'
              '<head><title py:content="title()">TITLE</title></head>'
              '<body><div py:replace="body()">BODY</div></body>'
            '</html>')
        self.check(
            '<?xml version="1.0" encoding="iso-8859-2"?>'
            '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Frameset//EN" '
            '          "http://www.w3.org/TR/html4/frameset.dtd">'
            '<?py title = lambda: u"Test" ?>\n'
            '<?py x=12?>\n'
            '<doc py:layout="lout">'
              '<y py:def="body()" py:strip="">Body <r>R</r></y>'
            '</doc>\n',

            '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"'
            ' "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">'
            '<html>'
              '<head><title>Test</title></head>'
              '<body>Body <r>R</r></body>'
            '</html>',
            lout = layout,
        )

    def XXXtest30_StateReplacesInherited(self):
        macros = self.parse(
            '<m>'
              '<a2 py:def="a(A)" py:content="A"/>'
            '</m>',
        )
        layout = self.parse(
            '<l>${body()}</l>'
        )
        tmpl = self.parse(
            '<t py:layout="layout" py:extends="macros">'
              '<p py:def="body()">$a</p>'
            '</t>',
            macros=macros, layout=layout
        )
        html = self.serialize(tmpl, a=42)
        self.assertEqual(html, '<l><p>42</p></l>')

















class FillForm(templates.Domlet):
    def expand(self, state, values=None):
        if values is None:
            if 'values' in state:
                values = state['values']
            else:
                values = self.gdict['values']
        for e,a in self.child_domlet(state):
            if e == const.START_TAG and a[0] == 'input':
                tag, attrs = a
                name = attrs.get('name')
                v = ""
                if name is not None and name in values:
                    v = values[name]
                attrs.set('value', v or "")
            yield e,a

class Filter(templates.Domlet):
    def expand(self, state, tag='r'):
        for e,a in self.child_domlet(state):
            if e in (const.START_TAG, const.END_TAG) and a[0] == tag:
                continue
            yield e,a

class Signature(templates.Domlet):
    def expand(self, state, x=10, y=42):
        yield const.TEXT, '%r,%r' % (x,y)

class Dynattrs(templates.Domlet):
    def expand(self, state):
        attrs = list(self._attrs.expand(state))
        yield const.START_TAG, ('dynattrs', attrs)
        for child in self._children:
            for e in child.expand(state):
                yield e
        yield const.END_TAG, ('dynattrs', attrs)


class RoundCorners(templates.Domlet):
    def expand(self, state):
        d = xh.div(self.child_expander(), _class='round-corners')
        for e in d.expand(state):
            yield e


class TestDOMlets(TestBase):

    def test01_Signature(self):
        self.check(
         '<s my:xml4.tests.test_templates.Signature="" />', '10,42')
        self.check(
         '<s my:xml4.tests.test_templates.Signature="1" />', '1,42')
        self.check(
         '<s my:xml4.tests.test_templates.Signature="1,2" />', '1,2')


    def test10_FillFormWithExplicitValues(self):
        self.check(
            '<?python values=dict(a="A",b=None)?>'
            '<form my:xml4.tests.test_templates.FillForm="values">'
              '<input name="a" value="1" />'
              '<input name="b" value="2" />'
              '<input name="c" value="3" />'
              '<input type="submit" name="submit" />'
            '</form>',
            '<form>'
              '<input name="a" value="A" />'
              '<input name="b" value="" />'
              '<input name="c" value="" />'
              '<input type="submit" name="submit" value="" />'
            '</form>',
        )

    def test11_FillFormWithImplicitValues(self):
        self.check(
            '<?python values=dict(a="A",b=None)?>'
            '<form my:xml4.tests.test_templates.FillForm="">'
              '<input name="a" value="1" />'
              '<input name="b" value="2" />'
              '<input name="c" value="3" />'
              '<input type="submit" name="submit" />'
            '</form>',
            '<form>'
              '<input name="a" value="A" />'
              '<input name="b" value="" />'
              '<input name="c" value="" />'
              '<input type="submit" name="submit" value="" />'
            '</form>',
        )

    def test15_RoundCorners(self):
        self.check(
            '<body>'
                '<div my:xml4.tests.test_templates.RoundCorners="">'
                    '<i>content</i>'
                '</div>'
            '</body>',
            '<body>'
                '<div class="round-corners">'
                    '<i>content</i>'
                '</div>'
            '</body>',
        )

    def test20_Dynattrs(self):
        self.check(
            '<div my:xml4.tests.test_templates.Dynattrs="" '
            '     a="%${\'A\'*2}%" py:attrs="dict(x=\'X\')">'
            '<p>content</p>'
            '</div>',
            '<dynattrs a="%AA%" x="X">'
            '<p>content</p>'
            '</dynattrs>'
        )

    def test30_Filter(self):
        self.check(
            '<p my:xml4.tests.test_templates.Filter="">'
              '<r /><k /><r />'
            '</p>',
            '<p><k></k></p>'
        )

    def test31_FilterWithArgument(self):
        self.check(
            '<p my:xml4.tests.test_templates.Filter="\'k\'">'
              '<r /><k /><r />'
            '</p>',
            '<p><r></r><r></r></p>'
        )

    def test40_Complex(self):
        self.check(
            '<?py\n'
            '  from lib4.iterutils import iterDecorator\n'
            '  items=["a","b","c"]\n'
            '?>'
            '<ul my:xml4.tests.test_templates.Filter="">'
               '<li py:for="i in iterDecorator(items)" '
               '''  class="${('Odd','Even')[i.index % 2]}">'''
                  '${i.index+1}. $i.item '
                  '<r i="$i.index">R</r>'
               '</li>'
            '</ul>',
            '<ul>'
              '<li class="Odd">1. a R</li>'
              '<li class="Even">2. b R</li>'
              '<li class="Odd">3. c R</li>'
            '</ul>',
        )


    def test60_InlineDomletDefinition(self):
        src = (
            '<div>'
               '<p py:def="my:foo.bar()">FOO.BAR</p>'
               '<span my:foo.bar="" />'
            '</div>')
        self.check(src, '<div><p>FOO.BAR</p></div>')

    def test61_InlineDomletDefinitionWithStrip(self):
        src = (
            '<div>'
               '<p py:def="my:foo.bar()" py:strip="">FOO.BAR</p>'
               '<span my:foo.bar="" />'
            '</div>')
        self.check(src, '<div>FOO.BAR</div>')

    def test62_InlineDomletDefinitionWithFor(self):
        src = (
            '<div>'
               '<p py:def="my:foo.bar()" py:for="i in range(3)">FOO.$i</p>'
               '<span my:foo.bar="" />'
            '</div>')
        self.check(src, '<div><p>FOO.0</p><p>FOO.1</p><p>FOO.2</p></div>')

    def test64_InlineDomletDefinitionWithExplicitNamespace(self):
        src = (
            '<div>'
               '<p py:def="{http://foo.com}foo.bar()">FOO.BAR</p>'
               '<span my:foo.bar="" />'
            '</div>')
        self.check(src, '<div><p>FOO.BAR</p></div>')

    def test65_InlineDomletLocals(self):
        src = (
            '<div>'
               '<p py:def="my:foo.bar()">$domlet.value</p>'
               '<span my:foo.bar="baz" />'
            '</div>')
        self.check(src, '<div><p>baz</p></div>')

    def XXXtest66_InlineDomletWithArguments(self):
        # FIXME Domlet getter nema pristup ke State,
        # takze to zatim nefunguje
        src = (
            '<div>'
               '<p py:def="my:foo.bar()">FOOO</p>'
               '${ns.my.foo.bar()}'
            '</div>')
        self.check(src, '<div><p>baz</p></div>')











class TestUnicode(TestBase):

    def test10_UnicodeInPI(self):
        self.check(
            '<?xml version="1.0" encoding="ISO-8859-2"?>\n'
            '<?py x=u"" ?><p>$x</p>',
            u'\n<p></p>'
        )

    def test20_UnicodeInExpression(self):
        self.check(
            '<?xml version="1.0" encoding="ISO-8859-2"?>\n'
            '<p>${u""}</p>',
            u'\n<p></p>'
        )

    def test30_UnicodeInPyContent(self):
        self.check(
            '<?xml version="1.0" encoding="ISO-8859-2"?>\n'
            '<p py:content="u\'\'">zzzzz</p>',
            u'\n<p></p>'
        )




def top_frame():
    _,_,tb = sys.exc_info()
    while tb.tb_next:
        tb = tb.tb_next
    return tb.tb_frame

class TestErrorReporting(TestBase):

    filename = os.path.join(THIS_DIR, 'error.wt')

    def test10_ErrorInPI(self):
        tmpl = self.parseFile(self.filename)
        try:
            self.serialize(tmpl, case='pi')
        except ZeroDivisionError:
            f = top_frame()
            self.assertEqual(f.f_lineno, 6)
            self.assertEqual(f.f_code.co_filename, self.filename)
        else:
            self.fail('Exception was not raised')

    def test20_SyntaxErrorInSubstitution(self):
        try:
            self.parse('''<p>

            ${a b}

            </p>''')
        except SyntaxError:
            f = top_frame()
            self.assertEqual(f.f_lineno, 3)
            self.assertEqual(f.f_code.co_filename, '<xml source>')
            self.assertEqual(f.f_code.co_name, '${a b}')
        else:
            self.fail('SyntaxError was not raised')

    def test21_ErrorInSubstitution(self):
        tmpl = self.parseFile(self.filename)
        try:
            self.serialize(tmpl, case='subst')
        except NameError:
            f = top_frame()
            self.assertEqual(f.f_lineno, 9)
            self.assertEqual(f.f_code.co_filename, self.filename)
        else:
            self.fail('Exception was not raised')

    def test30_SyntaxErrorInDynAttr(self):
        try:
            self.parse('''<p>

            <div class="xx${a b}yy" />

            </p>''')
        except SyntaxError:
            f = top_frame()
            self.assertEqual(f.f_lineno, 3)
            self.assertEqual(f.f_code.co_filename, '<xml source>')
            self.assertEqual(f.f_code.co_name, 'class="xx${a b}yy"')
        else:
            self.fail('SyntaxError was not raised')

    def test31_ErrorInDynAttr(self):
        tmpl = self.parseFile(self.filename)
        try:
            self.serialize(tmpl, case='attr')
        except NameError:
            f = top_frame()
            self.assertEqual(f.f_lineno, 12)
            self.assertEqual(f.f_code.co_filename, self.filename)
            self.assertEqual(f.f_code.co_name, 'class="${noname}"')
        else:
            self.fail('Exception was not raised')

    def test40_SyntaxErrorInContent(self):
        try:
            self.parse('''<p>

            <div py:content="a b" />

            </p>''')
        except SyntaxError:
            f = top_frame()
            self.assertEqual(f.f_lineno, 3)
            self.assertEqual(f.f_code.co_filename, '<xml source>')
            self.assertEqual(f.f_code.co_name, 'py:content')
        else:
            self.fail('SyntaxError was not raised')

    def test41_ErrorInContent(self):
        tmpl = self.parseFile(self.filename)
        try:
            self.serialize(tmpl, case='content')
        except NameError:
            f = top_frame()
            self.assertEqual(f.f_lineno, 17)
            self.assertEqual(f.f_code.co_filename, self.filename)
            self.assertEqual(f.f_code.co_name, 'py:content')
        else:
            self.fail('Exception was not raised')

    def test42_SyntaxErrorInStrip(self):
        try:
            self.parse('''<p>

            <div py:strip="a b" />

            </p>''')
        except SyntaxError:
            f = top_frame()
            self.assertEqual(f.f_lineno, 3)
            self.assertEqual(f.f_code.co_filename, '<xml source>')
            self.assertEqual(f.f_code.co_name, 'py:strip')
        else:
            self.fail('SyntaxError was not raised')

    def test43_ErrorInStrip(self):
        tmpl = self.parseFile(self.filename)
        try:
            self.serialize(tmpl, case='strip')
        except NameError:
            f = top_frame()
            self.assertEqual(f.f_lineno, 26)
            self.assertEqual(f.f_code.co_filename, self.filename)
            self.assertEqual(f.f_code.co_name, 'py:strip')
        else:
            self.fail('Exception was not raised')

    def test44_SyntaxErrorInPyAttrs(self):
        try:
            self.parse('''<p>

            <div py:attrs="a b" />

            </p>''')
        except SyntaxError:
            f = top_frame()
            self.assertEqual(f.f_lineno, 3)
            self.assertEqual(f.f_code.co_filename, '<xml source>')
            self.assertEqual(f.f_code.co_name, 'py:attrs')
        else:
            self.fail('SyntaxError was not raised')

    def test45_ErrorInPyAttrs(self):
        tmpl = self.parseFile(self.filename)
        try:
            self.serialize(tmpl, case='attrs')
        except NameError:
            f = top_frame()
            self.assertEqual(f.f_lineno, 32)
            self.assertEqual(f.f_code.co_filename, self.filename)
            self.assertEqual(f.f_code.co_name, 'py:attrs')
        else:
            self.fail('Exception was not raised')

    def test46_SyntaxErrorInReplace(self):
        try:
            self.parse('''<p>

            <div py:replace="a b" />

            </p>''')
        except SyntaxError:
            f = top_frame()
            self.assertEqual(f.f_lineno, 3)
            self.assertEqual(f.f_code.co_filename, '<xml source>')
            self.assertEqual(f.f_code.co_name, 'py:content')
        else:
            self.fail('SyntaxError was not raised')

    def test47_ErrorInReplace(self):
        tmpl = self.parseFile(self.filename)
        try:
            self.serialize(tmpl, case='replace')
        except NameError:
            f = top_frame()
            self.assertEqual(f.f_lineno, 29)
            self.assertEqual(f.f_code.co_filename, self.filename)
            self.assertEqual(f.f_code.co_name, 'py:content')
        else:
            self.fail('Exception was not raised')

    def test50_SyntaxErrorInFor(self):

        for xml in [
            '''<p>\n\n<div py:for="axb" /></p>''',
            '''<p>\n\n<div py:for="a in i j" /></p>''',
            '''<p>\n\n<div py:for="a b in i" /></p>''',
        ]:
            try:
                self.parse(xml)
            except SyntaxError:
                f = top_frame()
                self.assertEqual(f.f_lineno, 3)
                self.assertEqual(f.f_code.co_filename, '<xml source>')
                self.assertEqual(f.f_code.co_name, 'py:for')
            else:
                self.fail('SyntaxError was not raised')

    def test51_ErrorInFor(self):
        tmpl = self.parseFile(self.filename)
        try:
            self.serialize(tmpl, case='for')
        except NameError:
            f = top_frame()
            self.assertEqual(f.f_lineno, 20)
            self.assertEqual(f.f_code.co_filename, self.filename)
            self.assertEqual(f.f_code.co_name, 'py:for')
        else:
            self.fail('Exception was not raised')

    def test52_SyntaxErrorInIf(self):
        try:
            self.parse('''<p>

            <div py:if="a b" />

            </p>''')
        except SyntaxError:
            f = top_frame()
            self.assertEqual(f.f_lineno, 3)
            self.assertEqual(f.f_code.co_filename, '<xml source>')
            self.assertEqual(f.f_code.co_name, 'py:if')
        else:
            self.fail('SyntaxError was not raised')

    def test53_ErrorInIf(self):
        tmpl = self.parseFile(self.filename)
        try:
            self.serialize(tmpl, case='if')
        except NameError:
            f = top_frame()
            self.assertEqual(f.f_lineno, 23)
            self.assertEqual(f.f_code.co_filename, self.filename)
            self.assertEqual(f.f_code.co_name, 'py:if')
        else:
            self.fail('Exception was not raised')

    def test60_SyntaxErrorInDef(self):
        for xml in [
            '''<p>\n\n<div py:def="a" /></p>'''
            '''<p>\n\n<div py:def="a(" /></p>'''
            '''<p>\n\n<div py:def="a())" /></p>'''
            '''<p>\n\n<div py:def="a(b c)" /></p>'''
        ]:
            try:
                self.parse(xml)
            except SyntaxError:
                f = top_frame()
                self.assertEqual(f.f_lineno, 3)
                self.assertEqual(f.f_code.co_filename, '<xml source>')
                self.assertEqual(f.f_code.co_name, 'py:def=a')
            else:
                self.fail('SyntaxError was not raised')

    def test61_ErrorInDefInit(self):
        tmpl = self.parseFile(self.filename)
        try:
            self.serialize(tmpl, case='def_init')
        except TypeError, e:
            # XXX 1 argument, not 2
            self.assertEqual(
                str(e),
                'py:def=x(a)() takes exactly 2 arguments (3 given)'
            )
        else:
            self.fail('TypeError was not raised')

    def test62_ErrorInDefExpand(self):
        # XXX what info in traceback?
        tmpl = self.parseFile(self.filename)
        try:
            self.serialize(tmpl, case='def_expand')
        except NameError, e:
            f = top_frame()
            self.assertEqual(f.f_lineno, 42)
            self.assertEqual(f.f_code.co_filename, self.filename)
            self.assertEqual(f.f_code.co_name, '${b}')
        else:
            self.fail('NameError was not raised')

    def test63_SyntaxErrorInWith(self):
        try:
            self.parse('''<p py:with="a:3">$a</p>''')
        except SyntaxError:
            f = top_frame()
            self.assertEqual(f.f_lineno, 1)
            self.assertEqual(f.f_code.co_filename, '<xml source>')
            self.assertEqual(f.f_code.co_name, 'py:with=a:3')
        else:
            self.fail('SyntaxError was not raised')

    def test64_ErrorInWith(self):
        tmpl = self.parseFile(self.filename)
        try:
            self.serialize(tmpl, case='with')
        except NameError, e:
            f = top_frame()
            self.assertEqual(f.f_lineno, 47)
            self.assertEqual(f.f_code.co_filename, self.filename)
            self.assertEqual(f.f_code.co_name, 'py:with=a=novar')
        else:
            self.fail('NameError was not raised')








class TestDeferred(TestBase):

    def test01(self):
        s = {}
        self.check(
            '<p><d py:content="d"/><e py:content="s.setdefault(1,\'A\')"/></p>',
            '<p><d>A</d><e>A</e></p>',
            s = s, d = templates.Deferred(lambda: unicode(s[1])),
        )
        self.check(
            '<p><d>$d</d><e>${s.setdefault(2,\'B\')}</e></p>',
            '<p><d>B</d><e>B</e></p>',
            s = s, d = templates.Deferred(lambda: unicode(s[2])),
        )






def suite():
    return TestSuite((
        makeSuite(TestTemplateParsing),
        makeSuite(TestProcessingInstructions),
        makeSuite(TestInterpolation),
        makeSuite(TestControllSturctures),
        makeSuite(TestInheritance),
        makeSuite(TestLayout),
        makeSuite(TestDOMlets),
        makeSuite(TestUnicode),
        makeSuite(TestErrorReporting),
        makeSuite(TestDeferred),
    ))

if __name__ == '__main__':
    main(defaultTest='suite')
